@@ -12,7 +12,7 @@
 #include <functional>
 #include <algorithm>
 
-class MQueue_eq : public unary_function<MQueue*, bool> {
+class MQueue_eq  {
     string name;
   public:
     explicit MQueue_eq(const string& name_) : name(name_) {}
@@ -21,7 +21,7 @@ class MQueue_eq : public unary_function<MQueue*, bool> {
 
 set<MQueue*> MQueue::knownQueues;
 
-string MQueue::popMessage() throw (MQueueException)
+string MQueue::popMessage()
 {
     if (q.empty()) {
         throw MQueueException("attempt to pop from an empty queue");
