@@ -32,7 +32,7 @@
   base system that the k-th digit after point has the weight of k/(2k+1)
   of the previous digit, [1] means pi can be written as [2.222222...]
   in such a base system.
-  
+
     Base:   [1/3, 2/5, 3/7, ... , k/(2k+1), ...]
     Number: [2 . 222222222 ...]
 
@@ -40,7 +40,7 @@
     Number: [3 . 141592653 ...]
 
   Now, calculating pi is effectively a base conversion problem.
-  
+
   The actual calculation process is explained in detail at:
    http://membres.lycos.fr/bgourevitch/mathematiciens/goutte/goutte.html
 
@@ -148,9 +148,9 @@ ScmObj Spigot_calculate_e(int digits)
 
 /*
  * Module initialization function.
- * This is called when spigot.so is dynamically loaded into gosh.
+ * This is called when math--spigot.so is dynamically loaded into gosh.
  */
-ScmObj Scm_Init_spigot(void)
+void Scm_Init_math__spigot(void)
 {
     ScmModule *mod;
 
@@ -160,7 +160,6 @@ ScmObj Scm_Init_spigot(void)
     /* Create "spigot" module if it doesn't exist yet. */
     mod = SCM_MODULE(SCM_FIND_MODULE("spigot", TRUE));
 
-    /* Register stub-generated procedures inside "spigot" module */
-    Scm_Init_spigotlib(mod);
+    /* Initialize spigotlib.scm */
+    Scm_Init_spigotlib();
 }
-
