@@ -2,5 +2,5 @@
 
 (define (main argc)
   (print "Welcome to my standalone REPL application!")
-  (read-eval-print-loop #f #f #f (^[] (display "sample-repl> ") (flush)))
+  (read-eval-print-loop #f #f #f (cut display "sample-repl> "))
   0)
