@@ -3,7 +3,7 @@
 (use os.windows)
 
 (define (main args)
-  (guard (e [else (sys-message-box #f (format "ERROR: ~a" (~ e'message))
+  (guard (e [else (sys-message-box #f (format "ERROR: ~a" (condition-message e))
                                    (sys-basename *program-name*)
                                    (logior MB_OK MB_ICONERROR))])
     (messagebox-sample (sys-basename (car args))))
