@@ -1,7 +1,7 @@
 ;;;
 ;;; charconv - character code conversion module
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,8 +32,8 @@
 ;;;
 
 (define-module gauche.charconv
-  (use srfi-1)
-  (use srfi-13)
+  (use scheme.list)
+  (use srfi.13)
   (use gauche.sequence)
   (export open-input-conversion-port
           open-output-conversion-port
@@ -52,4 +52,3 @@
 (select-module gauche.charconv)
 
 (dynamic-load "gauche--charconv")
-
