@@ -74,20 +74,16 @@
 
 (define (test-input file from to :key (guesser #f) illegal-output)
   (let* ((realfrom (or guesser from))
-         (infostr  (format #f "~a.~a (~a) => ~a" file from realfrom to))
+         (infostr  (format #f "~a.~a (~a) => ~a [~a]" file from realfrom to
+                           illegal-output))
          (fromfile (format #f "~a.~a" file from))
          (tofile   (format #f "~a.~a" file to)))
     (if (ces-conversion-supported? from to)
-      (if (supported-character-encoding? to)
-        (test infostr
-              (file->string tofile)
-              (lambda () (file->string-conv/in fromfile realfrom
-                                               :illegal-output illegal-output)))
-        (test infostr
-              (file->string tofile)
-              (lambda () (file->string-conv/in fromfile realfrom
-                                               :to-code to
-                                               :illegal-output illegal-output))))
+      (test infostr
+            (file->string tofile)
+            (lambda () (file->string-conv/in fromfile realfrom
+                                             :to-code to
+                                             :illegal-output illegal-output)))
       (test infostr "(not supported)"
             (lambda () "(not supported)")))
     ))
@@ -124,12 +120,12 @@
           "data/lat1"
           '("ISO8859-1")
           '("ASCII" "ISO8859-1" "EUCJP" "SJIS" "ISO2022JP" "UTF-8"
-            "UTF-16" "UTF-16BE" "UTF-16LE" 
+            "UTF-16" "UTF-16BE" "UTF-16LE"
             "UTF-32" "UTF-32BE" "UTF-32LE"))
 (map-test (lambda (file from to)
             (test-input file from to :illegal-output 'replace))
           "data/lat1x"
-          '("ISO8859-1" "EUCJP" "UTF-8" "SJIS" 
+          '("ISO8859-1" "EUCJP" "UTF-8" "SJIS"
             "UTF-16" "UTF-16BE" "UTF-16LE"
             "UTF-32" "UTF-32BE" "UTF-32LE")
           '("ISO8859-1"))
@@ -193,12 +189,7 @@
 (define (test-output/chunk20 file from to)
   (test-output "chunk20" (^p (read-block 20 p)) display file from to))
 
-(define internal-enc
-  (case (gauche-character-encoding)
-    ((euc-jp) '("EUCJP"))
-    ((sjis)   '("SJIS"))
-    ((utf-8)  '("UTF-8"))
-    (else '())))
+(define internal-enc '("UTF-8"))
 
 (map-test test-output/byte "data/jp1"
           '("EUCJP" "UTF-8" "SJIS" "ISO2022JP" "UTF-16" "UTF-16BE" "UTF-16LE")
@@ -278,6 +269,24 @@
        (with-output-to-file "_NOT_EXIST_" (lambda () (error "PANIC!"))
                             :if-does-not-exist #f :encoding "UTF-8"))
 
+;;--------------------------------------------------------------------
+(test-section "utf8 optional BOM handling")
+
+(let ()
+  (define (file->lines path encoding)
+    (call-with-input-file #"~(sys-dirname (current-load-path))/~path"
+      port->string-list
+      :encoding encoding))
+
+  (test* "utf8 optional BOM handling (utf8bom)"
+         '("abc" "\xfeff;def")
+         (file->lines "data/utf8bom.txt" 'utf-8-bom))
+
+  (test* "utf8 optional BOM handling (utf8)"
+         '("\xfeff;abc" "\xfeff;def")
+         (file->lines "data/utf8bom.txt" 'utf8))
+  )
+
 ;;--------------------------------------------------------------------
 (test-section "code guessing")
 
@@ -412,12 +421,7 @@
                 (call-with-input-conversion p port->string
                                             :encoding code))))))
   (if (ces-conversion-supported? (gauche-character-encoding) code)
-    (let ((s (call-with-input-file (format "~a.~a" file
-                                           (case (gauche-character-encoding)
-                                             ((euc-jp) "EUCJP")
-                                             ((sjis)   "SJIS")
-                                             ((utf-8)  "UTF-8")
-                                             ((none)   "UTF-8")))
+    (let ((s (call-with-input-file (format "~a.~a" file "UTF-8")
                port->string)))
       (test* #"call-with-output-conversion (~(gauche-character-encoding) -> ~code)"
              (string-complete->incomplete
@@ -444,13 +448,10 @@
 (test "ucs <-> char" #xa1
       (lambda () (char->ucs (ucs->char #xa1))))
 
-(case (gauche-character-encoding)
-  ((euc-jp sjis utf-8)
-   (test "ucs <-> char" #x0391
-         (lambda () (char->ucs (ucs->char #x0391))))
-   (test "ucs <-> char" #x3042
-         (lambda () (char->ucs (ucs->char #x3042))))
-   ))
+(test "ucs <-> char" #x0391
+      (lambda () (char->ucs (ucs->char #x0391))))
+(test "ucs <-> char" #x3042
+      (lambda () (char->ucs (ucs->char #x3042))))
 
 (test "ucs char syntax" #\a
       (lambda () (read-from-string "#\\u0061")))
@@ -458,37 +459,28 @@
       (lambda () (read-from-string "#\\u00000062")))
 (test "ucs char syntax" #xa1
       (lambda () (char->ucs (read-from-string "#\\u00a1"))))
-(case (gauche-character-encoding)
-  ((euc-jp sjis utf-8)
-   (test "usc char syntax" #x0391
-         (lambda () (char->ucs (read-from-string "#\\u0391"))))
-   (test "ucs char syntax" #x3042
-         (lambda () (char->ucs (read-from-string "#\\u3042"))))
-   ))
+
+(test "usc char syntax" #x0391
+      (lambda () (char->ucs (read-from-string "#\\u0391"))))
+(test "ucs char syntax" #x3042
+      (lambda () (char->ucs (read-from-string "#\\u3042"))))
 
 (test "ucs string syntax" "abcde"
       (lambda () (read-from-string "\"\\u0061bcde\"")))
 (test "ucs string syntax" "abcde"
       (lambda () (read-from-string "\"\\U00000061bcde\"")))
-(case (gauche-character-encoding)
-  ((euc-jp sjis utf-8)
-   (test "usc string syntax" #x0391
-         (lambda () (char->ucs (string-ref (read-from-string "\"\\u03911\"") 0))))
-   (test "ucs string syntax" #x3042
-         (lambda () (char->ucs (string-ref (read-from-string "\"\\u30421\"") 0))))
-   ))
+
+(test "usc string syntax" #x0391
+      (lambda () (char->ucs (string-ref (read-from-string "\"\\u03911\"") 0))))
+(test "ucs string syntax" #x3042
+      (lambda () (char->ucs (string-ref (read-from-string "\"\\u30421\"") 0))))
 
 ;;--------------------------------------------------------------------
 (test-section "coding-aware port")
 
 (define *target-string*
-  (case (gauche-character-encoding)
-    ((euc-jp sjis utf-8)
-     (read-from-string
-      "\"\\u3053\\u3093\\u306b\\u3061\\u306f\\u3001\\u4e16\\u754c\""))
-    ((none)
-     (read-from-string
-      "\"\\xe3\\x81\\x93\\xe3\\x82\\x93\\xe3\\x81\\xab\\xe3\\x81\\xa1\\xe3\\x81\\xaf\\xe3\\x80\\x81\\xe4\\xb8\\x96\\xe7\\x95\\x8c\""))))
+  (read-from-string
+   "\"\\u3053\\u3093\\u306b\\u3061\\u306f\\u3001\\u4e16\\u754c\""))
 
 (define (test-coding-aware-port num encoding)
   (when (ces-conversion-supported? (gauche-character-encoding) encoding)
@@ -511,4 +503,22 @@
     (dolist (enc '("EUCJP" "SJIS" "UTF-8"))
       (test-coding-aware-port num enc))))
 
+;;--------------------------------------------------------------------
+(test-section "default-file-encoding")
+
+(let1 srcdir (sys-dirname (current-load-path))
+  (test* "default-file-encoding (input)"
+         (call-with-input-file #"~|srcdir|/data/lat1.UTF-8" port->string)
+         (parameterize ([default-file-encoding 'latin1])
+           (call-with-input-file #"~|srcdir|/data/lat1.ISO8859-1" port->string)))
+  (sys-unlink "testdata.o")
+  (test* "default-file-encoding (output)"
+         (call-with-input-file #"~|srcdir|/data/lat1.ISO8859-1" port->byte-string)
+         (let1 data (call-with-input-file #"~|srcdir|/data/lat1.UTF-8" port->string)
+           (parameterize ([default-file-encoding 'latin1])
+             (call-with-output-file "testdata.o"
+               (^p (display data p))))
+           (call-with-input-file "testdata.o" port->byte-string)))
+  (sys-unlink "testdata.o"))
+
 (test-end)
