@@ -1,7 +1,7 @@
 ;;;
 ;;; gdbm - gdbm interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -174,7 +174,8 @@
 (inline-stub
  (declcode
   (.include <gdbm.h>)
-  (.include <stdlib.h>))
+  (.include <stdlib.h>)
+  (.include <gauche/priv/configP.h>))
 
  (define-ctype ScmGdbmFile::(.struct
                              (SCM_HEADER :: ""
@@ -187,8 +188,7 @@
    [printer
     (Scm_Printf port "#<gdbm-file %S>" (-> (SCM_GDBM_FILE obj) name))])
 
- (define-cfn gdbm_finalize (obj data::void*) ::void :static
-   (cast void data)                     ; suppress unused var warning
+ (define-cfn gdbm_finalize (obj _::void*) ::void :static
    (let* ((g :: ScmGdbmFile* (SCM_GDBM_FILE obj)))
      (when (-> g dbf)
        (gdbm_close (-> g dbf))
@@ -325,5 +325,3 @@
  (define-enum GDBM_CENTFREE)
  (define-enum GDBM_COALESCEBLKS)
  )
-
-
