@@ -1,7 +1,7 @@
 ;;;
 ;;; ndbm - ndbm interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 (define-module dbm.ndbm
   (extend dbm)
-  (use srfi-1)
+  (use scheme.list)
   (use util.match)
   (export <ndbm>
           ;; low level funcions
@@ -169,6 +169,7 @@
 (inline-stub
  (declcode
   (.include <fcntl.h>)
+  (.include <gauche/priv/configP.h>)
   (.include "dbmconf.h")
   (.cond
    ["HAVE_NDBM_H" (.include <ndbm.h>)]
@@ -187,8 +188,7 @@
    [printer
     (Scm_Printf port "#<ndbm-file %S>" (-> (SCM_NDBM_FILE obj) name))])
 
- (define-cfn ndbm_finalize (obj data::void*) ::void :static
-   (cast void data)                     ; suppress unused var warning
+ (define-cfn ndbm_finalize (obj _::void*) ::void :static
    (let* ((n::ScmNdbmFile* (SCM_NDBM_FILE obj)))
      (when (-> n dbf)
        (dbm_close (-> n dbf))
@@ -294,4 +294,3 @@
  (define-enum O_CREAT)
  (define-enum O_TRUNC)
  )
-
