@@ -1,7 +1,7 @@
 ;;;
 ;;; odbm - original dbm interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 (define-module dbm.odbm
   (extend dbm)
-  (use srfi-1)
+  (use scheme.list)
   (export <odbm>
           ;; low level funcions
           odbm-init    odbm-close    odbm-closed?
@@ -162,11 +162,16 @@
 
 (inline-stub
  (declcode
+  (.include <gauche/priv/configP.h>)
   (.include "dbmconf.h")
   (.cond
    ["HAVE_DBM_H" (.include <dbm.h>)]
    ["HAVE_GDBM_SLASH_DBM_H" (.include <gdbm/dbm.h>)]
-   ["HAVE_GDBM_MINUS_DBM_H" (.include <gdbm-dbm.h>)]))
+   ["HAVE_GDBM_MINUS_DBM_H" (.include <gdbm-dbm.h>)])
+
+  ;; Avoid conflicting with C++ keyword
+  (.define delete_ delete)
+  )
 
  ;; data conversion macros
  (define-cise-stmt TO_DATUM
@@ -219,7 +224,7 @@
 
  (define-cproc odbm-delete (key::<string>) ::<int>
    (let* ([dkey::datum])
-     (CHECK_ODBM) (TO_DATUM dkey key) (return (delete dkey))))
+     (CHECK_ODBM) (TO_DATUM dkey key) (return (delete_ dkey))))
 
  (define-cproc odbm-firstkey ()
    (let* ([dkey::datum])
@@ -230,4 +235,3 @@
      (CHECK_ODBM) (TO_DATUM dkey key) (set! dnkey (nextkey dkey))
      (FROM_DATUM SCM_RESULT dnkey)))
  )
-
