@@ -88,8 +88,8 @@
 
 (inline-stub
  (declcode
-  (.include <gauche/class.h>)
-  (.include "md5.h")
+  (.include <gauche/priv/configP.h>
+            "md5.h")
 
   (.define LIBGAUCHE_EXT_BODY)
   (.include <gauche/extern.h>)      ; fix SCM_EXTERN in SCM_CLASS_DECL
@@ -128,4 +128,3 @@
                              (logior SCM_STRING_INCOMPLETE
                                      SCM_STRING_COPYING)))))
  )
-
