@@ -3,11 +3,13 @@
 ;;
 
 (use gauche.test)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (use gauche.uvector)
 
-(test-start "file.util")
+(test-start "file.*")
+
+(test-section "file.util")
 (use file.util)
 (cond-expand
  [gauche.os.windows
@@ -17,7 +19,7 @@
 
 ;; shorthand of normalizing pathname.  this doesn't do anything on
 ;; unix, but on Windows the separator in PATHNAME is replaced.
-;; NB: we remove the drive letter from DOS path to accomodate
+;; NB: we remove the drive letter from DOS path to accommodate
 ;; different installations.
 (define (n pathname . pathnames)
   (define (normalize s)
@@ -27,6 +29,9 @@
     (normalize pathname)
     (map normalize (cons pathname pathnames))))
 
+(define (na pathname . pathnames)       ;make it absolute
+  (sys-normalize-pathname (apply n pathname pathnames) :absolute #t))
+
 ;; mingw doesn't have fully compatible permissions as unix.
 ;; this procedure compensates it.
 (define (P perm)
@@ -49,6 +54,7 @@
 (test* "build-path" (n "foo") (build-path "" "foo"))
 (test* "build-path" (n "foo/bar") (build-path "" "foo" "" "bar"))
 (test* "build-path" (n "foo") (build-path "" "foo" ""))
+(test* "build-path" (n "./a/./b") (build-path #f "a" #f "b"))
 
 (test* "decompose-path" '("/a/b/c" "d" "e")
        (values->list (decompose-path "/a/b/c/d.e")))
@@ -117,19 +123,6 @@
               [cur** (current-directory)])
          (list root root* (string=? cur cur*) (string=? cur* cur**))))
 
-;; rm -rf
-(define (cmd-rmrf dir)
-  ;; shorthand of normalizing pathname.  this doesn't do anything on
-  ;; unix, but on Windows the separator in PATHNAME is replaced.
-  (define (n pathname) (sys-normalize-pathname pathname))
-
-  (cond-expand
-   [gauche.os.windows
-    (sys-system #"rmdir /q /s ~(n dir) > NUL 2>&1")
-    (sys-system #"del /q ~(n dir) > NUL 2>&1")]
-   [else
-    (sys-system #"rm -rf ~dir > /dev/null")]))
-
 ;; prepare test data set
 (define *test-tree*
   `(test.out
@@ -159,7 +152,7 @@
               test12.o))
      )))
 
-(cmd-rmrf "test.out")
+(test-remove-files "test.out")
 
 (test* "create-directory-tree" #t
        (begin (create-directory-tree "." *test-tree*)
@@ -338,7 +331,7 @@
                                                     files))))))
        )
 
-(cmd-rmrf "test.out")
+(test-remove-files "test.out")
 
 ;;=====================================================================
 (test-section "file utils")
@@ -626,6 +619,40 @@
    [else])
   )
 
+;;
+;; temporary file/dir
+;;
+
+(test* "call-with-temporary-file" '(#t #t #f)
+       (unwind-protect
+           (begin
+             (remove-files '("test.o"))
+             (make-directory* "test.o")
+             (receive (exists name)
+                 (call-with-temporary-file
+                  (^[oport name] (values (file-exists? name) name))
+                  :directory "test.o"
+                  :prefix "ooo")
+               (list exists
+                     (boolean (#/test.o[\/\\]ooo\w+$/ name))
+                     (file-exists? name))))
+         (remove-files '("test.o"))))
+
+(test* "call-with-temporary-directory" '(#t #t #f)
+       (unwind-protect
+           (begin
+             (remove-files '("test.o"))
+             (make-directory* "test.o")
+             (receive (exists name)
+                 (call-with-temporary-directory
+                  (^[name] (values (file-is-directory? name) name))
+                  :directory "test.o"
+                  :prefix "ooo")
+               (list exists
+                     (boolean (#/test.o[\/\\]ooo\w+$/ name))
+                     (file-exists? name))))
+         (remove-files '("test.o"))))
+
 ;;
 ;; file->*
 ;;
@@ -736,24 +763,24 @@
 
   (unwind-protect
       (begin
-        (test* "find-file-in-paths" (n "test.out/bin/foo")
+        (test* "find-file-in-paths" (na "test.out/bin/foo")
                (find-file-in-paths "foo"
                                    :paths paths
                                    :pred file-exists?))
-        (test* "find-file-in-paths" (n "test.out/usr/bin/foo")
+        (test* "find-file-in-paths" (na "test.out/usr/bin/foo")
                (find-file-in-paths "foo"
                                    :paths (cdr paths)
                                    :pred file-exists?))
-        (test* "find-file-in-paths" (n "test.out/usr/sbin/ban")
+        (test* "find-file-in-paths" (na "test.out/usr/sbin/ban")
                (find-file-in-paths "ban"
                                    :paths paths
                                    :pred file-exists?))
-        (test* "find-file-in-paths" (n "test.out/usr/bin/ban.exe")
+        (test* "find-file-in-paths" (na "test.out/usr/bin/ban.exe")
                (find-file-in-paths "ban"
                                    :paths paths
                                    :pred file-exists?
                                    :extensions '("exe")))
-        (test* "find-file-in-paths" (n "test.out/bin/ban.com")
+        (test* "find-file-in-paths" (na "test.out/bin/ban.com")
                (find-file-in-paths "ban"
                                    :paths paths
                                    :pred file-exists?
@@ -762,4 +789,23 @@
     (cleanup))
   )
 
+;;=====================================================================
+(test-section "file.event")
+(use file.event)
+(test-module 'file.event)
+(test-module 'file.event.common)        ;loaded via file.event
+
+(define-module file-event-generic-test
+  (use gauche.test)
+  (use file.event.generic)
+  (test-module 'file.event.generic))
+
+(define-module file-event-specific-test
+  (use gauche.test)
+  ;; Those submodules are already loaded via file.event
+  (cond-expand
+   [gauche.sys.inotify (test-module 'file.event.inotify)]
+   [gauche.sys.kqueue (test-module 'file.event.kqueue)]
+   [else]))
+
 (test-end)
