@@ -4,8 +4,8 @@
 (test-start "collections and sequences")
 
 (use data.queue)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (use gauche.collection)
 (use gauche.sequence)
 
@@ -27,6 +27,14 @@
     (proc (^[] (null? p))
           (^[] (pop! p)))))
 
+(define-method call-with-reverse-iterator ((seq <string-seq>) proc
+                                           :key (start #f) (end #f)
+                                           :allow-other-keys)
+  (let* ([ss (slot-ref seq 'strings)]
+         [p  (reverse (subseq ss (or start 0) (or end (length ss))))])
+    (proc (^[] (null? p))
+          (^[] (pop! p)))))
+
 (define-method call-with-builder ((seq <string-seq-meta>) proc . _)
   (let1 q (make-queue)
     (proc (^[item] (enqueue! q (x->string item)))
@@ -40,6 +48,9 @@
   (lambda (o i v)
     (set! (ref (slot-ref o 'strings) i) (x->string v))))
 
+(define-method object-equal? ((a <string-seq>) (b <string-seq>))
+  (equal? (slot-ref a 'strings) (slot-ref b 'strings)))
+
 (define (sseq . elements)
   (make <string-seq> :strings (map x->string elements)))
 
@@ -389,6 +400,10 @@
                                      (#\r #\r) (#\c) (#\d))
        (group-collection "abracadabra"))
 
+(test* "group-collection->alist" '((3 "a" "d") (2 "b") (1 "c" "e"))
+       (group-collection->alist '((3 "a") (2 "b") (1 "c") (3 "d") (1 "e"))
+                                :key car :value cadr))
+
 (test-section "sequence operations")
 
 (test* "ref (list)" 3     (ref '(1 2 3 4 5) 2))
@@ -570,6 +585,25 @@
 (test* "fold-right (different lengths)" '(#\a 0 u #\b 1 v)
        (fold-right cons* '() "ab" '#(0 1 2 3 4) '(u v w x y z)))
 
+;; reverse iterator
+(define (generic-reverse seq start end)
+  ($ call-with-builder (class-of seq)
+     (^[add! get]
+       ($ call-with-reverse-iterator seq
+          (^[end? next]
+            (let loop ()
+              (if (end?)
+                (get)
+                (begin (add! (next)) (loop)))))
+          :start start :end end))))
+
+(test* "reverse iterator (vector)" '#(d c)
+       (generic-reverse '#(a b c d e) 2 4))
+(test* "reverse iterator (string)" "dc"
+       (generic-reverse "abcde" 2 4))
+(test* "reverse iterator (custom)" (sseq 'd 'c)
+       (generic-reverse (sseq 'a 'b 'c 'd 'e) 2 4))
+
 (test* "group-sequence" '((1 1 1) (2) (3) (4 4) (2 2) (3) (1 1) (3))
        (group-sequence '(1 1 1 2 3 4 4 2 2 3 1 1 3)))
 (test* "group-sequence w/key" '((1 1 1) (2) (3) (4 4 2 2) (3 1 1 3))
@@ -704,11 +738,11 @@
                         (null? a)
                         (eq? a data))  ; check if head part is shared
                     (list a b))))))
-    
+
   (dolist [datum data]
     (apply tester #f datum)
     (apply tester #t datum)))
-  
+
 (test* "common-prefix" '(a b c)
        (common-prefix-to <list> '(a b c d e) '(a b c e d)))
 (test* "common-prefix" '(a b c)
@@ -724,6 +758,11 @@
 (test* "common-prefix" "ABC"
        (common-prefix "ABCE" "abcd" :test char-ci=?))
 
+(test* "inverse-permuter"  '#(4 3 0 2 1)
+       (inverse-permuter '#(2 4 3 1 0)))
+(test* "inverse-permuter" '(4 3 0 2 1)
+       (inverse-permuter '(2 4 3 1 0)))
+
 (define (permute-tester msg expected source order . fallback)
   (define (unit type elt-coercer order-type)
     (test* #"permute~msg ~type by ~order-type"
@@ -753,6 +792,11 @@
 (permute-tester " (short)" '(d a) '(a b c d) '(3 0))
 (permute-tester " (long)"  '(d a z c b) '(a b c d) '(3 0 4 2 1) 'z)
 
+(test* "unpermute" '#(a b c d e f g)
+       (unpermute (permute '#(a b c d e f g)
+                           '#(3 6 2 0 1 5 4))
+                  '#(3 6 2 0 1 5 4)))
+
 (define (shuffle-tester source)
   (define (cmp a b)
     (lset= eqv? (coerce-to <list> a) (coerce-to <list> b)))
