@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.unicode - Various Unicode-specific operations
 ;;;
-;;;   Copyright (c) 2011-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2011-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -54,7 +54,9 @@
           codepoints-upcase codepoints-downcase codepoints-titlecase
           codepoints-foldcase
 
-          char-east-asian-width
+          char-east-asian-width string-east-asian-width
+          string-take-width
+          string-drop-width
 
           string-ci=? string-ci<? string-ci<=? string-ci>? string-ci>=?
           )
@@ -461,7 +463,7 @@
          (eof-object)
          (begin0 (combine bvec start) (inc! start 2)))))
 
-(define (string->utf16 str :optional (endian 'big-endian) 
+(define (string->utf16 str :optional (endian 'big-endian)
                                      (add-bom? #f)
                                      (start 0) end)
   (with-builder (<u8vector> add! get)
@@ -499,7 +501,7 @@
       (if (zero? (mod start 4))
         (u32vector->string (uvector-alias <u32vector> bvec start end)
                            0 -1 #f endian)
-        (map-to <string> ucs->char 
+        (map-to <string> ucs->char
                 ($ generator->lseq
                    $ %u8->u32-generator bvec start end endian))))))
 
@@ -522,19 +524,19 @@
 (define (string->utf32 str :optional (endian 'big-endian)
                                      (add-bom? #f)
                                      (start 0) end)
-  (let* ([s ((with-module gauche.internal %maybe-substring) str start end)]
+  (let* ([s (opt-substring str start end)]
          [len (string-length s)]
          [r (make-u8vector (* 4 (if add-bom? (+ len 1) len)))]
          [rr (uvector-alias <u32vector> r (if add-bom? 4 0))])
     (string->u32vector! rr 0 s 0 -1 endian)
     (when add-bom?
       (case endian
-        [(big big-endian) 
+        [(big big-endian)
          (u8vector-set! r 0 #x00)
          (u8vector-set! r 1 #x00)
          (u8vector-set! r 2 #xfe)
          (u8vector-set! r 3 #xff)]
-        [else 
+        [else
          (u8vector-set! r 0 #xff)
          (u8vector-set! r 1 #xfe)
          (u8vector-set! r 2 #x00)
@@ -545,16 +547,18 @@
 ;;;  Character properties
 ;;;
 
+;; unicode-attr.scm is created with src/gen-unicode.scm during build.
+(include "unicode-attr.scm")
+
 ;;=========================================================================
 ;; Low level code to look up properties
 ;;
 
 (inline-stub
  (declcode
-  (.include "gauche/priv/unicode_attr.h"))
+  (.include <gauche/priv/configP.h>
+            "gauche/priv/unicode_attr.h"))
 
- (initcode "init_GB_symbols(Scm_CurrentModule());")
- (initcode "init_WB_symbols(Scm_CurrentModule());")
  (initcode "init_WIDTH_symbols(Scm_CurrentModule());")
 
  (define-cise-stmt get-arg
@@ -572,14 +576,11 @@
  (define-cproc gb-property (scode) ::<int>
    (let* ([ch::int SCM_CHAR_INVALID])
      (get-arg ch scode)
-     (cond [(== ch #x0a) (return GB_LF)]
-           [(== ch #x0d) (return GB_CR)]
-           [(< ch #x20000)
-            (let* ([k::u_char (aref break_table (>> ch 8))])
-              (if (== k 255)
-                (return GB_Other)
-                (let* ([b::u_char (aref break_subtable k (logand ch #xff))])
-                  (return (>> b 4)))))]
+     (cond [(< ch #x20000)
+            (let* ([k::u_char (aref GB_break_table (>> ch 8))])
+              (if (>= k BOFF)
+                (return (- k BOFF)) ; all block is the same
+                (return (aref GB_break_subtable k (logand ch #xff)))))]
            [(or (== #xE0001 ch)
                 (and (<= #xE0020 ch) (<= ch #xE007F))) (return GB_Control)]
            [(and (<= #xE0100 ch) (<= ch #xE01EF)) (return GB_Extend)]
@@ -588,16 +589,11 @@
  (define-cproc wb-property (scode) ::<int>
    (let* ([ch::int SCM_CHAR_INVALID])
      (get-arg ch scode)
-     (cond [(== ch #x0a) (return WB_LF)]
-           [(== ch #x0d) (return WB_CR)]
-           [(== ch #x22) (return WB_Double_Quote)]
-           [(== ch #x27) (return WB_Single_Quote)]
-           [(< ch #x20000)
-            (let* ([k::u_char (aref break_table (>> ch 8))])
-              (if (== k 255)
-                (return WB_Other)
-                (let* ([b::u_char (aref break_subtable k (logand ch #xff))])
-                  (return (logand b #x0f)))))]
+     (cond [(< ch #x20000)
+            (let* ([k::u_char (aref WB_break_table (>> ch 8))])
+              (if (>= k BOFF)
+                (return (- k BOFF)) ; all block is the same
+                (return (aref WB_break_subtable k (logand ch #xff)))))]
            [(or (== #xE0001 ch)
                 (and (<= #xE0020 ch) (<= ch #xE007F))) (return WB_Format)]
            [(and (<= #xE0100 ch) (<= ch #xE01EF)) (return WB_Extend)]
@@ -672,7 +668,7 @@
 (define (make-sequence-splitter cluster-reader-maker)
   (^[seq]
     (let1 gen (x->generator seq)
-      (generator->list 
+      (generator->list
        (cluster-reader-maker gen (cut coerce-to (class-of seq) <>))))))
 
 ;;=========================================================================
@@ -683,6 +679,12 @@
 ;; by input.  The element of inner vector is a pair of
 ;; (<output> . <next-state-index>)
 ;;
+;; For our purpose, <output> can be either one of these:
+;;   #f   - Do not break before the current input.
+;;   #t   - Break before the current input.
+;;   hold - Do not consume the current input and just go to the next state.
+;;   wb6, wb7b, wb12 - This triggers special handling in word breaker
+;;
 ;; Each description is in the form of (state-name arc ...)
 ;; where arc is (input -> output next-state-name)
 ;; and input may be a symbol, (or input ...), or :else.
@@ -694,6 +696,7 @@
 ;; The input-index-map argument is a map to look up input index from
 ;; the input symbol.
 
+(eval-when (:compile-toplevel)
 (define (compile-state-transition-description state-desc default-next-state
                                               input-index-map)
   (define (expand-or alist)
@@ -729,6 +732,7 @@
                [_ (errorf "Can't find transition from state ~s with input ~s"
                           (car state) input-symbol)])))))
     (list->vector (map build-inner-vec states))))
+)
 
 ;;=========================================================================
 ;; Word breaker state transition tables
@@ -742,7 +746,8 @@
 ;; The flag may be #t (break), #f (not break), wb6 (special lookahead for
 ;; WB6) or wb12 (special lookahead for WB12).
 
-(define *word-break-default-next-states*
+(eval-when (:compile-toplevel)
+(define-constant *word-break-default-next-states*
   '((CR                 . :cr)
     ((or Newline LF)    . :newline-lf)
     (Regional_Indicator . :regional-indicator)
@@ -751,48 +756,77 @@
     (ALetter            . :a-letter)
     (Numeric            . :numeric)
     (ExtendNumLet       . :extend-num-let)
+    (WSegSpace          . :wsegspace)
     (:else              . :other)))
 
-(define *word-break-states*
+(define-constant *word-break-states*
   '(;; Initial state (sot)
     ;; From WB1, we always break.
     (:sot
      ((or Newline LF)     -> #t :newline-lf)
      (CR                  -> #t :cr)
      ((or Extend Format)  -> #t :sep-extend-format)
+     (ZWJ                 -> #t :sep-zwj)
      (:else               -> #t :default))
     ;; CR (WB3, WB3a) - break except followed by LF.
     (:cr
      (LF                  -> #f :newline-lf)
-     ((or Extend Format)  -> #t :sep-extend-format)
+     ((or Extend Format ) -> #t :sep-extend-format)
+     (ZWJ                 -> #t :sep-zwj)
      (:else               -> #t :default))
     ;; Newline|LF  (WB3, WB3a)
     (:newline-lf
      ((or Extend Format)  -> #t :sep-extend-format)
+     (ZWJ                 -> #t :sep-zwj)
+     (:else               -> #t :default))
+    ;; Sep + ZWJ x ExtendedPict (WB3c)
+    (:sep-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :sep-extend-format)
+     (ZWJ                 -> #f :sep-zwj)
      (:else               -> #t :default))
     ;; Sep + (Extend|Format)*  (WB4)
     (:sep-extend-format
      ((or Extend Format)  -> #f :sep-extend-format)
+     (ZWJ                 -> #f :sep-zwj)
      (:else               -> #t :default))
     ;; ALetter
     (:a-letter
      ((or Extend Format)  -> #f :a-letter)       ; WB4
+     (ZWJ                 -> #f :a-letter-zwj)   ; WB4
      (ALetter             -> #f :a-letter)       ; WB5
      (Hebrew_Letter       -> #f :hebrew-letter)  ; WB5
      ((or MidLetter MidNumLet Single_Quote) -> wb6 :a-letter+mid) ; WB6
      (Numeric             -> #f :numeric)        ; WB9
      (ExtendNumLet        -> #f :extend-num-let) ; WB13a
      (:else               -> #t :default))
-    ;; ALetter + (MidLetter|MidNumLetQ)
+    ;; ALetter x ZWJ x ExtendedPict (WB3c)
+    (:a-letter-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :a-letter)
+     (ZWJ                 -> #f :a-letter-zwj)
+     (:else               -> hold :a-letter))
+    ;; ALetter x (MidLetter|MidNumLetQ)
     (:a-letter+mid
      ((or Extend Format)  -> #f :a-letter+mid)   ; WB4
+     (ZWJ                 -> #f :a-letter+mid-zwj) ; WB4
      (ALetter             -> #f :a-letter)       ; WB7
      (Hebrew_Letter       -> #f :hebrew-letter)  ; WB7
      (:else               -> #t :default))
+    ;; ALetter + (MidLetter|MidNumLetQ) + ZWJ x ExtendedPict (WB3c)
+    (:a-letter+mid-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :a-letter+mid)
+     (ZWJ                 -> #f :a-letter+mid-zwj)
+     (:else               -> hold :a-letter+mid))
     ;; Hebrew_Letter
     ;; NB: WB7a is actually covered by WB6.
     (:hebrew-letter
      ((or Extend Format)  -> #f :hebrew-letter)  ; WB4
+     (ZWJ                 -> #f :hebrew-letter-zwj) ; WB4
      (ALetter             -> #f :a-letter)       ; WB5
      (Hebrew_Letter       -> #f :hebrew-letter)  ; WB5
      (Single_Quote        -> #f :a-letter+mid)   ; WB7a
@@ -801,69 +835,138 @@
      (Numeric             -> #f :numeric)        ; WB9
      (ExtendNumLet        -> #f :extend-num-let) ; WB13a
      (:else               -> #t :default))
+    ;; Hebrew_Letter x ZWJ x ExtendedPict (WB3c)
+    (:hebrew-letter-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :hebrew-letter)
+     (ZWJ                 -> #f :hebrew-letter-zwj)
+     (:else               -> hold :hebrew-letter))
     ;; Hebrew_Letter + Double_Quote
     (:hebrew-letter+dq
      ((or Extend Format)  -> #f :hebrew-letter+dq) ; WB4
+     (ZWJ                 -> #f :hebrew-letter+dq-zwj) ; WB4
      (Hebrew_Letter       -> #f :hebrew-letter)    ; WB7c
      (:else               -> #t :default))
+    ;; Hebrew_Letter + Double_Quote + ZWJ x ExtendedPict (WB3c)
+    (:hebrew-letter+dq-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :hebrew-letter+dq) ; WB4
+     (ZWJ                 -> #f :hebrew-letter+dq-zwj)
+     (:else               -> hold :hebrew-letter+dq))
     (:numeric
      ((or Extend Format)  -> #f :numeric)        ; WB4
+     (ZWJ                 -> #f :numeric-zwj)    ; WB4
      (Numeric             -> #f :numeric)        ; WB8
      (ALetter             -> #f :a-letter)       ; WB10
      (Hebrew_Letter       -> #f :hebrew-letter)  ; WB10
-     ((or MidLetter MidNumLet Single_Quote) -> wb12 :numeric+mid) ; WB12
+     ((or MidNum MidNumLet Single_Quote) -> wb12 :numeric+mid) ; WB12
      (ExtendNumLet        -> #f :extend-num-let) ; WB13a
      (:else               -> #t :default))
+    (:numeric-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :numeric)        ; WB4
+     (ZWJ                 -> #f :numeric-zwj)
+     (:else               -> hold :numeric))
     (:numeric+mid
      ((or Extend Format)  -> #f :numeric+mid)    ; WB4
+     (ZWJ                 -> #f :numeric+mid-zwj) ; WB4
      (Numeric             -> #f :numeric)        ; WB11
      (:else               -> #t :default))
+    (:numeric+mid-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :numeric+mid)    ; WB4
+     (ZWJ                 -> #f :numeric+mid-zwj)
+     (:else               -> hold :numeric+mid))
     (:katakana
      ((or Extend Format)  -> #f :katakana)       ; WB4
+     (ZWJ                 -> #f :katakana-zwj)   ; WB4
      (Katakana            -> #f :katakana)       ; WB13
+     (ExtendNumLet        -> #f :extend-num-let) ; WB13a
      (:else               -> #t :default))
+    (:katakana-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :katakana)       ; WB4
+     (ZWJ                 -> #f :katakana-zwj)
+     (:else               -> hold :katakana))
     (:extend-num-let
      ((or Extend Format)  -> #f :extend-num-let) ; WB4
+     (ZWJ                 -> #f :extend-num-let-zwj) ; WB4
      (ExtendNumLet        -> #f :extend-num-let) ; WB13a
      (ALetter             -> #f :a-letter)       ; WB13b
      (Hebrew_Letter       -> #f :hebrew-letter)  ; WB13b
      (Numeric             -> #f :numeric)        ; WB13b
      (Katakana            -> #f :katakana)       ; WB13b
      (:else               -> #t :default))
+    (:extend-num-let-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :extend-num-let)
+     (ZWJ                 -> #f :extend-num-let-zwj)
+     (:else               -> hold :extend-num-let))
     (:regional-indicator
      ((or Extend Format)  -> #f :regional-indicator) ; WB4
-     (Regional_Indicator  -> #f :regional-indicator) ; WB13c
+     (ZWJ                 -> #f :regional-indicator-zwj) ; WB4
+     (Regional_Indicator  -> #f :regional-indicator-1) ; WB15, WB16
      (:else               -> #t :default))
+    (:regional-indicator-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :regional-indicator)
+     (ZWJ                 -> #f :regional-indicator-zwj)
+     (:else               -> hold :regional-indicator))
+    (:regional-indicator-1
+     (Regional_Indicator  -> #t :regional-indicator) ; WB15, WB16
+     ((or Extend Format)  -> #f :regional-indicator-1)
+     (ZWJ                 -> #f :regional-indicator-1-zwj)
+     (:else               -> #t :default))
+    (:regional-indicator-1-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :regional-indicator-1)
+     (ZWJ                 -> #f :regional-indicator-1-zwj)
+     (:else               -> hold :regional-indicator-1))
+    ;; WSegSpace (WB3d)
+    ;; NB: WordBreakTest.txt indicates WSegSpace x Extend ÷ WSegSpace,
+    ;; so we need :wsegspace-ex
+    (:wsegspace
+     (WSegSpace           -> #f :wsegspace)      ; WB3d
+     ((or Extend Format)  -> #f :wsegspace-ex)
+     (ZWJ                 -> #f :wsegspace-zwj)  ; WB4
+     (:else               -> #t :default))
+    (:wsegspace-ex
+     ((or Extend Format)  -> #f :wsegspace-ex)
+     (ZWJ                 -> #f :wsegspace-zwj)
+     (:else               -> #t :default))
+    (:wsegspace-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :wsegspace-ex)   ; WB4
+     (ZWJ                 -> #f :wsegspace-zwj)
+     (:else               -> hold :wsegspace))
     (:other
+     ((or Extend Format)  -> #f :other)          ; WB4
+     (ZWJ                 -> #f :other-zwj)      ; WB4
      (:else               -> #t :default))       ; WB14
+    (:other-zwj
+     (ExtPict             -> #f :other)
+     (ExtPict_ALetter     -> #f :a-letter)
+     ((or Extend Format)  -> #f :other)
+     (ZWJ                 -> #f :other-zwj)
+     (:else               -> hold :other))
     ))
 
-(define *word-break-fa*
+(define-constant *word-break-fa*
   (compile-state-transition-description *word-break-states*
                                         *word-break-default-next-states*
-                                        ;; NB: This mapping should be provided
-                                        ;; by text.unicode.ucd, but we hack it
-                                        ;; up for now.
-                                        (hash-table 'eq?
-                                                    '(Newline . 0)
-                                                    '(Extend . 1)
-                                                    '(Regional_Indicator . 2)
-                                                    '(Format . 3)
-                                                    '(Katakana . 4)
-                                                    '(Hebrew_Letter . 5)
-                                                    '(ALetter . 6)
-                                                    '(MidLetter . 7)
-                                                    '(MidNum . 8)
-                                                    '(MidNumLet . 9)
-                                                    '(Numeric . 10)
-                                                    '(ExtendNumLet . 11)
-                                                    '(WSegSpace . 12)
-                                                    '(ZWJ . 13)
-                                                    '(Other . 14)
-                                                    '(CR . 16)
-                                                    '(LF . 17)
-                                                    '(Single_Quote . 18)
-                                                    '(Double_Quote . 19))))
+                                        (alist->hash-table
+                                         *WB-properties* ; in unicode-attr.scm
+                                         'eq?)))
+);eval-when
 
 ;; Simple queue to avoid depending data.queue
 (define (makeq) (cons '() '()))
@@ -875,6 +978,10 @@
 (define (deq! q) (if (q-empty? q)
                    (error "[internal] queue is empty")
                    (rlet1 p (caar q) (set-car! q (cdar q)))))
+(define (pushq! q item) (if (q-empty? q)
+                          (let1 p (list item)
+                            (set-car! q p) (set-cdr! q p))
+                          (set-car! q (cons item (car q)))))
 
 ;; API
 ;; Returns a generator procedure.  Every time it is called, it returns
@@ -886,13 +993,14 @@
 (define (make-word-breaker generator)
   (define current-state (vector-ref *word-break-fa* 0))
   (define q (makeq))  ;characters looked ahead.
-  (^[]
+  (rec (gen)
     (let1 ch (if (q-empty? q) (generator) (deq! q))
       (if (eof-object? ch)
         (values ch #t) ; WB2
         (let1 p (vector-ref current-state (wb-property ch))
           (set! current-state (vector-ref *word-break-fa* (cdr p)))
           (cond [(boolean? (car p)) (values ch (car p))]
+                [(eq? (car p) 'hold) (pushq! q ch) (gen)]
                 [(eq? (car p) 'wb6)
                  (wb-lookahead generator q ch `(,WB_ALetter ,WB_Hebrew_Letter))]
                 [(eq? (car p) 'wb12)
@@ -912,6 +1020,7 @@
         (cond [(memv prop next-props) (values ch #f)]
               [(eqv? prop WB_Extend) (loop (generator))]
               [(eqv? prop WB_Format) (loop (generator))]
+              [(eqv? prop WB_ZWJ) (loop (generator))]
               [else (values ch #t)])))))
 
 ;; API
@@ -927,7 +1036,8 @@
 ;; and entry is a cons of boolean value (#t - break, #f - not break),
 ;; and the next state vector index.
 
-(define *grapheme-break-default-next-states*
+(eval-when (:compile-toplevel)
+(define-constant *grapheme-break-default-next-states*
   '((CR                 . :cr)
     (LF                 . :control-lf)
     (Control            . :control-lf)
@@ -936,9 +1046,10 @@
     ((or LVT T)         . :lvt-t)
     (Regional_Indicator . :regional-indicator)
     (Prepend            . :prepend)
+    (ExtPict            . :ext-pict)
     (:else              . :other)))
 
-(define *grapheme-break-states*
+(define-constant *grapheme-break-states*
   '(;; Initial state - sot
     (:sot
      (:else              -> #t :default))   ; GB1
@@ -952,48 +1063,56 @@
      (L                  -> #f :l)          ; GB6
      ((or V LV)          -> #f :lv-v)       ; GB6
      (LVT                -> #f :lvt-t)      ; GB6
+     ((or Extend ZWJ)    -> #f :other)      ; GB9
+     (SpacingMark        -> #f :other)      ; GB9a
      (:else              -> #t :default))
     (:lv-v
      (V                  -> #f :lv-v)       ; GB7
      (T                  -> #f :lvt-t)      ; GB7
+     ((or Extend ZWJ)    -> #f :other)      ; GB9
+     (SpacingMark        -> #f :other)      ; GB9a
      (:else              -> #t :default))
     (:lvt-t
      (T                  -> #f :lvt-t)      ; GB8
+     ((or Extend ZWJ)    -> #f :other)      ; GB9
+     (SpacingMark        -> #f :other)      ; GB9a
      (:else              -> #t :default))
     (:regional-indicator
-     (Regional_Indicator -> #f :regional-indicator) ; GB8a
+     (Regional_Indicator -> #f :regional-indicator-1) ; GB12, GB13
+     ((or Extend ZWJ)    -> #f :other)                ; GB9
+     (SpacingMark        -> #f :other)                ; GB9a
+     (:else              -> #t :default))
+    (:regional-indicator-1
+     (Regional_Indicator -> #t :regional-indicator) ; GB12, GB13
+     ((or Extend ZWJ)    -> #f :other)              ; GB9
+     (SpacingMark        -> #f :other)              ; GB9a
      (:else              -> #t :default))
     (:prepend
      (CR                 -> #t :cr)         ; GB5
      ((or Control LF)    -> #t :control-lf) ; GB5
      (:else              -> #f :default))   ; GB9b
+    (:ext-pict
+     (Extend             -> #f :ext-pict)   ; GB11
+     (ZWJ                -> #f :ext-pict-1) ; GB9
+     (SpacingMark        -> #f :other)      ; GB9a
+     (:else              -> #t :default))
+    (:ext-pict-1
+     (ExtPict            -> #f :ext-pict)   ; GB11
+     (SpacingMark        -> #f :other)      ; GB9a
+     (:else              -> #t :default))
     (:other
-     (Extend             -> #f :other)      ; GB9
+     ((or Extend ZWJ)    -> #f :other)      ; GB9
      (SpacingMark        -> #f :other)      ; GB9a
      (:else              -> #t :default))   ; GB10
     ))
 
-(define *grapheme-break-fa*
+(define-constant *grapheme-break-fa*
   (compile-state-transition-description *grapheme-break-states*
                                         *grapheme-break-default-next-states*
-                                        ;; NB: This mapping should be provided
-                                        ;; by text.unicode.ucd, but we hack it
-                                        ;; up for now.
-                                        (hash-table 'eq?
-                                                    '(Control . 0)
-                                                    '(Extend . 1)
-                                                    '(Regional_Indicator . 2)
-                                                    '(Prepend . 3)
-                                                    '(SpacingMark . 4)
-                                                    '(L . 5)
-                                                    '(V . 6)
-                                                    '(T . 7)
-                                                    '(LV . 8)
-                                                    '(LVT . 9)
-                                                    '(ZWJ . 10)
-                                                    '(Other . 11)
-                                                    '(CR . 16)
-                                                    '(LF . 17))))
+                                        (alist->hash-table
+                                         *GB-properties* ; in unicode_attr.scm
+                                         'eq?)))
+);eval-when
 
 ;; API
 ;; Returns a generator, that returns two values, a character and
@@ -1026,6 +1145,49 @@
 (define (char-east-asian-width char-or-code)
   (vector-ref *east-asian-widths* (width-property char-or-code)))
 
+(define (%char->width ch kargs)
+  (let-keywords* kargs ((F 2)  ;full-width
+                        (H 1)  ;half-width
+                        (W 2)  ;wide
+                        (Na 1) ;narrow
+                        (N 1)  ;neutral
+                        (A 2)) ;ambiguous
+    (case (char-east-asian-width ch)
+      [(F) F]
+      [(H) H]
+      [(W) W]
+      [(Na) Na]
+      [(N) N]
+      [(A) A])))
+
+;; compute the on-screen 'width' according to the East Asian Width property.
+;; See UAX #11 https://www.unicode.org/reports/tr11/tr11-40.html
+(define (string-east-asian-width str :key F H W Na N A :rest keys)
+  (generator-fold (^[ch sum]
+                    (+ sum (%char->width ch keys)))
+                  0 (string->generator str)))
+
+(define (string-take-width str width :key F H W Na N A :rest keys)
+  (with-string-io str
+    (^[] (let loop ([w 0] [ch (read-char)])
+           (unless (eof-object? ch)
+             (let1 w1 (%char->width ch keys)
+               (unless (> (+ w w1) width)
+                 (write-char ch)
+                 (loop (+ w w1) (read-char)))))))))
+
+(define (string-drop-width str width :key F H W Na N A :rest keys)
+  (call-with-string-io str
+    (^[in out]
+      (let loop ([w 0] [ch (read-char in)])
+        (unless (eof-object? ch)
+          (let1 w1 (%char->width ch keys)
+            (if (> (+ w w1) width)
+              (begin
+                (write-char ch out)
+                (copy-port in out))
+              (loop (+ w w1) (read-char in)))))))))
+
 ;;;
 ;;; String casing
 ;;;
@@ -1100,18 +1262,8 @@
     (generator-for-each (^[ch] (%tr ch buf CHAR_UPCASE sink char?)) generator)))
 
 ;; Greek capital sigma U+03a3 and final sigma U+03c2.
-;; We can't use character literal #\u03a3 etc., for they're not valid
-;; when internal encoding is 'none'.  We intend to make utf-8 support
-;; mandatory in future, and this limitation will be lifted.
-;; For now, they'll be #f when the internal encoding is 'none'.
-(define-constant .capital-sigma.
-  (cond-expand
-   [gauche.ces.none #f]
-   [else (ucs->char #x03a3)]))
-(define-constant .final-sigma.
-  (cond-expand
-   [gauche.ces.none #f]
-   [else (ucs->char #x03c2)]))
+(define-constant .capital-sigma. #\u03a3)
+(define-constant .final-sigma. #\u03c2)
 
 ;; Greek capital sigma U+03a3 needs context-sensitive conversion.
 (define (%downcase generator sink char?)
