@@ -65,8 +65,8 @@
 #define UPPER_MASK 0x80000000UL /* most significant w-r bits */
 #define LOWER_MASK 0x7fffffffUL /* least significant r bits */
 
-/* initializes mt[N] with a seed */
-void Scm_MTInitByUI(ScmMersenneTwister *mt, unsigned long s)
+/* initializes mt[N] with a seed.  caller needs to hold lock */
+static void init_by_ui(ScmMersenneTwister *mt, unsigned long s)
 {
     int mti;
     mt->mt[0]= s & 0xffffffffUL;
@@ -81,47 +81,19 @@ void Scm_MTInitByUI(ScmMersenneTwister *mt, unsigned long s)
         /* for >32 bit machines */
     }
     mt->mti = mti;
+    mt->seed = Scm_MakeIntegerU(s);
 }
 
-void Scm_MTSetSeed(ScmMersenneTwister *mt, ScmObj seed)
-{
-    if (SCM_INTP(seed)) {
-        /* For the backward compatibility, we only use the lower 32bit. */
-        Scm_MTInitByUI(mt, (uint32_t)Scm_GetUInteger(seed));
-    } else if (SCM_BIGNUMP(seed)) {
-#if SIZEOF_LONG == 4
-        Scm_MTInitByArray(mt, (int32_t*)SCM_BIGNUM(seed)->values,
-                          (int)SCM_BIGNUM_SIZE(seed));
-#elif SIZEOF_LONG == 8
-        /* We can't just pass seed->values as (int32_t*), for the result
-           would differ by endianness. */
-        int32_t seedv[SCM_BIGNUM_SIZE(seed)*2];
-        for (size_t i=0; i<SCM_BIGNUM_SIZE(seed); i++) {
-            seedv[i*2] = (int32_t)(SCM_BIGNUM(seed)->values[i]&0xffffffff);
-            seedv[i*2+1] = (int32_t)(SCM_BIGNUM(seed)->values[i]>>32);
-        }
-        Scm_MTInitByArray(mt, seedv, (int)SCM_BIGNUM_SIZE(seed)*2);
-#else
-#error "sizeof(long) must be 4 or 8"
-#endif
-    } else if (SCM_U32VECTORP(seed)) {
-        Scm_MTInitByArray(mt, (int32_t*)SCM_U32VECTOR_ELEMENTS(seed),
-                          SCM_U32VECTOR_SIZE(seed));
-    } else {
-        Scm_TypeError("random seed", "an exact integer or u32vector", seed);
-    }
-}
-
-
 /* initialize by an array with array-length */
 /* init_key is the array for initializing keys */
 /* key_length is its length */
-void Scm_MTInitByArray(ScmMersenneTwister *mt,
-                       int32_t init_key[],
-                       unsigned long key_length)
+/* caller needs to hold lock */
+static void init_by_array(ScmMersenneTwister *mt,
+                          int32_t init_key[],
+                          unsigned long key_length)
 {
     int i, j, k;
-    Scm_MTInitByUI(mt, 19650218UL);
+    init_by_ui(mt, 19650218UL);
     i=1; j=0;
     k = (N>key_length ? N : key_length);
     for (; k; k--) {
@@ -141,10 +113,64 @@ void Scm_MTInitByArray(ScmMersenneTwister *mt,
     }
 
     mt->mt[0] = 0x80000000UL; /* MSB is 1; assuring non-zero initial array */
+    /* NB: We coerce the seed value to u32vector, for the compability of
+       Scheme-level mt-random-set-seed!. */
+    mt->seed = Scm_MakeU32VectorFromArray(key_length, (uint32_t*)init_key);
 }
 
-/* generates a random number on [0,0xffffffff]-interval */
-unsigned long Scm_MTGenrandU32(ScmMersenneTwister *mt)
+#define LOCK(mt)                                        \
+    do {                                                \
+        if (SCM_MERSENNE_TWISTER_NEED_LOCK(mt)) {       \
+            SCM_INTERNAL_MUTEX_LOCK(mt->lock);          \
+        }                                               \
+    } while (0)
+
+#define UNLOCK(mt)                                      \
+    do {                                                \
+        if (SCM_MERSENNE_TWISTER_NEED_LOCK(mt)) {       \
+            SCM_INTERNAL_MUTEX_UNLOCK(mt->lock);        \
+        }                                               \
+    } while (0)
+
+
+void Scm_MTSetSeed(ScmMersenneTwister *mt, ScmObj seed)
+{
+    if (!(SCM_INTP(seed) || SCM_BIGNUMP(seed) || SCM_U32VECTORP(seed))) {
+        Scm_TypeError("random seed", "an exact integer or u32vector", seed);
+    }
+
+    LOCK(mt);
+    if (SCM_INTP(seed)) {
+        /* For the backward compatibility, we only use the lower 32bit. */
+        init_by_ui(mt, (uint32_t)Scm_GetUInteger(seed));
+    } else if (SCM_BIGNUMP(seed)) {
+#if SIZEOF_LONG == 4
+        init_by_array(mt, (int32_t*)SCM_BIGNUM(seed)->values,
+                          (int)SCM_BIGNUM_SIZE(seed));
+#elif SIZEOF_LONG == 8
+        /* We can't just pass seed->values as (int32_t*), for the result
+           would differ by endianness. */
+        int32_t seedv[SCM_BIGNUM_SIZE(seed)*2];
+        for (size_t i=0; i<SCM_BIGNUM_SIZE(seed); i++) {
+            seedv[i*2] = (int32_t)(SCM_BIGNUM(seed)->values[i]&0xffffffff);
+            seedv[i*2+1] = (int32_t)(SCM_BIGNUM(seed)->values[i]>>32);
+        }
+        init_by_array(mt, seedv, (int)SCM_BIGNUM_SIZE(seed)*2);
+#else
+#error "sizeof(long) must be 4 or 8"
+#endif
+    } else {
+        /* seed must be u32vector */
+        init_by_array(mt, (int32_t*)SCM_U32VECTOR_ELEMENTS(seed),
+                          SCM_U32VECTOR_SIZE(seed));
+    }
+    /* NB: mt->seed is set by either init_by_ui or init_by_array.  */
+
+    UNLOCK(mt);
+}
+
+/* internal.  caller holds the lock */
+static u_long genrand_u32(ScmMersenneTwister *mt)
 {
     unsigned long y;
     int mti = mt->mti;
@@ -154,8 +180,9 @@ unsigned long Scm_MTGenrandU32(ScmMersenneTwister *mt)
     if (mti >= N) { /* generate N words at one time */
         int kk;
 
-        if (mti == N+1)   /* if Scm_MTInitByUI() has not been called, */
-            Scm_MTInitByUI(mt, 5489UL); /* a default initial seed is used */
+        if (mti == N+1) {   /* if init_by_ui() has not been called, */
+            init_by_ui(mt, 5489UL); /* a default initial seed is used */
+        }
 
         for (kk=0;kk<N-M;kk++) {
             y = (mt->mt[kk]&UPPER_MASK)|(mt->mt[kk+1]&LOWER_MASK);
@@ -180,33 +207,87 @@ unsigned long Scm_MTGenrandU32(ScmMersenneTwister *mt)
     y ^= (y >> 18);
 
     mt->mti = mti;
+
     return y;
 }
 
+ScmObj Scm_MTGetState(ScmMersenneTwister *mt)
+{
+    ScmObj v = Scm_MakeU32Vector(N+1, 0);
+    LOCK(mt);
+    for (int i=0; i < N; i++) {
+        SCM_U32VECTOR_ELEMENT(v, i) = mt->mt[i];
+    }
+    SCM_U32VECTOR_ELEMENT(v, N) = mt->mti;
+    UNLOCK(mt);
+    return v;
+}
+
+void Scm_MTSetState(ScmMersenneTwister *mt, ScmU32Vector *v)
+{
+    if (SCM_U32VECTOR_SIZE(v) != N+1) {
+        Scm_Error("u32vector of length %d is required, but got length %d",
+                  N+1, SCM_U32VECTOR_SIZE(v));
+    }
+    LOCK(mt);
+    for (int i=0; i < N; i++) {
+        mt->mt[i] = SCM_U32VECTOR_ELEMENT(v, i);
+    }
+    mt->mti = SCM_U32VECTOR_ELEMENT(v, N);
+    UNLOCK(mt);
+}
+
+
+/* generates a random number on [0,0xffffffff]-interval */
+unsigned long Scm_MTGenrandU32(ScmMersenneTwister *mt)
+{
+    u_long r;
+    LOCK(mt);
+    r = genrand_u32(mt);
+    UNLOCK(mt);
+    return r;
+}
+
 /* generates a random number on (0,1) or [0,1) -real-interval */
-float Scm_MTGenrandF32(ScmMersenneTwister *mt, int exclude0)
+static float genrand_f32(ScmMersenneTwister *mt, int exclude0)
 {
     float r;
     do {
-        r = (float)(Scm_MTGenrandU32(mt)*(1.0/4294967296.0));
+        r = (float)(genrand_u32(mt)*(1.0/4294967296.0));
         /* divided by 2^32 */
     } while (exclude0 && r == 0.0); /*if we get 0.0, try another one. */;
     return r;
 }
 
+float Scm_MTGenrandF32(ScmMersenneTwister *mt, int exclude0)
+{
+    LOCK(mt);
+    float r = genrand_f32(mt, exclude0);
+    UNLOCK(mt);
+    return r;
+}
+
 /* generates a random number on (0,1) or [0,1) with 53-bit resolution*/
-double Scm_MTGenrandF64(ScmMersenneTwister *mt, int exclude0)
+static double genrand_f64(ScmMersenneTwister *mt, int exclude0)
 {
     double r;
     unsigned long a, b;
     do {
-        a = Scm_MTGenrandU32(mt)>>5;
-        b = Scm_MTGenrandU32(mt)>>6;
+        a = genrand_u32(mt)>>5;
+        b = genrand_u32(mt)>>6;
         r = (a*67108864.0+b)*(1.0/9007199254740992.0);
     } while (exclude0 && r == 0.0); /*if we get 0.0, try another one. */;
     return r;
 }
 
+double Scm_MTGenrandF64(ScmMersenneTwister *mt, int exclude0)
+{
+    LOCK(mt);
+    double r = genrand_f64(mt, exclude0);
+    UNLOCK(mt);
+    return r;
+}
+
 /*
  * Generic integer routine for [0, n-1], 0 < n <= 2^32
  */
@@ -288,36 +369,47 @@ ScmObj Scm_MTGenrandInt(ScmMersenneTwister *mt, ScmObj n)
     }
 #endif
   err:
-    Scm_Error("bad type of argument for n: positive integer up to 2^32 is required, but got %S", n);
+    Scm_Error("bad type of argument for n: positive exact integer up to 2^32 is required, but got %S", n);
     return SCM_UNDEFINED; /*dummy*/
 }
 
+ScmObj Scm_MTFillUvector(ScmMersenneTwister *mt, ScmObj v)
+{
+    if (SCM_U32VECTORP(v)) {
+        LOCK(mt);
+        for (int i = 0; i < SCM_U32VECTOR_SIZE(v); i++) {
+            SCM_U32VECTOR_ELEMENT(v, i) = genrand_u32(mt);
+        }
+        UNLOCK(mt);
+    } else if (SCM_F32VECTORP(v)) {
+        LOCK(mt);
+        for (int i = 0; i < SCM_F32VECTOR_SIZE(v); i++) {
+            SCM_F32VECTOR_ELEMENT(v, i) = genrand_f32(mt, TRUE);
+        }
+        UNLOCK(mt);
+    } else if (SCM_F64VECTORP(v)) {
+        LOCK(mt);
+        for (int i = 0; i < SCM_F64VECTOR_SIZE(v); i++) {
+            SCM_F64VECTOR_ELEMENT(v, i) = genrand_f64(mt, TRUE);
+        }
+        UNLOCK(mt);
+    }
+    return v;
+}
+
 /*
  * Gauche specific stuff
  */
-static ScmObj key_seed;
-static ScmObj mt_allocate(ScmClass *klass, ScmObj initargs);
-SCM_DEFINE_BUILTIN_CLASS(Scm_MersenneTwisterClass,
-                         NULL, NULL, NULL, mt_allocate,
-                         NULL);
-
-static ScmObj mt_allocate(ScmClass *klass SCM_UNUSED, ScmObj initargs)
+ScmObj Scm_MakeMT(ScmObj seed, u_int flags)
 {
-    ScmObj seed = Scm_GetKeyword(key_seed, initargs, SCM_FALSE);
-    ScmMersenneTwister *mt;
-
-    mt = SCM_NEW(ScmMersenneTwister);
+    ScmMersenneTwister *mt = SCM_NEW(ScmMersenneTwister);
     SCM_SET_CLASS(mt, &Scm_MersenneTwisterClass);
+    mt->flags = flags;
     mt->mti = N+1;
+    mt->seed = SCM_UNDEFINED;
     if (!SCM_FALSEP(seed)) Scm_MTSetSeed(mt, seed);
+    if (SCM_MERSENNE_TWISTER_NEED_LOCK(mt)) {
+        SCM_INTERNAL_MUTEX_INIT(mt->lock);
+    }
     return SCM_OBJ(mt);
 }
-
-void Scm_Init_mt_random(void)
-{
-    ScmModule *mod = SCM_FIND_MODULE("math.mt-random", SCM_FIND_MODULE_CREATE);
-    Scm_InitStaticClass(&Scm_MersenneTwisterClass, "<mersenne-twister>",
-                        mod, NULL, 0);
-    key_seed = SCM_MAKE_KEYWORD("seed");
-}
-
