@@ -1,7 +1,7 @@
 ;;;
 ;;; mt-random - Mersenne Twister interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,8 +32,9 @@
 ;;;
 
 (define-module math.mt-random
-  (use gauche.uvector)
   (export <mersenne-twister>
+          make-mersenne-twister
+          mt-random-get-seed
           mt-random-set-seed!
           mt-random-get-state
           mt-random-set-state!
@@ -49,32 +50,38 @@
 (inline-stub
  (declcode
   (.include "mt-random.h"))
- (initcode (Scm_Init_mt_random))
- 
- (define-type <mersenne-twister> "ScmMersenneTwister*")
- (define-type <u32vector> "ScmU32Vector*")
- (define-type <f32vector> "ScmF32Vector*")
- (define-type <f64vector> "ScmF64Vector*")
+
+ (define-cclass <mersenne-twister> "ScmMersenneTwister*"
+   "Scm_MersenneTwisterClass"
+   (c "SCM_CLASS_DEFAULT_CPL")
+   ()                                   ;no external slot
+   (allocator (let* ([seed (Scm_GetKeyword ':seed initargs '#f)]
+                     [priv (Scm_GetKeyword ':private? initargs '#f)]
+                     [flags::u_int (?: (SCM_FALSEP priv)
+                                       0
+                                       SCM_MERSENNE_TWISTER_PRIVATE)])
+                (cast void klass)
+                (return (Scm_MakeMT seed flags)))))
+
+ (define-cproc make-mersenne-twister (:optional (seed #f)
+                                                (private?::<boolean> #f))
+   (let* ([flags::u_int 0])
+     (when private?
+       (set! flags SCM_MERSENNE_TWISTER_PRIVATE))
+     (return (Scm_MakeMT seed flags))))
+
+ (define-cproc mt-random-get-seed (mt::<mersenne-twister>)
+   (return (-> mt seed)))
 
  (define-cproc mt-random-set-seed! (mt::<mersenne-twister> init) ::<void>
    Scm_MTSetSeed)
 
  (define-cproc mt-random-get-state (mt::<mersenne-twister>)
-   (let* ([v (Scm_MakeU32Vector (+ N 1) 0)])
-     (dotimes (i N)
-       (set! (aref (SCM_U32VECTOR_ELEMENTS v) i) (aref (-> mt mt) i)))
-     (set! (aref (SCM_U32VECTOR_ELEMENTS v) N) (-> mt mti))
-     (return v)))
+   Scm_MTGetState)
 
  (define-cproc mt-random-set-state! (mt::<mersenne-twister> state::<u32vector>)
    ::<void>
-   (unless (== (SCM_U32VECTOR_SIZE state) (+ N 1))
-     (Scm_Error "u32vector of length %d is required, but got length %d"
-                (+ N 1) (SCM_U32VECTOR_SIZE state)))
-   (dotimes [i N]
-     (set! (aref (-> mt mt) i)
-           (aref (SCM_U32VECTOR_ELEMENTS state) i)))
-   (set! (-> mt mti) (aref (SCM_U32VECTOR_ELEMENTS state) N)))
+   Scm_MTSetState)
 
  (define-cproc mt-random-real (mt::<mersenne-twister>) ::<double>
    (return (Scm_MTGenrandF64 mt TRUE)))
@@ -89,22 +96,13 @@
    Scm_MTGenrandU32)
 
  (define-cproc mt-random-fill-u32vector! (mt::<mersenne-twister> v::<u32vector>)
-   (let* ([p::uint32_t* (SCM_U32VECTOR_ELEMENTS v)])
-     (dotimes [i (SCM_U32VECTOR_SIZE v)]
-       (set! (* (post++ p)) (Scm_MTGenrandU32 mt)))
-     (return (SCM_OBJ v))))
+   (return (Scm_MTFillUvector mt (SCM_OBJ v))))
 
  (define-cproc mt-random-fill-f32vector! (mt::<mersenne-twister> v::<f32vector>)
-   (let* ([p::float* (SCM_F32VECTOR_ELEMENTS v)])
-     (dotimes [i (SCM_F32VECTOR_SIZE v)]
-       (set! (* (post++ p)) (Scm_MTGenrandF32 mt TRUE)))
-     (return (SCM_OBJ v))))
+   (return (Scm_MTFillUvector mt (SCM_OBJ v))))
 
  (define-cproc mt-random-fill-f64vector! (mt::<mersenne-twister> v::<f64vector>)
-   (let* ([p::double* (SCM_F64VECTOR_ELEMENTS v)])
-     (dotimes [i (SCM_F64VECTOR_SIZE v)]
-       (set! (* (post++ p)) (Scm_MTGenrandF64 mt TRUE)))
-     (return (SCM_OBJ v))))
+   (return (Scm_MTFillUvector mt (SCM_OBJ v))))
  )
 
 (define (%get-nword-random-int mt n)
@@ -125,4 +123,3 @@
         (if (< r qn)
           (quotient r q)
           (loop (%get-nword-random-int mt siz)))))))
-
