@@ -4,8 +4,8 @@
 
 (use gauche.test)
 (use gauche.collection)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (use gauche.uvector)
 
 (test-start "mt-random")
@@ -61,11 +61,6 @@
                      (make-random-sequence <list> 100
                                            (^[] (mt-random-real m1)))))))
 
-;;; NB: Cygwin version doesn't suppor some procedures.  Exit here.
-(when (string-contains (gauche-architecture) "cygwin")
-  (test-end)
-  (exit 0))
-
 (test "seed" #t
       (^[] (let ([m0 (make <mersenne-twister>)]
                  [m1 (make <mersenne-twister>)])
@@ -86,6 +81,17 @@
                      (make-random-sequence <list> 100
                                            (^[] (mt-random-real m1)))))))
 
+(test "seed transplant" #t
+      (^[] (let ([m0 (make <mersenne-twister>)]
+                 [m1 (make <mersenne-twister>)])
+             (mt-random-set-seed! m0 '#u32(472346 37429385 72))
+             (let1 s0 (make-random-sequence <list> 100
+                                            (^[] (mt-random-real m0)))
+               (mt-random-set-seed! m1 (mt-random-get-seed m0))
+               (equal? s0
+                       (make-random-sequence <list> 100
+                                             (^[] (mt-random-real m1))))))))
+
 (test "u32vector" #t
       (^[] (let ([m0 (make <mersenne-twister> :seed 1)]
                  [m1 (make <mersenne-twister> :seed 1)])
@@ -110,46 +116,5 @@
                      (make-random-sequence <list> 100 (^[] (mt-random-real m2)))
                      ))))
 
-;;-------------------------------------------------------------------
-;; srfi-27 is built on top of mt-random, so we test it here.
-(test-section "srfi-27")
-
-(use srfi-27)
-(test-module 'srfi-27)
-
-(let ([s0 (make-random-source)]
-      [s1 (make-random-source)]
-      [saved #f])
-  (define (get-random-series source)
-    (let* ([random0 (random-source-make-integers source)]
-           [random1 (random-source-make-reals source)])
-      (append-map (^_ (let* ([r0 (random0 10)]
-                             [r1 (random1)])
-                        (list r0 r1)))
-                  (iota 10))))
-  (define (unit-random source unit)
-    (let1 random (random-source-make-reals source unit)
-      (test* "real random unit exactness" (exact? unit)
-             (exact? (random)))
-      (test* "real random unit range" #t
-             (every (^_ (< 0 (random) 1)) (iota 100)))))
-  
-  (test* "random-source?" #t (random-source? s0))
-
-  (random-source-randomize! s0)
-  (random-source-randomize! s1)
-
-  (set! saved (random-source-state-ref s0))
-
-  (let1 result (get-random-series s0)
-    (test* "random-state" result
-           (begin
-             (random-source-state-set! s1 saved)
-             (get-random-series s1))))
-
-  (unit-random s0 0.1)
-  (unit-random s0 1/3)
-  )
 
 (test-end)
-
