@@ -2,7 +2,7 @@
 ;;; peg.scm - Parser Expression Grammar Parser
 ;;;
 ;;;   Copyright (c) 2006 Rui Ueyama (rui314@gmail.com)
-;;;   Copyright (c) 2008-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2008-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,8 +33,9 @@
 ;;;
 
 (define-module parser.peg
-  (use srfi-13)
-  (use srfi-14)
+  (use scheme.list)
+  (use scheme.charset)
+  (use srfi.13)
   (use gauche.collection)
   (use gauche.generator)
   (use gauche.lazy)
@@ -45,17 +46,17 @@
 
           peg-run-parser
           peg-parse-string peg-parse-port
-          peg-parser->generator ;experimental
+          peg-parser->generator peg-parser->lseq
 
           parse-success?
           return-result return-failure
           return-failure/expect return-failure/unexpect
           return-failure/message return-failure/compound
-          
+
           $bind $return $fail $expect $lift $lift* $debug
           $let $let* $try $assert $not
           $or $fold-parsers $fold-parsers-right
-          $seq $seq0 $between
+          $seq $seq0 $list $list* $between
           $many $many1 $many_ $many1_ $repeat $repeat_
           $many-till $many-till_
           $optional
@@ -63,11 +64,14 @@
           $chain-left $chain-right
           $lazy $parameterize
 
+          $cut $raise
+
           $any $eos $.
-          
-          $symbol $string $string-ci
+
+          $string $string-ci
           $char $one-of $none-of
           $satisfy $match1 $match1*
+          $binding $lbinding
 
           $->rope $->string $->symbol rope->string rope-finalize
           )
@@ -133,7 +137,7 @@
 ;;;  That is,
 ;;;    Parser :: [a] -> (Status, Value, [a])
 ;;;
-;;;  Error status and value
+;;;  Failure status and value
 ;;;
 ;;;    status           value
 ;;;    ------------------------------------------------
@@ -141,6 +145,7 @@
 ;;;    fail-expect      string (message), char, char-set
 ;;;    fail-unexpect    string (message)
 ;;;    fail-compound    ((Status . Value) ...)
+;;;    fail-error       (tag (Status . Value ...))
 ;;;
 ;;;  A DRIVER is a wrapper to take a parser and an input.
 ;;;  DRIVER applies the parser on the input, and on success, it returns the
@@ -155,14 +160,17 @@
 ;; An error object for the external API.  Note that a parser won't raise
 ;; this error; it just returns a 'failure' value, so that other options
 ;; can be tried if there's any.   Creating, throwing, and catching an error
-;; is expesive operation compared to the simple call/return.  It is the
-;; parser driver that sees the failure as the final result, then constructs
-;; and creates <parse-error> to raise.
+;; is expensive operation compared to the simple call/return.  It is the
+;; parser driver that sees the failure as the final result, and constructs
+;; <parse-error> to raise.
+;; NB: 'Token' slot is (car rest) or #<eof>.  It's redundant, but kept for
+;; the backward compatibility.
 (define-condition-type <parse-error> <error> #f
   (position)                            ;stream position
   (type)                                ;fail type
   (objects)                             ;expecting/unexpecting items
-  (token))                              ;token that caused error
+  (token)                               ;token that caused error
+  (rest))                               ;rest of input string, including token
 
 (define-method write-object ((o <parse-error>) out)
   (format out "#<<parse-error> ~S>" (~ o 'message)))
@@ -180,9 +188,11 @@
    (^[f r c]
      (match f
        [(_ ('quote x) v s)
-        (unless (memq x '(fail-message fail-expect fail-unexpect fail-compound))
+        (unless (memq x '(fail-message fail-expect fail-unexpect
+                                       fail-error fail-compound))
           (error "Invalid failure type; must be one of fail-message, \
-                  fail-expect, fail-unexpect or fail-compund, but got: " x))
+                  fail-expect, fail-unexpect, fail-compund or \
+                  fail-error, but got: " x))
         (quasirename r
           `(values ',x ,v ,s))]
        [(_ t v s)
@@ -193,6 +203,11 @@
 (define-inline (return-failure/expect v s)   (values 'fail-expect v s))
 (define-inline (return-failure/unexpect v s) (values 'fail-unexpect v s))
 (define-inline (return-failure/compound v s) (values 'fail-compound v s))
+(define (return-error-from-failure tag failure payload s)
+  (case failure
+    [(fail-error) (match-let1 (_ . alist) payload
+                    (values 'fail-error `(,tag . ,alist) s))]
+    [else         (values 'fail-error `(,tag (,failure . ,payload)) s)]))
 
 (define (make-peg-parse-error type objs pos seq)
   (define (flatten-compound-error objs)
@@ -202,19 +217,16 @@
                 objs))
   (define (analyze-compound-error objs pos)
     (let1 grps (map (^g (cons (caar g) (map cdr g)))
-                    (group-collection objs :key car))
-      ;; If we have explicit $fail, we just use it.  Other info is
-      ;; preserved in objects slot.
+                    (group-collection (flatten-compound-error objs) :key car))
       (let ([msgs (assoc-ref grps 'fail-message)]
             [exps (assoc-ref grps 'fail-expect)]
             [unexps (assoc-ref grps 'fail-unexpect)])
         (tree->string
-         (if msgs
-           (cons (or-concat msgs) (format " at ~s" pos))
-           (cons (or-concat (cond-list
-                             [exps (compound-exps exps)]
-                             [unexps (compound-unexps unexps)]))
-                 (format " at ~s" pos)))))))
+         (cons (or-concat (cond-list
+                           [exps (compound-exps exps)]
+                           [unexps (compound-unexps unexps)]
+                           [msgs msgs]))
+               (format " at ~a" pos))))))
   (define (or-concat lis)
     (define (rec lis)
       (match lis
@@ -232,35 +244,43 @@
     (match (delete-duplicates unexps equal?)
       [(x) (format "not expecting ~s" x)]
       [(xs ...) (format "not expecting any of ~s" xs)]))
+  (define pos-fmt
+    (if (sequence-position? pos)
+      (format "~s:~a:~a"
+              (or (sequence-position-source pos) "(unknown input)")
+              (sequence-position-line pos)
+              (sequence-position-column pos))
+      pos))
   (define (message objs pos nexttok)
     (case type
-      [(fail-message)  (format "~a at ~s" objs pos)] ;objs is a string message
+      [(fail-message)  (format "~a at ~a" objs pos-fmt)] ;objs is a string message
       [(fail-expect)
        (if (char? objs)
-         (format "expecting ~s at ~a, but got ~s" objs pos nexttok)
-         (format "expecting ~a at ~a, but got ~s" objs pos nexttok))]
+         (format "expecting ~s at ~a, but got ~s" objs pos-fmt nexttok)
+         (format "expecting ~a at ~a, but got ~s" objs pos-fmt nexttok))]
       [(fail-unexpect)
        (if (char? objs)
-         (format "expecting but ~s at ~a, and got ~s" objs pos nexttok)
-         (format "expecting but ~a at ~a, and got ~s" objs pos nexttok))]
+         (format "expecting but ~s at ~a, and got ~s" objs pos-fmt nexttok)
+         (format "expecting but ~a at ~a, and got ~s" objs pos-fmt nexttok))]
       [(fail-compound) (analyze-compound-error objs pos)]
-      [else (format "unknown parser error at ~a: ~a" pos objs)]  ;for safety
+      [(fail-error) (analyze-compound-error (cdr objs) pos)] ;car is a tag
+      [else (format "unknown parser error at ~a: ~a" pos-fmt objs)]  ;for safety
       ))
-  (let ([nexttok (if (pair? seq) (car seq) (eof-object))]
-        [objs (if (eq? type 'fail-compound)
-                (flatten-compound-error objs)
-                objs)])
+  (let ([nexttok (if (pair? seq) (car seq) (eof-object))])
     (make-condition <parse-error>
-                    'position pos 'type type 'objects objs 'token nexttok
+                    'position pos 'type type 'objects objs
+                    'token nexttok 'rest seq
                     'message (message objs pos nexttok))))
 
+(define (%get-input-pos head s)
+  (or (lseq-position s)
+      (let loop ([c 0] [head head])
+        (cond [(eq? head s) c]
+              [(null? head) "(unknown position)"]
+              [else (loop (+ c 1) (cdr head))]))))
+
 (define (construct-peg-parser-error r v s s1)
-  (make-peg-parse-error r v
-                        (let loop ([c 0] [s s])
-                          (cond [(eq? s1 s) c]
-                                [(null? s) 0] ;!!
-                                [else (loop (+ c 1) (cdr s))]))
-                        s1))
+  (make-peg-parse-error r v (%get-input-pos s s1) s1))
 
 ;; API
 ;;   Default driver.  Returns parsed value and next stream
@@ -289,7 +309,7 @@
 (define (peg-parse-string parser str :optional (cont #f))
   (check-arg string? str)
   (receive (r rest) (peg-run-parser parser (x->lseq str))
-    (if cont 
+    (if cont
       (cont r rest)
       r)))
 ;; API
@@ -312,6 +332,24 @@
                    [(eof-object? v) (set! s '()) v]
                    [else (set! s s1) (rope-finalize v)]))))))
 
+;; API
+;;  Returns an lseq
+;;  Input position info is propagated if available
+(define (peg-parser->lseq parser src)
+  (let1 s (%->lseq src)
+    (generator->lseq
+     (^[] (if (null? s)
+            (eof-object)
+            (receive (r v s1) (parser s)
+              ;; TODO: Refactor this part and similar part above
+              (let1 val (cond [(not (parse-success? r))
+                               (raise (construct-peg-parser-error r v s s1))]
+                              [(eof-object? v) (set! s '()) v]
+                              [else (set! s s1) (rope-finalize v)])
+                (if-let1 pos (lseq-position s)
+                  (values val pos)
+                  val))))))))
+
 ;;;============================================================
 ;;; Lazily-constructed string
 ;;;
@@ -352,13 +390,20 @@
 
 (define ($fail msg) (^s (return-failure/message msg s)))
 
+(define $raise
+  (case-lambda
+    [(tag msg)
+     (^s (return-failure 'fail-error `(,tag (fail-message . ,msg)) s))]
+    [(msg)
+     (^s (return-failure 'fail-error `(error (fail-message . ,msg)) s))]))
+
 ;; return a parser that tries PARSE.  On success, returns what it
 ;; returned.  On failure, returns 'fail-expect with MSG.
 (define ($expect parse msg)
   (^s (receive (r v ss) (parse s)
         (if (parse-success? r)
           (return-result v ss)
-          (return-failure/expect msg s)))))
+          (return-failure/expect msg ss)))))
 
 ;; A convenience utility to check the upper bound, allowing unlimited
 ;; upper bound by #f.
@@ -381,27 +426,71 @@
 ;; API
 ;; $lift :: (a,...) -> b, (Parser,..) -> Parser
 ;; ($lift f parser) == ($let ([x parser]) ($return (f x)))
-(define-inline ($lift f . parsers)
-  ;; We don't use the straightforward definition (using $let or $bind)
-  ;; to reduce closure construction.
-  (^s (let accum ([s s] [parsers parsers] [vs '()])
-        (if (null? parsers)
-          (return-result (apply f (reverse vs)) s)
-          (receive [r v s1] ((car parsers) s)
-            (if (parse-success? r)
-              (accum s1 (cdr parsers) (cons v vs))
-              (return-failure r v s1)))))))
+;;
+;; $lift* is like $lift, but f gets semantic values as a single list.
+;;
+;; The number of parsers is almost always fixed, and we macro-expand into
+;; chain of parser calls in such cases.
 
-;; API
-;; Like $lift, but f gets single list argument
-(define-inline ($lift* f . parsers)
-  (^s (let accum ([s s] [parsers parsers] [vs '()])
-        (if (null? parsers)
-          (return-result (f (reverse vs)) s)
-          (receive [r v s1] ((car parsers) s)
-            (if (parse-success? r)
-              (accum s1 (cdr parsers) (cons v vs))
-              (return-failure r v s1)))))))
+;; Aux function for macro expander
+(define (expand-$lift *?)
+)
+
+(define-hybrid-syntax $lift
+  (^[f . parsers]
+    ;; We don't use the straightforward definition (using $let or $bind)
+    ;; to reduce closure construction.
+    (^s (let accum ([s s] [parsers parsers] [vs '()])
+          (if (null? parsers)
+            (return-result (apply f (reverse vs)) s)
+            (receive [r v s1] ((car parsers) s)
+              (if (parse-success? r)
+                (accum s1 (cdr parsers) (cons v vs))
+                (return-failure r v s1)))))))
+  (er-macro-transformer
+   (^[f r c]
+     (define (generate-1 fn vs ps s)
+       (if (null? ps)
+         (quasirename r `(return-result (,fn ,@(reverse vs)) ,s))
+         (let* ([v1 (gensym "v")]
+                [inner (generate-1 fn (cons v1 vs) (cdr ps) s)])
+           (quasirename r `(receive [res ,v1 ,s] (,(car ps) ,s)
+                             (if (parse-success? res)
+                               ,inner
+                               (return-failure res ,v1 ,s)))))))
+     (match f
+       [(_ fn . ps)
+        (let1 s (r's)
+          (quasirename r
+            `(^[,s] ,(generate-1 fn '() ps s))))]
+       [_ f]))))
+
+(define-hybrid-syntax $lift*
+  (^[f . parsers]
+    (^s (let accum ([s s] [parsers parsers] [vs '()])
+          (if (null? parsers)
+            (return-result (f (reverse vs)) s)
+            (receive [r v s1] ((car parsers) s)
+              (if (parse-success? r)
+                (accum s1 (cdr parsers) (cons v vs))
+                (return-failure r v s1)))))))
+  (er-macro-transformer
+   (^[f r c]
+     (define (generate-1 fn vs ps s)
+       (if (null? ps)
+         (quasirename r `(return-result (,fn (list ,@(reverse vs))) ,s))
+         (let* ([v1 (gensym "v")]
+                [inner (generate-1 fn (cons v1 vs) (cdr ps) s)])
+           (quasirename r `(receive [res ,v1 ,s] (,(car ps) ,s)
+                             (if (parse-success? res)
+                               ,inner
+                               (return-failure res ,v1 ,s)))))))
+     (match f
+       [(_ fn . ps)
+        (let1 s (r's)
+          (quasirename r
+            `(^[,s] ,(generate-1 fn '() ps s))))]
+       [_ f]))))
 
 ;; API
 ;; For debugging
@@ -412,6 +501,16 @@
         (debug-print-post (list r v s))
         (values r v s))))
 
+;; API
+;; Convert failure to error
+(define $cut
+  (case-lambda
+    [(tag parser) (^s (receive (r v s) (parser s)
+                        (if (parse-success? r)
+                          (return-result v s)
+                          (return-error-from-failure tag r v s))))]
+    [(parser) ($cut 'error parser)]))
+
 ;; API
 ;; $let (bind ...) body ...
 ;;   where
@@ -424,7 +523,7 @@
    (^[f r c]
      (match f
        [(_ (bind ...) body ...)
-        (let1 vars&parsers 
+        (let1 vars&parsers
             (map (^b (match b
                        [(var parser) `(,var ,(gensym "parser") ,parser)]
                        [(parser) `(,(gensym "_") ,(gensym "parser") ,parser)]
@@ -481,23 +580,34 @@
 
 ;; API
 ;; $or p1 p2 ...
+;; $or p1 p2 ... :else pz
 ;;   Ordered choice.
+;;   When all of p1 p2 ... fail without consuming input, returns compound
+;;   failure of all of them, except in the latter form, where all the failures
+;;   before pz is discarded.  Usually, pz is ($fail ...) to give a nicer
+;;   failure message than the compound one.
 (define ($or . parsers)
   (define (fail vs s)
     (match vs
-      [((r v s)) (values r v s)] ;; no need to create compound error
-      [vs        (return-failure/compound (reverse vs) s)]))
+      [((r . v)) (values r v s)] ;; no need to create compound error
+      [vs      (return-failure/compound (reverse vs) s)]))
   (match parsers
     [()  (^s (return-failure/message "empty $or" s))]
     [(p) p]
     [(ps ...)
      (^s (let loop ([vs '()] [ps ps])
-           (if (null? ps)
-             (fail vs s)
+           (if (eq? (car ps) :else)
+             (if (null? (cdr ps))
+               (error "$or - no parsers after :else")
+               (loop '() (cdr ps)))     ;discard previous errors
              (receive (r v s1) ((car ps) s)
                (cond [(parse-success? r) (values r v s1)]
+                     [(eq? r 'fail-error) (values r v s1)]
+                     [(null? (cdr ps))
+                      (if (eq? s s1)
+                        (fail (acons r v vs) s1)
+                        (return-failure r v s1))] ; last branch consumed input
                      [(eq? s s1) (loop (acons r v vs) (cdr ps))]
-                     [(null? vs) (values r v s1)]
                      [else (fail (acons r v vs) s1)])))))]))
 
 ;; API
@@ -541,6 +651,14 @@
 (define ($seq0 parse . followers)
   (apply $lift (^[v . _] v) parse followers))
 
+;; API
+;; $list P ...
+;; $list* P ...
+;;   Same as ($lift list P ...) and ($list list* P ...), but we see them
+;;   a lot, so it's worth to have them.
+(define-inline ($list . parsers) (apply $lift list parsers))
+(define-inline ($list* . parsers) (apply $lift list* parsers))
+
 ;; API
 ;; $try parser
 ;;   Try to match parsers.  If it fails, backtrack to
@@ -551,9 +669,9 @@
 ;;   would try a, b, ... even some of them consumes the input.
 (define-inline ($try p)
   (^[s0] (receive (r v s) (p s0)
-           (if (parse-success? r)
-             (return-result v s)
-             (return-failure r v s0)))))
+           (cond [(parse-success? r) (return-result v s)]
+                 [(eq? r 'fail-error) (return-failure r v s)]
+                 [else (return-failure r v s0)]))))
 
 ;; API
 ;; $assert parser
@@ -730,14 +848,14 @@
 ;; API
 ;; $satisfy PRED EXPECT [RESULT])
 ;;   - Returns a parser such that ...
-;;   - If the head of input stream satisfies PRED, 
+;;   - If the head of input stream satisfies PRED,
 ;;     call (RESULT head (PRED head)) and let its result
 ;;     as the value of successulf parsing.
 ;;   - Otherwise, returns failure with EXPECT as the expected input.
 (define-inline ($satisfy pred expect :optional (result #f))
   (^s (if-let1 v (and (pair? s) (pred (car s)))
         (return-result (if result
-                         (result (car s) v) 
+                         (result (car s) v)
                          (car s))
                        (cdr s))
         (return-failure/expect expect s))))
@@ -749,8 +867,8 @@
 ;;   - $match1 takes one item from stream and see if it matches
 ;;     with PATTERN.
 ;;   - $match1* applys PATTERN on the entire input stream.
-;;   - If matched, RESULT is evaluated in the environment where pattern 
-;;     variables in PATTERN are bound, and its result becomes the result 
+;;   - If matched, RESULT is evaluated in the environment where pattern
+;;     variables in PATTERN are bound, and its result becomes the result
 ;;     value of the parser.
 ;;   - If RESULT is omitted, the matched item is returned.
 ;;   - If (=> FAIL) is given, FAIL (must be an identifier) is bound to
@@ -845,9 +963,168 @@
         (build-parser pat (quasirename r `(car s)) #f)]
        ))))
 
-;;;============================================================
-;;; Intermediate structure constructor
-;;;
+;; $binding  <peg-bind-expression> ... [(=> FAIL)] <body>
+;; $lbinding <peg-bind-expression> ... [(=> FAIL)] <body>
+;;   EXPERIMENTAL
+;;   <peg-bind-expression> is an expression that yields a parser,
+;;   but allowed to contain a form ($: var <peg-bind-expression>).
+;;   When inner <peg-bind-expression> accepts an input, its semantic
+;;   value is set to var.
+;;
+;;   The parser runs as if ($seq <peg-bind-expression> ...), then
+;;   if it succeeds, evaluate <body> with all such
+;;   vars to be bound.  If a branch of <peg-bind-expression> isn't
+;;   tried, vars included in it is bound to #<undef>.
+;;
+;;   $lbinding is similar to $binding, except that it behaves as if
+;;   entire expression is enclosed in $lazy.
+;;
+;;   An obvious translation is to create an enclosing scope, binds all
+;;   vars to some default value, and convert ($: var expr) to
+;;   ($lift (^x (set! var x) x) expr).  However, it causes allocation
+;;   of closures for each parser combinator every time the entire parser
+;;   is called, which incurs performance overhead.
+;;
+;;   Instead we rely on dynamic environment.  All closure allocation is
+;;   static.  The main parser sets up a vector for the storage in a parameter,
+;;   and each $: form store the value to the corresponding slot.
+;;   If match succeeds, the stored values are retrieved and bound to VARs.
+;;
+;;   If (=> FAIL) form is given, FAIL, which should be an identifier, is
+;;   bound to a closure that can be called as (FAIL message).  If you
+;;   determine that the parser should fail inside <body>, you can call FAIL
+;;   at a tail position, with a suitable message.
+;;
+;;   Limitation - since $binding walks code, it can't be statically nested.
+
+(define %binding-storage
+  (make-parameter 'accessing-binding-storage-out-of-context))
+
+(define-syntax $lbinding
+  (syntax-rules ()
+    ([_ x ...] ($lazy ($binding x ...)))))
+
+(define-syntax $binding
+  (er-macro-transformer
+   (^[f r c]
+     (define $:? (let1 $:. (r'$:) (^x (c $:. (r x)))))
+     (define =>? (let1 =>. (r'=>) (^x (c =>. (r x)))))
+     (define fail-mark (gensym))
+     ;; walk down form, gather variables and modify $: forms.
+     ;; returns modified form and list of (var . index)
+     (define (walk f vs)
+       (match f
+         [((? $:?) var expr)
+          (unless (identifier? var)
+            (error "($: var expr) form requires an identifier as var, but got:"
+                   var))
+          (receive (expr_ vs) (walk expr vs)
+            (receive (vs ind) (add-var var vs)
+              (values (quasirename r
+                        `($lift (^[x] (set! (vector-ref (%binding-storage) ,ind) x) x)
+                                ,expr_))
+                      vs)))]
+         [(xs ...) (map-accum walk vs xs)]
+         [obj (values obj vs)]))
+     (define (add-var var vs) ; returns var-alist and index
+       (if-let1 p (assq var vs)
+         (values vs (cdr p))
+         (let1 ind (length vs)
+           (values (acons var ind vs) ind))))
+     (define (build-fail-decl fail)
+       (if fail
+         (quasirename r
+           ;; NB: We can't optimize dynamically created case-lambda yet.
+           ;; `((define ,fail
+           ;;     (case-lambda
+           ;;       ((msg) (list* ',fail-mark #f msg))
+           ;;       ((tag msg) (list* ',fail-mark tag msg)))))
+           ;; With the following definition, the branch is likely
+           ;; to be optimized out when call to fail is inlined:
+           `((define (,fail a . b)
+               (if (null? b)
+                 (list* ',fail-mark #f a) ;a=msg
+                 (list* ',fail-mark a (car b)))))) ;a=tag, (car b)=msg
+         '()))
+     (define (build-return result fail tail)
+       (if fail
+         (quasirename r
+           `(let ((r ,result))
+              (if (and (pair? r) (eq? (car r) ',fail-mark))
+                (if (symbol? (cadr r))
+                  (return-error-from-failure (cadr r) 'fail-message (cddr r) s)
+                  (return-failure/message (cddr r) s))
+                (return-result r ,tail))))
+         (quasirename r
+           `(return-result ,result ,tail))))
+
+     (define (build-parser parsers vs body fail)
+       (quasirename r
+         `(let ([parser ($seq ,@parsers)])
+            (^[s]
+              ,@(build-fail-decl fail)
+              (let1 storage (make-vector ,(length vs))
+                (parameterize ((%binding-storage storage))
+                  (receive (rr v s) (parser s)
+                    (if (parse-success? rr)
+                      ,(build-return
+                        (quasirename r
+                          `(let ,(map
+                                  (^p (let ([var (car p)]
+                                            [ind (cdr p)])
+                                        (quasirename r
+                                          `(,var (vector-ref storage ,ind)))))
+                                  vs)
+                             ,body))
+                        fail
+                        (r's))
+                      (return-failure rr v s)))))))))
+
+     ;; optimization - if all the variable binding form is in the first level,
+     ;; e.g.
+     ;;   ($binding ($: x parser1)
+     ;;             ($: y parser2)
+     ;;             body)
+     ;; We can simply use $lift.
+     ;;   ($lift (lambda (x y) body) parser1 parser2)
+     ;; This saves setting up value vector.
+     ;; NB: we can't do this optimization if (=> fail) is used.
+
+     ;; check if lift optimization is possible.  note that duplicate vars
+     ;; are allowed, but that can be detected with the following, for
+     ;; vars are deduped.
+     (define (liftable? first-level-parsers vars)
+       (= (length vars)
+          (count (^p (match p
+                       [((? $:?) v _) #t]
+                       [_ #f]))
+                 first-level-parsers)))
+
+     (define (build-lift-form first-level-parsers body)
+       ;; ((var . parser) ...).
+       (define vars&parsers (map (^p (match p
+                                       [((? $:?) v p) `(,v . ,p)]
+                                       [p `(,(gensym) . ,p)]))
+                                 first-level-parsers))
+       (quasirename r
+         `($lift (lambda ,(map car vars&parsers) ,body)
+                 ,@(map cdr vars&parsers))))
+
+     (define (bad) (error "Malformed $binding:" f))
+
+     (let loop ([forms (cdr f)] [ps '()])
+       (match forms
+         [(((? =>?) fail) body)
+          (receive (parsers. vars) (walk (reverse ps) '())
+            (build-parser parsers. vars body fail))]
+         [(((? =>?) fail) body ...) (bad)]
+         [(body)
+          (receive (parsers. vars) (walk (reverse ps) '())
+            (if (liftable? ps vars)
+              (build-lift-form (reverse ps) body)
+              (build-parser parsers. vars body #f)))]
+         [(parser . rest) (loop rest (cons parser ps))]
+         [() (bad)])))))
 
 ;;;============================================================
 ;;; String parsers
@@ -906,28 +1183,16 @@
                    (char-ci=? (car s) (car lis)))
             (loop (cons (car s) r) (cdr s) (cdr lis))
             (return-failure/expect str s0)))))))
-   
+
 (define-inline ($char c)
-  (assume-type c <char>) 
+  (assume-type c <char>)
   ($satisfy (cut eqv? c <>) c))
 
 (define ($char-ci c)
-  (assume-type c <char>) 
+  (assume-type c <char>)
   ($satisfy (^x (and (char? x) (char-ci=? c x)))
             (list->char-set c (char-upcase c) (char-downcase c))))
 
-(define ($one-of charset)
-  (assume-type charset <char-set>)
-  ($satisfy (^x (and (char? x) (char-set-contains? charset x)))
-            charset))
-
-(define ($symbol sym) 
-  (assume-type sym <symbol>)
-  ($seq ($string (symbol->string sym)) ($return sym)))
-
-(define ($none-of charset)
-  ($one-of (char-set-complement charset)))
-
 ;; Anything except end of stream.
 (define-inline ($any)
   (^s (if (pair? s)
@@ -945,6 +1210,37 @@
    [(char-set? item) ($one-of item)]
    [(char? item) ($char item)]
    [(string? item) ($string item)]
-   [(symbol? item) ($symbol item)]
+   [(symbol? item) ($satisfy (cut eq? <> item) item)]
    [else (error "Bad item: $. requires a char, a char-set, a string, \
                  or a symbol, but got:" item)]))
+
+(define-hybrid-syntax $one-of
+  (^[items]
+    (cond
+     [(char-set? items)
+      ($satisfy (^x (and (char? x) (char-set-contains? items x)))
+                items)]
+     [(list? items) ($expect (apply $or (map $. items)) items)]
+     [else (error "$one-of requires a charset or a list of items, \
+                   but got:" items)]))
+  (er-macro-transformer
+   (^[f r c]
+     (match f
+       [(_ (? char-set? items))
+        (quasirename r
+          `($satisfy (^[x] (and (char? x) (char-set-contains? ,items x)))
+                     ',items))]
+       [(_ ('quote (items ...)))
+        (if (every (^x (or (char? x) (symbol? x))) items)
+          (quasirename r
+            `($satisfy (^[x] (memv x ',items)) ',items))
+          f)]
+       [_ f]))))
+
+;; NB: $none-of can't be generalized like $one-of, for the value to return
+;; on success is ambiguous except the case we're matching a single character.
+;; The asymmetry doesn't look good... we may drop $none-of in future.
+(define ($none-of charset)
+  (assume-type charset <char-set>)
+  ($satisfy (^x (or (not (char? x)) (not (char-set-contains? charset x))))
+            `(none-of ,charset)))
