@@ -32,9 +32,10 @@
 ;;;
 
 (use gauche.test)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (use gauche.generator)
+(use gauche.lazy)
 
 (test-start "parser.peg")
 
@@ -114,8 +115,16 @@
 (test-fail "$char" '(0 #\a) ($char #\a) "012")
 (test-succ "$one-of" #\j ($one-of #[a-z]) "j")
 (test-fail "$one-of" '(0 #[A-Z]) ($one-of #[A-Z]) "j")
+(test-succ "$one-of" "abd" ($one-of '("abc" "abd" "acd")) "abd")
+(test-succ "$one-of" #\c ($one-of '(#\a #\b #\c)) "c")
+(test-succ "$one-of" #\c ($one-of '("ab" #\c)) "c")
+(test-fail "$one-of" '(0 (#\a #\b #\c)) ($one-of '(#\a #\b #\c)) "z")
+(test-fail "$one-of" '(0 ("abc" "abd" "acd"))
+           ($one-of '("abc" "abd" "acd")) "a")
+(let ((var #f))
+  (set! var '("a" "b" "c")) ;suppress optimization
+  (test-succ "$one-of" "b" ($one-of var) "b"))
 (test-succ "$none-of" #\j ($none-of #[A-Z]) "j")
-(test-fail "$none-of" '(0 #[^a-z]) ($none-of #[a-z]) "j")
 (test-succ "$string-ci" "aBC" ($string-ci "abc") "aBCdef")
 (test-fail "$string-ci" '(0 "abc") ($string-ci "abc") "012")
 
@@ -143,6 +152,10 @@
            ($or ($string "foo") ($string "bar"))
            "012345")
 
+(test-fail "$or" '(0 "bar")
+           ($or ($string "foo") :else ($string "bar"))
+           "012345")
+
 ;; $try
 (test-succ "$try" "foo"
            ($try ($string "foo"))
@@ -157,6 +170,28 @@
            ($try ($seq ($string "foo") ($string "bar")))
            "foobaz...")
 
+;; $cut and $raise
+(test-fail "$cut" '(0 (error (fail-expect . "a")))
+           ($or ($cut ($. "a")) ($. "b"))
+           "b")
+(test-succ "$cut" "a"
+           ($or ($cut ($. "a")) ($. "b"))
+           "a")
+
+(test-fail "$cut" '(1 (error (fail-expect . "b")))
+           ($or ($try ($seq ($. "a") ($cut ($. "b")))) ($. "a"))
+           "ac")
+(test-succ "$cut" "b"
+           ($or ($try ($seq ($. "a") ($cut ($. "b")))) ($. "a"))
+           "ab")
+
+(test-fail "$raise" '(1 (error (fail-message . "oops")))
+           ($or ($seq ($. "a") ($raise "oops")) ($. "a"))
+           "a")
+(test-fail "$raise" '(1 (error (fail-message . "oops")))
+           ($or ($try ($seq ($. "a") ($raise "oops"))) ($. "a"))
+           "a")
+
 ;; $let, $let*
 (test-succ "$let" "j"
            ($let ([j ($string "j")])
@@ -236,19 +271,27 @@
            "lineament")
 
 ;; $lift and $lift*
-(test-succ "$lift" '(#\a . #\b)
-           ($lift cons ($any) ($any))
-           "abc")
-(test-fail "$lift" '(1 #\z)
-           ($lift cons ($any) ($char #\z))
-           "abc")
+;; NB: They are inline expanded, so we also call via apply
+(let ()
+  (define-syntax test-lift
+    (syntax-rules ()
+      [(test-lift tester name expect (exp ...) input)
+       (begin (tester name expect (exp ...) input)
+              (tester #"~|name| via apply"
+                      expect (apply exp ... '()) input))]))
 
-(test-succ "$lift*" "abc"
-           ($lift* list->string ($any) ($any) ($any))
-           "abc")
-(test-fail "$lift" '(2 #\z)
-           ($lift* list->string ($any) ($any) ($char #\z))
-           "abc")
+  (test-lift test-succ "$lift" '(#\a . #\b)
+             ($lift cons ($any) ($any))
+             "abc")
+  (test-lift test-fail "$lift" '(1 #\z)
+             ($lift cons ($any) ($char #\z))
+             "abc")
+  (test-lift test-succ "$lift*" "abc"
+             ($lift* list->string ($any) ($any) ($any))
+             "abc")
+  (test-lift test-fail "$lift" '(2 #\z)
+             ($lift* list->string ($any) ($any) ($char #\z))
+             "abc"))
 
 ;; $fold-parsers and $fold-parsers-right
 (test-succ "$fold-parsers" '()                  ; base case
@@ -445,7 +488,7 @@
 (test-section "backtrack control")
 
 (test-succ "$or and $try" "abc"
-           ($or ($try ($lift list ($string "abc") ($string "foo")))
+           ($or ($try ($list ($string "abc") ($string "foo")))
                 ($string "abc"))
            "abcdefg")
 
@@ -466,8 +509,8 @@
 ;;;
 (test-section "lookahead assertion")
 
-(let ([parser ($seq0 
-               ($->rope ($many 
+(let ([parser ($seq0
+               ($->rope ($many
                          ($try ($seq0 ($any)
                                              ($assert ($. #[a-z]))))))
                ($any))])
@@ -500,6 +543,23 @@
          (test-error <parse-error> "expecting one of (#\\c #\\a #\\b) at 1")
          (peg-parse-string p2 "dd")))
 
+;; position
+(test* "error position handling"
+       (test-error <parse-error>
+                   "expecting #[\\u000aa] at \"input.txt\":2:3, but got #\\b")
+       (peg-run-parser ($many ($. #[a\n]) 10)
+                       (port->char-lseq/position
+                        (open-input-string "aaaaa\naabaa")
+                        :source-name "input.txt")))
+
+(test* "error position handling (<parse-error> slots)"
+       #t
+       (guard (e [(<parse-error> e) (sequence-position? (~ e'position))])
+         (peg-run-parser ($many ($. #[a\n]) 10)
+                         (port->char-lseq/position
+                          (open-input-string "aaaaa\naabaa")
+                          :source-name "input.txt"))))
+
 ;;;============================================================
 ;;; Optional arg for peg-parse-string and peg-parse-port
 ;;;
@@ -512,6 +572,23 @@
        (call-with-input-string "aaaabbc"
          (cut peg-parse-port ($->rope ($many ($. #\a))) <> cons)))
 
+;;;============================================================
+;;; Token parsing
+;;;
+
+(test-section "token parsing")
+
+(let ()
+  (define %a ($many1 ($. 'a)))
+  (define %bcd ($many1 ($one-of '(b c d))))
+  (define %p ($many ($list %a %bcd)))
+
+  (test* "token parsing"
+         '(((a) (b c)) ((a) (d)) ((a a) (b)) ((a) (c d c)))
+         (peg-run-parser %p '(a b c a d a a b a c d c)))
+  )
+
+
 ;;;============================================================
 ;;; Examples
 ;;;
@@ -535,9 +612,9 @@
   (define integer
     ($let ([ds digits])
       ($return (x->integer (list->string ds)))))
-  (define integers1 
+  (define integers1
     ($seq integer ($many ($seq separator integer))))
-  (define integers2 
+  (define integers2
     ($let ([n  integer]
            [ns ($many ($seq separator integer))])
       ($return (cons n ns))))
@@ -556,14 +633,14 @@
            ($return (cons n ns)))
          ($return '())))
   (define integers5 (sep-by integer separator))
-  (define begin-list 
+  (define begin-list
     ($seq0 ($. #[\(\[\{]) ws))
-  (define (end-list opener) 
+  (define (end-list opener)
     ($seq ws (case opener
                [(#\() ($. #\))]
                [(#\[) ($. #\])]
                [(#\{) ($. #\})])))
-  
+
   (define int-list
     ($let* ([opener begin-list]
             [ints ($sep-by integer separator)]
@@ -609,14 +686,14 @@
   (define paren  ($. #\())
   (define thesis ($. #\)))
 
-  (test* "$or example" 
+  (test* "$or example"
          (test-error <parse-error> "expecting ab at 1, but got #\\c")
 
          (peg-parse-string ($or ($seq paren ($."ab") thesis)
                                 ($seq paren ($."cd") thesis))
                            "(cd)"))
   )
-  
+
 (let ()
   (define integer   ($many1 ($. #[\d])))
   (define ws        ($many_ ($. #[\s])))
@@ -721,21 +798,21 @@
              records "\"a  \"\" \n\" , b  , c"))
 
 ;; Poor-man's XML parser
-(letrec ([open-tag ($->rope ($between ($char #\<)
-                                      ($one-of #[[:alnum:]])
-                                      ($char #\>)))]
-         [close-tag (^[tagname]
-                      ($seq ($string "</") ($string tagname) ($char #\>)))]
-         [text ($->rope ($many ($none-of #[<])))]
-         [body ($let ([t text]
-                      [r ($many ($let ([e element]
-                                       [t text])
-                                  ($return (list e t))))])
-                 ($return `(,t ,@(apply append r))))]
-         [element ($let* ([tagname ($try open-tag)]
-                          [body body]
-                          [ (close-tag (rope-finalize tagname)) ])
-                    ($return (cons tagname body)))])
+(letrec* ([open-tag ($->rope ($between ($char #\<)
+                                       ($one-of #[[:alnum:]])
+                                       ($char #\>)))]
+          [close-tag (^[tagname]
+                       ($seq ($string "</") ($string tagname) ($char #\>)))]
+          [text ($->rope ($many ($none-of #[<])))]
+          [body ($let ([t text]
+                       [r ($many ($let ([e ($lazy element)]
+                                        [t text])
+                                       ($return (list e t))))])
+                      ($return `(,t ,@(apply append r))))]
+          [element ($let* ([tagname ($try open-tag)]
+                           [body body]
+                           [ (close-tag (rope-finalize tagname)) ])
+                          ($return (cons tagname body)))])
   (test-succ "tag element" '("a" "")
              element "<a></a>")
   (test-succ "tag element" '("a" "foo" ("b" "bar") "baz")
@@ -762,7 +839,7 @@
 
 (test-section "deprecated")
 ;; kludge - we can't say (use parser.peg.deprecated) until peg/deprecated.scm
-;; is installed in place.  
+;; is installed in place.
 (or (load "peg/deprecated" :error-if-not-found #f)
     (load "parser/peg/deprecated" :error-if-not-found #f))
 (import parser.peg.deprecated)
@@ -833,10 +910,65 @@
                   (if (eqv? x #\0)
                     'ok
                     (F "0 expected")))))
-  (test-succ "$match1 fail proc" 'ok p "012") 
+  (test-succ "$match1 fail proc" 'ok p "012")
   (test-fail "$match1 fail proc" '(0 "0 expected") p "123")
   (test-fail "$match1 not enough input" '(0 "(? char-numeric? x)") p ""))
 
+
+(let ((p ($binding ($: x ($or ($seq ($: y ($."a")) ($."b"))
+                              ($seq ($: z ($."c")) ($."d"))))
+                   (list x y z))))
+  (test-succ "$binding" `("b" "a" ,(undefined)) p "ab")
+  (test-succ "$binding" `("d" ,(undefined) "c") p "cd")
+  (test-fail "$binding" '(1 "b") p "az")
+  (test-fail "$binding" '(1 "d") p "cy")
+  (test-fail "$binding" '(0 ((fail-expect . "a") (fail-expect . "c"))) p "pq"))
+
+(let ((p ($binding ($: x ($or ($seq ($: y ($."a")) ($."b"))
+                              ($seq ($: y ($."c")) ($."d"))))
+                   (list x y))))
+  (test-succ "$binding" `("b" "a") p "ab")
+  (test-succ "$binding" `("d" "c") p "cd")
+  (test-fail "$binding" '(1 "d") p "cy"))
+
+(let ((p ($binding ($: x ($or ($seq ($: y ($."a")) ($."b"))
+                              ($seq ($: y ($."c")) ($."d"))))
+                   (=> F)
+                   (if (equal? y "c")
+                     (F "no-no")
+                     (list x y)))))
+  (test-succ "$binding fail proc" `("b" "a") p "ab")
+  (test-fail "$binding fail proc" `(2 "no-no") p "cd")
+  (test-fail "$binding fail proc" '(1 "d") p "cy"))
+
+(let ((p ($or ($binding ($: x ($. "ab"))
+                        (=> F)
+                        (F 'error "ugh"))
+              ($. "ab"))))
+  (test-fail "$binding fail/error" '(2 (error (fail-message . "ugh")))
+             p "ab"))
+
+(letrec ([p (lambda (q)
+              ($binding ($. "a")
+                        ($: y q)
+                        ($seq ($: x ($any)) ($. "."))
+                        ($. "z")
+                        (list x y)))]
+         [q ($binding ($many ($or ($: x ($. "b")) ($: y ($. "c"))))
+                      (list x y))])
+  (test-succ "$binding dynamically nested" '(#\d ("b" "c"))
+             (p q) "abcbcd.z")
+  )
+
+(test-succ "$binding -> lift optimization" '("b" "d")
+           ($binding ($. "a")
+                     ($: b ($. "b"))
+                     ($. "c")
+                     ($: d ($. "d"))
+                     ($. "e")
+                     (list b d))
+           "abcde")
+
 (test-succ "eof" (eof-object) eof "")
 (test-fail "eof" '(0 "end of input") eof "a")
 
@@ -865,6 +997,8 @@
   (t "{\"x\":null}" 'null)
   (t "{\"x\": \"abc\\\"\\\\\\/\\b\\f\\n\\r\\t\\u0040abc\"}"
      "abc\"\\/\u0008\u000c\u000a\u000d\u0009@abc")
+  (t "{\"x\": \"\\ud800\\udc01\"}"
+     "\x10001;")
   )
 
 (let ()
@@ -875,6 +1009,14 @@
   (t "{x : 100}}")
   )
 
+(let ()
+  (define (t str)
+    (test* #"parse error (unpaired surrogate) ~str"
+           (test-error <json-parse-error> #/unpaired (low|high) surrogate/)
+           (parse-json-string str)))
+  (t "{\"a\":\"\\ud800abc\"}")
+  (t "{\"a\":\"abc\\ude01xyz\"}"))
+
 (test* "parsing an object"
        '(("Image"
           ("Width"  . 800)
@@ -991,22 +1133,19 @@
                    ("Country"   . "US"))))
   )
 
-(cond-expand
- [gauche.ces.utf8
-  (let1 data `(("[\"\\u03bb\"]" #("\x3bb;"))
-               ("[\"\\ud800\"]" ,(test-error <json-parse-error>))
-               ("[\"\\ud867\\ude3d\\u03bb\"]" #("\x29e3d;\x3bb;"))
-               ("[\"\\ude3d\\ud867\"]" ,(test-error <json-parse-error>))
-               ("[\"\\uf020\\u03bb\"]"  #("\xf020;\x3bb;")))
-    (dolist [d data]
-      (test* (format "unicode escape reading (~s)" (car d))
-             (cadr d)
-             (parse-json-string (car d)))
-      (when (vector? (cadr data))
-        (test* (format "unicode escape writing (~s)" (cadr d))
-               (car d)
-               (construct-json-string (cadr d))))))]
- [else])
+(let1 data `(("[\"\\u03bb\"]" #("\x3bb;"))
+             ("[\"\\ud800\"]" ,(test-error <json-parse-error>))
+             ("[\"\\ud867\\ude3d\\u03bb\"]" #("\x29e3d;\x3bb;"))
+             ("[\"\\ude3d\\ud867\"]" ,(test-error <json-parse-error>))
+             ("[\"\\uf020\\u03bb\"]"  #("\xf020;\x3bb;")))
+  (dolist [d data]
+    (test* (format "unicode escape reading (~s)" (car d))
+           (cadr d)
+           (parse-json-string (car d)))
+    (when (vector? (cadr data))
+      (test* (format "unicode escape writing (~s)" (cadr d))
+             (car d)
+             (construct-json-string (cadr d))))))
 
 (let ()
   (define (t obj)
@@ -1042,5 +1181,5 @@
          (parse-json-string "{\"x\":123}")))
 
 (include "test-srfi-180")
-       
+
 (test-end)
