@@ -4,7 +4,7 @@
 
 (use gauche.test)
 (use util.match)
-(use srfi-19)
+(use srfi.19)
 (use gauche.sequence)
 
 (test-start "precompiled rfc modules")
@@ -56,6 +56,23 @@ Content-Length: 4349
        (equal? rfc822-header1-list
                (rfc822-read-headers (open-input-string rfc822-header1))))
 
+(test* "rfc822-header-ref" "Bogus Tester"
+       (rfc822-header-ref rfc822-header1-list "subject"))
+(test* "rfc822-header-ref" "Missing"
+       (rfc822-header-ref rfc822-header1-list "cc" "Missing"))
+(test* "rfc822-header-ref"
+       "by foo.bar.com id ZZZ55555; Thu, 31 May 2001 16:38:04 -1000 (HST)"
+       (rfc822-header-ref rfc822-header1-list "received"))
+(test* "rfc822-header-ref*"
+       '("by foo.bar.com id ZZZ55555; Thu, 31 May 2001 16:38:04 -1000 (HST)"
+         "from ooo.ooo.com (ooo.ooo.com [1.2.3.4])\tby foo.bar.com (9.9.9+3.2W/3.7W-) with ESMTP id ZZZ55555\tfor <yoo@bar.com>; Thu, 31 May 2001 16:38:02 -1000 (HST)"
+         "from zzz ([1.2.3.5]) by ooo.ooo.com  with Maccrosoft SMTPSVC(5.5.1877.197.19);\t Thu, 31 May 2001 22:33:16 -0400")
+       (rfc822-header-ref* rfc822-header1-list "received"))
+(test* "rfc822-header-put" '("Redacted")
+       (rfc822-header-ref*
+        (rfc822-header-put rfc822-header1-list "received" "Redacted")
+        "received"))
+
 ;; token parsers
 (test* "rfc822-field->tokens (basic)"
        '(("aa") ("bb") ("cc") ("dd") ("ee") (" a\"aa\\aa (a)"))
@@ -190,7 +207,7 @@ Content-Length: 4349
                  (begin
                    (test* (format "rfc822-write-headers error (~a)" expect)
                           expect
-                          (guard (e (else (rxmatch-case 
+                          (guard (e (else (rxmatch-case
                                               (condition-message e "")
                                             [#/\(([\w-]+)\)/ (_ m) m]
                                             [else e])))
@@ -210,7 +227,7 @@ Content-Length: 4349
   (test-reason "stray-crlf" '(("name" "Shiro\nKawai")))
   (test-reason "line-too-long" `(("name" ,(make-string 1000 #\a))))
   )
-        
+
 ;;--------------------------------------------------------------------
 (test-section "rfc.mime")
 (use rfc.mime)
@@ -343,8 +360,7 @@ Content-Length: 4349
                            :line-width 30))
   )
 
-(when (and (memq (gauche-character-encoding) '(euc-jp sjis utf-8))
-           (ces-conversion-supported? "iso-2022-jp" #f))
+(when (ces-conversion-supported? "iso-2022-jp" #f)
   (test* "mime-decode-word" "\u5ddd\u5408 \u53f2\u6717"
          (mime-decode-word "=?ISO-2022-JP?B?GyRCQG45ZxsoQiAbJEI7S08vGyhC?="))
   (test* "mime-decode-text" "(\u5ddd\u5408 \u53f2\u6717)"
@@ -354,13 +370,11 @@ Content-Length: 4349
                            :charset 'iso-2022-jp))
   )
 
-;; this tests whether illegal input sequence is handled gracefully
-(when (memq (gauche-character-encoding) '(euc-jp sjis utf-8))
-  (test* "mime-decode-word" "=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?="
-         (mime-decode-word "=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?="))
-  (test* "mime-decode-text" "(=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?=)"
-         (mime-decode-text "(=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?=)"))
-  )
+;; these tests whether illegal input sequence is handled gracefully
+(test* "mime-decode-word" "=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?="
+       (mime-decode-word "=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?="))
+(test* "mime-decode-text" "(=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?=)"
+       (mime-decode-text "(=?ISO-2022-JP?B?GyRCJDkbKBsoQg==?=)"))
 
 ;; NB: this assumes the test is run in-place
 (define (mime-message-tester num headers)
@@ -450,7 +464,7 @@ Content-Length: 4349
                        ("content-type" ,#"multipart/mixed; boundary=\"~|boundary|\""))
                      (cut mime-body->string <> <>)))
               #f)))))
-                     
+
 (dotimes (n 8) (mime-roundtrip-tester n))
-    
+
 (test-end)
