@@ -1,7 +1,7 @@
 ;;;
 ;;; data.sparse - sparse data structures
 ;;;
-;;;   Copyright (c) 2007-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2007-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -40,7 +40,8 @@
           sparse-table-clear! sparse-table-delete! sparse-table-copy
           sparse-table-update! sparse-table-push! sparse-table-pop!
           sparse-table-fold sparse-table-map sparse-table-for-each
-          sparse-table-keys sparse-table-values sparse-table-comparator
+          sparse-table-find sparse-table-keys sparse-table-values
+          sparse-table-comparator
           %sparse-table-dump %sparse-table-check
 
           <sparse-vector-base> <sparse-vector> <sparse-s8vector>
@@ -55,7 +56,7 @@
           sparse-vector-update! sparse-vector-inc!
           sparse-vector-push! sparse-vector-pop!
           sparse-vector-fold sparse-vector-map sparse-vector-for-each
-          sparse-vector-keys sparse-vector-values
+          sparse-vector-find sparse-vector-keys sparse-vector-values
           %sparse-vector-dump
 
           <sparse-matrix-base> <sparse-matrix> <sparse-s8matrix>
@@ -85,6 +86,7 @@
   (let ([x-fold     (string->symbol #"~|type|-fold")]
         [x-map      (string->symbol #"~|type|-map")]
         [x-for-each (string->symbol #"~|type|-for-each")]
+        [x-find     (string->symbol #"~|type|-find")]
         [x-keys     (string->symbol #"~|type|-keys")]
         [x-values   (string->symbol #"~|type|-values")]
         [x-update!  (string->symbol #"~|type|-update!")]
@@ -104,6 +106,14 @@
          (,x-fold st (^[k v s] (cons (proc k v) s)) '()))
        (define (,x-for-each st proc)
          (,x-fold st (^[k v _] (proc k v)) #f))
+       (define (,x-find pred st :optional (failure (lambda () (values #f #f))))
+         (let ([iter (,iter st)]
+               [end (list #f)])
+           (let loop ()
+             (receive (key val) (iter end)
+               (cond [(eq? key end) (failure)]
+                     [(pred key val) (values key val)]
+                     [else (loop)])))))
        (define (,x-keys st)
          (,x-fold st (^[k v s] (cons k s)) '()))
        (define (,x-values st)
@@ -147,7 +157,7 @@
 (inline-stub
  (initcode "Scm_Init_sptab(Scm_CurrentModule());")
 
- (define-type <sparse-table> "SparseTable*" "sparse table"
+ (declare-stub-type <sparse-table> "SparseTable*" "sparse table"
    "SPARSE_TABLE_P" "SPARSE_TABLE")
 
  (define-cproc %make-sparse-table (type cmpr::<comparator>)
@@ -162,7 +172,7 @@
 
  (define-cproc sparse-table-comparator (st::<sparse-table>)
    (return (SCM_OBJ (-> st comparator))))
- 
+
  (define-cproc sparse-table-num-entries (st::<sparse-table>) ::<ulong>
    (return (-> st numEntries)))
 
@@ -188,8 +198,7 @@
 
  (define-cproc sparse-table-copy (sv::<sparse-table>) SparseTableCopy)
 
- (define-cfn sparse-table-iter (args::ScmObj* nargs::int data::void*) :static
-   (cast void nargs)                    ; suppress unused var warning
+ (define-cfn sparse-table-iter (args::ScmObj* _::int data::void*) :static
    (let* ([iter::SparseTableIter* (cast SparseTableIter* data)]
           [r (SparseTableIterNext iter)]
           [eofval (aref args 0)])
@@ -258,11 +267,10 @@
 (inline-stub
  (initcode "Scm_Init_spvec(Scm_CurrentModule());")
 
- (define-type <sparse-vector> "SparseVector*" "sparse vector"
+ (declare-stub-type <sparse-vector> "SparseVector*" "sparse vector"
    "SPARSE_VECTOR_BASE_P" "SPARSE_VECTOR")
 
- (define-cproc %make-sparse-vector (type default-value flags::<ulong>)
-   (cast void flags)                     ; suppress unused var warning
+ (define-cproc %make-sparse-vector (type default-value _::<ulong>)
    (let* ([klass::ScmClass* NULL])
      (cond [(SCM_CLASSP type)  (set! klass (SCM_CLASS type))]
            [(SCM_FALSEP type)  (set! klass SCM_CLASS_SPARSE_VECTOR)]
@@ -336,8 +344,7 @@
                                    :optional fallback)
    SparseVectorInc)
 
- (define-cfn sparse-vector-iter (args::ScmObj* nargs::int data::void*) :static
-   (cast void nargs)                    ; suppress unused var warning
+ (define-cfn sparse-vector-iter (args::ScmObj* _::int data::void*) :static
    (let* ([iter::SparseVectorIter* (cast SparseVectorIter* data)]
           [r (SparseVectorIterNext iter)]
           [eofval (aref args 0)])
@@ -387,7 +394,7 @@
 (define-refset <sparse-f64vector>)
 
 ;; sparse vector comparator is just a singleton.
-(define *sparse-vector-comparator* 
+(define *sparse-vector-comparator*
   (make-comparator/compare (^x (and (exact? x) (integer? x) (>= x 0)))
                            eqv? compare eqv-hash
                            'sparse-vector-comparator))
@@ -406,7 +413,7 @@
 ;; Then we don't need to worry about the range of each index.
 
 (inline-stub
-  (define-type <sparse-matrix> "SparseVector*" "sparse matrix"
+  (declare-stub-type <sparse-matrix> "SparseVector*" "sparse matrix"
     "SPARSE_MATRIX_BASE_P" "SPARSE_MATRIX")
 
   (.define X_OOR 1)
@@ -417,7 +424,7 @@
      `(if (== ,oor NULL)
         (Scm_Error ,#"~|which| index is out of range: %S" ,which)
         (begin (set! (* ,oor) ,oorval) (return 0)))])
-  
+
   ;; Calculate linear index from 2d indexes (x,y) by interleaving the bits.
   ;; When the input is out of range, if OOR is provided it is set;
   ;; otherwise an error is signaled.
@@ -428,7 +435,7 @@
   ;; So we adopt 4 bits interleaving instead.
   (.define INTERLEAVE_SHIFT 4)
   (.define INTERLEAVE_MASK (- (<< 1UL INTERLEAVE_SHIFT) 1))
-  
+
   (define-cfn index-combine-2d (x y oor::int*)
     ::u_long :static
     (unless (SCM_INTEGERP x)
@@ -546,8 +553,7 @@
   (return (SparseVectorInc sv (index-combine-2d x y NULL) delta fallback)))
 
 (inline-stub
- (define-cfn sparse-matrix-iter (args::ScmObj* nargs::int data::void*) :static
-   (cast void nargs)                    ; suppress unused var warning
+ (define-cfn sparse-matrix-iter (args::ScmObj* _::int data::void*) :static
    (let* ([iter::SparseVectorIter* (cast SparseVectorIter* data)]
           [r (SparseVectorIterNext iter)]
           [eofval (aref args 0)])
@@ -565,7 +571,7 @@
     (SparseVectorIterInit iter sv)
     (return
      (Scm_MakeSubr sparse-matrix-iter iter 1 0 '"sparse-matrix-iterator"))))
-       
+
 (define-cproc sparse-matrix-update! (sv::<sparse-matrix> x y proc
                                                          :optional fallback)
   ;; NB: We use j to pass the index to continuation proc, for it will be
@@ -624,7 +630,7 @@
 
 (define (sparse-matrix-values sv)
   (sparse-matrix-fold sv (^[x y v s] (cons v s)) '()))
-  
+
 ;;===============================================================
 ;; protocols
 ;;
@@ -658,3 +664,5 @@
   :push!     sparse-vector-push!
   :update!   sparse-vector-update!)
 
+(define-method size-of ((s <sparse-table>)) (sparse-table-num-entries s))
+(define-method size-of ((s <sparse-vector-base>)) (sparse-vector-num-entries s))
