@@ -1,7 +1,7 @@
 ;;;
 ;;; termios - termios interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,22 +33,58 @@
 
 (define-module gauche.termios
   (use gauche.uvector)
-  (export-all)
+  (export has-windows-console? without-echoing with-terminal-mode)
   )
 (select-module gauche.termios)
 
-(inline-stub
- (declcode
-  (.include "gauche-termios.h"))
+(without-precompiling
+ (cond-expand
+  [gauche.os.windows
+   (export msys-get-stty)]
+  [else
+   (export-if-defined
+    <sys-termios-meta> <sys-termios> B0 B110 B115200 B1200 B134 B150
+    B1800 B19200 B200 B230400 B2400 B300 B38400 B4800 B50 B57600
+    B600 B75 B9600 BRKINT BS0 BS1 BSDLY CIBAUD CLOCAL CR0 CR1 CR2
+    CR3 CRDLY CREAD CRTSCTS CS5 CS6 CS7 CS8 CSIZE CSTOPB ECHO
+    ECHOCTL ECHOE ECHOK ECHOKE ECHONL ECHOPRT FF0 FF1 FFDLY HUPCL
+    ICANON ICRNL IEXTEN IGNBRK IGNCR IGNPAR IMAXBEL INLCR INPCK
+    ISIG ISTRIP IUCLC IXANY IXOFF IXON NCCS NL0 NL1 NLDLY NOFLSH
+    OCRNL OFDEL OFILL OLCUC ONLCR ONLRET ONOCR OPOST PARENB PARODD
+    PENDIN TCIFLUSH TCIOFF TCIOFLUSH TCION TCOFLUSH TCOOFF TCOON
+    TCSADRAIN TCSAFLUSH TCSANOW TOSTOP VDISCARD VEOF VEOL VEOL2
+    VERASE VINTR VKILL VLNEXT VMIN VQUIT VREPRINT VSTART VSTOP
+    VSUSP VSWTC VT0 VT1 VTDLY VTIME VWERASE XCASE _POSIX_VDISABLE
+    sys-cfgetispeed sys-cfgetospeed
+    sys-cfsetispeed sys-cfsetospeed sys-forkpty
+    sys-forkpty-and-exec sys-openpty sys-tcdrain sys-tcflow
+    sys-tcflush sys-tcgetattr sys-tcgetpgrp sys-tcsendbreak
+    sys-tcsetattr sys-tcsetpgrp sys-termios-copy)])
+ )
 
  ;;---------------------------------------------------------------------
  ;; termios.h
 
- (define-type <sys-termios> "ScmSysTermios*")
- (define-type <sys-sigset>  "ScmSysSigset*")
-
+(inline-stub
  (.unless (defined GAUCHE_WINDOWS)
 
+   (declcode
+    (.include <gauche/priv/portP.h>)
+    (.include <termios.h>)
+    (.when (defined HAVE_PTY_H)
+      (.include <pty.h>))
+    (.when (defined HAVE_UTIL_H)
+      (.include <util.h>))
+    ;; libutil.h is superseded by bsd/libutil.h
+    (.if (defined HAVE_BSD_LIBUTIL_H)
+         (.include <bsd/libutil.h>)
+         (.if (defined HAVE_LIBUTIL_H)
+              (.include <libutil.h>)))
+    (.when (defined HAVE_UNISTD_H)
+      (.include <unistd.h>)))
+
+   ;; Constants for termios.  Non-POSIX symbols are defined conditionally.
+
    (define-enum TCSANOW)
    (define-enum TCSADRAIN)
    (define-enum TCSAFLUSH)
@@ -60,6 +96,109 @@
    (define-enum TCIOFF)
    (define-enum TCION)
 
+   ;; c_iflag masks
+   (define-enum IGNBRK)
+   (define-enum BRKINT)
+   (define-enum IGNPAR)
+   (define-enum INPCK)
+   (define-enum ISTRIP)
+   (define-enum INLCR)
+   (define-enum IGNCR)
+   (define-enum ICRNL)
+   (define-enum IXON)
+   (define-enum IXOFF)
+   (define-enum-conditionally IXANY)
+   (define-enum-conditionally IUCLC)
+   (define-enum-conditionally IMAXBEL)
+
+   ;; c_oflag masks
+   (define-enum OPOST)
+   (define-enum-conditionally OLCUC)
+   (define-enum-conditionally ONLCR)
+   (define-enum-conditionally OCRNL)
+   (define-enum-conditionally ONOCR)
+   (define-enum-conditionally ONLRET)
+   (define-enum-conditionally OFILL)
+   (define-enum-conditionally OFDEL)
+   (define-enum-conditionally NLDLY)
+   (define-enum-conditionally NL0)
+   (define-enum-conditionally NL1)
+   (define-enum-conditionally CRDLY)
+   (define-enum-conditionally CR0)
+   (define-enum-conditionally CR1)
+   (define-enum-conditionally CR2)
+   (define-enum-conditionally CR3)
+   (define-enum-conditionally BSDLY)
+   (define-enum-conditionally BS0)
+   (define-enum-conditionally BS1)
+   (define-enum-conditionally VTDLY)
+   (define-enum-conditionally VT0)
+   (define-enum-conditionally VT1)
+   (define-enum-conditionally FFDLY)
+   (define-enum-conditionally FF0)
+   (define-enum-conditionally FF1)
+
+   ;; c_cflag masks
+   (define-enum CLOCAL)
+   (define-enum CREAD)
+   (define-enum CSIZE)
+   (define-enum CS5)
+   (define-enum CS6)
+   (define-enum CS7)
+   (define-enum CS8)
+   (define-enum CSTOPB)
+   (define-enum HUPCL)
+   (define-enum PARENB)
+   (define-enum PARODD)
+   (define-enum-conditionally CIBAUD)
+   (define-enum-conditionally CRTSCTS)
+
+   ;; c_lflag masks
+   (define-enum ECHO)
+   (define-enum ECHOE)
+   (define-enum ECHOK)
+   (define-enum ECHONL)
+   (define-enum ICANON)
+   (define-enum ISIG)
+   (define-enum NOFLSH)
+   (define-enum TOSTOP)
+   (define-enum IEXTEN)
+   (define-enum-conditionally XCASE)
+   (define-enum-conditionally ECHOCTL)
+   (define-enum-conditionally ECHOPRT)
+   (define-enum-conditionally ECHOKE)
+   (define-enum-conditionally FLUSH0)
+   (define-enum-conditionally PENDIN)
+
+   ;; c_cc size
+   (define-enum NCCS)
+
+   ;; disable character
+   (define-enum _POSIX_VDISABLE)
+
+   ;; c_cc subscripts
+   (define-enum VEOF)
+   (define-enum VEOL)
+   (define-enum VERASE)
+   (define-enum VINTR)
+   (define-enum VKILL)
+   (define-enum VMIN)
+   (define-enum VQUIT)
+   (define-enum VSTART)
+   (define-enum VSTOP)
+   (define-enum VSUSP)
+   (define-enum VTIME)
+   (define-enum-conditionally VDISCARD)
+   (define-enum-conditionally VDSUSP)
+   (define-enum-conditionally VEOL2)
+   (define-enum-conditionally VLNEXT)
+   (define-enum-conditionally VREPRINT)
+   (define-enum-conditionally VSTATUS)
+   (define-enum-conditionally VWERASE)
+   (define-enum-conditionally VSWTCH)
+   (define-enum-conditionally VSWTC)
+
+   ;; baudrates
    (define-enum B0)
    (define-enum B50)
    (define-enum B75)
@@ -76,18 +215,28 @@
    (define-enum B9600)
    (define-enum B19200)
    (define-enum B38400)
+   (define-enum-conditionally B57600)
+   (define-enum-conditionally B115200)
+   (define-enum-conditionally B230400)
 
-   (define-cproc sys-tcgetattr (port-or-fd)
+   (define-cstruct <sys-termios> "struct termios"
+     (iflag::<ulong> "c_iflag"
+      oflag::<ulong> "c_oflag"
+      cflag::<ulong> "c_cflag"
+      lflag::<ulong> "c_lflag"
+      cc::(.array <uint8> :as <u8vector>) "c_cc[NCCS]"))
+
+   (define-cproc sys-tcgetattr (port-or-fd) ::<sys-termios>
      (let* ([fd::int (Scm_GetPortFd port-or-fd TRUE)]
-            [term::ScmSysTermios* (SCM_SYS_TERMIOS (Scm_MakeSysTermios))])
-       (when (< (tcgetattr fd (& (-> term term))) 0)
+            [term::(struct termios)])
+       (when (< (tcgetattr fd (& term)) 0)
          (Scm_SysError "tcgetattr failed"))
-       (return (SCM_OBJ term))))
+       (return (& term))))
 
    (define-cproc sys-tcsetattr (port-or-fd option::<fixnum> term::<sys-termios>)
      ::<void>
      (let* ([fd::int (Scm_GetPortFd port-or-fd TRUE)])
-       (when (< (tcsetattr fd option (& (-> term term))) 0)
+       (when (< (tcsetattr fd option term) 0)
          (Scm_SysError "tcsetattr failed"))))
 
    (define-cproc sys-tcsendbreak (port-or-fd duration::<fixnum>) ::<boolean>
@@ -117,46 +266,72 @@
        (when (< (tcsetpgrp fd pgrp) 0) (Scm_SysError "tcsetpgrp failed"))))
 
    (define-cproc sys-cfgetispeed (term::<sys-termios>) ::<int>
-     (let* ([s::speed_t (cfgetispeed (& (-> term term)))])
+     (let* ([s::speed_t (cfgetispeed term)])
        (return s)))
 
    (define-cproc sys-cfsetispeed (term::<sys-termios> speed::<int>) ::<void>
-     (when (< (cfsetispeed (& (-> term term)) speed) 0)
+     (when (< (cfsetispeed term speed) 0)
        (Scm_SysError "cfsetispeed failed")))
 
    (define-cproc sys-cfgetospeed (term::<sys-termios>) ::<int>
-     (let* ([s::speed_t (cfgetospeed (& (-> term term)))])
+     (let* ([s::speed_t (cfgetospeed term)])
        (return s)))
 
    (define-cproc sys-cfsetospeed (term::<sys-termios> speed::<int>) ::<void>
-     (when (< (cfsetospeed (& (-> term term)) speed) 0)
+     (when (< (cfsetospeed term speed) 0)
        (Scm_SysError "cfsetospeed failed")))
 
    ;; Returns fresh copy of <sys-termios>
    ;; The fields of struct termios are system-dependent, and it may contain
    ;; more fields than specified in POSIX.  However, we assume they're safe
-   ;; for bitwise-copy.
-   (define-cproc sys-termios-copy (src::<sys-termios>)
-     (let* ([dest::(ScmSysTermios*) (SCM_SYS_TERMIOS (Scm_MakeSysTermios))])
-       (set! (-> dest term) (-> src term))
-       (return (SCM_OBJ dest))))
+   ;; for bitwise-copy.  The cstruct boxer copies *src into the fresh object.
+   (define-cproc sys-termios-copy (src::<sys-termios>) ::<sys-termios>
+     (return src))
 
    ;; pty interface
    (.when (defined HAVE_OPENPTY)
-     (define-cproc sys-openpty (:optional term) Scm_Openpty)
+     (define-cproc sys-openpty (:optional (term::<sys-termios>? #f))
+       ::(<int> <int>)
+       (let* ([master::int 0] [slave::int 0])
+         (when (< (openpty (& master) (& slave) NULL term NULL) 0)
+           (Scm_SysError "openpty failed"))
+         (return master slave)))
      )
    (.when (defined HAVE_FORKPTY)
-     (define-cproc sys-forkpty (:optional term) Scm_Forkpty)
+     (define-cproc sys-forkpty (:optional (term::<sys-termios>? #f))
+       ::(<int> <int>)
+       (let* ([master::int 0]
+              [pid::pid_t (forkpty (& master) NULL term NULL)])
+         (when (< pid 0)
+           (Scm_SysError "forkpty failed"))
+         (return pid master)))
      (define-cproc sys-forkpty-and-exec (program::<string> args::<list>
-                                         :key (iomap ()) 
-                                              term 
+                                         :key (iomap ())
+                                              (term::<sys-termios>? #f)
                                               (sigmask::<sys-sigset>? #f))
-       Scm_ForkptyAndExec)
+       ::(<int> <int>)
+       (let* ([argc::ScmSmallInt (Scm_Length args)])
+         (when (< argc 1)
+           (Scm_Error "argument list must have at least one element: %S" args))
+         (let* ([argv::char** (Scm_ListToCStringArray args TRUE NULL)]
+                [cprogram::(const char *) (Scm_GetStringConst program)]
+                [fds::int* (Scm_SysPrepareFdMap iomap)]
+                [master::int 0]
+                [pid::pid_t (forkpty (& master) NULL term NULL)])
+           (when (< pid 0)
+             (Scm_SysError "forkpty failed"))
+           (when (== pid 0)
+             (Scm_SysSwapFds fds)
+             (when sigmask
+               (Scm_ResetSignalHandlers (& (-> sigmask set)))
+               (Scm_SysSigmask SIG_SETMASK sigmask))
+             (execvp cprogram (cast (char * const *) argv))
+             ;; here, we failed
+             (Scm_Panic "exec failed: %s: %s" cprogram (strerror errno)))
+           (return pid master))))
      )
 
    ) ;; !defined(GAUCHE_WINDOWS)
-
- (initcode (Scm_Init_termios))
  ) ;; inline-stub
 
 ;;
@@ -173,9 +348,7 @@
  (define (has-windows-console?)
    (cond-expand
     [gauche.os.windows
-     (or (sys-isatty (standard-input-port))
-         (sys-isatty (standard-output-port))
-         (sys-isatty (standard-error-port)))]
+     (sys-has-windows-console?)]
     [else #f]))
 
  (cond-expand
@@ -291,40 +464,55 @@
          (define saved-buffering (port-buffering port))
          (define new-attr
            (rlet1 attr (sys-termios-copy saved-attr)
+             ;; NB: See cfmakeraw(3) for raw mode setting.
              (case mode
                [(raw)
-                (set! (~ attr'iflag)
-                      (logand (~ attr'iflag)
-                              (lognot (logior BRKINT ICRNL INPCK ISTRIP IXON))))
-                (set! (~ attr'oflag) (logand (~ attr'oflag) (lognot OPOST)))
-                (set! (~ attr'cflag)
-                      (logior (logand (~ attr'cflag)
-                                      (lognot (logior CSIZE PARENB)))
-                              CS8))
-                (set! (~ attr'lflag)
-                      (logand (~ attr'lflag)
-                              (lognot (logior ECHO ICANON IEXTEN ISIG))))]
+                (update! (~ attr'iflag)
+                         (cut logset+clear <>
+                              0
+                              (logior BRKINT ICRNL INPCK ISTRIP IXON)))
+                (update! (~ attr'oflag)
+                         (cut logset+clear <>
+                              0
+                              OPOST))
+                (update! (~ attr'cflag)
+                         (cut logset+clear <>
+                              CS8
+                              (logior CSIZE PARENB)))
+                (update! (~ attr'lflag)
+                         (cut logset+clear <>
+                              0
+                              (logior ECHO ICANON IEXTEN ISIG)))]
                [(rare)
-                (set! (~ attr'iflag)
-                      (logior BRKINT
-                              (logand (~ attr'iflag)
-                                      (lognot (logior ICRNL INPCK ISTRIP IXON)))))
-                (set! (~ attr'oflag) (logand (~ attr'oflag) (lognot OPOST)))
-                (set! (~ attr'cflag)
-                      (logior (logand (~ attr'cflag)
-                                      (lognot (logior CSIZE PARENB)))
-                              CS8))
-                (set! (~ attr'lflag)
-                      (logior ISIG
-                              (logand (~ attr'lflag)
-                                      (lognot (logior ECHO ICANON IEXTEN)))))]
+                (update! (~ attr'iflag)
+                         (cut logset+clear <>
+                              BRKINT
+                              (logior ICRNL INPCK ISTRIP IXON)))
+                (update! (~ attr'oflag)
+                         (cut logset+clear <>
+                              OPOST
+                              0))
+                (update! (~ attr'cflag)
+                         (cut logset+clear <>
+                              CS8
+                              (logior CSIZE PARENB)))
+                (update! (~ attr'lflag)
+                         (cut logset+clear <>
+                              ISIG
+                              (logior ECHO ICANON IEXTEN)))]
                [(cooked)
-                (set! (~ attr'iflag)
-                      (logior (~ attr'iflag)
-                              BRKINT ICRNL INPCK ISTRIP IXON))
-                (set! (~ attr'oflag) (logior (~ attr'oflag) OPOST))
-                (set! (~ attr'lflag)
-                      (logior (~ attr'lflag) ECHO ICANON IEXTEN ISIG))]
+                (update! (~ attr'iflag)
+                         (cut logset+clear <>
+                              (logior BRKINT ICRNL INPCK ISTRIP IXON)
+                              0))
+                (update! (~ attr'oflag)
+                         (cut logset+clear <>
+                              OPOST
+                              0))
+                (update! (~ attr'lflag)
+                         (cut logset+clear <>
+                              (logior ECHO ICANON IEXTEN ISIG)
+                              0))]
                [else
                 (error "terminal mode needs to be one of cooked, rare or raw, \
                           but got:" mode)])))
@@ -337,5 +525,5 @@
            (set! (port-buffering port) saved-buffering)
            (when cleanup (cleanup)))
          (unwind-protect (begin (set) (proc port)) (reset))))]))
- 
+
  ) ; without-precompiling
