@@ -7,8 +7,8 @@
 
 (test-start "termios")
 (use gauche.termios)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (test-module 'gauche.termios)
 
 (cond-expand
@@ -24,8 +24,8 @@
         (map cons
              syms
              (let1 p (open-input-string (string-append "(list " str ")"))
-               (eval (read p) (interaction-environment))))
-        (if (global-variable-bound? 'gauche.termios (car cans))
+               (eval (read p) (current-module))))
+        (if (module-binds? (current-module) (car cans))
           (loop (cdr cans)
                 (cons (car cans) syms)
                 (string-append str " " (symbol->string (car cans))))
@@ -154,6 +154,20 @@
                  )))
            ccs)))
 
+  ;; Check openpty basics
+  ;; sys-openpty isn't used in the tests above if we have console, so we check
+  ;; it here.
+  (cond-expand
+   [gauche.sys.openpty
+    (test* "openpty w/o term spec" '(#t #t #t)
+           (receive (m s) (sys-openpty)
+             (begin0 (list (integer? m)
+                           (integer? s)
+                           (sys-isatty s))
+               (sys-close m)
+               (sys-close s))))
+    ]
+   [else])
   ]
  [else
   ;;
