@@ -2,7 +2,7 @@
 ;;; rfc.tls.mbed - mbedTLS binding
 ;;; This is autoloaded from rfc.tls.  Don't directly use this module.
 ;;;
-;;;   Copyright (c) 2018-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2018-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,3 +38,12 @@
 (select-module rfc.tls.mbed)
 (dynamic-load "rfc--tls--mbed")
 
+;; This module may be loaded even Gauche doesn't support mbedtls, because
+;; of autoload (We need to set <mbed-tls> as autoloaded symbol).
+;; If Gauche isn't compiled with mbed support, mbed-set-debug-level! isn't
+;; defined.
+(cond-expand
+ [gauche.net.tls.mbedtls
+  ((with-module rfc.tls %tls-register-debug-level-callback)
+   mbed-set-debug-level!)]
+ [else])
