@@ -53,17 +53,17 @@ typedef struct ScmZlibInfoRec {
     int ownerp;
     int flush;
     int stream_endp;
-    int bufsiz;
+    ScmSize bufsiz;
     char *buf;
     char *ptr;
     unsigned char *dict;
-    int dictlen;
+    ScmSize dictlen;
     int level;
     int strategy;
     ScmObj dict_adler;
 } ScmZlibInfo;
 
-#define SCM_PORT_ZLIB_INFO(p) ((ScmZlibInfo*)(p)->src.buf.data)
+#define SCM_PORT_ZLIB_INFO(p) ((ScmZlibInfo*)Scm_PortBufferStruct(p)->data)
 #define SCM_PORT_ZSTREAM(p)   (SCM_PORT_ZLIB_INFO(p)->strm)
 
 SCM_CLASS_DECL(Scm_DeflatingPortClass);
@@ -76,8 +76,8 @@ SCM_CLASS_DECL(Scm_InflatingPortClass);
 extern ScmObj Scm_MakeDeflatingPort(ScmPort *source, int level,
                                     int window_bits, int memlevel,
                                     int strategy, ScmObj dict,
-                                    int bufsiz, int ownerp);
-extern ScmObj Scm_MakeInflatingPort(ScmPort *sink, int bufsiz,
+                                    ScmSize bufsiz, int ownerp);
+extern ScmObj Scm_MakeInflatingPort(ScmPort *sink, ScmSize bufsiz,
                                     int window_bits, ScmObj dict,
                                     int ownerp);
 
