@@ -379,7 +379,7 @@ CORD CORD_substr_closure(CORD x, size_t i, size_t n, CORD_fn f)
 
 # define SUBSTR_LIMIT (10 * SHORT_LIMIT)
         /* Substrings of function nodes and flat strings shorter than   */
-        /* this are flat strings.  Otherwise we use a functional         */
+        /* this are flat strings.  Othewise we use a functional         */
         /* representation, which is significantly slower to access.     */
 
 /* A version of CORD_substr that assumes i >= 0, n > 0, and i + n < length(x).*/
@@ -816,20 +816,20 @@ void CORD__next(CORD_pos p)
         if (cur_pos < end_pos) {
           /* Fill cache and return. */
             size_t i;
-            size_t limit = cur_pos + FUNCTION_BUF_SZ;
+            size_t limit = FUNCTION_BUF_SZ;
             CORD_fn fn = f -> fn;
             void * client_data = f -> client_data;
 
-            if (limit > end_pos) {
-                limit = end_pos;
+            if (end_pos - cur_pos < FUNCTION_BUF_SZ) {
+                limit = end_pos - cur_pos;
             }
-            for (i = cur_pos; i < limit; i++) {
-                p[0].function_buf[i - cur_pos] =
-                        (*fn)(i - start_pos, client_data);
+            for (i = 0; i < limit; i++) {
+                p[0].function_buf[i] = (*fn)(i + cur_pos - start_pos,
+                                             client_data);
             }
             p[0].cur_start = cur_pos;
             p[0].cur_leaf = p[0].function_buf;
-            p[0].cur_end = limit;
+            p[0].cur_end = cur_pos + limit;
             return;
         }
     }
