@@ -11,7 +11,7 @@
  * modified is included with the above copyright notice.
  */
 
-# include "gc.h"    /* For GC_INIT() only */
+# include "gc.h"
 # include "cord.h"
 
 # include <stdarg.h>
@@ -40,7 +40,8 @@ int count;
 
 int test_fn(char c, void * client_data)
 {
-    if (client_data != (void *)13) ABORT("bad client data");
+    if (client_data != (void *)(GC_word)13)
+        ABORT("bad client data");
     if (count < 64*1024+1) {
         if ((count & 1) == 0) {
             if (c != 'b') ABORT("bad char");
@@ -81,7 +82,8 @@ void test_basics(void)
     if (CORD_len(x) != 128*1024+1) ABORT("bad length");
 
     count = 0;
-    if (CORD_iter5(x, 64*1024-1, test_fn, CORD_NO_FN, (void *)13) == 0) {
+    if (CORD_iter5(x, 64*1024-1, test_fn, CORD_NO_FN,
+                   (void *)(GC_word)13) == 0) {
         ABORT("CORD_iter5 failed");
     }
     if (count != 64*1024 + 2) ABORT("CORD_iter5 failed");
@@ -89,8 +91,8 @@ void test_basics(void)
     count = 0;
     CORD_set_pos(p, x, 64*1024-1);
     while(CORD_pos_valid(p)) {
-        (void) test_fn(CORD_pos_fetch(p), (void *)13);
-    CORD_next(p);
+        (void)test_fn(CORD_pos_fetch(p), (void *)(GC_word)13);
+        CORD_next(p);
     }
     if (count != 64*1024 + 2) ABORT("Position based iteration failed");
 
@@ -113,7 +115,8 @@ void test_basics(void)
     if (CORD_len(x) != 128*1024+1) ABORT("bad length");
 
     count = 0;
-    if (CORD_iter5(x, 64*1024-1, test_fn, CORD_NO_FN, (void *)13) == 0) {
+    if (CORD_iter5(x, 64*1024-1, test_fn, CORD_NO_FN,
+                   (void *)(GC_word)13) == 0) {
         ABORT("CORD_iter5 failed");
     }
     if (count != 64*1024 + 2) ABORT("CORD_iter5 failed");
@@ -128,7 +131,7 @@ void test_basics(void)
     while(CORD_pos_valid(p)) {
         char c = CORD_pos_fetch(p);
 
-        if ((unsigned char)c != i)
+        if ((size_t)(unsigned char)c != i)
             ABORT("Traversal of function node failed");
         CORD_next(p);
         i++;
@@ -218,16 +221,15 @@ void test_extras(void)
     *(CORD volatile *)&w = CORD_EMPTY;
     *(CORD volatile *)&z = CORD_EMPTY;
     GC_gcollect();
-    GC_invoke_finalizers();
+#   ifndef GC_NO_FINALIZATION
+      GC_invoke_finalizers();
             /* Of course, this does not guarantee the files are closed. */
+#   endif
     if (remove(FNAME1) != 0) {
         /* On some systems, e.g. OS2, this may fail if f1 is still open. */
         /* But we cannot call fclose as it might lead to double close.   */
         fprintf(stderr, "WARNING: remove failed: " FNAME1 "\n");
     }
-    if (remove(FNAME2) != 0) {
-        fprintf(stderr, "WARNING: remove failed: " FNAME2 "\n");
-    }
 }
 
 int wrap_vprintf(CORD format, ...)
@@ -267,6 +269,8 @@ int wrap_vfprintf(FILE * f, CORD format, ...)
 # endif
 #endif
 
+/* no static */ /* no const */ char *zu_format = (char*)"%zu";
+
 void test_printf(void)
 {
     CORD result;
@@ -274,14 +278,15 @@ void test_printf(void)
     long l = -1;
     short s = (short)-1;
     CORD x;
+    int res;
 
     if (CORD_sprintf(&result, "%7.2f%ln", 3.14159F, &l) != 7)
         ABORT("CORD_sprintf failed 1");
-    if (CORD_cmp(result, "   3.14") != 0)ABORT("CORD_sprintf goofed 1");
+    if (CORD_cmp(result, "   3.14") != 0) ABORT("CORD_sprintf goofed 1");
     if (l != 7) ABORT("CORD_sprintf goofed 2");
     if (CORD_sprintf(&result, "%-7.2s%hn%c%s", "abcd", &s, 'x', "yz") != 10)
         ABORT("CORD_sprintf failed 2");
-    if (CORD_cmp(result, "ab     xyz") != 0)ABORT("CORD_sprintf goofed 3");
+    if (CORD_cmp(result, "ab     xyz") != 0) ABORT("CORD_sprintf goofed 3");
     if (s != 7) ABORT("CORD_sprintf goofed 4");
     x = "abcdefghij";
     x = CORD_cat(x,x);
@@ -296,7 +301,26 @@ void test_printf(void)
         (void)sprintf(result2, "->%-120.78s!\n", CORD_to_char_star(x));
 #   endif
     result2[sizeof(result2) - 1] = '\0';
-    if (CORD_cmp(result, result2) != 0)ABORT("CORD_sprintf goofed 5");
+    if (CORD_cmp(result, result2) != 0) ABORT("CORD_sprintf goofed 5");
+
+#   ifdef GC_SNPRINTF
+        /* Check whether "%zu" specifier is supported; pass the format  */
+        /* string via a variable to avoid a compiler warning if not.    */
+        res = GC_SNPRINTF(result2, sizeof(result2), zu_format, (size_t)0);
+#   else
+        res = sprintf(result2, zu_format, (size_t)0);
+#   endif
+    result2[sizeof(result2) - 1] = '\0';
+    if (res == 1) /* is "%z" supported by printf? */ {
+        if (CORD_sprintf(&result, "%zu %zd 0x%0zx",
+                         (size_t)123, (size_t)4567, (size_t)0x4abc) != 15)
+            ABORT("CORD_sprintf failed 5");
+        if (CORD_cmp(result, "123 4567 0x4abc") != 0)
+            ABORT("CORD_sprintf goofed 5");
+    } else {
+        (void)CORD_printf("printf lacks support of 'z' modifier\n");
+    }
+
     /* TODO: Better test CORD_[v][f]printf.     */
     (void)CORD_printf(CORD_EMPTY);
     (void)wrap_vfprintf(stdout, CORD_EMPTY);
@@ -317,6 +341,11 @@ int main(void)
     test_basics();
     test_extras();
     test_printf();
+
+    GC_gcollect(); /* to close f2 before the file removal */
+    if (remove(FNAME2) != 0) {
+        fprintf(stderr, "WARNING: remove failed: " FNAME2 "\n");
+    }
     CORD_fprintf(stdout, "SUCCEEDED\n");
     return(0);
 }
