@@ -1,7 +1,7 @@
 ;;;
 ;;; control.job - common job description
 ;;;
-;;;   Copyright (c) 2010-2016  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2010-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -123,13 +123,15 @@
       (job-result-set! job result)
       (job-status-set! job status)))
   (define (start)
-    (job-status-set! job 'running)
-    (%job-touch! job :start))
+    (and (not (eq? (job-status job) 'killed))
+         (begin (job-status-set! job 'running)
+                (%job-touch! job :start)
+                #t)))
 
-  (safely job (start))
-  (guard (e [else (safely job (finish 'error e))])
-    (let1 r ((job-thunk job))
-      (safely job (finish 'done r)))))
+  (and (safely job (start))
+       (guard (e [else (safely job (finish 'error e))])
+         (let1 r ((job-thunk job))
+           (safely job (finish 'done r))))))
 
 ;; this doesn't actually kill the process/threads that running the job,
 ;; but merely marks the job killed and notify it to the waiter.
@@ -144,4 +146,3 @@
         (job-status-set! job 'killed)
         (when cv
           (condition-variable-broadcast! cv))))))
-
