@@ -35,8 +35,8 @@
 (define-module dbi
   (use text.sql)
   (use file.util)
-  (use srfi-1)
-  (use srfi-13)
+  (use scheme.list)
+  (use srfi.13)
   (use util.match)
   (extend util.relation)
   (export <dbi-error> <dbi-nonexistent-driver-error>
@@ -212,7 +212,7 @@
                    [driver-class
                     (begin (eval `(require ,(path-sans-extension path))
                                  (current-module))
-                           (global-variable-ref module class-name #f))])
+                           (module-binding-ref module class-name #f))])
           (make driver-class :driver-name driver-name))
         (errorf <dbi-nonexistent-driver-error>
                 :driver-name driver-name
@@ -323,4 +323,3 @@
   (make <dbi-query> :connection c))
 (define-method dbi-execute-query ((q <dbi-query>) (s <string>))
   (dbi-do (ref q 'connection) s '(:pass-through #t)))
-
