@@ -1,7 +1,7 @@
 ;;;
 ;;; dbm - abstract base class for dbm interface
 ;;;
-;;;   Copyright (c) 2000-2018  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -241,4 +241,4 @@
          (guard (e [else #f])
            ((with-module gauche.internal %require)
             (module-name->path module-name))
-           (global-variable-ref (find-module module-name) class-name)))))
+           (module-binding-ref module-name class-name)))))
