@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.cgen.precomp - Precompile Scheme into C data
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -35,13 +35,13 @@
 ;;; as C code.
 
 (define-module gauche.cgen.precomp
-  (use srfi-1)
-  (use srfi-13)
+  (use scheme.list)
+  (use srfi.13)
   (use gauche.cgen)
   (use gauche.cgen.stub)
   (use gauche.cgen.tmodule)
   (use gauche.vm.insn)
-  (use gauche.parameter)
+  (use gauche.vm.debug-info)
   (use gauche.sequence)
   (use gauche.experimental.lamb)
   (use file.util)
@@ -151,6 +151,12 @@
 ;;      which isn't exported) are not included in the output.  But sometimes
 ;;      hygienic public macros expands to a call of private macros, and
 ;;      gauche.cgen.precomp cannot detect such dependencies yet.
+;;
+;; load-paths : Extra load paths prepended to the system's *load-path* during
+;;      compilation.  NB: Currnently, precompilation environment isn't isolated
+;;      enough from the host environment.  You can't load a different version
+;;      of the same library the host is using just for the precompiled file.
+;;      This option is mainly for 'include's in the source.
 
 (define (cgen-precompile src . keys)
   (with-tmodule-recording
@@ -164,9 +170,13 @@
                                :key (ext-initializer #f)
                                     ((:strip-prefix prefix) #f)
                                     ((:dso-name dso) #f)
+                                    (omit-line-directives (cise-omit-source-line))
+                                    ((:omit-debug-source-info no-source) #f)
                                     (predef-syms '())
                                     (macros-to-keep '())
                                     (single-sci-file #f)
+                                    (load-paths '())
+                                    (target-parameters '())
                                     (extra-optimization #f))
   (define (precomp-1 main src)
     (let* ([out.c (cgen-scm-path->c-file (strip-prefix src prefix))]
@@ -175,12 +185,16 @@
       (%cgen-precompile src
                         :out.c out.c
                         :dso-name (or dso (basename-sans-extension main))
+                        :omit-line-directives omit-line-directives
+                        :omit-debug-source-info no-source
                         :predef-syms predef-syms
                         :strip-prefix prefix
                         :macros-to-keep macros-to-keep
                         :extra-optimization extra-optimization
                         :ext-initializer (and (equal? src main)
                                               ext-initializer)
+                        :target-parameters target-parameters
+                        :load-paths load-paths
                         :initializer-name initname)))
   (define (tweak-sci sci) ; for consolidating sci files. returns sexp list
     (if (not sci)
@@ -217,23 +231,39 @@
 (define (%cgen-precompile src
                           :key (out.c #f)
                                (out.sci #f)
+                               (omit-line-directives (cise-omit-source-line))
+                               ((:omit-debug-source-info no-source) #f)
                                ((:strip-prefix prefix) #f)
                                (ext-initializer #f)
                                ((:dso-name dso) #f)
                                (initializer-name #f)
                                (sub-initializers '())
                                (predef-syms '())
+                               (load-paths '())
                                (macros-to-keep '())
+                               ((:target-parameters tparams) '())
                                (extra-optimization #f))
   (define (do-it)
-    (parameterize ([omitted-code '()])
+    (parameterize ([omitted-code '()]
+                   [omit-debug-source-info no-source]
+                   [target-parameters tparams])
       (setup ext-initializer sub-initializers)
       (with-input-from-file src
-        (cut emit-toplevel-executor
-             (reverse (generator-fold compile-toplevel-form '() read))))
+        (^[]
+          ;; NB: *load-path* isn't a parameter (yet), so we hack it.
+          ;; This shouldn't be the way; do not copy it.
+          (define load-path-save *load-path*)
+          (unwind-protect
+              (begin
+                (set! *load-path* (fold cons *load-path* (reverse load-paths)))
+                (emit-toplevel-executor
+                 (reverse (generator-fold compile-toplevel-form '() read))))
+            (set! *load-path* load-path-save))))
       (finalize sub-initializers)
+      (setup-debug-info)
+      (show-debug-info-stat (cgen-current-unit))
       (cgen-emit-c (cgen-current-unit))))
-  (let ([out.c   (or out.c (path-swap-extension (sys-basename src) "c"))]
+  (let ([out.c   (or out.c (cgen-scm-path->c-file (sys-basename src)))]
         [out.sci (or out.sci
                      (and (check-first-form-is-define-module src)
                           (strip-prefix (path-swap-extension src "sci")
@@ -247,6 +277,7 @@
                                        (basename-sans-extension out.c)))
                               initializer-name)]
                    [vm-eval-situation SCM_VM_COMPILING]
+                   [cise-omit-source-line omit-line-directives]
                    [private-macros-to-keep macros-to-keep]
                    [run-extra-optimization-passes extra-optimization])
       (select-tmodule 'gauche)
@@ -277,7 +308,10 @@
      "--"))
 
 (define (cgen-c-file->initfn c-name)
-  (let1 initfn (string-tr (path-sans-extension c-name) "-+." "___")
+  (let1 initfn (regexp-replace-all*
+                (path-sans-extension
+                 (sys-normalize-pathname c-name :canonicalize #t))
+                #/[-+.]/ "_")
     #"Scm_Init_~initfn"))
 
 ;;================================================================
@@ -300,8 +334,10 @@
   (eval expr (~ (current-tmodule)'module)))
 
 ;; Expr -> CompiledCode
+;;  target-params parameter is passed to 'compile'.
 (define (compile-in-current-tmodule expr)
-  (compile expr (~ (current-tmodule)'module)))
+  (compile expr (~ (current-tmodule)'module)
+           :target-params (target-parameters)))
 
 ;;================================================================
 ;; Parameters
@@ -337,6 +373,12 @@
 ;; Experimental: Run extra optimization during AOT compilation.
 (define run-extra-optimization-passes (make-parameter #f))
 
+;; Compile target parameters.  See 'compile' entry in compile.scm
+(define target-parameters (make-parameter '(:cont-frame-size 7)))
+
+;; Do not emit debug-source-info with the compiled code
+(define omit-debug-source-info (make-parameter #f))
+
 ;;================================================================
 ;; Bridge to the internal stuff
 ;;
@@ -350,8 +392,6 @@
   (with-module gauche.internal vm-eval-situation))
 (define global-eq?? (with-module gauche.internal global-eq??))
 (define make-identifier (with-module gauche.internal make-identifier))
-(define pair-attributes (with-module gauche.internal pair-attributes))
-(define pair-attribute-get (with-module gauche.internal pair-attribute-get))
 
 (define-constant SCM_VM_COMPILING 2) ;; must match with vm.h
 
@@ -364,6 +404,7 @@
         [initfn (cgen-c-file->initfn out.c)])
     (rlet1 u (make <cgen-stub-unit>
                :name base :c-name-prefix (cgen-safe-name base)
+               :language (if (equal? (path-extension out.c) "cpp") 'c++ 'c)
                :preamble `(,#"/* Generated automatically from ~|src|.  DO NOT EDIT */")
                :init-prologue (format "~avoid ~a() {"
                                       (if ext-init? "SCM_EXTENSION_ENTRY " "")
@@ -388,7 +429,7 @@
 
 ;; Read the first form.
 ;; We don't read the entire content of the file, since it may contain
-;; srfi-10 read-time constructor that we don't know about yet.
+;; SRFI-10 read-time constructor that we don't know about yet.
 (define (first-form src) (with-input-from-file src read))
 
 ;; Check if the first form is define-module.
@@ -566,7 +607,7 @@
       [((? =export-all?)) (write-ext-module form) (compile-module-exports #t)]
       [((? =export-if-defined?) . _) (write-ext-module form) seed]
       [((? =provide?) arg) (write-ext-module form) seed]
-      [((? =extend?) . _) 
+      [((? =extend?) . _)
        (write-ext-module form)
        (eval-in-current-tmodule form)
        seed]
@@ -621,9 +662,9 @@
                                           ~a,\
                                           ~a);\n"
                        (if mod (tmodule-cname mod) "Scm_CurrentModule()")
-                       (if mod (cgen-safe-comment (~ mod'name)) "")
+                       (if mod (cgen-safe-comment-sexp (~ mod'name)) "")
                        (cgen-cexpr sym)
-                       (cgen-safe-comment (unwrap-syntax id))
+                       (cgen-safe-comment-sexp id)
                        tmp
                        (case flags
                          [(2) 'SCM_BINDING_CONST]
@@ -650,10 +691,12 @@
       (error "include requires literal string, but got:" filename))
     (with-input-from-port
         ($ (with-module gauche.internal pass1/open-include-file)
-           filename 
+           filename
            (sys-dirname (port-name (current-input-port))))
       (^[]
-        (port-case-fold-set! (current-input-port) case-fold?)
+        (set! ((with-module gauche.internal port-case-fold)
+               (current-input-port))
+              case-fold?)
         (unwind-protect
             (generator-fold compile-toplevel-form seed read)
           (close-port (current-input-port))))))
@@ -688,6 +731,9 @@
         ((with-module gauche.cgen.stub cgen-stub-parse-form)
          (unwrap-syntax s)))
       (undefined))
+    (define-macro (in-module name)
+      ((with-module gauche.cgen.precomp select-tmodule) name)
+      (undefined))
     (define-macro (define-cproc . args)
       ((with-module gauche.cgen.stub cgen-stub-parse-form)
        (unwrap-syntax (cons 'define-cproc args)))
@@ -754,7 +800,7 @@
           (,%define ,name (,%macro ',name
                                    (,%lambda (,form ,env)
                                              (,%apply ,expr (,%cdr ,form)))
-                                   ',expr))
+                                   '((source . ,expr))))
           ((with-module gauche define-macro) ,name ,expr)))
       `((with-module gauche define-macro) ,name ,expr)))
 
@@ -777,7 +823,7 @@
      ;; Otherwise, we emit the form to *.sci file.
      (when (or (symbol-exported? name)
                (memq name (private-macros-to-keep)))
-       (let1 val (global-variable-ref (~ (current-tmodule)'module) name #f)
+       (let1 val (module-binding-ref (~ (current-tmodule)'module) name #f)
          (or (and-let* ([ ((with-module gauche.internal macro?) val) ]
                         [tx ((with-module gauche.internal macro-transformer) val)]
                         [ (closure? tx) ])
@@ -797,7 +843,7 @@
       `((with-module gauche define-syntax) ,name ,xformer-spec))
      ;; Now generate code.  NB: Once we make macro serializable, we won't
      ;; need all these manual wiring.
-     (let1 val (global-variable-ref (~ (current-tmodule)'module) name #f)
+     (let1 val (module-binding-ref (~ (current-tmodule)'module) name #f)
        (or (and-let* ([ ((with-module gauche.internal macro?) val) ]
                       [tx ((with-module gauche.internal macro-transformer) val)]
                       [ (closure? tx) ])
@@ -856,6 +902,7 @@
    [code-vector-c-name :init-keyword :code-vector-c-name]
    [literals           :init-keyword :literals]
    [signature-info     :init-keyword :signature-info]
+   [debug-info-cname   :init-keyword :debug-info-cname]
    )
   (make (value)
     (let* ([code (if (run-extra-optimization-passes)
@@ -869,16 +916,17 @@
            [inliner (check-packed-inliner code)])
       (define (init-thunk)
         (format #t "    SCM_COMPILED_CODE_CONST_INITIALIZER(  /* ~a */\n"
-                (cgen-safe-comment (~ code'name)))
+                (cgen-safe-comment-sexp (~ code'name)))
         (format #t "            (ScmWord*)(~a), ~a,\n"
                 cvn (length cv))
-        (format #t "            ~a, ~a, ~a, ~a, SCM_NIL, ~a,\n"
+        (format #t "            ~a, ~a, ~a, ~a, ~a, ~a,\n"
                 (~ code'max-stack)
                 (~ code'required-args)
                 (~ code'optional-args)
                 (if (cgen-literal-static? code-name)
                   (cgen-cexpr code-name)
                   "SCM_FALSE")
+                "SCM_NIL"
                 (if (cgen-literal-static? signature-info)
                   (cgen-cexpr signature-info)
                   "SCM_FALSE"))
@@ -896,15 +944,20 @@
             :code-vector-c-name cvn
             :code-name code-name
             :signature-info signature-info
+            :debug-info-cname (and (not (omit-debug-source-info))
+                                   (serializable-debug-info code))
             :literals lv)))
   (init (self)
     (unless (cgen-literal-static? (~ self'code-name))
       (print "  SCM_COMPILED_CODE("(~ self'c-name)")->name = "
              (cgen-cexpr (~ self'code-name))";"
-             "/* "(cgen-safe-comment (~ self'value'full-name))" */"))
+             "/* "(cgen-safe-comment-sexp (~ self'value'full-name))" */"))
     (unless (cgen-literal-static? (~ self'signature-info))
       (print "  SCM_COMPILED_CODE("(~ self'c-name)")->signatureInfo = "
              (cgen-cexpr (~ self'signature-info)) ";"))
+    (when (~ self'debug-info-cname)
+      (print "  SCM_COMPILED_CODE("(~ self'c-name)")->debugInfo = "
+             (~ self'debug-info-cname) ";"))
     (fill-code self))
   (static (self) #t)
   )
@@ -926,6 +979,21 @@
             (unwrap-syntax (cdr sig)))
       (unwrap-syntax sig))))
 
+;; Construct serializable debug info.
+;;   See code.c Scm_CompiledCodePushInfo for the format of debug-info.
+(define (serializable-debug-info code)
+  (and-let1 di (~ code'debug-info)
+    (let1 codevec (encode-debug-info (cgen-current-unit)
+                                     (unwrap-syntax di))
+      (record-debug-info-stat! (cgen-current-unit) codevec #f)
+      (let ([codevec-lit (cgen-literal codevec)]
+            [datum-cname (cgen-allocate-static-datum 'runtime)])
+        (cgen-init (format "  ~a = Scm_MakePackedDebugInfo(SCM_U8VECTOR(~a),\
+                                         SCM_debug_info_const_vector);"
+                           datum-cname
+                           (cgen-cexpr codevec-lit)))
+        datum-cname))))
+
 ;; Returns a list of the same length of CODE, which includes the
 ;; <cgen-literal>s corresponding to the literal values in CODE.
 ;; #f is filled in the places that don't have corresponding litaral value.
@@ -938,12 +1006,12 @@
              [info (vm-find-insn-info (car insn))])
         (case (~ info'operand-type)
           [(none) (loop (cdr code)  (cons #f lits))]
-          [(addr) (loop (cddr code) (list* #f #f lits))]
+          [(label) (loop (cddr code) (list* #f #f lits))]
           [(code codes) (loop (cddr code)
                               (list* (cgen-literal (cadr code)) #f lits))]
           [(obj) (loop (cddr code)
                         (list* (cgen-literal (cadr code)) #f lits))]
-          [(obj+addr)
+          [(obj+label obj+native)
            (loop (cdddr code)
                  (list* #f (cgen-literal (cadr code)) #f lits))]
           )))))
@@ -961,7 +1029,7 @@
              [insnval (vm-build-insn insn)]
              [name-info (if first-cexpr
                           ""
-                          (format "/* ~a */\n    " (cgen-safe-comment full-name)))]
+                          (format "/* ~a */\n    " (cgen-safe-comment-sexp full-name)))]
              [insn-cexpr
               (alloc-word
                ;; We emit it as signed integer so that 64bit machine
@@ -974,10 +1042,10 @@
                          name-info insnval count
                          (cgen-safe-comment insn))))]
              [first-cexpr (or first-cexpr insn-cexpr)])
-        (case (~ info'operand-type)
+        (ecase (~ info'operand-type)
           [(none)
            (loop (cdr cv) (cdr lv) (+ count 1) first-cexpr)]
-          [(addr)
+          [(label)
            (alloc-word
             (format "SCM_WORD((ScmWord*)~a + ~d)"
                     first-cexpr (cadr cv)))
@@ -987,22 +1055,25 @@
             (if (cgen-literal-static? (cadr lv))
               (format "SCM_WORD(~a) /* ~a */"
                       (cgen-cexpr (cadr lv))
-                      (cgen-safe-comment (write-to-string (cadr cv))))
+                      (cgen-safe-comment-sexp (cadr cv)))
               (format "SCM_WORD(SCM_UNDEFINED) /* ~a */"
-                      (cgen-safe-comment (write-to-string (cadr cv))))))
+                      (cgen-safe-comment-sexp (cadr cv)))))
            (loop (cddr cv) (cddr lv) (+ count 2) first-cexpr)]
-          [(obj+addr)
+          [(obj+label)
            (alloc-word
             (if (cgen-literal-static? (cadr lv))
               (format "SCM_WORD(~a) /* ~a */"
                       (cgen-cexpr (cadr lv))
-                      (cgen-safe-comment (write-to-string (cadr cv))))
+                      (cgen-safe-comment-sexp (cadr cv)))
               (format "SCM_WORD(SCM_UNDEFINED) /* ~a */"
-                      (cgen-safe-comment (write-to-string (cadr cv))))))
+                      (cgen-safe-comment-sexp (cadr cv)))))
            (alloc-word
             (format "SCM_WORD((ScmWord*)~a + ~d)  /*    ~3d */"
                     first-cexpr (caddr cv) (caddr cv)))
            (loop (cdddr cv) (cdddr lv) (+ count 3) first-cexpr)]
+          [(obj+native)
+           (errorf "~a instruction shouldn't appear in precompiled code"
+                   (~ info'name))]
           ))))
 
   (loop cv lv 0 #f))
@@ -1030,75 +1101,28 @@
                          il insns)])
            (cgen-literal packed)))))
 
-;;----------------------------------------------------------------
-;; <module>
-;;
-
-;; We recover modules at runtime.  Anonymous modules can be recovered
-;; if they are used as a proxy of a real module (via <tmodule>).
-(define-cgen-literal <cgen-module> <module>
-  ((name  :init-keyword :name)
-   (name-literal :init-keyword :name-literal))
-  (make (value)
-    (if-let1 n (predefined-module value)
-      (make <cgen-module> :value value
-            :c-name n
-            :name (~ value'name))
-      (let1 name (or (~ value'name)
-                     (any (^[tm] (and (eq? (~ tm'module) value) (~ tm'name)))
-                          (all-tmodules))
-                     (error "Cannot emit literal for anonymous module:" value))
-        (make <cgen-module> :value value
-              :c-name (cgen-allocate-static-datum)
-              :name name :name-literal (cgen-literal name)))))
-  (init (self)
-    (unless (predefined-module (~ self'value))
-      ;; Module literals are emitted only related with macro hygiene.  In that
-      ;; case, the module is created without loading its actual body, so
-      ;; we mark it as 'placeholding'.  The placeholding module won't show
-      ;; up in (all-modules).  Once (define-module ...) is evaluated by loading
-      ;; the module, it becomes 'real' module.
-      (format #t "  ~a = SCM_OBJ(Scm_FindModule(SCM_SYMBOL(~a), \
-                         SCM_FIND_MODULE_CREATE|SCM_FIND_MODULE_PLACEHOLDING)); \
-                  /* module ~a */\n"
-              (~ self'c-name)
-              (cgen-cexpr (~ self'name-literal))
-              (~ self'name))))
-  (static (self) (boolean (predefined-module (~ self'value))))
-  )
-
-(define (predefined-module module)
-  (assq-ref `((,(find-module 'null)   . "Scm_NullModule()")
-              (,(find-module 'scheme) . "Scm_SchemeModule()")
-              (,(find-module 'gauche) . "Scm_GaucheModule()")
-              (,(find-module 'gauche.internal) . "Scm_GaucheInternalModule()")
-              )
-            module))
-
-;;----------------------------------------------------------------
-;; <identifier>
-;;
-
-(define-cgen-literal <cgen-scheme-identifier> <identifier>
-  ((id-name        :init-keyword :id-name)
-   (module-literal :init-keyword :module-literal))
-  (make (value)
-    (unless ((with-module gauche.internal identifier-toplevel?) value)
-      (error "identifier with compiler environment can't be compiled" value))
-    (make <cgen-scheme-identifier> :value value
-          :c-name (cgen-allocate-static-datum)
-          :id-name (cgen-literal (unwrap-syntax value))
-          :module-literal (cgen-literal (~ value'module))))
-  (init (self)
-    (let ([name (cgen-cexpr (~ self'id-name))]
-          [cname (~ self'c-name)])
-      (format #t "  ~a = Scm_MakeIdentifier(~a, SCM_MODULE(~a), \
-                                            SCM_NIL); /* ~a#~a */\n"
-              cname name (cgen-cexpr (~ self'module-literal))
-              (cgen-safe-comment (~ self'module-literal'name))
-              (cgen-safe-comment (~ self'id-name'value)))))
-  (static (self) #f)
-  )
+;; Debug info generation.  This must be called once, after all compiled codes
+;; are processed.
+(define (setup-debug-info)
+  (when (has-unit-debug-info? (cgen-current-unit))
+    (let* ([cvec (get-debug-info-const-vector (cgen-current-unit))]
+           [vec-literal (cgen-literal (make-vector (vector-length cvec) #f))]
+           [vec-name (cgen-cexpr vec-literal)]
+           [elt-literals (map cgen-literal cvec)])
+      (cgen-decl "static ScmObj SCM_debug_info_const_vector();")
+      (cgen-body "ScmObj SCM_debug_info_const_vector()"
+                 "{"
+                 ;; This is idempotent.  No need to mutex.
+                 "  static _Bool initialized = FALSE;"
+                 "  if (!initialized) {"
+                 "    int i = 0;")
+      (dolist [e elt-literals]
+        (cgen-body #"    SCM_VECTOR_ELEMENT(~|vec-name|, i++) = \
+                           ~(cgen-cexpr e);"))
+      (cgen-body "    initialized = TRUE;"
+                 "  }"
+                 #"  return ~|vec-name|;"
+                 "}"))))
 
 ;;----------------------------------------------------------------
 ;; <macro>
@@ -1168,5 +1192,5 @@
   (init (self)
     (format #t "  ~a = Scm_MakeClosure(~a, NULL); /* ~a */\n"
             (cgen-cexpr self) (cgen-cexpr (~ self'code))
-            (cgen-safe-comment (write-to-string (~ self'value)))))
+            (cgen-safe-comment-sexp (~ self'value))))
   (static (self) #f))
