@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.cgen.unit - cgen-unit
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,11 +32,10 @@
 ;;;
 
 (define-module gauche.cgen.unit
-  (use srfi-13)
-  (use srfi-42)
+  (use srfi.13)
+  (use srfi.42)
   (use util.match)
   (use file.util)
-  (use gauche.parameter)
   (use gauche.sequence)
   (export <cgen-unit> cgen-current-unit
           cgen-unit-c-file cgen-unit-init-name cgen-unit-h-file
@@ -49,7 +48,7 @@
           cgen-include cgen-define cpp-condition->string
 
           cgen-safe-name cgen-safe-name-friendly
-          cgen-safe-string cgen-safe-comment
+          cgen-safe-string cgen-safe-comment cgen-safe-comment-sexp
 
           ;; semi-private routines
           cgen-emit-static-data)
@@ -70,6 +69,7 @@
    (name     :init-keyword :name   :init-value "cgen")
    (c-file   :init-keyword :c-file :init-value #f)
    (h-file   :init-keyword :h-file :init-value #f)
+   (language :init-keyword :language :init-value 'c) ; c or c++
    (preamble :init-keyword :preamble
              :init-value '("/* Generated by gauche.cgen */"))
    (init-prologue :init-keyword :init-prologue :init-value #f)
@@ -79,15 +79,26 @@
    (transients :init-value '())  ;; transient variables
    (literals  :init-form #f)     ;; literals. see gauche.cgen.literal
    (static-data-list :init-value '()) ;; static C data, see below
+   (debug-info-bin :init-value #f)   ;; see gauche.vm.debug-info
    ))
 
+;; A dummy instance to show a better error message  when cgen routine is
+;; used without setting a proper unit
+(define-class <cgen-dummy-unit> () ())
+
+(define-method slot-missing ((c <class>) (obj <cgen-dummy-unit>) slot . _)
+  (error "You need to bind an instance of <cgen-unit> to use \
+          cgen code generation procedures."))
+
 ;; API
-(define cgen-current-unit (make-parameter #f))
+(define cgen-current-unit (make-parameter (make <cgen-dummy-unit>)))
 
 ;; API
 (define-method cgen-unit-c-file ((unit <cgen-unit>))
   (or (~ unit'c-file)
-      #"~(~ unit 'name).c"))
+      (ecase (~ unit'language)
+        [(c)   #"~(~ unit 'name).c"]
+        [(c++) #"~(~ unit 'name).cpp"])))
 
 ;; API
 (define-method cgen-unit-init-name ((unit <cgen-unit>))
@@ -318,7 +329,7 @@
 ;; Call thunk while binding current-output-port to a temp file,
 ;; then calls (finisher tmpfile file).
 (define (cgen-with-output-file file finisher thunk)
-  (call-with-temporary-file 
+  (call-with-temporary-file
    (^[port tmpfile]
      (with-output-to-port port thunk)
      (close-output-port port)
@@ -359,7 +370,7 @@
                [(#\!) (display #\X) (loop (read-char))]
                [(#\<) (display "_LT") (loop (read-char))]
                [(#\>) (display "_GT") (loop (read-char))]
-               [(#\* #\> #\@ #\$ #\% #\^ #\& #\* #\+ #\= #\. #\/ #\~)
+               [(#\* #\> #\@ #\$ #\% #\^ #\& #\* #\+ #\= #\. #\/ #\~ #\:)
                 (display #\_)
                 (display (number->string (char->integer c) 16))
                 (loop (read-char))]
@@ -382,3 +393,8 @@
 ;; Escape  '*/' so that str can be inserted safely within a comment.
 (define (cgen-safe-comment str)
   (regexp-replace-all* (x->string str) #/\/\*/ "/ *" #/\*\// "* /"))
+
+;; Useful to include sexp in C comment.  Strip #<identifier ...> away
+;; for the readability.
+(define (cgen-safe-comment-sexp obj)
+  (cgen-safe-comment (write-to-string (unwrap-syntax obj))))
