@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.interactive - useful stuff in the interactive session
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -42,6 +42,8 @@
   )
 (select-module gauche.interactive)
 
+(autoload gauche.uvector f64vector uvector-alias u64vector-ref)
+
 ;;;
 ;;; Apropos - search bound symbols matching given pattern
 ;;;
@@ -86,11 +88,11 @@
              (found mod symbol))))))
 
     (define (visible? sym)
-      (global-variable-bound? module sym))
-    
+      (module-binds? module sym))
+
     (define (found module symbol)
       (push! result
-             (format #f "~30s (~a~a)~%" symbol 
+             (format #f "~30s (~a~a)~%" symbol
                      (if (visible? symbol) "" "*")
                      (module-name module))))
 
@@ -155,27 +157,94 @@
     (format #t ")\n")
     (values)))
 
+(define-method describe ((r <rational>))
+  (describe-common r)
+  (when (ratnum? r)
+    (format #t "  inexact: ~s\n" (inexact r))
+    (and-let* ([abs-r (abs r)]
+               [ (> abs-r 1) ]
+               [intpart (floor abs-r)]
+               [mixed `(+ ,intpart ,(- abs-r intpart))]
+               [mixed. (if (negative? r) `(- ,mixed) mixed)])
+      (format #t "    mixed: ~s\n" mixed.)))
+  (values))
+
+(define-method describe ((d <real>))
+  (describe-common d)
+  (when (flonum? d)
+    (let* ([components (decode-float d)]
+           [buf (f64vector d)])
+      (if (finite? d)
+        (format #t "  mantissa: #x~a~14,'0,'_,4:x   exponent: ~d"
+                (if (positive? (vector-ref components 2)) "+" "-")
+                (vector-ref components 0)
+                (vector-ref components 1))
+        (format #t "  ~s" d))
+      (format #t "    hex: #x~16,'0,'_,4:x\n"
+              (u64vector-ref (uvector-alias <u64vector> buf) 0)))
+    (when (finite? d)
+      (format #t "  exact: ~s\n" (exact d))))
+  (values))
+
+(define-method describe ((c <complex>))
+  (describe-common c)
+  (when (and (finite? (real-part c)) (finite? (imag-part c)))
+    (let1 ang (angle c)
+      (format #t "  polar: ~s@~s  (@~spi)\n" (magnitude c) ang
+              (/ ang (* 4 (atan 1))))))
+  (values))
+
 (define-method describe ((g <generic>))
-  (next-method) ; common object description
+  (describe-common g)
+  (describe-slots g)
   (print "methods:")
   (dolist [m (~ g'methods)]
     (let ([spnames (map class-name (~ m'specializers))]
-          [has-optional? (~ m'optional)])
-      (format #t "  ~s\n"
+          [has-optional? (~ m'optional)]
+          [srcloc (source-location m)])
+      (format #t "  ~20s ~a\n"
               (if has-optional?
                 (append spnames '_) ; this works even spnames is ()
-                spnames))))
+                spnames)
+              (if srcloc
+                (format "; ~s:~d" (car srcloc) (cadr srcloc))
+                ""))))
   (and-let1 dis ((with-module gauche.object generic-dispatcher-info) g)
     (format #t "dispatcher:\n  ~s\n" dis))
   (values))
 
+(define-method describe ((m <method>))
+  (describe-common m)
+  (and-let1 source (source-location m)
+    (format #t "Defined at ~s:~d\n" (car source) (cadr source)))
+  (describe-slots m)
+  (values))
+
 (define-method describe ((p <procedure>))
   (describe-common p)
-  (if-let1 source (source-location p)
+  (and-let1 source (source-location p)
     (format #t "Defined at ~s:~d\n" (car source) (cadr source)))
+  (format #t "type: ~s\n" (procedure-type p))
   (describe-slots p)
   (values))
 
+(define-method describe ((m <macro>))
+  (describe-common m)
+  (and-let1 source (assq-ref (~ m'info-alist) 'source-info)
+    (format #t "Defined at ~s:~d\n" (car source) (cadr source)))
+  (describe-slots m)
+  (values))
+
+(define-method describe ((c <compound-condition>))
+  ;; TODO: We might want to customize simple condition as well to
+  ;; get better display
+  (describe-common c)
+  (let loop ([i 0] (cs (~ c'%conditions)))
+    (unless (null? cs)
+      (format #t "Condition[~d]: " i)
+      (describe (car cs))
+      (loop (+ i 1) (cdr cs)))))
+
 (define d describe)
 
 ;;;
@@ -190,8 +259,11 @@
 ;; modules in REPL without losing access to the history.  So we "inject"
 ;; those variables into #<module gauche>.  It is not generally recommended
 ;; way, though.
-;; We also export those history variables, so the modules that does not
-;; inherit gauche can still use them by (import gauche :only (*1 ...)).
+;;
+;; We also export those history variables from gauche.base, if it has already
+;; been loaded.  Moduels that imports gauche.base should see all variables
+;; defined in #<module gauche>.  This runtime hack is also a kludge.
+
 (define-in-module gauche *1 #f)
 (define-in-module gauche *1+ '())
 (define-in-module gauche *2 #f)
@@ -204,8 +276,8 @@
   (display "*2: ") (repl-print *2) (newline)
   (display "*3: ") (repl-print *3) (newline)
   (values))
-(with-module gauche
-  (export *1 *1+ *2 *2+ *3 *3+ *e *history))
+(and-let1 m (find-module 'gauche.base)
+  (eval '(export *1 *1+ *2 *2+ *3 *3+ *e *history) m))
 
 (define (%set-history-expr! r)
   (unless (null? r)
@@ -215,6 +287,9 @@
 
 (define (%set-history-exception! e) (set! *e e))
 
+;; This is kluge for Windows deferred console creation.
+(define *line-edit-ctx* #f)
+
 ;; Will be extended for fancier printer
 (define (repl-print x) (write/ss x) (flush))
 
@@ -270,35 +345,62 @@
           [h    (sys-normalize-pathname h :absolute #t :expand #t)]
           [else (sys-normalize-pathname "~/.gosh_history" :expand #t)])))
 
-;; The variable *read-edit* is #f by default, #t if env var 
-;; GAUCHE_READ_EDIT is set.  It is also controlled by -fread-edit or
-;; -fno-read-edit flag.
-(define-values (%prompter %reader %line-edit-ctx)
+;; Returns three values, prompter, reader, and line-edit-context,
+;; depending on the editable repl setting, terminal capability, and an
+;; optional user-provided prompt generator.
+;;
+;; Editable repl setting is indicated by the variable *read-edit*.  It is
+;; on by default, but can be turned off by the env var GAUCHE_NO_READ_EDIT
+;; or -fno-read-edit flag.
+;;
+(define (make-suitable-prompter/reader/ctx given-prompter)
+  ;; fallback reader when we don't use editable repl.  We sill handle
+  ;; toplevel commands.
+  (define vanilla-reader
+    (make-repl-reader read read-line consume-trailing-whitespaces))
+
+  ;; Returns a prompt string.
+  (define prompt-string-editable
+    (if given-prompter
+      (cut with-output-to-string given-prompter)
+      (cut default-prompt-string "$")))
+  (define prompt-string-noneditable
+    (if given-prompter
+      (cut with-output-to-string given-prompter)
+      (cut default-prompt-string ">")))
+
+  ;; Try creating editable reader.
   (receive (r rl skipper ctx)
-      (if (with-module gauche.internal *read-edit*)
-        (make-editable-reader (^[] (default-prompt-string "$"))
-                              (get-history-filename))
-        (values #f #f #f #f))
-    (define vanilla-reader
-      (make-repl-reader read read-line consume-trailing-whitespaces))
+    (if (with-module gauche.internal *read-edit*)
+      (make-editable-reader prompt-string-editable
+                            (get-history-filename))
+      (values #f #f #f #f))
     (if (and r rl skipper ctx)
       (let1 editing-reader (make-repl-reader r rl skipper)
         (set! *read-edit-state* 'editable)
+        ;; Editable repl
         (values (^[]
-                  ;; if input editor is running, prompt is shown by the reader
+                  ;; We only let prompter write prompt when we're in
+                  ;; vanilla mode.
                   (when (eq? *read-edit-state* 'vanilla)
-                    (display (default-prompt-string)) (flush)))
+                    (display (prompt-string-noneditable))))
                 (^[]
                   (if (eq? *read-edit-state* 'editable)
                     (editing-reader)
                     (vanilla-reader)))
                 ctx))
-      (values (^[] (display (default-prompt-string)) (flush))
+      ;; Non-editabl repl
+      (values (^[] (display (prompt-string-noneditable)) (flush))
               vanilla-reader
               #f))))
 
 ;; error printing will be handled by the original read-eval-print-loop
 (define (%evaluator expr env)
+  ;; Kludge - If read edit mode is ON, the final '\n' is consumed by
+  ;; the read-line/edit and the column count isn't reset.  To ensure
+  ;; proper indentation of output during eval, we forcibly reset the
+  ;; column count.
+  (set! (~ (current-output-port)'current-column) 0)
   (guard (e [else (%set-history-exception! e) (raise e)])
     (receive r (eval expr env)
       (%set-history-expr! r)
@@ -306,7 +408,7 @@
 
 ;; <write-controls> used for the printer.
 (define-constant *default-controls*
-  (make-write-controls :length 50 :level 10 :width 79
+  (make-write-controls :length 50 :level 10 :width 79 :string-length 256
                        :pretty (not (sys-getenv "GAUCHE_REPL_NO_PPRINT"))))
 (define *controls* *default-controls*)
 
@@ -331,38 +433,46 @@
        (set! *controls* (apply write-controls-copy *controls* kvs)))]))
 
 ;; This shadows gauche#read-eval-print-loop
+;; NB: We ignore reader argument,  The whole point of thsi procedure
+;; is to provide an editable reader, so the user doesn't need to customize it.
 (define (read-eval-print-loop :optional (reader #f)
                                         (evaluator #f)
                                         (printer #f)
                                         (prompter #f))
-  (let ([reader    (or reader %reader)]
-        [evaluator (or evaluator %evaluator)]
-        [printer   (or printer %printer)]
-        [prompter  (or prompter %prompter)])
+  (receive (%prompter %reader ctx)
+      (make-suitable-prompter/reader/ctx prompter)
+    (set! *line-edit-ctx* ctx)          ;kludge
     ((with-module gauche read-eval-print-loop)
-     reader evaluator printer prompter)))
+     %reader
+     (or evaluator %evaluator)
+     (or printer %printer)
+     %prompter)))
 
 ;;;
 ;;; Misc. setup
 ;;;
 
 ;; EXPERIMENTAL: windows console code page support for text.line-edit
-;; NB: ces (character encoding scheme) conversion is not implemented.
 (cond-expand
  [gauche.os.windows
-  (autoload os.windows sys-get-console-output-cp)
-  ;; check if we have a windows console.
-  (when (or (sys-isatty (standard-input-port))
-            (sys-isatty (standard-output-port))
-            (sys-isatty (standard-error-port)))
-    ;; wide character settings for text.line-edit
-    (if-let1 ctx %line-edit-ctx
+  (autoload os.windows
+            sys-get-console-output-cp
+            sys-has-windows-console?
+            sys-windows-terminal?
+            sys-windows-console-legacy?)
+  ;; wide character settings for text.line-edit
+  (if-let1 ctx *line-edit-ctx*
+    ;; check if we have a windows console
+    ;; (except for windows terminal (windows 10))
+    (when (and (sys-has-windows-console?)
+               (not (sys-windows-terminal?)))
       (case (sys-get-console-output-cp)
         [(65001)
          (set! (~ ctx 'wide-char-disp-setting 'mode) 'Surrogate)
          (set! (~ ctx 'wide-char-pos-setting  'mode) 'Surrogate)
          (set! (~ ctx 'wide-char-disp-setting 'wide-char-width) 2)
-         (set! (~ ctx 'wide-char-pos-setting  'wide-char-width) 1)
+         (set! (~ ctx 'wide-char-pos-setting  'wide-char-width)
+               (if (sys-windows-console-legacy?) 1 2))
          (set! (~ ctx 'wide-char-disp-setting 'surrogate-char-width) 2)
          (set! (~ ctx 'wide-char-pos-setting  'surrogate-char-width) 2)]
         [else ; 932 etc.
@@ -392,4 +502,3 @@
 
 ;; This might help first time users
 (define-in-module user help "Type ,help (comma and help) for help")
-
