@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.interactive.editable-reader
 ;;;
-;;;   Copyright (c) 2016-2018  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2016-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -36,12 +36,14 @@
 (define-module gauche.interactive.editable-reader
   (use text.console)
   (use util.match)
-  (use gauche.listener :only (complete-sexp?))
+  (use file.util)
   (export make-editable-reader))
 (select-module gauche.interactive.editable-reader)
 
-;; Delay loading line-edit in case editable console isn't available.
-(autoload text.line-edit <line-edit-context> read-line/edit)
+;; Delay loading some modules in case editable console isn't available.
+(autoload text.line-edit <line-edit-context> read-line/edit
+          load-line-edit-history save-line-edit-history)
+(autoload gauche.interactive.completion list-completions)
 
 ;; Internal API, to be used by gauche.interactive.
 ;; Because of toplevel commands, we can't just provide alternative 'read'
@@ -51,24 +53,38 @@
 ;  They are suitable to be passed to make-repl-reader.
 ;; NB: Currently we assume we use default console.  Might be useful
 ;; to allow other console (e.g. over pty).
-(define (make-editable-reader get-prompt-string)
+(define (make-editable-reader get-prompt-string hist-file)
   (if-let1 console (make-default-console :if-not-available #f)
+    ;; We wrap list-completions with lambda to delay loading completion module
     (let ([ctx (make <line-edit-context>
                  :console console
-                 :prompt (^[] (putstr console (get-prompt-string)))
-                 :input-continues (^s (not (%input-complete? s))))]
+                 :prompt (^[] (display (get-prompt-string)))
+                 :input-continues (^s (not (%input-complete? s)))
+                 :completion-word-constituent? char-word-constituent?
+                 :completion-lister (^[word gbuf start end]
+                                      (list-completions word gbuf start end)))]
           [buffer (open-input-string "")])
       (define (read-1 reader)
         (rec (try)
-          (let1 x (reader buffer)
+          (let1 x (guard (e [(is-a? e <read-error>)
+                             ;; discard remaining input so that subsequent
+                             ;; read won't be confused
+                             (set! buffer (open-input-string ""))
+                             (raise e)])
+                    (reader buffer))
             (if (eof-object? x)
               (let1 input (read-line/edit ctx)
                 (if (eof-object? input)
-                  input                     ; EOF is typed
-                  (begin 
+                  (begin                ; EOF is typed
+                    (if hist-file
+                      (save-line-edit-history ctx (expand-path hist-file)))
+                    input)
+                  (begin
                     (set! buffer (open-input-string (string-append input "\n")))
                     (try))))
               x))))
+      (when hist-file
+        (load-line-edit-history ctx (expand-path hist-file)))
       (values (read-1 read)
               (read-1 read-line)
               (^[] (consume-trailing-whitespaces buffer))
@@ -78,7 +94,6 @@
 ;; We have to handle both toplevel command (begins with comma, ends with
 ;; newline) and the complete sexp.
 (define (%input-complete? s)
-  (and (not (equal? s ""))
-       (if-let1 m (#/^\s*,(.*)/ s)
-         (not (#/^\s*$/ (m 1)))
-         (complete-sexp? s))))
+  (if-let1 m (#/^\s*,(.*)/ s)
+    (not (#/^\s*$/ (m 1)))
+    (complete-sexp? s)))
