@@ -3,10 +3,8 @@
 ;;;  The definition is moved to gauche.procedure.   This file is for
 ;;;  backward compatibility.
 ;;;
-;;;  $Id: let-opt.scm,v 1.3 2002/10/26 08:41:31 shirok Exp $
-;;;
 
 (define-module gauche.let-opt
   (extend gauche.procedure))
 
-(provide "gauche/let-opt")
+(warn "gauche.let-opt is obsoleted.  You don't need to use it anymore.  The functionality is supported in the core.")
