@@ -1,7 +1,7 @@
 ;;;
 ;;; module related utility functions.  to be autoloaded.
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -42,13 +42,13 @@
   ;; before this form.   Furthermore, the semantics of this form is ambigous
   ;; when used except top-level.  It's not very nice, so you should
   ;; avoid this form unless you really need it.
-  ;; NB: filter is in srfi-1, and we don't want to load it here.  Ugh.
+  ;; NB: filter is in SRFI-1, and we don't want to load it here.  Ugh.
   `(export
     ,@(let loop ([syms symbols] [r '()])
         (cond [(null? syms) (reverse! r)]
               [(not (symbol? (car syms)))
                (error "non-symbol in export-if-defined form:" (car syms))]
-              [(global-variable-bound? #f (car syms))
+              [(module-binds? #f (car syms))
                (loop (cdr syms) (cons (car syms) r))]
               [else (loop (cdr syms) r)]))))
 
@@ -78,12 +78,12 @@
               (guard [e (else "#<unprintable>")]
                 (format "~,,,,50:a" val)))))
   (let1 bindings (filter-map (^m (and-let* ([g (find-b m sym #t)]
-                                            [ (global-variable-bound? m sym) ])
-                                   (list m g (global-variable-ref m sym))))
+                                            [ (module-binds? m sym) ])
+                                   (list m g (module-binding-ref m sym))))
                              (all-modules))
     (if (null? bindings)
       (format #t "No known bindings for variable ~a.\n" sym)
-      (begin (format #t "Known binding~a for variable ~a:\n"
-                     (if (null? (cdr bindings)) "" "s") sym)
+      (begin (format #t "Known binding~p for variable ~a:\n"
+                     (length bindings) sym)
              (for-each (cut apply describe-binding <>) bindings))))
   (values))
