@@ -1,7 +1,7 @@
 ;;;
 ;;; propagate.scm - propagate slot option
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,7 +32,7 @@
 ;;;
 
 (define-module gauche.mop.propagate
-  (use srfi-1)
+  (use scheme.list)
   (use util.match)
   (export <propagate-meta> <propagate-mixin>)
   )
@@ -78,4 +78,3 @@
 (define-class <propagate-mixin> ()
   ()
   :metaclass <propagate-meta>)
-
