@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.package - package management
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -31,7 +31,6 @@
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;
 
-;; *EXPERIMENTAL*
 ;; The spec might be changed in future.  Do not count on the current version.
 ;;
 ;; This module handles package meta-information.
@@ -63,15 +62,16 @@
 ;; the package's installed gpd file.)
 ;;
 ;;   (define-gauche-package NAME
+;;     :repository URI-STRING               ; uniquely identifies this module
 ;;     :version VERSION                     ; version of this module
 ;;     :require ((PACKAGE VERSION-SPEC)...) ; dependency
 ;;     :maintainers (STRING ...)            ;
 ;;     :authors (STRING ...)
 ;;     :licenses (STRING ...)
 ;;     :homepage URI-STRING
-;;     :repository URI-STRING
 ;;     :description STRING
 ;;     :providing-modules (SYMBOL ..)     ; list of providing modules
+;;     :superseded-by URI-STRING          ; this package is no longer active
 ;;
 ;;     ;; The following attributes are added when *.gpd file is generated.
 ;;     :gauche-version VERSION            ; Gauche version used to build
@@ -79,7 +79,6 @@
 ;;
 ;;     ;; In the *.gpd file, the attributes marked with '*' above
 ;;     ;; are consolidated to the following plural attributes.
-;;     
 ;;     )
 ;;
 ;;   NAME _must_ match the filename sans suffix.
@@ -95,6 +94,7 @@
   (use file.util)
   (use util.match)
   (export <gauche-package-description>
+          <package-description-error>
           make-gauche-package-description
           path->gauche-package-description
           gauche-package-description-paths
@@ -114,6 +114,7 @@
 ;; 0.9.4.
 (define-class <gauche-package-description> ()
   ((name           :init-keyword :name        :init-value #f)
+   (repository     :init-keyword :repository  :init-value #f)
    (version        :init-keyword :version     :init-value #f)
    (description    :init-keyword :description :init-value #f)
    (require        :init-keyword :require     :init-value '())
@@ -121,13 +122,16 @@
    (authors        :init-keyword :authors     :init-value '())
    (licenses       :init-keyword :licenses    :init-value '())
    (homepage       :init-keyword :homepage    :init-value #f)
-   (repository     :init-keyword :repository  :init-value #f)
    (providing-modules :init-keyword :providing-modules :init-value '())
+   (superseded-by  :init-keyword :superseded-by :init-value #f)
    ;; The slots filled by cf-make-gpd
    (gauche-version :init-keyword :gauche-version :init-form (gauche-version))
    (configure      :init-keyword :configure   :init-value #f)
    ))
 
+;; API
+(define-condition-type <package-description-error> <error> #f)
+
 ;; API
 (define (make-gauche-package-description name
                                          :key (version "0.0")
@@ -141,6 +145,7 @@
                                               (gauche-version (gauche-version))
                                               (configure #f)
                                               (providing-modules '())
+                                              (superseded-by #f)
                                          :allow-other-keys unknown-keys)
   (when (not (null? unknown-keys))
     (warn "Package description has unrecognized key-value pairs: ~s\n"
@@ -153,45 +158,53 @@
   (check-maybe-string 'repository repository)
   (check-maybe-string 'description description)
   (check-symbol-list 'providing-modules providing-modules)
+  (check-maybe-string 'superseded-by superseded-by)
   (make <gauche-package-description>
     :name name :version version :require require :maintainers maintainers
     :authors authors :licenses licenses :homepage homepage
     :repository repository :description description
     :gauche-version gauche-version :configure configure
-    :providing-modules providing-modules))
+    :providing-modules providing-modules :superseded-by superseded-by))
 
 (define (check-maybe-string key val)
   (unless (or (string? val) (not val))
-    (errorf "String or #f is required for ~a, but got: ~s" key val)))
+    (errorf <package-description-error>
+            "String or #f is required for ~a, but got: ~s" key val)))
 
 (define (check-string-list key val)
   (unless (every string? val)
-    (errorf "String list is required for ~a, but got: ~s" key val)))
+    (errorf <package-description-error>
+            "String list is required for ~a, but got: ~s" key val)))
 
 (define (check-symbol-list key val)
   (unless (every symbol? val)
-    (errorf "Symbol list is required for ~a, but got: ~s" key val)))
+    (errorf <package-description-error>
+            "Symbol list is required for ~a, but got: ~s" key val)))
 
 (define (check-require-syntax req)
   (define (check-require-1 clause)
     (match clause
       [((? string?) v)
        (unless (valid-version-spec? v)
-         (error "Invalid version spec in require clause:" clause))]
-      [else (error "Invalid require clause:" clause)]))
+         (error <package-description-error>
+                "Invalid version spec in require clause:" clause))]
+      [else (error <package-description-error>
+                   "Invalid require clause:" clause)]))
   (match req
     [(x ...) (for-each check-require-1 x)]
-    [else (error "Invalid require form:" req)]))
+    [else (error <package-description-error> "Invalid require form:" req)]))
 
 ;; API
 (define (path->gauche-package-description path)
-  (guard (e [(or (<io-error> e) (<read-error> e))
-             (errorf "couldn't read the package description ~s: ~a"
+  (guard (e [(or (<io-error> e) (<read-error> e) (<system-error> e))
+             (errorf <package-description-error>
+                     "couldn't read the package description ~s: ~a"
                      path (ref e 'message))])
     (match (file->sexp-list path)
       [(('define-gauche-package (? string? name) . attrs))
        (apply make-gauche-package-description name attrs)]
-      [_ (errorf "malformed gauche package description file. \
+      [_ (errorf <package-description-error>
+                 "malformed gauche package description file. \
                   It must contain a define-gauche-package form and \
                   nothing else: ~a" path)])))
 
