@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.package.fetch - fetch a package
 ;;;
-;;;   Copyright (c) 2004-2016  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -31,7 +31,6 @@
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;
 
-;; *EXPERIMENTAL*
 ;; gauche.package.fetch module is intended to automate fetching package
 ;; distribution from the Net.
 ;;
@@ -44,27 +43,26 @@
   (use rfc.uri)
   (use file.util)
   (use gauche.package.util)
-  (use gauche.parameter)
   (export gauche-package-ensure))
 (select-module gauche.package.fetch)
 
 ;; Default programs
-(define *wget-program*     (find-file-in-paths "wget"))
-(define *ncftpget-program* (find-file-in-paths "ncftpget"))
+(define *wget-program*     (%find-file-in-paths "wget"))
+(define *ncftpget-program* (%find-file-in-paths "ncftpget"))
 
 (define (gauche-package-ensure uri :key (config '()))
   (let* ([build-dir (assq-ref config 'build-dir ".")]
-         [wget      (assq-ref config 'wget *wget-program*)]
-         [ncftpget  (assq-ref config 'ncftpget *ncftpget-program*)]
+         [wget      (%fix-path (assq-ref config 'wget *wget-program*))]
+         [ncftpget  (%fix-path (assq-ref config 'ncftpget *ncftpget-program*))]
          [dest      (build-path build-dir (sys-basename uri))])
     (rxmatch-case uri
       (#/^https?:/ (#f)
                    (sys-unlink dest)
-                   (run #"~wget -P \"~build-dir\" \"~uri\"")
+                   (run (%shell #"~wget -P ~(%fix-path build-dir) \"~uri\""))
                    dest)
       (#/^ftp:/ (#f)
                 (guard (e (else (sys-unlink dest) (raise e)))
-                  (run #"~ncftpget -c \"~uri\" > \"~dest\""))
+                  (run (%shell #"~ncftpget -c \"~uri\" > ~(%fix-path dest)")))
                 dest)
       (else
        (unless (file-is-readable? uri)
