@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.pp - pretty printer
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -31,13 +31,14 @@
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;
 
-;; Experimental
-
 (define-module gauche.pputil
   (use util.match)
   (export pprint))
 (select-module gauche.pputil)
 
+(autoload gauche.collection size-of)
+(autoload gauche.dictionary dict->alist dict-comparator)
+
 ;; List printing modes:
 ;;
 ;; oneline:
@@ -83,10 +84,7 @@
 (define-inline (rp-length c) (~ c 'controls 'length))
 (define-inline (rp-level c)  (~ c 'controls 'level))
 (define-inline (rp-width c)  (~ c 'controls 'width))
-
-(define simple-obj?
-  (any-pred number? boolean? char? port? symbol? null?
-            (cut member <> '("" #()))))
+(define-inline (rp-indent c) (~ c 'controls 'indent))
 
 ;; Maybe-ish monadic comparison
 (define-inline (=* a b) (and a b (= a b)))
@@ -98,15 +96,13 @@
   (and a b (if (null? args) (- a b) (apply -* (- a b) args))))
 (define-inline (min* a b) (if a (if b (min a b) a) b))
 
-;; Recurse to the system's writer to handle objects other than
-;; lists and vectors.  We want to pass down the controls (for
-;; print-base etc.), but the system's writer directly recurse into
-;; %pretty-print if print-pretty is true, causing infinite loop.
-;; So we drop pretty-print.
-(define (rec-writer c)
-  (let ([w  (~ c'writer)]
-        [c2 (write-controls-copy (~ c'controls) :pretty #f)])
-    (^x (w x c2))))
+;; Render OBJ into a string.  The resulting string is then used with
+;; layout-simple.  OBJ must be a 'simple' object (system's writer won't
+;; immediately call to %pretty-print on it).
+(define (stringify obj c)
+  (let ([w (~ c'writer)]
+        [c2 (~ c'controls)])
+    (write-to-string obj (^x (w x c2)))))
 
 (define (need-label? obj c) (> (hash-table-get (rp-shared c) obj 0) 1))
 (define (has-label? obj c) (<= (hash-table-get (rp-shared c) obj 1) 0))
@@ -156,37 +152,39 @@
 ;; layout :: (Obj, Integer, Context) -> Layouter
 (define (layout obj level c)
   (cond [(has-label? obj c) (layout-ref obj c)]
-        [(simple-obj? obj) (layout-simple (write-to-string obj (rec-writer c)))]
-        [(>=* level (rp-level c)) (layout-simple "#")]
+        [(and (>=* level (rp-level c))
+              (or (pair? obj)
+                  (vector? obj)
+                  (uvector? obj)
+                  (is-a? obj <dictionary>)))
+         (layout-simple "#")]
         [else (layout-misc obj (cute layout <> (+ level 1) c) c)]))
 
 ;; layout-misc :: (Obj, (Obj -> Layouter), Context) -> Layouter
 (define (layout-misc obj rec c)
 
-  ;; mapi :: (Obj -> Layouter), Vector -> Layouter
+  ;; mapi :: (Obj -> Layouter), Vector -> [Layouter]
   (define (mapi fn vec)
     (let* ([s (vector-length vec)]
            [rs (map (^i (fn (vector-ref vec i)))
                     (iota (min* s (rp-length c))))])
       (if (>=* s (rp-length c)) `(,@rs ,dots) rs)))
 
-  ;; mapu :: (Obj -> Layouter), UVector -> Layouter
+  ;; mapu :: (Obj -> Layouter), UVector -> [Layouter]
   (define (mapu fn vec)
     (let* ([s (uvector-length vec)]
            [rs (map (^i (fn (uvector-ref vec i)))
                     (iota (min* s (rp-length c))))])
       (if (>=* s (rp-length c)) `(,@rs ,dots) rs)))
-  
-  ;; map+ :: (Obj -> Layouter), List -> Layouter
+
+  ;; map+ :: (Obj -> Layouter), List -> [Layouter]
   ;; map considering dotted list, print-length, and shared structure
   (define (map+ fn lis)
     (let loop ([lis lis] [len 0] [rs '()])
       (match lis
         [() (reverse rs)]
-        [((and (or 'quote 'quasiquote 'unquote 'unquote-splicing) p) _)
-         (if (has-label? (cdr lis) c)
-           (reverse (list* (layout-ref (cdr lis) c) dot (fn p) rs))
-           (reverse (list* (layout-shorthand lis fn c) dot rs)))]
+        [(? (cut may-layout-shorthand? <> c))
+         (reverse (list* (layout-shorthand lis fn c) dot rs))]
         [(l . lis)
          (if (>=* len (rp-length c))
            (reverse `(,dots ,@rs))
@@ -199,10 +197,7 @@
         [x (reverse `(,(fn x) ,dot ,@rs))])))
 
   (cond [(pair? obj)
-         (or (and (pair? (cdr obj))
-                  (null? (cddr obj))
-                  (not (has-label? (cdr obj) c))
-                  (memq (car obj) '(quote quasiquote unquote unquote-splicing))
+         (or (and (may-layout-shorthand? obj c)
                   (layout-shorthand obj rec c))
              (layout-list (sprefix obj "(" c) (map+ rec obj) c))]
         [(vector? obj) (layout-list (sprefix obj "#(" c) (mapi rec obj) c)]
@@ -210,11 +205,16 @@
          (let1 tag (rxmatch-substring
                      (#/[csuf]\d+/ (x->string (class-name (class-of obj)))))
            (layout-list (sprefix obj (format "#~a(" tag) c) (mapu rec obj) c))]
+        [(is-a? obj <dictionary>)
+         (layout-dict obj (map+ rec (dict->alist obj)) c)]
         [else
-         (layout-simple (sprefix obj (write-to-string obj (rec-writer c)) c))]))
+         (layout-simple (sprefix obj (stringify obj c) c))]))
 
 ;; :: Layouter
-(define dots (^[w m] '("...." . 4)))
+(define dots
+  (let* ([elli (with-module gauche.internal (string-ellipsis))]
+         [val (cons elli (string-length elli))])
+    (^[w m] val)))
 (define dot  (^[w m] '("." . 1)))
 
 ;; layout-simple :: String -> Layouter
@@ -230,11 +230,11 @@
 ;;   The second argument is to recursively layout X.
 ;;   The case that cdr of the form is shared should be handled by the caller.
 (define (layout-shorthand form rec c)
-  (let* ([pfx (sprefix form 
+  (let* ([pfx (sprefix form
                        (assq-ref '((quote . "'")
                                    (quasiquote . "`")
                                    (unquote . ",")
-                                   (unquote-splicing . ",@")) 
+                                   (unquote-splicing . ",@"))
                                  (car form))
                        c)]
          [plen (string-length pfx)]
@@ -245,6 +245,13 @@
            (match-let1 (s . w) (inner (-* room plen) memo)
              (cons (list pfx s) (and w (+ w plen)))))))
 
+(define (may-layout-shorthand? obj c)
+  (and (pair? obj)
+       (pair? (cdr obj))
+       (null? (cddr obj))
+       (not (has-label? (cdr obj) c))
+       (memq (car obj) '(quote quasiquote unquote unquote-splicing))))
+
 ;; layout-list :: (String, [Layouter], Context) -> Layouter
 (define (layout-list prefix elts c)
   (let1 plen (string-length prefix)
@@ -252,6 +259,25 @@
            (match-let1 (s . w) (do-layout-elements (-* room plen) memo elts)
              (cons (cons prefix (reverse s)) (and w (+ w plen)))))))
 
+;; layout-dict :: (Dict, [Layouter], Context) -> Layouter
+(define (layout-dict dict content-layouters c)
+  (let* ([cname (regexp-replace #/^<(.*)>$/
+                                (symbol->string (class-name (class-of dict)))
+                                (^m (m 1)))]
+         [cmpname (if-let1 cmp (dict-comparator dict)
+                    (or (~ cmp 'name) 'custom)
+                    '||)]
+         [tag (format "~a ~a[~d](" cname cmpname (size-of dict))]
+         [tag-layouter (layout-simple tag)]
+         [prefix "#<"]
+         [plen (string-length prefix)])
+    (memo^ [room memo]
+           (match-let1 (s . w)
+               (do-layout-elements (-* room plen 1) memo
+                                   (cons tag-layouter content-layouters))
+             (cons `(,prefix ,@(reverse s) ">")
+                   (and w (+ w plen 1)))))))
+
 ;; sprefix :: (Object, String, Context) -> String
 (define (sprefix obj s c)
   (if (need-label? obj c)
@@ -294,7 +320,7 @@
 ;; B's insert a newline and a proper indentation.
 (define (render stree indent port)
   (define (next-line col) (newline port) (dotimes [i col] (display " " port)))
-  (define (drop-while pred xs) ; avoid depending on srfi-1 (for now)
+  (define (drop-while pred xs) ; avoid depending on SRFI-1 (for now)
     (cond [(null? xs) '()]
           [(pred (car xs)) (drop-while pred (cdr xs))]
           [else xs]))
@@ -310,7 +336,9 @@
            [(s . es)  (render s ind port) (loop es)])))]
     [else (display stree port)]))
 
-;; Stitch together.  This is called from Scm_Write() family.
+;; Stitch together.  This is called from Scm_Write() family, replacing
+;; internal write_rec().   Note that we already done pass 1 (write_walk)
+;; to set up shared-table for shared/circular structure reference.
 (define-in-module gauche (%pretty-print obj port shared-table controls)
   (assume shared-table)
   (let1 context (make <pp-context>
@@ -318,8 +346,10 @@
                   :shared shared-table)
     (let* ([layouter (layout obj 0 context)]
            [memo (make-memo-hash)]
-           [fstree (car (layouter (rp-width context) memo))])
-      (render fstree 0 port))))
+           [fstree (car (layouter (-* (rp-width context)
+                                      (rp-indent context))
+                                  memo))])
+      (render fstree (rp-indent context) port))))
 
 ;; Write controls used by pprint
 (define *default-controls* (make-write-controls :length #f
@@ -331,12 +361,13 @@
 (define (pprint obj
                 :key (port (current-output-port))
                      (controls *default-controls*)
-                     width length level
+                     width length level indent
                      ((:newline nl) #t))
   (let1 controls (write-controls-copy controls
                                       :width width
                                       :length length
                                       :level level
-                                      :pretty #t)
+                                      :pretty #t
+                                      :indent indent)
     (write obj port controls)
     (when nl (newline port))))
