@@ -1,7 +1,7 @@
 ;;;
 ;;; redefutil.scm - class redefinition protocol (autoloaded)
 ;;;
-;;;   Copyright (c) 2003-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2003-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,7 +34,7 @@
 ;; This file is autoloaded
 
 (select-module gauche.object)
-(use srfi-1)           ;; used by instance update protocol
+(use scheme.list)           ;; used by instance update protocol
 
 ;;----------------------------------------------------------------
 ;; Class Redefinition protocol
@@ -57,14 +57,23 @@
 
 (define-generic class-redefinition)
 (define-method class-redefinition ((old <class>) (new <class>))
+  ;; Be careful to access slot values of OLD.  If OLD has a metaclass
+  ;; which itself is redefined, using slot-ref triggers update-instance
+  ;; on OLD and we lose old'redefined.
   (for-each (^m (if (is-a? m <accessor-method>)
                   (delete-method! (slot-ref m 'generic) m)
                   (update-direct-method! m old new)))
-            (class-direct-methods old))
+            (slot-ref-using-class (current-class-of old)
+                                  old
+                                  'direct-methods))
   (for-each (^[sup] (%delete-direct-subclass! sup old))
-            (class-direct-supers old))
+            (slot-ref-using-class (current-class-of old)
+                                  old
+                                  'direct-supers))
   (for-each (^[sub] (update-direct-subclass! sub old new))
-            (class-direct-subclasses old))
+            (slot-ref-using-class (current-class-of old)
+                                  old
+                                  'direct-subclasses))
   )
 
 (define-generic update-direct-subclass!)
@@ -82,7 +91,7 @@
             [(eq? (car args) :metaclass)
              (loop (cddr args) (list* metaclass :metaclass r))]
             [else (loop (cddr args) (list* (cadr args) (car args) r))])))
-  
+
   (let* ([initargs (slot-ref sub 'initargs)]
          [supers   (new-supers (class-direct-supers sub))]
          ;; NB: this isn't really correct!
@@ -195,4 +204,3 @@
 (define-in-module gauche class-redefinition class-redefinition)
 (define-in-module gauche update-direct-subclass! update-direct-subclass!)
 (define-in-module gauche change-object-class change-object-class)
-
