@@ -1,7 +1,7 @@
 ;;;
 ;;; sysutil - Auxiliary system interface, autoloaded
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,15 +38,12 @@
 ;; Src/libsys.scm is pre-compiled at the time of distribution,
 ;; so we cannot have platform-dependent cond-expand in it.
 
-;; These are better to be in src/scmlib.scm, but right now we don't have
-;; a nice way to make cond-expand work (when compiling src/scmlib.scm
-;; cond-expand uses the host gosh's feature set, not the target gosh's.)
 (cond-expand
  [gauche.sys.select
   (define (sys-fdset . pfs)
     (list->sys-fdset pfs))
   (define (sys-fdset->list fdset)
-    (check-arg (cut is-a? <> <sys-fdset>) fdset)
+    (assume-type fdset <sys-fdset>)
     (do ([i (sys-fdset-max-fd fdset) (- i 1)]
          [fds '() (if (sys-fdset-ref fdset i) (cons i fds) fds)])
         [(< i 0) fds]
@@ -112,3 +109,61 @@
 
     (resolve '() (append! (if (absolute? path) '() (decompose (sys-getcwd)))
                           (decompose path))))])
+
+;; Find a file that satisfies a predicate in given list of paths.
+;; This is a generalization of what execvp does to search executable from PATH.
+;; This used to be file.util#find-file-in-paths.  We moved it here so that
+;; sys-exec can rely on it, as we switch to execve(2) from execvp(3).
+;;
+;; The default value handling of pred and extensions are a bit involved.
+;; The default of pred is to check the file is executable, and the default of
+;; extensions is '().  However, on Windows, if both are omitted, we default
+;; extensions to be '("exe" "com" "cmd" "bad"), to mimic the behaivor of
+;; execvp().
+(define (sys-find-file name
+                       :key (paths (cond [(sys-getenv "PATH")
+                                          => (cut string-split <>
+                                                  (cond-expand
+                                                   [gauche.os.windows #\;]
+                                                   [else #\:]))]
+                                         [else '()]))
+                            pred
+                            extensions)
+  (define real-predicate
+    (if (undefined? pred)
+      (cute sys-access <> X_OK)
+      pred))
+  (define real-extensions
+    (if (undefined? extensions)
+      (cond-expand
+       [gauche.os.windows (if (undefined? pred)
+                            '("exe" "com" "cmd" "bat")
+                            '())]
+       [else '()])
+      (assume-type extensions (<List> <string>)
+                   "List of strings expected, but got:" extensions)))
+  (define names
+    (if (null? real-extensions)
+      `(,name)
+      (cons name
+            (map (^e (string-append name "." e)) real-extensions))))
+  (define (try n) (and (real-predicate n) n))
+  (define (abspath? path)               ;dupe of file.util#absolute-path?
+    (cond-expand
+     [gauche.os.windows (#/^[\/\\]|^[A-Za-z]:/ path)]
+     [else              (#/^\// path)]))
+  (define (mkpath dir file)             ;dupe of file.util#build-path
+    (string-append (sys-normalize-pathname dir)
+                   (cond-expand
+                    [gauche.os.windows "\\"]
+                    [else              "/"])
+                   file))
+
+  (if (abspath? name)
+    (any try names)
+    (and-let1 found
+        (let loop ((paths paths))
+          (and (not (null? paths))
+               (or (any try (map (cute mkpath (car paths) <>) names))
+                   (loop (cdr paths)))))
+      (sys-normalize-pathname found :absolute #t))))
