@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.test - test framework
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -61,11 +61,11 @@
 ;;  (test-end)
 ;;
 ;; To run a test interactively, just load the file.
-;; It is also recommended to have a "test" target in your Makefile, so that
+;; It is also recommended to have a "check" target in your Makefile, so that
 ;; the user of your program can run a test easily.  The rule may look like
 ;; this:
 ;;
-;;   test :
+;;   check :
 ;;        gosh my-feature-test.scm > test.log
 ;;
 ;; If stdout is redirected to other than tty, all the verbose logs will go
@@ -83,15 +83,25 @@
 
 (define-module gauche.test
   (export test test* test-start test-end test-running? test-section test-log
-          test-module test-script
-          test-error test-one-of test-none-of
-          test-check test-record-file test-summary-check
+          test-module test-script test*/diff
+          test-error test-one-of test-none-of test-truthy
+          test-check test-check-diff
+          test-include-r7
+
+          test-report-failure test-report-failure-plain
+          test-report-failure-diff
+          test-record-file test-summary-check
           *test-error* *test-report-error* test-error? prim-test
-          test-count++ test-pass++ test-fail++))
+          test-count++ test-pass++ test-fail++
+
+          test-remove-files test-with-temporary-directory))
 (select-module gauche.test)
 
-;; Autoload test-script to avoid depending other modules
+;; Autoloads to avoid depending other modules
 (autoload "gauche/test/script" test-script)
+(autoload "gauche/test/diff"
+          (:macro test*/diff) test-check-diff test-report-failure-diff)
+(autoload "gauche/test/include" (:macro test-include-r7))
 
 ;; An object to represent error.  This class isn't exported; the user
 ;; must use `test-error' procedure to create an instance.
@@ -140,6 +150,14 @@
 ;; API
 (define (test-none-of . choices) (make <test-none-of> :choices choices))
 
+;; An object to represent "any true value"
+(define-class <test-truthy> () ())
+(define-method write-object ((obj <test-truthy>) out)
+  (format out "#<test-truthy>"))
+
+;; API
+(define (test-truthy) (make <test-truthy>))
+
 ;; API
 ;; We don't use generic function dispatch (at least for the time being),
 ;; to make it easy to troubleshoot when object system gets messed up.
@@ -162,6 +180,7 @@
         [(is-a? expected <test-none-of>)
          (every (lambda (choice) (not (test-check choice result fallback)))
                 (slot-ref expected 'choices))]
+        [(is-a? expected <test-truthy>) result]
         [else (fallback expected result)]))
 
 (define *test-error* (make <test-error>)) ;DEPRECATED
@@ -172,6 +191,8 @@
 (define (test-record-file file) (set! *test-record-file* file))
 
 ;; List of discrepancies
+;; ((<report> <msg> <expected> <result>) ...)
+;; (We intentially avoid using parameters)
 (define *discrepancy-list* '())
 
 (define *test-counts* (vector 0 0 0 0)) ; total/pass/fail/abort
@@ -180,9 +201,10 @@
   (vector-set! *test-counts* 0 (+ (vector-ref *test-counts* 0) 1)))
 (define (test-pass++)
   (vector-set! *test-counts* 1 (+ (vector-ref *test-counts* 1) 1)))
-(define (test-fail++ msg expected result)
+(define (test-fail++ msg expected result :optional (report test-report-failure))
   (vector-set! *test-counts* 2 (+ (vector-ref *test-counts* 2) 1))
-  (set! *discrepancy-list* (cons (list msg expected result) *discrepancy-list*)))
+  (set! *discrepancy-list*
+        (cons (list report msg expected result) *discrepancy-list*)))
 (define (format-summary)
   (format "Total: ~5d tests, ~5d passed, ~5d failed, ~5d aborted.\n"
           (vector-ref *test-counts* 0)
@@ -218,11 +240,34 @@
 
 ;; Tests ------------------------------------------------------------
 
+;; test msg expect thunk :optional check report hook
+;;   check is called (check expected actual-result).  default is test-check.
+;;   report is called (report msg expected actual-result) when failed.
+;;                  The output must end with newline.
+;;   hook is called (hook 'fail|'pass msg expected actual-result)
+;; NB: In 0.9.10, we have a signature ':optional check hook' (no report).
+;; We adopted ':optional check report hook', for it makes more sense.
+;; To keep the backward compatibility, we check the arity of the second
+;; optional argument.
+
 ;; Primitive test.  This uses neither with-error-handler nor the
 ;; object system, so it can be used _before_ those constructs are tested.
 (define (prim-test msg expect thunk . args)
-  (let ([cmp  (if (pair? args) (car args) test-check)]
-        [hook (and (pair? args) (pair? (cdr args)) (cadr args))])
+  (let ([cmp  (or (and (pair? args) (car args))
+                  test-check)]
+        [report (or (and (pair? args) (pair? (cdr args))
+                         (cadr args))
+                    test-report-failure)]
+        [hook (and (pair? args) (pair? (cdr args)) (pair? (cddr args))
+                   (caddr args))])
+    ;; TRANSIENT: Backward compatibility of 'hook'.  Remove by 1.0.
+    (when (and (not (eq? report test-report-failure))
+               (not hook)
+               (eqv? (arity report) 4))
+      (warn "gauche.test: `test' is called with old signature (hook).\n")
+      (set! hook report)
+      (set! report test-report-failure))
+    ;; End transient code
     (format/ss #t "test ~a, expects ~s ==> " msg expect)
     (flush)
     (test-count++)
@@ -232,8 +277,10 @@
              (test-pass++)
              (when hook (hook 'pass msg expect r))]
             [else
-             (format/ss #t "ERROR: GOT ~S\n" r)
-             (test-fail++ msg expect r)
+             (display "ERROR: GOT ")
+             (report msg expect r)
+             (newline)
+             (test-fail++ msg expect r report)
              (when hook (hook 'fail msg expect r))])
       (flush))))
 
@@ -252,7 +299,7 @@
          args))
 
 ;; A convenient macro version
-;; We use er-macro-transformer, so test* should be used after macro
+;; We use er-macro-transformer, so test* should be used after the macro
 ;; subsystem is tested with more primitive framework.
 (define-syntax test*
   (er-macro-transformer
@@ -296,7 +343,7 @@
     (hash-table-for-each (module-table mod)
                          (lambda (sym val)
                            (guard (_ (else (push! bad-autoload sym)))
-                             (global-variable-ref mod sym))))
+                             (module-binding-ref mod sym))))
     ;; 2. Check if all exported symbols are properly defined.
     ;; We create an anonymous moudle and import the tested module.  By this
     ;; way, we can test renaming export (in which case, the exported name
@@ -308,7 +355,7 @@
         (eval `(extend) m)
         (for-each (lambda (sym)
                     (guard (_ [else (push! bad-export sym)])
-                      (global-variable-ref m sym)))
+                      (module-binding-ref m sym)))
                   (module-exports mod))))
     ;; 3. Check if all global references are resolvable, and if it is
     ;; called, gets valid number of arguments.
@@ -379,7 +426,7 @@
 (define (toplevel-closures module)
   (filter closure?
           (map (lambda (sym)
-                 (global-variable-ref module sym #f))
+                 (module-binding-ref module sym #f))
                (hash-table-keys (module-table module)))))
 
 ;; Combs closure's instruction list to extract references for the global
@@ -436,7 +483,7 @@
 (define (arity-invalid? gref numargs src-code)
   (and-let* ([ numargs ]
              ;; TODO: What if GREF is nested identifier?
-             [proc (global-variable-ref
+             [proc (module-binding-ref
                     (slot-ref gref'module)
                     (unwrap-syntax gref)
                     #f)]
@@ -456,11 +503,20 @@
 
 ;; Logging and bookkeeping -----------------------------------------
 
-;; private global flag, true during we're running tests.
-;; (we avoid using parameters intentionally.)
-(define *test-running* #f)
+;; API: Default report procedure
+(define (test-report-failure msg expected actual)
+  (write actual))
 
-(define (test-running?) *test-running*)
+;; API: Use ~a instead of ~s
+(define (test-report-failure-plain msg expected actual)
+  (display actual))
+
+;; Private global stack, count test nesting level.
+;; ((<name> <file>) ...)
+;; (We avoid using parameters intentionally.)
+(define *test-nesting* '())
+
+(define (test-running?) (pair? *test-nesting*))
 
 (define (test-section msg)
   (let ([msglen (string-length msg)])
@@ -478,24 +534,29 @@
     (sys-isatty port)]))
 
 (define (test-start msg)
-  (set! *test-running* #t)
-  (let* ([s (format #f "Testing ~a ... " msg)]
-         [pad (make-string (max 3 (- 65 (string-length s))) #\space)])
-    (display s (current-error-port))
-    (display pad (current-error-port))
-    (flush (current-error-port))
-    (read-summary)
-    (prepare-summary)
-    (when (and (not-redirected? (current-error-port))
-               (not-redirected? (current-output-port)))
-      (newline (current-error-port))))
-  (set! *discrepancy-list* '())
-  (unless (and (not-redirected? (current-error-port))
-               (not-redirected? (current-output-port)))
-    (let ([msglen (string-length msg)])
-      (format #t "Testing ~a ~a\n" msg (make-string (max 5 (- 70 msglen)) #\=)))
-    (flush))
-  )
+  (push! *test-nesting* (list msg (current-load-path)))
+  (cond
+   [(null? (cdr *test-nesting*))
+    (let* ([s (format #f "Testing ~a ... " msg)]
+           [pad (make-string (max 3 (- 65 (string-length s))) #\space)])
+      (display s (current-error-port))
+      (display pad (current-error-port))
+      (flush (current-error-port))
+      (read-summary)
+      (prepare-summary)
+      (when (and (not-redirected? (current-error-port))
+                 (not-redirected? (current-output-port)))
+        (newline (current-error-port))))
+    (set! *discrepancy-list* '())
+    (unless (and (not-redirected? (current-error-port))
+                 (not-redirected? (current-output-port)))
+      (let ([msglen (string-length msg)])
+        (format #t "Testing ~a ~a\n" msg
+                (make-string (max 5 (- 70 msglen)) #\=)))
+      (flush))]
+   [else
+    (format #t "Nested testing[~a] ~a\n" (length *test-nesting*) msg)
+    (flush)]))
 
 ;; test-log fmt arg ...
 ;; The formatted output, prefixed by ";;",  goes to stdout for the logging.
@@ -508,6 +569,18 @@
 ;; test-end :key :exit-on-failure
 ;; avoid using extended formal list since we need to test it.
 (define (test-end . args)
+  (if (pair? *test-nesting*)
+    (if (pair? (cdr *test-nesting*))
+      (begin
+        (format #t "End nested testing[~a] ~a\n"
+                (length *test-nesting*) (caar *test-nesting*))
+        (pop! *test-nesting*))
+      (begin
+        (pop! *test-nesting*)
+        (%test-true-end args)))
+    (error "Test nesting count disrepancy; missing test-start?")))
+
+(define (%test-true-end args)
   (let ([e (current-error-port)]
         [o (current-output-port)]
         [exit-on-failure (get-keyword :exit-on-failure args #f)])
@@ -520,17 +593,19 @@
     (if (null? *discrepancy-list*)
       (fmt "passed.\n")
       (begin
-        (fmt "failed.\ndiscrepancies found.  Errors are:\n")
+        (fmt "failed.\n~a discrepanc~:*~@p found:\n" (length *discrepancy-list*))
         (for-each (lambda (r)
-                    (apply fmt "test ~a: expects ~s => got ~s\n" r))
+                    (apply (lambda (report msg expect actual)
+                             (fmt "test ~a: expects ~s => got ~a\n" msg expect
+                                  (with-output-to-string
+                                    (cut report msg expect actual))))
+                           r))
                   (reverse *discrepancy-list*))))
     (flush)
 
     (when *test-record-file*
       (write-summary))
 
-    (set! *test-running* #f)
-
     ;; the number of failed tests.
     (let ([num-failures (length *discrepancy-list*)])
       (when (and (> num-failures 0)
@@ -545,3 +620,28 @@
   (unless (and (zero? (vector-ref *test-counts* 2))
                (zero? (vector-ref *test-counts* 3)))
     (exit 1)))
+
+;; Temporary files and cleanup  --------------------------------------
+
+;; Remove files and directories.  Functionally same as 'remove-files' in
+;; file.util; but this can be used _before_ we test file.util.
+(define (test-remove-files . paths)
+  (define (remove-1 path)
+    (if (and (file-is-directory? path)
+             (not (eq? (slot-ref (sys-lstat path)'type) 'symlink)))
+      (begin (for-each (^e (unless (member e '("." ".."))
+                             (remove-1 (string-append path "/" e))))
+                       (sys-readdir path))
+             (sys-rmdir path))
+      (sys-unlink path)))
+  (for-each remove-1 paths))
+
+;; Create a fresh DIR, call THUNK (without cd'ing), then remove DIR.
+(define (test-with-temporary-directory dir thunk)
+  (test-remove-files dir)
+  (sys-mkdir dir #o755)
+  ;; Avoid relying on unwind-protect
+  (dynamic-wind
+    (^[] #f)
+    thunk
+    (^[] (test-remove-files dir))))
