@@ -1,7 +1,7 @@
 ;;;
 ;;; version.scm - deal with version numbers
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -31,6 +31,10 @@
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;
 
+;; NB: This module is autoloaded from the core (cond-expand) when 'package'
+;; clause is used with version spec.  If you modify this, keep the dependencies
+;; minimum.
+
 ;;; Defines a few procedures that deal with popular version
 ;;; numbers, such as 1.2.3.
 ;;;
@@ -40,7 +44,7 @@
 ;;; <principal-release> := <relnum>
 ;;; <post-subrelease>   := [.-] <relnum>
 ;;; <pre-subrelease>    := _ <relnum>
-;;; <relnum>       := [0-9A-Za-z]+
+;;; <relnum>       := [^._-]+
 ;;;
 ;;; Typically <relnum> is composed by numeric part and extension part.
 ;;; For example, "23a" is composed by an integer 23 and extension "a".
@@ -61,7 +65,7 @@
 ;;;      remove it from both.  Repeat this until the head of the lists
 ;;;      differ.
 ;;;   3. Now we have the following cases.
-;;;     (1) Both lists are empty: versions are the same.
+;;;     (1) Both lists are empty. Versions are the same.
 ;;;     (2) One list (A) is empty and the other list (B) has post-subrelease
 ;;;         at head: A is prior to B
 ;;;     (3) One list (A) is empty and the other list (B) has pre-subrelease
@@ -79,8 +83,6 @@
 ;;;    "1.1-patch112" < "1.2_alpha"
 
 (define-module gauche.version
-  (use srfi-11)                         ;let-values
-  (use srfi-13)
   (use util.match)
   (export relnum-compare
           version-compare
@@ -90,48 +92,46 @@
 (select-module gauche.version)
 
 (define (relnum-decompose vn)
-  (rxmatch-if (rxmatch #/^(\d*)(\w*)$/ vn) (#f num suffix)
+  (rxmatch-if (rxmatch #/^(\d*)([^-_.]*)$/ vn) (#f num suffix)
     (values (if (string=? num "") -1 (x->integer num)) suffix)
     (errorf "unparsable release number: ~s" vn)))
 
 (define (relnum-compare a b)
-  (let-values (((an ax) (relnum-decompose a))
-               ((bn bx) (relnum-decompose b)))
-    (cond ((= an bn)
+  (let-values ([(an ax) (relnum-decompose a)]
+               [(bn bx) (relnum-decompose b)])
+    (cond [(= an bn)
            (cond ((string=? ax bx) 0)
                  ((string<? ax bx) -1)
-                 (else 1)))
-          ((< an bn) -1)
-          (else 1))))
+                 (else 1))]
+          [(< an bn) -1]
+          [else 1])))
 
 (define (next-component ver)
-  (cond ((rxmatch #/[._-]/ ver)
+  (cond [(rxmatch #/[._-]/ ver)
          => (^m (values (rxmatch-before m)
-			(if (string=? (rxmatch-substring m) "_")
-			    'pre 'post)
-			(rxmatch-after m))))
-        (else (values ver #f #f))))
+                        (if (string=? (rxmatch-substring m) "_")
+                            'pre 'post)
+                        (rxmatch-after m)))]
+        [else (values ver #f #f)]))
 
 (define (version-compare va vb)
   (define (sep-cmp a-sep a-ver b-sep b-ver)
     (cond
-     ((not a-sep)
-      (cond ((eq? b-sep 'post) -1)
-            ((eq? b-sep 'pre) 1)
-            (else 0)))
-     ((eq? a-sep 'post)
-      (cond ((eq? b-sep 'post) (ver-cmp a-ver b-ver))
-            (else 1)))
-     (else
-      (cond ((eq? b-sep 'pre)  (ver-cmp a-ver b-ver))
-            (else -1)))))
+     [(not a-sep)
+      (cond [(eq? b-sep 'post) -1]
+            [(eq? b-sep 'pre) 1]
+            [else 0])]
+     [(eq? a-sep 'post)
+      (if (eq? b-sep 'post) (ver-cmp a-ver b-ver) 1)]
+     [else
+      (if (eq? b-sep 'pre)  (ver-cmp a-ver b-ver) -1)]))
   (define (ver-cmp a-ver b-ver)
-    (let*-values (((a-rel a-sep a-next) (next-component a-ver))
-                  ((b-rel b-sep b-next) (next-component b-ver))
-                  ((c) (relnum-compare a-rel b-rel)))
+    (let*-values ([(a-rel a-sep a-next) (next-component a-ver)]
+                  [(b-rel b-sep b-next) (next-component b-ver)]
+                  [(c) (relnum-compare a-rel b-rel)])
       (if (zero? c)
-          (sep-cmp a-sep a-next b-sep b-next)
-          c)))
+        (sep-cmp a-sep a-next b-sep b-next)
+        c)))
   (ver-cmp va vb))
 
 (define (version=? a b)  (zero? (version-compare a b)))
