@@ -1,7 +1,7 @@
 ;;;
 ;;; r7rs-setup - R7RS compatibility
 ;;;
-;;;   Copyright (c) 2013-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2013-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -45,8 +45,8 @@
 ;; in this module.
 (define-module r7rs.import
   (use util.match)
-  (use srfi-1)
-  (export (rename import r7rs-import))
+  (use scheme.list)
+  (export r7rs-import)
 
   ;; A trick - must be replaced once we have explicit-renaming macro.
   (define import.  ((with-module gauche.internal make-identifier)
@@ -85,7 +85,7 @@
                   If you intend to import Gauche's built-in bindings, \
                   say (import (gauche base)).\n"))
          (list (library-name->module-name import-set))]))
-    (let1 import-spec (rec import-set)
+    (let1 import-spec (rec (unwrap-syntax import-set))
       `((,require-if-module-doesnt-exist. ,(car import-spec))
         (,import. ,import-spec)))))
 
@@ -110,7 +110,7 @@
   (define r7rs-import.    ((with-module gauche.internal make-identifier)
                            'r7rs-import (find-module 'r7rs.import) '()))
   (define use.            (global-id 'use))
-  
+
   (define-macro (define-library name . decls)
     `(,define-module. ,(library-name->module-name name)
        (,extend.)
@@ -137,7 +137,7 @@
       [(include-ci)  `(,include-ci. ,@(cdr decl))]
       [(cond-expand)
        ;; cond-expand needs special handling.  The expansion logic is the
-       ;; same as srfi-0 cond-expand, but we have to treat the expanded
+       ;; same as SRFI-0 cond-expand, but we have to treat the expanded
        ;; form as library-declarations instead of ordinary Scheme expressions.
        ;; The current implementation relies on how cond-expand constructs
        ;; the output; if we change cond-expand, we may need to tweak this
@@ -196,9 +196,9 @@
   (use gauche.uvector)
   (use gauche.record)
   (use gauche.unicode)
-  (use srfi-13)
+  (use srfi.13)
 
-  (require "srfi-43")
+  (require "srfi/43")
 
   (import r7rs.aux)
   (export * + - ... / < <= = => > >= _ abs and append apply assoc assq
@@ -239,7 +239,7 @@
           syntax-rules truncate truncate-remainder u8-ready?  unquote
           utf8->string vector vector->string vector-copy vector-fill!
           vector-length vector-ref vector?
-          (rename r7rs:with-exception-handler with-exception-handler)
+          with-exception-handler
           write-char write-u8 string-fill!  string-length string-ref string<=?
           string=?  string>?  substring symbol=?  syntax-error textual-port?
           truncate-quotient truncate/ unless unquote-splicing values
@@ -322,24 +322,11 @@
 
   ;; 6.11 Exceptions
   ;; error - built-in
+  ;; with-exception-handler - built-in
+  ;; raise-continuable - built-in
 
-  ;; NB: In Gauche, 'with-exception-handler' is srfi-18 version.
-  (define (r7rs:with-exception-handler handler thunk)
-    (let* ([old (current-exception-handler)]
-           [new (^[exc]
-                  (with-exception-handler
-                   old
-                   (^[]
-                     (if (condition-has-type? exc <serious-condition>)
-                       (begin
-                         (handler exc)
-                         (raise exc))
-                       (handler exc)))))])
-      (with-exception-handler new thunk)))
-
-  ;; NB: In Gauche, 'raise' is continuable as far as the thrown exception
-  ;; isn't fatal.
-  (define (raise-continuable c) (raise c))
+  ;; NB: Gauche's internal 'raise' let the condition choose whether it is
+  ;; continuable or not.  R7RS raise forbid 'raise' to continue.
   (define (r7rs:raise c) ((with-module gauche.internal %raise) c #t))
 
   (define (error-object? e) (condition-has-type? e <error>))
@@ -408,9 +395,9 @@
   (define+ string-ci>?  gauche.unicode)   ; not gauche's.
   (define+ string-ci<=? gauche.unicode)   ; not gauche's.
   (define+ string-ci>=? gauche.unicode)   ; not gauche's.
-  (define+ string-upcase gauche.unicode)   ; not srfi-13's.
-  (define+ string-downcase gauche.unicode) ; not srfi-13's.
-  (define+ string-foldcase gauche.unicode) ; not srfi-13's.
+  (define+ string-upcase gauche.unicode)   ; not SRFI-13's.
+  (define+ string-downcase gauche.unicode) ; not SRFI-13's.
+  (define+ string-foldcase gauche.unicode) ; not SRFI-13's.
   (provide "scheme/char"))
 
 (define-module scheme.complex
@@ -442,8 +429,10 @@
           open-binary-input-file open-binary-output-file
           open-input-file open-output-file
           with-input-from-file with-output-to-file)
-  (define open-binary-input-file open-input-file)
-  (define open-binary-output-file open-output-file)
+  (define (open-binary-input-file filename)
+    (open-input-file filename :element-type :binary))
+  (define (open-binary-output-file filename)
+    (open-output-file filename :element-type :binary))
   (define+ delete-file file.util)
   (provide "scheme/file"))
 
@@ -456,7 +445,8 @@
   (import r7rs.aux)
   (export delay force delay-force promise? make-promise)
   (define-syntax delay-force (with-module gauche lazy))
-  (define (make-promise obj) (if (promise? obj) obj (delay obj)))
+  (define (make-promise obj)
+    (if (promise? obj) obj (delay obj)))
   (provide "scheme/lazy"))
 
 (define-module scheme.load
@@ -469,7 +459,7 @@
   (import r7rs.aux)
   (export command-line emergency-exit exit
           get-environment-variable get-environment-variables)
-  (define (emergency-exit :optional (obj 0)) (sys-exit obj))
+  (define (emergency-exit :optional (code #t)) (sys-exit code))
   (define get-environment-variable  sys-getenv)
   (define get-environment-variables sys-environ->alist)
   (provide "scheme/process-context"))
@@ -486,7 +476,7 @@
 
 (define-module scheme.time
   (export current-jiffy jiffies-per-second current-second)
-  (define-constant tai-utc 35) ; TAI is ahead of this amount as of 2014
+  (define-constant tai-utc 37) ; TAI is ahead of this amount as of 2017
   (define-constant tai-off 8)  ; TAI epoch is ahead of this amount
   ;; We reduce resolution in 32bit platform so that we have more time
   ;; before current-jiffy falls out of fixnum range.  On 32bit machines,
