@@ -1,7 +1,7 @@
 ;;;
 ;;; http.scm - HTTP 1.1
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -44,11 +44,11 @@
 ;; The features required for typical client usage are implemented first.
 
 (define-module rfc.http
-  (use srfi-11)
-  (use srfi-13)
+  (use srfi.13)
   (use rfc.822)
   (use rfc.uri)
   (use rfc.base64)
+  (use rfc.tls)
   (use gauche.net)
   (use gauche.charconv)
   (use gauche.sequence)
@@ -82,11 +82,10 @@
           mime-compose-message-string
           mime-compose-parameters
           mime-parse-content-type)
-(autoload rfc.tls make-tls tls-connect)
 (autoload rfc.http.tunnel probe-stunnel make-stunnel-connection)
 (autoload file.util file-size find-file-in-paths null-device)
 
-;;==============================================================
+;==============================================================
 ;; Conditions
 ;;
 
@@ -171,7 +170,7 @@
 ;;             appropriate error is thrown from http-request.
 ;;
 ;;   sender  - A procedure that handles sending the request body.
-;;             It takes three arguments: tantative list of headers,
+;;             It takes three arguments: tentative list of headers,
 ;;             the value given to :request-encoding argument,
 ;;             and a procedure HEADER-SINK.
 ;;
@@ -574,15 +573,9 @@
 ;; In reality it is used very frequently with http, so we put it here.
 (define (http-compose-query path params
                             :optional (encoding (gauche-character-encoding)))
-  (define (esc s) (uri-encode-string (x->string s) :encoding encoding))
-  (define (query-1 n&v)
-    (match n&v
-      [(name value) #"~(esc name)=~(esc value)"]
-      [_ (error "Invalid request-uri form:" params)]))
-  (define (query) (string-concatenate (intersperse "&" (map query-1 params))))
-  (cond [(not path) (query)]
+  (cond [(not path) (uri-compose-query params)]
         [(null? params) path]
-        [else #"~|path|?~(query)"]))
+        [else #"~|path|?~(uri-compose-query params)"]))
 
 ;; multipart/form-data composition [RFC2388]
 ;; <params> : (<param> ...)
@@ -732,24 +725,27 @@
       [#/^([^:]+)(?::(\d+))?$/ (_ host port) (values host port #f)]
       [else (error "Unrecognized http server address:" addr)]))
   (define (connect-socket)
-    (receive (host port path) (parse-address (or (~ conn'proxy) (~ conn'server)))
-      (if path
-        (make-client-socket 'unix path)
-        (make-client-socket host (if port
-                                   (x->integer port)
-                                   (if (~ conn'secure) 443 80))))))
-  (ecase (~ conn'secure)
-    [(#f) (set! (~ conn'socket) (connect-socket))]
-    [(tls)
-     (let1 tls (make-tls :server-name (~ conn'server)
-                         :options (if (tls-ca-bundle-path)
-                                    0
-                                    SSL_SERVER_VERIFY_LATER))
-       (set! (~ conn'socket) (tls-connect tls (connect-socket))))]
-    [(stunnel)
-     (let* ([rhost      (or (~ conn'proxy) (~ conn'server))]
-            [rhost:port (if (string-index rhost #\:) rhost #"~|rhost|:https")])
-       (set! (~ conn'socket) (make-stunnel-connection rhost:port)))]))
+    (receive (host port path)
+        (parse-address (or (~ conn'proxy) (~ conn'server)))
+      (cond [path (make-client-socket 'unix path)]
+            [(eq? (~ conn'secure) 'tls)
+             (cond-expand
+              [gauche.net.tls
+               (tls-connect (make-tls :server-name host
+                                      :options (if (tls-ca-bundle-path)
+                                                 0
+                                                 SSL_SERVER_VERIFY_LATER))
+                            host (or port "443") 'tcp)]
+              [else
+               (error "TLS support is not available on this Gauche (you need to recompile Gauche with TLS support on.)")])]
+            [(eq? (~ conn'secure) 'stunnel)
+             (let1 host:port (cond [port #"~|host|:~|port|"]
+                                    [(string-index host #\:) host]
+                                    [else #"~|host|:https"])
+               (make-stunnel-connection host:port))]
+            [else (make-client-socket host (if port (x->integer port) 80))])))
+
+  (set! (~ conn'socket) (connect-socket)))
 
 (define (with-connection conn proc)
   (unwind-protect
