@@ -36,12 +36,14 @@
 (define-module rfc.json
   (use gauche.sequence)
   (use gauche.generator)
-  (use parser.peg)
   (use gauche.unicode)
-  (use srfi-13)
-  (use srfi-14)
-  (use srfi-113)
+  (use scheme.charset)
+  (use parser.peg)
+  (use srfi.13)
+  (use srfi.113)
   (export <json-parse-error> <json-construct-error>
+          <json-mixin>
+
           parse-json parse-json-string
           parse-json*
           construct-json construct-json-string
@@ -70,10 +72,19 @@
 (define (build-object pairs) ((json-object-handler) pairs))
 (define (build-special symbol) ((json-special-handler) symbol))
 
-;; This is to bail out the input nesting is too deep (see srfi-180)
+;; This is to bail out the input nesting is too deep (see SRFI-180)
 (define json-nesting-depth-limit (make-parameter +inf.0))
 (define json-current-nesting-depth (make-parameter 0))
 
+;; JSON-serializable class.
+;;  In fact, we can just use object inspection to provide the default
+;;  serializer without a specialized mixin class.  However, such implicit
+;;  handling tends to be error-prone.  We'd rather be explicit.
+;;  In a class inheriting <json-mixin>, a slot can have :json-name
+;;  to be serializable.
+(define-class <json-mixin> () ())
+
+
 ;;;============================================================
 ;;; Parser
 ;;;
@@ -104,30 +115,23 @@
 (define %array
   ($depth-check
    ($lift ($ build-array $ rope-finalize $)
-          ($between %begin-array 
+          ($between %begin-array
                     ($sep-by ($seq0 %value %ws) %value-separator)
                     %end-array))))
 
+(define %sign ($optional ($. #[+-])))
+(define %digits ($many1 ($. #[\d])))
+
 (define %number
-  (let* ([%sign ($optional ($->rope ($one-of #[+-])))]
-         [%digits ($->rope ($many ($. #[\d]) 1))]
-         [%frac ($->rope ($. #\.) ($many ($. #[\d]) 1))]
-         [%exp ($->rope ($. #[eE]) %sign %digits)])
-    ($lift ($ string->number $ string-concatenate
-              $ map rope->string $ list $*)
-           %sign %digits ($optional %frac) ($optional %exp))))
+  ($binding ($: sign   %sign)
+            ($: digits %digits)
+            ($: frac   ($optional ($->rope ($. #\.) ($many1 ($. #[\d])))))
+            ($: exp    ($optional ($->rope ($. #[eE]) %sign %digits)))
+            (string->number (rope->string (list sign digits frac exp)))))
 
 (define %unicode
   (let ([%hex4 ($lift (^s (string->number (list->string s) 16))
-                      ($many ($. #[\da-fA-F]) 4 4))]
-        ;; NB: If we just $fail, the higher-level parser may conceal the
-        ;; direct cause of the error by backtracking.  Unpaired surrogate
-        ;; is an unrecoverable error, so we throw <json-parse-error> directly.
-        ;; There may be a better way to integrate this kind of error in
-        ;; the combinators; let's see.
-        [err (^c (errorf <json-parse-error>
-                         :position #f :object c
-                         "unpaired surrogate: \\u~4,'0x" c))])
+                      ($many ($. #[\da-fA-F]) 4 4))])
     ($let ([ ($. #\u) ]
            [c %hex4])
       (cond [(<= #xd800 c #xdbff)
@@ -137,16 +141,9 @@
                               (utf16->ucs4 `(,c ,c2) 'permissive)
                             (and (null? x)
                                  ($return (ucs->char cc))))))
-                  ;; NB: We wrap (err c) with dummy $let to put the call
-                  ;; to the err into a parser monad.  Simple ($return (err c))
-                  ;; or ($fail (err c)) won't do, since (err c) is evaluated
-                  ;; at the parser-construction time, not the actual parsing
-                  ;; time.  We need a dummy ($return #t) clause to ensure
-                  ;; (err c) is wrapped; ($let () x) is expanded to just x.
-                  ;; Definitely we need something better to do this kind of
-                  ;; operation.
-                  ($let ([($return #t)]) (err c)))]
-            [(<= #xdc00 c #xdfff) (err c)]
+                  ($raise (format "unpaired high surrogate: \\u~4,'0x" c)))]
+            [(<= #xdc00 c #xdfff)
+             ($raise (format "unpaired low surrogate: \\u~4,'0x" c))]
             [else ($return (ucs->char c))]))))
 
 (define %string
@@ -169,12 +166,12 @@
     ($between %dquote %string-body %dquote)))
 
 (define %object
-  (let1 %member ($let ([k %string]
-                       [ %ws ]
-                       [ %name-separator ]
-                       [v %value]
-                       [ %ws ])
-                  ($return (cons k v)))
+  (let1 %member ($binding ($: k %string)
+                          %ws
+                          %name-separator
+                          ($: v %value)
+                          %ws
+                          (cons k v))
     ($depth-check
      ($between %begin-object
                ($lift ($ build-object $ rope-finalize $)
@@ -228,31 +225,51 @@
         [(number? obj)    (print-number obj)]
         [(is-a? obj <dictionary>) (print-object obj)]
         [(is-a? obj <sequence>)   (print-array obj)]
+        [(is-a? obj <json-mixin>) (print-instance obj)]
         [else (error <json-construct-error> :object obj
                      "can't convert Scheme object to json:" obj)]))
 
 (define (print-object obj)
-  (display "{")
-  (fold (^[attr comma]
+  (write-char #\{)
+  (fold (^[attr comma-needed?]
           (unless (pair? attr)
             (error <json-construct-error> :object obj
                    "construct-json needs an assoc list or dictionary, \
                     but got:" obj))
-          (display comma)
+          (when comma-needed?
+            (write-char #\,))
           (print-string (x->string (car attr)))
-          (display ":")
+          (write-char #\:)
           (print-value (cdr attr))
-          ",")
-        "" obj)
-  (display "}"))
+          #t)
+        #f obj)
+  (write-char #\}))
 
 (define (print-array obj)
-  (display "[")
+  (write-char #\[)
   (for-each-with-index (^[i val]
-                         (unless (zero? i) (display ","))
+                         (unless (zero? i) (write-char #\,))
                          (print-value val))
                        obj)
-  (display "]"))
+  (write-char #\]))
+
+(define (print-instance obj)            ;<json-mixin>
+  (let1 class (class-of obj)
+    (write-char #\{)
+    (fold (^[slot comma-needed?]
+            (if-let1 json-name (slot-definition-option slot :json-name #f)
+              (begin
+                (when comma-needed?
+                  (write-char #\,))
+                (print-string (if (eqv? json-name #t)
+                                (x->string (slot-definition-name slot))
+                                (x->string json-name)))
+                (write-char #\:)
+                (print-value (slot-ref obj (slot-definition-name slot)))
+                #t)
+              comma-needed?))
+          #f (class-slots class))
+    (write-char #\})))
 
 (define (print-number num)
   (cond [(or (not (real? num)) (not (finite? num)))
@@ -277,9 +294,9 @@
              (if (>= code #x10000)
                (for-each hexescape (ucs4->utf16 code))
                (hexescape code)))]))
-  (display "\"")
+  (write-char #\")
   (string-for-each print-char str)
-  (display "\""))
+  (write-char #\"))
 
 (define (construct-json x :optional (oport (current-output-port)))
   (with-output-to-port oport
@@ -287,4 +304,3 @@
 
 (define (construct-json-string x)
   (call-with-output-string (cut construct-json x <>)))
-
