@@ -1,4 +1,4 @@
 ;;
 ;; This is a cushion module to use R7RS (scheme r5rs) from Gauche
 ;;
-(define-module scheme.r5rs (use r7rs))
+(define-module scheme.r5rs (use r7rs-setup))
