@@ -1,7 +1,7 @@
 ;;;
 ;;; text.sql - SQL parsing
 ;;;
-;;;   Copyright (c) 2005-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2005-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,7 +38,7 @@
 ;; modules.
 
 (define-module text.sql
-  (use srfi-13)
+  (use srfi.13)
   (use util.match)
   (export <sql-parse-error>
           sql-tokenize
@@ -214,5 +214,3 @@
   ;; main entry
   ;;
   (entry (skip-ws sql-string) '()))
-
-
