@@ -23,13 +23,14 @@
   (use gauche.uvector)
   (use gauche.charconv)
   (use gauche.generator)
-  (use srfi-13)
-  (use srfi-42)
+  (use srfi.13)
+  (use srfi.42)
   (use util.match)
   (use file.util)
   (use text.csv)
   (export ucd-parse-files
-          ucd-get-entry ucd-get-category-ranges ucd-get-break-property
+          ucd-get-entry ucd-get-category-ranges
+          ucd-get-break-property ucd-get-east-asian-width
           ucd-map-entries
 
           ucd-save-db ucd-load-db ucd-version
@@ -47,7 +48,7 @@
           ucd-extended-case-map-simple-map
           ucd-extended-case-map-special-map
           ucd-extended-case-map-code
-          
+
           ucd-break-property
           ucd-break-property-grapheme
           ucd-break-property-word
@@ -181,6 +182,11 @@
 (define (ucd-get-break-property db code)
   (dict-get (unichar-db-break-table db) code #f))
 
+;; API
+;; return East Asian Width
+(define (ucd-get-east-asian-width db code)
+  (dict-get (unichar-db-width-table db) code 'N))
+
 (define-record-type ucd-simple-case-map #t #t
   case      ; upper or lower, for this character itself (title==upper)
   offset)   ; offset to convert to the opposite case
@@ -211,20 +217,21 @@
 (define (ucd-east-asian-widths)
   '(A F H N Na W))
 
-;; Break properties.  NB: In order to squeeze grapheme and word break
-;; properties into a single byte, we treat GB_CR, GB_LF, WB_CR, WB_LF,
-;; WB_Single_Quote and WB_Double_Quote specially; they have precisely
-;; one character each, so the lookup procedure recognizes those characters
-;; directly instead of table lookup.  We don't store those properties
-;; in the table.  The '#f' in the list marks boundary of these propetries.
+;; Break properties.  The actual property values are in
+;; src/gauche/priv/unicode_attr.h, which is generated by src/gen-unicode.scm.
+;; NB: We use 'Other instead of 'Any in UAX#29.
+;; NB: 'ExtPict is not defined in Grapheme Break Properties, but in
+;; Emoji Properties.  It is required to implement GB11 and WB3c  rule
+;; (see ext/gauche/unicode.scm)
 (define (ucd-grapheme-break-properties)
-  '(Control Extend Regional_Indicator Prepend SpacingMark
-    L V T LV LVT Other #f CR LF))
+  '(CR LF Control Extend ZWJ Regional_Indicator Prepend SpacingMark
+    L V T LV LVT ExtPict Other))
 
 (define (ucd-word-break-properties)
-  '(Newline Extend Regional_Indicator Format Katakana Hebrew_Letter ALetter
-    MidLetter MidNum MidNumLet Numeric ExtendNumLet Other
-    #f CR LF Single_Quote Double_Quote))
+  '(CR LF Newline Extend ZWJ Regional_Indicator
+    Format Katakana Hebrew_Letter ALetter Single_Quote Double_Quote
+    MidLetter MidNum MidNumLet Numeric ExtendNumLet WSegSpace
+    ExtPict ExtPict_ALetter Other))
 
 ;;
 ;; CES encoding conversions.
@@ -245,15 +252,17 @@
                           (logand (~ vec 3) #x3f))])))
 
 (define (get-char/ces vec ces)
-  (utf8vec->ucs4
-   (string->u8vector
-    (ces-convert (u8vector->string vec) ces 'utf-8))))
+  (guard (e [else #f])
+    (utf8vec->ucs4
+     (string->u8vector
+      (ces-convert (u8vector->string vec) ces 'utf-8)))))
 
 (define (2b->u8vector code)
   (u8vector (ash code -8) (logand code #xff)))
 (define (3b->u8vector code)
   (u8vector (ash code -16) (logand (ash code -8) #xff) (logand code #xff)))
 
+;; can return #f
 (define (eucjp->ucs code)
   ;; NB: The codes in *jisx0213-extras* should be filtered before
   ;; calling this.
@@ -264,6 +273,7 @@
                       [else #f])])
     (get-char/ces v 'euc-jp)))
 
+;; can return #f
 (define (sjis->ucs code)
   ;; NB: The codes in *jisx0213-extras* should be filtered before
   ;; calling this.
@@ -313,17 +323,20 @@
                   (rlet1 e (make-ucd-entry cat '(#f #f #f) #f)
                     (when (eq? cat 'Ll) (set! (ucd-entry-lowercase e) #t))
                     (when (eq? cat 'Lo) (set! (ucd-entry-alphabetic e) #t)))))]
-        [else
-         (dict-get (unichar-db-table db) (->ucs code) #f)]))
+        [(->ucs code)
+         => (^c (dict-get (unichar-db-table db) c #f))]
+        [else #f]))
 
 ;; The following APIs are needed since we have to handle jisx0213 extra
 ;; characters.
 
 ;; API
+;; can return #f if CODE is 'hole' (unassigned)
 (define (eucjp->ucd-entry db code)
   (xxx->entry db code cadr eucjp->ucs))
 
 ;; API
+;; can return #f if CODE is 'hole' (unassigned)
 (define (sjis->ucd-entry db code)
   (xxx->entry db code car sjis->ucs))
 
@@ -345,6 +358,8 @@
     (word-break-property
      (ucd-word-break-properties)
      (build-path datadir "auxiliary/WordBreakProperty.txt") db)
+    (augument-break-property
+     (build-path datadir "emoji/emoji-data.txt") db)
     (east-asian-width (build-path datadir "EastAsianWidth.txt") db)
     db))
 
@@ -559,6 +574,30 @@
 (define grapheme-break-property (break-property ucd-break-property-grapheme))
 (define word-break-property     (break-property ucd-break-property-word))
 
+;; Grapheme break property needs to know emoji Extended_Pictographic
+(define (augument-break-property file db)
+  (define (ensure c)
+    (let1 e (ensure-ucd-break-property db c)
+      (case (ucd-break-property-grapheme e)
+        [(Other) (set! (ucd-break-property-grapheme e) 'ExtPict)]
+        [else (warn "Overriding non-'Other' grapheme property of U+~4,'0x (~s)\n"
+                    c (ucd-break-property-grapheme e))])
+      (case (ucd-break-property-word e)
+        [(Other)   (set! (ucd-break-property-word e) 'ExtPict)]
+        [(ALetter) (set! (ucd-break-property-word e) 'ExtPict_ALetter)]
+        [else (warn "Overriding non-'Other' word property of U+~4,'0x (~s)\n"
+                    c (ucd-break-property-word e))])))
+  (define (handle line)
+    (rxmatch-case line
+      [#/^([0-9a-fA-F]{4,6})\.\.([0-9a-fA-F]{4,6})\s+\;\s+Extended_Pictographic/
+          (_ ss ee)
+        (do-ec (: c (parse-code ss) (+ (parse-code ee) 1))
+               (ensure c))]
+      [#/^([0-9a-fA-F]{4,6})\s+\;\s+Extended_Pictographic/ (_ ss)
+        (ensure (parse-code ss))]))
+  (with-input-from-file file
+    (cut generator-for-each handle read-line)))
+
 ;;;
 ;;;  Dumping and restoring database
 ;;;
@@ -576,7 +615,7 @@
 ;;
 ;; 1. The low codepoint table part
 ;;
-;;  RLE compressed list of <ucd-entry> 
+;;  RLE compressed list of <ucd-entry>
 ;;
 ;;  <ucd-entry>  : <basic-entry>
 ;;  <basic-entry> : #f ; unassigned
@@ -630,7 +669,7 @@
   (print ";; unicode version")
   (write (unichar-db-version db))
   (print)
-  
+
   (print ";; unichar-db-table")
   (print "(")
   (let1 tab (unichar-db-table db)
