@@ -1,7 +1,7 @@
 ;;;
 ;;; isomorph.scm - check isomorphism
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,8 +32,8 @@
 ;;;
 
 (define-module util.isomorph
-  (use srfi-1)
-  (use srfi-133)
+  (use scheme.list)
+  (use scheme.vector)
   (export isomorphic? object-isomorphic?)
   )
 (select-module util.isomorph)
@@ -76,4 +76,3 @@
 
 (define-method object-isomorphic? (a b context)
   (equal? a b))                         ;default
-
