@@ -1,7 +1,7 @@
 ;;;
 ;;; util.levenshtein - Calculate edit distance between two sequences
 ;;;
-;;;   Copyright (c) 2016-2018  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2016-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,8 +32,8 @@
 ;;;
 
 (define-module util.levenshtein
-  (use srfi-1)
-  (use srfi-133)
+  (use scheme.list)
+  (use scheme.vector)
   (use gauche.array)
   (use gauche.sequence)
   (export l-distance l-distances
@@ -45,7 +45,7 @@
 ;; Restricted edit distances of two sequences
 ;; (not limited to strings).
 
-;; Note: 
+;; Note:
 ;;  Those widely known algorithms were extensively researched in 60s
 ;;  to 70s, and sometimes different researchers reported same or similar
 ;;  algorithms without attaching specific name.  We took widely known
@@ -55,7 +55,7 @@
 ;;  problem, JACM 21(1), 1974, shows DP algorithm to calculate this, but
 ;;  there's no mention of "Levenshtein" in the paper.
 ;;
-;;  Damerau-Levenshtein-distance: Lowrance & Wagner, An extention of
+;;  Damerau-Levenshtein-distance: Lowrance & Wagner, An extension of
 ;;  the string-to-string correction problem, JACM 22(2), 1975.  Again,
 ;;  there's no mention of "Damerau" nor "Levenshtein".
 ;;
@@ -69,7 +69,7 @@
 ;;  hand, this one restricts how transposition is applied and is not
 ;;  fully compatible to Damerau-Levenshtein.
 
-;; It is often explaned using (N+k)x(M+k) array for dynamic programming
+;; It is often explained using (N+k)x(M+k) array for dynamic programming
 ;; (k=1 or 2), but we only need to refer to look back at most k rows,
 ;; so we can run the algorithm with k+1 rows and rotating them.
 ;;
@@ -145,7 +145,7 @@
         (cut run <> #f)))
 
     (map f Bs)))
-      
+
 (define (l-distance A B :key (elt= eqv?) (cutoff #f))
   (car (l-base A (list B) elt= cutoff)))
 
@@ -217,7 +217,7 @@
         (cut run <> #f)))
 
     (map f Bs)))
-      
+
 (define (re-distance A B :key (elt= eqv?) (cutoff #f))
   (car (re-base A (list B) elt= cutoff)))
 
