@@ -1,7 +1,7 @@
 ;;;
 ;;; css.scm - Generate and parse CSS
 ;;;
-;;;   Copyright (c) 2014-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2014-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,8 +38,8 @@
   (use parser.peg)
   (use util.match)
   (use text.tree)
-  (use srfi-1)
-  (use srfi-13)
+  (use scheme.list)
+  (use srfi.13)
   (export construct-css simple-selector?
           parse-css parse-css-file
           parse-css-selector-string)
@@ -543,7 +543,7 @@
 (define %declaration-list
   ($lazy
    ($seq %WS* ($/ ($sep-by ($optional %declaration) ($tok 'SEMICOLON))
-                  ($lift list %at-rule %declaration-list)))))
+                  ($list %at-rule %declaration-list)))))
 
 (define %inter-rule-spaces
   ($many_ ($/ ($tok 'CDO) ($tok 'CDC) ($tok 'WHITESPACE))))
