@@ -1,7 +1,7 @@
 ;;;
 ;;; file/util.scm - filesystem utility functions
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -39,9 +39,8 @@
 #!no-fold-case
 
 (define-module file.util
-  (use srfi-13)
+  (use srfi.13)
   (use util.match)
-  (use gauche.parameter)
   (export current-directory directory-list directory-list2 directory-fold
           home-directory temporary-directory
           make-directory* create-directory* remove-directory* delete-directory*
@@ -118,12 +117,12 @@
 (define (call-with-temporary-file proc
                                   :key (directory (temporary-directory))
                                        (prefix "gtemp"))
-  (receive (oport name) (sys-mkstemp (build-path directory prefix))  
+  (receive (oport name) (sys-mkstemp (build-path directory prefix))
     (unwind-protect (proc oport name)
       (close-output-port oport)
       (sys-unlink name))))
 
-(define (call-with-temporary-directory proc 
+(define (call-with-temporary-directory proc
                                        :key (directory (temporary-directory))
                                             (prefix "gtemp"))
   (let1 dir (sys-mkdtemp (build-path directory prefix))
@@ -205,7 +204,7 @@
   (define (err-notdir p)
     (errorf "non-directory ~s is found while creating a directory ~s"
             (sys-basename p) dir))
-    
+
   (define (rec p)
     (if (file-exists? p)
       (unless (file-is-directory? p) (err-notdir p))
@@ -230,15 +229,24 @@
 (define create-directory* make-directory*)
 
 ;; rm -rf
-(define (remove-directory* dir)
+(define (remove-directory* dir :key (if-does-not-exist :error))
   (define (rec d)
     (receive (dirs files)
         (directory-list2 d :add-path? #t :children? #t :follow-link? #f)
       (for-each rec dirs)
       (for-each sys-rmdir dirs)
       (for-each sys-unlink files)))
-  (rec dir)
-  (sys-rmdir dir))
+  (case if-does-not-exist
+    [(:error #f)]
+    [else (error ":if-does-not-exist must be either :error or #f, bot got:"
+                 if-does-not-exist)])
+  (if (file-exists? dir)
+    (if (file-is-directory? dir)
+      (begin (rec dir) (sys-rmdir dir) #t)
+      (error "Not a directory:" dir))
+    (if if-does-not-exist
+      (error "No such directory:" dir)
+      #f)))
 
 (define delete-directory* remove-directory*)
 
@@ -400,6 +408,18 @@
 
 (define (build-path base-path . components)
   (define path-separator-string (string (path-separator)))
+  (define real-base-path
+    (case base-path
+      [(cwd) (current-directory)]
+      [(cld) (if-let1 clp (current-load-path)
+               (sys-dirname clp)
+               (current-directory))]
+      [(#f) "."]
+      [else (unless (string? base-path)
+              (error "base-path must be a string, #f, \
+                      or a symbol 'cwd or 'cld, \
+                      but got:" base-path))
+            base-path]))
   (define (rec base components)
     (if (null? components)
       base
@@ -407,6 +427,7 @@
           (match (car components)
             ['up   ".."]
             ['same "."]
+            ['#f "."]
             [(? absolute-path? p)
              (error "can't append absolute path after other path" p)]
             [p (sys-normalize-pathname p)])
@@ -416,7 +437,7 @@
                    [else
                     (string-append base path-separator-string component)])
           (cdr components)))))
-  (rec (sys-normalize-pathname base-path) components))
+  (rec (sys-normalize-pathname real-base-path) components))
 
 (define (expand-path path)
   (sys-normalize-pathname path :expand #t))
@@ -489,27 +510,8 @@
     (string-append (path-sans-extension path) "." ext)
     (path-sans-extension path)))
 
-(define (find-file-in-paths name
-                            :key (paths (cond [(sys-getenv "PATH")
-                                               => (cut string-split <>
-                                                       (cond-expand
-                                                        [gauche.os.windows #\;]
-                                                        [else #\:]))]
-                                              [else '()]))
-                                 (pred file-is-executable?)
-                                 (extensions '()))
-  (define names
-    (if (null? extensions)
-      `(,name)
-      (cons name
-            (map (^e (string-append name "." e)) extensions))))
-  (define (try n) (and (pred n) n))
-  (if (absolute-path? name)
-    (any try names)
-    (let loop ((paths paths))
-      (and (not (null? paths))
-           (or (any try (map (cute build-path (car paths) <>) names))
-               (loop (cdr paths)))))))
+;; find-file-in-paths is now in core.  This is for the backward compatibility.
+(define find-file-in-paths sys-find-file)
 
 ;;;=============================================================
 ;;; File attributes
