@@ -1,7 +1,7 @@
 ;;;
 ;;; collection.scm - collection generics
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -42,13 +42,13 @@
           find find-min find-max find-min&max
           filter filter-to remove remove-to partition partition-to
           size-of lazy-size-of coerce-to
-          group-collection)
+          group-collection group-collection->alist)
   )
 (select-module gauche.collection)
 
 ;; Avoid hairy dependency issues
 (autoload gauche.uvector
-          u8vector-length 
+          u8vector-length
           s8vector-length
           u16vector-length
           s16vector-length
@@ -63,6 +63,7 @@
           c64vector-length
           c128vector-length)
 
+;; utility - we can't depend on data.queue, so this is a simple
 ;; alternative.
 (define (make-queue)   (let1 anchor (list #f) (cons anchor anchor)))
 (define (enqueue! q x) (set! (cddr q) (list x)) (set! (cdr q) (cddr q)))
@@ -127,7 +128,7 @@
 
 (define-method call-with-iterator ((coll <string>) proc
                                    :key (start #f) :allow-other-keys)
-  (let* ([s  (open-input-string (if start (string-copy coll start) coll))]
+  (let* ([s  (open-input-string (opt-substring coll start))]
          [ch (read-char s)])
     (proc (cut eof-object? ch)
           (^[] (rlet1 c ch
@@ -184,30 +185,32 @@
 
 (define-method call-with-builder ((class <vector-meta>) proc
                                   :key (size #f) :allow-other-keys)
-  (if size
-    (let ([v (make-vector size)]
-          [i 0])
-      (proc (^[item] (when (< i size)
-                       (vector-set! v i item)
-                       (inc! i)))
-            (^[] v)))
-    (let1 q (make-queue)
-      (proc (cut enqueue! q <>)
-            (cut list->vector (queue->list q))))))
+  (let1 size (force size)
+    (if size
+      (let ([v (make-vector size)]
+            [i 0])
+        (proc (^[item] (when (< i size)
+                         (vector-set! v i item)
+                         (inc! i)))
+              (^[] v)))
+      (let1 q (make-queue)
+        (proc (cut enqueue! q <>)
+              (cut list->vector (queue->list q)))))))
 
 (define-syntax *vector-builder
   (syntax-rules ()
     [(_ %make %set! %list-> class proc size)
-     (if size
-       (let ([v (%make size)]
-             [i 0])
-         (proc (^[item] (when (< i size)
-                          (%set! v i item)
-                          (inc! i)))
-               (^[] v)))
-       (let1 q (make-queue)
-         (proc (cut enqueue! q <>)
-               (cut %list-> (queue->list q)))))]))     
+     (let1 size (force size)
+       (if size
+         (let ([v (%make size)]
+               [i 0])
+           (proc (^[item] (when (< i size)
+                            (%set! v i item)
+                            (inc! i)))
+                 (^[] v)))
+         (let1 q (make-queue)
+           (proc (cut enqueue! q <>)
+                 (cut %list-> (queue->list q))))))]))
 
 ;; NB: We don't have list->uvector in core yet, so builder definition is
 ;; in gauche.uvector.  It should be here, though.
@@ -218,7 +221,7 @@
 
 (define-method call-with-builder ((class <bitvector-meta>) proc
                                   :key (size #f) :allow-other-keys)
-  (*vector-builder make-bitvector bitvector-set! list->bitvector 
+  (*vector-builder make-bitvector bitvector-set! list->bitvector
                    class proc size))
 
 (define-method call-with-builder ((class <weak-vector-meta>) proc
@@ -336,7 +339,7 @@
 (define-fold-k fold3 (knil1 knil2 knil3))
 
 ;; partial applied version
-(define fold$ 
+(define fold$
   (case-lambda
     ([proc] (^[knil . lists] (apply fold proc knil lists)))
     ([proc knil] (^ lists (apply fold proc knil lists)))))
@@ -470,14 +473,18 @@
 (define-method size-of ((coll <char-set>))    (char-set-size coll))
 (define-method size-of ((coll <uvector>))     (uvector-length coll))
 (define-method size-of ((coll <bitvector>))   (bitvector-length coll))
+(define-method size-of ((coll <hash-table>))  (hash-table-num-entries coll))
+(define-method size-of ((coll <tree-map>))    (tree-map-num-entries coll))
 
-(define-method lazy-size-of ((coll <list>))        (length coll))
+(define-method lazy-size-of ((coll <list>))        (delay (length coll)))
 (define-method lazy-size-of ((coll <vector>))      (vector-length coll))
 (define-method lazy-size-of ((coll <weak-vector>)) (weak-vector-length coll))
 (define-method lazy-size-of ((coll <string>))      (string-length coll))
 (define-method lazy-size-of ((coll <char-set>))    (char-set-size coll))
 (define-method lazy-size-of ((coll <uvector>))     (uvector-length coll))
 (define-method lazy-size-of ((coll <bitvector>))   (bitvector-length coll))
+(define-method lazy-size-of ((coll <hash-table>))  (hash-table-num-entries coll))
+(define-method lazy-size-of ((coll <tree-map>))    (tree-map-num-entries coll))
 
 ;; find -------------------------------------------------
 
@@ -645,16 +652,26 @@
 ;;  gather elements with the same key value.
 ;;  cf. group-sequence in gauche.sequence
 
+;; common utility.  returns ((key elt ...) (key elt ...))
+(define (%group-by col key-proc test-proc)
+  (let1 alist (fold (^[elt buckets]
+                      (let1 key (key-proc elt)
+                        (cond [(assoc key buckets test-proc)
+                               => (^p (push! (cdr p) elt) buckets)]
+                              [else (cons (list key elt) buckets)])))
+                    '()
+                    col)
+    (dolist [p alist] (set-cdr! p (reverse! (cdr p))))
+    (reverse alist)))
+
 (define-method group-collection ((col <collection>)
                                  :key ((:key key-proc) identity)
                                       ((:test test-proc) eqv?))
-  (fold (^[b r] (cons (reverse! (cdr b)) r))
-        '()
-        (fold (^[elt buckets]
-                (let1 key (key-proc elt)
-                  (cond [(assoc key buckets test-proc)
-                         => (^p (push! (cdr p) elt) buckets)]
-                        [else (cons (list key elt) buckets)])))
-              '()
-              col)))
+  (map cdr (%group-by col key-proc test-proc)))
 
+(define-method group-collection->alist ((col <collection>)
+                                        :key ((:key key-proc) identity)
+                                             ((:value value-proc) identity)
+                                             ((:test test-proc) eqv?))
+  (map (^p (cons (car p) (map value-proc (cdr p))))
+       (%group-by col key-proc test-proc)))
