@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.dictionary - dictionary generics
 ;;;
-;;;   Copyright (c) 2007-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2007-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,9 +32,15 @@
 ;;;
 
 (define-module gauche.dictionary
+  ;; TRANSIENT - somehow we need both 'use' and 'extend' to make this work.
+  ;; Should only 'extend' be enough.  This is just a workaround until we
+  ;; resolve the issue.
   (use gauche.collection)
+  (use gauche.libdict)
+  (extend gauche.libdict)
+  (extend gauche.collection)
   (export dict-get dict-put! |setter of dict-get|
-          dict-immutable? dict-exists?
+          dict-immutable? dict-transparent? dict-exists?
           dict-delete!
           dict-seek dict-find dict-any
           dict-fold dict-fold-right
@@ -53,6 +59,10 @@
           ))
 (select-module gauche.dictionary)
 
+;;;
+;;; Generic dictionary API
+;;;
+
 ;; Generic dictionary interface.
 ;; Required methods:
 ;;
@@ -80,129 +90,17 @@
 ;;    dict-push! dict key value
 ;;    dict-update! dict key proc [default]
 ;;    dict->alist
-
-;; A convenient macro to define dictionary methods.
-
-(define-macro (define-dict-interface class . clauses)
-  (define (gen-def kind specific)
-    (let ([dict (gensym)]
-          [key (gensym)]
-          [val (gensym)]
-          [default (gensym)]
-          [proc (gensym)]
-          [seed (gensym)]
-          [succ (gensym)]
-          [fail (gensym)])
-      (case kind
-        [(:get)
-         `(define-method dict-get ((,dict ,class) ,key . ,default)
-            (if (null? ,default)
-              (,specific ,dict ,key)
-              (,specific ,dict ,key (car ,default))))]
-        [(:put!)
-         `(define-method dict-put! ((,dict ,class) ,key ,val)
-            (,specific ,dict ,key ,val))]
-        [(:exists?)
-         `(define-method dict-exists? ((,dict ,class) ,key)
-            (,specific ,dict ,key))]
-        [(:immutable?)
-         `(define-method dict-immutable ((,dict ,class)) (,specific ,dict))]
-        [(:delete!)
-         `(define-method dict-delete! ((,dict ,class) ,key)
-            (,specific ,dict ,key))]
-        [(:clear!)
-         `(define-method dict-clear! ((,dict ,class))
-            (,specific ,dict))]
-        [(:fold)
-         `(define-method dict-fold ((,dict ,class) ,proc ,seed)
-            (,specific ,dict ,proc ,seed))]
-        [(:fold-right)
-         `(define-method dict-fold-right ((,dict ,class) ,proc ,seed)
-            (,specific ,dict ,proc ,seed))]
-        [(:map)
-         `(define-method dict-map ((,dict ,class) ,proc)
-            (,specific ,dict ,proc))]
-        [(:for-each)
-         `(define-method dict-for-each ((,dict ,class) ,proc)
-            (,specific ,dict ,proc))]
-        [(:keys)
-         `(define-method dict-keys ((,dict ,class))
-            (,specific ,dict))]
-        [(:values)
-         `(define-method dict-values ((,dict ,class))
-            (,specific ,dict))]
-        [(:push!)
-         `(define-method dict-push! ((,dict ,class) ,key ,val)
-            (,specific ,dict ,key ,val))]
-        [(:pop!)
-         `(define-method dict-pop! ((,dict ,class) ,key . ,default)
-            (apply ,specific ,dict ,key ,default))]
-        [(:update!)
-         `(define-method dict-update! ((,dict ,class) ,key . ,default)
-            (apply ,specific ,dict ,key ,default))]
-        [(:->alist)
-         `(define-method dict->alist ((,dict ,class))
-            (,specific ,dict))]
-        [(:comparator)
-         `(define-method dict-comparator ((,dict ,class))
-            (,specific ,dict))]
-        [(:seek)
-         `(define-method dict-seek ((,dict ,class) ,proc ,succ ,fail)
-            (,specific ,dict ,proc ,succ ,fail))]
-        [else (error "invalid kind in define-dict-interface:" kind)])))
-  `(begin
-     ,@(map (^p (gen-def (car p) (cadr p))) (slices clauses 2))))
-
-;;-----------------------------------------------
-;; Methods for hash-table, tree-map
-;;
-
-(define-dict-interface <hash-table>
-  :get        hash-table-get
-  :put!       hash-table-put!
-  :delete!    hash-table-delete!
-  :clear!     hash-table-clear!
-  :exists?    hash-table-exists?
-  :seek       hash-table-seek
-  :fold       hash-table-fold
-  :for-each   hash-table-for-each
-  :map        hash-table-map
-  :keys       hash-table-keys
-  :values     hash-table-values
-  :pop!       hash-table-pop!
-  :push!      hash-table-push!
-  :update!    hash-table-update!
-  :->alist    hash-table->alist
-  :comparator hash-table-comparator)
-
-(define-dict-interface <tree-map>
-  :get        tree-map-get
-  :put!       tree-map-put!
-  :delete!    tree-map-delete!
-  :clear!     tree-map-clear!
-  :exists?    tree-map-exists?
-  :seek       tree-map-seek
-  :fold       tree-map-fold
-  :fold-right tree-map-fold-right
-  :for-each   tree-map-for-each
-  :map        tree-map-map
-  :keys       tree-map-keys
-  :values     tree-map-values
-  :pop!       tree-map-pop!
-  :push!      tree-map-push!
-  :update!    tree-map-update!
-  :->alist    tree-map->alist
-  :comparator tree-map-comparator)
-
-;;-----------------------------------------------
-;; Fallback methods
 ;;
+;; The macro 'define-dict-interface' is defined in libdict.scm.
 
 (define %unique (list #f))
 
 ;; Default is mutable.  Immutable dict should override.
 (define-method dict-immutable? ((dict <dictionary>)) #f)
 
+;; Default is opaque (the content isn't shown by the printer)
+(define-method dict-transparent? ((dict <dictionary>)) #f)
+
 (define-method dict-exists? ((dict <dictionary>) key)
   (not (eq? (dict-get dict key %unique) %unique)))
 
@@ -213,7 +111,7 @@
 (define-method dict-fold-right ((dict <ordered-dictionary>) proc seed)
   (fold-right (^[kv seed] (proc (car kv) (cdr kv) seed)) dict seed))
 
-(define-method dict-seek ((dict <dictionary>) pred succ fail)
+(define-method dict-seek ((dict <dictionary>) pred fail succ)
   (let/cc return
     (dict-fold dict
                (^[k v _] (if-let1 r (pred k v)
@@ -225,10 +123,10 @@
 
 (define-method dict-find ((dict <dictionary>) pred
                           :optional (fail (^[] (values #f #f))))
-  (dict-seek dict pred (^[r k v] (values k v)) fail))
+  (dict-seek dict pred fail (^[r k v] (values k v))))
 
 (define-method dict-any ((dict <dictionary>) pred)
-  (dict-seek dict pred (^[r k v] r) (^[] #f)))
+  (dict-seek dict pred (^[] #f) (^[r k v] r)))
 
 (define-method dict-map ((dict <dictionary>) proc)
   (reverse (dict-fold dict (^[k v s] (cons (proc k v) s)) '())))
@@ -285,6 +183,7 @@
 
 (define-method dict-comparator ((dict <dictionary>)) #f)
 
+
 ;;;
 ;;; Bidirectional map
 ;;;
@@ -366,6 +265,7 @@
   (dict-fold (bimap-left dict) proc seed))
 (define-method dict-comparator ((dict <bimap>))
   (dict-comparator (bimap-left dict)))
+(define-method dict-transparent? ((dict <bimap>)) #t)
 
 ;; Collection protocol.   We just redirect methods to left map.
 (define-method call-with-iterator ((coll <bimap>) proc . args)
@@ -403,12 +303,12 @@
 (define (make-stacked-map cmpr . maps)
   (if (comparator? cmpr)
     (make <stacked-map> :key-comparator cmpr :stack maps)
-    (make <stacked-map> 
+    (make <stacked-map>
       :key-comparator (dict-comparator cmpr) ; cmpt is the first map
       :stack (cons cmpr maps))))
 
 (define-method stacked-map-stack ((smap <stacked-map>) (dict <dictionary>))
-  (make <stacked-map> 
+  (make <stacked-map>
     :key-comparator (slot-ref smap 'key-comparator)
     :stack (cons dict (slot-ref smap 'stack))))
 
@@ -420,7 +320,7 @@
   (length (slot-ref smap 'stack)))
 
 ;; These two alters the specific map that contains the key.
-(define-method stacked-map-entry-update! ((smap <stacked-map>) key proc 
+(define-method stacked-map-entry-update! ((smap <stacked-map>) key proc
                                           :optional fallback)
   (let loop ([ms (slot-ref smap 'stack)])
     (cond [(null? ms) #f]
@@ -435,6 +335,8 @@
 (define-method dict-exists? ((smap <stacked-map>) key)
   (any (cut dict-exists? <> key) (slot-ref smap 'stack)))
 
+(define-method dict-transparent? ((dict <stacked-map>)) #t)
+
 (define-method dict-get ((smap <stacked-map>) key . maybe-default)
   (let1 maybe-r (any (^m (let1 r (dict-get m key %unique)
                            (and (not (eq? r %unique)) (list r))))
