@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.generator - generator framework
 ;;;
-;;;   Copyright (c) 2011-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2011-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -60,7 +60,7 @@
           gtake gtake* gdrop gtake-while gdrop-while grxmatch gslices
           glet* glet1 do-generator
 
-          ;; srfi-121 compatibility
+          ;; SRFI-121 compatibility
           generator make-iota-generator make-range-generator
           make-coroutine-generator bytevector->generator
           make-for-each-generator make-unfold-generator gcombine
@@ -113,8 +113,7 @@
     (^[] (if (< i start) (eof-object) (%begin0 (vector-ref vec i) (dec! i))))))
 
 (define (string->generator str :optional (start #f) (end #f))
-  (let1 p (open-input-string
-           ((with-module gauche.internal %maybe-substring) str start end))
+  (let1 p (open-input-string (opt-substring str start end))
     (^[] (read-char p))))
 
 (define (uvector->generator uvec :optional (start #f) (end #f))
@@ -286,7 +285,7 @@
                (if (<= v end)
                  (eof-object)
                  (begin (inc! k) v))))))))
-        
+
 
 ;;;
 ;;; Generator operations
@@ -509,7 +508,7 @@
 
 ;; NB: The signature of gtake used to be:
 ;;   gtake gen n :optional (fill? #f) (padding #f)
-;; It is changed to be in sync with srfi-121
+;; It is changed to be in sync with SRFI-121
 ;;   gtake gen n :optional fill
 ;; We keep the old signature under the name gtake*, for it's
 ;; consistent with take*.
@@ -641,7 +640,7 @@
                  [fill? (append elts (make-list (- k len) padding))]
                  [else elts])))))
 
-(define (gindex vgen igen) ;srfi-121
+(define (gindex vgen igen) ;SRFI-121
   (let ([vgen (%->gen vgen)]
         [igen (%->gen igen)])
     (define (skip n)
@@ -658,7 +657,7 @@
                    (set! prev i))))
              (eof-object))))))
 
-(define (gselect vgen bgen) ;srfi-121
+(define (gselect vgen bgen) ;SRFI-121
   (let ([vgen (%->gen vgen)]
         [bgen (%->gen bgen)])
     (rec (g)
@@ -672,7 +671,7 @@
 
 ;; NB: The following consumers are built-in (libproc.scm)
 ;; generator-fold generator-fold-right
-;; genergenerator-for-each generator-map generator-find 
+;; genergenerator-for-each generator-map generator-find
 
 (define (%generator->list gen reverse? n)
   (if (integer? n)
@@ -734,7 +733,7 @@
 (define (generator->bytevector! vec at gen)
   (check-arg u8vector? vec)
   (generator->uvector! vec at gen))
-  
+
 (define (generator->string gen :optional (n #f))
   (with-output-to-string
     (^[]
@@ -768,13 +767,13 @@
 (define (generator-count pred gen)
   (rlet1 n 0
     (do-generator [v gen] (when (pred v) (inc! n)))))
-  
+
 (define (generator-unfold gen unfold . args)
   (apply unfold eof-object? identity (^_ (gen)) (gen) args))
 
-;; srfi-121 compatibility aliases
+;; SRFI-121 compatibility aliases
 ;; NB: We're not sure if we should put them here, or split them to
-;; srfi-121 module.
+;; SRFI-121 module.
 
 (define (generator . args) (list->generator args))
 (define (make-iota-generator count . args) (apply giota count args))
@@ -788,7 +787,7 @@
 (define (make-unfold-generator p f g seed) (gunfold p f g seed))
 (define (gcombine proc seed gen . gens) (apply gmap-accum proc seed gen gens))
 
-;; srfi-158 auxiliary procedures
+;; SRFI-158 auxiliary procedures
 (define (ggroup gen k :optional padding)
   (if (undefined? padding)
     (gslices gen k)
