@@ -1,7 +1,7 @@
 ;;;
 ;;; hook.scm - hook procedures
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -40,7 +40,7 @@
 
 (define-module gauche.hook
   (use gauche.mop.validator)
-  (use srfi-1)
+  (use scheme.list)
   (export <hook> make-hook hook? hook-empty? add-hook!
           remove-hook! delete-hook! reset-hook!
           hook->list run-hook)
@@ -99,4 +99,3 @@
 ;; make hook applicable.
 (define-method object-apply ((hook <hook>) . args)
   (apply run-hook hook args))
-
