@@ -1,7 +1,7 @@
 ;;;
 ;;; parameter.scm - parameter support
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -31,18 +31,23 @@
 ;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ;;;
 
-;; The basic parameter support is now built-in.  The following procedures
-;; and a macro is defined in src/libparam.scm:
+;; This module is for the backward compatibility.  Parameter is a built-in now,
+;; so the new code no longer needs to use this module.
 ;;
-;;    <parameter>, make-parameter, parameter?, procedure-parameter
-;;    parameterize
+;; An important compatibility node: As of 0.9.13, Gauche's built-in parameterize
+;; becomes to be based on SRFI-226 rather than dynamic-wind.  The former
+;; has a nicer property that the body of parameterize can be tail context.
+;; Other than that, the two works almost the same, except when a delimited
+;; continuation is involved.  See the discussion of SRFI-226.
 ;;
-;; This module serves two purposes:
-;;   - Make existing code that uses gauche.parameter keep working
-;;   - Augument parameters with the less-frequently used feature, observers.
+;; Since legacy code uses this module, we export backward-compatible
+;; parameterize/dynwind as parameterize from this module.  That is,
+;; if the code (use gauche.parameter), parameterize works exactly as the
+;; older versions.
 ;;
-;; New code doesn't need to use gauche.parameter unless it uses the
-;; observer feature.
+;; This module also provides the 'hook' feature of parameters; they're
+;; deprecated and we'll eventually drop its support; but meantime, legacy
+;; code can keep working as far as it uses gauche.parameter.
 ;;
 ;; Note: (object-apply <parameter> ...) is only defined here.  Since
 ;; make-parameter returns a procedure rather than <parameter>, it is no
@@ -50,11 +55,11 @@
 ;; doing something tricky--such code surely uses gauche.parameter.
 
 (define-module gauche.parameter
-  (export <parameter>
-          make-parameter 
-          parameter?
-          procedure-parameter
-          parameterize
+  (export <parameter>                   ;built-in
+          parameter?                    ;built-in
+          procedure-parameter           ;built-in
+          make-parameter                ;core's make-legacy-parameter
+          parameterize                  ;core's parameterize/dynwind
           parameter-pre-observers
           parameter-post-observers
           parameter-observer-add!
@@ -64,6 +69,11 @@
 
 (autoload gauche.hook make-hook add-hook! delete-hook! run-hook)
 
+(declare (keep-private-macro parameterize))
+
+(define-syntax parameterize parameterize/dynwind)
+(define make-parameter make-legacy-parameter)
+
 ;; When an observer is first set, replace setter and restorer
 ;; to take into account of observers.
 (define (%ensure-hooks param)
@@ -90,7 +100,7 @@
       (slot-set! param 'pre-observers (make-hook 2))
       (slot-set! param 'post-observers (make-hook 2)))))
 
-;; For the backward compatibility.  
+;; For the backward compatibility.
 (define-method object-apply ((self <parameter>))
   ((with-module gauche.internal %primitive-parameter-ref) self))
 (define-method object-apply ((self <parameter>) newval)
