@@ -1,7 +1,7 @@
 ;;;
 ;;; process.scm - process interface
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -39,8 +39,8 @@
 (define-module gauche.process
   (use gauche.generator)
   (use gauche.connection)
-  (use srfi-13)
-  (use srfi-14)
+  (use scheme.charset)
+  (use srfi.13)
   (export <process> <process-abnormal-exit>
           run-process do-process do-process!
           process? process-alive? process-pid
@@ -57,7 +57,7 @@
           with-input-from-process   with-output-to-process
           call-with-process-io
           process-output->string    process-output->string-list
-          ;; shell utilities
+          ;; shell utilities (re-export from text.sh)
           shell-escape-string shell-tokenize-string
           ;; deprecated
           run-process-pipeline
@@ -71,17 +71,18 @@
           ))
 (select-module gauche.process)
 
-;; Delay-load gauche.charconv
-(autoload gauche.charconv
-          wrap-with-input-conversion wrap-with-output-conversion)
-(autoload gauche.uvector
-          write-uvector)
+;; Delay-load to avoid circular dependency
+(autoload gauche.charconv wrap-with-input-conversion wrap-with-output-conversion)
+(autoload gauche.uvector write-uvector)
+(autoload gauche.threads make-thread thread-start!)
+
+;; These two are moved to text.sh
+(autoload text.sh shell-escape-string shell-tokenize-string)
 
 ;; Avoid build dependency issue
-(autoload srfi-19 
+(autoload srfi.19
           make-time add-duration time>?)
 
-
 (define-class <process> ()
   ((pid       :init-value -1 :getter process-pid)
    (command   :init-value #f :getter process-command :init-keyword :command)
@@ -145,21 +146,24 @@
 
 ;; A typical use case---run process synchronously, returns #t for
 ;; success, #f for failure.
-(define (do-process command . args)
-  (let* ([eflag (case (get-keyword :on-abnormal-exit args #f)
-                  [(#f) #f]
-                  [(:error) #t]
-                  [else => (cut error
-                                "Value for on-abnormal-exit argument \
-                                 must be either #f or :error, but got:" <>)])]
-         [p (apply run-process command (delete-keyword :on-abnormal-exit args))])
-    (process-wait p #f eflag)
-    (zero? (process-exit-status p))))
+(define (do-process command :key (on-abnormal-exit #f) :allow-other-keys args)
+  (let* ([raise-flag (ecase on-abnormal-exit
+                       [(#f :exit-code) #f]
+                       [(:error) #t])]
+         [p (apply run-process command args)])
+    (process-wait p #f raise-flag)
+    (let1 status (process-exit-status p)
+      (cond [(zero? status) #t]
+            [(eq? on-abnormal-exit :exit-code)
+             (if (sys-wait-exited? status)
+               (sys-wait-exit-status status)
+               (%check-normal-exit p))] ;this raises <process-abnormal-exit>
+            [else #f]))))
 
 ;; Similar to do-process, but raise an error on abnormal exit.
 ;; This use case is typical, so we add a separate API.
 (define (do-process! command . args)
-  (apply do-process command :on-abnormal-exit :error 
+  (apply do-process command :on-abnormal-exit :error
          (delete-keyword :on-abnormal-exit args)))
 
 ;; Note: I/O redirection
@@ -213,7 +217,8 @@
                        (redirects '())
                        (wait   #f) (fork   #t)
                        (host   #f)    ;remote execution
-                       (sigmask #f) (directory #f) (detached #f))
+                       (sigmask #f) (directory #f) (detached #f)
+                       (environment #t))
     (let* ([redirs (%canon-redirects redirects input output error)]
            [argv (map x->string command)]
            [proc (make <process> :command (car argv))]
@@ -230,7 +235,8 @@
           (let1 pid (sys-fork-and-exec (car argv) argv
                                        :iomap iomap :directory dir
                                        :sigmask (%ensure-mask sigmask)
-                                       :detached detached)
+                                       :detached detached
+                                       :environment environment)
             (push! (ref proc 'processes) proc)
             (set!  (ref proc 'pid) pid)
             (dolist [p toclose]
@@ -246,7 +252,8 @@
           (sys-exec (car argv) argv
                     :iomap iomap :directory dir
                     :sigmask (%ensure-mask sigmask)
-                    :detached detached))))))
+                    :detached detached
+                    :enrivonment environment))))))
 
 (define (%canon-redirects redirects in out err)
   (rlet1 redirs
@@ -294,7 +301,7 @@
        (unless (or (string? a2) (symbol? a2) (eq? a2 :null) (integer? a2)
                    (and (output-port? a2) (port-file-number a2)))
          (errorf "output redirection '~a' requires a filename, a symbol, \
-                  :null, integer file descriptor or file input port, \
+                  :null, integer file descriptor or file output port, \
                   but got ~s for file descriptor ~s"
                  (car arg) a2 (cadr arg)))]
       [(<& >&) (unless (integer? a2)
@@ -427,33 +434,16 @@
   (values iomap toclose ipipes opipes tmpfiles))
 
 ;; Set up source of << and <<< redirection, used inside %setup-iomap.
-;; We don't want to depend on gauche.threads at compile time, so we split
-;; this part to be evaluated at runtime.
-;; Returns updated iomap, toclose, and tmpfiles.
-(without-precompiling
- (define (%setup-<< fd do-file dir arg iomap toclose tmpfiles)
-   (define (write-arg o)
-     (unwind-protect
-         (cond [(eq? dir '<<<) (write arg o)]
-               [(string? arg) (display arg o)]
-               [else (write-uvector arg o)])
-       (close-output-port o)))
-   (cond-expand
-    [gauche.sys.threads
-     (receive (in out) (sys-pipe)
-       (thread-start! (make-thread (cut write-arg out)))
-       (values (acons fd in iomap) (cons in toclose) tmpfiles))]
-    [else
-     ;; If we can't use threads, we fall back to using temporary
-     ;; files.  It has some drawbacks: (1) If we 'exec'-ing
-     ;; (not forking), we don't have chance to remove them, and
-     ;; (2) On Windows we won't be able to remove temp file unless
-     ;; we wait the child process to complete.  We'll do our best
-     ;; anyway.
-     (receive (out nam) (sys-mkstemp (%temp-path-prefix))
-       (write-arg out)
-       (do-file '< fd nam)
-       (values iomap toclose (cons nam tmpfiles)))])))
+(define (%setup-<< fd do-file dir arg iomap toclose tmpfiles)
+  (define (write-arg o)
+    (unwind-protect
+        (cond [(eq? dir '<<<) (write arg o)]
+              [(string? arg) (display arg o)]
+              [else (write-uvector arg o)])
+      (close-output-port o)))
+  (receive (in out) (sys-pipe)
+    (thread-start! (make-thread (cut write-arg out)))
+    (values (acons fd in iomap) (cons in toclose) tmpfiles)))
 
 (define (%ensure-mask mask)
   (cond
@@ -526,7 +516,7 @@
                                         (max-wait #f) ; ns
                                         (continue-test #f)
                                         (raise-error? #f))
-  (define limit (and max-wait 
+  (define limit (and max-wait
                      (add-duration (current-time)
                                    (make-time 'time-duration
                                               (modulo max-wait #e1e9)
@@ -551,11 +541,11 @@
    [gauche.os.windows (undefined)]
    [else (process-send-signal process SIGCONT)]))
 
-(define (process-shutdown process 
+(define (process-shutdown process
                           :key (ask #f)
                                (ask-interval #e50e6) ;ns
                                (ask-retry 1)
-                               (signals `(,SIGTERM 
+                               (signals `(,SIGTERM
                                           ,SIGTERM
                                           ,SIGKILL))
                                (signal-interval #e50e6) ;ns
@@ -584,7 +574,8 @@
 ;; would cover typical use case...
 (define (run-pipeline commands
                       :key (input #f) (output #f) (error #f)
-                      (wait #f) (sigmask #f) (directory #f) (detached #f))
+                      (wait #f) (sigmask #f) (directory #f) (detached #f)
+                      (environment #f))
   (when (null? commands)
     (error "At least one command is required to run-command-pipeline"))
   (and-let1 offending (any (^c (and (not (pair? c)) (list c))) commands)
@@ -595,7 +586,8 @@
          [cmds (map (^[cmdline in out]
                       `(,cmdline :input ,in :output ,out :error ,error
                                  :sigmask ,sigmask
-                                 :directory ,directory :detached ,detached))
+                                 :directory ,directory :detached ,detached
+                                 :environment ,environment))
                     commands
                     (cons input (map car pipe-pairs))
                     (fold-right cons (list output) (map cdr pipe-pairs)))]
@@ -644,16 +636,40 @@
   (let1 p (%apply-run-process command input :pipe err host rest)
     (values (wrap-input-process-port p rest) p)))
 
+(define-syntax %with-handling-abnormal-exit
+  (syntax-rules ()
+    [(_ process closer on-abnormal-exit body ...)
+     (let ([p process]
+           [on-abnormal-exit on-abnormal-exit])
+       (guard (e [else
+                  (closer)
+                  (process-wait p)
+                  (case on-abnormal-exit
+                    [(:error) (%check-normal-exit p) (raise e)]
+                    [(:ignore) (raise e)]
+                    [(#f) #f]
+                    [else (unless (zero? (process-exit-status p))
+                            (on-abnormal-exit p))])])
+         (receive r (begin body ...)
+           (closer)
+           (process-wait p)
+           (case on-abnormal-exit
+             [(:error) (%check-normal-exit p) (apply values r)]
+             [(:ignore) (apply values r)]
+             [(#f) (and (zero? (process-exit-status p))
+                        (apply values r))]
+             [else (unless (zero? (process-exit-status p))
+                     (on-abnormal-exit p))]))))]))
+
 (define (call-with-input-process command proc :key (input *nulldev*)
-                                 ((:error err) #f) (host #f) (on-abnormal-exit :error)
+                                 ((:error err) #f) (host #f)
+                                 (on-abnormal-exit :error)
                                  :allow-other-keys rest)
-  (let* ((p (%apply-run-process command input :pipe err host rest))
-         (i (wrap-input-process-port p rest)))
-    (unwind-protect (proc i)
-      (begin
-        (close-input-port i)
-        (process-wait p)
-        (handle-abnormal-exit on-abnormal-exit p)))))
+  (let* ([p (%apply-run-process command input :pipe err host rest)]
+         [i (wrap-input-process-port p rest)])
+    (%with-handling-abnormal-exit
+     p (^[] (close-input-port i)) on-abnormal-exit
+     (proc i))))
 
 (define (with-input-from-process command thunk . opts)
   (apply call-with-input-process command
@@ -670,13 +686,11 @@
                                   ((:error err) #f) (host #f)
                                   (on-abnormal-exit :error)
                                   :allow-other-keys rest)
-  (let* ((p (%apply-run-process command :pipe output err host rest))
-         (o (wrap-output-process-port p rest)))
-    (unwind-protect (proc o)
-      (begin
-        (close-output-port o)
-        (process-wait p)
-        (handle-abnormal-exit on-abnormal-exit p)))))
+  (let* ([p (%apply-run-process command :pipe output err host rest)]
+         [o (wrap-output-process-port p rest)])
+    (%with-handling-abnormal-exit
+     p (^[] (close-output-port o)) on-abnormal-exit
+     (proc o))))
 
 (define (with-output-to-process command thunk . opts)
   (apply call-with-output-process command
@@ -686,15 +700,12 @@
 (define (call-with-process-io command proc :key ((:error err) #f)
                               (host #f) (on-abnormal-exit :error)
                               :allow-other-keys rest)
-  (let* ((p (%apply-run-process command :pipe :pipe err host rest))
-         (i (wrap-input-process-port p rest))
-         (o (wrap-output-process-port p rest)))
-    (unwind-protect (proc i o)
-      (begin
-        (close-output-port o)
-        (close-input-port i)
-        (process-wait p)
-        (handle-abnormal-exit on-abnormal-exit p)))))
+  (let* ([p (%apply-run-process command :pipe :pipe err host rest)]
+         [i (wrap-input-process-port p rest)]
+         [o (wrap-output-process-port p rest)])
+    (%with-handling-abnormal-exit
+     p (^[] (close-output-port o) (close-input-port i)) on-abnormal-exit
+     (proc i o))))
 
 ;;---------------------------------------------------------------------
 ;; Convenient thingies that can be used like `command` in shell scripts
@@ -709,152 +720,6 @@
 (define (process-output->string-list command . opts)
   (apply call-with-input-process command port->string-list opts))
 
-;;---------------------------------------------------------------------
-;; Shell utility
-;;
-
-(define (shell-escape-string str :optional (flavor
-                                            (cond-expand
-                                             [gauche.os.windows 'windows]
-                                             [else 'posix])))
-  (ecase flavor
-   [(windows)
-    ;; This is supported in src/libsys.scm.  See the comment in it.
-    (%sys-escape-windows-command-line str)]
-   [(posix)
-    ;; We follow standard unix shell convention: if STR contains special
-    ;; chars, we quote the entire STR by single-quotes.  If STR contains
-    ;; a single quote, we replace it with '"'"'.
-    (cond [(string-null? str) "''"]
-          [(string-index str #[\s\\\"\'`*?$<>!\[\](){}])
-           (string-append "'" (regexp-replace-all #/'/ str "'\"'\"'") "'")]
-          [else str])]))
-
-;; Aux procedure to tokenize command-line string into command name and
-;; arglist.   We don't deal with shell metacharacters; just recognize
-;; quotes.  For the time being, we reject any unquoted shell metacharacters.
-;; POSIX shell syntax:
-;; http://pubs.opengroup.org/onlinepubs/009695399/utilities/xcu_chap02.html
-;; About wicked MS shell syntax, see the comment of
-;; %sys-escape-windows-command-line in src/libsys.scm.
-;; NB: We don't expect STR to be too long, so just recurse.
-(define (shell-tokenize-string str :optional (flavor
-                                              (cond-expand
-                                               [gauche.os.windows 'windows]
-                                               [else 'posix])))
-  (define (skip-ws)
-    (let loop ([c (peek-char)])
-      (when (and (char? c) (char-whitespace? c))
-        (begin (read-char) (loop (peek-char))))))
-  (define (err fmt . args)
-    (apply errorf #"Bad shell syntax - ~fmt" args))
-  (define (err-meta c)
-    (err "shell metacharacter ~s not allowed, in ~s" c str))
-  (define (err-backslash)
-    (err "stray backslash in ~s" str))
-
-  (define (tokenize-posix)
-    (with-input-from-string str
-      (^[]
-        (define (read-next)
-          (let1 c (read-char)
-            (cond [(eof-object? c) '()]
-                  [(char-whitespace? c) (skip-ws) '()]
-                  [(eqv? c #\') (read-sq)]
-                  [(eqv? c #\") (read-dq)]
-                  [(eqv? c #\\) (read-escaped)]
-                  [(#[|&\;<>()$`] c) (err-meta c)]
-                  [else (read-unquoted c)])))
-        (define (read-escaped)
-          (let1 c (read-char)
-            (if (eof-object? c)
-              (err-backslash)
-              (read-unquoted c))))
-        (define (read-unquoted c0)
-          (let1 c (peek-char)
-            (cond [(eof-object? c) (list c0)]
-                  [(char-whitespace? c) (skip-ws) (list c0)]
-                  [(eqv? c #\') (read-char) (cons c0 (read-sq))]
-                  [(eqv? c #\") (read-char) (cons c0 (read-dq))]
-                  [(eqv? c #\\) (read-char) (cons c0 (read-escaped))]
-                  [(#[|&\;<>()$`] c) (err-meta c)]
-                  [else (read-char) (cons c0 (read-unquoted c))])))
-        (define (read-sq) 
-          (let loop ([c (read-char)] [cs '()]) 
-            (cond [(eof-object? c) (err "unclosed single quote: ~s" str)]
-                  [(eqv? c #\') (reverse cs (read-next))]
-                  [else (loop (read-char) (cons c cs))])))
-        (define (read-dq)
-          (let loop ([c (read-char)] [cs '()]) 
-            (cond [(eof-object? c) (err "unclosed double quote: ~s" str)]
-                  [(eqv? c #\") (reverse cs (read-next))]
-                  [(eqv? c #\\) (let1 c (read-char) ; section 2.2.3
-                                  (cond [(eof-object? c) (err-backslash)]
-                                        [(#[$`\"\\\n] c)
-                                         (loop (read-char) (cons c cs))]
-                                        [else
-                                         (loop (read-char) (cons* c #\\ cs))]))]
-                  [(#[$`] c) (err-meta c)]
-                  [else (loop (read-char) (cons c cs))])))
-
-        ;; body
-        (skip-ws)
-        (let loop ([words '()])
-          (if (eof-object? (peek-char))
-            (reverse words)
-            (loop (cons (list->string (read-next)) words)))))))
-
-  (define (tokenize-windows-cmd)
-    (with-input-from-string str
-      (^[]
-        (define (read-next)
-          (let1 c (read-char)
-            (cond [(eof-object? c) '()]
-                  [(char-whitespace? c) (skip-ws) '()]
-                  [(eqv? c #\") (read-quoted)]
-                  [else (read-unquoted c)])))
-        (define (read-unquoted c)
-          (let loop ([c c] [cs '()])
-            (cond [(eof-object? c) (reverse cs)]
-                  [(char-whitespace? c) (skip-ws) (reverse cs)]
-                  [(eqv? c #\\) (let1 cs (read-backslash cs)
-                                  (loop (read-char) cs))]
-                  [(eqv? c #\") (reverse cs (read-quoted))]
-                  [else (loop (read-char) (cons c cs))])))
-        (define (read-quoted)
-          (let loop ([c (read-char)] [cs '()])
-            (cond [(eof-object? c) (reverse cs)] ;implicitly closed
-                  [(eqv? c #\\) (let1 cs (read-backslash cs)
-                                  (loop (read-char) cs))]
-                  [(eqv? c #\") (read-after-closing cs)]
-                  [else (loop (read-char) (cons c cs))])))
-        (define (read-backslash cs)
-          (let loop ([n 1] [cs cs])
-            (let1 c (peek-char)
-              (cond [(eqv? c #\\) (read-char) (loop (+ n 1) cs)]
-                    [(eqv? c #\")
-                     (if (even? n)
-                       (append (make-list (ash n -1) #\\) cs)
-                       (cons (read-char)
-                             (append (make-list (ash n -1) #\\) cs)))]
-                    [else (append (make-list n #\\) cs)]))))
-        (define (read-after-closing cs)
-          (let1 c (peek-char)
-            (if (eqv? c #\") ; this is treated as literal dq
-              (begin (read-char)
-                     (reverse (cons c cs) (read-next)))
-              (reverse cs (read-next)))))
-        ;; body
-        (skip-ws)
-        (let loop ([words '()])
-          (if (eof-object? (peek-char))
-            (reverse words)
-            (loop (cons (list->string (read-next)) words)))))))
-
-  (ecase flavor
-    [(windows) (tokenize-windows-cmd)]
-    [(posix) (tokenize-posix)]))
-
 ;;----------------------------------------------------------------------
 ;; Internal utilities for process ports
 ;;
@@ -908,13 +773,6 @@
                                   :buffer-size conversion-buffer-size)
       (process-input process))))
 
-(define (handle-abnormal-exit on-abnormal-exit process)
-  (case on-abnormal-exit
-    [(:error) (%check-normal-exit process)]
-    [(:ignore)]
-    [else (unless (zero? (process-exit-status process))
-            (on-abnormal-exit process))]))
-
 ;;----------------------------------------------------------------------
 ;; Process connection
 ;;
@@ -959,7 +817,7 @@
                (not (process-exit-status p)))
       ;; Usually, closing process-input should cause the process to exit.
       ;; We try 150ms after closing, then send signals.
-      (process-shutdown p 
+      (process-shutdown p
                         :ask identity ;; doesn't matter
                         :ask-interval #e150e6 ; 50ms
                         :signal-interval #e100e6))))
