@@ -1,7 +1,7 @@
 ;;;
 ;;; gauche.record - record implementation
 ;;;
-;;;   Copyright (c) 2010-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2010-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -45,7 +45,7 @@
   )
 (select-module gauche.record)
 
-;; <record> is srfi-9 and srif-99 compatible record types, integrated
+;; <record> is SRFI-9 and srif-99 compatible record types, integrated
 ;; into Gauche object system, and implemented for efficiency.
 ;;
 ;; rtd is a Gauche class with metaclass <record-meta>.  It has more
@@ -164,14 +164,7 @@
   ;; trick: we let (next-method) adjust the instance slot count, but
   ;; we uses :index option of the slot to determine the actual slot number.
   (next-method)
-  (let1 s
-      (compute-slot-accessor class slot (slot-definition-option slot :index))
-    (if (slot-definition-option slot :immutable #f)
-      `(,(^o (slot-ref-using-accessor o s))
-        ,(^[o v] (errorf "slot ~a of ~a is immutable"
-                         (slot-definition-name slot) o))
-        ,(^o (slot-bound-using-accessor? o s)))
-      s)))
+  (compute-slot-accessor class slot (slot-definition-option slot :index)))
 
 (define (%valid-fieldspecs? specs signal-error?)
   (if (not (vector? specs))
@@ -261,8 +254,10 @@
 ;;; Procedural layer
 ;;;
 
+;; NB: If record is created procedurally, we don't know what to set
+;; in :defined-module.
 (define (make-rtd name fieldspecs :optional (parent #f)
-                  :key (mixins '()) (metaclass #f))
+                  :key (mixins '()) (metaclass #f) (module #f))
   ;; We only allow to inherit either other record class, or classes that
   ;; don't add slots.  The metaclass be alwayas <record-meta>
   (when parent
@@ -285,7 +280,8 @@
                     (cond [(is-a? parent <record-meta>) (list parent)]
                           [else (list <record>)]))
     :slots ($ fieldspecs->slotspecs fieldspecs
-              $ if parent (length (class-slots parent)) 0)))
+              $ if parent (length (class-slots parent)) 0)
+    :defined-modules (if module (list module) '())))
 
 (define (rtd? obj) (is-a? obj <record-meta>))
 
@@ -400,7 +396,7 @@
            (errorf "slot ~a of record ~s is immutable" field rtd))
          (^(o v) (,@mutator* o k v))))
      ))
-  
+
 (define-rtd-methods <record-meta>
   %record-ctor
   ((with-module gauche.object %record-ref) rtd)
@@ -448,7 +444,7 @@
 ;;; Syntactic layer
 ;;;
 
-(define-syntax define-record-type 
+(define-syntax define-record-type
   (er-macro-transformer
    (^[f r c]
      (define (parse-type-spec type-spec)
@@ -459,7 +455,7 @@
           (let-keywords opts ([mixins #f]
                               [metaclass #f])
             (values name parent
-                    (cond-list [mixins @ (quasirename r 
+                    (cond-list [mixins @ (quasirename r
                                            `(:mixins (list ,@mixins)))]
                                [metaclass @ (quasirename r
                                               `(:metaclass ,metaclass))])))]
@@ -476,12 +472,13 @@
        (quasirename r
          `(define-inline ,typename
             (make-rtd ',typename ',(build-field-spec field-specs) ,parent
+                      ':module (current-module)
                       ,@opts))))
      (define (build-ctor typename ctor-spec)
        (match ctor-spec
          [#f '()]
          [#t (quasirename r
-               `((define-inline ,(sym+ 'make- typename) 
+               `((define-inline ,(sym+ 'make- typename)
                    (rtd-constructor ,typename))))]
          [((? identifier? ctor-name) field ...)
           (quasirename r
@@ -513,7 +510,7 @@
                               (rtd-accessor ,typename ',f)))])
             field-specs))
      (define (build-mutators typename field-specs)
-       (append-map 
+       (append-map
         (match-lambda
           [(f a m) (quasirename r
                      `((define-inline ,m
