@@ -1,7 +1,7 @@
 ;;;
 ;;; sequence.scm - sequence operations
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -38,6 +38,7 @@
 (define-module gauche.sequence
   (extend gauche.collection)
   (export referencer modifier subseq
+          call-with-reverse-iterator call-with-reverse-iterators
           fold-right
           fold-with-index map-with-index map-to-with-index for-each-with-index
           find-index find-with-index group-sequence group-contiguous-sequence
@@ -46,13 +47,18 @@
           sequence->kmp-stepper sequence-contains
           break-list-by-sequence! break-list-by-sequence
           common-prefix-to common-prefix
+          inverse-permuter
           permute-to permute permute!
+          unpermute-to unpermute unpermute!
           shuffle-to shuffle shuffle!)
   )
 (select-module gauche.sequence)
 
 ;; used by shuffle
-(autoload srfi-27 default-random-source random-source-make-integers)
+(autoload srfi.27 default-random-source random-source-make-integers)
+
+(autoload srfi.152 string-fold-right)
+(autoload scheme.vector vector-fold-right)
 
 (define-method referencer ((obj <list>))   list-ref)
 (define-method referencer ((obj <vector>)) vector-ref)
@@ -133,26 +139,119 @@
 (define-method (setter subseq) ((seq <bitvector>) start end (vals <bitvector>))
   (bitvector-copy! seq start vals 0 (- end start)))
 
+;; reverse iterator ------------------------------------
+
+(define-method call-with-reverse-iterator ((str <string>) proc
+                                           :key
+                                           (start (string-cursor-start str))
+                                           (end (string-cursor-end str))
+                                           :allow-other-keys)
+  (let ((start (string-index->cursor str start))
+        (cur (string-index->cursor str end)))
+    (proc (cut string-cursor<=? cur start)
+          (^[]
+            (set! cur (string-cursor-prev str cur))
+            (string-ref str cur)))))
+
+(define-syntax *vector-reverse-iter
+  (syntax-rules ()
+    [(_ %ref coll proc start end)
+     (proc (cut <= end start)
+           (^[] (%ref coll (dec! end))))]))
+
+(define-syntax define-vector-reverse-iterator
+  (er-macro-transformer
+   (^[f r c]
+     (let* ([type (cadr f)]
+            [%class (r (symbol-append '< type '>))]
+            [%length (r (symbol-append type '-length))]
+            [%ref    (r (if (c (r type) (r'bitvector))
+                          'bitvector-ref/int
+                          (symbol-append type '-ref)))])
+       (quasirename r
+         `(define-method call-with-reverse-iterator ((coll ,%class) proc
+                                                     ,':key
+                                                     (start 0)
+                                                     (end (,%length coll))
+                                                     ,':allow-other-keys)
+            (*vector-reverse-iter ,%ref coll proc start end)))))))
+
+(define-vector-reverse-iterator vector)
+(define-vector-reverse-iterator u8vector)
+(define-vector-reverse-iterator s8vector)
+(define-vector-reverse-iterator u16vector)
+(define-vector-reverse-iterator s16vector)
+(define-vector-reverse-iterator u32vector)
+(define-vector-reverse-iterator s32vector)
+(define-vector-reverse-iterator u64vector)
+(define-vector-reverse-iterator s64vector)
+(define-vector-reverse-iterator f16vector)
+(define-vector-reverse-iterator f32vector)
+(define-vector-reverse-iterator f64vector)
+(define-vector-reverse-iterator c32vector)
+(define-vector-reverse-iterator c64vector)
+(define-vector-reverse-iterator c128vector)
+(define-vector-reverse-iterator bitvector)
+(define-vector-reverse-iterator weak-vector)
+
+(define-method call-with-reverse-iterator ((coll <tree-map>) proc :allow-other-keys)
+  (let ([eof-marker (cons #f #f)]
+        [iter ((with-module gauche.internal %tree-map-iter) coll)])
+    (receive (k v) (iter eof-marker #t)
+      (proc (cut eq? k eof-marker)
+            (^[] (begin0 (cons k v)
+                   (set!-values (k v) (iter eof-marker #t))))))))
+
+(define (call-with-reverse-iterators colls proc)
+  (let loop ([colls colls]
+             [eprocs '()]
+             [nprocs '()])
+    (if (null? colls)
+      (proc (reverse! eprocs) (reverse! nprocs))
+      (call-with-reverse-iterator
+       (car colls)
+       (^[end? next]
+         (loop (cdr colls) (cons end? eprocs) (cons next nprocs)))))))
+
 ;; fold-right --------------------------------------------
 
 ;;  (proc e1 (proc e2 ... (proc eN seed)))
 
 (define-method fold-right (proc seed (seq <sequence>) . more)
   (if (null? more)
-    (with-iterator (seq end? next)
-      (let rec ()
-        (if (end?)
-          seed
-          (let1 elt (next)
-            (proc elt (rec))))))
-    (call-with-iterators
-     (cons seq more)
-     (^[ends? nexts]
-       (let rec ()
-         (if (any (cut <>) ends?)
-           seed
-           (let1 elts (map (cut <>) nexts)
-             (apply proc (append! elts (list (rec)))))))))))
+    (if (applicable? call-with-reverse-iterator (class-of seq) <bottom>)
+      (call-with-reverse-iterator
+       seq
+       (^[end? next]
+         (let loop ([acc seed])
+           (if (end?)
+             acc
+             (loop (proc (next) acc))))))
+      (with-iterator (seq end? next)
+        (let rec ()
+          (if (end?)
+            seed
+            (let1 elt (next)
+              (proc elt (rec)))))))
+    (let1 seqs (cons seq more)
+      (if (every (^ (seq)
+                   (applicable? call-with-reverse-iterator (class-of seq) <bottom>))
+                 seqs)
+        (call-with-reverse-iterators
+         seqs
+         (^[ends? nexts]
+           (let loop ([acc seed])
+             (if (any (cut <>) ends?)
+               acc
+               (loop (apply proc (append! (map (cut <>) nexts) (list acc))))))))
+        (call-with-iterators
+            seqs
+          (^[ends? nexts]
+            (let rec ()
+              (if (any (cut <>) ends?)
+                seed
+                (let1 elts (map (cut <>) nexts)
+                  (apply proc (append! elts (list (rec)))))))))))))
 
 ;; for list arguments, built-in fold-right is faster.
 (define-method fold-right (proc seed (seq <list>))
@@ -161,6 +260,16 @@
 (define-method fold-right (proc seed (seq1 <list>) (seq2 <list>))
   ((with-module gauche fold-right) proc seed seq1 seq2))
 
+;; shortcut
+(define-method fold-right (proc seed (seq <string>))
+  (string-fold-right proc seed seq))
+
+(define-method fold-right (proc seed (seq <vector>))
+  (vector-fold-right (^ (r e) (proc e r)) seed seq))
+
+(define-method fold-right (proc seed (seq1 <vector>) (seq2 <vector>))
+  (vector-fold-right (^ (r e1 e2) (proc e1 e2 r)) seed seq1 seq2))
+
 ;; mapping with index ------------------------------------
 
 (define-method fold-with-index (proc knil (seq <sequence>) . more)
@@ -345,7 +454,7 @@
             (reverse! results)
             (reverse! `((,start ,last) ,@(cdr results)))))
         (reverse! (map reverse! results))))))
-                                          
+
 (define-method delete-neighbor-dups ((seq <sequence>)
                                      :key ((:key key-proc) identity)
                                           ((:test test-proc) eqv?)
@@ -395,7 +504,7 @@
 ;; NB: We can't define generic version, since there's no generic way
 ;; for length-changing mutation.  Each capable sequence should implement
 ;; the method.  Here we only provide for <list>.
-(define-method delete-neighbor-dups-squeeze! ((seq <list>)            
+(define-method delete-neighbor-dups-squeeze! ((seq <list>)
                                               :key ((:key key-proc) identity)
                                                    ((:test test-proc) eqv?)
                                                    (start 0)
@@ -445,7 +554,7 @@
                  (if (= i -1) (values 0 #f) (loop i)))))))))
 
 ;; Search NEEDLE from SEQ.  Returns index if found, #f if not.
-;; The name is aligned to string-contains in srfi-13.
+;; The name is aligned to string-contains in SRFI-13.
 (define-method sequence-contains ((hay <sequence>) (needle <sequence>)
                                   :key ((:test test-proc) eqv?))
   (if-let1 stepper (sequence->kmp-stepper needle :test test-proc)
@@ -518,7 +627,31 @@
 
 ;; TODO: suffix
 
-;; permute -------------------------------------------------------
+;;
+;; Permuter
+;;
+
+;; 'permuter' is a sequence of exact integers, [X_0 X_1 ... X_i ... X_n],
+;; specifying the i-th element of the original sequence to be mapped to X_i.
+;; `inverse-permuter` returns a sequence of exact integers that does the
+;; inverse, that is, X_i-th index contains i.
+;; For now, we only support surjective mappings.
+
+(define-method inverse-permuter ((permuter <sequence>))
+  (rlet1 r (subseq permuter)            ; generic copy
+    (with-iterator (permuter end? next)
+      (do ([i 0 (+ i 1)])
+          [(end?)]
+        (set! (~ r (next)) i)))))
+
+(define-method inverse-permuter ((permuter <list>))
+  ;; for list, avoid random access
+  (vector->list
+   (rlet1 r (make-vector (size-of permuter))
+     (with-iterator (permuter end? next)
+       (do ([i 0 (+ i 1)])
+           [(end?)]
+         (set! (~ r (next)) i))))))
 
 (define-method permute-to ((class <class>) (src <sequence>) (ord <sequence>)
                            . maybe-fallback)
@@ -552,6 +685,16 @@
   (%permute!-arg-check seq ord)
   ((with-module gauche.internal %string-replace-body!) seq (permute seq ord)))
 
+(define-method unpermute-to ((class <class>) (src <sequence>) (ord <sequence>)
+                             . maybe-fallback)
+  (apply permute-to class src (inverse-permuter ord) maybe-fallback))
+
+(define-method unpermute ((src <sequence>) (ord <sequence>) . maybe-fallback)
+  (apply permute src (inverse-permuter ord) maybe-fallback))
+
+(define-method unpermute! ((seq <sequence>) (ord <sequence>))
+  (permute! seq (inverse-permuter ord)))
+
 (define (%permute!-arg-check seq ord)
   (unless (= (size-of seq) (size-of ord))
     (error "permute! needs the length of sequence and length of permuter the same: sequence is ~,,,,30:s, permuter is ~,,,,30:s" seq ord)))
@@ -596,4 +739,3 @@
   ;; string-set! is super-slow, so we provide the alternative.
   ((with-module gauche.internal %string-replace-body!)
    seq (apply shuffle seq args)))
-
