@@ -1,7 +1,7 @@
 ;;;
 ;;; text.line-edit - line editing
 ;;;
-;;;   Copyright (c) 2015-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2015-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,8 +33,8 @@
 
 (define-module text.line-edit
   (use gauche.generator)
-  (use gauche.parameter)
   (use gauche.threads)
+  (use gauche.sequence)
   (use gauche.unicode)
   (use data.ring-buffer)
   (use data.queue)
@@ -58,6 +58,7 @@
 (autoload text.console.framebuffer <framebuffer-console>
                                    init-framebuffer
                                    draw-framebuffer)
+(autoload text.multicolumn display-multicolumn)
 
 (define *kill-ring-size* 60)
 (define *history-size* 200)
@@ -92,6 +93,20 @@
    (wide-char-pos-setting :init-keyword :wide-char-pos-setting
                           :init-form (make <wide-char-setting>))
 
+   ;; A predicate or a char-set to determine word constituent characters
+   (completion-word-constituent? :init-keyword :completion-word-constituent?
+                                 :init-value #[-\w])
+   ;; <^ <string> <gap-buffer> <integer> <integer>
+   ;;  -> </ <string> <List <string>>>
+   ;; Returns completion candidate(s) that match the given word.
+   ;; If a single final completed value is determined, it is returned as
+   ;; a string.  Otherwise, a list of candidates (including zero or one
+   ;; cases) is returned.  The difference between a single string and
+   ;; a list of a string is that the latter needs further completion
+   ;; (e.g. complete to a directory boundary; further completion will go
+   ;; down to the directory).
+   (completion-lister :init-keyword :completion-lister :init-value #f)
+
    ;; Following slots are private.
    ;; When we enter read/edit, we record the cursor position and screen size.
    (initpos-y)
@@ -104,6 +119,13 @@
    (lastpos-y)
    (lastpos-x)
 
+   ;; A position of a highlighting parenthesis
+   (paren-pos :init-value #f)
+
+   ;; The last executed editor command.  Completion command behaves differently
+   ;; if invoked more than once immediately.
+   (last-command :init-value #f)
+
    ;; Experimental framebuffer support.  Only enabled by the env var.
    (framebuffer :init-form (and (sys-getenv "GAUCHE_READ_EDIT_WITH_FB")
                                 (make <framebuffer-console>)))
@@ -131,8 +153,13 @@
    (undo-stack :init-form (make-queue))
    (redo-queue :init-form (make-queue))
 
-   ;; Keystroke queue
+   ;; Keyboard input
+   ;; timer-callback is, if set, fired after waiting for keyboard input
+   ;; for 1 sec, or a new keyboard input, whichever comes first.  It is
+   ;; invoked with the key input or #f (timeout); it is one-shot, so
+   ;; automatially cleared.
    (keystroke-queue :init-form (make-queue))
+   (timed-callback :init-value #f)
 
    ;; History
    ;; Global history is kept in the ring buffer, index 0 being the most recent.
@@ -149,7 +176,7 @@
 
 (define (%load-history ctx path)
   (with-input-from-file path
-    (lambda ()
+    (^[]
       (read)                            ; ignore version for now
       (do-generator [line read-line]    ; for now, one sexpr per line
         (guard (e [else #f])
@@ -167,14 +194,14 @@
 
 (define (%save-history ctx path)
   (call-with-temporary-file
-    (lambda (port name)
+    (^[port name]
       (write '(gauche-history-version 1) port)
       (display #\newline port)
-      (for-each
-       (lambda (i)
-         (write (ring-buffer-ref (~ ctx'history) i) port)
-         (display "\n" port))
-       (reverse (iota (history-size ctx))))
+      (for-each (^i
+                 (write (ring-buffer-ref (~ ctx'history) i) port)
+                 (display "\n" port))
+                (reverse (iota (history-size ctx))))
+      (close-output-port port) ;; necessary on MinGW
       (sys-rename name path))
     :directory (sys-dirname path)))
 
@@ -186,7 +213,7 @@
     (%save-history ctx path)))
 
 ;; Global SIGCONT handler
-;;  Since we, as a library, don't have control of sigmasks in all threads, 
+;;  Since we, as a library, don't have control of sigmasks in all threads,
 ;;  so we don't know which thread the sigcont handler is invoked.
 
 ;;  *sigcont-observer* is an atom of these values:
@@ -207,7 +234,7 @@
 
 ;; Windows doesn't define SIGCONT, and we precompile this module so we can't
 ;; use cond-expand.
-(define *sigcont* (global-variable-ref (find-module 'gauche) 'SIGCONT 0))
+(define *sigcont* (module-binding-ref 'gauche 'SIGCONT 0))
 
 (define (%sigcont-handler sig)
   (when (eqv? sig *sigcont*)
@@ -250,7 +277,7 @@
       (eof-object)))
   ;; Command handler.
   ;; The command may return one of the following values.
-  ;;   visible - the command changed something visible so we need to 
+  ;;   visible - the command changed something visible so we need to
   ;;            redisplay, but we don't need to save the change in
   ;;            the actual buffer.
   ;;            The selection is cleared.
@@ -289,19 +316,19 @@
      [(applicable? h <top> <top> <top>)
       (match (h ctx buffer ch)
         [(? eof-object?) (eofread)]
-        ['nop       (loop redisp)]
+        ['nop       (loop redisp h)]
         ['visible   (reset-last-yank! ctx)
                     (clear-mark! ctx buffer)
                     (break-undo-sequence! ctx)
-                    (loop #t)]
+                    (loop #t h)]
         ['unchanged (reset-last-yank! ctx)
                     (break-undo-sequence! ctx)
-                    (loop redisp)]
+                    (loop redisp h)]
         ['redraw    (reset-cursor-pos ctx)
-                    (loop #t)]
+                    (loop #t h)]
         ['moved     (reset-last-yank! ctx)
                     (break-undo-sequence! ctx)
-                    (loop #t)]
+                    (loop #t h)]
         ['commit
          ;; We move the cursor to the last of input and redisplay,
          ;; so that the output of the client program won't overwrite
@@ -312,18 +339,18 @@
          (commit-history ctx buffer)]
         ['undone (reset-last-yank! ctx)
                  (clear-mark! ctx buffer)
-                 (loop #t)] ; don't break undo sequence
+                 (loop #t h)] ; don't break undo sequence
         [('yanked edit-command)
          (break-undo-sequence! ctx)
          (clear-mark! ctx buffer)
          (push-undo! ctx edit-command)
-         (loop #t)]
+         (loop #t h)]
         [(? list? edit-command)
          (reset-last-yank! ctx)
          (break-undo-sequence! ctx)
          (clear-mark! ctx buffer)
          (push-undo! ctx edit-command)
-         (loop #t)]
+         (loop #t h)]
         [x (error "[internal] invalid return value from a command:" x)])]
      [(is-a? h <keymap>)
       (let* ([ch (next-keystroke ctx)]
@@ -336,7 +363,7 @@
   ;; Pass true to redisp to cause buffer to be redisplayed before
   ;; accepting a new edit command.
   ;; Returns a string or #<eof>
-  (define (edit-loop redisp)
+  (define (edit-loop redisp last-command)
     ;; NB: If next char is ready, don't bother to redisplay and
     ;; just carry over redisp flag.
     (let* ([redisp (if (and redisp (not (chready? (~ ctx'console))))
@@ -344,6 +371,7 @@
                      redisp)]
            [ch (next-keystroke ctx)]
            [h (keymap-ref (~ ctx'keymap) ch)])
+      (set! (~ ctx'last-command) last-command)
       (handle-command h ch edit-loop redisp)))
 
   ;; The 'main' routine of the editor.  Called while the console is set up
@@ -357,7 +385,7 @@
         (when redisp (reset-terminal (~ ctx'console)))
         (show-prompt ctx)
         (init-screen-params ctx)
-        (edit-loop redisp))
+        (edit-loop redisp #f))
       %sigcont-unobserve))
 
   ;;
@@ -384,11 +412,15 @@
     (set! (~ ctx'lastpos-y) y)
     (set! (~ ctx'initpos-x) x)
     (set! (~ ctx'lastpos-x) x))
-  (receive (h w) (query-screen-size (~ ctx'console))
+  (reset-screen-size! ctx))
+
+(define (reset-screen-size! ctx)
+  (receive (h w) (query-screen-size (~ ctx'console) #t)
     (set! (~ ctx'screen-height) h)
     (set! (~ ctx'screen-width) w)
-    (if (~ ctx'framebuffer)
-      (init-framebuffer (~ ctx'framebuffer) w h (~ ctx'wide-char-pos-setting)))))
+    (when (~ ctx'framebuffer)
+      (init-framebuffer (~ ctx'framebuffer) w h
+                        (~ ctx'wide-char-pos-setting)))))
 
 ;; If the cursor position has been moved from the supposed position,
 ;; redisplay the prompt and reset initial position.  This is called
@@ -412,15 +444,25 @@
 ;; those keys, then raises 'sigcont-received, which causes read-line/edit
 ;; to reset the terminal mode and redisplay the buffer.
 (define (next-keystroke ctx)
+  (define (call-timed-callback arg)
+    (and-let1 cb (~ ctx'timed-callback)
+      (set! (~ ctx'timed-callback) #f)
+      (cb arg)))
   (if (queue-empty? (~ ctx'keystroke-queue))
-    (let1 ch (getch (~ ctx'console))
-      (if (and (%sigcont-received?)
-               (canonical-mode? (~ ctx'console)))
-        (let loop ()
-          (let1 ch (getch (~ ctx'console) 10000) ;10ms wait
-            (unless ch (raise 'sigcont-received))
-            (loop)))
-        ch))
+    (let1 ch (getch (~ ctx'console)
+                    (and (~ ctx'timed-callback) 1.0e6)) ;1s timeout delay
+      (cond [(not ch)                   ;timeout
+             (call-timed-callback #f)
+             (next-keystroke ctx)]
+            [(and (%sigcont-received?)
+                  (canonical-mode? (~ ctx'console)))
+             (let loop ()
+               (let1 ch (getch (~ ctx'console) 10000) ;10ms wait
+                 (unless ch (raise 'sigcont-received))
+                 (loop)))]
+            [else
+             (call-timed-callback ch)
+             ch]))
     (queue-pop! (~ ctx'keystroke-queue))))
 
 ;; Show prompt.  Returns the current column.
@@ -453,7 +495,7 @@
 (define (%redisplay ctx buffer)
 
   ;; check a initial position
-  (if (< (~ ctx'initpos-y) 0)
+  (when (< (~ ctx'initpos-y) 0)
     (set! (~ ctx'initpos-y) 0))
 
   (let* ([con     (~ ctx'console)]
@@ -462,7 +504,6 @@
          [w       (~ ctx'screen-width)]
          [h       (~ ctx'screen-height)]
          [sel     (selected-range ctx buffer)]
-         [oparen  (buffer-find-matching-paren-on-cursor buffer)]
          [oldattr '(#f #f)]
          [newattr '(#f #f)]
          [disp-x  x]
@@ -507,7 +548,7 @@
       (glet1 ch (g)
 
         ;; set character attributes
-        (set! newattr (current-char-attr n sel oparen))
+        (set! newattr (current-char-attr n sel (~ ctx'paren-pos)))
         (switch-char-attr-when-needed con oldattr newattr)
         (set! oldattr newattr)
 
@@ -566,23 +607,58 @@
       (reset-character-attribute con)
       (set-character-attribute con newattr))))
 
-;; Show message at the starting point of the current 
-;; redraw the current buffer below it.
+;; Show message at the starting point of the current input.
+;; MSG may be a string, or list of strings (lines).
+;; Redraw the current buffer below it.
+;; (Existing input is overwritten.)
 (define (show-message ctx buffer msg :optional (attr #f))
   (move-cursor-to (~ ctx'console) (~ ctx'initpos-y) 0)
+  (%draw-lines-with-attr ctx msg attr #t))
+
+;; Show message below the line where the cursor is at.
+;; Redraw the entire input after that.
+(define (show-message-below ctx buffer msg :optional (attr #f))
+  (ensure-bottom-room (~ ctx'console))
+  (cursor-down/scroll-up (~ ctx'console))
+  (putch (~ ctx'console) #\return)
+  (%draw-lines-with-attr ctx msg attr #t))
+
+(define (%draw-lines-with-attr ctx msg attr newline-after?)
+  (define lines
+    (cond [(string? msg) (string-split msg #\newline)]
+          [(list? msg) msg]
+          [else (error "[internal] string or list of strings required but got:"
+                       msg)]))
   (when attr
     (set-character-attribute (~ ctx'console) attr))
   (unwind-protect
-      (dolist [line (string-split msg #\newline)]
-        (clear-to-eos (~ ctx'console))
-        (putstr (~ ctx'console) line)
+      (let loop ([lines lines])
+        (unless (null? lines)
+          (clear-to-eos (~ ctx'console))
+          (putstr (~ ctx'console) (car lines))
+          (when (or (pair? (cdr lines)) newline-after?)
+            (cursor-down/scroll-up (~ ctx'console))
+            (putch (~ ctx'console) #\return)
+            (ensure-bottom-room (~ ctx'console))
+            (when (< (~ ctx'initpos-y) (- (~ ctx'screen-height) 1))
+              (inc! (~ ctx'initpos-y))))
+          (loop (cdr lines))))
+    (when attr
+      (reset-character-attribute (~ ctx'console)))))
+
+(define (ask-y-or-n ctx buffer msg)
+  (ensure-bottom-room (~ ctx'console))
+  (cursor-down/scroll-up (~ ctx'console))
+  (putch (~ ctx'console) #\return)
+  (%draw-lines-with-attr ctx #"~msg (y/n): " #f #f)
+  (or (eqv? #\y (getch (~ ctx'console)))
+      (begin
         (cursor-down/scroll-up (~ ctx'console))
         (putch (~ ctx'console) #\return)
         (ensure-bottom-room (~ ctx'console))
         (when (< (~ ctx'initpos-y) (- (~ ctx'screen-height) 1))
-          (inc! (~ ctx'initpos-y))))
-    (when attr
-      (reset-character-attribute (~ ctx'console)))))
+          (inc! (~ ctx'initpos-y)))
+        #f)))
 
 ;;
 ;; Key combinations
@@ -804,6 +880,32 @@
              '(#\) #\] #\}))
        (buffer-find-matching-paren buf 0 (gap-buffer-pos buf))))
 
+;; Pick word at the current position of BUF.
+;; Returns three values: The word, start position, and end position.
+;; The first word can be #f if the word isn't found.
+;; If the position doesn't have a word-constituent character, but
+;; the immediate left position has, then we pick the word immediately
+;; left on the position.
+;; This is mainly to extract a word to be used for completion.  We don't
+;; do precise word segmentation (considering grapheme clusters); we just
+;; define 'word' as a consecutive characters in word-chars charset.
+(define (buffer-find-word buf word-chars)
+  (define (scan-backward pos)
+    (cond [(= pos 0) 0]
+          [(word-chars (gap-buffer-ref buf (- pos 1))) (scan-backward (- pos 1))]
+          [else pos]))
+  (define (scan-forward pos)
+    (cond [(gap-buffer-pos-at-end? buf pos) pos]
+          [(word-chars (gap-buffer-ref buf pos)) (scan-forward (+ pos 1))]
+          [else pos]))
+
+  (let* ([pos (gap-buffer-pos buf)]
+         [start (scan-backward pos)]
+         [end (scan-forward pos)])
+    (values (if (= start end) #f (gap-buffer->string buf start end))
+            start
+            end)))
+
 ;;;
 ;;; Commands
 ;;;
@@ -833,7 +935,7 @@
     [(_ (name . args) docstring . body)
      (define name (make <edit-command>
                     :name 'name
-                    :handler (lambda args . body)
+                    :handler (^ args . body)
                     :docstring docstring))]))
 
 
@@ -849,6 +951,19 @@
   (let1 ch (next-keystroke ctx) ; TODO: octal digits input
     (gap-buffer-edit! buf `(i #f ,(x->string ch)))))
 
+(define-edit-command (insert-closing-paren ctx buf key)
+  "Insert closing paren, and highlight corresponding opening paren."
+  (reset-last-yank! ctx)
+  (break-undo-sequence! ctx)
+  (clear-mark! ctx buf)
+  (begin0
+      (gap-buffer-edit! buf `(i #f ")"))
+    (let1 open-pos (buffer-find-matching-paren buf 0 (- (gap-buffer-pos buf) 1))
+      (set! (~ ctx'paren-pos) open-pos)
+      (set! (~ ctx'timed-callback)
+            (^_ (set! (~ ctx'paren-pos) #f)
+                (redisplay ctx buf))))))
+
 (define-edit-command (insert-parentheses ctx buf key)
   "Insert a pair of parentheses at the cursor, and position the cursor \
    between them."
@@ -1011,6 +1126,7 @@
 (define-edit-command (refresh-display ctx buf key)
   "Redraw contents of the buffer."
   (reset-terminal (~ ctx'console))
+  (reset-screen-size! ctx)
   (move-cursor-to (~ ctx'console) 0 0) ;redundant, but mintty has problem without this
   (show-prompt ctx)
   (init-screen-params ctx)
@@ -1061,8 +1177,8 @@
     (if (= lines (buffer-num-lines buf))
       (next-history ctx buf key)
       (begin (buffer-set-line&col! buf (+ lines 1) col) 'moved))))
-  
-       
+
+
 (define-edit-command (transpose-chars ctx buf key)
   "Exchange characters on the cursor and before the cursor and position \
    the cursor one character before.  If the buffer is empty, do nothing. \
@@ -1085,7 +1201,7 @@
   (if (> (ring-buffer-num-entries (~ ctx'kill-ring)) 0)
     (begin
       (set! (~ ctx'last-yank) 0)
-      (set! (~ ctx'last-yank-pos) (gap-buffer-gap-start buf))
+      (set! (~ ctx'last-yank-pos) (gap-buffer-pos buf))
       (let1 yanked-text (get-yank-line ctx)
         (set! (~ ctx'last-yank-size) (string-length yanked-text))
         `(yanked ,(gap-buffer-edit! buf `(i #f ,yanked-text)))))
@@ -1141,7 +1257,7 @@
   (beep (~ ctx'console))
   (show-message ctx buf
                 #"Unknown keystroke ~(keys->string '() key). \
-                  Type M-h b for list of key bindings."
+                  Type M-h b for the list of key bindings."
                 '(#f #f reverse))
   'redraw)
 
@@ -1157,6 +1273,56 @@
   "Do nothing."
   'nop)
 
+(define-edit-command (completion-command ctx buf key)
+  "Try to complete the (partial) word at the cursor."
+  (receive (word start-pos end-pos)
+      (buffer-find-word buf (~ ctx'completion-word-constituent?))
+    (or (and-let* ([ word ]
+                   [lister (~ ctx'completion-lister)]
+                   [words (lister word buf start-pos end-pos)])
+          (match words
+            [() #f]
+            [(? string?)
+             ;; Final candidate.  We append whitespace after it.
+             (gap-buffer-move! buf start-pos)
+             (gap-buffer-replace! buf (- end-pos start-pos) #"~|words| ")
+             'redraw]
+            [(w)
+             ;; Single candidate, but may be completed further.
+             (gap-buffer-move! buf start-pos)
+             (gap-buffer-replace! buf (- end-pos start-pos) w)
+             'redraw]
+            [(w ws ...)
+             (when (eq? (~ ctx'last-command) completion-command)
+               ;; We split string for..
+               ;; - Output of display-multicolumn contains newline at the
+               ;;   end.  We don't want to display it.
+               ;; - We need to know # of rows to switch how to display it
+               (let1 candidates (string-split
+                                 (with-output-to-string
+                                   (cut display-multicolumn (cons w ws)
+                                        :width (~ ctx'screen-width)))
+                                 #\newline 'suffix)
+                 (if (length<? candidates 20)
+                   (show-message-below ctx buf candidates)
+                   (if ($ ask-y-or-n ctx buf
+                          $ format "Display all ~d possibilities?"
+                          $ + 1 (length ws))
+                     (display/pager (string-join candidates "\n"))))))
+             ;; Replace input with the longest common prefix.  This allows
+             ;; to complete 'defi' to 'define', while continuing the input.
+             ;; Note that if segmented match is used, common prefix can
+             ;; be shorter than the current input; e.g. 'c-w' lists candidates
+             ;; such as 'call-with-values' and 'char-word-consituent?', and
+             ;; the common prefix is 'c'.  In such case we leave the current
+             ;; input.
+             (let1 pre (fold common-prefix w ws)
+               (when (> (string-length pre) (- end-pos start-pos))
+                 (gap-buffer-move! buf start-pos)
+                 (gap-buffer-replace! buf (- end-pos start-pos) pre))
+               'redraw)]))
+        'nop)))
+
 (define-edit-command (help-binding-command ctx buf key)
   "Show key bindings."
   (display/pager
@@ -1239,19 +1405,35 @@
    (table   :init-keyword :table)         ;hashtable
    (default :init-keyword :default)))
 
+;; Stacked keymap is like <keymap>, except that default slot is
+;; always a <keymap>, and if a keystroke is not in the table,
+;; lookup is delegated to that keymap.
+;; (It is distinguished from a <keymap> whose default value is
+;; a secondary keymap - in that case, a keystroke falls on the default
+;; replaces the current keymap to it.)
+(define-class <stacked-keymap> (<keymap>) ())
+
 (define (make-keymap :optional (name #f) (default undefined-command))
   (make <keymap> :name name :table (make-hash-table 'equal?) :default default))
 
+(define (make-stacked-keymap base-keymap :optional (name #f))
+  (make <stacked-keymap> :name name :table (make-hash-table 'equal?)
+        :default base-keymap))
+
 (define (copy-keymap km :optional (name #f))
-  (make <keymap>
+  (make (class-of km)
     :name name
     :table (hash-table-copy (~ km'table))
     :default (~ km'default)))
 
-(define (keymap-ref km keystroke)
+(define-method keymap-ref ((km <keymap>) keystroke)
   (or (hash-table-get (~ km'table) keystroke #f)
       (~ km'default)))
 
+(define-method keymap-ref ((km <stacked-keymap>) keystroke)
+  (or (hash-table-ref (~ km'table) keystroke #f)
+      (keymap-ref (~ km'default) keystroke)))
+
 (define (command-name->keystrokes km command-name)
   (assume-type km <keymap>)
   (assume-type command-name <symbol>)
@@ -1274,8 +1456,8 @@
 (define (command-name->keystroke-string km command-name)
   (and-let1 ks (command-name->keystrokes km command-name)
     (let1 s (keys->string ks #f)
-      (substring s 0 (- (string-length s) 1))))) ;avoid depending srfi-13
-  
+      (substring s 0 (- (string-length s) 1))))) ;avoid depending SRFI-13
+
 (define-syntax define-key
   (syntax-rules ()
     [(_ km keystroke command)
@@ -1358,7 +1540,7 @@
   (define (rec km prefixes)
     (print (or (~ km'name) "(anonymous keymap)"))
     (keymap-describe km prefixes)
-    (hash-table-for-each 
+    (hash-table-for-each
      (~ km'table)
      (^[k v] (when (is-a? v <keymap>)
                (print)
@@ -1390,7 +1572,7 @@
 (define-key *default-keymap* (ctrl #\f) forward-char)
 (define-key *default-keymap* (ctrl #\g) keyboard-quit)
 (define-key *default-keymap* (ctrl #\h) delete-backward-char)
-(define-key *default-keymap* (ctrl #\i) self-insert-command) ; tab
+(define-key *default-keymap* (ctrl #\i) completion-command)  ; tab
 (define-key *default-keymap* (ctrl #\j) self-insert-command) ; newline
 (define-key *default-keymap* (ctrl #\k) kill-line)
 (define-key *default-keymap* (ctrl #\l) refresh-display)
@@ -1416,7 +1598,10 @@
 
 (do ([c #x20 (+ c 1)])
     [(= c #x7f)]
-  (define-key *default-keymap* (integer->char c) self-insert-command))
+  (unless (= c #x29) ;close paren
+    (define-key *default-keymap* (integer->char c) self-insert-command)))
+
+(define-key *default-keymap* #\)   insert-closing-paren)
 
 (define-key *default-keymap* (alt #\null) nop-command) ; for windows (ime on/off)
 ;;(define-key *default-keymap* (alt #\space) undefined-command) ; should be set-mark
@@ -1452,7 +1637,7 @@
 (define-key *default-keymap* (alt #\>) move-end-of-buffer)
 ;;(define-key *default-keymap* (alt #\?) undefined-command)
 
-              
+
 ;;(define-key *default-keymap* (alt #\@) undefined-command)
 ;;(define-key *default-keymap* (alt #\A) undefined-command)
 ;;(define-key *default-keymap* (alt #\B) undefined-command)
@@ -1589,7 +1774,7 @@ Undo semantics
  Here's the description of algorithm.
 
  - A, B, C, ... is an edit commans
- - Applying command A to buffer v is v*A.  
+ - Applying command A to buffer v is v*A.
  - ^X is an inversion of command X.  X*^X = ^X*X = I (identity)
 
  1. Suppose the user did three buffer-changing operations, A, B and C.
@@ -1634,8 +1819,6 @@ Undo semantics
     empty.
 
          buffer: v*A*B*C
-     undo stack: 
+     undo stack:
      redo queue: ^A ^B ^C C B ^B ^C C B A
 |#
-
-
