@@ -1,7 +1,7 @@
 /*
  * boolean.c
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,6 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/vmP.h"    /* for ScmVMUndefinedBool() */
 
 int Scm_EqP(ScmObj x, ScmObj y)
@@ -53,7 +54,16 @@ int Scm_EqvP(ScmObj x, ScmObj y)
                another is not. */
             if (SCM_FLONUMP(x)) {
                 if (SCM_FLONUMP(y)) {
-                    return (SCM_FLONUM_VALUE(x) == SCM_FLONUM_VALUE(y));
+                    double rx = SCM_FLONUM_VALUE(x);
+                    double ry = SCM_FLONUM_VALUE(y);
+                    /* Comparison of signbit is for (eq? 0.0 -0.0) => #f.
+                       R7RS 6.1: eqv? returns #f if ... obj1 and obj2 are
+                       both inexact numbers such that ... they do not yield
+                       the same results when passed as arguments to any
+                       other procedure that can be defined as a finite
+                       composition of Scheme's standard arithmetic
+                       procedures, ... */
+                    return (rx == ry && signbit(rx) == signbit(ry));
                 } else {
                     return FALSE;
                 }
@@ -191,7 +201,7 @@ int Scm_EqualP(ScmObj x, ScmObj y)
     if (cx == cy && cx->compare) return (cx->compare(x, y, TRUE) == 0);
     else                         return FALSE;
 
- fallback: 
+ fallback:
     {
         /* Fall back to Scheme version. */
         static ScmObj equal_interleave_proc = SCM_UNDEFINED;
