@@ -1,7 +1,7 @@
 /*
  * collection.c - common stuff for collection & other aggregate types.
  *
- *   Copyright (c) 2007-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2007-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 
 /*
  * Class stuff
