@@ -1,7 +1,7 @@
 /*
  * compare.c - comparison & sort
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -34,7 +34,7 @@
 #include <stdlib.h>
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 
 /*
  * Comparator
@@ -134,17 +134,17 @@ int Scm_Compare(ScmObj x, ScmObj y)
         if (SCM_TRUEP(y)) return  0;
         goto distinct_types;
     }
-    
-    if (Scm_SubtypeP(cx, cy)) {
+
+    if (Scm_SubclassP(cx, cy)) {
         if (cy->compare) return cy->compare(x, y, FALSE);
-    } else if (Scm_SubtypeP(cy, cx)) {
+    } else if (Scm_SubclassP(cy, cx)) {
         if (cx->compare) return cx->compare(x, y, FALSE);
     }
     if (cx == cy) {
         /* x and y are of the same type, and they can't be ordered. */
         Scm_Error("can't compare %S and %S", x, y);
     }
-    
+
  distinct_types:
     /* x and y are of distinct types.  Follow the srfi-114 rule:
        () < pairs < booleans < chars < strings < symbols < numbers
