@@ -1,7 +1,7 @@
 /*
  * compaux.c - C API bridge for the compiler
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -37,10 +37,11 @@
 #include <stdlib.h>
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/vm.h"
 #include "gauche/vminsn.h"
-#include "gauche/class.h"
 #include "gauche/code.h"
+#include "gauche/priv/glocP.h"
 #include "gauche/priv/identifierP.h"
 #include "gauche/priv/builtin-syms.h"
 #include "gauche/priv/readerP.h" /* for Scm_MakeReadReference */
@@ -89,70 +90,25 @@ void Scm_CompileFinish(ScmCompiledCode *cc)
     }
 }
 
-/*-------------------------------------------------------------
- * Syntactic closure object
- */
-
-static void synclo_print(ScmObj obj, ScmPort *port,
-                         ScmWriteContext *ctx SCM_UNUSED)
-{
-    Scm_Printf(port, "#<syntactic-closure %S>",
-               SCM_SYNTACTIC_CLOSURE(obj)->expr);
-}
-
-SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SyntacticClosureClass, synclo_print);
-
-ScmObj Scm_MakeSyntacticClosure(ScmObj env, ScmObj literals, ScmObj expr)
-{
-    ScmSyntacticClosure *s = SCM_NEW(ScmSyntacticClosure);
-    SCM_SET_CLASS(s, SCM_CLASS_SYNTACTIC_CLOSURE);
-    s->env = env;
-    s->literals = literals;
-    s->expr = expr;
-    return SCM_OBJ(s);
-}
-
-static ScmObj synclo_env_get(ScmObj obj)
-{
-    return SCM_SYNTACTIC_CLOSURE(obj)->env;
-}
-
-static ScmObj synclo_literals_get(ScmObj obj)
-{
-    return SCM_SYNTACTIC_CLOSURE(obj)->literals;
-}
-
-static ScmObj synclo_expr_get(ScmObj obj)
-{
-    return SCM_SYNTACTIC_CLOSURE(obj)->expr;
-}
-
-static ScmClassStaticSlotSpec synclo_slots[] = {
-    SCM_CLASS_SLOT_SPEC("env", synclo_env_get, NULL),
-    SCM_CLASS_SLOT_SPEC("literals", synclo_literals_get, NULL),
-    SCM_CLASS_SLOT_SPEC("expr", synclo_expr_get, NULL),
-    SCM_CLASS_SLOT_SPEC_END()
-};
-
 /*-------------------------------------------------------------
  * Identifier object
  */
 
-/* 
+/*
  * About identifier's ENV slot.
- * We close identifier's binding environment (list of frames), so that later 
+ * We close identifier's binding environment (list of frames), so that later
  * we can look up its bindings hygienically.  We truncate the frames up to
  * where the binding occur, for the efficient lookup and comparison.
- * Notably, the identifiers that are unbound or refer to toplevel variable 
+ * Notably, the identifiers that are unbound or refer to toplevel variable
  * has () in its env.
  * A caveat--we can't truncate frames at the time of construction, since
  * the entire frame structure may not be fixed while we're processing internal
- * defines.  The identifier may refer to another identifier that will be 
+ * defines.  The identifier may refer to another identifier that will be
  * inserted later.  Thus, we delay the truncation operation until it is
  * needed.
  * The ENV slot itself now contains (<flag> . <frames>), where <flag> is #f
  * if truncation hasn't be done, #t otherwise.  ENV should be treated as
- * an opaque data for the others; you should always get it with 
+ * an opaque data for the others; you should always get it with
  * Scm_IdentifierEnv(), and never directly access ENV slot itself.
  */
 
@@ -227,13 +183,70 @@ ScmSymbol *Scm_UnwrapIdentifier(ScmIdentifier *id)
     return SCM_SYMBOL(z);
 }
 
-/* returns global binding of the identifier */
+/* Returns global binding of the identifier */
 ScmGloc *Scm_IdentifierGlobalBinding(ScmIdentifier *id)
 {
     ScmIdentifier *z = Scm_OutermostIdentifier(id);
     return Scm_FindBinding(z->module, SCM_SYMBOL(z->name), 0);
 }
 
+/* Runtime reference/mutation of global variables.  The are called
+   from VM instructoins GREF and GSET, and will also be used from
+   the AOT compiled code.
+
+   We cache the result of global variable to GLOC lookup in VM instructions,
+   so these procedures can also store the looked up GLOC in *pgloc.  You can
+   pass NULL if you don't need it.
+
+   If the referenced variable is unbound, or is immutable when attempted to
+   set!, an error is thrown.
+ */
+ScmObj Scm_IdentifierGlobalRef(ScmIdentifier *id,
+                               ScmGloc **pgloc /* out */)
+{
+    ScmGloc *gloc = Scm_IdentifierGlobalBinding(id);
+    if (gloc == NULL) {
+        Scm_VMThrowException(Scm_VM(),
+                             Scm_MakeUnboundVariableError(id->name),
+                             0);
+    }
+    if (pgloc != NULL) *pgloc = gloc;
+
+    ScmObj v = Scm_GlocGetValue(gloc);
+    if (SCM_AUTOLOADP(v)) {
+        return Scm_ResolveAutoload(SCM_AUTOLOAD(v), 0);
+    } else {
+        return v;
+    }
+}
+
+void Scm_IdentifierGlobalSet(ScmIdentifier *id,
+                             ScmObj val,
+                             ScmGloc **pgloc /* out */)
+{
+    /* If runtime flag LIMIT_MODULE_MUTATION is set,
+       we search only for the id's module, so that set! won't
+       mutate bindings in the other module. */
+    ScmIdentifier *z = Scm_OutermostIdentifier(id);
+    int limit = SCM_VM_RUNTIME_FLAG_IS_SET(Scm_VM(), SCM_LIMIT_MODULE_MUTATION);
+    ScmGloc *gloc = Scm_FindBinding(z->module, SCM_SYMBOL(z->name),
+                                    (limit ? SCM_BINDING_STAY_IN_MODULE : 0));
+    if (gloc == NULL) {
+        /* Search again for meaningful error message */
+        if (limit) {
+            gloc = Scm_FindBinding(z->module, SCM_SYMBOL(z->name), 0);
+            if (gloc != NULL) {
+                Scm_Error("Can't mutate binding of %S, which is in another module",
+                          z->name);
+            }
+            /* FALLTHROUGH */
+        }
+        Scm_Error("Symbol not defined: %S", z->name);
+    }
+    SCM_GLOC_SET(gloc, val);
+    if (pgloc != NULL) *pgloc = gloc;
+}
+
 /* returns true if SYM has the same binding with ID in ENV. */
 int Scm_IdentifierBindingEqv(ScmIdentifier *id, ScmSymbol *sym, ScmObj env)
 {
@@ -314,8 +327,8 @@ typedef struct unwrap_ctx_rec {
     int immutable;              /* flag */
 } unwrap_ctx;
 
-static void register_location(unwrap_ctx *ctx, 
-                              ScmObj *loc, 
+static void register_location(unwrap_ctx *ctx,
+                              ScmObj *loc,
                               ScmObj ref)
 {
     if (!SCM_READ_REFERENCE_P(ref)) return;
@@ -359,8 +372,8 @@ static void patch_locations(unwrap_ctx *ctx)
 static ScmObj unwrap_rec(ScmObj form, unwrap_ctx *ctx)
 {
     if (!SCM_PTRP(form)) return form;
-    ScmDictEntry *e = Scm_HashCoreSearch(&ctx->history, 
-                                         (intptr_t)form, 
+    ScmDictEntry *e = Scm_HashCoreSearch(&ctx->history,
+                                         (intptr_t)form,
                                          SCM_DICT_GET);
     if (e) {
         /* We've visited FORM before.  If this is the second time,
@@ -379,11 +392,24 @@ static ScmObj unwrap_rec(ScmObj form, unwrap_ctx *ctx)
         if (ca == SCM_CAR(form) && cd == SCM_CDR(form)
             && (!ctx->immutable || Scm_ImmutablePairP(form))) {
             fill_history(e, form);
+            /* no need to create a new cons */
             return form;
         }
-        ScmObj p = (ctx->immutable
-                    ? Scm_MakeImmutablePair(ca, cd)
-                    : Scm_Cons(ca, cd));
+        /* We have to create a new cons.  Make sure to keep attributes. */
+        ScmObj p = SCM_NIL;
+        if (SCM_EXTENDED_PAIR_P(form)) {
+            if (ctx->immutable || Scm_ImmutablePairP(form)) {
+                p = Scm_MakeImmutablePair(ca, cd, Scm_PairAttr(SCM_PAIR(form)));
+            } else {
+                p = Scm_MakeExtendedPair(ca, cd, Scm_PairAttr(SCM_PAIR(form)));
+            }
+        } else {
+            if (ctx->immutable) {
+                p = Scm_MakeImmutablePair(ca, cd, SCM_NIL);
+            } else {
+                p = Scm_Cons(ca, cd);
+            }
+        }
         fill_history(e, p);
         register_location(ctx, &SCM_PAIR(p)->car, ca);
         register_location(ctx, &SCM_PAIR(p)->cdr, cd);
@@ -426,15 +452,15 @@ static ScmObj unwrap_rec(ScmObj form, unwrap_ctx *ctx)
     return form;
 }
 
-#if GAUCHE_API_VERSION < 1000
+#if GAUCHE_API_VERSION < 98
 ScmObj Scm_UnwrapSyntax(ScmObj form)
 {
     return Scm_UnwrapSyntax2(form, FALSE);
 }
 ScmObj Scm_UnwrapSyntax2(ScmObj form, int immutablep)
-#else  /* GAUCHE_API_VERSION >= 1000 */
+#else  /* GAUCHE_API_VERSION >= 98 */
 ScmObj Scm_UnwrapSyntax(ScmObj form, int immutablep)
-#endif /* GAUCHE_API_VERSION >= 1000 */
+#endif /* GAUCHE_API_VERSION >= 98 */
 {
     unwrap_ctx ctx;
     Scm_HashCoreInitSimple(&ctx.history, SCM_HASH_EQ, 0, NULL);
@@ -463,8 +489,6 @@ void Scm__InitCompaux(void)
     ScmModule *g = Scm_GaucheModule();
     ScmModule *gi = Scm_GaucheInternalModule();
 
-    Scm_InitStaticClass(SCM_CLASS_SYNTACTIC_CLOSURE, "<syntactic-closure>", g,
-                        synclo_slots, 0);
     Scm_InitStaticClass(SCM_CLASS_IDENTIFIER, "<identifier>", g,
                         identifier_slots, 0);
 
