@@ -1,7 +1,7 @@
 ;;;
 ;;; compile-3.scm - The compiler: Pass 3
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -91,6 +91,16 @@
 (define (pass3/$CONST iform labels) iform)
 (define (pass3/$IT iform labels) iform)
 
+(define (pass3/$DYNENV iform labels)
+  (let ([key (pass3/rec ($dynenv-key iform) labels)]
+        [val (pass3/rec ($dynenv-value iform) labels)]
+        [expr (pass3/rec ($dynenv-body iform) labels)])
+    (unless (and (eq? key ($dynenv-key iform))
+                 (eq? val ($dynenv-value iform)))
+      ($dynenv-kvs-set! iform `(,key ,val ,@($dynenv-opts iform))))
+    ($dynenv-body-set! iform expr)
+    iform))
+
 ;; If optimization:
 ;;
 ;;  If the 'test' clause of $IF node contains another $IF that has $IT in
@@ -240,6 +250,14 @@
   ($lambda-body-set! iform (pass3/rec ($lambda-body iform) labels))
   iform)
 
+(define (pass3/$CLAMBDA iform labels)
+  (let loop ([cs ($clambda-closures iform)] [r '()])
+    (if (null? cs)
+      (begin
+        ($clambda-closures-set! iform (reverse r))
+        iform)
+      (loop (cdr cs) (cons (pass3/rec (car cs) labels) r)))))
+
 (define (pass3/$LABEL iform labels)
   (unless (label-seen? labels iform)
     (label-push! labels iform)
@@ -301,6 +319,16 @@
            ;; call ($call ($lambda ...) args ...)
            (has-tag? proc $LAMBDA)
            (pass3/inline-call iform proc args labels)]
+          [;; Similar to above, but ($call ($seq ... ($lambda ..)) args ...)
+           ;; We can convert it ot ($seq ... ($call ($lamba ...) args...))
+           (and (has-tag? proc $SEQ)
+                (has-tag? (last ($seq-body proc)) $LAMBDA))
+           (pass3/$SEQ ($seq `(,@(drop-right ($seq-body proc) 1)
+                               ,($call ($*-src iform)
+                                       (last ($seq-body proc)) ;$lambda
+                                       ($call-args iform)
+                                       ($call-flag iform))))
+                       labels)]
           [;; If we get ($call ($gref proc) args ...) and proc is inlinable,
            ;; we can inline the call.
            (and-let* ([ (has-tag? proc $GREF) ]
@@ -352,9 +380,9 @@
 ;; we can't get such value from any one, returns #f.
 (define (pass3/constant-arguments iforms)
   (let loop ([iforms iforms] [args '()])
-    (cond 
+    (cond
      [(null? iforms) (reverse args)]
-     [($const? (car iforms)) 
+     [($const? (car iforms))
       (loop (cdr iforms) (cons ($const-value (car iforms)) args))]
      [(has-tag? (car iforms) $GREF)
       (and-let1 gloc (gref-inlinable-gloc (car iforms))
@@ -464,4 +492,3 @@
 
 ;; Dispatch table.
 (define *pass3-dispatch-table* (generate-dispatch-table pass3))
-
