@@ -1,7 +1,7 @@
 ;;;
 ;;; compile-4.scm - The compiler: Pass 4
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -115,12 +115,15 @@
                 ($let #f 'rec
                       (imap (^[x] ($lambda-lifted-var x)) lifted)
                       lifted
-                      iform.)))))))))    
+                      iform.)))))))))
 
 ;; Pass4 step1 - scan
 ;;   bs - List of lvars whose binding is introduced in the current scope.
 ;;   fs - List of free lvars found so far in the current scope.
-;;   t? - #t if we're in the top level, #f otherwise.
+;;   t? - #t in toplevel expression, not inside another $lambda or $let.
+;;        $lambda node appears in this context doesn't need to be lifted,
+;;        for they're created at the time the entire toplevel expression
+;;        is created anyway.
 ;;   labels - label-dic.  the info field is used to hold $LAMBDA nodes.
 ;; Eacl call returns a list of free lvars.
 
@@ -139,8 +142,10 @@
 
 (define/case (pass4/scan iform bs fs t? labels)
   (iform-tag iform)
-  [($DEFINE) (unless t? (error "[internal] pass4 $DEFINE in non-toplevel"))
-             (pass4/scan ($define-expr iform) bs fs #t labels)]
+  [($DEFINE) (unless t?
+               (warn "define appears in non-top-level expression: ~s"
+                     (unravel-syntax ($define-src iform))))
+             (pass4/scan ($define-expr iform) bs fs t? labels)]
   [($LREF)   (pass4/add-lvar ($lref-lvar iform) bs fs)]
   [($LSET)   (let1 fs (pass4/scan ($lset-expr iform) bs fs t? labels)
                (pass4/add-lvar ($lset-lvar iform) bs fs))]
@@ -150,9 +155,9 @@
                (pass4/scan ($if-else iform) bs fs t? labels))]
   [($LET)    (let* ([new-bs (append ($let-lvars iform) bs)]
                     [bs (if (memv ($let-type iform) '(rec rec*)) new-bs bs)]
-                    [fs (pass4/scan* ($let-inits iform) bs fs t? labels)])
+                    [fs (pass4/scan* ($let-inits iform) bs fs #f labels)])
                (pass4/scan ($let-body iform) new-bs fs #f labels))]
-  [($RECEIVE)(let ([fs (pass4/scan ($receive-expr iform) bs fs t? labels)]
+  [($RECEIVE)(let ([fs (pass4/scan ($receive-expr iform) bs fs #f labels)]
                    [bs (append ($receive-lvars iform) bs)])
                (pass4/scan ($receive-body iform) bs fs #f labels))]
   [($LAMBDA) (let1 inner-fs (pass4/scan ($lambda-body iform)
@@ -161,7 +166,7 @@
                ;; we don't need to lift it, nor need to set free-lvars.
                ;; We just mark it by setting lifted-var to #t so that
                ;; pass4/lift phase can treat it specially.
-               (unless (eq? ($lambda-flag iform) 'dissolved)
+               (unless ($lambda-dissolved? iform)
                  (label-dic-info-push! labels iform) ;save the lambda node
                  (when t?                            ;mark this is toplevel
                    ($lambda-lifted-var-set! iform #t)))
@@ -172,9 +177,10 @@
                                fs
                                (loop (cdr inner-fs)
                                      (pass4/add-lvar (car inner-fs) bs fs))))]))]
+  [($CLAMBDA) (pass4/scan* ($clambda-closures iform) bs fs t? labels)]
   [($LABEL)  (cond [(label-seen? labels iform) fs]
                    [else (label-push! labels iform)
-                         (pass4/scan ($label-body iform) bs fs #f labels)])]
+                         (pass4/scan ($label-body iform) bs fs t? labels)])]
   [($SEQ)    (pass4/scan* ($seq-body iform) bs fs t? labels)]
   [($CALL)   (let1 fs (if (eq? ($call-flag iform) 'jump)
                         fs
@@ -184,6 +190,9 @@
   [($CONS $APPEND $MEMV $EQ? $EQV?) (pass4/scan2 iform bs fs t? labels)]
   [($VECTOR $LIST $LIST*) (pass4/scan* ($*-args iform) bs fs t? labels)]
   [($LIST->VECTOR) (pass4/scan ($*-arg0 iform) bs fs t? labels)]
+  [($DYNENV) (let* ([fs (pass4/scan ($dynenv-key iform) bs fs #f labels)]
+                    [fs (pass4/scan ($dynenv-value iform) bs fs #f labels)])
+               (pass4/scan ($dynenv-body iform) bs fs t? labels))]
   [else fs])
 
 (define (pass4/scan2 iform bs fs t? labels)
@@ -228,28 +237,35 @@
               (set! top-name (identifier->symbol n))
               #t)
             (loop (cdr lms)))))
-    (rlet1 results '()
-      (let loop ([lms lambda-nodes])
-        (cond [(null? lms)]
+    ;; Loop over liftable lambda candidates.  Lifting one lambda node
+    ;; can turn another lambda nodes liftable, so we repeat the pass until
+    ;; we can't find liftable ones.
+    (rlet1 lifted '()
+      (let loop ([lms lambda-nodes]
+                 [lifted? #f]
+                 [remaining '()])
+        (cond [(null? lms) (when lifted? (loop remaining #f '()))] ;repeat
               [($lambda-lifted-var (car lms)) ; this is toplevel node
                ($lambda-lifted-var-set! (car lms) #f)
-               (loop (cdr lms))]
+               (loop (cdr lms) lifted? remaining)]
               [else
                (let* ([lm (car lms)]
                       [fvs ($lambda-free-lvars lm)])
-                 (when (or (null? fvs)
-                           (and (null? (cdr fvs))
-                                (lvar-immutable? (car fvs))
-                                (eq? (lvar-initval (car fvs)) lm)))
+                 (if (or (null? fvs)
+                         (and (null? (cdr fvs))
+                              (lvar-immutable? (car fvs))
+                              (memq (lvar-initval (car fvs)) lifted)))
                    (let1 lvar (make-lvar (gensym))
                      ($lambda-name-set! lm (list top-name
                                                  (or ($lambda-name lm)
                                                      (lvar-name lvar))))
                      ($lambda-lifted-var-set! lm lvar)
-                     (push! results lm)))
-                 (loop (cdr lms)))])))))
+                     (push! lifted lm)
+                     (loop (cdr lms) #t remaining))
+                   (loop (cdr lms) lifted? (cons lm remaining))))])))))
 
-;; Final touch of pass4 - replace lifted lambda nodes to $GREFs.
+;; Final touch of pass4 - replace lifted lambda nodes to the reference of
+;; pre-bound lvars.
 ;; Returns (possibly modified) IForm.
 (define-macro (pass4/subst! access-form labels)
   (match-let1 (accessor expr) access-form
@@ -301,6 +317,7 @@
              (or (and-let* ([lifted ($lambda-lifted-var iform)])
                    ($lref lifted))
                  iform)]
+  [($CLAMBDA) (pass4/subst*! ($clambda-closures iform) labels) iform]
   [($LABEL)  (unless (label-seen? labels iform)
                (label-push! labels iform)
                (pass4/subst! ($label-body iform) labels))
@@ -313,5 +330,8 @@
              (pass4/subst! ($*-arg1 iform) labels)]
   [($VECTOR $LIST $LIST*) (pass4/subst*! ($*-args iform) labels) iform]
   [($LIST->VECTOR) (pass4/subst! ($*-arg0 iform) labels)]
+  [($DYNENV) (pass4/subst! (car ($dynenv-kvs iform)) labels)
+             (pass4/subst! (car (cdr ($dynenv-kvs iform))) labels)
+             (pass4/subst! ($dynenv-body iform) labels)
+             iform]
   [else iform])
-
