@@ -1,7 +1,7 @@
 ;;;
 ;;; compile-i.scm - Inliners of builtin procedures
 ;;;
-;;;   Copyright (c) 2004-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2004-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -151,7 +151,7 @@
 (define-builtin-inliner--/ /. NUMIDIV2 ensure-inexact-const)
 
 ;; NB: If we detect exact division-by-zero case, we shouldn't fold
-;; the constant and let it be handled at runtime.  
+;; the constant and let it be handled at runtime.
 (define-builtin-inliner /
   (^[src args]
     (match args
@@ -199,6 +199,7 @@
     (match (classify-args args)
       [(#f)         ($const unit)]
       [(constval)   ($const constval)]
+      [(#f x)       (undefined)]
       [(constval x) ($asm src `(,opcode) (list ($const constval) x))]
       [(#f x y)     ($asm src `(,opcode) (list x y))]
       [_ (undefined)])))
@@ -234,7 +235,7 @@
          (match args
            [(vec ind) ($asm src `(,UVEC-REF ,,%type) `(,vec ,ind))]
            [else (undefined)])))))
-    
+
 (define-builtin-inliner-uvref s8 S8)
 (define-builtin-inliner-uvref u8 U8)
 (define-builtin-inliner-uvref s16 S16)
@@ -247,6 +248,12 @@
 (define-builtin-inliner-uvref f32 F32)
 (define-builtin-inliner-uvref f64 F64)
 
+(define-builtin-inliner uvector-ref
+  (^[src args]
+    (match args
+      [(vec ind) ($asm src `(,UVEC-REF ,SCM_UVECTOR_GENERIC) `(,vec ,ind))]
+      [else (undefined)])))
+
 (define-builtin-inliner zero?
   (^[src args]
     (match args
@@ -283,6 +290,17 @@
       [() ($asm src `(,CURERR) '())]
       [else (undefined)])))
 
+(define-builtin-inliner with-exception-handler
+  (^[src args]
+    (match args
+      [(handler (? (cut has-tag? <> $LAMBDA) lmda))
+       ($dynenv src
+                (list ($call #f ($gref %exception-handler-mark-key.) '())
+                      handler
+                      '(push))
+                ($lambda-body lmda))]
+      [else (undefined)])))
+
 (define-builtin-inliner dynamic-wind
   (^[src args]
     (match args
@@ -293,23 +311,25 @@
              [r (make-lvar 'tmp)])
          (if (constant-lambda? a)
            ;; when after thunk is dummy, we don't bother to call it.
+           ;; NB: POP-HANDLERS does not affect value registers, so we don't
+           ;; need to save the result of thunk.
            ($let src 'let `(,at ,bt ,tt) `(,a ,b ,t)
                  ($seq
-                  `(,($call ($*-src b) ($lref bt) '())
-                    ,($asm src `(,PUSH-HANDLERS) `(,($lref bt) ,($lref at)))
-                    ,($call ($*-src t) ($lref tt) '()))))
+                  `(,($call (iform-src b) ($lref bt) '())
+                    ,($asm src `(,PUSH-HANDLERS 0) `(,($lref bt) ,($lref at)))
+                    ,($call (iform-src t) ($lref tt) '())
+                    ,($asm src `(,POP-HANDLERS) '()))))
            ;; normal path
            ($let src 'let `(,at ,bt ,tt) `(,a ,b ,t)
                  ($seq
-                  `(,($call ($*-src b) ($lref bt) '())
-                    ,($asm src `(,PUSH-HANDLERS) `(,($lref bt) ,($lref at)))
+                  `(,($call (iform-src b) ($lref bt) '())
+                    ,($asm src `(,PUSH-HANDLERS 0) `(,($lref bt) ,($lref at)))
                     ,($receive #f 0 1 (list r)
-                               ($call ($*-src t) ($lref tt) '())
+                               ($call (iform-src t) ($lref tt) '())
                                ($seq
                                 `(,($asm src `(,POP-HANDLERS) '())
-                                  ,($call ($*-src a) ($lref at) '())
+                                  ,($call (iform-src a) ($lref at) '())
                                   ,($asm #f `(,TAIL-APPLY 2)
                                          (list ($gref values.) ($lref r))))))
                     )))))]
       [_ (undefined)])))
-
