@@ -1,7 +1,7 @@
 /*
  * core.c - core kernel interface
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,6 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/builtin-syms.h"
 
 /* GC_print_static_roots() is declared in private/gc_priv.h.  It is too much
@@ -69,6 +70,7 @@ extern void Scm__InitSymbol(void);
 extern void Scm__InitNumber(void);
 extern void Scm__InitChar(void);
 extern void Scm__InitClass(void);
+extern void Scm__InitMemoTable(void);
 extern void Scm__InitList(void);
 extern void Scm__InitExceptions(void);
 extern void Scm__InitPort(void);
@@ -76,6 +78,7 @@ extern void Scm__InitWrite(void);
 extern void Scm__InitCompaux(void);
 extern void Scm__InitMacro(void);
 extern void Scm__InitLoad(void);
+extern void Scm__InitThreadLocal(void);
 extern void Scm__InitParameter(void);
 extern void Scm__InitProc(void);
 extern void Scm__InitRegexp(void);
@@ -87,9 +90,17 @@ extern void Scm__InitAutoloads(void);
 extern void Scm__InitCollection(void);
 extern void Scm__InitComparator(void);
 extern void Scm__InitExecenv(void);
+extern void Scm__InitLazy(void);
+extern void Scm__InitNative(void);
+extern void Scm__InitNet(void);
+extern void Scm__InitNetAddr(void);
+extern void Scm__InitNetDb(void);
+extern void Scm__InitMutex(void);
+extern void Scm__InitThreads(void);
 
 extern void Scm_Init_libalpha(void);
 extern void Scm_Init_libbool(void);
+extern void Scm_Init_libbox(void);
 extern void Scm_Init_libchar(void);
 extern void Scm_Init_libcode(void);
 extern void Scm_Init_libcmp(void);
@@ -97,11 +108,15 @@ extern void Scm_Init_libdict(void);
 extern void Scm_Init_libeval(void);
 extern void Scm_Init_libexc(void);
 extern void Scm_Init_libfmt(void);
+extern void Scm_Init_libhash(void);
 extern void Scm_Init_libio(void);
 extern void Scm_Init_liblazy(void);
 extern void Scm_Init_liblist(void);
+extern void Scm_Init_libmemo(void);
 extern void Scm_Init_libmisc(void);
 extern void Scm_Init_libmod(void);
+extern void Scm_Init_libnative(void);
+extern void Scm_Init_libnet(void);
 extern void Scm_Init_libnum(void);
 extern void Scm_Init_libobj(void);
 extern void Scm_Init_libproc(void);
@@ -110,6 +125,8 @@ extern void Scm_Init_libsrfis(void);
 extern void Scm_Init_libstr(void);
 extern void Scm_Init_libsym(void);
 extern void Scm_Init_libsys(void);
+extern void Scm_Init_libthr(void);
+extern void Scm_Init_libtype(void);
 extern void Scm_Init_libvec(void);
 extern void Scm_Init_libmacbase(void);
 extern void Scm_Init_compile(void);
@@ -117,6 +134,8 @@ extern void Scm_Init_libmacro(void);
 extern void Scm_Init_libparam(void);
 extern void Scm_Init_libomega(void);
 
+extern void Scm__FinishModuleInitialization(void);
+
 static void finalizable(void);
 static void init_cond_features(void);
 
@@ -127,15 +146,17 @@ static int (*ptr_pthread_create)(pthread_t *,
                                  void *(*)(void *), void * /* arg */) = NULL;
 #endif
 
-/* flag to see if Scheme infrastructure is fully initialized or not */
-static int scheme_initialized = FALSE;
+/* To track how far we've gone through booting stages.  See gauche.h
+   This variable is changed before any other threads are created, so no mutex
+   is needed. */
+static ScmRuntimeState runtime_state = SCM_RUNTIME_INITIALIZING;
 
 /*
  * Entry point of initlalizing Gauche runtime
  */
 void Scm_Init(const char *signature)
 {
-    if (scheme_initialized) return;
+    if (runtime_state > SCM_RUNTIME_INITIALIZING) return;
 
     int debug_init = FALSE;
     if (Scm_GetEnv("GAUCHE_DEBUG_INITIALIZATION") != NULL) {
@@ -146,7 +167,7 @@ void Scm_Init(const char *signature)
         if (debug_init) fprintf(stderr, "Calling %s...\n", #f); \
         f();                                                    \
     } while (0)
-    
+
     /* make sure the main program links the same version of libgauche */
     if (strcmp(signature, GAUCHE_SIGNATURE) != 0) {
         Scm_Panic("libgauche ABI version mismatch: libgauche %s, expected %s",
@@ -160,14 +181,20 @@ void Scm_Init(const char *signature)
     /* Set up GC parameters.  We need to call finalizers at the safe
        point of VM loop, so we disable auto finalizer invocation, and
        ask GC to call us back when finalizers are queued. */
-    GC_oom_fn = oom_handler;
-    GC_finalize_on_demand = TRUE;
-    GC_finalizer_notifier = finalizable;
+    GC_set_oom_fn(oom_handler);
+    GC_set_finalize_on_demand(TRUE);
+    GC_set_finalizer_notifier(finalizable);
+
+    /* Newer bdwgc delays spawning marker threads until the client creates
+       first thread.  We can take advantage of parallel markers even with
+       single-threaded program, so we ask them to go parallel now.  */
+    GC_start_mark_threads();
 
     (void)SCM_INTERNAL_MUTEX_INIT(cond_features.mutex);
 
     /* Initialize components.  The order is important, for some components
        rely on the other components to be initialized. */
+    CALL_INIT(Scm__InitThreadLocal);
     CALL_INIT(Scm__InitParameter);
     CALL_INIT(Scm__InitVM);
     CALL_INIT(Scm__InitHash);
@@ -176,6 +203,7 @@ void Scm_Init(const char *signature)
     CALL_INIT(Scm__InitNumber);
     CALL_INIT(Scm__InitChar);
     CALL_INIT(Scm__InitClass);
+    CALL_INIT(Scm__InitMemoTable);
     CALL_INIT(Scm__InitList);
     CALL_INIT(Scm__InitCollection);
     CALL_INIT(Scm__InitExceptions);
@@ -190,21 +218,31 @@ void Scm_Init(const char *signature)
     CALL_INIT(Scm__InitSystem);
     CALL_INIT(Scm__InitComparator);
     CALL_INIT(Scm__InitExecenv);
+    CALL_INIT(Scm__InitNative);
+    CALL_INIT(Scm__InitNet);
+    CALL_INIT(Scm__InitNetAddr);
+    CALL_INIT(Scm__InitNetDb);
+    CALL_INIT(Scm__InitMutex);
+    CALL_INIT(Scm__InitThreads);
 
     CALL_INIT(Scm_Init_libalpha);
     CALL_INIT(Scm_Init_libbool);
+    CALL_INIT(Scm_Init_libbox);
     CALL_INIT(Scm_Init_libchar);
     CALL_INIT(Scm_Init_libcode);
     CALL_INIT(Scm_Init_libcmp);
-    CALL_INIT(Scm_Init_libdict);
     CALL_INIT(Scm_Init_libeval);
     CALL_INIT(Scm_Init_libexc);
     CALL_INIT(Scm_Init_libfmt);
+    CALL_INIT(Scm_Init_libhash);
     CALL_INIT(Scm_Init_libio);
     CALL_INIT(Scm_Init_liblazy);
     CALL_INIT(Scm_Init_liblist);
+    CALL_INIT(Scm_Init_libmemo);
     CALL_INIT(Scm_Init_libmisc);
     CALL_INIT(Scm_Init_libmod);
+    CALL_INIT(Scm_Init_libnative);
+    CALL_INIT(Scm_Init_libnet);
     CALL_INIT(Scm_Init_libnum);
     CALL_INIT(Scm_Init_libobj);
     CALL_INIT(Scm_Init_libproc);
@@ -216,8 +254,13 @@ void Scm_Init(const char *signature)
     CALL_INIT(Scm_Init_libvec);
     CALL_INIT(Scm_Init_libmacbase);
     CALL_INIT(Scm_Init_compile);
+    /* lib* that uses er-macro-transformer must be initialized
+       after libmacro */
     CALL_INIT(Scm_Init_libmacro);
     CALL_INIT(Scm_Init_libparam);
+    CALL_INIT(Scm_Init_libtype);
+    CALL_INIT(Scm_Init_libthr);
+    CALL_INIT(Scm_Init_libdict);
     CALL_INIT(Scm_Init_libomega);
 
     CALL_INIT(Scm__InitCompaux);
@@ -235,14 +278,57 @@ void Scm_Init(const char *signature)
     ptr_pthread_create = GC_pthread_create;
 #endif
 
-    scheme_initialized = TRUE;
+    Scm__FinishModuleInitialization();
+    runtime_state = SCM_RUNTIME_INITIALIZED;
 }
 
-int Scm_InitializedP()
+/*
+ * Runtime state
+ */
+
+ScmRuntimeState Scm_RuntimeState()
 {
-    return scheme_initialized;
+    return runtime_state;
 }
 
+/* This should only be called from main.c or equivalent.  */
+void Scm_SetRuntimeReplState(int full)
+{
+    runtime_state = full? SCM_RUNTIME_FULL_REPL : SCM_RUNTIME_MINI_REPL;
+}
+
+/*=============================================================
+ * Memory utilities
+ */
+
+#if SCM_ATOMIC_NEED_HELPER
+/* This is only needed if we fall back to libatomic_ops and not GNUC */
+#include "gauche/priv/atomicP.h"
+
+int Scm__AtomicCompareExchange(ScmAtomicVar *loc,
+                               ScmAtomicWord *expectedloc,
+                               ScmAtomicWord newval)
+{
+    ScmAtomicWord expected = *expectedloc;
+    ScmAtomicWord oldval = AO_fetch_compare_and_swap_full(loc,
+                                                          expected,
+                                                          newval);
+    *expectedloc = expected;
+    return oldval == expected;
+}
+
+ScmAtomicWord Scm__AtomicExchange(ScmAtomicVar *loc,
+                                  ScmAtomicWord newval)
+{
+    ScmAtomicWord current;
+    do {
+        current = *loc;
+    } while (AO_compare_and_swap_full(loc, current, newval));
+    return current;
+}
+#endif /*SCM_ATOMIC_NEED_HELPER*/
+
+
 /*=============================================================
  * GC utilities
  */
@@ -372,13 +458,11 @@ void Scm_DeleteCleanupHandler(void *handle)
     }
 }
 
-#if defined(GAUCHE_HAS_THREADS)
 void Scm__MutexCleanup(void *mutex_)
 {
     ScmInternalMutex *mutex = mutex_;
     (void)SCM_INTERNAL_MUTEX_UNLOCK(*mutex);
 }
-#endif /* GAUCHE_HAS_THREADS */
 
 
 /* Scm_Cleanup and Scm_Exit
@@ -401,17 +485,7 @@ void Scm_Cleanup(void)
     if (!cleanup.dirty) return;
     cleanup.dirty = FALSE;
 
-    /* Execute pending dynamic handlers.
-       NB: we ignore errors here.  It may not be accurate behavior, though.
-       (A handler may intentionally raise an error to skip the rest of
-       handlers).  We may change to break from SCM_FOR_EACH in case if
-       we detect error in Scm_Apply. */
-    ScmVM *vm = Scm_VM();
-    ScmObj hp;
-    SCM_FOR_EACH(hp, vm->handlers) {
-        vm->handlers = SCM_CDR(hp);
-        Scm_Apply(SCM_CDAR(hp), SCM_NIL, NULL);
-    }
+    Scm_VMFlushDynamicHandlers();
 
     /* Call the C-registered cleanup handlers. */
     for (struct cleanup_handler_rec *ch = cleanup.handlers; ch; ch = ch->next) {
@@ -422,6 +496,21 @@ void Scm_Cleanup(void)
     Scm_FlushAllPorts(TRUE);
 }
 
+int Scm_ObjToExitCode(ScmObj obj)
+{
+    if (SCM_EQ(obj, SCM_TRUE)) {
+        return 0;
+    } else if (SCM_EQ(obj, SCM_FALSE)) {
+        return 1;
+    } else if (SCM_INTP(obj)) {
+        return SCM_INT_VALUE(obj) & 0xff;
+    } else if (SCM_INTEGERP(obj)) {
+        return Scm_GetInteger(Scm_LogAnd(obj, SCM_MAKE_INT(0xff)));
+    } else {
+        return 70;              /* EX_SOFTWARE */
+    }
+}
+
 void Scm_Panic(const char *msg, ...)
 {
     va_list args;
@@ -520,7 +609,7 @@ init_cond_features()
         /* NB: We should probably make checking 'r7rs' trigger loading
            r7rs module and setting up the environment. */
         { "r7rs", NULL },
-            
+
         /* R7RS Appendix B */
         { "exact-closed", NULL },
         // { "exact-complex", NULL }, /* not yet */
@@ -532,7 +621,8 @@ init_cond_features()
 #else
         { "posix", NULL },
 #endif
-        /* TODO: OS Flags, CPU arch flags and C memory model flags */
+        /* OS/CPU arch flags are generated to features.c by gen-features.sh */
+        /* TODO: C memory model flags */
 #if   defined(WORDS_BIGENDIAN)
         { "big-endian", NULL },
 #else
@@ -541,7 +631,8 @@ init_cond_features()
 
         /* SRFIs that are not libraries */
         { "srfi-22", NULL },    /* Scheme scripts */
-        
+        { "srfi-138", NULL },    /* compile-r7rs script */
+
         /* Threads */
 #if   defined(GAUCHE_USE_PTHREADS)
         { "gauche.sys.threads", "gauche.threads" },
@@ -553,12 +644,9 @@ init_cond_features()
 
         /* TLS/SSL.  This nees to be in the core in order to switch
            code _before_ loading rfc.tls */
-#if defined(GAUCHE_USE_AXTLS) || defined(GAUCHE_USE_MBEDTLS)
+#if defined(GAUCHE_USE_MBEDTLS)
         { "gauche.net.tls", "rfc.tls" },
 #endif
-#if defined(GAUCHE_USE_AXTLS)
-        { "gauche.net.tls.axtls", "rfc.tls" },
-#endif
 #if defined(GAUCHE_USE_MBEDTLS)
         /* NB: Kludge - mbedTLS is implemented in a separate module rfc.tls.mbed,
            but we don't autoload it with cond-expand gauche.net.tls.mbedtls.
@@ -581,7 +669,7 @@ init_cond_features()
         { "regexp-unicode", "srfi.115" },
 
 #include "features.c"
-        
+
         { NULL, NULL }
     };
 
@@ -590,36 +678,13 @@ init_cond_features()
     }
 }
 
-/* TRANSIENT: These used to be in paths.c but no longer needed.  We keep
-   the entries just for ABI compatibility.  These shouldn't be called.
-   Remove on 1.0 release.
-*/
-#if GAUCHE_API_VERSION < 1000
-const char *Scm_GetLibraryDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
+#include "buildinfo.c"
+
+ScmObj Scm_BuildGoshVersion()
 {
-    Scm_Abort("Scm_GetLibraryDirectory is called (it shouldn't).");
+    return SCM_MAKE_STR(build_gosh_version);
 }
 
-const char *Scm_GetArchitectureDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
-{
-    Scm_Abort("Scm_GetArchitectureDirectory is called (it shouldn't).");
-}
-
-const char *Scm_GetSiteLibraryDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
-{
-    Scm_Abort("Scm_GetSiteLibraryDirectory is called (it shouldn't).");
-}
-
-const char *Scm_GetSiteArchitectureDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
-{
-    Scm_Abort("Scm_GetSiteArchitectureDirectory is called (it shouldn't).");
-}
-
-const char *Scm_GetRuntimeDirectory(void (*errfn)(const char *, ...) SCM_UNUSED)
-{
-    Scm_Abort("Scm_GetRuntimeDirectory is called (it shouldn't).");
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
 
 /*=============================================================
  * 'Main'
@@ -662,9 +727,12 @@ void Scm_SimpleMain(int argc, const char *argv[],
 
         ScmEvalPacket epak;
         int r = Scm_Apply(run_main_proc, SCM_LIST2(mainproc, args), &epak);
-        SCM_ASSERT(r == 1 && SCM_INTP(epak.results[0]));
-        Scm_Exit(SCM_INT_VALUE(epak.results[0]));
+        if (r >= 0 && SCM_INTP(epak.results[0])) {
+            Scm_Exit(SCM_INT_VALUE(epak.results[0]));
+        } else {
+            Scm_Exit(70);       /* EX_SOFTWARE */
+        }
     } else {
-        Scm_Exit(70);
+        Scm_Exit(0);
     }
 }
