@@ -1,7 +1,7 @@
 /*
  * dispatch.c - method dispatch accelerator
  *
- *   Copyright (c) 2017-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2017-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,6 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/atomicP.h"
 #include "gauche/priv/dispatchP.h"
 
@@ -44,7 +45,7 @@
  *
  *  In the current implementation, we use special open-addressing hash table.
  *  We don't use ScmHashCore for a few reasons:
- *  
+ *
  *   - We mutex modification of the table by gf->mutex, but we don't want
  *     to lock for searching it.  ScmHashCore doesn't guarantee consistency
  *     of reading while modifying.
@@ -89,7 +90,7 @@
  *   - We don't need to allocate to construct the list of applicable
  *     methods.
  *
- *  Note that if the entry has at least one non-leaf methods, we have to
+ *  Note that if the entry has at least one non-leaf method, we have to
  *  fall back to the normal path, since there may be other applicable
  *  methods that might be called via next-method.  To cut the overhead
  *  of checking, we keep the list of leaf methods and the one of non-leaf
@@ -117,7 +118,7 @@
 struct ScmMethodDispatcherRec {
     int axis;                    /* Which argument we look at?
                                     This is immutable. */
-    ScmAtomicVar methodHash;	 /* mhash.  In case mhash is extended,
+    ScmAtomicVar methodHash;     /* mhash.  In case mhash is extended,
                                     we atomically swap reference. */
 };
 
@@ -131,7 +132,7 @@ typedef struct mhash_entry_rec {
 typedef struct mhash_rec {
     int size;                   /* # of bins.  power of 2. */
     int num_entries;            /* # of active entries.  */
-    ScmAtomicVar bins[1];	/* Table.  Each entry may have one of:
+    ScmAtomicVar bins[1];       /* Table.  Each entry may have one of:
                                      0 - free
                                      1 - deleted
                                      mhash_entry
@@ -165,10 +166,10 @@ static ScmObj mhash_probe(const mhash *h, ScmClass *k, int nargs)
     u_long j = mhashfn(k, nargs) & (h->size - 1);
     int i = 0;
     for (; i < h->size; i++) {
-	/* Need to strip 'const', because of C11 error
-	   http://www.open-std.org/jtc1/sc22/wg14/www/docs/summary.htm#dr_459 */
-	ScmAtomicVar *loc = (ScmAtomicVar*)&h->bins[j];
-        ScmWord w = SCM_WORD(AO_load(loc));
+        /* Need to strip 'const', because of C11 error
+           http://www.open-std.org/jtc1/sc22/wg14/www/docs/summary.htm#dr_459 */
+        ScmAtomicVar *loc = (ScmAtomicVar*)&h->bins[j];
+        ScmWord w = SCM_WORD(Scm_AtomicLoad(loc));
         if (w == 0) break;
         if (w != 1) {
             mhash_entry *e = (mhash_entry*)w;
@@ -189,7 +190,7 @@ static mhash *mhash_insert_1(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
     ScmObj ltail = SCM_NIL, ntail = SCM_NIL;
     int i = 0;
     for (; i < h->size; i++) {
-        ScmWord w = SCM_WORD(AO_load(&h->bins[j]));
+        ScmWord w = SCM_WORD(Scm_AtomicLoad(&h->bins[j]));
         if (w == 0) {           /* end of chain */
             if (free_slot < 0) free_slot = j;
             break;
@@ -207,7 +208,7 @@ static mhash *mhash_insert_1(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
             break;
         }
         j = (j + i + 1) & (h->size - 1);
-        
+
     }
     SCM_ASSERT(free_slot >= 0);
     mhash_entry *e = SCM_NEW(mhash_entry);
@@ -215,7 +216,7 @@ static mhash *mhash_insert_1(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
     e->nargs = nargs;
     e->leaves = SCM_METHOD_LEAF_P(m)? Scm_Cons(SCM_OBJ(m), ltail) : ltail;
     e->nonleaves = SCM_METHOD_LEAF_P(m) ? ntail : Scm_Cons(SCM_OBJ(m), ntail);
-    AO_store_full(&h->bins[free_slot], (ScmAtomicWord)e);
+    Scm_AtomicStoreFull(&h->bins[free_slot], (ScmAtomicWord)e);
     h->num_entries++;
     return h;
 }
@@ -249,10 +250,10 @@ static mhash *mhash_insert(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
 static mhash *mhash_delete(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
 {
     u_long j = mhashfn(k, nargs) & (h->size - 1);
-    
+
     int i = 0;
     for (; i < h->size; i++) {
-        ScmWord w = SCM_WORD(AO_load(&h->bins[j]));
+        ScmWord w = SCM_WORD(Scm_AtomicLoad(&h->bins[j]));
         if (w == 0) break;
         if (w == 1) continue;
         mhash_entry *e = (mhash_entry*)w;
@@ -272,14 +273,14 @@ static mhash *mhash_delete(mhash *h, ScmClass *k, int nargs, ScmMethod *m)
 
             if (SCM_NULLP(ml) && SCM_NULLP(ml)) {
                 h->num_entries--;
-                AO_store(&h->bins[j], 1); /* mark as deleted */
+                Scm_AtomicStore(&h->bins[j], 1); /* mark as deleted */
             } else {
                 mhash_entry *e = SCM_NEW(mhash_entry);
                 e->klass = k;
                 e->nargs = nargs;
                 e->leaves = ml;
                 e->nonleaves = mn;
-                AO_store_full(&h->bins[j], (ScmAtomicWord)e);
+                Scm_AtomicStoreFull(&h->bins[j], (ScmAtomicWord)e);
             }
             break;
         }
@@ -299,7 +300,7 @@ static void mhash_print(mhash *h, ScmPort *out)
             Scm_Printf(out, "[%3d] deleted\n\n\n", i);
         } else {
             mhash_entry *e = (mhash_entry*)w;
-            Scm_Printf(out, "[%3d] %lu %S(%d)\n", i, 
+            Scm_Printf(out, "[%3d] %lu %S(%d)\n", i,
                        (mhashfn(e->klass, e->nargs) % h->size),
                        e->klass, e->nargs);
             Scm_Printf(out, "  Leaves:   %S\n", e->leaves);
@@ -339,7 +340,7 @@ static mhash *delete_method_from_dispatcher(mhash *h, int axis, ScmMethod *m)
 }
 
 /*
-    NB: We run through the method list twice, first process the 
+    NB: We run through the method list twice, first process the
     leaf methods, and then process non-leaf methods.  Non-leaf methods
     cancels the dispatcher entry and forces to go through normal route.
  */
@@ -364,16 +365,16 @@ ScmMethodDispatcher *Scm__BuildMethodDispatcher(ScmObj methods, int axis)
 
 void Scm__MethodDispatcherAdd(ScmMethodDispatcher *dis, ScmMethod *m)
 {
-    mhash *h = (mhash*)AO_load(&dis->methodHash);
+    mhash *h = (mhash*)Scm_AtomicLoad(&dis->methodHash);
     mhash *h2 = add_method_to_dispatcher(h, dis->axis, m);
-    if (h != h2) AO_store(&dis->methodHash, (ScmAtomicWord)h2);
+    if (h != h2) Scm_AtomicStore(&dis->methodHash, (ScmAtomicWord)h2);
 }
 
 void Scm__MethodDispatcherDelete(ScmMethodDispatcher *dis, ScmMethod *m)
 {
-    mhash *h = (mhash*)AO_load(&dis->methodHash);
+    mhash *h = (mhash*)Scm_AtomicLoad(&dis->methodHash);
     mhash *h2 = delete_method_from_dispatcher(h, dis->axis, m);
-    if (h != h2) AO_store(&dis->methodHash, (ScmAtomicWord)h2);
+    if (h != h2) Scm_AtomicStore(&dis->methodHash, (ScmAtomicWord)h2);
 }
 
 ScmObj Scm__MethodDispatcherLookup(ScmMethodDispatcher *dis,
@@ -381,7 +382,7 @@ ScmObj Scm__MethodDispatcherLookup(ScmMethodDispatcher *dis,
 {
     if (dis->axis <= argc) {
         ScmClass *selector = typev[dis->axis];
-        mhash *h = (mhash*)AO_load(&dis->methodHash);
+        mhash *h = (mhash*)Scm_AtomicLoad(&dis->methodHash);
         return mhash_probe(h, selector, argc);
     } else {
         return SCM_FALSE;
@@ -394,7 +395,7 @@ ScmObj Scm__MethodDispatcherInfo(const ScmMethodDispatcher *dis)
     /* Need to strip 'const', because of C11 error
        http://www.open-std.org/jtc1/sc22/wg14/www/docs/summary.htm#dr_459 */
     ScmAtomicVar *loc = (ScmAtomicVar*)&dis->methodHash;
-    const mhash *mh = (const mhash*)AO_load(loc);
+    const mhash *mh = (const mhash*)Scm_AtomicLoad(loc);
     SCM_APPEND1(h, t, SCM_MAKE_KEYWORD("axis"));
     SCM_APPEND1(h, t, SCM_MAKE_INT(dis->axis));
     SCM_APPEND1(h, t, SCM_MAKE_KEYWORD("num-entries"));
@@ -407,4 +408,3 @@ void Scm__MethodDispatcherDump(ScmMethodDispatcher *dis, ScmPort *port)
     Scm_Printf(port, "MethodDispatcher axis=%d\n", dis->axis);
     mhash_print((mhash*)dis->methodHash, port);
 }
-
