@@ -1,7 +1,7 @@
 /*
  * dl_darwin.c - dlopen() interface for MacOS X/Darwin
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -90,11 +90,11 @@ static const char *dl_error(void)
     return dl_open_errstr;
 }
 
-static ScmDynLoadInitFn dl_sym(void *handle, const char *name)
+static ScmDynLoadEntry dl_sym(void *handle, const char *name)
 {
     NSSymbol sym = NSLookupSymbolInModule((NSModule)handle, name);
     if (sym == NULL) return NULL;
-    return (ScmDynLoadInitFn)NSAddressOfSymbol(sym);
+    return (ScmDynLoadEntry)NSAddressOfSymbol(sym);
 }
 
 static void dl_close(void *handle)
