@@ -1,7 +1,7 @@
 /*
  * dl_win.c - windows LoadLibrary interface
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -53,9 +53,9 @@ static const char *dl_error(void)
     return SCM_STRDUP(buf);
 }
 
-static ScmDynLoadInitFn dl_sym(void *handle, const char *name)
+static ScmDynLoadEntry dl_sym(void *handle, const char *name)
 {
-    return (ScmDynLoadInitFn)GetProcAddress((HMODULE)handle, name);
+    return (ScmDynLoadEntry)GetProcAddress((HMODULE)handle, name);
 }
 
 static void dl_close(void *handle)
