@@ -1,7 +1,7 @@
 /*
  * error.c - error handling
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 #include "gauche/exception.h"
 #include "gauche/vm.h"
 #include "gauche/priv/builtin-syms.h"
@@ -46,8 +46,10 @@ static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj message_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs);
+static ScmObj unberror_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs);
+static ScmObj conterror_allocate(ScmClass *klass, ScmObj initargs);
 static ScmObj compound_allocate(ScmClass *klass, ScmObj initargs);
 
 /* Setting up CPL is a bit tricky, since we have multiple
@@ -75,7 +77,7 @@ static ScmClass *condition_cpl[] = {
 };
 
 SCM_DEFINE_BASE_CLASS(Scm_ConditionClass, ScmInstance,
-                      NULL, NULL, NULL, 
+                      NULL, NULL, NULL,
                       condition_allocate, SCM_CLASS_DEFAULT_CPL);
 SCM_DEFINE_BASE_CLASS(Scm_MessageConditionClass, ScmMessageCondition,
                       Scm_MessageConditionPrint, NULL, NULL,
@@ -93,11 +95,11 @@ static ScmObj condition_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 }
 
 /* We expose this, for other condition subclasses may share this. */
-void Scm_MessageConditionPrint(ScmObj obj, ScmPort *port, 
+void Scm_MessageConditionPrint(ScmObj obj, ScmPort *port,
                                ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmClass *k = Scm_ClassOf(obj);
-    Scm_Printf(port, "#<%A \"%30.1A\">",
+    Scm_Printf(port, "#<%A \"%150.1A\">",
                Scm_ShortClassName(k),
                SCM_ERROR_MESSAGE(obj));
 }
@@ -195,7 +197,7 @@ static ScmClass *decoding_error_cpl[] = {
     ERROR_CPL,
     NULL
 };
-    
+
 static ScmClass *encoding_error_cpl[] = {
     SCM_CLASS_STATIC_PTR(Scm_IOWriteErrorClass),
     SCM_CLASS_STATIC_PTR(Scm_PortErrorClass),
@@ -213,6 +215,9 @@ SCM_DEFINE_BASE_CLASS(Scm_SystemErrorClass, ScmSystemError,
 SCM_DEFINE_BASE_CLASS(Scm_UnhandledSignalErrorClass, ScmUnhandledSignalError,
                       Scm_MessageConditionPrint, NULL, NULL,
                       sigerror_allocate, error_cpl);
+SCM_DEFINE_BASE_CLASS(Scm_UnboundVariableErrorClass, ScmUnboundVariableError,
+                      Scm_MessageConditionPrint, NULL, NULL,
+                      unberror_allocate, error_cpl);
 SCM_DEFINE_BASE_CLASS(Scm_ReadErrorClass, ScmReadError,
                       Scm_MessageConditionPrint, NULL, NULL,
                       readerror_allocate, error_cpl);
@@ -244,6 +249,10 @@ SCM_DEFINE_BASE_CLASS(Scm_IOInvalidPositionErrorClass, ScmIOInvalidPositionError
                       Scm_MessageConditionPrint, NULL, NULL,
                       porterror_allocate, porterror_cpl);
 
+SCM_DEFINE_BASE_CLASS(Scm_ContinuationErrorClass, ScmContinuationError,
+                      Scm_MessageConditionPrint, NULL, NULL,
+                      conterror_allocate, error_cpl);
+
 static ScmObj syserror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
     ScmSystemError *e = SCM_NEW_INSTANCE(ScmSystemError, klass);
@@ -261,6 +270,15 @@ static ScmObj sigerror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
     return SCM_OBJ(e);
 }
 
+static ScmObj unberror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
+{
+    ScmUnboundVariableError *e = SCM_NEW_INSTANCE(ScmUnboundVariableError,
+                                                  klass);
+    e->common.message = SCM_FALSE; /* set by initialize */
+    e->identifier = SCM_FALSE;     /* set by initialize */
+    return SCM_OBJ(e);
+}
+
 static ScmObj readerror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
 {
     ScmReadError *e = SCM_NEW_INSTANCE(ScmReadError, klass);
@@ -279,6 +297,14 @@ static ScmObj porterror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
     return SCM_OBJ(e);
 }
 
+static ScmObj conterror_allocate(ScmClass *klass, ScmObj initargs SCM_UNUSED)
+{
+    ScmContinuationError *e = SCM_NEW_INSTANCE(ScmContinuationError, klass);
+    e->common.message = SCM_FALSE; /* set by initialize */
+    e->promptTag = SCM_FALSE;      /* set by initialize */
+    return SCM_OBJ(e);
+}
+
 static ScmObj syserror_number_get(ScmSystemError *obj)
 {
     return SCM_MAKE_INT(obj->error_number);
@@ -305,6 +331,19 @@ static void sigerror_signal_set(ScmUnhandledSignalError *obj, ScmObj val)
     obj->signal = SCM_INT_VALUE(val);
 }
 
+static ScmObj unberror_identifier_get(ScmUnboundVariableError *obj)
+{
+    return obj->identifier;
+}
+
+static void unberror_identifier_set(ScmUnboundVariableError *obj, ScmObj val)
+{
+    if (!(SCM_SYMBOLP(val) || !SCM_IDENTIFIERP(val))) {
+        Scm_Error("symbol or identifier required, but got %S", val);
+    }
+    obj->identifier = val;
+}
+
 static ScmObj readerror_port_get(ScmReadError *obj)
 {
     if (obj->port) return SCM_OBJ(obj->port);
@@ -399,6 +438,17 @@ static void porterror_position_set(ScmPortError *obj, ScmObj val)
     porterror_auxinfo_setter(obj, sym_position, val);
 }
 
+static ScmObj conterror_prompt_tag_get(ScmContinuationError *obj)
+{
+    return obj->promptTag;
+}
+
+static void conterror_prompt_tag_set(ScmContinuationError *obj, ScmObj val)
+{
+    obj->promptTag = val;
+}
+
+
 static ScmClassStaticSlotSpec syserror_slots[] = {
     SCM_CLASS_SLOT_SPEC("errno", syserror_number_get, syserror_number_set),
     SCM_CLASS_SLOT_SPEC_END()
@@ -409,6 +459,11 @@ static ScmClassStaticSlotSpec sigerror_slots[] = {
     SCM_CLASS_SLOT_SPEC_END()
 };
 
+static ScmClassStaticSlotSpec unberror_slots[] = {
+    SCM_CLASS_SLOT_SPEC("identifier", unberror_identifier_get, unberror_identifier_set),
+    SCM_CLASS_SLOT_SPEC_END()
+};
+
 static ScmClassStaticSlotSpec readerror_slots[] = {
     SCM_CLASS_SLOT_SPEC("port", readerror_port_get, readerror_port_set),
     SCM_CLASS_SLOT_SPEC("line", readerror_line_get, readerror_line_set),
@@ -425,20 +480,27 @@ static ScmClassStaticSlotSpec porterror_slots[] = {
 
 static ScmClassStaticSlotSpec encodingerror_slots[] = {
     SCM_CLASS_SLOT_SPEC("port", porterror_port_get, porterror_port_set),
-    SCM_CLASS_SLOT_SPEC("offending-char", 
-                        porterror_offending_char_get, 
+    SCM_CLASS_SLOT_SPEC("offending-char",
+                        porterror_offending_char_get,
                         porterror_offending_char_set),
     SCM_CLASS_SLOT_SPEC_END()
 };
-    
+
 static ScmClassStaticSlotSpec invalidpositionerror_slots[] = {
     SCM_CLASS_SLOT_SPEC("port", porterror_port_get, porterror_port_set),
-    SCM_CLASS_SLOT_SPEC("position", 
-                        porterror_position_get, 
+    SCM_CLASS_SLOT_SPEC("position",
+                        porterror_position_get,
                         porterror_position_set),
     SCM_CLASS_SLOT_SPEC_END()
 };
-    
+
+static ScmClassStaticSlotSpec conterror_slots[] = {
+    SCM_CLASS_SLOT_SPEC("prompt-tag",
+                        conterror_prompt_tag_get,
+                        conterror_prompt_tag_set),
+    SCM_CLASS_SLOT_SPEC_END()
+};
+
 
 /*------------------------------------------------------------
  * Compound conditions
@@ -451,7 +513,7 @@ static ScmClass *compound_cpl[] = {
     NULL
 };
 
-static void compound_print(ScmObj obj, ScmPort *port, 
+static void compound_print(ScmObj obj, ScmPort *port,
                            ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmClass *k = Scm_ClassOf(obj);
@@ -599,6 +661,21 @@ ScmObj Scm_MakeSystemError(ScmObj message, int en)
     return SCM_OBJ(e);
 }
 
+ScmObj Scm_MakeUnboundVariableError(ScmObj identifier)
+{
+    static ScmString *prefix = NULL;
+    if (prefix == NULL) {
+        prefix = SCM_STRING(SCM_MAKE_STR_IMMUTABLE("unbound variable"));
+    }
+
+    ScmUnboundVariableError *e =
+        SCM_UNBOUND_VARIABLE_ERROR(unberror_allocate(SCM_CLASS_UNBOUND_VARIABLE_ERROR, SCM_NIL));
+    ScmObj msg = Scm_Sprintf("%A: %S", SCM_OBJ(prefix), identifier);
+    e->common.message = SCM_LIST3(msg, SCM_OBJ(prefix), identifier);
+    e->identifier = identifier;
+    return SCM_OBJ(e);
+}
+
 ScmObj Scm_MakeReadError(ScmObj message, ScmPort *port, int line)
 {
     ScmReadError *e =
@@ -699,7 +776,7 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
 /* Double fault check
  * In order to aviod infinite loop when error throwing routine
  * throws an error, we use vm flag SCM_ERROR_BEING_HANDLED to
- * check that.  Ideally a common single API should handle it,   
+ * check that.  Ideally a common single API should handle it,
  * but for the time being, we add the check at the beginning
  * of Scm_*Error APIs.
  * The SCM_ERROR_BEING_HANDLED flag is cleared in Scm_VMThrowException().
@@ -718,7 +795,7 @@ ScmObj Scm_ConditionTypeName(ScmObj c)
         SCM_VM_RUNTIME_FLAG_SET(vm, SCM_ERROR_BEING_HANDLED);           \
     } while (0)
 
-/* Common part to format error message passed as varargs.  
+/* Common part to format error message passed as varargs.
  * ostr must be declared as ScmObj, and msg must be the last
  * arg of function before '...'.
  */
@@ -798,7 +875,7 @@ static int get_errno(void)
     return errno;
 #else  /*GAUCHE_WINDOWS*/
     int en;
-    
+
     if (errno == 0) {
         en = -(int)GetLastError();
     } else {
@@ -988,7 +1065,7 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
     va_list ap;
 
     if (!SCM_CLASSP(condition_type)
-        || !Scm_SubtypeP(SCM_CLASS(condition_type), SCM_CLASS_CONDITION)) {
+        || !Scm_SubclassP(SCM_CLASS(condition_type), SCM_CLASS_CONDITION)) {
         /* If we don't get a condition type, fallback to a normal error. */
         condition_type = SCM_OBJ(SCM_CLASS_ERROR);
     }
@@ -1028,7 +1105,7 @@ ScmObj Scm_RaiseCondition(ScmObj condition_type, ...)
  *   format   - SCM_STACK_TRACE_FORMAT_* enum value.  No longer used.
  */
 void Scm_ShowStackTrace(ScmPort *out, ScmObj stacklite,
-                        int maxdepth, int skip, int offset, 
+                        int maxdepth, int skip, int offset,
                         int format SCM_UNUSED)
 {
     static ScmObj show_stack_trace = SCM_UNDEFINED;
@@ -1046,7 +1123,7 @@ void Scm_ShowStackTrace(ScmPort *out, ScmObj stacklite,
 void Scm_DumpStackTrace(ScmVM *vm, ScmPort *port)
 {
     if (vm == NULL) vm = Scm_VM();
-    if (port == NULL) port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    if (port == NULL) port = SCM_CURERR;
     ScmObj stack = Scm_VMGetStackLite(vm);
     ScmObj calls = Scm_VMGetCallTraceLite(vm);
     SCM_PUTZ("Stack Trace:\n", -1, port);
@@ -1073,12 +1150,17 @@ void Scm_DumpStackTrace(ScmVM *vm, ScmPort *port)
 */
 static void Scm_PrintDefaultErrorHeading(ScmObj e, ScmPort *out)
 {
-    if (Scm_InitializedP()) {
+    if (Scm_RuntimeState() >= SCM_RUNTIME_INITIALIZED) {
         static ScmObj print_default_error_heading = SCM_UNDEFINED;
         SCM_BIND_PROC(print_default_error_heading,
                       "print-default-error-heading",
                       Scm_GaucheModule());
         Scm_ApplyRec2(print_default_error_heading, e, SCM_OBJ(out));
+        static ScmObj print_additional_error_heading = SCM_UNDEFINED;
+        SCM_BIND_PROC(print_additional_error_heading,
+                      "print-additional-error-heading",
+                      Scm_GaucheModule());
+        Scm_ApplyRec2(print_additional_error_heading, e, SCM_OBJ(out));
     } else {
         /* Error during initialization. */
         if (SCM_CONDITIONP(e)) {
@@ -1102,11 +1184,11 @@ static void Scm_PrintDefaultErrorHeading(ScmObj e, ScmPort *out)
 ScmObj Scm_ReportError(ScmObj e, ScmObj out)
 {
     ScmVM *vm = Scm_VM();
-    ScmPort *port = SCM_VM_CURRENT_ERROR_PORT(vm);
+    ScmPort *port = SCM_CURERR;
     if (SCM_FALSEP(out)) {
         port = SCM_PORT(Scm_MakeOutputStringPort(TRUE));
     } else if (SCM_TRUEP(out)) {
-        port = SCM_VM_CURRENT_OUTPUT_PORT(vm);
+        port = SCM_CUROUT;
     } else if (SCM_OPORTP(out)) {
         port = SCM_PORT(out);
     }
@@ -1156,6 +1238,10 @@ void Scm__InitExceptions(void)
                                 "<unhandled-signal-error>",
                                 mod, cond_meta, SCM_FALSE,
                                 sigerror_slots, 0);
+    Scm_InitStaticClassWithMeta(SCM_CLASS_UNBOUND_VARIABLE_ERROR,
+                                "<unbound-variable-error>",
+                                mod, cond_meta, SCM_FALSE,
+                                unberror_slots, 0);
     Scm_InitStaticClassWithMeta(SCM_CLASS_READ_ERROR,
                                 "<read-error>",
                                 mod, cond_meta, SCM_FALSE,
@@ -1197,6 +1283,11 @@ void Scm__InitExceptions(void)
                                 mod, cond_meta, SCM_FALSE,
                                 invalidpositionerror_slots, 0);
 
+    Scm_InitStaticClassWithMeta(SCM_CLASS_CONTINUATION_ERROR,
+                                "<continuation-violation>",
+                                mod, cond_meta, SCM_FALSE,
+                                conterror_slots, 0);
+
     Scm_InitStaticClassWithMeta(SCM_CLASS_COMPOUND_CONDITION,
                                 "<compound-condition>",
                                 mod, cond_meta, SCM_FALSE,
