@@ -1,7 +1,7 @@
 /*
  * code.h - Virtual machine code
  *
- *   Copyright (c) 2005-2018  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2005-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -62,7 +62,7 @@ struct ScmCompiledCodeRec {
                                    its name.  Otherwise #f. */
     ScmObj debugInfo;           /* debug info, that associates instructions
                                    and source code / other metainfo.  May be
-                                   () if no info is avaialble. (*3) */
+                                   () if no info is available. (*3) */
     ScmObj signatureInfo;       /* signature info, a metainfo related to the
                                    interface of this closure.   Maybe #f
                                    if no info is available. (*4) */
@@ -73,7 +73,7 @@ struct ScmCompiledCodeRec {
                                    to inline this procedure.  Only set if
                                    the procedure is defined with define-inline.
                                    #f otherwise. (*5) */
-    void *builder;              /* An opaque data used during consturcting
+    void *builder;              /* An opaque data used during constructing
                                    the code vector.  Usually NULL. */
 };
 
@@ -89,6 +89,10 @@ struct ScmCompiledCodeRec {
  *       <offset> is either an instruction offset or 'definition (for the
  *       entire closure).
  *       At this moment, only used <info> is (source-info . <source>).
+ *       For the precompiled code, this slot may contain <packed-debug-info>
+ *       object at the startup.  Scheme slot accessor unpacks the object
+ *       and returns the unpacked info, so Scheme program will never see
+ *       <packed-debug-info> objects.
  *   *4) (<signature> <info> ...)
  *       <signature> is (<procedure-name> <formal> ...)
  *       <procedure-name> may be just a symbol, or a list (in case of internal
@@ -96,18 +100,13 @@ struct ScmCompiledCodeRec {
  *       for debug information.  See Scm_CompiledCodeFullName().
  *       <formal> ... is the formal argument list, as appears in the original
  *       lambda form, including :key, :optional etc.
- *       At this moment we don't use <info> ... yet; the plan is to put
- *       metainfo about closure interface, e.g. types.
+ *       <info> is a form of (<key> . <value>).  For now, we have 'type
+ *       as a key, and <value> contains symbolic type signature such as
+ *       '(<list> <integer> * -> *).  We may add more <key>s later.
  *   *5) This IForm is a direct result of Pass1, i.e. non-optimized form.
  *       Pass2 scans it when IForm is inlined into the caller site.
  */
 
-/* TRANSIENT: ***WARNING - NAMESPACE POLLUTION*** Certain versions of
-   0.9.5_pre1 precompiler emits C code that refers to the old name
-   ScmCompiledCode.argInfo.  This is a dirty workaround to allow compiling
-   such precompiled code.   Make sure to remove this after 0.9.5 release.  */
-#define argInfo  signatureInfo
-
 SCM_CLASS_DECL(Scm_CompiledCodeClass);
 #define SCM_CLASS_COMPILED_CODE   (&Scm_CompiledCodeClass)
 
@@ -123,7 +122,7 @@ SCM_CLASS_DECL(Scm_CompiledCodeClass);
     { { SCM_CLASS_STATIC_TAG(Scm_CompiledCodeClass) },   \
       (code), NULL, (codesize), 0, (maxstack),           \
       (reqargs), (optargs), (name), (debuginfo), (signatureinfo),   \
-      (parent), (iform) }
+      (parent), (iform), NULL /*builder*/ }
 
 SCM_EXTERN void   Scm_CompiledCodeCopyX(ScmCompiledCode *dest,
                                         const ScmCompiledCode *src);
@@ -132,6 +131,10 @@ SCM_EXTERN ScmObj Scm_CompiledCodeToList(ScmCompiledCode *cc);
 SCM_EXTERN ScmObj Scm_CompiledCodeFullName(ScmCompiledCode *cc);
 SCM_EXTERN void   Scm_VMExecuteToplevels(ScmCompiledCode *cv[]);
 
+/* This is called in the precomp-generated C code */
+SCM_EXTERN ScmObj Scm_MakePackedDebugInfo(ScmUVector*,
+                                          ScmObj(*)(void));
+
 /*
  * VM instructions
  */
@@ -162,11 +165,12 @@ enum ScmVMInsnFlag {
 /* Operand type */
 enum {
     SCM_VM_OPERAND_NONE,        /* take no operand */
-    SCM_VM_OPERAND_OBJ,         /* take ScmObj */
-    SCM_VM_OPERAND_CODE,        /* take ScmCompiledCode */
-    SCM_VM_OPERAND_CODES,       /* take a list of ScmCompiledCodes */
-    SCM_VM_OPERAND_ADDR,        /* take address of next code */
-    SCM_VM_OPERAND_OBJ_ADDR     /* take an object and address of next code */
+    SCM_VM_OPERAND_OBJ,         /* an ScmObj */
+    SCM_VM_OPERAND_CODE,        /* an ScmCompiledCode */
+    SCM_VM_OPERAND_CODES,       /* a list of ScmCompiledCodes */
+    SCM_VM_OPERAND_LABEL,       /* a label address of next code */
+    SCM_VM_OPERAND_OBJ_LABEL,   /* an object and label address of next code */
+    SCM_VM_OPERAND_OBJ_NATIVE   /* an object and native code address */
 };
 
 SCM_EXTERN const char *Scm_VMInsnName(u_int code);
