@@ -1,7 +1,7 @@
 /*
  * gauche/port.h - Port API
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -87,7 +87,8 @@ typedef struct ScmPortBufferRec {
     int  (*ready)(ScmPort *p);
     int  (*filenum)(ScmPort *p);
     off_t (*seeker)(ScmPort *p, off_t offset, int whence);
-    void *data;
+    void *data;         /* aux data to be used in handlers (heap allocated) */
+    long idata;         /* aux data to be used in handlers (single word) */
 
     /* The following fields are added in 0.9.10.  When you pass ScmPortBuffer
        to the constructor, those fields are looked at only when you specify
@@ -111,7 +112,8 @@ typedef struct ScmPortVTableRec {
     void    (*Flush)(ScmPort *p);
     void    (*Close)(ScmPort *p);
     off_t   (*Seek)(ScmPort *p, off_t off, int whence);
-    void    *data;
+    void    *data;      /* aux data to be used in handlers (heap allocated) */
+    long    idata;      /* aux data to be used in handlers (single word) */
 
     /* The following fields are added in 0.9.10.  When you pass ScmPortBuffer
        to the constructor, those fields are looked at only when you specify
@@ -135,7 +137,7 @@ typedef struct ScmPortInputStringRec {
    both getting and setting the position.  R6RS introduced
    port-position and set-port-position! interface, which then adopted as
    SRFI-192.  It requires different strategy than lseek interface.
-   
+
    1) SRFI-192 allows port position can be gotten but not be set, and that
       needs to be queried by port-has-set-port-position!?.  With lseek API,
       port position is either totally unavaiable or can be gotten/set.
@@ -149,8 +151,8 @@ typedef struct ScmPortInputStringRec {
    3) Lseek allows position to be set relative to the current position or
       to the end of the file.  SRFI-192 set-port-position! is always the
       absolute position.
-   
-  To support both interface, we have these hooks in SCM_PORT_FILE and 
+
+  To support both interface, we have these hooks in SCM_PORT_FILE and
   SCM_PORT_PROC type of ports (the field names differ for the historical
   reasons):
 
@@ -196,16 +198,16 @@ typedef struct ScmPortInputStringRec {
     u_int error     : 1;        /* Error has been occurred */           \
                                                                         \
     u_int flags     : 5         /* see ScmPortFlags below */
-    
+
 struct ScmPortRec {
     SCM_PORT_HEADER;
     /* See portP.h for the real definition.  The magic number is chosen
        so that ScmPort has enough size to contain the real struct.
-       
+
        We tested to use a pointer to the real struct, but I/O intensive
        benchmark showed 2-3% reduction of speed.  So we avoid indirection.
      */
-    char opaque[29*sizeof(ScmWord) + sizeof(ScmInternalFastlock)];
+    char opaque[30*sizeof(ScmWord) + 2*sizeof(ScmInternalFastlock)];
 };
 
 /* Port direction.
@@ -222,6 +224,14 @@ enum ScmPortDirection {
     SCM_PORT_OUTPUT_TRANSIENT = 4+SCM_PORT_OUTPUT
 };
 
+/*
+ * The 'mode' flags
+ *
+ *   Bit 0, 1, 2 - ScmPortBufferingMode
+ *   Bit 3-7 - Reserved
+ *   Bit 8 - SCM_PORT_BUFFER_SIGPIPE_SENSITIVE
+ */
+
 /* Port buffering mode */
 enum ScmPortBufferMode {
     SCM_PORT_BUFFER_FULL = 0,       /* full buffering */
@@ -300,7 +310,7 @@ enum ScmPortFlags {
 
 /* The following macros are for the backward compatibility; their use is
    deprecated.  (We used to have these macros directly access ScmPort
-   structure to eliminate function call overhead.  It turned out 
+   structure to eliminate function call overhead.  It turned out
    it matters little.   Use function APIs.) */
 #define SCM_PUTB(b, p)     Scm_Putb(b, SCM_PORT(p))
 #define SCM_PUTC(c, p)     Scm_Putc(c, SCM_PORT(p))
@@ -347,9 +357,13 @@ SCM_EXTERN int    Scm_PortPositionable(ScmPort *port, int setp);
 
 SCM_EXTERN void   Scm_FlushAllPorts(int exitting);
 
+SCM_EXTERN void   Scm_LinkPorts(ScmPort *iport, ScmPort *oport);
+SCM_EXTERN void   Scm_UnlinkPorts(ScmPort *port);
+
 SCM_EXTERN ScmObj Scm_PortName(ScmPort *port);
 SCM_EXTERN ScmSize Scm_PortLine(ScmPort *port);
 SCM_EXTERN ScmSize Scm_PortBytes(ScmPort *port);
+SCM_EXTERN ScmSize Scm_PortColumn(ScmPort *port);
 SCM_EXTERN ScmObj Scm_GetPortPosition(ScmPort *port);
 SCM_EXTERN ScmObj Scm_GetPortPositionUnsafe(ScmPort *port);
 SCM_EXTERN ScmObj Scm_SetPortPosition(ScmPort *port, ScmObj pos);
@@ -381,16 +395,9 @@ SCM_EXTERN ScmObj Scm_VMWithPortLocking(ScmPort *port,
 
 SCM_EXTERN ScmObj Scm_PortAttrGet(ScmPort *port, ScmObj key,
                                   ScmObj fallback);
-SCM_EXTERN ScmObj Scm_PortAttrGetUnsafe(ScmPort *port, ScmObj key,
-                                        ScmObj fallback);
 SCM_EXTERN ScmObj Scm_PortAttrSet(ScmPort *port, ScmObj key, ScmObj val);
-SCM_EXTERN ScmObj Scm_PortAttrSetUnsafe(ScmPort *port, ScmObj key, ScmObj val);
-SCM_EXTERN ScmObj Scm_PortAttrCreate(ScmPort *port, ScmObj key, ScmObj get, ScmObj set);
-SCM_EXTERN ScmObj Scm_PortAttrCreateUnsafe(ScmPort *port, ScmObj key, ScmObj get, ScmObj set);
 SCM_EXTERN ScmObj Scm_PortAttrDelete(ScmPort *port, ScmObj key);
-SCM_EXTERN ScmObj Scm_PortAttrDeleteUnsafe(ScmPort *port, ScmObj key);
 SCM_EXTERN ScmObj Scm_PortAttrs(ScmPort *port);
-SCM_EXTERN ScmObj Scm_PortAttrsUnsafe(ScmPort *port);
 
 SCM_EXTERN void   Scm_Putb(ScmByte b, ScmPort *port);
 SCM_EXTERN void   Scm_Putc(ScmChar c, ScmPort *port);
@@ -442,9 +449,13 @@ SCM_EXTERN ScmObj Scm_SetStdin(ScmPort *port);
 SCM_EXTERN ScmObj Scm_SetStdout(ScmPort *port);
 SCM_EXTERN ScmObj Scm_SetStderr(ScmPort *port);
 
-#define SCM_CURIN    SCM_VM_CURRENT_INPUT_PORT(Scm_VM())
-#define SCM_CUROUT   SCM_VM_CURRENT_OUTPUT_PORT(Scm_VM())
-#define SCM_CURERR   SCM_VM_CURRENT_ERROR_PORT(Scm_VM())
+SCM_EXTERN ScmPort *Scm_CurrentInputPort(void);
+SCM_EXTERN ScmPort *Scm_CurrentOutputPort(void);
+SCM_EXTERN ScmPort *Scm_CurrentErrorPort(void);
+
+#define SCM_CURIN    Scm_CurrentInputPort()
+#define SCM_CUROUT   Scm_CurrentOutputPort()
+#define SCM_CURERR   Scm_CurrentErrorPort()
 
 SCM_EXTERN ScmObj Scm_SetCurrentInputPort(ScmPort *port);
 SCM_EXTERN ScmObj Scm_SetCurrentOutputPort(ScmPort *port);
@@ -506,4 +517,3 @@ SCM_EXTERN ScmObj Scm_MakePortWithFd(ScmObj name,
 SCM_EXTERN ScmObj Scm_MakeCodingAwarePort(ScmPort *iport);
 
 #endif /*GAUCHE_PORT_H*/
-
