@@ -1,7 +1,7 @@
 /*
  * arith.h - macros for arithmetic
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -114,16 +114,16 @@
 
 #ifndef SADDOV
 /* Portable version */
-#define SADDOV(r, v, x, y)                      \
-  do {                                          \
-    (r) = (x) + (y);                            \
-    if ((x) >= 0) {                             \
-      if ((y) >= 0 && (r) < 0) (v) = 1;         \
-      else (v) = 0;                             \
-    } else {                                    \
-      if ((y) < 0 && (r) >= 0) (v) = -1;        \
-      else (v) = 0;                             \
-    }                                           \
+#define SADDOV(r, v, x, y)                                      \
+  do {                                                          \
+    (r) = (long)((unsigned long)(x) + (unsigned long)(y));      \
+    if ((x) >= 0) {                                             \
+      if ((y) >= 0 && (r) < 0) (v) = 1;                         \
+      else (v) = 0;                                             \
+    } else {                                                    \
+      if ((y) < 0 && (r) >= 0) (v) = -1;                        \
+      else (v) = 0;                                             \
+    }                                                           \
   } while (0)
 #endif /*SADDOV*/
 
