@@ -1,7 +1,7 @@
 /*
  * charP.h - Character-related API private header
  *
- *   Copyright (c) 2018-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2018-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -34,10 +34,6 @@
 #ifndef GAUCHE_PRIV_CHARP_H
 #define GAUCHE_PRIV_CHARP_H
 
-/* In char.c; called by gauche.charconv */
-SCM_EXTERN void Scm__InstallCharconvHooks(ScmChar (*u2c)(int),
-                                          int (*c2u)(ScmChar));
-
 /* Predefined charsets
    The enum value may change across versions, so we keep them private.
 
