@@ -1,7 +1,7 @@
 /*
  * portP.h - Port private API
  *
- *   Copyright (c) 2013-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2013-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -70,18 +70,23 @@ typedef struct ScmPortImplRec {
     ScmVM *lockOwner;           /* for port mutex; owner of the lock */
     int lockCount;              /* for port mutex; # of recursive locks */
 
+    ScmInternalFastlock attrlock; /* for port attribute access
+                                     port attributes are simple alist, so
+                                     we don't consider recursive lock. */
+
     ScmWriteState *writeState;  /* used internally */
 
-    /* Input counters.  these doesn't take account of ungetting and
+    /* Counters.  These doesn't take account of ungetting and
        seeking: Ungetting doesn't affect those counters (you can think
        that ungetting are handled above the counting layer).
        Seeking invalidates counters; if you seek, the values of the counters
        become bogus.
-       We don't have character counter, since it is difficult to track
-       (read-line uses byte read; see Scm_ReadLine in portapi.c).
+       Gauche doesn't distinguish textual and binary ports.  But once
+       you mix binary and textual I/O, line and column becomes unreliable.
      */
-    ScmSize line;               /* line counter */
-    ScmSize bytes;              /* byte counter */
+    ScmSize line;               /* line counter (input only).  1-base */
+    ScmSize bytes;              /* byte counter (input only) */
+    ScmSize column;             /* column tracker (output only).  0-base */
 
     /* The source or the sink of the port.   Use specialized accessor
        functions to retrieve one of those union members. */
@@ -94,7 +99,15 @@ typedef struct ScmPortImplRec {
 
     /* Port attibutes.  Use Scm_PortAttr* API to access. */
     ScmObj attrs;
-    
+
+    /* Linked port. An input port and an output port may be linked.  Whenever
+       a read is attempted from the iport, the linked oport is automatically
+       flushed and its column count is reset to zero.
+     */
+    ScmObj link;
+
+    /* Flags for private use.  See below. */
+    u_long internalFlags;
 } ScmPortImpl;
 
 #define P_(p)   ((ScmPortImpl*)(p))
@@ -122,13 +135,59 @@ void Scm__SetupPortsForWindows(int has_console);
 #define PORT_LOCK_OWNER_P(port, vm) \
     (P_(port)->lockOwner == (vm))
 
-/* Internal intreface to retrieve src member.
+/* Internal interface to retrieve src member.
    For public use, we have Scm_PortBufferStruct() etc. */
 #define PORT_BUF(port)     (&P_(port)->src.buf)
 #define PORT_ISTR(port)    (&P_(port)->src.istr)
 #define PORT_OSTR(port)    (&P_(port)->src.ostr)
 #define PORT_VT(port)      (&P_(port)->src.vt)
-    
+
+/* internalFlags */
+enum {
+    SCM_PORT_FLUSHED = (1L << 0), /* output only. TRUE when the port is flushed,
+                                     FALSE if anything is written after the
+                                     flush.*/
+    SCM_PORT_FILE_EXTDATA = (1L << 1), /* src.buf.data points to heap-allocated
+                                          data. */
+    SCM_PORT_PROC_EXTDATA = (1L << 2), /* src.vt.data points to heap-allocated
+                                          data. */
+};
+
+#define PORT_FLUSHED_P(port)  (P_(port)->internalFlags & SCM_PORT_FLUSHED)
+#define PORT_FLUSHED_SET(port) (P_(port)->internalFlags |= SCM_PORT_FLUSHED)
+#define PORT_FLUSHED_CLEAR(port) (P_(port)->internalFlags &= ~SCM_PORT_FLUSHED)
+
+/* PORT_FILE_EXTDATA and PORT_PROC_EXTDATA concern port finalizers.
+   By the time a port's finalizer is called, the object pointed by src.buf.data
+   or src.vt.data may already have been corrected (since the port itself
+   is already garbage.)  When port_finalize detects that situation,
+   it avoids calling flush or close methods of the portImpl->src structure,
+   since those methods are likely to rely on the data field and will no
+   longer work.
+
+   If the implementaiton of buffered or virtual port requires proper
+   cleanup if it is GC-ed before properly closed, it must implement
+   its own finalization.
+ */
+
+#define PORT_FILE_EXTDATA_P(port) \
+    (P_(port)->internalFlags & SCM_PORT_FILE_EXTDATA)
+#define PORT_FILE_EXTDATA_SET(port) \
+    (P_(port)->internalFlags |= SCM_PORT_FILE_EXTDATA)
+#define PORT_FILE_EXTDATA_CLEAR(port) \
+    (P_(port)->internalFlags &= ~SCM_PORT_FILE_EXTDATA)
+#define PORT_FILE_EXTDATA_GONE_P(port) \
+    (PORT_FILE_EXTDATA_P(port) && PORT_BUF(port)->data == NULL)
+
+#define PORT_PROC_EXTDATA_P(port) \
+    (P_(port)->internalFlags & SCM_PORT_PROC_EXTDATA)
+#define PORT_PROC_EXTDATA_SET(port) \
+    (P_(port)->internalFlags |= SCM_PORT_PROC_EXTDATA)
+#define PORT_PROC_EXTDATA_CLEAR(port) \
+    (P_(port)->internalFlags &= ~SCM_PORT_PROC_EXTDATA)
+#define PORT_PROC_EXTDATA_GONE_P(port) \
+    (PORT_PROC_EXTDATA_P(port) && PORT_VT(port)->data == NULL)
+
 /*================================================================
  * Locking the ports
  *
