@@ -1,7 +1,7 @@
 /*
  * prof.h - Profiler
  *
- *   Copyright (c) 2005-2018  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2005-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -117,7 +117,7 @@ struct ScmVMProfilerRec {
     HANDLE hObserverThread;     /* observer thread */
     HANDLE hTimerEvent;         /* sampling timer event */
     char *samplerFileName;      /* temporary file name to remove the file */
-#endif
+#endif /* GAUCHE_WINDOWS */
     ScmProfSample samples[SCM_PROF_SAMPLES_IN_BUFFER];
     ScmProfCount  counts[SCM_PROF_COUNTER_IN_BUFFER];
 };
@@ -131,7 +131,7 @@ SCM_EXTERN void Scm_ProfilerCountBufferFlush(ScmVM *vm);
 #ifdef GAUCHE_PROFILE
 #define SCM_PROF_COUNT_CALL(vm, obj)                                    \
     do {                                                                \
-        if (vm->profilerRunning) {                                      \
+        if (MOSTLY_FALSE(vm->profilerRunning)) {                        \
             if (vm->prof->currentCount == SCM_PROF_COUNTER_IN_BUFFER) { \
                 Scm_ProfilerCountBufferFlush(vm);                       \
             }                                                           \
