@@ -1,7 +1,7 @@
 /*
  * pthread.h - pthread primitives
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -34,15 +34,28 @@
 #ifndef GAUCHE_PTHREAD_H
 #define GAUCHE_PTHREAD_H
 
-/* Abstract thread implementation based on pthreads.
- * See gauche/uthread.h for the detailed semantics of those macros.
- */
+/* Abstract thread implementation based on pthreads. */
 
 #include <pthread.h>
 
-#define GAUCHE_HAS_THREADS 1
+/* ScmInternalThread - a handle to the system's thread.
 
-/* Thread */
+   The variable of type ScmInternalThread must be initialized by
+   SCM_INTERNAL_THREAD_INIT.  You can check if the variable is
+   initialized or not by SCM_INTERNAL_THREAD_INITIALIZED_P.
+
+   SCM_INTERNAL_THREAD_GETCURRENT() should return the current thread.
+   SCM_INTERNAL_THREAD_SETSPECIFIC(key, val) sets the value to the
+   thread-local storage identified by key.  It should evaluate
+   to TRUE on success, FALSE on error.
+
+   SCM_INTERNAL_THREAD_CLEANUP_{PUSH|POP} is an abstraction similar
+   to pthread_cleanup_{push|pop}.  We assume no Scheme error is
+   signalled between these macros.
+
+   SCM_INTERNAL_THREAD_PROC_RET{TYPE|VAL} specify return type and value
+   of the thread entry procedure.  See ext/threads/threads.c.
+ */
 typedef pthread_t ScmInternalThread;
 #define SCM_INTERNAL_THREAD_INIT(thr)   ((thr) = (pthread_t)NULL)
 #define SCM_INTERNAL_THREAD_INITIALIZED_P(thr)  ((thr) != (pthread_t)NULL)
@@ -73,7 +86,29 @@ typedef pthread_t ScmInternalThread;
 #  endif
 #endif /*defined(SIGRTMIN) && !defined(GAUCHE_PTHREAD_SIGNAL)*/
 
-/* Mutex */
+/* ScmInternalMutex - a mutex structure.
+
+   The variable of type ScmInternaMutex must be initialized by
+   SCM_INTERNAL_MUTEX_INIT at runtime.  If a static initializer is
+   needed, SCM_INTERNAL_MUTEX_INITIALIZER---but it's merely for making
+   the compiler happy, and does not guarantee the mutex is initialized.
+   You still need to use SCM_INTERNAL_MUTEX_INIT.   The resources
+   associated to the mutex would be released by SCM_INTERNAL_MUTEX_DESTROY.
+
+   To lock/unlock the mutex, SCM_INTERNAL_MUTEX_LOCK and
+   SCM_INTERNAL_MUTEX_UNLOCK can be used.   They should be used
+   with care, especially the thread should not be canceled between
+   them.
+
+   If you need to have a cancellable code in the critical
+   section, you should use SCM_INTERNAL_MUTEX_SAFE_LOCK_BEGIN and
+   SCM_INTERNAL_MUTEX_SAFE_LOCK_END.  They use cleanup mechanism
+   provided by the underlying system, such as pthread_cleanup_push/pop,
+   to make sure the mutex is unlocked when the thread terminates.
+   These macros must be used in the pair at the same syntactic scope,
+   for they can be expanded to opening block and closing block tokens,
+   respectively.
+ */
 typedef pthread_mutex_t ScmInternalMutex;
 #define SCM_INTERNAL_MUTEX_INIT(mutex)    pthread_mutex_init(&(mutex), NULL)
 #define SCM_INTERNAL_MUTEX_LOCK(mutex)    pthread_mutex_lock(&(mutex))
@@ -93,7 +128,42 @@ SCM_EXTERN void Scm__MutexCleanup(void *); /* in core.c */
     pthread_cleanup_push(Scm__MutexCleanup, &(mutex))
 #define SCM_INTERNAL_MUTEX_SAFE_LOCK_END() /*dummy*/; pthread_cleanup_pop(1)
 
-/* Condition variable */
+/* ScmInternalFastlock - fast lock
+ *
+ *  We use this type of lock when we know lock contention is rare.
+ *  If the system provides some sort of spinlock, it would be a
+ *  nice fit.  Otherwise the implementation can make these an alias
+ *  to SCM_INTERNAL_MUTEX_*.
+ */
+#ifdef HAVE_PTHREAD_SPINLOCK_T
+typedef pthread_spinlock_t ScmInternalFastlock;
+#define SCM_INTERNAL_FASTLOCK_INIT(spin) \
+    pthread_spin_init(&(spin), PTHREAD_PROCESS_PRIVATE)
+#define SCM_INTERNAL_FASTLOCK_LOCK(spin) \
+    pthread_spin_lock(&(spin))
+#define SCM_INTERNAL_FASTLOCK_UNLOCK(spin) \
+    pthread_spin_unlock(&(spin))
+#define SCM_INTERNAL_FASTLOCK_DESTROY(spin) \
+    pthread_spin_destroy(&(spin))
+#define SCM_INTERNAL_FASTLOCK_FINALIZATION_NOT_REQUIRED
+#else  /*!HAVE_PTHREAD_SPINLOCK_T*/
+/* We don't provide fast lock */
+typedef pthread_mutex_t ScmInternalFastlock;
+#define SCM_INTERNAL_FASTLOCK_INIT(fl)   SCM_INTERNAL_MUTEX_INIT(fl)
+#define SCM_INTERNAL_FASTLOCK_LOCK(fl)   SCM_INTERNAL_MUTEX_LOCK(fl)
+#define SCM_INTERNAL_FASTLOCK_UNLOCK(fl) SCM_INTERNAL_MUTEX_UNLOCK(fl)
+#define SCM_INTERNAL_FASTLOCK_DESTROY(fl) SCM_INTERNAL_MUTEX_DESTROY(fl)
+#endif /*!HAVE_PTHREAD_SPINLOCK_T*/
+
+/* ScmInternalCond - condition variable
+ *
+ *  Pthreads-style condition variables.  SCM_INTERNAL_COND_INITIALIZER
+ *  may be used for the placeholder of static variables (but the
+ *  variable still needs to be initialized by SCM_INTERNAL_COND_INIT.
+ *
+ *  SCM_INTERNAL_COND_TIMEDWAIT should return either 0 (signalled),
+ *  SCM_INTERNAL_COND_TIMEDOUT, or SCM_INTERNAL_COND_INTR.
+ */
 typedef pthread_cond_t ScmInternalCond;
 #define SCM_INTERNAL_COND_INIT(cond)        pthread_cond_init(&(cond), NULL)
 #define SCM_INTERNAL_COND_SIGNAL(cond)      pthread_cond_signal(&(cond))
@@ -106,25 +176,6 @@ typedef pthread_cond_t ScmInternalCond;
 #define SCM_INTERNAL_COND_TIMEDOUT          ETIMEDOUT
 #define SCM_INTERNAL_COND_INTR              EINTR
 
-/* Fast lock */
-#ifdef HAVE_PTHREAD_SPINLOCK_T
-typedef pthread_spinlock_t ScmInternalFastlock;
-#define SCM_INTERNAL_FASTLOCK_INIT(spin) \
-    pthread_spin_init(&(spin), PTHREAD_PROCESS_PRIVATE)
-#define SCM_INTERNAL_FASTLOCK_LOCK(spin) \
-    pthread_spin_lock(&(spin))
-#define SCM_INTERNAL_FASTLOCK_UNLOCK(spin) \
-    pthread_spin_unlock(&(spin))
-#define SCM_INTERNAL_FASTLOCK_DESTROY(spin) \
-    pthread_spin_destroy(&(spin))
-#else  /*!HAVE_PTHREAD_SPINLOCK_T*/
-typedef pthread_mutex_t ScmInternalFastlock;
-#define SCM_INTERNAL_FASTLOCK_INIT(fl)   SCM_INTERNAL_MUTEX_INIT(fl)
-#define SCM_INTERNAL_FASTLOCK_LOCK(fl)   SCM_INTERNAL_MUTEX_LOCK(fl)
-#define SCM_INTERNAL_FASTLOCK_UNLOCK(fl) SCM_INTERNAL_MUTEX_UNLOCK(fl)
-#define SCM_INTERNAL_FASTLOCK_DESTROY(fl) SCM_INTERNAL_MUTEX_DESTROY(fl)
-#endif /*!HAVE_PTHREAD_SPINLOCK_T*/
-
 /* Issues a full memory barrier */
 #define SCM_INTERNAL_SYNC()                 __sync_synchronize()
 
