@@ -1,7 +1,7 @@
 /*
  * reader.h - Reader API
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -114,6 +114,11 @@ SCM_EXTERN long Scm_ReadDigitsAsLong(ScmPort *port, ScmChar ch, int radix,
 SCM_EXTERN long Scm_ParseDigitsAsLong(const char *buf, ScmSize len, int radix,
                                       ScmSize *numread);
 
+/*
+ * Useful utility
+ */
+SCM_EXTERN int    Scm_CharWordConstituent(ScmChar ch, u_long flags);
+
 /*
  * SRFI-10 hash-comma syntax
  */
@@ -127,4 +132,3 @@ SCM_EXTERN ScmObj Scm_GetReaderCtor(ScmObj symbol, ScmObj module);
 SCM_EXTERN ScmObj Scm_DefineReaderDirective(ScmObj symbol, ScmObj proc);
 
 #endif  /*GAUCHE_READER_H*/
-
