@@ -1,7 +1,7 @@
 /*
  * system.h - Gauche system interface
  *
- *   Copyright (c) 2000-2018  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -73,9 +73,6 @@
 
 SCM_EXTERN int Scm_GetPortFd(ScmObj port_or_fd, int needfd);
 
-SCM_EXTERN off_t  Scm_IntegerToOffset(ScmObj i);
-SCM_EXTERN ScmObj Scm_OffsetToInteger(off_t o);
-
 /*==============================================================
  * Filesystem
  */
@@ -136,7 +133,7 @@ SCM_EXTERN int  Scm_ClockGetResMonotonic(u_long *sec, u_long *nsec);
 typedef struct ScmTimeRec {
     SCM_HEADER;
     ScmObj type;       /* 'time-utc by default.  see SRFI-19 */
-    ScmInt64 sec;      /* seconds */
+    int64_t sec;       /* seconds */
     long nsec;         /* nanoseconds */
 } ScmTime;
 
@@ -147,9 +144,9 @@ SCM_CLASS_DECL(Scm_TimeClass);
 
 SCM_EXTERN ScmObj Scm_CurrentTime(void);
 SCM_EXTERN ScmObj Scm_MakeTime(ScmObj type, long sec, long nsec);
-SCM_EXTERN ScmObj Scm_MakeTime64(ScmObj type, ScmInt64 sec, long nsec);
+SCM_EXTERN ScmObj Scm_MakeTime64(ScmObj type, int64_t sec, long nsec);
 SCM_EXTERN ScmObj Scm_IntSecondsToTime(long sec);
-SCM_EXTERN ScmObj Scm_Int64SecondsToTime(ScmInt64 sec);
+SCM_EXTERN ScmObj Scm_Int64SecondsToTime(int64_t sec);
 SCM_EXTERN ScmObj Scm_RealSecondsToTime(double sec);
 SCM_EXTERN ScmObj Scm_TimeToSeconds(ScmTime *t);
 
@@ -164,28 +161,33 @@ typedef struct ScmTimeSpecRec {
 } ScmTimeSpec;
 #endif /*!(defined(HAVE_STRUCT_TIMESPEC) && (!defined(GAUCHE_WINDOWS) || defined(__MINGW64_VERSION_MAJOR)))*/
 
+SCM_EXTERN ScmTimeSpec *Scm_ToTimeSpec(ScmObj t, ScmTime *t0, ScmTimeSpec *spec);
 SCM_EXTERN ScmTimeSpec *Scm_GetTimeSpec(ScmObj t, ScmTimeSpec *spec);
 
+/* strftime() compatibility layer */
+SCM_EXTERN ScmObj Scm_StrfTime(const char *format,
+                               const struct tm *tm,
+                               ScmObj reserved);
+
 /* sched_yield */
 SCM_EXTERN void   Scm_YieldCPU(void);
 
-/* struct tm */
-typedef struct ScmSysTmRec {
-    SCM_HEADER;
-    struct tm tm;
-} ScmSysTm;
-
-SCM_CLASS_DECL(Scm_SysTmClass);
-#define SCM_CLASS_SYS_TM      (&Scm_SysTmClass)
-#define SCM_SYS_TM(obj)       ((ScmSysTm*)(obj))
-#define SCM_SYS_TM_P(obj)     (SCM_XTYPEP(obj, SCM_CLASS_SYS_TM))
-#define SCM_SYS_TM_TM(obj)    SCM_SYS_TM(obj)->tm
-
-SCM_EXTERN ScmObj Scm_MakeSysTm(struct tm *);
-
 SCM_EXTERN int    Scm_NanoSleep(const ScmTimeSpec *req,
                                 ScmTimeSpec *rem);
 
+#if !HAVE_UTIMENSAT
+int utimensat(int, const char*, const ScmTimeSpec[2], int);
+#  if !defined(AT_FDCWD)
+#  define AT_FDCWD -1
+#  endif
+#  if !defined(UTIME_NOW)
+#  define UTIME_NOW -1
+#  endif
+#  if !defined(UTIME_OMIT)
+#  define UTIME_OMIT -1
+#  endif
+#endif /*!HAVE_UTIMENSAT*/
+
 /*==============================================================
  * Groups and users
  */
@@ -243,7 +245,7 @@ enum {
 
 SCM_EXTERN ScmObj Scm_SysExec(ScmString *file, ScmObj args,
                               ScmObj iomap, ScmSysSigset *mask,
-                              ScmString *dir, int flags);
+                              ScmString *dir, ScmObj env, u_long flags);
 SCM_EXTERN int   *Scm_SysPrepareFdMap(ScmObj iomap);
 SCM_EXTERN void   Scm_SysSwapFds(int *fds);
 
@@ -311,4 +313,3 @@ SCM_EXTERN ScmObj Scm_WinGetPipeName(HANDLE h);
 #endif /* GAUCHE_WINDOWS */
 
 #endif /* GAUCHE_SYSTEM_H */
-
