@@ -1,7 +1,7 @@
 /*
  * vm.h - Virtual machine
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -74,11 +74,13 @@ typedef struct ScmCompiledCodeRec ScmCompiledCode;
 /* Actual structure is defined in priv/vmP.h */
 typedef struct ScmCallTraceRec ScmCallTrace;
 
-/* Actual structure is defined in priv/parameterP.h */
-typedef struct ScmVMParameterTableRec ScmVMParameterTable;
+/* Actual structure is defined in priv/vmP.h */
+typedef struct ScmPromptTagRec ScmPromptTag;
+typedef struct ScmThreadLocalRec ScmThreadLocal;
+typedef struct ScmVMThreadLocalTableRec ScmVMThreadLocalTable;
 
 /*
- * Environment frame
+ * Static environment frame
  *
  *   :        :
  *   +--------+
@@ -104,6 +106,31 @@ typedef struct ScmEnvFrameRec {
 #define ENV_FP(env)        (((ScmObj*)(env))-((env)->size))
 #define ENV_DATA(env, num) (*(((ScmObj*)(env))-(num)-1))
 
+/*
+ * Dynamic environment
+ *
+ *  Dynamic environment is just an alist pointed from vm->denv.
+ *  It is a part of continuation---when a continuation frame is pushed,
+ *  the tip of denv is saved with it.  When a continuation frame is
+ *  popped, the saved denv is restored.
+ *  This means that you don't ever need to worry about popping denv.
+ */
+
+SCM_EXTERN void   Scm_VMPushDynamicEnv(ScmObj key, ScmObj val);
+SCM_EXTERN ScmObj Scm_VMFindDynamicEnv(ScmObj key, ScmObj fallback);
+
+typedef struct ScmContinuationMarkSetRec ScmContinuationMarkSet;
+
+SCM_CLASS_DECL(Scm_ContinuationMarkSetClass);
+#define SCM_CLASS_CONTINUATION_MARK_SET (&Scm_ContinuationMarkSetClass)
+#define SCM_CONTINUATION_MARK_SET(obj)  ((ScmContinuationMarkSet*)obj)
+#define SCM_CONTINUATION_MARK_SET_P(obj) SCM_ISA(obj,SCM_CLASS_CONTINUATION_MARK_SET)
+
+SCM_EXTERN ScmObj Scm_ContinuationMarks(ScmObj contProc, ScmObj promptTag);
+SCM_EXTERN ScmObj Scm_CurrentContinuationMarks(ScmObj promptTag);
+SCM_EXTERN ScmObj Scm_ContinuationMarkSetToList(const ScmContinuationMarkSet *,
+                                                ScmObj, ScmObj);
+
 /*
  * Continuation frame
  *
@@ -116,6 +143,7 @@ typedef struct ScmEnvFrameRec {
  *   |  cpc   |
  *   | marker |
  *   | size=N |
+ *   |  denv  |
  *   |  env   |
  *   |..prev..|<--- ScmContFrame* cont
  *   |arg[N-1]|
@@ -125,15 +153,61 @@ typedef struct ScmEnvFrameRec {
  *   +--------+
  *   :        :
  *
- *  If env is a special value (&ccEnvMark), this is a C continuation
- *  and pc contains a C function pointer.
+ *  If env is a special value (&ccEnvMark), this is a C continuation.
+ *  Some fields are used differently:
+ *
+ *   |  base   |
+ *   |   pc    |  <-- PCont procedure
+ *   |  cpc    |  <-- CCont procedure or NULL
+ *   | marker  |
+ *   | size=N  |
+ *   |  denv   |
+ *   |  env    |  <-- &ccEnvMark
+ *   |..prev.. |
+ *   |data[N-1]|
+ *   |data[N-2]|
+ *   :         :
+ *   | data[0] |
+ *   +---------+
+ *
+ *  The difference between CCont and PCont is purely historical.
+ *  CCont is the original C continuation.  It takes one result and
+ *  the data array.  PCont is introduced to support precompile-to-C
+ *  code, and it takes ScmVM* in addition to the result and data array.
+ *
+ *  When C continuation frame is popped, the value of cpc is temporaily
+ *  saved in vm->ccont, and PCont procedure is invoked with vm, val0,
+ *  and pointer to the data array.  If the C continuation frame is
+ *  pushed with Scm_pc_PushCC, vm->ccont is NULL (unused).
+ *  If the C continuation frame is pushed with Scm_VMPushCC, PCont
+ *  procedure is set to ccont_adapter() which invokes vm->ccont as
+ *  CCont procedure.
+ *
+ *  A special case is a C continuation frame pushed for execution of
+ *  body thunk of dynamic-wind.  It contains #<dynamic-handler>,
+ *  so that stack rewinder can find dynamic handlers.
+ */
+
+/* NB: The size of continuation frame (in words) is embedded in the precompiled
+ * compile.scm in order to calculate maximum stack size in procedure.
+ * Precompilation can be done on a different platform from the target,
+ * so the size of continuation frame in words must be kept the same across
+ * different platforms.  Hence the #if's in size and marker field.
+ * We could've done it by using full word for size and marker, but one more
+ * word in each continuation frame impacts performance.
  */
 
 typedef struct ScmContFrameRec {
     struct ScmContFrameRec *prev; /* previous frame */
     ScmEnvFrame *env;             /* saved environment */
+    ScmObj denv;                  /* dynamic environment links */
+#if SIZEOF_LONG == 4
+    long size : 30;               /* size of argument frame */
+    u_long marker : 2;            /* end marker of partial continuation */
+#else
     int size;                     /* size of argument frame */
     int marker;                   /* end marker of partial continuation */
+#endif
     SCM_PCTYPE cpc;               /* current PC (for debugging info) */
     SCM_PCTYPE pc;                /* next PC */
     ScmCompiledCode *base;        /* base register value */
@@ -147,28 +221,47 @@ SCM_EXTERN int  Scm__VMProtectStack(ScmVM *vm);
 SCM_EXTERN void Scm__VMUnprotectStack(ScmVM *vm);
 
 /*
- * Syntactic closure
- *
- *   Syntactic closure encapsulates compile-time environment for
- *   hygienic macro expansion.
- *   See Bawden & Rees, Syntactic Closures, MIT AI Memo 1049, June 1988.
+ * Thread Locals
+ *  See also priv/vmP.h for private definitions
  */
 
-typedef struct ScmSyntacticClosureRec {
-    ScmObj env;                 /* compile-time environment */
-    ScmObj literals;            /* literal symbols */
-    ScmObj expr;                /* expression */
-} ScmSyntacticClosure;
+SCM_CLASS_DECL(Scm_ThreadLocalClass);
+#define SCM_CLASS_THREAD_LOCAL  (&Scm_ThreadLocalClass)
+#define SCM_THREAD_LOCAL(obj)   ((ScmThreadLocal*)obj)
+#define SCM_THREAD_LOCAL_P(obj) SCM_ISA(obj,SCM_CLASS_THREAD_LOCAL)
 
-SCM_CLASS_DECL(Scm_SyntacticClosureClass);
-#define SCM_CLASS_SYNTACTIC_CLOSURE   (&Scm_SyntacticClosureClass)
+/* Flag value for Scm_MakeThreadLocal */
+enum {
+    /* Whether the initial value is inherited from the parent thread.
+       Thread locals are noninheritable by default, while parameters are
+       inheritable. */
+    SCM_THREAD_LOCAL_INHERITABLE = (1UL << 0)
 
-#define SCM_SYNTACTIC_CLOSURE(obj)   ((ScmSyntacticClosure*)(obj))
-#define SCM_SYNTACTIC_CLOSURE_P(obj) SCM_XTYPEP(obj, SCM_CLASS_SYNTACTIC_CLOSURE)
+};
 
-SCM_EXTERN ScmObj Scm_MakeSyntacticClosure(ScmObj env,
-                                           ScmObj literals,
-                                           ScmObj expr);
+SCM_EXTERN ScmThreadLocal *Scm_MakeThreadLocal(ScmObj name,
+                                               ScmObj initval,
+                                               u_long flags);
+SCM_EXTERN ScmObj Scm_ThreadLocalRef(ScmVM *vm,
+                                     const ScmThreadLocal *tl);
+SCM_EXTERN ScmObj Scm_ThreadLocalSet(ScmVM *vm,
+                                     const ScmThreadLocal *tl,
+                                     ScmObj val);
+
+
+/*
+ * Prompt tags
+ */
+
+SCM_CLASS_DECL(Scm_PromptTagClass);
+#define SCM_CLASS_PROMPT_TAG   (&Scm_PromptTagClass)
+#define SCM_PROMPT_TAG(obj)    ((ScmPromptTag*)obj)
+#define SCM_PROMPT_TAG_P(obj)  SCM_XTYPEP(obj, SCM_CLASS_PROMPT_TAG)
+
+SCM_EXTERN ScmObj Scm_MakePromptTag(ScmObj name);
+SCM_EXTERN ScmObj Scm_DefaultPromptTag();
+
+typedef struct ScmContinuationPromptRec ScmContinuationPrompt;
 
 /*
  * Identifier
@@ -192,13 +285,13 @@ SCM_CLASS_DECL(Scm_IdentifierClass);
 #define SCM_IDENTIFIERP(obj)    SCM_XTYPEP(obj, SCM_CLASS_IDENTIFIER)
 
 /* Create an identifier.
-   NAME can be a symbol or an identifier.  
+   NAME can be a symbol or an identifier.
    MOD is the toplevel module to close; can be NULL to use the current module.
    ENV is the local local environment (list of frames) to close. */
 SCM_EXTERN ScmObj Scm_MakeIdentifier(ScmObj name,
                                      ScmModule *mod,
                                      ScmObj env);
-/* Returns the minimal local environment (list of frames) where this 
+/* Returns the minimal local environment (list of frames) where this
    identifier is bound.  If the identifier isn't bound locally, it would be ().
    The returned value may not be the same as ENV argument passed to the
    constructor; we truncate irrelevant frames. */
@@ -211,6 +304,11 @@ SCM_EXTERN int    Scm_IdentifierBindingEqv(ScmIdentifier *id,
 SCM_EXTERN ScmIdentifier *Scm_OutermostIdentifier(ScmIdentifier *id);
 SCM_EXTERN ScmSymbol     *Scm_UnwrapIdentifier(ScmIdentifier *id);
 SCM_EXTERN ScmGloc       *Scm_IdentifierGlobalBinding(ScmIdentifier *id);
+SCM_EXTERN ScmObj         Scm_IdentifierGlobalRef(ScmIdentifier *id,
+                                                  ScmGloc **pgloc);
+SCM_EXTERN void           Scm_IdentifierGlobalSet(ScmIdentifier *id,
+                                                  ScmObj val,
+                                                  ScmGloc **pgloc);
 
 /*
  * Escape handling
@@ -313,6 +411,9 @@ typedef struct ScmVMStatRec {
 /* The profiler structure is defined in prof.h */
 typedef struct ScmVMProfilerRec ScmVMProfiler;
 
+/* Some experimental stuff (native.c) */
+typedef struct ScmCodeCacheRec ScmCodeCache;
+
 /*
  * VM structure
  *
@@ -353,14 +454,8 @@ struct ScmVMRec {
                                    is used only in compilation. */
     ScmCStack *cstack;          /* current escape point.  see the comment of
                                    "C stack rewinding" below. */
-    /* TRANSIENT: Change these flags to u_long in 1.0 */
-#if GAUCHE_API_VERSION < 1000    
-    unsigned int runtimeFlags;  /* Runtime flags */
-    unsigned int compilerFlags; /* Compiler flags */
-#else /*GAUCHE_API_VERSION >= 1000*/
-    u_long runtimeFlags;  /* Runtime flags */
-    u_long compilerFlags; /* Compiler flags */
-#endif /*GAUCHE_API_VERSION >= 1000*/
+    u_long runtimeFlags;        /* Runtime flags */
+    u_long compilerFlags;       /* Compiler flags */
     intptr_t attentionRequest;  /* Flag if VM needs to check signal queue,
                                    finalizer queue, or stop request.
                                    This flag can be turned on asynchronously.
@@ -380,26 +475,33 @@ struct ScmVMRec {
                                    ext/threads/threads.c, and turned off by
                                    process_queued_requests() in vm.c */
 
-    ScmPort *curin;             /* current input port */
-    ScmPort *curout;            /* current output port */
-    ScmPort *curerr;            /* current error port */
-    ScmVMParameterTable *parameters; /* parameter table */
+    ScmVMThreadLocalTable *threadLocals; /* thread local table */
 
     /* Registers */
     ScmCompiledCode *base;      /* Current executing closure's code packet. */
     SCM_PCTYPE pc;              /* Program pointer.  Points into the code
                                    vector. (base->code) */
     ScmEnvFrame *env;           /* Current environment.                      */
+    ScmObj denv;                /* Current dynamic environment. */
     ScmContFrame *cont;         /* Current continuation.                     */
     ScmObj *argp;               /* Current argument pointer.  Points
                                    to the incomplete environment frame
                                    being accumulated.  This is a part of
                                    continuation.                             */
+    void *ccont;                /* This is to support old C continuation
+                                   protocol with the new protool.  It is
+                                   set with CCont procedure right before
+                                   PCont procedure is called.  See the
+                                   comment on Scm_VMPushCC in vm.c.
+                                */
     ScmObj val0;                /* Value register.                           */
     ScmObj vals[SCM_VM_MAX_VALUES]; /* Value register for multiple values */
     int    numVals;             /* # of values */
+    int    trampoline;          /* Usually -1. C-compiled subr may set >=0
+                                   before returning to trampoline to another
+                                   subr. */
 
-    ScmObj handlers;            /* chain of active dynamic handlers          */
+    ScmObj dynamicHandlers;     /* chain of active dynamic handlers          */
 
     ScmObj *sp;                 /* stack pointer */
     ScmObj *stack;              /* bottom of allocated stack area */
@@ -423,19 +525,21 @@ struct ScmVMRec {
 #endif
 
     /* Escape handling */
-    ScmObj exceptionHandler;    /* the current exception handler installed by
-                                   with-exception-handler. */
-    ScmEscapePoint *escapePoint;/* chain of escape points (a kind of one-shot
-                                   continuation).  used by system's default
-                                   exception handler to escape from the error
-                                   handlers. */
-    ScmEscapePoint *escapePointFloating;
-                                /* reverse link of escape point chain
-                                   to keep 'active' EPs.
-                                   See ScmEscapePoint definition above. */
-    int escapeReason;           /* temporary storage to pass data across
-                                   longjmp(). */
-    void *escapeData[2];        /* ditto. */
+    ScmObj floatingEscapePoints; /* List of escapePoints that point to
+                                    in-stack continuation frames.  We only need
+                                    it so that we can adjust pointers to cont
+                                    frames when they are moved to the heap.
+                                    The list is cleared once cont frames
+                                    are moved to the heap.
+                                 */
+    int escapeReason;            /* temporary storage to pass data across
+                                    longjmp(). */
+    void *escapeData[2];         /* ditto. */
+    int errorHandlerContinuable; /* A transient flag, set by the error handler
+                                    as a result of 'guard' expansion to tell
+                                    that the control should return to 'raise'
+                                    after handler returns.  See %reraise
+                                    in libexc.scm as well. */
 
     /* Custom debugger or error reporter */
     ScmObj customErrorReporter; /* If set, Scm_ReportError (report-error) calls
@@ -468,8 +572,10 @@ struct ScmVMRec {
                                    Set by vm_register. */
 
     ScmCallTrace *callTrace;
+    ScmCodeCache *codeCache;
 
     /* for reset/shift */
+    ScmContinuationPrompt *currentPrompt;
     ScmObj resetChain;          /* list of reset information,
                                    where reset information is
                                    (delimited . <dynamic handlers chain>).
@@ -484,10 +590,16 @@ SCM_EXTERN int    Scm_AttachVM(ScmVM *vm);
 SCM_EXTERN void   Scm_DetachVM(ScmVM *vm);
 SCM_EXTERN void   Scm_VMDump(ScmVM *vm);
 SCM_EXTERN ScmObj Scm_VMDefaultExceptionHandler(ScmObj exc);
+SCM_EXTERN ScmObj Scm_VMCurrentExceptionHandler();
+SCM_EXTERN ScmObj Scm_VMExceptionHandlerStack();
+SCM_EXTERN void   Scm_VMPushExceptionHandler(ScmObj eh);
+SCM_EXTERN ScmObj Scm_VMPopExceptionHandler();
+SCM_EXTERN ScmObj Scm_VMWithExceptionHandler(ScmObj eh, ScmObj thunk);
 SCM_EXTERN ScmObj Scm_VMThrowException(ScmVM *vm, ScmObj exc, u_long flags);
 SCM_EXTERN ScmObj Scm_VMGetSourceInfo(ScmCompiledCode *code, SCM_PCTYPE pc);
 SCM_EXTERN ScmObj Scm_VMGetBindInfo(ScmCompiledCode *code, SCM_PCTYPE pc);
 SCM_EXTERN void   Scm_VMSetResult(ScmObj obj);
+SCM_EXTERN ScmVM *Scm_VMTakeSnapshot(ScmVM *master);
 
 SCM_CLASS_DECL(Scm_VMClass);
 #define SCM_CLASS_VM              (&Scm_VMClass)
@@ -589,6 +701,10 @@ SCM_EXTERN long Scm_VMUnwindProtect(ScmVM *vm, ScmCStack *cstack);
 SCM_EXTERN void Scm_VMNextHandler(ScmVM *vm);
 SCM_EXTERN void Scm_VMRewindProtect(ScmVM *vm);
 
+SCM_EXTERN ScmObj Scm_VMGetDynamicHandlers(void);
+SCM_EXTERN void   Scm_VMSetDynamicHandlers(ScmObj handlers);
+SCM_EXTERN void   Scm_VMFlushDynamicHandlers(void);
+
 /*
  * Runtime flags
  */
@@ -624,14 +740,13 @@ SCM_EXTERN void Scm_VMPushCC(ScmCContinuationProc *func,
                              void **data,
                              int datasize);
 
-#define SCM_CCONT_DATA_SIZE 6   /* Maximum datasize for VMPushCC */
-
 /*
  * Compiler flags
  */
 
 enum {
-    SCM_COMPILE_NOINLINE_GLOBALS = (1L<<0),/* Do not inline global procs */
+    SCM_COMPILE_NOINLINE_GLOBALS = (1L<<0),/* Do not inline global procs
+                                              (implies NOINLINE_INLINER, too) */
     SCM_COMPILE_NOINLINE_LOCALS = (1L<<1), /* Do not inline local procs */
     SCM_COMPILE_NOINLINE_CONSTS = (1L<<2), /* Do not inline constants */
     SCM_COMPILE_NOSOURCE = (1L<<3),        /* Do not insert source info */
@@ -649,8 +764,12 @@ enum {
     SCM_COMPILE_LEGACY_DEFINE = (1L<<11),  /* Do not insert toplevel binding
                                               at compile-time. */
     SCM_COMPILE_MUTABLE_LITERALS = (1L<<12),/* Literal pairs are mutable */
-    SCM_COMPILE_SRFI_FEATURE_ID = (1L<<13)  /* Allow srfi-N feature id in
+    SCM_COMPILE_SRFI_FEATURE_ID = (1L<<13), /* Allow srfi-N feature id in
                                                cond-expand */
+    SCM_COMPILE_NOINLINE_INLINER = (1L<<14)/* (internal) Do not invoke custom
+                                              inliner and ASM inliners.
+                                              hybrid macro is still expanded.
+                                              used for macroexpand-all */
 };
 
 #define SCM_VM_COMPILER_FLAG_IS_SET(vm, flag) ((vm)->compilerFlags & (flag))
@@ -672,6 +791,4 @@ SCM_EXTERN ScmObj Scm_CallProcedureInliner(ScmObj obj, ScmObj form, ScmObj env);
 /* This is in module.c, but it's not for public, so declaration is here. */
 SCM_EXTERN ScmModule* Scm_GaucheInternalModule(void);
 
-SCM_EXTERN ScmObj Scm__VMInsnOffsets(void);
-
 #endif /* GAUCHE_VM_H */
