@@ -1,7 +1,7 @@
 /*
  * hash.c - hash table implementation
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -34,7 +34,7 @@
 #define LIBGAUCHE_BODY
 #include <math.h>
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/atomicP.h"
 
 /*============================================================
@@ -238,7 +238,7 @@ static u_long internal_uvector_hash(ScmUVector *u, u_long salt, int portable)
         size_t s = SCM_UVECTOR_SIZE(u);
 
         switch (uvtype) {
-            /* We can use siphash if u is u8vector or s8vector. */ 
+            /* We can use siphash if u is u8vector or s8vector. */
         case SCM_UVECTOR_S8:
         case SCM_UVECTOR_U8:
             return Scm__DwSipPortableHash((uint8_t*)SCM_UVECTOR_ELEMENTS(u),
@@ -249,13 +249,13 @@ static u_long internal_uvector_hash(ScmUVector *u, u_long salt, int portable)
                for we need to allocate a buffer.
                (Unless we directly access siphash setup/round function
                and feed one word at a time--which is tedious.)
-               The current code may not be ideal, but just as good as our 
+               The current code may not be ideal, but just as good as our
                other primitive portable hash functions. */
 
             /* Initial hash value.  The seed value for each uvector type
                is just a random value I generated. */
 #define INIT_R(r, seed) SMALL_INT_HASH(r, seed^salt)
-                                                         
+
         case SCM_UVECTOR_S16:
             seed = 3499211612ul;
             goto do16;
@@ -297,7 +297,7 @@ static u_long internal_uvector_hash(ScmUVector *u, u_long salt, int portable)
             goto do64;
         case SCM_UVECTOR_U64:
             seed = 1323567403ul;
-            do64: 
+            do64:
             {
                 INIT_R(r, seed);
                 for (size_t i=0; i<s; i++) {
@@ -319,13 +319,14 @@ static u_long internal_uvector_hash(ScmUVector *u, u_long salt, int portable)
             }
             return r;
         }
-#undef INIT_R            
+#undef INIT_R
         case SCM_UVECTOR_RESERVED1:
         case SCM_UVECTOR_C32:
         case SCM_UVECTOR_C64:
         case SCM_UVECTOR_C128:
         case SCM_UVECTOR_RESERVED2:
         case SCM_UVECTOR_INVALID:
+        case SCM_UVECTOR_GENERIC:
             Scm_Panic("invalid uvector class.");
         }
         return 0;           /* dummy */
@@ -338,7 +339,7 @@ static u_long internal_uvector_hash(ScmUVector *u, u_long salt, int portable)
 
 /* equal-hash, which satisfies
      forall x, y: equal(x,y) => hash(x) = hash(y)
-  
+
    Both default-hash and portable-hash have this property but their
    requirements are slightly different, so here's the common part.
 */
@@ -446,17 +447,17 @@ static u_long equal_hash_common(ScmObj obj, u_long salt, int portable)
    and object-hash definitions in libomega.scm. */
 static ScmPrimitiveParameter *current_recursive_hash;
 
-ScmObj Scm_CurrentRecursiveHash(ScmObj newval)
+/* This must be called after symbols and modules are initialized.
+   See Scm__InitModule(). */
+void Scm__InitCurrentRecursiveHash()
 {
-    if (newval == SCM_UNBOUND) {
-        return Scm_PrimitiveParameterRef(Scm_VM(), current_recursive_hash);
-    } else {
-        return Scm_PrimitiveParameterSet(Scm_VM(), current_recursive_hash, newval);
-    }
+    SCM_DEFINE(Scm_GaucheInternalModule(),
+               "%current-recursive-hash",
+               Scm_MakePrimitiveParameterSubr(current_recursive_hash));
 }
 
 /* 'Portable' general hash function.
-   
+
    It is guaranteed that the hash value won't change for the same objects
    (roughly, indistinguishable in their external representation)
    across the runs of the program, and among different platforms.
@@ -484,7 +485,7 @@ ScmSmallInt Scm_RecursiveHash(ScmObj obj, ScmSmallInt salt, u_long flags)
     }
 }
 
-ScmSmallInt Scm_SmallIntHash(ScmSmallInt val, 
+ScmSmallInt Scm_SmallIntHash(ScmSmallInt val,
                              ScmSmallInt salt SCM_UNUSED,
                              u_long flags SCM_UNUSED)
 {
@@ -658,7 +659,7 @@ static u_long address_hash(const ScmHashCore *ht SCM_UNUSED, intptr_t obj)
     return hashval;
 }
 
-static int address_cmp(const ScmHashCore *ht SCM_UNUSED, 
+static int address_cmp(const ScmHashCore *ht SCM_UNUSED,
                        intptr_t key, intptr_t k2)
 {
     return (key == k2);
@@ -733,43 +734,6 @@ static int string_cmp(const ScmHashCore *ht SCM_UNUSED, intptr_t k1, intptr_t k2
                        SCM_STRING_BODY_SIZE(b1)) == 0));
 }
 
-/*
- * Accessor function for multiword raw hashtable.
- * Key points to an array of N words.
- */
-#if 0                           /* not used yet */
-static u_long multiword_hash(const ScmHashCore *table, intptr_t key)
-{
-    ScmWord keysize = (ScmWord)table->data;
-    ScmWord *keyarray = (ScmWord*)key;
-    u_long h = 0, h1;
-    for (int i=0; i<keysize; i++) {
-        ADDRESS_HASH(h1, keyarray[i]);
-        h = COMBINE(h, h1);
-    }
-    return h;
-}
-#endif
-
-#if 0
-static Entry *multiword_access(ScmHashCore *table, intptr_t k, ScmDictOp op)
-{
-    u_long hashval, index;
-    ScmWord keysize = (ScmWord)table->data;
-
-    hashval = multiword_hash(table, k);
-    index = HASH2INDEX(table->numBuckets, table->numBucketsLog2, hashval);
-    Entry **buckets = (Entry**)table->buckets;
-
-    for (Entry *e = buckets[index], *p = NULL; e; p = e, e = e->next) {
-        if (memcmp((void*)k, (void*)e->key, keysize*sizeof(ScmWord)) == 0)
-            FOUND(table, op, e, p, index);
-    }
-    NOTFOUND(table, op, k, hashval, index);
-}
-#endif
-
-
 /*
  * Accessor function for general case
  *    (hashfn and cmpfn are given by user)
@@ -1114,35 +1078,25 @@ ScmObj Scm_HashTableStat(ScmHashTable *table)
  * Printer
  */
 
-static void hash_print(ScmObj obj, ScmPort *port, 
+/* Note: When pprint is used, hash table priting is directly handled by pprint
+   and this is never called. */
+static void hash_print(ScmObj obj, ScmPort *port,
                        ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmHashTable *ht = (ScmHashTable*)obj;
     char *str = "";
 
     switch (ht->type) {
-    case SCM_HASH_EQ:      str = "eq?"; break;
-    case SCM_HASH_EQV:     str = "eqv?"; break;
-    case SCM_HASH_EQUAL:   str = "equal?"; break;
-    case SCM_HASH_STRING:  str = "string=?"; break;
+    case SCM_HASH_EQ:      str = "eq"; break;
+    case SCM_HASH_EQV:     str = "eqv"; break;
+    case SCM_HASH_EQUAL:   str = "equal"; break;
+    case SCM_HASH_STRING:  str = "string"; break;
     case SCM_HASH_GENERAL: str = "general"; break;
     default: Scm_Panic("something wrong with a hash table");
     }
 
-#if 0
-    /* Use read-time constructor so that table can be read back
-       --- is it necessary?  I'm not sure yet. */
-    Scm_Printf(port, "#,(<hash-table> %s", str);
-    if (ht->numEntries > 0) {
-        Scm_HashIterInit(&iter, ht);
-        while ((e = Scm_HashIterNext(&iter)) != NULL) {
-            Scm_Printf(port, " %S %S", e->key, e->value);
-        }
-    }
-    SCM_PUTZ(")", -1, port);
-#else
-    Scm_Printf(port, "#<hash-table %s %p>", str, ht);
-#endif
+    Scm_Printf(port, "#<hash-table %s[%d] @%p>", str,
+               SCM_HASH_TABLE_CORE(ht)->numEntries, ht);
 }
 
 static unsigned int round2up(unsigned int val)
@@ -1168,15 +1122,15 @@ void Scm__InitHash()
     u_long salt = ((u_long)getpid() * ((u_long)t.tv_sec^(u_long)t.tv_usec));
     ADDRESS_HASH(salt, salt);
     salt &= SCM_SMALL_INT_MAX;
-    /* 
+    /*
      * We can't use Scm_BindPrimitiveParameter here, since symbol table
      * is not initialized yet (symbol table uses hashtable!)
      */
-    hash_salt = 
-        Scm_MakePrimitiveParameter(SCM_CLASS_PRIMITIVE_PARAMETER, SCM_FALSE, 
+    hash_salt =
+        Scm_MakePrimitiveParameter(SCM_CLASS_PRIMITIVE_PARAMETER, SCM_FALSE,
                                    Scm_MakeIntegerU(salt), 0);
-    current_recursive_hash = 
-        Scm_MakePrimitiveParameter(SCM_CLASS_PRIMITIVE_PARAMETER, SCM_FALSE, 
+    current_recursive_hash =
+        Scm_MakePrimitiveParameter(SCM_CLASS_PRIMITIVE_PARAMETER, SCM_FALSE,
                                    SCM_FALSE, 0);
 }
 
@@ -1237,15 +1191,6 @@ ScmHashEntry *Scm__HashIterNextCompat(ScmHashIter *iter)
     return (ScmHashEntry*)e;
 }
 
-#if 0
-ScmObj Scm_MakeHashTableMultiWord(int keysize, int initsize)
-{
-    return make_hash_table(SCM_CLASS_HASH_TABLE, SCM_HASH_MULTIWORD,
-                           multiword_access, multiword_hash,
-                           NULL, initsize, (void*)SCM_WORD(keysize));
-}
-#endif
-
 /* Legacy constructor.  DEPRECATED.  Will go away soon. */
 ScmObj Scm_MakeHashTable(ScmHashProc *hashfn,
                          ScmHashCompareProc *cmpfn SCM_UNUSED,
@@ -1260,14 +1205,7 @@ ScmObj Scm_MakeHashTable(ScmHashProc *hashfn,
     } else if (hashfn == (ScmHashProc*)SCM_HASH_STRING) {
         return Scm_MakeHashTableSimple(SCM_HASH_STRING, initSize);
     }
-#if 0
-    else {
-        return Scm_MakeHashTableFull(SCM_CLASS_HASH_TABLE, SCM_HASH_GENERAL,
-                                     hashfn, cmpfn, initSize, NULL);
-    }
-#else
     return SCM_UNDEFINED;
-#endif
 }
 #endif /*GAUCHE_API_VERSION < 1000*/
 
@@ -1392,13 +1330,13 @@ static u_long legacy_flonum_hash(double f)
         two_pow_63 = ldexp(1.0, 63);
         minus_two_pow_63 = -ldexp(1.0, 63);
         two_pow_32 = ldexp(1.0, 32);
-        AO_nop_full();
+        Scm_AtomicThreadFence();
         initialized = TRUE;
     }
     /* This condition eliminates NaN as well. */
     if (!(minus_two_pow_63 < d && d < two_pow_63)) return 0;
     if (-0.5 < d && d < 0.5) return 0;
-                                 
+
     double dm = trunc(fmod(d, two_pow_32));
     if (dm < 0) dm += two_pow_32;
     return (u_long)trunc(dm);
