@@ -1,7 +1,7 @@
 ;;;
 ;;; Interface to VM code generator (code.c)
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -39,7 +39,7 @@
 (define-module gauche.vm.code
   (export vm-dump-code vm-code->list vm-insn-build
           vm-insn-code->name vm-insn-name->code
-          
+
           make-compiled-code-builder
           compiled-code-emit0! compiled-code-emit0o!
           compiled-code-emit0i! compiled-code-emit0oi!
@@ -59,11 +59,10 @@
 ;;
 
 (inline-stub
- (declcode
-  (.include <gauche/code.h>
-            <gauche/priv/codeP.h>
-            <gauche/class.h>
-            <gauche/vminsn.h>))
+ (.include "gauche/priv/configP.h"
+           "gauche/code.h"
+           "gauche/priv/codeP.h"
+           "gauche/vminsn.h")
 
  (define-cproc vm-dump-code (code::<compiled-code>) ::<void>
    Scm_CompiledCodeDump)
@@ -173,8 +172,7 @@
 (define (%original-source src) ; internal
   (and (pair? src)
        (let loop ([src src])
-         (if-let1 orig ((with-module gauche.internal pair-attribute-get)
-                        src 'original #f)
+         (if-let1 orig (pair-attribute-get src 'original #f)
            (loop orig)
            src))))
 
@@ -185,42 +183,82 @@
     ((with-module gauche.vm.code %original-source)
      (assq-ref def 'source-info))))
 
+(define (compiled-code-type code)
+  (and-let* ([sig (~ code'signature-info)]
+             [ (pair? sig) ])
+    (alist-ref (cdr sig) 'type)))
+
+(define-cproc %decode-packed-debug-info (code-vector::<u8vector>
+                                         const-vector::<vector>)
+  (return (Scm_DecodePackedDebugInfo (SCM_U8VECTOR_ELEMENTS code-vector)
+                                     (SCM_U8VECTOR_SIZE code-vector)
+                                     const-vector)))
+
 ;; Called from pass5 of the compiler.  SRC is $lambda-src of the iform
 ;; (could be #f)  Does two things:
 ;;  - Push the entire source into debug-info
 ;;  - Create extended pair of signature, with source location info attached
 ;;    as 'source-info pair attribute, and store it into signature-info.
-(define (compiled-code-attach-source-info! code src)
+(define (compiled-code-attach-source-info! code src type)
   (if src
     (let ([orig ((with-module gauche.vm.code %original-source) src)]
-          [sig (extended-cons (~ code'full-name)
+          [sig (extended-cons (slot-ref code'full-name)
                               (pair-attribute-get src 'arg-info '_))])
       (if-let1 si (pair-attribute-get orig 'source-info #f)
         (pair-attribute-set! sig 'source-info si))
       (compiled-code-push-info! code `(definition (source-info . ,src)))
-      (slot-set! code 'signature-info (list sig)))
-    (slot-set! code 'signature-info (list (cons (~ code'full-name) '_)))))
+      (slot-set! code 'signature-info
+                 (cons sig
+                       (cond-list [type `(type . ,type)]))))
+    (slot-set! code 'signature-info (list (cons (slot-ref code'full-name) '_)))))
 
 (select-module gauche)
 (inline-stub
-  ;; <compiled-code> class is visible 
+  ;; <compiled-code> class is visible
  (define-cclass <compiled-code>
    "ScmCompiledCode*" "Scm_CompiledCodeClass"
    (c "SCM_CLASS_DEFAULT_CPL")
    ((parent :setter #f)
     (signature-info :c-name "signatureInfo")
-    (debug-info :c-name "debugInfo" :setter #f)
+    (debug-info :c-spec "Scm_CodeDebugInfo(obj)" :setter #f)
     (required-args :type <fixnum> :c-name "requiredArgs" :setter #f)
     (optional-args :type <fixnum> :c-name "optionalArgs" :setter #f)
     (name :setter #f)
     (full-name :c-spec "Scm_CompiledCodeFullName(obj)" :setter #f)
     (size :type <fixnum> :c-name "codeSize" :setter #f)
     (max-stack :type <fixnum> :c-name "maxstack" :setter #f)
-    (intermediate-form :c-name "intermediateForm" :setter #f)
-    ;; TRANSIENT: For the backward compatibility - will be gone soon
-    (info :c-name "debugInfo" :setter #f)
-    (arg-info :c-name "signatureInfo" :setter #f))
+    (intermediate-form :c-name "intermediateForm" :setter #f))
    (printer (Scm_Printf port "#<compiled-code %S@%p>"
                         (Scm_CompiledCodeFullName (SCM_COMPILED_CODE obj))
                         obj)))
  )
+
+(select-module gauche.internal)
+(inline-stub
+
+ (define-cfn get_packed_vector (s::ScmPackedDebugInfo*) :static
+   ;; This is called only during precomp, or once during decoding,
+   ;; so we don't need to worry about allocation.
+   (return (Scm_MakeU8VectorFromArray (-> s codeSize) (-> s codeVector))))
+
+  ;; <packed-debug-info> is internal
+ (define-cclass <packed-debug-info>
+   "ScmPackedDebugInfo*" "Scm_PackedDebugInfoClass"
+   (c "SCM_CLASS_DEFAULT_CPL")
+   ((packed-vector :setter #f
+                   :getter (c "get_packed_vector"))
+    (const-vector  :setter #f
+                   :c-name "constVector"))
+   )
+
+ (define-cproc decode-packed-debug-info (p::<packed-debug-info>)
+   (when (SCM_FALSEP (-> p decoded))
+     (let* ([decode-debug-info SCM_UNDEFINED]
+            [_ (SCM_BIND_PROC decode-debug-info "decode-debug-info"
+                              (Scm_GaucheInternalModule))]
+            [decoded (Scm_ApplyRec2 decode-debug-info
+                                    (get_packed_vector p)
+                                    (-> p constVector))])
+       (set! (-> p decoded) decoded)))
+   (return (-> p decoded)))
+ )
