@@ -1,7 +1,7 @@
 ;;;
 ;;; libeval.scm - eval, load and related stuff
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -32,11 +32,13 @@
 ;;;
 
 (select-module gauche.internal)
+(use util.match)
 
 (inline-stub
- (declcode (.include <gauche/vminsn.h>
-                     <gauche/class.h>
-                     <gauche/priv/readerP.h>)))
+ (.include "gauche/priv/configP.h"
+           "gauche/vminsn.h"
+           "gauche/priv/readerP.h"
+           "gauche/priv/vmP.h"))
 
 (declare (keep-private-macro autoload add-load-path))
 
@@ -79,9 +81,12 @@
     (let* ([path (car r)]
            [remaining-paths (cadr r)]
            [hooked? (pair? (cddr r))]
-           [opener (if hooked? (caddr r) open-input-file)]
+           [opener (if hooked?
+                     (caddr r)
+                     (cut open-input-file <>
+                          :encoding (gauche-character-encoding)))]
            [port (guard (e [else e]) (opener path))])
-      (when main-script 
+      (when main-script
         ;; record full path of the script
         (script-file (sys-normalize-pathname path
                                              :absolute #t :canonicalize #t)))
@@ -92,11 +97,13 @@
                 (if hooked? " (hooked) " "")))
       (if (not (input-port? port))
         (and error-if-not-found (raise port))
-        (load-from-port (if ignore-coding
-                          port
-                          (open-coding-aware-port port))
-                        :environment environment
-                        :paths remaining-paths)))))
+        (begin
+          (load-from-port (if ignore-coding
+                            port
+                            (open-coding-aware-port port))
+                          :environment environment
+                          :paths remaining-paths)
+          path)))))
 
 
 (select-module gauche.internal)
@@ -152,39 +159,40 @@
       (begin0 (read port)
         (current-read-context prev-read-context)))
 
-    (guard (e [else (let1 e2 (if (condition? e)
-                               ($ make-compound-condition e
-                                  $ make <load-condition-mixin>
-                                  :history (current-load-history)
-                                  :port (current-load-port)
-                                  :expr #f)
-                               e)
-                      (restore-load-context)
-                      (raise e2))])
-      (setup-load-context)
-      ;; Discard BOM
-      (when (eq? (gauche-character-encoding) 'utf-8)
-        (when (eqv? (peek-char port) #\ufeff)
-          (read-char port)))
-      (do ([s (read+ port) (read+ port)])
-          [(eof-object? s)]
-        (eval s #f)))
+    (with-exception-handler
+     (^e (let1 e2 (if (condition? e)
+                    ($ make-compound-condition e
+                       $ make <load-condition-mixin>
+                       :history (current-load-history)
+                       :port (current-load-port)
+                       :expr #f)
+                    e)
+           (restore-load-context)
+           (raise e2)))
+     (^[]
+       (setup-load-context)
+       ;; Discard BOM
+       (when (eqv? (peek-char port) #\ufeff)
+         (read-char port))
+       (do ([s (read+ port) (read+ port)])
+           [(eof-object? s)]
+         (eval s #f))))
     (restore-load-context)
     #t))
 
 ;; A few helper procedures
 (define-cproc %record-load-stat (path) ::<void>
-  (.if "defined(HAVE_GETTIMEOFDAY)"
-       (let* ([vm::ScmVM* (Scm_VM)])
-         (when (SCM_VM_RUNTIME_FLAG_IS_SET vm SCM_COLLECT_LOAD_STATS)
-           (let* ([t0::(struct timeval)])
-             (gettimeofday (& t0) NULL)
-             (let* ([t (Scm_MakeIntegerU (+ (* (ref t0 tv_sec) 1000000)
-                                            (ref t0 tv_usec)))])
-               (set! (ref (-> vm stat) loadStat)
-                     (Scm_Cons
-                      (?: (SCM_FALSEP path) t (Scm_Cons path t))
-                      (ref (-> vm stat) loadStat)))))))))
+  (.when "defined(HAVE_GETTIMEOFDAY)"
+    (let* ([vm::ScmVM* (Scm_VM)])
+      (when (SCM_VM_RUNTIME_FLAG_IS_SET vm SCM_COLLECT_LOAD_STATS)
+        (let* ([t0::(struct timeval)]
+               [_ (gettimeofday (& t0) NULL)]
+               [t (Scm_MakeIntegerU (+ (* (ref t0 tv_sec) 1000000)
+                                       (ref t0 tv_usec)))])
+          (set! (ref (-> vm stat) loadStat)
+                (Scm_Cons
+                 (?: (SCM_FALSEP path) t (Scm_Cons path t))
+                 (ref (-> vm stat) loadStat))))))))
 
 (define-cproc %new-read-context-for-load ()
   (let* ([ctx::ScmReadContext* (Scm_MakeReadContext NULL)])
@@ -207,7 +215,7 @@
     (let* ([base (sys-basename dsopath)]
            [stem (or (string-scan-right base #\. 'before) base)])
       (string-append "_Scm_Init_"
-                     ;; string-map is in srfi-13, so we roll our own.
+                     ;; string-map is in SRFI-13, so we roll our own.
                      ;; This looks awful, but we don't call this frequently.
                      (list->string
                       (map (^c (cond [(char-alphabetic? c) (char-downcase c)]
@@ -219,12 +227,18 @@
 (select-module gauche)
 
 ;; API
+;; returns #<dlobj>
 (define-cproc dynamic-load (file::<string>
-                            :key (init-function #f)
+                            :key (init-function #t)
                             (export-symbols #f)); for backward compatibility
   (cast void export-symbols) ; suppress unused var warning
   (return (Scm_DynLoad file init_function 0)))
 
+;; API (experimental)
+;; returns #<dlptr>
+(define-cproc dlobj-get-entry-address (dlo::<dlobj> name::<string>)
+  Scm_DLOGetEntryAddress)
+
 ;; API
 (define-cproc provide (feature)   Scm_Provide)
 ;; API
@@ -422,8 +436,13 @@
 ;;; Repl
 ;;;
 
+;; Note: `gauche.interactive` also defines `read-eval-print-loop` which is
+;; capable of input editing.  `Gosh` calls either one depending on the
+;; availability of capable terminal.
+
 (select-module gauche.internal)
 
+;; Fallback
 (define (%repl-print . vals) (for-each (^e (write e) (newline)) vals))
 (define (%repl-prompt) (display "gosh> ") (flush))
 
@@ -464,6 +483,8 @@
 ;;;
 ;;; Invoke other version of gosh
 ;;;   This is called when gosh gets -vVERSION option.
+;;;   On success, this never returns.
+;;;   On failure, returns a globpath that is tried to find the other versions.
 ;;;
 (select-module gauche.internal)
 (define (%invoke-other-version version args)
@@ -477,8 +498,14 @@
   (let* ([prefix (sys-normalize-pathname
                   (string-append (gauche-architecture-directory) "/../../..")
                   :canonicalize #t)]
+         [windows? (or (assq 'gauche.os.windows (cond-features))
+                       (assq 'gauche.os.cygwin (cond-features)))]
          [globpath (string-append prefix "/gauche-*/" version "/"
-                                  (gauche-architecture) "/gosh")]
+                                  (gauche-architecture) "/"
+                                  (if windows?
+                                   ;; glob requires a file extension on Windows
+                                    "gosh.exe"
+                                    "gosh"))]
          [goshes (glob globpath)]
          ;; Remove -v option from args
          [args (let loop ([args (cdr args)]
@@ -487,23 +514,30 @@
                        [(equal? (car args) "-v") (loop (cddr args) r)]
                        [(#/^-v/ (car args)) (loop (cdr args) r)]
                        [else (loop (cdr args) (cons (car args) r))]))])
-    (unless (pair? goshes)
-      (exit 1 "No installed Gauche with version ~a under ~a.\n"
-            version prefix))
-    (sys-exec (car goshes) (cons (car goshes) args))))
+    (if (pair? goshes)
+      (if windows?
+        ;; On windows, sys-exec (execvp) doesn't work like unix.
+        ;; Notably, exit status of the exec'ed process is lost.
+        ;; So we use sys-fork-and-exec (CreateProcess) and wait for the
+        ;; child process to exit.
+        (let1 pid (sys-fork-and-exec (car goshes) (cons (car goshes) args))
+          (receive (_ status) (sys-waitpid pid)
+            (cond
+             [(sys-wait-exited? status)
+              (exit (sys-wait-exit-status status))] ; return exit code
+             [(sys-wait-signaled? status)
+              (sys-kill (sys-getpid) (sis-wait-termsig status))]
+             [else (exit 70)])))        ;EX_SOFTWARE
+        (sys-exec (car goshes) (cons (car goshes) args))) ;never return
+      prefix)))
 
 ;;;
 ;;; System termination
 ;;;
 
-(select-module gauche)
+(select-module gauche.internal)
 (define-cproc %exit (:optional (code 0)) ::<void>
-  (let* ([status::int 0])
-    (cond
-     [(SCM_EQ code SCM_TRUE)]  ; status == 0
-     [(SCM_INTP code) (set! status (SCM_INT_VALUE code))]
-     [else (set! status 70)])  ; EX_SOFTWARE
-    (Scm_Exit status)))
+  (Scm_Exit (Scm_ObjToExitCode code)))
 
 ;; API
 ;; exit handler.
@@ -569,25 +603,7 @@
 (define-cproc %gauche-runtime-directory () Scm_RuntimeDirectory)
 (define-cproc %gauche-libgauche-path () Scm_LibgauchePath)
 (define-cproc %gauche-executable-path () Scm_ExecutablePath)
-
-;; srfi-176
-(define-in-module gauche (version-alist)
-  (let1 fs ((with-module gauche.internal cond-features))
-    `((version ,(gauche-version))
-      (command "gosh")
-      (scheme.id gauche)
-      (languages scheme r5rs r7rs)
-      (encodings ,(gauche-character-encoding))
-      (website "https://practical-scheme.net/gauche")
-      (build.platform ,(gauche-architecture))
-      (scheme.path ,@*load-path*)
-      (gauche.threads ,(cond
-                        [(assq 'gauche.sys.pthreads fs) 'pthreads]
-                        [(assq 'gauche.sys.wthreads fs) 'wthreads]
-                        [else 'none]))
-      (gauche.net.tls ,@(cond-list
-                         [(assq 'gauche.net.tls.axtls fs) 'axtls]
-                         [(assq 'gauche.net.tls.mbedtls fs) 'mbedtls])))))
+(define-cproc %build-gosh-version () Scm_BuildGoshVersion)
 
 ;;
 ;; External view of VM.
@@ -642,7 +658,10 @@
    trace port maxdepth skip offset))
 
 (select-module gauche.internal)
-(define-cproc %vm-get-insn-offsets () Scm__VMInsnOffsets)
+(define-cproc %vm-get-insn-address (code::<int>
+                                    :optional (offset?::<boolean> #f))
+  ::<integer>
+  (return (Scm__VMInsnAddress code offset?)))
 
 ;; This is also called from C's Scm_ShowStackTrace
 ;; TRACE is what vm-get-stack-trace-lite returns.
@@ -660,14 +679,24 @@
                [else #f])]
       (unless (> skip 0)
         (format port "~3d  ~,,,,65:s\n" depth (unwrap-syntax (car trace)))
+        ;; Each source info consists of (<file> <line> <form>).
+        ;; The first <form> is the same as (car trace), so we omit it.
         (let1 sis (%source-info (car trace))
-          (if-let1 si (find (^[si] (and (car si) (cadr si))) (reverse sis))
-            (begin
-              (when (and (pair? (cddr si))
-                         (not (eq? (car trace) (caddr si))))
-                (format port "        expanded from ~,,,,60s\n" (caddr si)))
-              (format port "        at ~s:~d\n" (car si) (cadr si)))
-            (format port "        [unknown location]\n")))))))
+          (match sis
+            [((file line  _) . rest)
+             (when (and file line)
+               (format port "        at ~s:~d\n" file line))
+             (let loop ([sis rest])
+               (match sis
+                 [() #f]
+                 [((file line form) . rest)
+                  (format port "        expanded from ~,,,,57:s\n"
+                          (unwrap-syntax form))
+                  (when (and file line)
+                    (format port "        at ~s:~d\n" file line))
+                  (loop rest)]
+                 [(_ . rest) (loop rest)]))]
+            [_ #f]))))))
 
 ;; API
 (select-module gauche.internal)
@@ -680,7 +709,7 @@
 ;; Returns ((<file1> <line1> <form1>) (<file2> <line2> <form2>) ...)
 ;; Where <form1> is a result of macro expansion of <form2> etc.
 ;; <fileN> and <lineN> can be #f if unknown.
-;; For public use, gauche.vm.debugger#debug-source-info is provided.
+;; For public use, debug-source-info below should be provided.
 (select-module gauche.internal)
 (define (%source-info obj)
   (let loop ([obj obj] [r '()])
