@@ -1,7 +1,7 @@
 ;;;
 ;;; libfmt.scm - format
 ;;;
-;;;   Copyright (c) 2013-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2013-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -36,6 +36,8 @@
 (define-module gauche.format
   (use util.match))
 (select-module gauche.format)
+(inline-stub
+ (.include "gauche/priv/configP.h"))
 
 ;; Parsing format string
 ;; The whole parsing business can be written much cleaner using parser.peg,
@@ -58,7 +60,12 @@
 ;;         | ($ flags digits pre-digits width padchar)
 ;;         | (* flags count)
 ;;         | (? flags)
+;;         | (& flags count)
+;;         | (P flags)
 ;;         | (char flags <char> count) ; single-character insertion
+;;         | (bra flags)   ; ~[, beginning of conditional
+;;         | (sep flags)   ; ~;, separator of conditional
+;;         | (ket flags)   ; ~], end of conditional
 ;; flags : '() | (Cons #\@ flags) | (Cons #\: flags)
 
 ;; NB: This procedure is written in a way that all closures can be
@@ -67,12 +74,17 @@
 (define formatter-lex
   (let ()
     (define (fmtstr p) (port-attribute-ref p 'format-string))
-    (define (directive? c) (string-scan "sSaAcCwWdDbBoOxXrR*?fF$~%tT|" c))
-    (define directive-param-spec ; (type max-#-of-params)
+    (define (directive? c)
+      (string-scan "sSaAcCwWdDbBoOxXrR*?fF$~%&pPtT|(){}[];" c))
+    (define directive-param-spec ; (type max-#-of-params [token-id])
       '((S 5) (A 5) (W 0) (C 0)
         (D 4) (B 4) (O 4) (X 4) (x 4) (* 1) (R 5) (r 5) (F 5) ($ 4) (? 0)
+        (& 1) (P 0)
         ;; single-character instertion
-        (~ #\~) (% #\newline) (T #\tab) (|\|| #\page)))
+        (~ 1 #\~) (% 1 #\newline) (T 1 #\tab) (|\|| 1 #\page)
+        ;; tokens for structures
+        (|(| 0 paren) (|)| 0 thesis) (|[| 1 bra) (|]| 0 ket)
+        (|{| 0 curly) (|}| 0 brace) (|\;| 0 sep)))
     (define (flag? c) (memv c '(#\@ #\:)))
     (define (next p)
       (rlet1 c (read-char p)
@@ -82,9 +94,8 @@
       (let loop ([cs (list c)])
         (let1 c (read-char p)
           (cond [(eof-object? c) (list (list->string (reverse cs)))]
-                [(eqv? c #\~)
-                 (cons (list->string (reverse cs))
-                       (directive p (next p)))]
+                [(eqv? c #\~) (cons (list->string (reverse cs))
+                                    (directive p (next p)))]
                 [else (loop (cons c cs))]))))
     ;; NB: this keeps params and flags in reverse order.
     ;; check-param reverses them.
@@ -92,11 +103,19 @@
       (let loop ([params '()] [c c])
         (receive (param c) (fc-param p c params)
           (cond [param (loop (cons param params) c)]
+                [(eqv? c #\newline)  ;; tilde-newline
+                 (if (memq #\@ params)
+                   (if (memq #\: params)
+                     (error "@ and : flags can't be specified simultaneously \
+                             for the ~ + newline directive: " (fmtstr p))
+                     (begin (skip-ws p) (cons "\n" (init p)))) ; ~@\newline
+                   (if (memq #\: params)
+                     (init p)           ; ~:\newline
+                     (begin (skip-ws p) (init p))))] ; ~\newline
                 [(directive? c)
-                 (let1 directive
-                     ($ string->symbol $ string
-                        $ if (memv c '(#\x #\r)) c (char-upcase c))
-                   (acons directive params (init p)))]
+                 (let1 dir ($ string->symbol $ string
+                              $ if (memv c '(#\x #\r)) c (char-upcase c))
+                   (acons dir params (init p)))]
                 [else (errorf "Invalid format directive character `~a' in ~s"
                               c (fmtstr p))]))))
     ;; initial state
@@ -144,6 +163,11 @@
           (if-let1 d (digit->integer c)
             (loop (+ (* val 10) d))
             (values (* val sign) (ensure-param-delimiter p c))))))
+    ;; for tilde-newline.  skip following whitespaces.
+    (define (skip-ws p)
+      (let loop ([c (peek-char p)])
+        (when (char-whitespace? c)
+          (begin (read-char p) (loop (peek-char p))))))
     ;; Check directive parameters vailidity.  This also normalize
     ;; the use of 'empty.  Returns normalized directive.
     (define (check-param directive p)
@@ -154,14 +178,16 @@
                     [(flag? (car ps)) (outer (cdr ps) (cons (car ps) fs))]
                     [(eq? (car ps) 'empty) (outer (cdr ps) fs)]
                     [else (values fs (reverse! ps))]))
-          (let* ([spec (assq (car directive) directive-param-spec)]
-                 [nparams (if (char? (cadr spec)) 1 (cadr spec))])
+          (receive (nparams type)
+              (match (assq (car directive) directive-param-spec)
+                [(_ nparams) (values nparams (car directive))]
+                [(_ nparams dir) (values nparams dir)])
             (unless (length<=? params nparams)
               (errorf "Too many parameters for directive `~a' in ~s"
                       (car directive) (fmtstr p)))
-            (if (char? (cadr spec))
-              (char-node flags params (cadr spec))
-              (list* (car directive) flags params))))
+            (if (char? type)
+              (char-node flags params type)
+              (list* type flags params))))
         directive))
     ;; single character node optimization
     (define (char-node flags params char)
@@ -181,7 +207,7 @@
             (loop (cdr directives)
                   (cons (check-param (car directives) p) r))))))))
 
-;; formatter-parse :: [Directive] -> (Tree, argcnt::Maybe Int)
+;; formatter-parse :: [Directive] -> Tree
 ;;
 ;; Tree = String
 ;;      | (Seq Tree ...)
@@ -198,56 +224,93 @@
 ;;      | ($ flags digits pre-digits width padchar)
 ;;      | (* flags count)
 ;;      | (? flags)
+;;      | (& flags count)
+;;      | (P flags)
 ;;      | (char flags <char> count)
+;;      | (case (Tree ...) Tree)
+;;      | (if Tree Tree)    ; NB: if arg is #f, the first clause is chosen
+;;      | (maybe Tree)
 ;;
-;; argcnt : An integer if the formatter takes fixed number of arguments,
-;;          #f otherwise.
-;;
-;; At this moment we don't have much to do here, but when we introduce
-;; structured formatting directives such as ~{ ~}, this procedure will
-;; recognize it.
 (define formatter-parse
   (let ()
     ;; Plain string.  Concatenate consecutive strings.
-    (define (string-node node ds r cnt)
-      (if (null? ds)
-        (values (cons node r) cnt)
-        (if (string? (car ds))
-          (let loop ([ds (cdr ds)] [strs (list (car ds) node)])
-            (cond
-             [(null? ds)
-              (values (cons (apply string-append (reverse strs)) r) cnt)]
-             [(string? (car ds))
-              (loop (cdr ds) (cons (car ds) strs))]
-             [else
-              (receive (r cnt) (parse ds r cnt)
-                (values (cons (apply string-append (reverse strs)) r) cnt))]))
-          (receive (r cnt) (parse ds r cnt)
-            (values (cons node r) cnt)))))
+    (define (string-node node ds)
+      (let loop ([ds ds] [strs (list node)])
+        (match ds
+          [() (values (list (apply string-append (reverse strs))) '())]
+          [((? string? node) . rest) (loop rest (cons node strs))]
+          [_ (receive (trees rest) (parse ds)
+               (values
+                (cons (apply string-append (reverse strs)) trees)
+                rest))])))
     ;; Non-structured, argument consuming nodes
-    (define (simple-node node ds r cnt)
-      (receive (r cnt) (parse ds r cnt)
-        (values (cons node r)
-                (and cnt (+ 1 (num-variable-params (cdr node)) cnt)))))
-    ;; Jump nodes
-    (define (jump-node node ds r cnt)
-      (receive (r _) (parse ds r #f)
-        (values (cons node r) #f)))
-    ;; count "variable" parameters
-    (define (num-variable-params params)
-      (count (^p (eq? p 'variable)) params))
-    ;; master dispatcher
-    (define (parse ds r cnt)
-      (cond [(null? ds) (values ds cnt)]
-            [(string? (car ds)) (string-node (car ds) (cdr ds) r cnt)]
-            [(eq? (caar ds) '*) (jump-node (car ds) (cdr ds) r cnt)]
-            [else (simple-node (car ds) (cdr ds) r cnt)]))
-    
+    (define (simple-node node ds)
+      (receive (trees rest) (parse ds)
+        (values (cons node trees) rest)))
+    ;; Parsing ~[...~;...~]
+    (define (conditional-node node ds)
+      (receive (tree rest)
+          (match node
+            [(_ ())     (case-tree ds)]
+            [(_ (#\:))  (if-tree ds)]
+            [(_ (#\@))  (maybe-tree ds)]
+            [_ (error "Format directive ~[ may have either : or @ flag, \
+                       but not both")])
+        (receive (trees rest) (parse rest)
+          (values (cons tree trees) rest))))
+    (define (case-tree ds)
+      (let loop ([ds ds] [clauses '()] [next-default? #f])
+        (receive (trees rest) (parse ds)
+          (match rest
+            [(('ket . _) . rest)
+             (values (if next-default?
+                       `(case ,(reverse clauses) ,(treefy trees))
+                       `(case ,(reverse (cons (treefy trees) clauses)) #f))
+                     rest)]
+            [(('sep fs) . rest)
+             (when next-default?
+               (error "Extra ~; directive after ~:;"))
+             (loop rest (cons (treefy trees) clauses) (has-:? fs))]
+            [_ (unterminated "~[")]))))
+    (define (if-tree ds)
+      (receive (trees1 rest) (parse ds)
+        (match rest
+          [(('sep . _) . rest)
+           (receive (trees2 rest) (parse rest)
+             (match rest
+               [(('ket . _) . rest)
+                (values `(if ,(treefy trees1) ,(treefy trees2)) rest)]
+               [_ (unterminated "~:[")]))]
+          [_ (unterminated "~:[")])))
+    (define (maybe-tree ds)
+      (receive (trees rest) (parse ds)
+        (match rest
+          [(('ket . _) . rest) (values `(maybe ,(treefy trees)) rest)]
+          [_ (unterminated "~@[")])))
+    ;; [Tree] -> Tree
+    (define (treefy trees)
+      (match trees
+        [(tree) tree]
+        [(tree ...) `(Seq ,@tree)]))
+    (define (unterminated what) (errorf "Unterminated ~a directive" what))
+    ;; master dispatcher.  returns [Tree] and the rest of ds.
+    (define (parse ds)
+      (match ds
+        [() (values ds '())]
+        [((? string? s) . rest) (string-node s rest)]
+        [(n . rest)
+         (match n
+           [('bra flags) (conditional-node n rest)]
+           [('ket . _)   (values '() ds)]
+           [('sep . _)   (values '() ds)]
+           [_ (simple-node n rest)])]))
+    ;; Main body of formatter-parse.
+    ;; (We don't utilize # of parameters yet).
     (^[directives]
-      (let1 branches (parse directives '() 0)
-        (if (and (pair? branches) (null? (cdr branches)))
-          (car branches)
-          (cons 'Seq branches))))))
+      (receive (trees rest) (parse directives)
+        (unless (null? rest)
+          (errorf "Unbalanced formatter directive: ~~~a" (caar rest)))
+        (treefy trees)))))
 
 ;; Runtime
 ;; Formatters have canonical signature:
@@ -332,7 +395,7 @@
             (error "Bad type of argument for the format parameter ~a: ~s"
                    ',var ,var))))
   (define (gen-vbind var default :optional (check #f))
-    `(,var 
+    `(,var
       (if (eq? ,var 'variable)
         ,(if check
            `(rlet1 ,var (fr-next-arg! fmtstr argptr)
@@ -353,9 +416,8 @@
              (let* ,vbinds
                ,@body))
            (^[argptr port ctrl] ,@body))))))
-           
-;; ~S, ~A, ~W
-;; NB: ~W doesn't take params, so justifying part won't be called for ~W.
+
+;; ~S, ~A
 (define (make-format-expr fmtstr params flags printer)
   (if (null? params)
     (^[argptr port ctrl]
@@ -388,20 +450,71 @@
                (display s port)
                (unless (has-@? flags)
                  (dotimes [_ npad] (write-char padchar port))))))))))
-        
+
 (define (chop-and-out str limit flags port)
-  (if (has-:? flags)
+  ;; check if str chopped at limit contains terminated double-quote.
+  ;; Returns one of the tree values:
+  ;;  #t - Double-quote is terminated.  No special treatment needed.
+  ;;  #f - Double quote is not terminated, and the last character is not
+  ;;       a lone backslash.  We emit an extra double-quote.
+  ;;  #\\ - Double quote is not termianted, and the last character is
+  ;;       a lone backslash.  In this case, we need to emit extra backslash
+  ;;       before the terminating double-quote.  Otherwise, the program
+  ;;       that interprets the result (e.g. font colorizer) mistakes the
+  ;;       closing backquote as escaped.
+  (define (quote-terminated? str limit)
+    (scan-out (open-input-string str) limit))
+  (define (scan-out p limit)
+    (or (<= limit 0)
+        (let1 c (read-char p)
+          (or (eof-object? c)
+              (cond [(eqv? c #\") (scan-in p (- limit 1))]
+                    [(eqv? c #\\) (read-char p) (scan-out p (- limit 2))]
+                    [else (scan-out p (- limit 1))])))))
+  (define (scan-in p limit)
+    (and (> limit 0)
+         (let1 c (read-char p)
+           (and (not (eof-object? c))
+                (cond [(eqv? c #\") (scan-out p (- limit 1))]
+                      [(eqv? c #\\) (read-char p)
+                       (if (= limit 1)
+                         #\\
+                         (scan-in p (- limit 2)))]
+                      [else (scan-in p (- limit 1))])))))
+
+  (if (and (has-:? flags) (>= limit 4))
     (begin (display (substring str 0 (- limit 4)) port)
-           (display " ..." port))
+           (case (quote-terminated? str (- limit 4))
+             [(#t) (display " ..." port)]
+             [(#f) (display "\"..." port)]
+             [(#\\) (display "\\\".." port)]))
     (display (substring str 0 limit) port)))
 
+;; ~W
+;; Without flags, just behave like ~S.
+;; With ':'-flag, pretty print.
+;; With '@'-flag, remove level and length limits.
+(define (make-format-pprint fmtstr flags)
+  (let1 ctrl-args
+      (cond-list [(has-:? flags) @ '(:pretty #t :width 79)]
+                 [(has-@? flags) @ '(:length #f :level #f)])
+    (^[argptr port ctrl]
+      (let ([arg (fr-next-arg! fmtstr argptr)]
+            [ctrl-args (list* :indent
+                              ((with-module gauche.internal port-column) port)
+                              ctrl-args)])
+        (write-shared arg port
+                      (if ctrl
+                        (apply write-controls-copy ctrl ctrl-args)
+                        (apply make-write-controls ctrl-args)))))))
+
 ;; ~C
 ;; The "spelling out" mode (~:C) isn't supported yet.
 (define (make-format-char fmtstr flags)
   (define (char-formatter writer)
     (^[argptr port ctrl]
       (let1 c (fr-next-arg! fmtstr argptr)
-        (unless (char? c) 
+        (unless (char? c)
           (error "Character required for ~c format directive, but got:" c))
         (writer c port))))
   (char-formatter (if (has-@? flags) write display)))
@@ -464,7 +577,12 @@
 ;; ~R
 (define (make-format-r fmtstr params flags upcase)
   (if (null? params)
-    (^[argptr port ctrl] (error "Roman numerals not implemented yet"))
+    (^[argptr port ctrl]
+      (format-numeral-R port
+                        (fr-next-arg! fmtstr argptr)
+                        (has-@? flags)
+                        (has-:? flags)
+                        upcase))
     ($ with-format-params ([radix 10]
                            [mincol 0]
                            [padchar #\space]
@@ -496,17 +614,17 @@
               (let1 s (call-with-output-string
                         (^p
                          (flo-fmt (* (real-part arg) (expt 10 scale))
-                                  0 digits 0 ovchar padchar 
+                                  0 digits 0 ovchar padchar
                                   (has-@? flags) (has-:? flags) #f p)
                          (flo-fmt (* (imag-part arg) (expt 10 scale))
-                                  0 digits 0 ovchar padchar 
+                                  0 digits 0 ovchar padchar
                                   #t (has-:? flags) #f p)
                          (display "i" p)))
                 (let1 len (string-length s)
                   (when (< len width)
                     (dotimes [(- width len)]  (write-char padchar port)))
                   (display s port)))]
-             [else 
+             [else
               (let1 sarg (write-to-string arg display)
                 (let1 len (string-length sarg)
                   (when (< len width)
@@ -524,7 +642,7 @@
      (let1 arg (fr-peek-arg fmtstr argptr)
        (if (real? arg)
          (flo-fmt (inexact (fr-next-arg! fmtstr argptr))
-                  width digits pre-digits 
+                  width digits pre-digits
                   #f padchar (has-@? flags) #t (has-:? flags) port)
          ;; if arg isn't real, use ~wD.
          (format-num-body fmtstr argptr port ctrl 10 #f
@@ -541,7 +659,7 @@
               (let* ([m (#/^([+-]?)(\d+)/ s0)]
                      [pre (m 2)])
                 (if (> pre-digits (string-length pre))
-                  (string-append 
+                  (string-append
                    (m 1)
                    (make-string (- pre-digits (string-length pre)) #\0)
                    pre (m 'after))
@@ -567,7 +685,7 @@
       (error "Argument for ~? must be a string, but got" str))
     ($ formatter-compile-rec str $ formatter-parse $ formatter-lex str))
   (if (has-@? flags)
-    ;; take 
+    ;; take
     (^[argptr port ctrl]
       (let1 formatter (fcompile (fr-next-arg! fmtstr argptr))
         (formatter argptr port ctrl)))
@@ -575,7 +693,7 @@
       (let* ([formatter (fcompile (fr-next-arg! fmtstr argptr))]
              [xargptr (fr-make-argptr (fr-next-arg! fmtstr argptr))])
         (formatter xargptr port ctrl)))))
-           
+
 ;; ~*
 (define (make-format-jump fmtstr params flags)
   ($ with-format-params ([count 1])
@@ -584,11 +702,65 @@
          (fr-jump-arg! argptr count)
          (fr-jump-arg-relative! argptr count)))))
 
-;; ~t, ~%, ~~
+;; ~&
+(define (make-format-fresh-line fmtstr params flags)
+  ($ with-format-params ([count 1])
+     (when (>= count 1)
+       (fresh-line port)
+       (when (>= count 2)
+         (dotimes [(- count 1)] (display "\n" port))))))
+
+;; ~P
+(define (make-format-plural fmtstr flags)
+  (^[argptr port ctl]
+    (when (has-:? flags)
+      (fr-jump-arg-relative! argptr -1))
+    (let1 arg (fr-next-arg! fmtstr argptr)
+      ;; NB: Only exact 1 is singular.
+      (if (eqv? arg 1)
+        (when (has-@? flags)
+          (display "y" port))
+        (if (has-@? flags)
+          (display "ies" port)
+          (display "s" port))))))
+
+;; ~t, ~%, ~~, ~|
 (define (make-format-single fmtstr ch params flags)
   ($ with-format-params ([count 1])
      (display (make-string count ch) port)))
 
+;; ~[...~;...~]
+(define (make-format-case fmtstr clauses default-clause)
+  (define branches
+    (map (cut formatter-compile-rec fmtstr <>) clauses))
+  (define fallback
+    (if default-clause
+      (formatter-compile-rec fmtstr default-clause)
+      (^[argptr port ctl] argptr)))
+  (^[argptr port ctl]
+    (let1 arg (fr-next-arg! fmtstr argptr)
+      (unless (exact-integer? arg)
+        (error "Argument for ~[ must be an exact integer, but got" arg))
+      (let1 branch-formatter (list-ref branches arg fallback)
+        (branch-formatter argptr port ctl)))))
+
+(define (make-format-if fmtstr alternative consequent)
+  (define alt-formatter (formatter-compile-rec fmtstr alternative))
+  (define csq-formatter (formatter-compile-rec fmtstr consequent))
+  (^[argptr port ctl]
+    (let1 arg (fr-next-arg! fmtstr argptr)
+      (if arg
+        (csq-formatter argptr port ctl)
+        (alt-formatter argptr port ctl)))))
+
+(define (make-format-maybe fmtstr consequent)
+  (define csq-formatter (formatter-compile-rec fmtstr consequent))
+  (^[argptr port ctl]
+    (if-let1 arg (fr-peek-arg fmtstr argptr)
+      (csq-formatter argptr port ctl)
+      (begin (fr-jump-arg! argptr 1)
+             (^[argptr port ctl] argptr)))))
+
 ;; Tree -> Formatter
 ;; src : source format string for error message
 (define (formatter-compile-rec src tree)
@@ -600,7 +772,7 @@
     [('S fs . ps) (make-format-expr src ps fs write)]
     [('A fs . ps) (make-format-expr src ps fs display)]
     [('C fs . ps) (make-format-char src fs)]
-    [('W fs . ps) (make-format-expr src ps fs write-shared)]
+    [('W fs . ps) (make-format-pprint src fs)]
     [('D fs . ps) (make-format-num src ps fs 10 #f)]
     [('B fs . ps) (make-format-num src ps fs 2 #f)]
     [('O fs . ps) (make-format-num src ps fs 8 #f)]
@@ -611,8 +783,13 @@
     [((or 'R 'r) fs . ps) (make-format-r src ps fs (eq? (car tree) 'R))]
     [('? fs)      (make-format-recur src fs)]
     [('* fs . ps) (make-format-jump src ps fs)]
+    [('& fs . ps) (make-format-fresh-line src ps fs)]
+    [('P fs)      (make-format-plural src fs)]
     [('char fs c . ps) (make-format-single src c ps fs)]
-    [_ (error "boo!")]))
+    [('case cls default) (make-format-case src cls default)]
+    [('if alt csq) (make-format-if src alt csq)]
+    [('maybe csq) (make-format-maybe src csq)]
+    [_ (error "Unsupported formatter directive:" tree)]))
 
 ;; Toplevel compiler
 ;; Returns formatter procedure
@@ -638,8 +815,12 @@
         [locking? (with-port-locking port formatter args port ctrl)]
         [else (formatter args port ctrl)]))
 
-(define (format-2 shared? out control fmtstr args)
-  (let1 formatter (formatter-compile fmtstr)
+(define (format-2 formatter-cache shared? out control fmtstr args)
+  (let1 formatter (if formatter-cache
+                    (or (unbox formatter-cache)
+                        (rlet1 f (formatter-compile fmtstr)
+                          (set-box! formatter-cache f)))
+                    (formatter-compile fmtstr))
     (case out
       [(#t)
        (call-formatter shared? #t formatter (current-output-port) control args)]
@@ -648,28 +829,51 @@
               (get-output-string out))]
       [else (call-formatter shared? #t formatter out control args)])))
 
-;; handle optional destination arg
-(define (format-1 shared? args)
-  (let loop ([port 'unseen]
-             [control 'unseen]
-             [as args])
-    (cond [(null? as) (error "format: too few arguments" args)]
-          [(string? (car as))
-           (format-2 shared?
-                     (if (eq? port 'unseen) #f port)
-                     (if (eq? control 'unseen) #f control)
-                     (car as)
-                     (cdr as))]
-          [(or (port? (car as)) (boolean? (car as)))
-           (if (eq? port 'unseen)
-             (loop (car as) control (cdr as))
-             (error "format: multiple ports given" args))]
-          [(is-a? (car as) <write-controls>)
-           (if (eq? control 'unseen)
-             (loop port (car as) (cdr as))
-             (error "format: multiple controls given" args))]
-          [else (error "format: invalid argument" (car as))])))
+;; This is a bridge to the internal API.  Although this isn't exported,
+;; inliner or compiler macro may insert a call to this.  You shouldn't
+;; change the signature across versions.
+;;
+;; CONTEXT is a list parameters statically constructed during compilation
+;; phase.  ARGS is the actual parameter list passed.
+;;
+;; CONTEXT is currently three-argument list:
+;;    shared?  - a flag to distinguish format (#f) and format/ss (#t).
+;;    literal-str-pos - the position of the first string literal in
+;;               the actual parameters, up to 2.  If the format directive
+;;               string is a literal, we can cache the formatter.
+;;               We may not be able to statically determine the format
+;;               string, because of the optional port/write-controls arguments.
+;;    formatter-cache - if there's a string literal in the 0, 1 or 2nd argument,
+;;               the compiler macro allocates a box and passes it.
+;;               This will be used to cache the formatter.
+;;
+;; We make CONTEXT a list so that we can extend it in future without changing
+;; the signature.  Those CONTEXT can be computed at compile-time and becomes
+;; a static literal list.
+(define (format-internal context args)
+  (match-let1 (shared? literal-str-pos format-cache . _) context
+    (let loop ([port 'unseen]
+               [control 'unseen]
+               [as args]
+               [pos 0])
+      (cond [(null? as) (error "format: too few arguments" args)]
+            [(string? (car as))
+             (format-2 (and (eqv? pos literal-str-pos) format-cache)
+                       shared?
+                       (if (eq? port 'unseen) #f port)
+                       (if (eq? control 'unseen) #f control)
+                       (car as)
+                       (cdr as))]
+            [(or (port? (car as)) (boolean? (car as)))
+             (if (eq? port 'unseen)
+               (loop (car as) control (cdr as) (+ pos 1))
+               (error "format: multiple ports given" args))]
+            [(is-a? (car as) <write-controls>)
+             (if (eq? control 'unseen)
+               (loop port (car as) (cdr as) (+ pos 1))
+               (error "format: multiple controls given" args))]
+            [else (error "format: invalid argument" (car as))]))))
 
-;; API
-(define-in-module gauche (format . args) (format-1 #f args))
-(define-in-module gauche (format/ss . args) (format-1 #t args))
+;; The external API, format and format/ss, are defined in libmacro.scm.
+;; We want to make them hybrid macro, and the definition needs to come
+;; after macro system initialization.
