@@ -1,7 +1,7 @@
 ;;;
 ;;; libmisc.scm - miscellaneous built-in procedures
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -33,7 +33,8 @@
 
 (select-module gauche.internal)
 (inline-stub
- (declcode (.include <gauche/vminsn.h>)))
+ (.include "gauche/priv/configP.h"
+           "gauche/vminsn.h"))
 
 ;;;
 ;;; Miscellaneous
@@ -92,4 +93,3 @@
 
 (inline-stub
  (define-constant SLIB_DIR (c (SCM_MAKE_STR_IMMUTABLE SLIB_DIR))))
-
