@@ -1,7 +1,7 @@
 ;;;
 ;;; libnum.stub - builtin number libraries
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,13 +34,14 @@
 (select-module gauche.internal)
 
 (inline-stub
- (declcode (.include <gauche/vminsn.h>
-                     <gauche/bignum.h>
-                     <stdlib.h>
-                     <float.h>
-                     <math.h>))
- (when "!defined(M_PI)"
-   (declcode "#define M_PI 3.1415926535897932384")))
+ (.include "gauche/priv/configP.h"
+           "gauche/vminsn.h"
+           "gauche/priv/bignumP.h"
+           <stdlib.h>
+           <float.h>
+           <math.h>)
+ (.when (not (defined M_PI))
+   (.define M_PI 3.1415926535897932384)))
 
 ;;
 ;; Predicates
@@ -77,11 +78,12 @@
 (define-cproc fixnum? (x) ::<boolean> :fast-flonum SCM_INTP)
 (define-cproc bignum? (x) ::<boolean> :fast-flonum SCM_BIGNUMP)
 (define-cproc flonum? (x) ::<boolean> :fast-flonum :constant SCM_FLONUMP)
+(define-cproc ratnum? (x) ::<boolean> :fast-flonum :constant SCM_RATNUMP)
 
 (define-cproc finite?   (x::<number>) ::<boolean> :fast-flonum Scm_FiniteP)
 (define-cproc infinite? (x::<number>) ::<boolean> :fast-flonum Scm_InfiniteP)
 (define-cproc nan?      (x::<number>) ::<boolean> :fast-flonum Scm_NanP)
-(define-cproc -zero?    (x::<number>) ::<boolean> :fast-flonum :constant
+(define-cproc negative-zero? (x::<number>) ::<boolean> :fast-flonum :constant
   (return (and (SCM_FLONUMP x)          ;only flonums have -0.0
                (== (SCM_FLONUM_VALUE x) 0.0)
                (!= (signbit (SCM_FLONUM_VALUE x)) 0))))
@@ -91,7 +93,9 @@
 
 ;;
 ;; Platform introspection
-;;
+;; TODO: These are not :constant for now, because of cross-compilation.
+;; We may need the means to provide target-specific constants during
+;; cross compilation.
 
 (select-module gauche)
 ;; Names are from R6RS.
@@ -108,17 +112,17 @@
 (define-cproc native-endian () Scm_NativeEndian)
 
 ;; DBL_EPSILON, etc.
-(define-cproc flonum-epsilon () 
+(define-cproc flonum-epsilon ()
   (let* ([x::(static ScmObj) SCM_UNBOUND])
     (when (== x SCM_UNBOUND)
       (set! x (Scm_MakeFlonum DBL_EPSILON)))
     (return x)))
-(define-cproc flonum-min-normalized () 
+(define-cproc least-positive-normalized-flonum ()
   (let* ([x::(static ScmObj) SCM_UNBOUND])
     (when (== x SCM_UNBOUND)
       (set! x (Scm_MakeFlonum DBL_MIN)))
     (return x)))
-(define-cproc flonum-min-denormalized ()
+(define-cproc least-positive-flonum ()
   (let* ([x::(static ScmObj) SCM_UNBOUND])
     (when (== x SCM_UNBOUND)
       (let* ([m::double (Scm_EncodeFlonum (SCM_MAKE_INT 1) -1074 1)])
@@ -126,11 +130,17 @@
           (set! x (Scm_MakeFlonum DBL_MIN))
           (set! x (Scm_MakeFlonum m)))))
     (return x)))
+(define-cproc greatest-positive-flonum ()
+  (return (Scm_MakeFlonum DBL_MAX)))
+
+;; For backward compatibility
+(define flonum-min-normalized least-positive-normalized-flonum)
+(define flonum-min-denormalized least-positive-flonum)
 
 (select-module gauche.internal)
 (define-cproc %bignum-dump (obj) ::<void>
   (when (SCM_BIGNUMP obj)
-    (Scm_DumpBignum (SCM_BIGNUM obj) SCM_CUROUT)))
+    (Scm_BignumDump (SCM_BIGNUM obj) SCM_CUROUT)))
 
 ;;
 ;; Comparison
@@ -189,6 +199,10 @@
   (Scm_MinMax arg0 args (& SCM_RESULT) NULL))
 
 (select-module gauche)
+(define-cproc <=:< (x y z) ::<boolean> :fast-flonum :constant
+  (return (and (Scm_NumLE x y) (Scm_NumLT y z))))
+(define-cproc <:<= (x y z) ::<boolean> :fast-flonum :constant
+  (return (and (Scm_NumLT x y) (Scm_NumLE y z))))
 (define-cproc min&max (arg0 :rest args) ::(<top> <top>)
   (Scm_MinMax arg0 args (& SCM_RESULT0) (& SCM_RESULT1)))
 
@@ -201,8 +215,8 @@
 (define-cproc inexact->exact (obj) :fast-flonum :constant Scm_Exact)
 
 (select-module gauche)
-(define exact   inexact->exact)           ;R6RS
-(define inexact exact->inexact)           ;R6RS
+(define-inline exact   inexact->exact)           ;R6RS
+(define-inline inexact exact->inexact)           ;R6RS
 
 (select-module scheme)
 (define-cproc number->string (obj :optional (radix::<fixnum> 10)
@@ -214,7 +228,7 @@
          [o::ScmPort* (SCM_PORT (Scm_MakeOutputStringPort TRUE))])
     (when (or (< radix SCM_RADIX_MIN) (> radix SCM_RADIX_MAX))
       (Scm_Error "radix must be an integer between %d and %d, but got %d"
-                 SCM_RADIX_MIN SCM_RADIX_MAX radix))  
+                 SCM_RADIX_MIN SCM_RADIX_MAX radix))
     (cond [(or (SCM_FALSEP flags) (SCM_NULLP flags)) (set! f 0)]
           [(SCM_TRUEP flags) (set! f SCM_NUMBER_FORMAT_USE_UPPER)];compatibility
           [(SCM_PAIRP flags)
@@ -242,9 +256,9 @@
                                         (default-exactness #f))
   (let* ([flags::u_long 0])
     (cond
-     [(SCM_EQ default-exactness 'exact) 
+     [(SCM_EQ default-exactness 'exact)
       (set! flags SCM_NUMBER_FORMAT_EXACT)]
-     [(SCM_EQ default-exactness 'inexact) 
+     [(SCM_EQ default-exactness 'inexact)
       (set! flags SCM_NUMBER_FORMAT_INEXACT)]
      [(SCM_FALSEP default-exactness)]
      [else (Scm_Error "default-exactness must be either #f, exact or inexact, but got: %S" default-exactness)])
@@ -277,45 +291,64 @@
            (return v))]
         [else (SCM_TYPE_ERROR num "real number") (return SCM_UNDEFINED)]))
 
-(define-cproc fmod (x::<real> y::<real>) ::<double> fmod)
+(define-cproc fmod (x::<double> y::<double>)::<double> :constant fmod)
 
-(define-cproc frexp (d::<real>) ::(<double> <int>)
+(define-cproc frexp (d::<double>) ::(<double> <int>) :constant
   (set! SCM_RESULT0 (frexp d (& SCM_RESULT1))))
 
-(define-cproc modf (x::<real>) ::(<double> <double>)
+(define-cproc modf (x::<double>) ::(<double> <double>) :constant
   (set! SCM_RESULT0 (modf x (& SCM_RESULT1))))
 
-(define-cproc ldexp (x::<real> exp::<int>) ::<real> ldexp)
+(define-cproc ldexp (x::<double> exp::<int>) ::<double> :constant ldexp)
 
-(define-cproc log10 (x::<real>) ::<real> log10)
+(define-cproc log10 (x::<double>) ::<double> :constant log10)
 
 ;; NB: Alternative implemenation of gamma and log-abs-gamma functions are
 ;; provided in Scheme (lib/gauche/numerical.scm).
 (select-module gauche.internal)
 (inline-stub
- (if "defined(HAVE_TGAMMA)"
-   (define-cproc %gamma (x::<real>) ::<real> :fast-flonum :constant
+ (.when (defined HAVE_TGAMMA)
+   (define-cproc %gamma (x::<double>) ::<double> :fast-flonum :constant
      tgamma))
- (if "defined(HAVE_LGAMMA)"
-   (define-cproc %lgamma (x::<real>) ::<real> :fast-flonum :constant
+ (.when (defined HAVE_LGAMMA)
+   (define-cproc %lgamma (x::<double>) ::<double> :fast-flonum :constant
      lgamma)))
 ;; Returns the HalfFloat representation as integer.  For now,
 ;; we keep it in gauche.internal.
-(define-cproc flonum->f16bits (x::<real>) ::<int> Scm_DoubleToHalf)
+(define-cproc flonum->f16bits (x::<double>) ::<int> :constant Scm_DoubleToHalf)
 
 ;;
 ;; Arithmetics
 ;;
 
+;; NB: Compile-time constant folding for these four procedures are
+;; handled in the compiler, so we don't need :constant flag here.
+
+;; For - and /, one-argument case is handled by separate C procedures.
+;; (Scm_Negate(), Scm_Reciprocal()).  For + and *, we return the argument
+;; as is if it is a number, or delegate it to unary object-+ / object-*
+;; method.
 (select-module scheme)
 (define-cproc * (:rest args) ::<number> :fast-flonum
   (cond [(not (SCM_PAIRP args)) (return (SCM_MAKE_INT 1))]
+        [(and (not (SCM_PAIRP (SCM_CDR args)))
+              (not (SCM_NUMBERP (SCM_CAR args))))
+         (let* ([unary-proc SCM_UNDEFINED])
+           (SCM_BIND_PROC unary-proc "%apply-unary-*"
+                          (Scm_GaucheInternalModule))
+           (return (Scm_ApplyRec1 unary-proc (SCM_CAR args))))]
         [else (let* ([r::ScmObj (SCM_CAR args)])
                 (dolist [v (SCM_CDR args)] (set! r (Scm_Mul r v)))
                 (return r))]))
 
 (define-cproc + (:rest args) ::<number> :fast-flonum
   (cond [(not (SCM_PAIRP args)) (return (SCM_MAKE_INT 0))]
+        [(and (not (SCM_PAIRP (SCM_CDR args)))
+              (not (SCM_NUMBERP (SCM_CAR args))))
+         (let* ([unary-proc SCM_UNDEFINED])
+           (SCM_BIND_PROC unary-proc "%apply-unary-+"
+                          (Scm_GaucheInternalModule))
+           (return (Scm_ApplyRec1 unary-proc (SCM_CAR args))))]
         [else (let* ([r::ScmObj (SCM_CAR args)])
                 (dolist [v (SCM_CDR args)] (set! r (Scm_Add r v)))
                 (return r))]))
@@ -332,6 +365,24 @@
     (begin (dolist [v args] (set! arg1 (Scm_Div arg1 v)))
            (return arg1))))
 
+(select-module gauche.internal)
+(define (%apply-unary-generic gf who obj)
+  ;; TRANSIENT: Up to 0.9.14, we return the argument as is from unary + and *
+  ;; unconditionally.  For the backward compatibility if there's no unary
+  ;; object-+ / object-*, we warn so and returns the argument instead of
+  ;; raising an error.  Should remove that part after a few releases.
+  ;; https://github.com/shirok/Gauche/issues/1012
+  (if (applicable? gf (class-of obj))
+    (gf obj)
+    (begin
+      (warn "operation ~a is not defined on ~s.  \
+             This will be an error in future releases.\n" who obj)
+      obj)))
+(define (%apply-unary-+ obj) (%apply-unary-generic object-+ '+ obj))
+(define (%apply-unary-* obj) (%apply-unary-generic object-* '* obj))
+
+(select-module scheme)
+
 (define-cproc abs (obj) :fast-flonum :constant Scm_VMAbs)
 
 (define-cproc quotient (n1 n2) :fast-flonum :constant
@@ -402,39 +453,38 @@
   (return (Scm_Round v SCM_ROUND_ROUND)))
 
 ;; Transcedental functions.   First, real-only versions.
+;; The name real-* is coined in SRFI-94.
+;; NB: We don't support 'real-log' of SRFI-94.  It takes base number first,
+;; which is reverse of R7RS 'log'.  It would be too confusing.
+(select-module gauche)
+(define-cproc real-exp (x::<double>) ::<double> :fast-flonum :constant exp)
+(define-cproc real-ln (x::<number>) ::<double> :fast-flonum :constant
+  (unless (SCM_REALP x) (SCM_TYPE_ERROR x "real number"))
+  (when (< (Scm_Sign x) 0)
+    (Scm_Error "Argument must be nonnegative real number: %S" x))
+  ;; We can't simply cast x to double, for x can be a large bignum
+  ;; outside of double.
+  (let* ([d::double (Scm_GetDouble x)])
+    (when (== d SCM_DBL_POSITIVE_INFINITY)
+      (if (SCM_BIGNUMP x)
+        (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM x) values))]
+               [scale::long (- (Scm_BitsHighest1 z 0 (* (SCM_BIGNUM_SIZE x) SCM_WORD_BITS)) 53)])
+          (return (+ (log (Scm_GetDouble
+                           (Scm_DivInexact x (Scm_Ash (SCM_MAKE_INT 1) scale))))
+                     (* scale (log 2.0)))))
+        (return SCM_DBL_POSITIVE_INFINITY)))
+    (return (log d))))
 
 (select-module gauche)
-(define-cproc %exp (x::<real>) ::<real> :fast-flonum :constant exp)
+(define-cproc real-sin (x::<double>) ::<double> :fast-flonum :constant sin)
+(define-cproc real-cos (x::<double>) ::<double> :fast-flonum :constant cos)
+(define-cproc real-tan (x::<double>) ::<double> :fast-flonum :constant tan)
 
-(define-cproc %log (x) ::<number> :fast-flonum :constant
-  (unless (SCM_REALP x) (SCM_TYPE_ERROR x "real number"))
-  (when (Scm_InfiniteP x)
-    (if (> (Scm_Sign x) 0)
-      (return SCM_POSITIVE_INFINITY)
-      (return (Scm_MakeComplex SCM_DBL_POSITIVE_INFINITY M_PI))))
-  (let* ([d::double (Scm_GetDouble x)]
-         [shift::double 0.0])
-    (when (or (== d SCM_DBL_POSITIVE_INFINITY)  ; SCM_IS_INF isn't visible...
-              (== d SCM_DBL_NEGATIVE_INFINITY))
-      (SCM_ASSERT (SCM_BIGNUMP x))  ; the arg is too big to represent in double
-      (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM x) values))]
-             [scale::long (Scm_BitsHighest1 z 0 (* (SCM_BIGNUM_SIZE x) SCM_WORD_BITS))])
-        (set! shift (* scale (log 2.0)))
-        (set! d (Scm_GetDouble
-                 (Scm_DivInexact x (Scm_Ash (SCM_MAKE_INT 1) scale))))))
-    (if (< (Scm_FlonumSign d) 0)
-      (return (Scm_MakeComplex (+ (log (- d)) shift) M_PI))
-      (return (Scm_VMReturnFlonum (+ (log d) shift))))))
+(define-cproc real-sinpi (x::<double>) ::<double> :fast-flonum :constant Scm_SinPi)
+(define-cproc real-cospi (x::<double>) ::<double> :fast-flonum :constant Scm_CosPi)
+(define-cproc real-tanpi (x::<double>) ::<double> :fast-flonum :constant Scm_TanPi)
 
-(define-cproc %sin (x::<real>) ::<real> :fast-flonum :constant sin)
-(define-cproc %cos (x::<real>) ::<real> :fast-flonum :constant cos)
-(define-cproc %tan (x::<real>) ::<real> :fast-flonum :constant tan)
-
-(define-cproc %sinpi (x::<real>) ::<real> :fast-flonum :constant Scm_SinPi)
-(define-cproc %cospi (x::<real>) ::<real> :fast-flonum :constant Scm_CosPi)
-(define-cproc %tanpi (x::<real>) ::<real> :fast-flonum :constant Scm_TanPi)
-
-(define-cproc %asin (x::<real>) ::<number> :fast-flonum :constant
+(define-cproc real-asin (x::<double>) ::<number> :fast-flonum :constant
   (cond [(> x 1.0)
          (return (Scm_MakeComplex (/ M_PI 2.0)
                                   (- (log (+ x (sqrt (- (* x x) 1.0)))))))]
@@ -443,24 +493,26 @@
                                   (- (log (- (- x) (sqrt (- (* x x) 1.0)))))))]
         [else (return (Scm_VMReturnFlonum (asin x)))]))
 
-(define-cproc %acos (x::<real>) ::<number> :fast-flonum :constant
+(define-cproc real-acos (x::<double>) ::<number> :fast-flonum :constant
   (cond [(> x 1.0)
          (return (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
         [(< x -1.0)
          (return (Scm_MakeComplex 0 (log (+ x (sqrt (- (* x x) 1.0))))))]
         [else (return (Scm_VMReturnFlonum (acos x)))]))
 
-(define-cproc %atan (z::<real> :optional x) ::<double> :fast-flonum :constant
+(define-cproc real-atan (z::<double> :optional x) ::<double> :fast-flonum :constant
   (cond [(SCM_UNBOUNDP x) (return (atan z))]
         [else (unless (SCM_REALP x) (SCM_TYPE_ERROR x "real number"))
               (return (atan2 z (Scm_GetDouble x)))]))
 
-(define-cproc %sinh (x::<real>) ::<real> :fast-flonum :constant sinh)
-(define-cproc %cosh (x::<real>) ::<real> :fast-flonum :constant cosh)
-(define-cproc %tanh (x::<real>) ::<real> :fast-flonum :constant tanh)
-;; NB: asinh and acosh are not in POSIX.
+(define-cproc real-sinh (x::<double>) ::<double> :fast-flonum :constant sinh)
+(define-cproc real-cosh (x::<double>) ::<double> :fast-flonum :constant cosh)
+(define-cproc real-tanh (x::<double>) ::<double> :fast-flonum :constant tanh)
+(define-cproc real-asinh (x::<double>) ::<double> :fast-flonum :constant asinh)
+(define-cproc real-acosh (x::<double>) ::<double> :fast-flonum :constant acosh)
+(define-cproc real-atanh (x::<double>) ::<double> :fast-flonum :constant atanh)
 
-(define-cproc %sqrt (x::<real>) :fast-flonum :constant
+(define-cproc real-sqrt (x::<double>) :fast-flonum :constant
   (if (< x 0)
     (return (Scm_MakeComplex 0.0 (sqrt (- x))))
     (return (Scm_VMReturnFlonum (sqrt x)))))
@@ -487,19 +539,29 @@
         [else (return SCM_FALSE)]))
 
 (select-module gauche)
-(define-cproc %expt (x y) :fast-flonum :constant Scm_Expt)
+(define-cproc real-expt (x y) :fast-flonum :constant Scm_Expt)
+(define-cproc exact-expt (x y::<integer>) :constant
+  (unless (SCM_EXACTP x) (SCM_TYPE_ERROR x "exact real number"))
+  (return (Scm_ExactIntegerExpt x y)))
+(define-cproc integer-expt (x::<integer> y::<integer>) ::<integer> :constant
+  Scm_ExactIntegerExpt)
 
 ;; Now, handles complex numbers.
 ;;  Cf. Teiji Takagi: "Kaiseki Gairon" pp.193--198
 (define-in-module scheme (exp z)
-  (cond [(real? z) (%exp z)]
-        [(complex? z) (make-polar (%exp (real-part z)) (imag-part z))]
+  (cond [(real? z) (real-exp z)]
+        [(complex? z) (make-polar (real-exp (real-part z)) (imag-part z))]
         [else (error "number required, but got" z)]))
 
 (define-in-module scheme (log z . base)
   (if (null? base)
-    (cond [(real? z) (%log z)]
-          [(complex? z) (make-rectangular (%log (magnitude z)) (angle z))]
+    (cond [(real? z)
+           (if (<= 0 z)
+             (real-ln z)
+             ;; The constant pi is not avaialble here, but (* 4 (real-atan 1))
+             ;; is constand-foled to pi.
+             (make-rectangular (real-ln (- z)) (* 4 (real-atan 1))))]
+          [(complex? z) (make-rectangular (real-ln (magnitude z)) (angle z))]
           [else (error "number required, but got" z)])
     (/ (log z) (log (car base)))))  ; R6RS addition
 
@@ -511,7 +573,7 @@
     ;; Gauche doesn't have exact complex, so we have real z.
     (if (integer? z)
       (receive (s r) (%exact-integer-sqrt z)
-        (if (= r 0) s (%sqrt z)))
+        (if (= r 0) s (real-sqrt z)))
       (let ([n (numerator z)]
             [d (denominator z)])
         (if-let1 nq (%sqrt-fast-path n)
@@ -520,30 +582,30 @@
             (receive (ds dr) (%exact-integer-sqrt d)
               (if (= dr 0)
                 (/ nq ds)
-                (%sqrt z))))
+                (real-sqrt z))))
           (receive (ns nr) (%exact-integer-sqrt n)
             (if (= nr 0)
               (receive (ds dr) (%exact-integer-sqrt d)
                 (if (= dr 0)
-                  (/ ns ds)
-                  (%sqrt z)))
-              (%sqrt z))))))]
-   [(real? z) (%sqrt z)]
-   [(complex? z) (make-polar (%sqrt (magnitude z)) (/ (angle z) 2.0))]
+                 (/ ns ds)
+                  (real-sqrt z)))
+              (real-sqrt z))))))]
+   [(real? z) (real-sqrt z)]
+   [(complex? z) (make-polar (real-sqrt (magnitude z)) (/ (angle z) 2.0))]
    [else (error "number required, but got" z)]))
 
 (define-in-module gauche.internal (%exact-integer-sqrt k) ; k >= 0
   (if (< k 4503599627370496)            ;2^52
     ;; k can be converted to a double without loss.
-    (let1 s (floor->exact (%sqrt k))
+    (let1 s (floor->exact (real-sqrt k))
       (values s (- k (* s s))))
     ;; use Newton-Rhapson
-    ;; If k is representable with double, we use (%sqrt k) as the initial
+    ;; If k is representable with double, we use (real-sqrt k) as the initial
     ;; estimate, for calculating double sqrt is fast.  If k is too large,
     ;; we use 2^floor((log2(k)+1)/2) as the initial value.
-    (let loop ([s (let1 ik (%sqrt k)
+    (let loop ([s (let1 ik (real-sqrt k)
                     (if (finite? ik)
-                      (floor->exact (%sqrt k))
+                      (floor->exact (real-sqrt k))
                       (ash 1 (quotient (integer-length k) 2))))])
       (let1 s2 (* s s)
         (if (< k s2)
@@ -557,22 +619,21 @@
   (cond [(and (exact? x) (exact? y))
          ((with-module gauche.internal %exact-expt) x y)]
         [(real? x)
-         (cond [(real? y) (%expt x y)]
+         (cond [(real? y) (real-expt x y)]
                [(number? y)
                 (let1 ry (real-part y)
                   (if (and (zero? x) (positive? ry))
                     (if (exact? x) 0 0.0)
-                    (* (%expt x ry)
-                       (exp (* +i (imag-part y) (%log x))))))]
+                    (* (real-expt x ry)
+                       (exp (* +i (imag-part y) (real-ln x))))))]
                [else (error "number required, but got" y)])]
         [(number? x) (exp (* y (log x)))]
         [else (error "number required, but got" x)]))
 
 (select-module gauche.internal)
-(define-cproc %exact-integer-expt (x y) :constant Scm_ExactIntegerExpt)
 (define (%exact-expt x y) ;; x, y :: exact
-  (cond [(integer? y) (%exact-integer-expt x y)]
-        [(< x 0) (%expt x y)] ; we don't have exact compnum  
+  (cond [(integer? y) (exact-expt x y)]
+        [(< x 0) (real-expt x y)] ; we don't have exact compnum
         [(< y 0) (/ (%exact-expt x (- y)))]
         [(integer? x)
          (let ([a (numerator y)]
@@ -593,24 +654,24 @@
            ;;     ( r_N+1 - r_N == 1 AND r_N^b < x < r_{N+1}^b )
            ;;  OR ( r_N - r_N+1 == 1 AND r_{N+1}^b < x < r_N^b )
            ;;
-           ;; In that case, we fall back to inexact calculation by %expt.
+           ;; In that case, we fall back to inexact calculation by real-expt.
            (let* ([r (ash 1 (round->exact (/ (log x 2) b)))]
-                  [s (%exact-integer-expt r b)])
+                  [s (integer-expt r b)])
              (if (= s x)
-               (%exact-integer-expt r a)
+               (integer-expt r a)
                (let loop ([r r] [s s])
-                 (let* ([deliv (* b (%exact-integer-expt r (- b 1)))]
+                 (let* ([deliv (* b (integer-expt r (- b 1)))]
                         [err   (- s x)]
                         [delta (if (> err 0)
                                  (quotient (+ err deliv -1) deliv)
                                  (quotient (- err deliv -1) deliv))]
                         [r2 (- r delta)]
-                        [s2 (%exact-integer-expt r2 b)])
+                        [s2 (integer-expt r2 b)])
                    (if (= s2 x)
-                     (%exact-integer-expt r2 a)
+                     (integer-expt r2 a)
                      (if (or (and (= delta 1)  (< s2 x s))
                              (and (= delta -1) (< s x s2)))
-                       (%expt x y)
+                       (real-expt x y)
                        (loop r2 s2))))))))]
         [else
          ;; x is rational
@@ -619,47 +680,47 @@
                         [xd (%exact-expt (denominator x) y)]
                         [ (exact? xd) ])
                (/ xn xd))
-             (%expt x y))]))
+             (real-expt x y))]))
 
 (select-module gauche)
 (define-in-module scheme (cos z)
-  (cond [(real? z) (if (eqv? 0 z) 1 (%cos z))]
+  (cond [(real? z) (real-cos z)]
         [(number? z)
          (let ([x (real-part z)]
                [y (imag-part z)])
-           (make-rectangular (* (%cos x) (%cosh y))
-                             (- (* (%sin x) (%sinh y)))))]
+           (make-rectangular (* (real-cos x) (real-cosh y))
+                             (- (* (real-sin x) (real-sinh y)))))]
         [else (error "number required, but got" z)]))
 
 (define (cosh z)
-  (cond [(real? z) (if (eqv? z 0) 1 (%cosh z))]
+  (cond [(real? z) (real-cosh z)]
         [(number? z)
          (let ([x (real-part z)]
                [y (imag-part z)])
-           (make-rectangular (* (%cosh x) (%cos y))
-                             (* (%sinh x) (%sin y))))]
+           (make-rectangular (* (real-cosh x) (real-cos y))
+                             (* (real-sinh x) (real-sin y))))]
         [else (error "number required, but got" z)]))
 
 (define-in-module scheme (sin z)
-  (cond [(real? z) (if (eqv? 0 z) 0 (%sin z))]
+  (cond [(real? z) (real-sin z)]
         [(number? z)
          (let ([x (real-part z)]
                [y (imag-part z)])
-           (make-rectangular (* (%sin x) (%cosh y))
-                             (* (%cos x) (%sinh y))))]
+           (make-rectangular (* (real-sin x) (real-cosh y))
+                             (* (real-cos x) (real-sinh y))))]
         [else (error "number required, but got" z)]))
 
 (define (sinh z)
-  (cond [(real? z) (if (eqv? 0 z) 0 (%sinh z))]
+  (cond [(real? z) (real-sinh z)]
         [(number? z)
          (let ([x (real-part z)]
                [y (imag-part z)])
-           (make-rectangular (* (%sinh x) (%cos y))
-                             (* (%cosh x) (%sin y))))]
+           (make-rectangular (* (real-sinh x) (real-cos y))
+                             (* (real-cosh x) (real-sin y))))]
         [else (error "number required, but got" z)]))
 
 (define-in-module scheme (tan z)
-  (cond [(real? z) (if (eqv? 0 z) 0 (%tan z))]
+  (cond [(real? z) (real-tan z)]
         [(number? z)
          (let1 iz (* +i z)
            (* -i
@@ -668,14 +729,13 @@
         [else (error "number required, but got" z)]))
 
 (define (tanh z)
-  (cond [(real? z) (if (eqv? 0 z) 0 (%tanh z))]
-        [(number? z)
-         (/ (- (exp z) (exp (- z)))
-            (+ (exp z) (exp (- z))))]
+  (cond [(real? z) (real-tanh z)]
+        [(number? z) (/ (- (exp z) (exp (- z)))
+                        (+ (exp z) (exp (- z))))]
         [else (error "number required, but got" z)]))
 
 (define-in-module scheme (asin z)
-  (cond [(real? z) (if (eqv? z 0) 0 (%asin z))]
+  (cond [(real? z) (real-asin z)]
         [(number? z)
          ;; The definition of asin is
          ;;   (* -i (log (+ (* +i z) (sqrt (- 1 (* z z))))))
@@ -692,15 +752,16 @@
         [else (error "number required, but got" z)]))
 
 (define (asinh z)
-  (if (eqv? z 0)
-    0
-    (let1 zz (+ z (sqrt (+ (* z z) 1)))
-      (if (< (/. (magnitude zz) (magnitude z)) 1.0e-8)
-        (make-rectangular +nan.0 +nan.0)
-        (log (+ z (sqrt (+ (* z z) 1))))))))
+  (cond [(real? z) (real-asinh z)]
+        [(number? z)
+         (let1 zz (+ z (sqrt (+ (* z z) 1)))
+           (if (< (/. (magnitude zz) (magnitude z)) 1.0e-8)
+             (make-rectangular +nan.0 +nan.0)
+             (log zz)))]
+        [else (error "number required, but got" z)]))
 
 (define-in-module scheme (acos z)
-  (cond [(real? z) (if (eqv? z 1) 0 (%acos z))]
+  (cond [(real? z) (real-acos z)]
         [(number? z)
          ;; The definition of acos is
          ;;  (* -i (log (+ z (* +i (sqrt (- 1 (* z z)))))))))
@@ -711,36 +772,46 @@
         [else (error "number required, but got" z)]))
 
 (define (acosh z)
-  (if (eqv? z 1)
-    0
-    ;; See the discussion of CLtL2, pp. 313-314
-    (* 2 (log (+ (sqrt (/ (+ z 1) 2))
-                 (sqrt (/ (- z 1) 2)))))))
+  (cond [(real? z) (real-acosh z)]
+        [(number? z);; See the discussion of CLtL2, pp. 313-314
+         (* 2 (log (+ (sqrt (/ (+ z 1) 2))
+                      (sqrt (/ (- z 1) 2)))))]
+        [else (error "number required, but got" z)]))
 
 (define-in-module scheme (atan z . x)
-  (if (eq? z 0)
-    0
-    (if (null? x)
-      (cond [(real? z) (%atan z)]
-            [(number? z)
-             (let1 iz (* z +i)
-               (/ (- (log (+ 1 iz))
-                     (log (- 1 iz)))
-                  +2i))]
-            [else (error "number required, but got" z)])
-      (%atan z (car x)))))
+  (if (null? x)
+    (cond [(real? z) (real-atan z)]
+          [(number? z)
+           (let1 iz (* z +i)
+             (/ (- (log (+ 1 iz))
+                   (log (- 1 iz)))
+                +2i))]
+          [else (error "number required, but got" z)])
+    (cond [(and (eq? (car x) 0) (zero? z))
+           ;; Special case.  See R7RS.
+           (cond [(eq? z 0) +nan.0]
+                 [(negative-zero? z) -1.5707963267948966]
+                 [else +1.5707963267948966])]
+          [(and (real? z) (real? (car x))) (real-atan z (car x))]
+          [else (error "two-argument atan requires real numbers, but got"
+                       (list z (car x)))])))
 
 (define (atanh z)
-  (if (eq? z 0)
-    0
-    (/ (- (log (+ 1 z)) (log (- 1 z))) 2)))
+  (cond [(real? z) (real-atanh z)]
+        [(number? z) (/ (- (log (+ 1 z)) (log (- 1 z))) 2)]
+        [else (error "number required, but got" z)]))
 
+(select-module gauche)
+(define-cproc radians->degrees (r::<double>) ::<double> :constant :fast-flonum
+  (return (* r (/ 180 M_PI))))
+(define-cproc degrees->radians (d::<double>) ::<double> :constant :fast-flonum
+  (return (* d (/ M_PI 180))))
 
 (select-module gauche)
 
-(define-cproc ash (num cnt::<fixnum>) :constant Scm_Ash)
+(define-cproc ash (num cnt::<fixnum>) ::<integer> :constant Scm_Ash)
 
-(define-cproc lognot (x) :constant Scm_LogNot)
+(define-cproc lognot (x::<integer>) ::<integer>  :constant Scm_LogNot)
 
 (inline-stub
  (define-cise-stmt logop
@@ -756,14 +827,14 @@
               (return r))])])
  )
 
-(define-cproc logand (:optarray (arg2 optcnt 2) :rest args) :constant
+(define-cproc logand (:optarray (arg2 optcnt 2) :rest args) ::<integer> :constant
   (logop Scm_LogAnd (SCM_MAKE_INT -1)))
-(define-cproc logior (:optarray (arg2 optcnt 2) :rest args) :constant
+(define-cproc logior (:optarray (arg2 optcnt 2) :rest args) ::<integer> :constant
   (logop Scm_LogIor (SCM_MAKE_INT 0)))
-(define-cproc logxor (:optarray (arg2 optcnt 2) :rest args) :constant
+(define-cproc logxor (:optarray (arg2 optcnt 2) :rest args) ::<integer> :constant
   (logop Scm_LogXor (SCM_MAKE_INT 0)))
 
-(define-cproc logcount (n) ::<int> :constant
+(define-cproc logcount (n::<integer>) ::<int> :constant
   (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (return 0)]
         [(SCM_INTP n)
          (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
@@ -773,29 +844,16 @@
         [(SCM_BIGNUMP n) (return (Scm_BignumLogCount (SCM_BIGNUM n)))]
         [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
 
-(define-cproc integer-length (n) ::<int> :constant
-  (cond [(SCM_INTP n)
-         (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
-           (if (>= (SCM_INT_VALUE n) 0)
-             (return (+ (Scm_BitsHighest1 (& z) 0 SCM_WORD_BITS) 1))
-             (return (+ (Scm_BitsHighest0 (& z) 0 SCM_WORD_BITS) 1))))]
-        [(SCM_BIGNUMP n)
-         ;; 2's complement adjustment.
-         (when (< (SCM_BIGNUM_SIGN n) 0)
-           (set! n (Scm_Add n (SCM_MAKE_INT 1))))
-         ;; The above operation may change n to fixnum, so we check again
-         (if (SCM_BIGNUMP n)
-           (let* ([z::ScmBits* (cast ScmBits* (-> (SCM_BIGNUM n) values))]
-                  [k::int (SCM_BIGNUM_SIZE n)])
-             (return (+ (Scm_BitsHighest1 z 0 (* k SCM_WORD_BITS)) 1)))
-           ;; If n+1 becomes fixnum, we know n was the least fixnum.
-           (return (+ SCM_SMALL_INT_SIZE 1)))]
-        [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
+(define-cproc logset+clear (n::<integer> sets::<integer> clears::<integer>)
+  ::<integer> :constant
+  (return (Scm_LogAnd (Scm_LogIor n sets) (Scm_LogNot clears))))
+
+(define-cproc integer-length (n::<integer>) ::<ulong> :constant Scm_IntegerLength)
 
 ;; Returns maximum s where (expt 2 s) is a factor of n.
 ;; This can be (- (integer-length (logxor n (- n 1))) 1), but we can save
 ;; creating intermediate numbers by providing this natively.
-(define-cproc twos-exponent-factor (n) ::<int> :constant
+(define-cproc twos-exponent-factor (n::<integer>) ::<int> :constant
   (cond [(SCM_EQ n (SCM_MAKE_INT 0)) (return -1)]
         [(SCM_INTP n)
          (let* ([z::ScmBits (cast ScmBits (cast long (SCM_INT_VALUE n)))])
@@ -806,11 +864,9 @@
            (return (Scm_BitsLowest1 z 0 (* k SCM_WORD_BITS))))]
         [else (SCM_TYPE_ERROR n "exact integer") (return 0)]))
 
-(define-cproc twos-exponent (n) :constant
-  (if (SCM_INTEGERP n) ; exact integer only
-    (let* ([i::long (Scm_TwosPower n)])
-      (return (?: (>= i 0) (Scm_MakeInteger i) SCM_FALSE)))
-    (begin (SCM_TYPE_ERROR n "exact integer") (return SCM_FALSE))))
+(define-cproc twos-exponent (n::<integer>) ::<integer>? :constant
+  (let* ([i::long (Scm_TwosPower n)])
+    (return (?: (>= i 0) (Scm_MakeInteger i) SCM_FALSE))))
 
 ;; As of 0.8.8 we started to support exact rational numbers.  Some existing
 ;; code may count on exact integer division to be coerced to flonum
@@ -873,8 +929,10 @@
 ;;
 
 (select-module scheme)
-(define-cproc make-rectangular (a::<real> b::<real>) :constant Scm_MakeComplex)
-(define-cproc make-polar (r::<real> t::<real>) :constant Scm_MakeComplexPolar)
+(define-cproc make-rectangular (a::<double> b::<double>) ::<number> :constant
+  Scm_MakeComplex)
+(define-cproc make-polar (r::<double> t::<double>) ::<number> :constant
+  Scm_MakeComplexPolar)
 
 ;; we don't use Scm_RealPart and Scm_ImagPart, for preserving exactness
 ;; and avoiding extra allocation.
@@ -888,19 +946,19 @@
         [(SCM_REALP z)  (return (Scm_VMReturnFlonum 0.0))]
         [else (return (Scm_VMReturnFlonum (SCM_COMPNUM_IMAG z)))]))
 
-(define-cproc magnitude (z) :fast-flonum :constant Scm_VMAbs)
-(define-cproc angle (z)     ::<double> :fast-flonum :constant Scm_Angle)
+(define-cproc magnitude (z::<number>) ::<number> :fast-flonum :constant Scm_VMAbs)
+(define-cproc angle (z::<number>) ::<double> :fast-flonum :constant Scm_Angle)
 
 ;; Utility to recognize clamp mode symbols and returns C enum
 ;; (Mostly used in uvector-related apis)
 
 (inline-stub
- ;; NB: This is dupe from ext/uvector/uvector.scm
  (define-cfn Scm_ClampMode (clamp) ::int
    (cond [(SCM_EQ clamp 'both) (return SCM_CLAMP_BOTH)]
          [(SCM_EQ clamp 'low)  (return SCM_CLAMP_LO)]
          [(SCM_EQ clamp 'high) (return SCM_CLAMP_HI)]
+         [(SCM_EQ clamp 'wraparound) (return SCM_CLAMP_WRAPAROUND)]
          [(not (or (SCM_FALSEP clamp) (SCM_UNBOUNDP clamp)))
-          (Scm_Error "clamp argument must be either 'both, 'high, 'low or #f, \
-                    but got %S" clamp)])
+          (Scm_Error "clamp argument must be either 'both, 'high, 'low, \
+                      'wraparound or #f, but got %S" clamp)])
    (return SCM_CLAMP_ERROR)))
