@@ -1,7 +1,7 @@
 ;;;
 ;;; object.scm - object system
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -35,12 +35,12 @@
 
 ;; preparing inline stub code
 (inline-stub
- (declcode (.include <gauche/class.h>
-                     <gauche/priv/classP.h>
-                     <gauche/vminsn.h>))
- (define-type <slot-accessor> "ScmSlotAccessor*")
- (define-type <generic> "ScmGeneric*")
- (define-type <method> "ScmMethod*")
+ (.include "gauche/priv/configP.h"
+           "gauche/priv/classP.h"
+           "gauche/vminsn.h")
+ (declare-stub-type <slot-accessor> "ScmSlotAccessor*")
+ (declare-stub-type <generic> "ScmGeneric*")
+ (declare-stub-type <method> "ScmMethod*")
  )
 
 ;; This module is not meant to be `use'd.   It is just to hide
@@ -85,9 +85,7 @@
 ;; Generic function
 ;;
 
-;(define-macro (define-generic name . opts)
-;  (%expand-define-generic name opts))
-
+;; Called from define-generic macro expander in libmacro.scm
 (define (%expand-define-generic name opts)
   (receive (true-name getter-name) (%check-setter-name name)
     (let ([class (get-keyword :class opts <generic>)]
@@ -123,10 +121,8 @@
 ;; Method
 ;;
 
-;(define-macro (define-method name [quals ...] specs . body)
-;  (%expand-define-macro name quals specs body))
-
-(define (%expand-define-method name quals specs body)
+;; Called from define-method macro expander in libmacro.scm
+(define (%expand-define-method name quals specs body form)
   ;; check qualifiers.  we only support :locked for now
   ;; TRANSIENT: Must use hygienic compare.
   (if-let1 bad (find (^q (not (memq q '(:locked)))) quals)
@@ -149,26 +145,39 @@
            [real-args    (if rest
                            `(,@reqargs ,rest next-method)
                            `(,@reqargs next-method))]
-           [real-body (if opts
-                        (quasirename %id
-                          `(lambda ,real-args
-                             (apply (lambda ,opts ,@body) ,rest)))
-                        (quasirename %id
-                          `(lambda ,real-args ,@body)))])
+           [real-body ((with-module gauche.internal with-source-info)
+                       (if opts
+                         (quasirename %id
+                           `(lambda ,real-args
+                              (apply (lambda ,opts ,@body) ,rest)))
+                         (quasirename %id
+                           `(lambda ,real-args ,@body)))
+                       form)]
+           ;; We bind method body to an artificially weird module-local name,
+           ;; so that test-module can find and scan the method's body.
+           [method-unique-name (gensym
+                                (symbol->string
+                                 (symbol-append (unwrap-syntax name)
+                                                (unwrap-syntax specs)
+                                                '|.|)))]
+           )
       (receive (true-name getter-name) (%check-setter-name name)
         (let1 gf (gensym "gf")
           (quasirename %id
-            `(rlet1 ,gf (%ensure-generic-function ',true-name (current-module))
-               (add-method! ,gf (make <method>
-                                  ':generic ,gf
-                                  ':specializers (list ,@specializers)
-                                  ':lambda-list ',lambda-list
-                                  ':method-locked (boolean (memq ':locked ',quals))
-                                  ':body ,real-body))
-               ,@(cond-list [getter-name
-                             (quasirename %id
-                               `(unless (has-setter? ,getter-name)
-                                  (set! (setter ,getter-name) ,gf)))]))))))))
+            `(define ,method-unique-name
+               (rlet1 ,method-unique-name ,real-body
+                 (let1 ,gf (%ensure-generic-function ',true-name (current-module))
+                   (add-method! ,gf (make <method>
+                                      ':generic ,gf
+                                      ':specializers (list ,@specializers)
+                                      ':lambda-list ',lambda-list
+                                      ':method-locked (boolean (memq ':locked ',quals))
+                                      ':body ,method-unique-name))
+                   ,@(cond-list [getter-name
+                                 (quasirename %id
+                                   `(unless (has-setter? ,getter-name)
+                                      (set! (setter ,getter-name) ,gf)))]))))
+            ))))))
 
 (inline-stub
  ;; internal for %ensure-generic-function
@@ -190,7 +199,7 @@
    (let* ([val (Scm_GlobalVariableRef module name 0)])
      (cond
       [(Scm_TypeP val SCM_CLASS_GENERIC)
-       (when (SCM_UNBOUNDP (Scm_GlobalVariableRef module name 
+       (when (SCM_UNBOUNDP (Scm_GlobalVariableRef module name
                                                   SCM_BINDING_STAY_IN_MODULE))
          (Scm_Define module name val))]
       [else
@@ -230,15 +239,7 @@
           [(eqv? key (unwrap-syntax-1 (car as))) (cadr as)]
           [else (loop (cddr as))])))
 
-;; TRANSIENT: We should employ er-macro-transformer to expand
-;; define-class etc., but this file need to be compiled by 0.9.5
-;; and we've improved the transformer since then.  So we postponed
-;; rewriting these after 0.9.6 release.  Meanwhile, we manually
-;; replacing identifiers.
-
-;(define-macro (define-class name supers slots . options)
-;  (%expand-define-class name supers slots options))
-
+;; Called from define-class macro expander in libmacro.scm
 (define (%expand-define-class name supers slots options)
   (let* ([metaclass (or (%get-keyword :metaclass options #f)
                         (quasirename %id
@@ -247,7 +248,12 @@
          [class     (gensym)]
          [slot      (gensym)])
     (quasirename %id
-      `(define ,name
+      ;; We expand this into define-inline, to make the binding inlinable.
+      ;; It allows the compiler to use the information of the class
+      ;; reference.  Note that class redefinition does override the
+      ;; inlinable binding (which should be rare, though).  We handle
+      ;; redefinition of class specially in gloc.c.
+      `(define-inline ,name
          (rlet1 ,class (make ,metaclass
                          ':name ',name ':supers (list ,@supers)
                          ':slots (list ,@slot-defs)
@@ -494,11 +500,23 @@
 (define (class-slot-bound? class slot-name)
   (apply (%class-slot-gns class slot-name 'bound?) '(#f)))
 
-;; default class printer.  Avoid using class-name so that in case
-;; when obj's class has been redefined, this wouldn't trigger updating obj.
+;; Default class printer.  This is only called if the class is an instance
+;; of a metaclass other than <class>.  (If it's <class>, built-in print
+;; method is called; see class.c)
+;; Avoid using directly access the slots of obj, so that it won't trigger
+;; updating obj in case when obj's class has been redefined.
+;; We also need a cproc to check if OBJ is redefined; for, accessing
+;; 'redefined' slot needs to lock OBJ, which blocks when OBJ is being
+;; redefined.
 (define-method write-object ((obj <class>) out)
-  (format out "#<class ~a>"
-          (slot-ref-using-class (current-class-of obj) obj 'name)))
+  (let1 name (slot-ref-using-class (current-class-of obj) obj 'name)
+    (if (%class-redefined? obj)
+      (format out "#<class ~a (redefined)>" name)
+      (format out "#<class ~a>" name))))
+
+(define-cproc %class-redefined? (obj) ::<boolean>
+  (return (and (SCM_ISA obj SCM_CLASS_CLASS)
+               (not (SCM_FALSEP (-> (cast ScmClass* obj) redefined))))))
 
 ;; convenient routine to push/pop a value to the slot.
 ;; this can be optimized later.
@@ -535,8 +553,11 @@
 ;; which has been redefined.  Should only be used in class redefinition
 ;; routines.
 (define-cproc current-class-of (obj) (return (SCM_OBJ (Scm_ClassOf obj))))
+
 (define-cproc is-a? (obj klass::<class>) (inliner IS-A) Scm_VMIsA)
-(define-cproc subtype? (c1::<class> c2::<class>) ::<boolean> Scm_SubtypeP)
+(define-cproc subclass? (c1::<class> c2::<class>) ::<boolean> Scm_SubclassP)
+
+;; NB: subtype? and of-type? is in libtype.scm.
 
 (define-cproc slot-ref (obj slot)
   (inliner SLOT-REF) (setter slot-set!)
@@ -561,7 +582,7 @@
 
 ;; Internal API - undocumented
 (define-cproc instance-slot-ref (obj num::<fixnum> :optional fallback)
-  (.if "GAUCHE_API_VERSION < 1000"
+  (.if "GAUCHE_API_VERSION < 98"
        (return (Scm_InstanceSlotRef3 obj num fallback))
        (return (Scm_InstanceSlotRef obj num fallback))))
 ;; Internal API - undocumented
@@ -591,7 +612,7 @@
 (define-cproc %record-ref (klass::<class> obj k::<fixnum>)
   (unless (SCM_ISA obj klass)
     (Scm_Error "record-ref: instance of %S expected, got %S" klass obj))
-  (.if "GAUCHE_API_VERSION < 1000"
+  (.if "GAUCHE_API_VERSION < 98"
        (return (Scm_InstanceSlotRef3 obj k SCM_UNBOUND))
        (return (Scm_InstanceSlotRef obj k SCM_UNBOUND))))
 
@@ -774,16 +795,25 @@
 (define-method (setter ref) :locked ((obj <tree-map>) key value)
   (tree-map-put! obj key value))
 
-(define-method ref  ((obj <box>) field) ; srfi-123
-  (unless (eq? field '*)
-    (error "Box content can be accessed with a field name '*, but got:"
-           field))
-  (unbox obj))
+(define-method ref :locked ((obj <box>) field)
+  (case field
+    [(* 0) (unbox obj)]                 ;'* for srfi-123
+    [else
+     (error "Bad field name for box reference. Exact integer or '*, but got:"
+            field)]))
 (define-method (setter ref) :locked ((obj <box>) field value)
-  (unless (eq? field '*)
-    (error "Box content can be accessed with a field name '*, but got:"
-           field))
-  (set-box! obj value))
+  (case field
+    [(* 0) (set-box! obj value)]        ;'* for srfi-123
+    [else
+     (error "Bad field name for box reference. Exact integer or '*, but got:"
+            field)]))
+
+(define-method ref :locked ((obj <mv-box>) field)
+  (assume (exact-integer? field))
+  (unbox-value obj field))
+(define-method (setter ref) :locked ((obj <mv-box>) field value)
+  (assume (exact-integer? field))
+  (set-box-value! obj field value))
 
 ;; gauche.sequence has the generic version for <sequence>, but these
 ;; shortcuts would be faster.
@@ -872,7 +902,7 @@
           (cond [(char=? #\/ c) (display #\\ out) (display c out)]
                 [(memq (char-general-category c) '(Mn Mc Me Cc Cf Cs Co Cn))
                  (let1 code (char->ucs c)
-                   (if (< code #x10000)  
+                   (if (< code #x10000)
                      (format out "\\u~4,'0x" code)
                      (format out "\\U~8,'0x" code)))]
                 [else (display c out)])
@@ -886,7 +916,10 @@
             :base   ,(~ obj'base)
             :radix  ,(~ obj'radix)
             :pretty ,(~ obj'pretty)
-            :width  ,(~ obj'width))))
+            :width  ,(~ obj'width)
+            :indent ,(~ obj'indent)
+            :string-length ,(~ obj'string-length)
+            :exact-decimal ,(~ obj'exact-decimal))))
 
 ;;----------------------------------------------------------------
 ;; Describe
@@ -898,8 +931,22 @@
 ;; methods here.  Other specialized methods on built-in objects are
 ;; still defined in gauche.interactive.
 
+;; Setting this dynamic state to true makes describe to show more details,
+;; which may be too much for casual use.  (We don't use make-parameter,
+;; for parameter system is not initialized at this point.)
+(define describe-details
+  (let1 flag #f
+    (case-lambda
+      [() flag]
+      [(val)
+       (assume-type val <boolean>
+                    "Value of describe-details must be a boolean, but got:"
+                    val)
+       (begin0 flag (set! flag val))])))
+
 (define (describe-common obj)
-  (format #t "~s is an instance of class ~a\n" obj (class-name (class-of obj))))
+  (format #t "~s is an instance of class ~a\n" obj (class-name (class-of obj)))
+  (values))
 
 (define-method describe-slots (obj)
   (let* ([class (class-of obj)]
@@ -907,11 +954,14 @@
     (unless (null? slots)
       (format #t "slots:\n")
       (dolist [s (map slot-definition-name slots)]
-        (format #t "  ~10s: ~a\n" s
-                (if (slot-bound? obj s)
-                  (with-output-to-string
-                    (^[] (write-limited (slot-ref obj s) 60)))
-                  "#<unbound>"))))))
+        (when (or (describe-details)
+                  (not (#/^%/ (symbol->string s))))
+          (format #t "  ~10s: ~a\n" s
+                  (if (slot-bound? obj s)
+                    (with-output-to-string
+                      (^[] (write-limited (slot-ref obj s) 60)))
+                    "#<unbound>"))))))
+  (values))
 
 (define-method describe (object) ; default
   (describe-common object)
@@ -922,40 +972,32 @@
 ;;; Make exported symbol visible from outside
 ;;;
 
-(define-macro (insert-symbols . syms)
-  `(begin ,@(map (^[s] `(define-in-module gauche ,s ,s)) syms)))
-
-(insert-symbols ;define-generic define-method define-class
-                compute-slots compute-get-n-set compute-slot-accessor
-                class-slot-ref class-slot-set! class-slot-bound?
-                slot-push! slot-pop! slot-unbound slot-missing
-                slot-exists? slot-exists-using-class?
-                change-class
-                apply-generic sort-applicable-methods
-                apply-methods apply-method
-                class-of current-class-of is-a? subtype? slot-ref slot-set!
-                slot-bound? slot-ref-using-accessor slot-bound-using-accessor?
-                slot-set-using-accessor! slot-initialize-using-accessor!
-                instance-slot-ref instance-slot-set! touch-instance!
-                class-name class-precedence-list class-direct-supers
-                class-direct-methods class-direct-subclasses
-                class-direct-slots class-slots
-                class-post-initialize
-                slot-definition-name slot-definition-options
-                slot-definition-option
-                slot-definition-allocation slot-definition-getter
-                slot-definition-setter slot-definition-accessor
-                class-slot-definition class-slot-accessor
-                x->string x->integer x->number ref |setter of ref|
-                ~ ref*
-                describe describe-common describe-slots
-
-                ;; These shouldn't be necessary to be injected into gauche
-                ;; module; unfortunately, the current define-method and
-                ;; define-class are unhygienic, and we need them visible
-                ;; from the expanded code.  Should be removed once we rewrite
-                ;; related macros hygienic.
-                %ensure-generic-function
-                %check-class-binding
-                )
-
+(let ((xfer (with-module gauche.internal %transfer-bindings)))
+  (xfer (current-module)
+        (find-module 'gauche)
+        '(compute-slots
+          compute-get-n-set compute-slot-accessor
+          class-slot-ref class-slot-set! class-slot-bound?
+          slot-push! slot-pop! slot-unbound slot-missing
+          slot-exists? slot-exists-using-class?
+          change-class
+          apply-generic sort-applicable-methods
+          apply-methods apply-method
+          class-of current-class-of is-a? subclass?
+          slot-ref slot-set!
+          slot-bound? slot-ref-using-accessor slot-bound-using-accessor?
+          slot-set-using-accessor! slot-initialize-using-accessor!
+          instance-slot-ref instance-slot-set! touch-instance!
+          class-name class-precedence-list class-direct-supers
+          class-direct-methods class-direct-subclasses
+          class-direct-slots class-slots
+          class-post-initialize
+          slot-definition-name slot-definition-options
+          slot-definition-option
+          slot-definition-allocation slot-definition-getter
+          slot-definition-setter slot-definition-accessor
+          class-slot-definition class-slot-accessor
+          x->string x->integer x->number ref |setter of ref|
+          ~ ref*
+          describe describe-common describe-slots describe-details
+          )))
