@@ -1,7 +1,7 @@
 ;;;
 ;;;libstr.scm - built-in string library
 ;;;
-;;;   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -34,9 +34,19 @@
 (select-module gauche.internal)
 
 (inline-stub
- (declcode (.include <gauche/bignum.h>
-                     <gauche/vminsn.h>
-                     <gauche/priv/stringP.h>)))
+ (.include "gauche/priv/configP.h"
+           "gauche/vminsn.h"
+           "gauche/priv/bignumP.h"
+           "gauche/priv/stringP.h"
+           "gauche/priv/writerP.h"))
+
+;; Ellipsis, used when output is truncated.
+(select-module gauche.internal)
+(define-cproc string-ellipsis ()
+  (let* ([elli::(static ScmObj) SCM_UNDEFINED])
+    (when (SCM_UNDEFINEDP elli)
+      (set! elli (SCM_MAKE_STR_IMMUTABLE SCM_WRITTEN_ELLIPSIS)))
+    (return elli)))
 
 ;;
 ;; Predicates
@@ -117,10 +127,10 @@
 (define-cproc string-size (str::<string>) ::<fixnum> :constant
   (return (SCM_STRING_BODY_SIZE (SCM_STRING_BODY str))))
 
-(define-cproc %maybe-substring (str::<string> :optional start end)
+(define-cproc opt-substring (str::<string> :optional start end)
   Scm_MaybeSubstring)
 
-;; bound argument is for srfi-13
+;; bound argument is for SRFI-13
 (define-cproc %hash-string (str::<string> :optional bound) ::<ulong>
   (let* ([modulo::u_long 0])
     (cond [(or (SCM_UNBOUNDP bound) (SCM_UNDEFINEDP bound))
@@ -172,8 +182,8 @@
 ;; string-split
 ;; Generic string-split
 ;;  splitter can be a character, a char-set, a string, or a regexp.
-;;  NB: To adapt to srfi-152, we changed the optional argument - now the
-;;  'grammar' argument comes before 'limit'.  For the bacward compatibility,
+;;  NB: To adapt to SRFI-152, we changed the optional argument - now the
+;;  'grammar' argument comes before 'limit'.  For the backward compatibility,
 ;;  we recognize an integer argument in 'grammar' as 'limit'.
 (define (string-split string splitter
                       :optional (grammar 'infix) (limit #f) start end)
@@ -193,7 +203,7 @@
     (error "grammar argument must be one of (infix strict-infix prefix suffix), but got" grammar))
   (unless (or (not limit) (and (integer? limit) (>= limit 0)))
     (error "limit argument must be a nonnegative integer or #f, but got" limit))
-  (let1 s ((with-module gauche.internal %maybe-substring) string start end)
+  (let1 s (opt-substring string start end)
     (if (equal? s "")
       (if (eq? grammar 'strict-infix)
         (error "string must not be empty with strict-infix grammar")
@@ -396,7 +406,7 @@
 
 ;; We have two kinds of string-{map|for-each}.  SRFI-13 takes one string
 ;; with optional start/end arguments.  R7RS takes one or more string arguments.
-;; In retrospect, R7RS is more consistent, but srfi-13 is very frequently
+;; In retrospect, R7RS is more consistent, but SRFI-13 is very frequently
 ;; imported and it is bothersome to avoid name conflicts.  So we have
 ;; single implementation that handles both.
 
@@ -406,14 +416,12 @@
         [(or (integer? (car rest))
              (string-cursor? (car rest)))
          (if (null? (cdr rest))
-           (proc-single proc (%maybe-substring str (car rest) (undefined)))
+           (proc-single proc (opt-substring str (car rest) (undefined)))
            (if (or (integer? (cadr rest))
                    (string-cursor? (cadr rest)))
              (if (null? (cddr rest))
-               (proc-single proc (%maybe-substring str
-                                                   (car rest)
-                                                   (cadr rest)))
-               (errorf "Too many arguments for srfi-13 style ~a" name))
+               (proc-single proc (opt-substring str (car rest) (cadr rest)))
+               (errorf "Too many arguments for SRFI-13 style ~a" name))
              (error "Integer or string-cursor expected, but got:" (cadr rest))))]
         [(string? (car rest))
          (if-let1 bad (find (^s (not (string? s))) (cdr rest))
@@ -480,8 +488,8 @@
     s))
 
 (define-cproc string-complete->incomplete (str::<string>)
-  (return (Scm_CopyStringWithFlags str 
-                                   SCM_STRING_INCOMPLETE 
+  (return (Scm_CopyStringWithFlags str
+                                   SCM_STRING_INCOMPLETE
                                    SCM_STRING_INCOMPLETE)))
 
 ;; handling : #f | :omit | :replace | :escape
@@ -596,6 +604,7 @@
                          [(eqv? c2 #\\)
                           (and (not (eof-object? (read-char)))
                                (begin (skip-token) (rec closer)))]
+                         [(eqv? c2 #\") (and (rec-escaped #\") (rec closer))]
                          [(eqv? c2 #\/) (and (rec-escaped #\/) (rec closer))]
                          [(eqv? c2 #\[) (and (rec-escaped #\]) (rec closer))]
                          [(eqv? c2 #\,)
@@ -644,50 +653,6 @@
       (rec #f)
       )))
 
-;;
-;; String pointers (OBSOLETED)
-;;
-
-(select-module gauche)
-
-(inline-stub
-(.if "GAUCHE_STRING_POINTER"
-     (begin
-       ;; string pointer
-       (define-type <string-pointer> "ScmStringPointer*" "string pointer"
-         "SCM_STRING_POINTERP" "SCM_STRING_POINTER")
-
-       (define-cproc make-string-pointer (str::<string>
-                                          :optional (index::<fixnum> 0)
-                                          (start::<fixnum> 0)
-                                          (end::<fixnum> -1))
-         Scm_MakeStringPointer)
-       (define-cproc string-pointer? (obj) ::<boolean> SCM_STRING_POINTERP)
-
-       (define-cproc string-pointer-ref (sp::<string-pointer>)
-         Scm_StringPointerRef)
-       (define-cproc string-pointer-next! (sp::<string-pointer>)
-         Scm_StringPointerNext)
-       (define-cproc string-pointer-prev! (sp::<string-pointer>)
-         Scm_StringPointerPrev)
-       (define-cproc string-pointer-set! (sp::<string-pointer> index::<fixnum>)
-         Scm_StringPointerSet)
-       (define-cproc string-pointer-substring (sp::<string-pointer> :key (after #f))
-         (return (Scm_StringPointerSubstring sp (not (SCM_FALSEP after)))))
-       (define-cproc string-pointer-index (sp::<string-pointer>) ::<int>
-         (return (-> sp index)))
-       (define-cproc string-pointer-copy (sp::<string-pointer>)
-         Scm_StringPointerCopy)
-       (define-cproc string-pointer-byte-index (sp::<string-pointer>) ::<int>
-         (return (cast int (- (-> sp current) (-> sp start))))))))
-
-(select-module gauche.internal)
-(inline-stub
-(.if "GAUCHE_STRING_POINTER"
-       (define-cproc %string-pointer-dump (sp::<string-pointer>) ::<void>
-         Scm_StringPointerDump)
-       ))
-
 ;;
 ;; String cursors
 ;;
