@@ -1,12 +1,12 @@
 /*
  * list.c - List related functions
  *
- *   Copyright (c) 2000-2009  Shiro Kawai  <shiro@acm.org>
- * 
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
+ *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
  *   are met:
- * 
+ *
  *   1. Redistributions of source code must retain the above copyright
  *      notice, this list of conditions and the following disclaimer.
  *
@@ -29,12 +29,11 @@
  *   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- *
- *  $Id: list.c,v 1.48 2008-05-10 13:36:19 shirok Exp $
  */
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/pairP.h"
 
 /*
  * Classes
@@ -65,8 +64,8 @@ ScmObj Scm_Cons(ScmObj car, ScmObj cdr)
        the register. */
     SCM_FLONUM_ENSURE_MEM(car);
     SCM_FLONUM_ENSURE_MEM(cdr);
-    SCM_SET_CAR(z, car);
-    SCM_SET_CDR(z, cdr);
+    SCM_SET_CAR_UNCHECKED(z, car);
+    SCM_SET_CDR_UNCHECKED(z, cdr);
     return SCM_OBJ(z);
 }
 
@@ -74,22 +73,20 @@ ScmObj Scm_Acons(ScmObj caar, ScmObj cdar, ScmObj cdr)
 {
     ScmPair *y = SCM_NEW(ScmPair);
     ScmPair *z = SCM_NEW(ScmPair);
-    SCM_SET_CAR(y, caar);
-    SCM_SET_CDR(y, cdar);
-    SCM_SET_CAR(z, SCM_OBJ(y));
-    SCM_SET_CDR(z, cdr);
+    SCM_SET_CAR_UNCHECKED(y, caar);
+    SCM_SET_CDR_UNCHECKED(y, cdar);
+    SCM_SET_CAR_UNCHECKED(z, SCM_OBJ(y));
+    SCM_SET_CDR_UNCHECKED(z, cdr);
     return SCM_OBJ(z);
 }
 
 ScmObj Scm_List(ScmObj elt, ...)
 {
-    va_list pvar;
-    ScmObj cdr;
-
     if (elt == NULL) return SCM_NIL;
-        
+
+    va_list pvar;
     va_start(pvar, elt);
-    cdr = Scm_VaList(pvar);
+    ScmObj cdr = Scm_VaList(pvar);
     va_end(pvar);
     return Scm_Cons(elt, cdr);
 }
@@ -97,13 +94,11 @@ ScmObj Scm_List(ScmObj elt, ...)
 
 ScmObj Scm_Conses(ScmObj elt, ...)
 {
-    va_list pvar;
-    ScmObj cdr;
-
     if (elt == NULL) return SCM_NIL;
-    
+
+    va_list pvar;
     va_start(pvar, elt);
-    cdr = Scm_VaCons(pvar);
+    ScmObj cdr = Scm_VaCons(pvar);
     va_end(pvar);
     if (cdr == NULL) return elt;
     else             return Scm_Cons(elt, cdr);
@@ -113,56 +108,56 @@ ScmObj Scm_Conses(ScmObj elt, ...)
 ScmObj Scm_VaList(va_list pvar)
 {
     ScmObj start = SCM_NIL, cp = SCM_NIL, obj;
-    
+
     for (obj = va_arg(pvar, ScmObj);
-	 obj != NULL;
-	 obj = va_arg(pvar, ScmObj))
+         obj != NULL;
+         obj = va_arg(pvar, ScmObj))
     {
-	if (SCM_NULLP(start)) {
+        if (SCM_NULLP(start)) {
             start = SCM_OBJ(SCM_NEW(ScmPair));
-            SCM_SET_CAR(start, obj);
-            SCM_SET_CDR(start, SCM_NIL);
+            SCM_SET_CAR_UNCHECKED(start, obj);
+            SCM_SET_CDR_UNCHECKED(start, SCM_NIL);
             cp = start;
         } else {
             ScmObj item;
             item = SCM_OBJ(SCM_NEW(ScmPair));
-            SCM_SET_CDR(cp, item);
-            SCM_SET_CAR(item, obj);
-            SCM_SET_CDR(item, SCM_NIL);
+            SCM_SET_CDR_UNCHECKED(cp, item);
+            SCM_SET_CAR_UNCHECKED(item, obj);
+            SCM_SET_CDR_UNCHECKED(item, SCM_NIL);
             cp = item;
-	}
+        }
     }
     return start;
 }
 
 
-ScmObj Scm_VaCons(va_list pvar)
+ScmObj Scm_VaCons(va_list pvar SCM_UNUSED)
 {
     Scm_Panic("Scm_VaCons: not implemented");
     return SCM_UNDEFINED;
 }
 
-ScmObj Scm_ArrayToList(ScmObj *elts, int nelts)
+ScmObj Scm_ArrayToList(ScmObj *elts, ScmSize nelts)
 {
     return Scm_ArrayToListWithTail(elts, nelts, SCM_NIL);
 }
 
-ScmObj Scm_ArrayToListWithTail(ScmObj *elts, int nelts, ScmObj tail)
+ScmObj Scm_ArrayToListWithTail(ScmObj *elts, ScmSize nelts, ScmObj tail)
 {
     ScmObj h = SCM_NIL, t = SCM_NIL;
     if (elts) {
-        int i;
-        for (i=0; i<nelts; i++) SCM_APPEND1(h, t, *elts++);
+        for (ScmSize i=0; i<nelts; i++) SCM_APPEND1(h, t, *elts++);
     }
     if (!SCM_NULLP(tail)) SCM_APPEND(h, t, tail);
     return h;
 }
 
-ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
+ScmObj *Scm_ListToArray(ScmObj list, ScmSize *nelts, ScmObj *store, int alloc)
 {
-    ScmObj *array, lp;
-    int len = Scm_Length(list), i;
+    ScmSize len = Scm_Length(list);
     if (len < 0) Scm_Error("proper list required, but got %S", list);
+
+    ScmObj *array;
     if (store == NULL) {
         array = SCM_NEW_ARRAY(ScmObj, len);
     } else {
@@ -174,7 +169,8 @@ ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
             array = store;
         }
     }
-    for (i=0, lp=list; i<len; i++, lp=SCM_CDR(lp)) {
+    ScmSize i = 0;
+    for (ScmObj lp=list; i<len; i++, lp=SCM_CDR(lp)) {
         array[i] = SCM_CAR(lp);
     }
     *nelts = len;
@@ -183,7 +179,7 @@ ScmObj *Scm_ListToArray(ScmObj list, int *nelts, ScmObj *store, int alloc)
 
 /* cXr stuff */
 
-#define	CXR(cname, sname, body)                 \
+#define CXR(cname, sname, body)                 \
 ScmObj cname (ScmObj obj)                       \
 {                                               \
    ScmObj obj2 = obj;                           \
@@ -191,11 +187,11 @@ ScmObj cname (ScmObj obj)                       \
    return obj2;                                 \
 }
 
-#define	A                                                       \
+#define A                                                       \
    if (!SCM_PAIRP(obj2)) Scm_Error("bad object: %S", obj);      \
    obj2 = SCM_CAR(obj2);
 
-#define	D                                                       \
+#define D                                                       \
    if (!SCM_PAIRP(obj2)) Scm_Error("bad object: %S", obj);      \
    obj2 = SCM_CDR(obj2);
 
@@ -206,6 +202,44 @@ CXR(Scm_Cadr, "cadr", D A)
 CXR(Scm_Cdar, "cdar", A D)
 CXR(Scm_Cddr, "cddr", D D)
 
+void Scm_SetCar(ScmObj pair, ScmObj obj)
+{
+    if (!SCM_PAIRP(pair)) {
+        Scm_Error("set-car!: Pair required, but got: %S", pair);
+    }
+    ScmExtendedPairDescriptor *d = Scm__GetExtendedPairDescriptor(pair);
+    if (d) {
+        if (d->flags & SCM_PAIR_IMMUTABLE) {
+            Scm_Error("attempt to mutate car of an immutable pair %S with %S",
+                      pair, obj);
+        }
+        if (d->setCar) {
+            d->setCar(pair, obj);
+            return;
+        }
+    }
+    SCM_CAR(pair) = obj;
+}
+
+void Scm_SetCdr(ScmObj pair, ScmObj obj)
+{
+    if (!SCM_PAIRP(pair)) {
+        Scm_Error("set-cdr!: Pair required, but got: %S", pair);
+    }
+    ScmExtendedPairDescriptor *d = Scm__GetExtendedPairDescriptor(pair);
+    if (d) {
+        if (d->flags & SCM_PAIR_IMMUTABLE) {
+            Scm_Error("attempt to mutate cdr of an immutable pair %S with %S",
+                      pair, obj);
+        }
+        if (d->setCdr) {
+            d->setCdr(pair, obj);
+            return;
+        }
+    }
+    SCM_CDR(pair) = obj;
+}
+
 /*
  * List manipulate routines:
  */
@@ -215,24 +249,24 @@ CXR(Scm_Cddr, "cddr", D D)
    If the argument is a dotted list, return -1.
    If the argument is a circular list, return -2. */
 
-int Scm_Length(ScmObj obj)
+ScmSize Scm_Length(ScmObj obj)
 {
     ScmObj slow = obj;
-    int len = 0;
+    ScmSize len = 0;
 
     for (;;) {
         if (SCM_NULLP(obj)) break;
         if (!SCM_PAIRP(obj)) return SCM_LIST_DOTTED;
-	
-	obj = SCM_CDR(obj);
-	len++;
+
+        obj = SCM_CDR(obj);
+        len++;
         if (SCM_NULLP(obj)) break;
         if (!SCM_PAIRP(obj)) return SCM_LIST_DOTTED;
 
-	obj = SCM_CDR(obj);
-	slow = SCM_CDR(slow);
-	if (obj == slow) return SCM_LIST_CIRCULAR;
-	len++;
+        obj = SCM_CDR(obj);
+        slow = SCM_CDR(slow);
+        if (obj == slow) return SCM_LIST_CIRCULAR;
+        len++;
     }
     return len;
 }
@@ -240,39 +274,90 @@ int Scm_Length(ScmObj obj)
 /* Scm_CopyList(list)
  *   Copy toplevel list LIST.  LIST can be improper.
  *   If LIST is not a pair, return LIST itself.
+ *   Note: For the naming consistency, this should be Scm_ListCopy().
  */
 
 ScmObj Scm_CopyList(ScmObj list)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
-
     if (!SCM_PAIRP(list)) return list;
-    
-    SCM_FOR_EACH(list, list) {
-        SCM_APPEND1(start, last, SCM_CAR(list));
+    ScmObj tortoise = list, hare = list;
+
+    ScmObj start = SCM_NIL, last = SCM_NIL;
+    for (;;) {
+        if (!SCM_PAIRP(hare)) break;
+        SCM_APPEND1(start, last, SCM_CAR(hare));
+        hare = SCM_CDR(hare);
+
+        if (!SCM_PAIRP(hare)) break;
+        SCM_APPEND1(start, last, SCM_CAR(hare));
+        hare = SCM_CDR(hare);
+        tortoise = SCM_CDR(tortoise);
+        if (hare == tortoise) {
+            Scm_Error("Attempt to copy a circular list: %S", list);
+        }
     }
-    if (!SCM_NULLP(list)) SCM_SET_CDR(last, list);
+    if (!SCM_NULLP(hare)) SCM_SET_CDR_UNCHECKED(last, hare);
     return start;
 }
 
+/* Scm_AlistCopy(alist)
+ *   Copy the spine and each pair of alist.  If there's non-pair element
+ *   it's included as is.  We detect a cycle.
+ *   NB: We won't allow 'dotted' alist.
+ */
+ScmObj Scm_AlistCopy(ScmObj alist)
+{
+    if (!SCM_PAIRP(alist)) return alist;
+    ScmObj tortoise = alist, hare = alist;
+
+    ScmObj start = SCM_NIL, last = SCM_NIL;
+    for (;;) {
+        if (!SCM_PAIRP(hare)) break;
+        if (SCM_PAIRP(SCM_CAR(hare))) {
+            ScmObj p = Scm_Cons(SCM_CAAR(hare), SCM_CDAR(hare));
+            SCM_APPEND1(start, last, p);
+        } else {
+            SCM_APPEND1(start, last, SCM_CAR(hare));
+        }
+        hare = SCM_CDR(hare);
+
+        if (!SCM_PAIRP(hare)) break;
+        if (SCM_PAIRP(SCM_CAR(hare))) {
+            ScmObj p = Scm_Cons(SCM_CAAR(hare), SCM_CDAR(hare));
+            SCM_APPEND1(start, last, p);
+        } else {
+            SCM_APPEND1(start, last, SCM_CAR(hare));
+        }
+        hare = SCM_CDR(hare);
+
+        tortoise = SCM_CDR(tortoise);
+        if (hare == tortoise) {
+            Scm_Error("Attempt to copy a circular alist: %S", alist);
+        }
+    }
+    if (!SCM_NULLP(hare)) Scm_Error("Dotted alist isn't allowed: %S", alist);
+    return start;
+}
+
+
 /* Scm_MakeList(len, fill)
  *    Make a list of specified length.
  *    Note that <len> is C-integer.
  */
 
-ScmObj Scm_MakeList(int len, ScmObj fill)
+ScmObj Scm_MakeList(ScmSmallInt len, ScmObj fill)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
     if (len < 0) {
-        Scm_Error("make-list: negative length given: %d", len);
+        Scm_Error("make-list: negative length given: %ld", len);
     }
+    ScmObj start = SCM_NIL, last = SCM_NIL;
     while (len--) {
         SCM_APPEND1(start, last, fill);
     }
     return start;
 }
-     
-    
+
+
 /* Scm_Append2X(list, obj)
  *    Replace cdr of last pair of LIST for OBJ.
  *    If LIST is not a pair, return OBJ.
@@ -282,8 +367,8 @@ ScmObj Scm_Append2X(ScmObj list, ScmObj obj)
 {
     ScmObj cp;
     SCM_FOR_EACH(cp, list) {
-	if (SCM_NULLP(SCM_CDR(cp))) {
-            SCM_SET_CDR(cp, obj);
+        if (SCM_NULLP(SCM_CDR(cp))) {
+            Scm_SetCdr(cp, obj);
             return list;
         }
     }
@@ -297,14 +382,16 @@ ScmObj Scm_Append2X(ScmObj list, ScmObj obj)
 
 ScmObj Scm_Append2(ScmObj list, ScmObj obj)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL;
+    if (SCM_NULLP(list)) return obj;
 
-    if (!SCM_PAIRP(list)) return obj;
-
-    SCM_FOR_EACH(list, list) {
-	SCM_APPEND1(start, last, SCM_CAR(list));
+    ScmObj cp, start = SCM_NIL, last = SCM_NIL;
+    SCM_FOR_EACH(cp, list) {
+        SCM_APPEND1(start, last, SCM_CAR(cp));
     }
-    SCM_SET_CDR(last, obj);
+    if (!SCM_NULLP(cp)) {
+        Scm_Error("proper list required, but got %S", list);
+    }
+    SCM_SET_CDR_UNCHECKED(last, obj);
 
     return start;
 }
@@ -315,7 +402,7 @@ ScmObj Scm_Append(ScmObj args)
     SCM_FOR_EACH(cp, args) {
         if (!SCM_PAIRP(SCM_CDR(cp))) {
             if (SCM_NULLP(start)) return SCM_CAR(cp);
-            SCM_SET_CDR(last, SCM_CAR(cp));
+            SCM_SET_CDR_UNCHECKED(last, SCM_CAR(cp));
             break;
         } else if (SCM_NULLP(SCM_CAR(cp))) {
             continue;
@@ -323,80 +410,94 @@ ScmObj Scm_Append(ScmObj args)
             Scm_Error("pair required, but got %S", SCM_CAR(cp));
         } else {
             SCM_APPEND(start, last, Scm_CopyList(SCM_CAR(cp)));
+            if (SCM_PAIRP(last) && !SCM_NULLP(SCM_CDR(last))) {
+                Scm_Error("proper list required, but got %S", SCM_CAR(cp));
+            }
         }
     }
     return start;
 }
 
-/* Scm_Reverse(list)
- *    Reverse LIST.  If LIST is not a pair, return LIST itself.
- *    If LIST is improper list, cdr of the last pair is ignored.
+/* Scm_Reverse2(list, tail)
+ *    Reverse LIST, and append TAIL to the result.
+ *    If LIST is an improper list, cdr of the last pair is ignored.
+ *    If LIST is not a pair, TAIL is returned.
+ * Scm_Reverse(list)
+ *    Scm_Reverse2(list, SCM_NIL).  Just for the backward compatibility.
  */
 
-ScmObj Scm_Reverse(ScmObj list)
+ScmObj Scm_Reverse2(ScmObj list, ScmObj tail)
 {
-    ScmObj cp, result;
-    ScmPair *p;
+    if (!SCM_PAIRP(list)) return tail;
 
-    if (!SCM_PAIRP(list)) return list;
-
-    p = SCM_NEW(ScmPair);
-    SCM_SET_CAR(p, SCM_NIL);
-    SCM_SET_CDR(p, SCM_NIL);
-    result = SCM_OBJ(p);
+    ScmPair *p = SCM_NEW(ScmPair);
+    SCM_SET_CAR_UNCHECKED(p, SCM_NIL);
+    SCM_SET_CDR_UNCHECKED(p, tail);
+    ScmObj result = SCM_OBJ(p);
+    ScmObj cp;
     SCM_FOR_EACH(cp, list) {
-	SCM_SET_CAR(result, SCM_CAR(cp));
+        SCM_SET_CAR_UNCHECKED(result, SCM_CAR(cp));
         p = SCM_NEW(ScmPair);
-        SCM_SET_CAR(p, SCM_NIL);
-        SCM_SET_CDR(p, result);
+        SCM_SET_CAR_UNCHECKED(p, SCM_NIL);
+        SCM_SET_CDR_UNCHECKED(p, result);
         result = SCM_OBJ(p);
     }
     return SCM_CDR(result);
 }
 
-    
-/* Scm_ReverseX(list)
+ScmObj Scm_Reverse(ScmObj list)
+{
+    return Scm_Reverse2(list, SCM_NIL);
+}
+
+
+/* Scm_Reverse2X(list, tail)
  *   Return reversed list of LIST.  Pairs in previous LIST is used to
- *   create new list.  If LIST is not a pair, return LIST itself.
+ *   create new list.  TAIL is appended to the result.
+ *   If LIST is not a pair, returns TAIL.
  *   If LIST is an improper list, cdr of the last cell is ignored.
  */
 
-ScmObj Scm_ReverseX(ScmObj list)
+ScmObj Scm_Reverse2X(ScmObj list, ScmObj tail)
 {
-    ScmObj first, next, result = SCM_NIL;
-    if (!SCM_PAIRP(list)) return list;
+    if (!SCM_PAIRP(list)) return tail;
+    ScmObj first, next, result = tail;
     for (first = list; SCM_PAIRP(first); first = next) {
         next = SCM_CDR(first);
-        SCM_SET_CDR(first, result);
+        Scm_SetCdr(first, result);
         result = first;
     }
     return result;
 }
 
+ScmObj Scm_ReverseX(ScmObj list)
+{
+    return Scm_Reverse2X(list, SCM_NIL);
+}
+
 /* Scm_ListTail(list, i, fallback)
  * Scm_ListRef(list, i, fallback)
  *    Note that i is C-INTEGER.  If i is out of bound, signal error.
  */
 
-ScmObj Scm_ListTail(ScmObj list, int i, ScmObj fallback)
+ScmObj Scm_ListTail(ScmObj list, ScmSmallInt i, ScmObj fallback)
 {
-    int cnt = i;
     if (i < 0) goto err;
+    ScmSmallInt cnt = i;
     while (cnt-- > 0) {
         if (!SCM_PAIRP(list)) goto err;
         list = SCM_CDR(list);
     }
     return list;
-  err:  
-    if (SCM_UNBOUNDP(fallback)) Scm_Error("argument out of range: %d", i);
+  err:
+    if (SCM_UNBOUNDP(fallback)) Scm_Error("argument out of range: %ld", i);
     return fallback;
 }
 
-ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
+ScmObj Scm_ListRef(ScmObj list, ScmSmallInt i, ScmObj fallback)
 {
-    int k;
     if (i < 0) goto err;
-    for (k=0; k<i; k++) {
+    for (ScmSmallInt k=0; k<i; k++) {
         if (!SCM_PAIRP(list)) goto err;
         list = SCM_CDR(list);
     }
@@ -404,7 +505,7 @@ ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
     return SCM_CAR(list);
   err:
     if (SCM_UNBOUNDP(fallback)) {
-        Scm_Error("argument out of range: %d", i);
+        Scm_Error("argument out of range: %ld", i);
     }
     return fallback;
 }
@@ -416,12 +517,16 @@ ScmObj Scm_ListRef(ScmObj list, int i, ScmObj fallback)
 
 ScmObj Scm_LastPair(ScmObj l)
 {
-    ScmObj cp;
-
     if (!SCM_PAIRP(l)) Scm_Error("pair required: %S", l);
-    SCM_FOR_EACH(cp, l) {
-	ScmObj cdr = SCM_CDR(cp);
-	if (!SCM_PAIRP(cdr)) return cp;
+
+    ScmObj slow = l;
+    for (;;) {
+        if (!SCM_PAIRP(SCM_CDR(l))) return l;
+        l = SCM_CDR(l);
+        if (!SCM_PAIRP(SCM_CDR(l))) return l;
+        l = SCM_CDR(l);
+        slow = SCM_CDR(slow);
+        if (l == slow) Scm_Error("list is circular: %S", l);
     }
     return SCM_UNDEFINED;       /* NOTREACHED */
 }
@@ -458,9 +563,9 @@ ScmObj Scm_Member(ScmObj obj, ScmObj list, int cmpmode)
 /* delete. */
 ScmObj Scm_Delete(ScmObj obj, ScmObj list, int cmpmode)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL, cp, prev = list;
-
     if (SCM_NULLP(list)) return SCM_NIL;
+
+    ScmObj start = SCM_NIL, last = SCM_NIL, cp, prev = list;
     SCM_FOR_EACH(cp, list) {
         if (Scm_EqualM(obj, SCM_CAR(cp), cmpmode)) {
             for (; prev != cp; prev = SCM_CDR(prev))
@@ -482,7 +587,7 @@ ScmObj Scm_DeleteX(ScmObj obj, ScmObj list, int cmpmode)
             if (SCM_NULLP(prev)) {
                 list = SCM_CDR(cp);
             } else {
-                SCM_SET_CDR(prev, SCM_CDR(cp));
+                Scm_SetCdr(prev, SCM_CDR(cp));
             }
         } else {
             prev = cp;
@@ -500,8 +605,8 @@ ScmObj Scm_DeleteX(ScmObj obj, ScmObj list, int cmpmode)
 
 ScmObj Scm_Assq(ScmObj obj, ScmObj alist)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assq: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
         ScmObj entry = SCM_CAR(cp);
         if (!SCM_PAIRP(entry)) continue;
@@ -512,20 +617,20 @@ ScmObj Scm_Assq(ScmObj obj, ScmObj alist)
 
 ScmObj Scm_Assv(ScmObj obj, ScmObj alist)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assv: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
-	ScmObj entry = SCM_CAR(cp);
-	if (!SCM_PAIRP(entry)) continue;
-	if (Scm_EqvP(obj, SCM_CAR(entry))) return entry;
+        ScmObj entry = SCM_CAR(cp);
+        if (!SCM_PAIRP(entry)) continue;
+        if (Scm_EqvP(obj, SCM_CAR(entry))) return entry;
     }
     return SCM_FALSE;
 }
 
 ScmObj Scm_Assoc(ScmObj obj, ScmObj alist, int cmpmode)
 {
-    ScmObj cp;
     if (!SCM_LISTP(alist)) Scm_Error("assoc: list required, but got %S", alist);
+    ScmObj cp;
     SCM_FOR_EACH(cp,alist) {
         ScmObj entry = SCM_CAR(cp);
         if (!SCM_PAIRP(entry)) continue;
@@ -537,12 +642,12 @@ ScmObj Scm_Assoc(ScmObj obj, ScmObj alist, int cmpmode)
 /* Assoc-delete */
 ScmObj Scm_AssocDelete(ScmObj elt, ScmObj alist, int cmpmode)
 {
-    ScmObj start = SCM_NIL, last = SCM_NIL, cp, p, prev = alist;
     if (!SCM_LISTP(alist)) {
         Scm_Error("assoc-delete: list required, but got %S", alist);
     }
     if (SCM_NULLP(alist)) return SCM_NIL;
-    
+
+    ScmObj start = SCM_NIL, last = SCM_NIL, cp, p, prev = alist;
     SCM_FOR_EACH(cp, alist) {
         p = SCM_CAR(cp);
         if (SCM_PAIRP(p)) {
@@ -561,10 +666,10 @@ ScmObj Scm_AssocDelete(ScmObj elt, ScmObj alist, int cmpmode)
 
 ScmObj Scm_AssocDeleteX(ScmObj elt, ScmObj alist, int cmpmode)
 {
-    ScmObj cp, prev = SCM_NIL;
     if (!SCM_LISTP(alist)) {
         Scm_Error("assoc-delete!: list required, but got %S", alist);
     }
+    ScmObj cp, prev = SCM_NIL;
     SCM_FOR_EACH(cp, alist) {
         ScmObj e = SCM_CAR(cp);
         if (SCM_PAIRP(e)) {
@@ -573,7 +678,7 @@ ScmObj Scm_AssocDeleteX(ScmObj elt, ScmObj alist, int cmpmode)
                     alist = SCM_CDR(cp);
                     continue;
                 } else {
-                    SCM_SET_CDR(prev, SCM_CDR(cp));
+                    Scm_SetCdr(prev, SCM_CDR(cp));
                 }
             }
         }
@@ -603,7 +708,7 @@ ScmObj Scm_DeleteDuplicatesX(ScmObj list, int cmpmode)
     SCM_FOR_EACH(lp, list) {
         ScmObj obj = SCM_CAR(lp);
         ScmObj tail = Scm_DeleteX(obj, SCM_CDR(lp), cmpmode);
-        if (SCM_CDR(lp) != tail) SCM_SET_CDR(lp, tail);
+        if (SCM_CDR(lp) != tail) Scm_SetCdr(lp, tail);
     }
     return list;
 }
@@ -615,7 +720,7 @@ ScmObj Scm_DeleteDuplicatesX(ScmObj list, int cmpmode)
  *  list.   If there's more than one way to order an element, choose the
  *  first one appears in the given list of lists.
  *  Returns SCM_FALSE if the lists are inconsistent to be ordered
- *  in the way. 
+ *  in the way.
  *
  *  START is an item of the starting point.  It is inserted into the result
  *  first.  SEQUENCES is a list of lists describing the order of preference.
@@ -627,30 +732,32 @@ ScmObj Scm_DeleteDuplicatesX(ScmObj list, int cmpmode)
  *  of the algorithm here.
  */
 
-ScmObj Scm_MonotonicMerge(ScmObj start, ScmObj sequences)
+ScmObj Scm_MonotonicMerge(ScmObj sequences)
 {
-    ScmObj result = Scm_Cons(start, SCM_NIL), next, h;
-    ScmObj *seqv, *sp, *tp;
+    ScmObj result = SCM_NIL;
     int nseqs = Scm_Length(sequences);
-
     if (nseqs < 0) Scm_Error("bad list of sequences: %S", sequences);
-    seqv = SCM_NEW_ARRAY(ScmObj, nseqs);
-    for (sp=seqv; SCM_PAIRP(sequences); sp++, sequences=SCM_CDR(sequences)) {
+    ScmObj *seqv = SCM_NEW_ARRAY(ScmObj, nseqs);
+    for (ScmObj *sp=seqv;
+         SCM_PAIRP(sequences);
+         sp++, sequences=SCM_CDR(sequences)) {
         *sp = SCM_CAR(sequences);
     }
 
     for (;;) {
         /* have we consumed all the inputs? */
+        ScmObj *sp;
         for (sp=seqv; sp<seqv+nseqs; sp++) {
             if (!SCM_NULLP(*sp)) break;
         }
         if (sp == seqv+nseqs) return Scm_ReverseX(result);
 
         /* select candidate */
-        next = SCM_FALSE;
+        ScmObj next = SCM_FALSE;
         for (sp = seqv; sp < seqv+nseqs; sp++) {
             if (!SCM_PAIRP(*sp)) continue;
-            h = SCM_CAR(*sp);
+            ScmObj h = SCM_CAR(*sp);
+            ScmObj *tp;
             for (tp = seqv; tp < seqv+nseqs; tp++) {
                 if (!SCM_PAIRP(*tp)) continue;
                 if (!SCM_FALSEP(Scm_Memq(h, SCM_CDR(*tp)))) {
@@ -676,9 +783,57 @@ ScmObj Scm_MonotonicMerge(ScmObj start, ScmObj sequences)
 }
 
 /*
- * Pair attributes
+ * Extended pairs
  */
 
+static ScmObj make_extended_pair(ScmExtendedPairDescriptor *desc,
+                                 ScmObj car, ScmObj cdr, ScmObj attrs)
+{
+    ScmRealExtendedPair *xp = SCM_NEW(ScmRealExtendedPair);
+    /* ScmRealExtendedPair is not an ScmObj, and
+       ScmExtendedPairDescriptor is not an ScmClass.   To avoid confusion,
+       we manually tweak tag bits.
+    */
+    xp->hiddenTag = SCM_WORD((ScmByte*)desc + 7);
+    xp->data.car = car;
+    xp->data.cdr = cdr;
+    xp->data.attributes = attrs;
+    return SCM_OBJ(&xp->data);  /* hide the first word  */
+}
+
+/* "vanilla" extended pair.  used mainly to hold extra attributes, but
+   otherwise behaves like normal pairs.
+   NB: Static initialization with ScmClass* requires extra care on
+   Windows.  To avoid complication, we initialize the klass field in
+   _Init() routine.
+*/
+static ScmExtendedPairDescriptor mpair_desc SCM_ALIGN8 = {
+    NULL,                       /* will be SCM_CLASS_PAIR */
+    0,
+    NULL,
+    NULL
+};
+
+ScmObj Scm_MakeExtendedPair(ScmObj car, ScmObj cdr, ScmObj attrs)
+{
+    return make_extended_pair(&mpair_desc, car, cdr, attrs);
+}
+
+/* Returns NULL if p isn't an extended pair. */
+ScmExtendedPairDescriptor *Scm__GetExtendedPairDescriptor(ScmObj p)
+{
+    if (!SCM_EXTENDED_PAIR_P(p)) return NULL;
+    ScmRealExtendedPair *z = (ScmRealExtendedPair*)(((ScmObj*)p) - 1);
+    SCM_ASSERT((z->hiddenTag&0x7) == 0x7);
+    return (ScmExtendedPairDescriptor *)(z->hiddenTag-7);
+}
+
+/* Returns vanilla extended pair descriptor */
+ScmExtendedPairDescriptor *Scm__GetDefaultExtendedPairDesctiptor()
+{
+    return &mpair_desc;
+}
+
 ScmObj Scm_PairAttr(ScmPair *pair)
 {
     if (SCM_EXTENDED_PAIR_P(pair)) {
@@ -688,23 +843,20 @@ ScmObj Scm_PairAttr(ScmPair *pair)
     }
 }
 
+/* The common scenario is to use ExtendedCons in place of Cons,
+   then add attributes later.  So we provide this.  */
 ScmObj Scm_ExtendedCons(ScmObj car, ScmObj cdr)
 {
-    ScmExtendedPair *xp = SCM_NEW(ScmExtendedPair);
-    xp->car = car;
-    xp->cdr = cdr;
-    xp->attributes = SCM_NIL;
-    return SCM_OBJ(xp);
+    return Scm_MakeExtendedPair(car, cdr, SCM_NIL);
 }
 
 ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback)
 {
-    ScmObj p;
     if (!SCM_EXTENDED_PAIR_P(pair)) {
         goto fallback;
     }
-    
-    p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
+
+    ScmObj p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
     if (SCM_PAIRP(p)) return SCM_CDR(p);
   fallback:
     if (fallback == SCM_UNBOUND)
@@ -715,17 +867,81 @@ ScmObj Scm_PairAttrGet(ScmPair *pair, ScmObj key, ScmObj fallback)
 
 ScmObj Scm_PairAttrSet(ScmPair *pair, ScmObj key, ScmObj value)
 {
-    ScmObj p;
     if (!SCM_EXTENDED_PAIR_P(pair)) {
         Scm_Error("Cannot set pair attribute (%S) to non-extended pair: %S",
                   key, SCM_OBJ(pair));
     }
-    
-    p = Scm_Assq(key, SCM_EXTENDED_PAIR(pair)->attributes);
-    if (SCM_PAIRP(p)) SCM_SET_CDR(p, value);
-    else SCM_EXTENDED_PAIR(pair)->attributes
-        = Scm_Acons(key, value, SCM_EXTENDED_PAIR(pair)->attributes);
+
+    /* We update the attribute list non-destructively.
+       In case if two threads update the list simultaneously, one change
+       can be lost, but no invalid access can occur.  Guaranteeing mutex
+       is the caller's responsibility, though.
+    */
+    ScmObj xs, attrs = SCM_EXTENDED_PAIR(pair)->attributes;
+    SCM_FOR_EACH(xs, attrs) {
+        SCM_ASSERT(SCM_PAIRP(SCM_CAR(xs)));
+        if (SCM_EQ(SCM_CAAR(xs), key)) {
+            /* need to copy up to here */
+            ScmObj h = SCM_NIL, t = SCM_NIL, ys;
+            SCM_FOR_EACH(ys, attrs) {
+                if (SCM_EQ(ys, xs)) {
+                    SCM_APPEND1(h, t, Scm_Cons(key, value));
+                    SCM_APPEND(h, t, SCM_CDR(xs));
+                    SCM_EXTENDED_PAIR(pair)->attributes = h;
+                    return SCM_UNDEFINED;
+                }
+                SCM_APPEND1(h, t, SCM_CAR(ys));
+            }
+            SCM_ASSERT(FALSE);  /* Shouldn't be here */
+        }
+    }
+    SCM_EXTENDED_PAIR(pair)->attributes = Scm_Acons(key, value, attrs);
     return SCM_UNDEFINED;
 }
 
+/*
+ * Immutable pairs
+ */
 
+/* not cheap.  usually you don't want to check this and let it catched
+   in set-car!/set-cdr!. */
+int Scm_ImmutablePairP(ScmObj obj)
+{
+    if (!SCM_EXTENDED_PAIR_P(obj)) return FALSE;
+    ScmExtendedPairDescriptor *d = Scm__GetExtendedPairDescriptor(obj);
+    return d->flags & SCM_PAIR_IMMUTABLE;
+}
+
+static ScmExtendedPairDescriptor ipair_desc SCM_ALIGN8 = {
+    NULL,                       /* will be SCM_CLASS_PAIR.  see above. */
+    SCM_PAIR_IMMUTABLE,
+    NULL,
+    NULL
+};
+
+ScmObj Scm_MakeImmutablePair(ScmObj car, ScmObj cdr, ScmObj attrs)
+{
+    return make_extended_pair(&ipair_desc, car, cdr, attrs);
+}
+
+void Scm__InitList()
+{
+    mpair_desc.klass = SCM_CLASS_PAIR;
+    ipair_desc.klass = SCM_CLASS_PAIR;
+}
+
+/* Temporary - Check if normal pairs are all aligned with 2-word boundary */
+#if GAUCHE_CHECK_PAIR_ALIGNMENT
+int Scm_CheckingPairP(ScmObj obj)
+{
+    if (SCM_HPTRP(obj)&&(SCM_HTAG(obj)!=7||Scm_PairP(SCM_OBJ(obj)))) {
+        if (SCM_WORD(obj) & SIZEOF_LONG) {
+            if ((SCM_WORD((ScmObj*)obj - 1) & 7) != 7) {
+                fprintf(stderr, "Unaligned pair %p\n", obj);
+            }
+        }
+        return TRUE;
+    }
+    return FALSE;
+}
+#endif /*GAUCHE_CHECK_PAIR_ALIGNMENT*/
