@@ -1,7 +1,7 @@
 /*
  * macro.c - macro implementation
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,6 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/code.h"
 #include "gauche/vminsn.h"
 #include "gauche/priv/macroP.h"
@@ -50,7 +51,7 @@
  * Syntax object
  */
 
-static void syntax_print(ScmObj obj, ScmPort *port, 
+static void syntax_print(ScmObj obj, ScmPort *port,
                          ScmWriteContext *mode SCM_UNUSED)
 {
     ScmSymbol *name = SCM_SYNTAX(obj)->name;
@@ -58,7 +59,7 @@ static void syntax_print(ScmObj obj, ScmPort *port,
     if (mod == NULL) {
         Scm_Printf(port, "#<syntax %A>", (name ? SCM_OBJ(name) : SCM_FALSE));
     } else {
-        Scm_Printf(port, "#<syntax %A#%A>", 
+        Scm_Printf(port, "#<syntax %A#%A>",
                    mod->name,
                    (name ? SCM_OBJ(name) : SCM_FALSE));
     }
@@ -81,14 +82,14 @@ ScmObj Scm_MakeSyntax(ScmSymbol *name, ScmModule *mod, ScmObj handler)
  */
 
 ScmObj Scm_MakeMacro(ScmObj name, ScmObj transformer,
-                     ScmObj src, ScmObj describer)
+                     ScmObj info, u_long flags)
 {
     ScmMacro *s = SCM_NEW(ScmMacro);
     SCM_SET_CLASS(s, SCM_CLASS_MACRO);
     s->name = name;
     s->transformer = transformer;
-    s->src = src;
-    s->describer = describer;
+    s->info_alist = info;
+    s->flags = flags;
     return SCM_OBJ(s);
 }
 
@@ -102,6 +103,26 @@ ScmObj Scm_MacroName(ScmMacro *mac)
     return (mac->name? SCM_OBJ(mac->name) : SCM_FALSE);
 }
 
+/* Called from syntax-parameterize.
+   This API is internal and subject to change.
+ */
+void Scm__SwapMacroTransformer(ScmMacro *mac,
+                               ScmObj *pxformer,
+                               u_long *pflags)
+{
+    if (!(mac->flags & SCM_MACRO_PARAMETERIZABLE)) {
+        Scm_Error("Macro %S is not parameterizable.", SCM_OBJ(mac));
+    }
+    ScmObj oxformer = mac->transformer;
+    u_long oflags = mac->flags;
+
+    mac->transformer = *pxformer;
+    mac->flags = *pflags | SCM_MACRO_PARAMETERIZABLE;
+
+    *pxformer = oxformer;
+    *pflags = oflags & (~SCM_MACRO_PARAMETERIZABLE);
+}
+
 /*===================================================================
  * SyntaxPattern object
  * Repesents a repeatable subpattern
@@ -130,7 +151,7 @@ SCM_CLASS_DECL(Scm_SyntaxPatternClass);
 #define SCM_SYNTAX_PATTERN(obj)   ((ScmSyntaxPattern*)(obj))
 #define SCM_SYNTAX_PATTERN_P(obj) SCM_XTYPEP(obj, SCM_CLASS_SYNTAX_PATTERN)
 
-static void pattern_print(ScmObj obj, ScmPort *port, 
+static void pattern_print(ScmObj obj, ScmPort *port,
                           ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(port, "#<pattern:%d%S %S%s>",
@@ -158,7 +179,7 @@ ScmSyntaxPattern *make_syntax_pattern(int level, int numFollowing)
  *   Internal object to construct pattern matcher
  */
 
-static void synrule_print(ScmObj obj, ScmPort *port, 
+static void synrule_print(ScmObj obj, ScmPort *port,
                           ScmWriteContext *mode SCM_UNUSED)
 {
     ScmSyntaxRules *r = SCM_SYNTAX_RULES(obj);
@@ -188,17 +209,6 @@ ScmSyntaxRules *make_syntax_rules(int nr)
  * Traditional Macro
  */
 
-/* TRANSIENT
-   This used to be called via make-macro-transformer, but no longer.
-   We leave stub here for ABI compatibility. */
-#if GAUCHE_API_VERSION < 1000
-ScmObj Scm_MakeMacroTransformerOld(ScmSymbol *name SCM_UNUSED,
-                                   ScmProcedure *proc SCM_UNUSED)
-{
-    Scm_Panic("Obsoleted Scm_MakeMacroTransformerOld called!  Something is wrong!");
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
-
 static ScmMacro *resolve_macro_autoload(ScmAutoload *adata)
 {
     ScmObj mac = Scm_ResolveAutoload(adata, 0);
@@ -234,7 +244,7 @@ ScmObj Scm_MakeMacroAutoload(ScmSymbol *name, ScmAutoload *adata)
                                      sizeof(AUTOLOAD_MACRO_SUFFIX)-1,
                                      sizeof(AUTOLOAD_MACRO_SUFFIX)-1);
     return Scm_MakeMacro(Scm_MakeSymbol(SCM_STRING(name1), FALSE),
-                         transformer, SCM_FALSE, SCM_FALSE);
+                         transformer, SCM_NIL, 0);
 }
 
 /*===================================================================
@@ -273,7 +283,7 @@ typedef struct {
     ScmObj literals;            /* list of literals */
     ScmObj pvars;               /* list of (pvar . pvref) */
     ScmObj renames;             /* list of (var . identifier)  Keep mapping
-                                   of input symbol/identifier to fresh 
+                                   of input symbol/identifier to fresh
                                    identifier */
     ScmObj ellipsis;            /* symbol/identifier/keyword for ellipsis
                                    SCM_TRUE means default (...)
@@ -313,6 +323,69 @@ static inline ScmObj add_pvar(PatternContext *ctx,
     return pvref;
 }
 
+/* Like SCM_APPEND1 and SCM_APPEND, but try to preserve source info
+   of the input.  SOURCE is the macro source in its definition, ORIGIN is
+   the macro expansion input in its entirety.  ORIGIN is #f while compiling
+   a macro; only given when expanding it.
+*/
+static ScmObj get_source_info_attrs(ScmObj source, ScmObj origin)
+{
+    ScmObj attrs = SCM_NIL;
+    if (SCM_PAIRP(source)) {
+        ScmObj info = Scm_PairAttrGet(SCM_PAIR(source),
+                                      SCM_SYM_SOURCE_INFO, SCM_FALSE);
+        if (SCM_PAIRP(info))
+            attrs = Scm_Acons(SCM_SYM_SOURCE_INFO, info, attrs);
+    }
+    if (SCM_PAIRP(origin)) {
+        attrs = Scm_Acons(SCM_SYM_ORIGINAL, origin, attrs);
+    }
+    return attrs;
+}
+
+
+static void template_append1(ScmObj *head, ScmObj *tail,
+                             ScmObj elt, ScmObj source, ScmObj origin)
+{
+    if (SCM_NULLP(*head)) {
+        ScmObj attrs = get_source_info_attrs(source, origin);
+        if (SCM_PAIRP(attrs)) {
+            ScmObj p = Scm_MakeExtendedPair(elt, SCM_NIL, attrs);
+            *head = *tail = p;
+            return;
+        }
+        *head = *tail = Scm_Cons(elt, SCM_NIL);
+        return;
+    }
+    ScmObj p = Scm_Cons(elt, SCM_NIL);
+    SCM_SET_CDR(*tail, p);
+    *tail = p;
+}
+
+static void template_append(ScmObj *head, ScmObj *tail,
+                            ScmObj rest, ScmObj source, ScmObj origin)
+{
+    if (SCM_NULLP(*head)) {
+        if (SCM_PAIRP(rest)) {
+            ScmObj attrs = get_source_info_attrs(source, origin);
+            if (SCM_PAIRP(attrs)) {
+                ScmObj p = Scm_MakeExtendedPair(SCM_CAR(rest), SCM_CDR(rest),
+                                                attrs);
+                *head = p;
+                *tail = Scm_LastPair(p);
+                return;
+            }
+        }
+        *head = rest;
+        *tail = SCM_PAIRP(rest)? Scm_LastPair(rest) : rest;
+        return;
+    }
+    SCM_SET_CDR(*tail, rest);
+    /* rest can be () */
+    if (SCM_PAIRP(rest)) *tail = Scm_LastPair(rest);
+}
+
+
 /* returns pvref corresponds to the given pvar in template compilation.
    if pvar is not a valid pvar, returns pvar itself. */
 static inline ScmObj pvar_to_pvref(PatternContext *ctx,
@@ -390,7 +463,7 @@ static ScmObj rename_variable(ScmObj var,
    In the pattern, we replace variables to identifiers.
    We also recognize pattern variables, and replace them for PVARs in
    the pattern, and for PVREFs in the template.
-   When encounters a repeatable subpattern, replace it with 
+   When encounters a repeatable subpattern, replace it with
    SyntaxPattern node.
 */
 static ScmObj compile_rule1(ScmObj form,
@@ -464,7 +537,7 @@ static ScmObj compile_rule1(ScmObj form,
                 }
                 outer->pattern = compile_rule1(base, outer, ctx, patternp);
                 outermost->vars = outer->vars;
-                SCM_APPEND1(h, t, SCM_OBJ(outermost));
+                template_append1(&h, &t, SCM_OBJ(outermost), form, SCM_FALSE);
                 if (!patternp) {
                     ScmObj vp;
                     if (SCM_NULLP(outermost->vars)) {
@@ -485,12 +558,14 @@ static ScmObj compile_rule1(ScmObj form,
                 }
                 spat->vars = Scm_Append2(spat->vars, outermost->vars);
             } else {
-                SCM_APPEND1(h, t,
-                            compile_rule1(SCM_CAR(pp), spat, ctx, patternp));
+                template_append1(&h, &t,
+                                 compile_rule1(SCM_CAR(pp), spat, ctx, patternp),
+                                 form, SCM_FALSE);
             }
         }
         if (!SCM_NULLP(pp))
-            SCM_APPEND(h, t, compile_rule1(pp, spat, ctx, patternp));
+            template_append(&h, &t, compile_rule1(pp, spat, ctx, patternp),
+                            form, SCM_FALSE);
         return h;
     }
     else if (SCM_VECTORP(form)) {
@@ -504,11 +579,11 @@ static ScmObj compile_rule1(ScmObj form,
         if (!SCM_FALSEP(Scm_Memq(form, ctx->literals))) {
             if (patternp)
                 return rename_variable(form, &ctx->renames, ctx->mod, ctx->env);
-            else        
+            else
                 return form;  /* template renaming is done in expansion time */
         }
         if (patternp && Scm__ERCompare(form, SCM_SYM_UNDERBAR,
-                                       ctx->mod, ctx->env)) { 
+                                       ctx->mod, ctx->env)) {
             return SCM_SYM_UNDERBAR;
         }
         if (patternp) {
@@ -580,8 +655,15 @@ static ScmSyntaxRules *compile_rules(ScmObj name,
         ctx.maxlev = 0;
 
         ctx.form = SCM_CAR(rule);
-        if (!SCM_PAIRP(ctx.form)) goto badform;
-        pat->pattern = compile_rule1(SCM_CDR(ctx.form), pat, &ctx, TRUE);
+        if (SCM_PAIRP(ctx.form)) {
+            pat->pattern = compile_rule1(SCM_CDR(ctx.form), pat, &ctx, TRUE);
+        } else {
+            if (!SCM_SYMBOLP(ctx.form) && !SCM_IDENTIFIERP(ctx.form)) goto badform;
+            /* This happens for identifier macro. we allow the entire input.
+               The entire input always match.  The identifier in the pattern
+               doesn't matter; we use underscore for placeholder.*/
+            pat->pattern = SCM_SYM_UNDERBAR;
+        }
 
         ctx.form = SCM_CADR(rule);
         tmpl->pattern = compile_rule1(ctx.form, tmpl, &ctx, FALSE);
@@ -859,8 +941,9 @@ static int match_synrule(ScmObj form, ScmObj pattern, ScmObj mod, ScmObj env,
  * pattern language transformer
  */
 
-/* If a pattern variable is exhausted, SCM_UNDEFINED is returned. */
-static ScmObj realize_template_rec(ScmSyntaxRules *sr,
+/* If a pattern variable is exhausted, SCM_UNBOUND is returned. */
+static ScmObj realize_template_rec(ScmObj src,
+                                   ScmSyntaxRules *sr,
                                    ScmObj template,
                                    MatchVar *mvec,
                                    int level,
@@ -873,21 +956,24 @@ static ScmObj realize_template_rec(ScmSyntaxRules *sr,
         while (SCM_PAIRP(template)) {
             ScmObj e = SCM_CAR(template);
             if (SCM_SYNTAX_PATTERN_P(e)) {
-                ScmObj r = realize_template_rec(sr, e, mvec, level, indices, idlist, exlev);
+                ScmObj r = realize_template_rec(src, sr, e, mvec, level,
+                                                indices, idlist, exlev);
                 if (SCM_UNBOUNDP(r)) return r;
-                SCM_APPEND(h, t, r);
+                template_append(&h, &t, r, template, src);
             } else {
-                ScmObj r = realize_template_rec(sr, e, mvec, level, indices, idlist, exlev);
+                ScmObj r = realize_template_rec(src, sr, e, mvec, level,
+                                                indices, idlist, exlev);
                 if (SCM_UNBOUNDP(r)) return r;
-                SCM_APPEND1(h, t, r);
+                template_append1(&h, &t, r, template, src);
             }
             template = SCM_CDR(template);
         }
         if (!SCM_NULLP(template)) {
-            ScmObj r = realize_template_rec(sr, template, mvec, level, indices, idlist, exlev);
+            ScmObj r = realize_template_rec(src, sr, template, mvec, level,
+                                            indices, idlist, exlev);
             if (SCM_UNBOUNDP(r)) return r;
             if (SCM_NULLP(h)) return r; /* (a ... . b) and a ... is empty */
-            SCM_APPEND(h, t, r);
+            template_append(&h, &t, r, template, src);
         }
         return h;
     }
@@ -899,7 +985,8 @@ static ScmObj realize_template_rec(ScmSyntaxRules *sr,
         ScmObj h = SCM_NIL, t = SCM_NIL;
         indices[level+1] = 0;
         for (;;) {
-            ScmObj r = realize_template_rec(sr, pat->pattern, mvec, level+1, indices, idlist, exlev);
+            ScmObj r = realize_template_rec(src, sr, pat->pattern, mvec,
+                                            level+1, indices, idlist, exlev);
             if (SCM_UNBOUNDP(r)) return (*exlev < pat->level)? r : h;
             if (SCM_SYNTAX_PATTERN_P(pat->pattern)) {
                 SCM_APPEND(h, t, r);
@@ -916,11 +1003,13 @@ static ScmObj realize_template_rec(ScmSyntaxRules *sr,
 
         for (int i=0; i<len; i++, pe++) {
             if (SCM_SYNTAX_PATTERN_P(*pe)) {
-                ScmObj r = realize_template_rec(sr, *pe, mvec, level, indices, idlist, exlev);
+                ScmObj r = realize_template_rec(src, sr, *pe, mvec, level,
+                                                indices, idlist, exlev);
                 if (SCM_UNBOUNDP(r)) return r;
                 SCM_APPEND(h, t, r);
             } else {
-                ScmObj r = realize_template_rec(sr, *pe, mvec, level, indices, idlist, exlev);
+                ScmObj r = realize_template_rec(src, sr, *pe, mvec, level,
+                                                indices, idlist, exlev);
                 if (SCM_UNBOUNDP(r)) return r;
                 SCM_APPEND1(h, t, r);
             }
@@ -935,7 +1024,8 @@ static ScmObj realize_template_rec(ScmSyntaxRules *sr,
 
 #define DEFAULT_MAX_LEVEL  10
 
-static ScmObj realize_template(ScmSyntaxRules *sr,
+static ScmObj realize_template(ScmObj src,
+                               ScmSyntaxRules *sr,
                                ScmSyntaxRuleBranch *branch,
                                MatchVar *mvec)
 {
@@ -946,7 +1036,7 @@ static ScmObj realize_template(ScmSyntaxRules *sr,
     if (branch->maxLevel > DEFAULT_MAX_LEVEL)
         indices = SCM_NEW_ATOMIC2(int*, (branch->maxLevel+1) * sizeof(int));
     for (int i=0; i<=branch->maxLevel; i++) indices[i] = 0;
-    return realize_template_rec(sr, branch->template, mvec, 0,
+    return realize_template_rec(src, sr, branch->template, mvec, 0,
                                 indices, &idlist, &exlev);
 }
 
@@ -962,12 +1052,19 @@ static ScmObj synrule_expand(ScmObj form, ScmObj mod, ScmObj env, ScmSyntaxRules
         Scm_Printf(SCM_CUROUT, "pattern #%d: %S\n", i, sr->rules[i].pattern);
 #endif
         init_matchvec(mvec, sr->rules[i].numPvars);
-        if (match_synrule(SCM_CDR(form), sr->rules[i].pattern, mod, env, mvec)) {
+
+        int matched = FALSE;
+        if (sr->rules[i].pattern == SCM_SYM_UNDERBAR) {
+            matched = TRUE;     /* unconditionally match */
+        } else {
+            matched = match_synrule(SCM_CDR(form), sr->rules[i].pattern, mod, env, mvec);
+        }
+        if (matched) {
 #ifdef DEBUG_SYNRULE
             Scm_Printf(SCM_CUROUT, "success #%d:\n", i);
             print_matchvec(mvec, sr->rules[i].numPvars, SCM_CUROUT);
 #endif
-            ScmObj expanded = realize_template(sr, &sr->rules[i], mvec);
+            ScmObj expanded = realize_template(form, sr, &sr->rules[i], mvec);
 #ifdef DEBUG_SYNRULE
             Scm_Printf(SCM_CUROUT, "result: %S\n", expanded);
 #endif
@@ -1001,27 +1098,26 @@ ScmObj Scm_CompileSyntaxRules(ScmObj name, ScmObj src, ScmObj ellipsis,
                                        SCM_MODULE(mod), env);
     ScmObj sr_xform = Scm_MakeSubr(synrule_transform, sr,
                                    2, 0, SCM_FALSE);
-    return Scm_MakeMacro(name, sr_xform, src, SCM_FALSE);
+
+    ScmObj info = SCM_NIL, info_tail = SCM_NIL;
+    if (SCM_PAIRP(src)) {
+        SCM_APPEND1(info, info_tail, Scm_Cons(SCM_SYM_SOURCE, src));
+        ScmObj source_info = Scm_PairAttrGet(SCM_PAIR(src),
+                                             SCM_SYM_SOURCE_INFO,
+                                             SCM_FALSE);
+        if (SCM_PAIRP(source_info)) {
+            SCM_APPEND1(info, info_tail,
+                        Scm_Cons(SCM_SYM_SOURCE_INFO, source_info));
+        }
+    }
+
+    return Scm_MakeMacro(name, sr_xform, info, 0);
 }
 
 /*===================================================================
  * macro-expand
  */
 
-/* TRANSIENT
-   Now it's in compile.scm (%internal-macro-expand).  This is kept
-   for ABI compatibility, but nobody is supposed to call this.
- */
-#if GAUCHE_API_VERSION < 1000
-ScmObj Scm_VMMacroExpand(ScmObj expr SCM_UNUSED,
-                         ScmObj env SCM_UNUSED,
-                         int oncep SCM_UNUSED)
-{
-    Scm_Error("Scm_VMMacroExpand is obsoleted.");
-    return SCM_UNDEFINED;
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
-
 ScmObj Scm_CallMacroExpander(ScmMacro *mac, ScmObj expr, ScmObj cenv)
 {
     SCM_ASSERT(SCM_VECTORP(cenv));
