@@ -1,7 +1,7 @@
 /*
  * port.c - port implementation
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/portP.h"
 #include "gauche/priv/builtin-syms.h"
 
@@ -60,8 +60,10 @@
 #define PORT_SCRATCH(p)   (P_(p)->scratch)
 #define PORT_LINE(p)      (P_(p)->line)
 #define PORT_BYTES(p)     (P_(p)->bytes)
+#define PORT_COLUMN(p)    (P_(p)->column)
 #define PORT_ATTRS(p)     (P_(p)->attrs)
 #define PORT_SAVED_POS(p) (P_(p)->savedPos)
+#define PORT_LINK(p)      (P_(p)->link)
 
 /* Parameter location for the global reader lexical mode, from which
    ports inherit. */
@@ -90,6 +92,18 @@ static ScmObj get_port_current_line(ScmPort *port)
     return SCM_MAKE_INT(Scm_PortLine(port));
 }
 
+static ScmObj get_port_current_column(ScmPort *port)
+{
+    return SCM_MAKE_INT(PORT_COLUMN(port));
+}
+
+static void set_port_current_column(ScmPort *port, ScmObj val)
+{
+    if (SCM_INTP(val)) {
+        PORT_COLUMN(port) = SCM_INT_VALUE(val);
+    }
+}
+
 static ScmObj get_port_buffering(ScmPort *port)
 {
     return Scm_GetPortBufferingModeAsKeyword(port);
@@ -113,6 +127,11 @@ static void set_port_sigpipe_sensitive(ScmPort *port, ScmObj val)
     Scm_SetPortBufferSigpipeSensitive(port, SCM_BOOL_VALUE(val));
 }
 
+static ScmObj get_port_link(ScmPort *port)
+{
+    return PORT_LINK(port);
+}
+
 static ScmClassStaticSlotSpec port_slots[] = {
     SCM_CLASS_SLOT_SPEC("name", get_port_name, NULL),
     SCM_CLASS_SLOT_SPEC("buffering", get_port_buffering,
@@ -120,6 +139,9 @@ static ScmClassStaticSlotSpec port_slots[] = {
     SCM_CLASS_SLOT_SPEC("sigpipe-sensitive?", get_port_sigpipe_sensitive,
                         set_port_sigpipe_sensitive),
     SCM_CLASS_SLOT_SPEC("current-line", get_port_current_line, NULL),
+    SCM_CLASS_SLOT_SPEC("current-column", get_port_current_column,
+                        set_port_current_column),
+    SCM_CLASS_SLOT_SPEC("link", get_port_link, NULL),
     SCM_CLASS_SLOT_SPEC_END()
 };
 
@@ -148,8 +170,16 @@ SCM_DEFINE_BASE_CLASS(Scm_CodingAwarePortClass,
  *   share the same file descriptor.  However, C code and Scheme port
  *   may share the same file descriptor for efficiency (e.g. stdios).
  *   In such cases, it is C code's responsibility to destroy the port.
+ *
+ *   Port_clenaup routine may be called when the port is closed or
+ *   finalized.  If port is finalized before it is properly closed,
+ *   we want to call close procedure of the underlying implementation.
+ *   However, the fact that the port is being finalized means the
+ *   extra structure used by the underlying implementation may have
+ *   already GC-ed.  The FINAL flag indicates that.
+
  */
-static void port_cleanup(ScmPort *port)
+static void port_cleanup(ScmPort *port, _Bool final)
 {
     if (SCM_PORT_CLOSED_P(port)) return;
 
@@ -161,7 +191,7 @@ static void port_cleanup(ScmPort *port)
     switch (SCM_PORT_TYPE(port)) {
     case SCM_PORT_FILE:
         if (SCM_PORT_DIR(port) == SCM_PORT_OUTPUT) {
-            if (!SCM_PORT_ERROR_OCCURRED_P(port)) {
+            if (!SCM_PORT_ERROR_OCCURRED_P(port) && !final) {
                 bufport_flush(port, 0, TRUE);
             }
             if (!(SCM_PORT_FLAGS(port) & SCM_PORT_TRANSIENT)) {
@@ -169,10 +199,10 @@ static void port_cleanup(ScmPort *port)
             }
         }
         ScmPortBuffer *buf = Scm_PortBufferStruct(port);
-        if (port->ownerp && buf->closer) buf->closer(port);
+        if (!final && port->ownerp && buf->closer) buf->closer(port);
         break;
     case SCM_PORT_PROC:
-        if (PORT_VT(port)->Close) PORT_VT(port)->Close(port);
+        if (!final && PORT_VT(port)->Close) PORT_VT(port)->Close(port);
         break;
     default:
         break;
@@ -184,10 +214,22 @@ static void port_cleanup(ScmPort *port)
     Scm_UnregisterFinalizer(SCM_OBJ(port));
 }
 
-/* called by GC */
+/* called by GC.
+   NB: We only attach finalizers for SCM_PORT_FILE and SCM_PORT_PROC.
+   Attaching finalizer has overhead, which is not negligible when
+   string ports are frequently created.
+   Cf. https://github.com/shirok/Gauche/issues/937
+*/
 static void port_finalize(ScmObj obj, void* data SCM_UNUSED)
 {
-    port_cleanup(SCM_PORT(obj));
+    if (PORT_FILE_EXTDATA_P(obj)) {
+        GC_unregister_disappearing_link((void**)&PORT_BUF(obj)->data);
+    }
+    if (PORT_PROC_EXTDATA_P(obj)) {
+        GC_unregister_disappearing_link((void**)&PORT_VT(obj)->data);
+    }
+    port_cleanup(SCM_PORT(obj),
+                 PORT_FILE_EXTDATA_GONE_P(obj)||PORT_PROC_EXTDATA_GONE_P(obj));
 }
 
 /*
@@ -214,18 +256,31 @@ static ScmPort *make_port(ScmClass *klass, ScmObj name, int dir, int type)
     (void)SCM_INTERNAL_FASTLOCK_INIT(port->lock);
     port->lockOwner = NULL;
     port->lockCount = 0;
+    (void)SCM_INTERNAL_FASTLOCK_INIT(port->attrlock);
     port->writeState = NULL;
     port->line = 1;
-    /* We set name attribute as read-only attribute.  See portapi.c
-       for the format of attrs. */
-    port->attrs = SCM_LIST1(Scm_Cons(SCM_SYM_NAME, Scm_Cons(name, SCM_FALSE)));
+    port->bytes = 0;
+    port->column = 0;
+    /* Initial port attributes.
+         ((name <name> . #f)   ; read-only
+          (reader-lexical-mode <lexical-mode>))
+       See "Port Attributes" section below for the format of attrs. */
+    port->attrs = SCM_LIST2(Scm_Cons(SCM_SYM_NAME,
+                                     Scm_Cons(name, SCM_FALSE)),
+                            SCM_LIST2(SCM_SYM_READER_LEXICAL_MODE,
+                                      Scm_ReaderLexicalMode()));
+    port->link = SCM_FALSE;
+    port->internalFlags = 0;
 
+    /* If fast lock doesn't require finalization, we can skip to register
+       a finalizer for every string port. It reduces significant overhead. */
+#if defined(SCM_INTERNAL_FASTLOCK_FINALIZATION_NOT_REQUIRED)
+    if (type == SCM_PORT_FILE || type == SCM_PORT_PROC) {
+        Scm_RegisterFinalizer(SCM_OBJ(port), port_finalize, NULL);
+    }
+#else
     Scm_RegisterFinalizer(SCM_OBJ(port), port_finalize, NULL);
-
-    /* Default reader lexical mode */
-    Scm_PortAttrSetUnsafe(SCM_PORT(port),
-                          SCM_SYM_READER_LEXICAL_MODE,
-                          Scm_ReaderLexicalMode());
+#endif
 
     return SCM_PORT(port);
 }
@@ -240,12 +295,87 @@ void Scm_ClosePort(ScmPort *port)
     PORT_SAFE_CALL(port,
                    do {
                        if (!SCM_PORT_CLOSED_P(port)) {
-                           port_cleanup(port);
+                           port_cleanup(port, FALSE);
                        }
                    } while (0), /*no cleanup*/);
     PORT_UNLOCK(port);
 }
 
+/*
+ * Link
+ *   NB: Always lock from iport, then oport, to avoid deadlock.
+ */
+void Scm_LinkPorts(ScmPort *iport, ScmPort *oport)
+{
+    if (!SCM_IPORTP(iport)) {
+        Scm_Error("input port required, but got: %S", iport);
+    }
+    if (!SCM_OPORTP(oport)) {
+        Scm_Error("output port required, but got: %S", oport);
+    }
+
+    int aborted = FALSE;
+    ScmVM *vm = Scm_VM();
+    PORT_LOCK(iport, vm);
+    PORT_LOCK(oport, vm);
+    if (!SCM_FALSEP(PORT_LINK(iport))
+        || !SCM_FALSEP(PORT_LINK(oport))) {
+        aborted = TRUE;
+    } else {
+        PORT_LINK(iport) = SCM_OBJ(oport);
+        PORT_LINK(oport) = SCM_OBJ(iport);
+    }
+    PORT_UNLOCK(oport);
+    PORT_UNLOCK(iport);
+
+    if (aborted) {
+        Scm_Error("Port(s) are already linked.  Unlink them first.  (%S, %S)",
+                  iport, oport);
+    }
+}
+
+void Scm_UnlinkPorts(ScmPort *port)
+{
+    ScmObj link = PORT_LINK(port);
+    if (SCM_PORTP(link)) {
+        ScmVM *vm = Scm_VM();
+        if (SCM_IPORTP(port)) {
+            PORT_LOCK(port, vm);
+            PORT_LOCK(link, vm);
+            if (PORT_LINK(port) == link) {
+                PORT_LINK(port) = SCM_FALSE;
+                PORT_LINK(link) = SCM_FALSE;
+            }
+            PORT_UNLOCK(link);
+            PORT_UNLOCK(port);
+        } else {
+            PORT_LOCK(link, vm);
+            PORT_LOCK(port, vm);
+            if (PORT_LINK(port) == link) {
+                PORT_LINK(port) = SCM_FALSE;
+                PORT_LINK(link) = SCM_FALSE;
+            }
+            PORT_UNLOCK(port);
+            PORT_UNLOCK(link);
+        }
+    }
+}
+
+static inline void flush_linked_port(ScmPort *iport)
+{
+    ScmObj p = PORT_LINK(iport);
+    if (SCM_OPORTP(p)) Scm_Flush(SCM_PORT(p));
+}
+
+static inline void reset_linked_column(ScmPort *iport)
+{
+    ScmObj p = PORT_LINK(iport);
+    /* We don't lock p.  If another thread is writing to a linked port
+       something's already wrong, and even so, stepping on each other
+       about the column count won't create havoc. */
+    if (SCM_OPORTP(p))  PORT_COLUMN(p) = 0;
+}
+
 /*===============================================================
  * Locking ports
  */
@@ -263,7 +393,6 @@ ScmObj Scm_VMWithPortLocking(ScmPort *port SCM_UNUSED, ScmObj closure)
 
 /*===============================================================
  * Getting information
- * NB: Port attribute access API is in portapi.c
  */
 
 ScmObj Scm_PortName(ScmPort *port)
@@ -281,7 +410,12 @@ ScmSize Scm_PortBytes(ScmPort *port)
     return PORT_BYTES(port);
 }
 
-static void port_print(ScmObj obj, ScmPort *port, 
+ScmSize Scm_PortColumn(ScmPort *port)
+{
+    return PORT_COLUMN(port);
+}
+
+static void port_print(ScmObj obj, ScmPort *port,
                        ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(port, "#<%s%sport%s %A %p>",
@@ -494,7 +628,7 @@ int Scm_PortPositionable(ScmPort *port, int setp)
         if (setp) return FALSE; /* we haven't supported setpos for ostr */
         else      return TRUE;
     }
-    return FALSE;		/* dummy */
+    return FALSE;               /* dummy */
 }
 
 /* Duplicates the file descriptor of the source port, and set it to
@@ -536,7 +670,7 @@ void Scm_PortFdDup(ScmPort *dst, ScmPort *src)
         Scm_Flush(dst);
     }
     /*  NB: We don't retry dup2().  By the time it returns EINTR, the
-        dstfd has actually been closed, and if other thread happens to 
+        dstfd has actually been closed, and if other thread happens to
         grab the same fd, retrying dup2() inadvertently closes that one.
     */
 
@@ -617,6 +751,103 @@ int Scm_FdReady(int fd, int dir)
 #endif /*!HAVE_SELECT && !GAUCHE_WINDOWS */
 }
 
+/*=================================================================
+ * Port Attributes
+ *
+ * Port attributes are stored in alist.  Each entry is either one
+ * of the following form:
+ *  (key value)          Just a value
+ *  (key value . #f)     Just a value, read-only.  This is a system
+ *                       attribute; no public interface to create this
+ *                       kind of API is provided.  It is also undeletable.
+ *                       (Currently, only the internal port constructor
+ *                       creates this kind of attributes.)
+ *
+ * Access to the port attributes is mutexed by a separate lock than
+ * the port's main lock, for the attributes can be accessed from a
+ * thread while another thread is reading from/writing to it (e.g. taking
+ * the port name for display).
+ */
+
+ScmObj Scm_PortAttrGet(ScmPort *p, ScmObj key, ScmObj fallback)
+{
+    ScmObj r = SCM_UNBOUND;
+
+    /* We assume pointer op is atomic */
+    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
+    if (SCM_PAIRP(v)) {
+        SCM_ASSERT(SCM_PAIRP(SCM_CDR(v)));
+        r = SCM_CADR(v);
+    } else {
+        r = fallback;
+    }
+
+    if (SCM_UNBOUNDP(r)) {
+        Scm_Error("No port attribute for key %S in port %S", key, SCM_OBJ(p));
+    }
+    return r;
+}
+
+ScmObj Scm_PortAttrSet(ScmPort *p, ScmObj key, ScmObj val)
+{
+    volatile int err_readonly = FALSE;
+    volatile int exists = FALSE;
+
+    (void)SCM_INTERNAL_FASTLOCK_LOCK(P_(p)->lock);
+    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
+    if (SCM_PAIRP(v)) {
+        SCM_ASSERT(SCM_PAIRP(SCM_CDR(v)));
+        exists = TRUE;
+        if (SCM_NULLP(SCM_CDDR(v))) {
+            SCM_SET_CAR_UNCHECKED(SCM_CDR(v), val);
+        } else {
+            err_readonly = TRUE;
+        }
+    } else {
+        PORT_ATTRS(p) = Scm_Cons(SCM_LIST2(key, val), PORT_ATTRS(p));
+    }
+    (void)SCM_INTERNAL_FASTLOCK_UNLOCK(P_(p)->lock);
+
+    if (err_readonly) {
+        Scm_Error("Port attribute '%A' is read-only in port: %S",
+                  key, SCM_OBJ(p));
+    }
+    return SCM_MAKE_BOOL(exists);
+}
+
+ScmObj Scm_PortAttrDelete(ScmPort *p, ScmObj key)
+{
+    int err_undeletable = FALSE;
+
+    (void)SCM_INTERNAL_FASTLOCK_LOCK(P_(p)->lock);
+    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
+    if (SCM_PAIRP(v) && SCM_PAIRP(SCM_CDR(v)) && SCM_FALSEP(SCM_CDDR(v))) {
+        err_undeletable = TRUE;
+    } else {
+        PORT_ATTRS(p) = Scm_AssocDelete(key, PORT_ATTRS(p), SCM_CMP_EQ);
+    }
+    (void)SCM_INTERNAL_FASTLOCK_UNLOCK(P_(p)->lock);
+    if (err_undeletable) {
+        Scm_Error("Port attribute '%A' is not deletable from port: %S",
+                  key, SCM_OBJ(p));
+    }
+    return SCM_UNDEFINED;       /* we may return more useful info in future */
+}
+
+ScmObj Scm_PortAttrs(ScmPort *p)
+{
+    ScmObj h = SCM_NIL, t = SCM_NIL;
+    ScmObj cp;
+    SCM_FOR_EACH(cp, PORT_ATTRS(p)) {
+        SCM_ASSERT(SCM_PAIRP(SCM_CAR(cp)) && SCM_PAIRP(SCM_CDAR(cp)));
+        ScmObj e = SCM_CAR(cp);
+        ScmObj k = SCM_CAR(e);
+        ScmObj v = SCM_CADR(e);
+        SCM_APPEND1(h, t, Scm_Cons(k, v));
+    }
+    return h;
+}
+
 /*===============================================================
  * buffered Port
  *  - mainly used for buffered file I/O, but can also be used
@@ -822,6 +1053,16 @@ ScmObj Scm_MakeBufferedPortFull(ScmClass *klass,
     PORT_BUF(p)->filenum = bufrec->filenum;
     PORT_BUF(p)->seeker = bufrec->seeker;
     PORT_BUF(p)->data = bufrec->data;
+    PORT_BUF(p)->idata = bufrec->idata;
+
+    /* When the port is GC'ed without properly closed, a finalizer is
+       run to clean up.  However, at that moment, PORT_BUF(p)->data
+       may be already collected.  We should detect the case.
+     */
+    if (GC_is_heap_ptr(bufrec->data)) {
+        PORT_FILE_EXTDATA_SET(p);
+        GC_register_disappearing_link((void**)&PORT_BUF(p)->data);
+    }
 
     if (flags & SCM_PORT_WITH_POSITION) {
         PORT_BUF(p)->getpos = bufrec->getpos;
@@ -1014,7 +1255,7 @@ static struct {
 } active_buffered_ports;
 
 #define PORT_HASH(port)  \
-    ((((SCM_WORD(port)>>3) * 2654435761UL)>>16) % PORT_VECTOR_SIZE)
+    ((u_long)(((SCM_WORD(port)>>3) * 2654435761UL)>>16) % PORT_VECTOR_SIZE)
 
 static void register_buffered_port(ScmPort *port)
 {
@@ -1190,17 +1431,14 @@ ScmObj Scm_GetBufferingMode(ScmPort *port)
  * File Port
  */
 
-/* This small piece of data is kept in port->src.buf.data. */
-typedef struct file_port_data_rec {
-    int fd;
-} file_port_data;
-
-#define FILE_PORT_DATA(p) ((file_port_data*)(PORT_BUF(p)->data))
+/* We keep fd to PORT_BUF(p)->idata */
+#define FILE_PORT_FD(p) ((int)(PORT_BUF(p)->idata))
+#define FILE_PORT_FD_SET(p, fd) (PORT_BUF(p)->idata = (fd))
 
 static ScmSize file_filler(ScmPort *p, ScmSize cnt)
 {
     ScmSize nread = 0;
-    int fd = FILE_PORT_DATA(p)->fd;
+    int fd = FILE_PORT_FD(p);
     char *datptr = PORT_BUF(p)->end;
     SCM_ASSERT(fd >= 0);
     while (nread == 0) {
@@ -1225,7 +1463,7 @@ static ScmSize file_flusher(ScmPort *p, ScmSize cnt, int forcep)
 {
     ScmSize nwrote = 0;
     ScmSize datsiz = PORT_BUFFER_AVAIL(p);
-    int fd = FILE_PORT_DATA(p)->fd;
+    int fd = FILE_PORT_FD(p);
     char *datptr = PORT_BUF(p)->buffer;
 
     SCM_ASSERT(fd >= 0);
@@ -1235,7 +1473,7 @@ static ScmSize file_flusher(ScmPort *p, ScmSize cnt, int forcep)
         errno = 0;
         SCM_SYSCALL(r, write(fd, datptr, datsiz-nwrote));
         if (r < 0) {
-            if (PORT_BUFFER_SIGPIPE_SENSITIVE_P(p)) {
+            if (errno == EPIPE && PORT_BUFFER_SIGPIPE_SENSITIVE_P(p)) {
                 /* (sort of) emulate termination by SIGPIPE.
                    NB: The difference is visible from the outside world
                    as the process exit status differ (WIFEXITED
@@ -1256,12 +1494,12 @@ static ScmSize file_flusher(ScmPort *p, ScmSize cnt, int forcep)
 
 static void file_closer(ScmPort *p)
 {
-    int fd = FILE_PORT_DATA(p)->fd;
+    int fd = FILE_PORT_FD(p);
     if (fd >= 0) {
         /* If close() fails, the port's CLOSED flag isn't set and file_closer
            may be called again (probably via finalizer).  We don't want to call
            close() again and raise an error. */
-        FILE_PORT_DATA(p)->fd = -1;
+        FILE_PORT_FD_SET(p, -1);
         if (close(fd) < 0) {
             Scm_SysError("close() failed on %S", SCM_OBJ(p));
         }
@@ -1270,19 +1508,19 @@ static void file_closer(ScmPort *p)
 
 static int file_ready(ScmPort *p)
 {
-    int fd = FILE_PORT_DATA(p)->fd;
+    int fd = FILE_PORT_FD(p);
     SCM_ASSERT(fd >= 0);
     return Scm_FdReady(fd, SCM_PORT_DIR(p));
 }
 
 static int file_filenum(ScmPort *p)
 {
-    return FILE_PORT_DATA(p)->fd;
+    return FILE_PORT_FD(p);
 }
 
 static off_t file_seeker(ScmPort *p, off_t offset, int whence)
 {
-    return lseek(FILE_PORT_DATA(p)->fd, offset, whence);
+    return lseek(FILE_PORT_FD(p), offset, whence);
 }
 
 /* Kludge: We should have better way */
@@ -1296,7 +1534,7 @@ static void file_buffered_port_set_fd(ScmPort *p, int fd)
     if (!file_buffered_port_p(p)) {
         Scm_Error("port is not directly conntect to fd: %S", p);
     }
-    FILE_PORT_DATA(p)->fd = fd;
+    FILE_PORT_FD_SET(p, fd);
 }
 
 ScmObj Scm_OpenFilePort(const char *path, int flags, int buffering, int perm)
@@ -1325,9 +1563,6 @@ ScmObj Scm_OpenFilePort(const char *path, int flags, int buffering, int perm)
     */
     if (flags & O_APPEND) (void)lseek(fd, 0, SEEK_END);
 
-    file_port_data *data = SCM_NEW(file_port_data);
-    data->fd = fd;
-    
     ScmPortBuffer bufrec;
     bufrec.mode = buffering;
     bufrec.buffer = NULL;
@@ -1338,7 +1573,8 @@ ScmObj Scm_OpenFilePort(const char *path, int flags, int buffering, int perm)
     bufrec.ready = file_ready;
     bufrec.filenum = file_filenum;
     bufrec.seeker = file_seeker;
-    bufrec.data = data;
+    bufrec.data = NULL;
+    bufrec.idata = fd;
     ScmObj p = Scm_MakeBufferedPort(SCM_CLASS_PORT, SCM_MAKE_STR_COPYING(path),
                                     dir, TRUE, &bufrec);
     return p;
@@ -1354,9 +1590,6 @@ ScmObj Scm_OpenFilePort(const char *path, int flags, int buffering, int perm)
 ScmObj Scm_MakePortWithFd(ScmObj name, int direction,
                           int fd, int bufmode, int ownerp)
 {
-    file_port_data *data = SCM_NEW(file_port_data);
-    data->fd = fd;
-
     ScmPortBuffer bufrec;
     bufrec.buffer = NULL;
     bufrec.size = 0;
@@ -1366,7 +1599,8 @@ ScmObj Scm_MakePortWithFd(ScmObj name, int direction,
     bufrec.closer = file_closer;
     bufrec.ready = file_ready;
     bufrec.filenum = file_filenum;
-    bufrec.data = data;
+    bufrec.data = NULL;
+    bufrec.idata = fd;
 
     /* Check if the given fd is seekable, and set seeker if so. */
     if (lseek(fd, 0, SEEK_CUR) < 0) {
@@ -1388,11 +1622,10 @@ ScmObj Scm_MakeInputStringPortFull(ScmString *str, ScmObj name,
                                    u_long flags)
 {
     ScmPort *p = make_port(SCM_CLASS_PORT, name, SCM_PORT_INPUT, SCM_PORT_ISTR);
-    ScmSmallInt size;
-    const char *s = Scm_GetStringContent(str, &size, NULL, NULL);
-    PORT_ISTR(p)->start = s;
-    PORT_ISTR(p)->current = s;
-    PORT_ISTR(p)->end = s + size;
+    const ScmStringBody *sb = SCM_STRING_BODY(str);
+    PORT_ISTR(p)->start = SCM_STRING_BODY_START(sb);
+    PORT_ISTR(p)->current = SCM_STRING_BODY_START(sb);
+    PORT_ISTR(p)->end = SCM_STRING_BODY_END(sb);
     if (flags&SCM_PORT_STRING_PRIVATE) PORT_PRELOCK(p, Scm_VM());
     return SCM_OBJ(p);
 }
@@ -1438,22 +1671,6 @@ ScmObj Scm_GetOutputStringUnsafe(ScmPort *port, int flags)
     return Scm_DStringGet(PORT_OSTR(port), flags);
 }
 
-#if GAUCHE_API_VERSION < 1000
-/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
-   Will be removed on 1.0 */
-ScmObj Scm__GetOutputStringCompat(ScmPort *port)
-{
-    return Scm_GetOutputString(port, 0);
-}
-
-/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
-   Will be removed on 1.0 */
-ScmObj Scm__GetOutputStringUnsafeCompat(ScmPort *port)
-{
-    return Scm_GetOutputStringUnsafe(port, 0);
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
-
 static ScmObj get_remaining_input_string_aux(const char *s, ScmSize ssiz,
                                              const char *p, ScmSize psiz,
                                              int flags);
@@ -1481,7 +1698,7 @@ ScmObj Scm_GetRemainingInputString(ScmPort *port, int flags)
             return Scm_MakeString(cp, (int)(ep-cp), -1, flags);
         } else {
             /* we need to copy */
-            return get_remaining_input_string_aux(cp, ep-cp, 
+            return get_remaining_input_string_aux(cp, ep-cp,
                                                   cbuf, nbytes, flags);
         }
     } else if (port->scrcnt > 0) {
@@ -1512,15 +1729,6 @@ static ScmObj get_remaining_input_string_aux(const char *s, ScmSize ssiz,
     return Scm_MakeString(b, psiz+ssiz, -1, flags);
 }
 
-#if GAUCHE_API_VERSION < 1000
-/* TRANSIENT: Pre-0.9 Compatibility routine.  Kept for the binary compatibility.
-   Will be removed on 1.0 */
-ScmObj Scm__GetRemainingInputStringCompat(ScmPort *port)
-{
-    return Scm_GetRemainingInputString(port, 0);
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
-
 /*===============================================================
  * Procedural port
  */
@@ -1534,12 +1742,12 @@ ScmObj Scm__GetRemainingInputStringCompat(ScmPort *port)
 /* default dummy procedures */
 static int null_getb(ScmPort *dummy SCM_UNUSED)
 {
-    return SCM_CHAR_INVALID;
+    return EOF;
 }
 
 static int null_getc(ScmPort *dummy SCM_UNUSED)
 {
-    return SCM_CHAR_INVALID;
+    return EOF;
 }
 
 static ScmSize null_getz(char *buf SCM_UNUSED,
@@ -1563,7 +1771,7 @@ static void null_putc(ScmChar c SCM_UNUSED, ScmPort *dummy SCM_UNUSED)
 }
 
 static void null_putz(const char *str SCM_UNUSED,
-                      ScmSize len SCM_UNUSED, 
+                      ScmSize len SCM_UNUSED,
                       ScmPort *dummy SCM_UNUSED)
 {
 }
@@ -1599,7 +1807,7 @@ ScmObj Scm_MakeVirtualPortFull(ScmClass *klass, ScmObj name,
     PORT_VT(p)->GetPos = NULL;
     PORT_VT(p)->SetPos = NULL;
     PORT_VT(p)->flags = 0;
-    
+
     if (vtable->Getb)  PORT_VT(p)->Getb   = vtable->Getb;
     if (vtable->Getc)  PORT_VT(p)->Getc   = vtable->Getc;
     if (vtable->Getz)  PORT_VT(p)->Getz   = vtable->Getz;
@@ -1611,7 +1819,17 @@ ScmObj Scm_MakeVirtualPortFull(ScmClass *klass, ScmObj name,
     if (vtable->Flush) PORT_VT(p)->Flush  = vtable->Flush;
     if (vtable->Close) PORT_VT(p)->Close  = vtable->Close;
     if (vtable->Seek)  PORT_VT(p)->Seek   = vtable->Seek;
-    PORT_VT(p)->data = vtable->data;
+    PORT_VT(p)->data  = vtable->data;
+    PORT_VT(p)->idata = vtable->idata;
+
+    /* When the port is GC'ed without properly closed, a finalizer is
+       run to clean up.  However, at that moment, PORT_VT(p)->data
+       may be already collected.  We should detect the case.
+     */
+    if (GC_is_heap_ptr(vtable->data)) {
+        PORT_PROC_EXTDATA_SET(p);
+        GC_register_disappearing_link((void**)&PORT_VT(p)->data);
+    }
 
     if (flags & SCM_PORT_WITH_POSITION) {
         if (vtable->GetPos) PORT_VT(p)->GetPos = vtable->GetPos;
@@ -1895,44 +2113,54 @@ DEFSTDPORT(Stdin, scm_stdin)
 DEFSTDPORT(Stdout, scm_stdout)
 DEFSTDPORT(Stderr, scm_stderr)
 
+/* Parameters */
+static ScmPrimitiveParameter *scm_curin = NULL;
+static ScmPrimitiveParameter *scm_curout = NULL;
+static ScmPrimitiveParameter *scm_curerr = NULL;
+
+ScmPort *Scm_CurrentInputPort()
+{
+    return SCM_PORT(Scm_PrimitiveParameterRef(Scm_VM(), scm_curin));
+}
+
+ScmPort *Scm_CurrentOutputPort()
+{
+    return SCM_PORT(Scm_PrimitiveParameterRef(Scm_VM(), scm_curout));
+}
+
+ScmPort *Scm_CurrentErrorPort()
+{
+    return SCM_PORT(Scm_PrimitiveParameterRef(Scm_VM(), scm_curerr));
+}
+
 ScmObj Scm_SetCurrentInputPort(ScmPort *port)
 {
-    ScmVM *vm = Scm_VM();
-    ScmObj oldp = SCM_OBJ(SCM_VM_CURRENT_INPUT_PORT(vm));
-    SCM_VM_CURRENT_INPUT_PORT(vm) = port;
-    return oldp;
+    return Scm_PrimitiveParameterSet(Scm_VM(), scm_curin, SCM_OBJ(port));
 }
 
 ScmObj Scm_SetCurrentOutputPort(ScmPort *port)
 {
-    ScmVM *vm = Scm_VM();
-    ScmObj oldp = SCM_OBJ(SCM_VM_CURRENT_OUTPUT_PORT(vm));
-    SCM_VM_CURRENT_OUTPUT_PORT(vm) = port;
-    return oldp;
+    return Scm_PrimitiveParameterSet(Scm_VM(), scm_curout, SCM_OBJ(port));
 }
 
 ScmObj Scm_SetCurrentErrorPort(ScmPort *port)
 {
-    ScmVM *vm = Scm_VM();
-    ScmObj oldp = SCM_OBJ(SCM_VM_CURRENT_ERROR_PORT(vm));
-    SCM_VM_CURRENT_ERROR_PORT(vm) = port;
-    return oldp;
+    return Scm_PrimitiveParameterSet(Scm_VM(), scm_curerr, SCM_OBJ(port));
 }
 
 /*===============================================================
  * Initialization
  */
-
 void Scm__InitPort(void)
 {
     if (sizeof(ScmPort) < sizeof(ScmPortImpl)) {
-	fprintf(stderr,
-		"sizeof(ScmPort) [%"PRIdPTR"] is smaller than "
-		"sizeof(ScmPortImpl) [%"PRIdPTR"]\n",
-		SCM_WORD(sizeof(ScmPort)), SCM_WORD(sizeof(ScmPortImpl)));
+        fprintf(stderr,
+                "sizeof(ScmPort) [%"PRIdPTR"] is smaller than "
+                "sizeof(ScmPortImpl) [%"PRIdPTR"]\n",
+                SCM_WORD(sizeof(ScmPort)), SCM_WORD(sizeof(ScmPortImpl)));
         Scm_Panic("Implementation error.  Exitting.");
     }
-    
+
     (void)SCM_INTERNAL_MUTEX_INIT(active_buffered_ports.mutex);
     active_buffered_ports.ports = SCM_WEAK_VECTOR(Scm_MakeWeakVector(PORT_VECTOR_SIZE));
 
@@ -1963,11 +2191,15 @@ void Scm__InitPort(void)
                                      | SCM_PORT_BUFFER_SIGPIPE_SENSITIVE),
                                     TRUE);
 
-    /* The root VM is initialized with bogus standard ports; we need to
-       reset them. */
-    Scm_VM()->curin  = SCM_PORT(scm_stdin);
-    Scm_VM()->curout = SCM_PORT(scm_stdout);
-    Scm_VM()->curerr = SCM_PORT(scm_stderr);
+    scm_curin  = Scm_BindPrimitiveParameter(Scm_SchemeModule(),
+                                            "current-input-port",
+                                            scm_stdin, 0);
+    scm_curout = Scm_BindPrimitiveParameter(Scm_SchemeModule(),
+                                            "current-output-port",
+                                            scm_stdout, 0);
+    scm_curerr = Scm_BindPrimitiveParameter(Scm_GaucheModule(),
+                                            "current-error-port",
+                                            scm_stderr, 0);
 
     key_full   = Scm_MakeKeyword(SCM_STRING(SCM_MAKE_STR("full")));
     key_modest = Scm_MakeKeyword(SCM_STRING(SCM_MAKE_STR("modest")));
@@ -2112,9 +2344,8 @@ static ScmObj make_trapper_port(ScmObj name, int direction,
     bufrec.closer = file_closer;
     bufrec.ready = file_ready;
     bufrec.filenum = file_filenum;
-    file_port_data *data = SCM_NEW(file_port_data);
-    data->fd = fd;
-    bufrec.data = data;
+    bufrec.data = NULL;
+    bufrec.idata = fd;
     bufrec.seeker = NULL;
     ScmObj p = Scm_MakeBufferedPort(SCM_CLASS_PORT, name, direction, TRUE,
                                     &bufrec);
@@ -2153,9 +2384,9 @@ void Scm__SetupPortsForWindows(int has_console)
             scm_stderr = make_trapper_port(SCM_MAKE_STR("(standard error output)"),
                                            SCM_PORT_OUTPUT, 2,
                                            SCM_PORT_BUFFER_NONE);
-            Scm_VM()->curin  = SCM_PORT(scm_stdin);
-            Scm_VM()->curout = SCM_PORT(scm_stdout);
-            Scm_VM()->curerr = SCM_PORT(scm_stderr);
+            Scm_SetCurrentInputPort(SCM_PORT(scm_stdin));
+            Scm_SetCurrentOutputPort(SCM_PORT(scm_stdout));
+            Scm_SetCurrentErrorPort(SCM_PORT(scm_stderr));
         }
         (void)orig_stdin;  /* suppress unused var warning */
         (void)orig_stdout; /* suppress unused var warning */
