@@ -1,7 +1,7 @@
 /*
  * portapi.c - port common API
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -104,6 +104,35 @@
     } while (0)
 #endif /* WALKER_CHECK */
 
+/* A utility to compute the current position for Puts and Putz.
+   We only need to count from the last newline.  If the string
+   doesn't contain a newline, returns -1.
+   TODO: Should we consider character width? */
+#ifndef COLUMN_COUNT
+#define COLUMN_COUNT
+static int length_count(const u_char *start, const u_char *end)
+{
+    int col = 0;
+    while (start < end) {
+        u_char b = *start;
+        int i = SCM_CHAR_NFOLLOWS(b);
+        if (i < 0) break;
+        col++;
+        start += i+1;
+    }
+    return col;
+}
+
+static int column_count(const u_char *start, const u_char *end)
+{
+    const u_char *p = end-1;
+    for (; p >= start; p--) {
+        if (*p == '\n') return length_count(p+1, end);
+    }
+    return -1;
+}
+#endif /*COLUMN_COUNT*/
+
 /*=================================================================
  * Putb
  */
@@ -119,6 +148,7 @@ void Scm_PutbUnsafe(ScmByte b, ScmPort *p)
     WALKER_CHECK(p);
     LOCK(p);
     CLOSE_CHECK(p);
+    PORT_FLUSHED_CLEAR(p);
 
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE:
@@ -145,6 +175,10 @@ void Scm_PutbUnsafe(ScmByte b, ScmPort *p)
         Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
                       "bad port type for output: %S", p);
     }
+
+    /* Binary output doesn't increment column count, but we can reset
+       it when newline is put. */
+    if (b == '\n') PORT_COLUMN(p) = 0;
 }
 
 /*=================================================================
@@ -162,6 +196,7 @@ void Scm_PutcUnsafe(ScmChar c, ScmPort *p)
     WALKER_CHECK(p);
     LOCK(p);
     CLOSE_CHECK(p);
+    PORT_FLUSHED_CLEAR(p);
 
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE: {
@@ -196,6 +231,10 @@ void Scm_PutcUnsafe(ScmChar c, ScmPort *p)
         Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
                       "bad port type for output: %S", p);
     }
+
+    /* TODO: Should we consider character width? */
+    if (c == '\n') PORT_COLUMN(p) = 0;
+    else PORT_COLUMN(p)++;
 }
 
 /*=================================================================
@@ -213,13 +252,13 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
     WALKER_CHECK(p);
     LOCK(p);
     CLOSE_CHECK(p);
+    PORT_FLUSHED_CLEAR(p);
 
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE: {
         ScmSmallInt size;
         const char *ss = Scm_GetStringContent(s, &size, NULL, NULL);
         SAFE_CALL(p, bufport_write(p, ss, size));
-
         if (PORT_BUFFER_MODE(p) == SCM_PORT_BUFFER_LINE) {
             const char *cp = PORT_BUF(p)->current;
             while (cp-- > PORT_BUF(p)->buffer) {
@@ -248,6 +287,12 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
         Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
                       "bad port type for output: %S", p);
     }
+
+    const ScmStringBody *sb = SCM_STRING_BODY(s);
+    int col = column_count((const u_char*)SCM_STRING_BODY_START(sb),
+                           (const u_char*)SCM_STRING_BODY_END(sb));
+    if (col < 0) PORT_COLUMN(p) += SCM_STRING_BODY_LENGTH(sb);
+    else PORT_COLUMN(p) = col;
 }
 
 /*=================================================================
@@ -255,7 +300,7 @@ void Scm_PutsUnsafe(ScmString *s, ScmPort *p)
  */
 
 #ifdef SAFE_PORT_OP
-void Scm_Putz(const char *s, volatile ScmSize siz, ScmPort *p)
+void Scm_Putz(const char * volatile s, volatile ScmSize siz, ScmPort *p)
 #else
 void Scm_PutzUnsafe(const char *s, volatile ScmSize siz, ScmPort *p)
 #endif
@@ -265,6 +310,7 @@ void Scm_PutzUnsafe(const char *s, volatile ScmSize siz, ScmPort *p)
     WALKER_CHECK(p);
     LOCK(p);
     CLOSE_CHECK(p);
+    PORT_FLUSHED_CLEAR(p);
     if (siz < 0) siz = (ScmSize)strlen(s);
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE:
@@ -296,6 +342,14 @@ void Scm_PutzUnsafe(const char *s, volatile ScmSize siz, ScmPort *p)
         Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
                       "bad port type for output: %S", p);
     }
+
+    {
+        const u_char *start = (const u_char*)s;
+        const u_char *end = (const u_char*)(s + siz);
+        int col = column_count(start, end);
+        if (col < 0) PORT_COLUMN(p) += length_count(start, end);
+        else PORT_COLUMN(p) = col;
+    }
 }
 
 /*=================================================================
@@ -312,6 +366,10 @@ void Scm_FlushUnsafe(ScmPort *p)
     SHORTCUT(p, Scm_FlushUnsafe(p); return);
     WALKER_CHECK(p);
     LOCK(p);
+    if (PORT_FLUSHED_P(p)) {
+        UNLOCK(p);
+        return;
+    }
     CLOSE_CHECK(p);
     switch (SCM_PORT_TYPE(p)) {
     case SCM_PORT_FILE:
@@ -330,6 +388,7 @@ void Scm_FlushUnsafe(ScmPort *p)
         Scm_PortError(p, SCM_PORT_ERROR_OUTPUT,
                       "bad port type for output: %S", p);
     }
+    PORT_FLUSHED_SET(p);
 }
 
 /*=================================================================
@@ -369,7 +428,7 @@ ScmChar Scm_PeekcUnsafe(ScmPort *p)
     if (ch == SCM_CHAR_INVALID) {
         volatile ScmObj saved_pos = SCM_UNBOUND;
         UNSAVE_POS(p);
-        if (SCM_PORT_TYPE(p) == SCM_PORT_PROC 
+        if (SCM_PORT_TYPE(p) == SCM_PORT_PROC
             && Scm_PortPositionable(p, FALSE)) {
             SAFE_CALL(p, saved_pos = Scm_PortSeekUnsafe(p, SCM_MAKE_INT(0), SEEK_CUR));
         }
@@ -436,11 +495,14 @@ int Scm_PeekbUnsafe(ScmPort *p)
     VMDECL;
     SHORTCUT(p, return Scm_PeekbUnsafe(p));
     LOCK(p);
+    flush_linked_port(p);
     if (p->scrcnt > 0) {
         b = (unsigned char)PORT_SCRATCH(p)[0];
     } else {
         SCM_GETB(b, p);
-        if (b >= 0) {
+        if (b == EOF) {
+            PORT_UNGOTTEN(p) = EOF;
+        } else if (b >= 0) {
             if (p->scrcnt > 0) {
                 /* unshift scratch buffer */
                 SCM_ASSERT(p->scrcnt < SCM_CHAR_MAX_BYTES);
@@ -525,11 +587,17 @@ int Scm_GetbUnsafe(ScmPort *p)
     SHORTCUT(p, return Scm_GetbUnsafe(p));
     LOCK(p);
     CLOSE_CHECK(p);
+    flush_linked_port(p);
 
     /* check if there's "pushed back" stuff */
     if (p->scrcnt) {
         b = getb_scratch(p);
     } else if (PORT_UNGOTTEN(p) != SCM_CHAR_INVALID) {
+        if (PORT_UNGOTTEN(p) == EOF) {
+            PORT_UNGOTTEN(p) = SCM_CHAR_INVALID;
+            UNLOCK(p);
+            return EOF;
+        }
         b = getb_ungotten(p);
     } else {
         switch (SCM_PORT_TYPE(p)) {
@@ -558,7 +626,10 @@ int Scm_GetbUnsafe(ScmPort *p)
         }
         PORT_BYTES(p)++;
         /* we may mix binary/textual input, so we keep lines updated too. */
-        if (b == '\n') PORT_LINE(p)++;
+        if (b == '\n') {
+            PORT_LINE(p)++;
+            reset_linked_column(p);
+        }
     }
     UNLOCK(p);
     return b;
@@ -618,6 +689,7 @@ int Scm_GetcUnsafe(ScmPort *p)
     SHORTCUT(p, return Scm_GetcUnsafe(p));
     LOCK(p);
     CLOSE_CHECK(p);
+    flush_linked_port(p);
     if (p->scrcnt > 0) {
         int r = GETC_SCRATCH(p);
         UNLOCK(p);
@@ -684,7 +756,10 @@ int Scm_GetcUnsafe(ScmPort *p)
             PORT_BYTES(p) += nb;
         } else {
             c = first;
-            if (c == '\n') PORT_LINE(p)++;
+            if (c == '\n') {
+                PORT_LINE(p)++;
+                reset_linked_column(p);
+            }
         }
         UNLOCK(p);
         return c;
@@ -710,7 +785,10 @@ int Scm_GetcUnsafe(ScmPort *p)
             PORT_BYTES(p) += nb;
         } else {
             c = first;
-            if (c == '\n') PORT_LINE(p)++;
+            if (c == '\n') {
+                PORT_LINE(p)++;
+                reset_linked_column(p);
+            }
         }
         UNLOCK(p);
         return c;
@@ -719,7 +797,10 @@ int Scm_GetcUnsafe(ScmPort *p)
         int c = 0;
         UNSAVE_POS(p);
         SAFE_CALL(p, c = PORT_VT(p)->Getc(p));
-        if (c == '\n') PORT_LINE(p)++;
+        if (c == '\n') {
+            PORT_LINE(p)++;
+            reset_linked_column(p);
+        }
         UNLOCK(p);
         return c;
     }
@@ -790,6 +871,7 @@ ScmSize Scm_GetzUnsafe(char *buf, ScmSize buflen, ScmPort *p)
     SHORTCUT(p, return Scm_GetzUnsafe(buf, buflen, p));
     LOCK(p);
     CLOSE_CHECK(p);
+    flush_linked_port(p);
 
     if (p->scrcnt) {
         ScmSize r = GETZ_SCRATCH(buf, buflen, p);
@@ -811,7 +893,7 @@ ScmSize Scm_GetzUnsafe(char *buf, ScmSize buflen, ScmPort *p)
         SAFE_CALL(p, siz = bufport_read(p, buf, buflen));
         PORT_BYTES(p) += siz;
         UNLOCK(p);
-        if (siz == 0) return EOF;
+        if (siz == 0 && buflen != 0) return EOF;
         else return siz;
     }
     case SCM_PORT_ISTR: {
@@ -878,6 +960,7 @@ ScmObj readline_body(ScmPort *p)
         b1 = Scm_GetbUnsafe(p);
     }
     PORT_LINE(p)++;
+    reset_linked_column(p);
     return Scm_DStringGet(&ds, 0);
 }
 #endif /* READLINE_AUX */
@@ -893,6 +976,7 @@ ScmObj Scm_ReadLineUnsafe(ScmPort *p)
     SHORTCUT(p, return Scm_ReadLineUnsafe(p));
 
     LOCK(p);
+    flush_linked_port(p);
     SAFE_CALL(p, r = readline_body(p));
     UNLOCK(p);
     return r;
@@ -1067,7 +1151,7 @@ ScmObj Scm_GetPortPositionUnsafe(ScmPort *p)
                 err_disabled = TRUE;
                 break;
             }
-            
+
             if (SCM_PORT_DIR(p)&SCM_PORT_INPUT) {
                 rr -= (off_t)(PORT_BUF(p)->end - PORT_BUF(p)->current);
             } else {
@@ -1076,7 +1160,7 @@ ScmObj Scm_GetPortPositionUnsafe(ScmPort *p)
             r = Scm_OffsetToInteger(rr);
         }
         break;
-    case SCM_PORT_ISTR: 
+    case SCM_PORT_ISTR:
         r = Scm_OffsetToInteger(PORT_ISTR(p)->current - PORT_ISTR(p)->start);
         break;
     case SCM_PORT_OSTR:
@@ -1104,11 +1188,11 @@ ScmObj Scm_GetPortPositionUnsafe(ScmPort *p)
     UNLOCK(p);
 
     if (err_disabled) {
-        Scm_PortError(p, SCM_PORT_ERROR_SEEK, 
+        Scm_PortError(p, SCM_PORT_ERROR_SEEK,
                       "getting port position is disabled");
     }
     if (err_badpos) {
-        Scm_PortError(p, SCM_PORT_ERROR_SEEK, 
+        Scm_PortError(p, SCM_PORT_ERROR_SEEK,
                       "getpos method returned invalid position: %S", r);
     }
 
@@ -1169,7 +1253,7 @@ static ScmObj set_port_position(ScmPort *p, ScmObj pos, int whence)
                     off = Scm_Sub(off, Scm_MakeIntegerU(PORT_BUF(p)->end - c));
                 }
                 PORT_BUF(p)->current = PORT_BUF(p)->end; /* invalidate buffer */
-                SAFE_CALL(p, rr = PORT_BUF(p)->seeker(p, 
+                SAFE_CALL(p, rr = PORT_BUF(p)->seeker(p,
                                                       Scm_IntegerToOffset(off),
                                                       whence));
                 if (rr == (off_t)-1) {
@@ -1178,7 +1262,7 @@ static ScmObj set_port_position(ScmPort *p, ScmObj pos, int whence)
                 }
             } else {
                 SAFE_CALL(p, bufport_flush(p, 0, TRUE));
-                SAFE_CALL(p, rr = PORT_BUF(p)->seeker(p, 
+                SAFE_CALL(p, rr = PORT_BUF(p)->seeker(p,
                                                       Scm_IntegerToOffset(off),
                                                       whence));
             }
@@ -1191,7 +1275,7 @@ static ScmObj set_port_position(ScmPort *p, ScmObj pos, int whence)
             err_disabled = TRUE;
         }
         break;
-    case SCM_PORT_ISTR: 
+    case SCM_PORT_ISTR:
         r = SEEK_ISTR(p, off, whence);
         break;
     case SCM_PORT_OSTR:
@@ -1218,7 +1302,7 @@ static ScmObj set_port_position(ScmPort *p, ScmObj pos, int whence)
     UNLOCK(p);
 
     if (err_disabled) {
-        Scm_PortError(p, SCM_PORT_ERROR_SEEK, 
+        Scm_PortError(p, SCM_PORT_ERROR_SEEK,
                       "setting port position is disabled");
     }
     return r;
@@ -1252,178 +1336,6 @@ ScmObj Scm_PortSeekUnsafe(ScmPort *p, ScmObj off, int whence)
     }
 }
 
-/*=================================================================
- * Port Attributes
- *
- * Port attributes are stored in alist.  Each entry is either one
- * of the following form:
- *  (key value)          Just a value
- *  (key value . #f)     Just a value, read-only.  This is a system
- *                       attribute; no public interface to create this
- *                       kind of API is provided.  It is also undeletable.
- *                       (Currently, only the internal port constructor
- *                       creates this kind of attributes.)
- *  (key getter setter)  Procedurally handled value.  Getter will be
- *                       called as (getter port [fallback]), and Setter
- *                       will be called as (setter port value).
- *                       Setter can be #f if the attr is read-only.
- *                       Port is locked while getter and setter is called.
- *
- * The third type of attribute can be created by Scm_PortAttrCreate.
- *
- * TODO: probably we want deletable and undeletable procedural values.
- */
-
-#ifdef SAFE_PORT_OP
-ScmObj Scm_PortAttrGet(ScmPort *p, ScmObj key, ScmObj fallback)
-#else
-ScmObj Scm_PortAttrGetUnsafe(ScmPort *p, ScmObj key, ScmObj fallback)
-#endif
-{
-    ScmObj r = SCM_UNBOUND;
-    VMDECL;
-    SHORTCUT(p, return Scm_PortAttrGetUnsafe(p, key, fallback););
-    LOCK(p);
-    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
-    if (SCM_PAIRP(v)) {
-        SCM_ASSERT(SCM_PAIRP(SCM_CDR(v)));
-        if (SCM_PAIRP(SCM_CDDR(v))) {
-            /* procedural */
-            ScmObj getter = SCM_CADR(v);
-            if (SCM_UNBOUNDP(fallback)) {
-                SAFE_CALL(p, r = Scm_ApplyRec1(getter, SCM_OBJ(p)));
-            } else {
-                SAFE_CALL(p, r = Scm_ApplyRec2(getter, SCM_OBJ(p), fallback));
-            }
-        } else {
-            r = SCM_CADR(v);
-        }
-    } else {
-        r = fallback;
-    }
-    UNLOCK(p);
-
-    if (SCM_UNBOUNDP(r)) {
-        Scm_Error("No port attribute for key %S in port %S", key, SCM_OBJ(p));
-    }
-    return r;
-}
-
-
-#ifdef SAFE_PORT_OP
-ScmObj Scm_PortAttrSet(ScmPort *p, ScmObj key, ScmObj val)
-#else
-ScmObj Scm_PortAttrSetUnsafe(ScmPort *p, ScmObj key, ScmObj val)
-#endif
-{
-    volatile int err_readonly = FALSE;
-    volatile int exists = FALSE;
-    VMDECL;
-    SHORTCUT(p, return Scm_PortAttrSetUnsafe(p, key, val););
-    LOCK(p);
-    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
-    if (SCM_PAIRP(v)) {
-        SCM_ASSERT(SCM_PAIRP(SCM_CDR(v)));
-        exists = TRUE;
-        if (SCM_PAIRP(SCM_CDDR(v))) {
-            /* procedural */
-            ScmObj setter = SCM_CAR(SCM_CDDR(v));
-            if (SCM_FALSEP(setter)) {
-                err_readonly = TRUE;
-            } else {
-                SAFE_CALL(p, Scm_ApplyRec2(setter, SCM_OBJ(p), val));
-            }
-        } else if (SCM_NULLP(SCM_CDDR(v))) {
-            SCM_SET_CAR_UNCHECKED(SCM_CDR(v), val);
-        } else {
-            err_readonly = TRUE;
-        }
-    } else {
-        PORT_ATTRS(p) = Scm_Cons(SCM_LIST2(key, val), PORT_ATTRS(p));
-    }
-    UNLOCK(p);
-    if (err_readonly) {
-        Scm_Error("Port attribute '%A' is read-only in port: %S",
-                  key, SCM_OBJ(p));
-    }
-    return SCM_MAKE_BOOL(exists);
-}
-
-#ifdef SAFE_PORT_OP
-ScmObj Scm_PortAttrCreate(ScmPort *p, ScmObj key, ScmObj get, ScmObj set)
-#else
-ScmObj Scm_PortAttrCreateUnsafe(ScmPort *p, ScmObj key, ScmObj get, ScmObj set)
-#endif
-{
-    int err_exists = FALSE;
-    VMDECL;
-    SHORTCUT(p, return Scm_PortAttrCreateUnsafe(p, key, get, set););
-
-    /* If get == #f, we create an ordinary attr entry.  Otherwise,
-       we create a procedural entry. */
-    ScmObj entry = (SCM_FALSEP(get)
-                    ? SCM_LIST2(key, SCM_FALSE)
-                    : SCM_LIST3(key, get, set));
-    LOCK(p);
-    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
-    if (SCM_FALSEP(v)) {
-        PORT_ATTRS(p) = Scm_Cons(entry, PORT_ATTRS(p));
-    } else {
-        err_exists = TRUE;
-    }
-    UNLOCK(p);
-    if (err_exists) {
-        Scm_Error("Couldn't create port attribute %A in %S: Named attribute already exists.",
-                  key, SCM_OBJ(p));
-    }
-    return SCM_UNDEFINED;       /* we may return more useful info in future */
-}
-
-#ifdef SAFE_PORT_OP
-ScmObj Scm_PortAttrDelete(ScmPort *p, ScmObj key)
-#else
-ScmObj Scm_PortAttrDeleteUnsafe(ScmPort *p, ScmObj key)
-#endif
-{
-    int err_undeletable = FALSE;
-    VMDECL;
-    SHORTCUT(p, return Scm_PortAttrDeleteUnsafe(p, key););
-    LOCK(p);
-    ScmObj v = Scm_Assq(key, PORT_ATTRS(p));
-    if (SCM_PAIRP(v) && SCM_PAIRP(SCM_CDR(v)) && SCM_FALSEP(SCM_CDDR(v))) {
-        err_undeletable = TRUE;
-    } else {
-        PORT_ATTRS(p) = Scm_AssocDelete(key, PORT_ATTRS(p), SCM_CMP_EQ);
-    }
-    UNLOCK(p);
-    if (err_undeletable) {
-        Scm_Error("Port attribute '%A' is not deletable from port: %S",
-                  key, SCM_OBJ(p));
-    }
-    return SCM_UNDEFINED;       /* we may return more useful info in future */
-}
-
-#ifdef SAFE_PORT_OP
-ScmObj Scm_PortAttrs(ScmPort *p)
-#else
-ScmObj Scm_PortAttrsUnsafe(ScmPort *p)
-#endif
-{
-    ScmObj h = SCM_NIL, t = SCM_NIL;
-    VMDECL;
-    SHORTCUT(p, return Scm_PortAttrsUnsafe(p););
-    LOCK(p);
-    ScmObj cp;
-    SCM_FOR_EACH(cp, PORT_ATTRS(p)) {
-        ScmObj k = SCM_CAAR(cp);
-        ScmObj v = Scm_PortAttrGetUnsafe(p, k, SCM_UNBOUND);
-        SCM_APPEND1(h, t, Scm_Cons(k, v));
-    }
-    UNLOCK(p);
-    return h;
-}
-
-
 #undef VMDECL
 #undef LOCK
 #undef UNLOCK
