@@ -10,27 +10,22 @@
 ;;
 
 (use gauche.collection)
+(use gauche.charconv)
 (use gauche.sequence)
 (use gauche.dictionary)
 (use gauche.hook)
-(use gauche.parameter)
 (use gauche.uvector)
 (use gauche.regexp)
 (use gauche.record)
 (use gauche.generator)
 (use gauche.interpolate)
-(cond-expand
- [gauche-0.9.9 (use gauche.stringutil) (use gauche.portutil)]
- [else])
+(use gauche.pputil)
+(use gauche.process)
 (use gauche.threads)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 (use file.util)
 (use text.tr)
 (use util.match)
 (use rfc.822)
 (use rfc.mime)
-
-
-
-
