@@ -1,7 +1,7 @@
 /*
  * proc.c - Procedures
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,9 +33,10 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 #include "gauche/code.h"
 #include "gauche/priv/builtin-syms.h"
+#include "gauche/priv/procP.h"
 
 /*=================================================================
  * Classes
@@ -45,9 +46,19 @@ static void proc_print(ScmObj obj, ScmPort *port, ScmWriteContext *);
 
 SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_ProcedureClass, proc_print);
 
-static void proc_print(ScmObj obj, ScmPort *port, 
+static ScmObj print_tag_key = SCM_FALSE; /* set by Scm__InitProc() */
+
+static void proc_print(ScmObj obj, ScmPort *port,
                        ScmWriteContext *ctx SCM_UNUSED)
 {
+    /* A procedure can have custom printer in the tag. */
+    ScmObj printer = Scm_Assq(print_tag_key, SCM_PROCEDURE(obj)->tagsAlist);
+    if (SCM_PAIRP(printer)) {
+        Scm_ApplyRec2(SCM_CDR(printer), obj, SCM_OBJ(port));
+        return;
+    }
+
+    /* Standard printer  */
     ScmObj info = SCM_PROCEDURE_INFO(obj);
     if (SCM_PROCEDURE_TYPE(obj) == SCM_PROC_SUBR) {
         SCM_PUTZ("#<subr", -1, port);
@@ -60,11 +71,39 @@ static void proc_print(ScmObj obj, ScmPort *port,
     }
 }
 
+/*=================================================================
+ * Common
+ */
+
+void Scm__ProcedureInit(ScmProcedure *proc,
+                        int type, /* ScmProcedureType */
+                        int required,
+                        int optional,
+                        ScmObj info)
+{
+    proc->required = required;
+    proc->optional = optional;
+    proc->type = type;
+    proc->locked = FALSE;
+    proc->currying = FALSE;
+    proc->constant = FALSE;
+    proc->leaf = FALSE;
+    proc->reserved = 0;
+    proc->info = info;
+    proc->setter = SCM_FALSE;
+    proc->inliner = SCM_FALSE;
+#if GAUCHE_API_VERSION >= 98
+    proc->reserved32 = 0;
+    proc->typeHint = SCM_FALSE;
+    proc->tagsAlist = SCM_NIL;
+#endif /*GAUCHE_API_VERSION >= 98*/
+}
+
 /*=================================================================
  * Closure
  */
 
-ScmObj Scm_MakeClosure(ScmObj code, ScmEnvFrame *env)
+ScmObj Scm_MakeClosureWithTags(ScmObj code, ScmEnvFrame *env, ScmObj tags)
 {
     ScmClosure *c = SCM_NEW(ScmClosure);
 
@@ -84,29 +123,45 @@ ScmObj Scm_MakeClosure(ScmObj code, ScmEnvFrame *env)
     int opt = SCM_COMPILED_CODE_OPTIONAL_ARGS(code);
 
     SCM_SET_CLASS(c, SCM_CLASS_PROCEDURE);
-    SCM_PROCEDURE_INIT(c, req, opt, SCM_PROC_CLOSURE, info);
+    Scm__ProcedureInit(SCM_PROCEDURE(c), SCM_PROC_CLOSURE, req, opt, info);
     c->code = code;
     c->env = env;
     SCM_PROCEDURE(c)->inliner = SCM_COMPILED_CODE(code)->intermediateForm;
+    SCM_PROCEDURE(c)->tagsAlist = tags;
 
     return SCM_OBJ(c);
 }
 
+ScmObj Scm_MakeClosure(ScmObj code, ScmEnvFrame *env)
+{
+    return Scm_MakeClosureWithTags(code, env, SCM_NIL);
+}
+
 /*=================================================================
  * Subr
  */
 
+ScmObj Scm_MakeSubrWithTags(ScmSubrProc *func,
+                            void *data,
+                            int required, int optional,
+                            ScmObj info, ScmObj tags)
+{
+    ScmSubr *s = SCM_NEW(ScmSubr);
+    SCM_SET_CLASS(s, SCM_CLASS_PROCEDURE);
+    Scm__ProcedureInit(SCM_PROCEDURE(s), SCM_PROC_SUBR,
+                       required, optional, info);
+    s->func = func;
+    s->data = data;
+    SCM_PROCEDURE(s)->tagsAlist = tags;
+    return SCM_OBJ(s);
+}
+
 ScmObj Scm_MakeSubr(ScmSubrProc *func,
                     void *data,
                     int required, int optional,
                     ScmObj info)
 {
-    ScmSubr *s = SCM_NEW(ScmSubr);
-    SCM_SET_CLASS(s, SCM_CLASS_PROCEDURE);
-    SCM_PROCEDURE_INIT(s, required, optional, SCM_PROC_SUBR, info);
-    s->func = func;
-    s->data = data;
-    return SCM_OBJ(s);
+    return Scm_MakeSubrWithTags(func, data, required, optional, info, SCM_NIL);
 }
 
 /*
@@ -137,12 +192,12 @@ ScmObj Scm_NullProc(void)
 
 /* Procedure is inherently an immutable entity, so copying doesn't
    make much sense.  However, sometimes we do need to mutate procedure
-   instances to add various bookkeeping infos during construction,
+   instances to add various bookkeeping info during construction,
    and it comes handy to do "copy, then modify" workflow. */
-ScmObj Scm_CopyProcedure(ScmProcedure *proc)
+ScmObj Scm__CopyProcedure(ScmProcedure *proc , ScmObj tagsAlist)
 {
     ScmObj n = SCM_UNDEFINED;
-    
+
     switch (proc->type) {
     case SCM_PROC_SUBR:
         n = Scm_MakeSubr(SCM_SUBR_FUNC(proc),
@@ -150,19 +205,11 @@ ScmObj Scm_CopyProcedure(ScmProcedure *proc)
                          SCM_PROCEDURE_REQUIRED(proc),
                          SCM_PROCEDURE_OPTIONAL(proc),
                          SCM_PROCEDURE_INFO(proc));
-        SCM_PROCEDURE_INLINER(n) = SCM_PROCEDURE_INLINER(proc);
-        SCM_PROCEDURE_SETTER(n) = SCM_PROCEDURE_SETTER(proc);
-        SCM_PROCEDURE_SETTER_LOCKED(n) = SCM_PROCEDURE_SETTER_LOCKED(proc);
-        SCM_PROCEDURE_CURRYING(n) = SCM_PROCEDURE_CURRYING(proc);
         SCM_SUBR_FLAGS(n) = SCM_SUBR_FLAGS(proc);
         break;
     case SCM_PROC_CLOSURE:
         n = Scm_MakeClosure(SCM_CLOSURE_CODE(proc),
                             SCM_CLOSURE_ENV(proc));
-        SCM_PROCEDURE_INLINER(n) = SCM_PROCEDURE_INLINER(proc);
-        SCM_PROCEDURE_SETTER(n) = SCM_PROCEDURE_SETTER(proc);
-        SCM_PROCEDURE_SETTER_LOCKED(n) = SCM_PROCEDURE_SETTER_LOCKED(proc);
-        SCM_PROCEDURE_CURRYING(n) = SCM_PROCEDURE_CURRYING(proc);
         break;
     case SCM_PROC_GENERIC:
     case SCM_PROC_METHOD:
@@ -172,6 +219,21 @@ ScmObj Scm_CopyProcedure(ScmProcedure *proc)
         Scm_Error("procedure-copy can only copy subr or closure: %S",
                   SCM_OBJ(proc));
     }
+    /* Copying common fields */
+    SCM_PROCEDURE_INLINER(n) = SCM_PROCEDURE_INLINER(proc);
+    SCM_PROCEDURE_SETTER(n) = SCM_PROCEDURE_SETTER(proc);
+    SCM_PROCEDURE_SETTER_LOCKED(n) = SCM_PROCEDURE_SETTER_LOCKED(proc);
+    SCM_PROCEDURE_CURRYING(n) = SCM_PROCEDURE_CURRYING(proc);
+#if GAUCHE_API_VERSION >= 98
+    SCM_PROCEDURE(n)->reserved32 = proc->reserved32;
+    SCM_PROCEDURE(n)->typeHint = proc->typeHint;
+    if (SCM_FALSEP(tagsAlist)) {
+        SCM_PROCEDURE(n)->tagsAlist = proc->tagsAlist;
+    } else {
+        SCM_ASSERT(SCM_LISTP(tagsAlist));
+        SCM_PROCEDURE(n)->tagsAlist = tagsAlist;
+    }
+#endif /*GAUCHE_API_VERSION >= 98*/
     return n;
 }
 
@@ -306,6 +368,8 @@ ScmObj Scm_CurryProcedure(ScmObj proc, ScmObj *given, int ngiven, int foldlen)
  * in 1.0.
  */
 
+#if GAUCHE_API_VERSION <= 97
+
 ScmObj Scm_ForEach1(ScmObj proc, ScmObj args)
 {
     return Scm_ForEach(proc, args, SCM_NIL);
@@ -330,6 +394,8 @@ ScmObj Scm_Map(ScmObj proc, ScmObj arg1, ScmObj args)
     return Scm_VMApply(stub, Scm_Cons(proc, Scm_Cons(arg1, args)));
 }
 
+#endif /*GAUCHE_API_VERSION <= 97*/
+
 /*=================================================================
  * Generic setter
  */
@@ -440,4 +506,6 @@ void Scm__InitProc(void)
     Scm_InitStaticClass(&Scm_ProcedureClass, "<procedure>",
                         Scm_GaucheModule(), proc_slots, 0);
     Scm_ProcedureClass.flags |= SCM_CLASS_APPLICABLE;
+
+    print_tag_key = Scm_MakeSymbol(SCM_STRING(SCM_MAKE_STR("print")), FALSE);
 }
