@@ -1,7 +1,7 @@
 /*
  * signal.c - signal handling
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,8 +33,8 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/vm.h"
-#include "gauche/class.h"
 
 /* Signals
  *
@@ -307,7 +307,7 @@ static ScmObj sigset_allocate(ScmClass *klass, ScmObj initargs);
 SCM_DEFINE_BUILTIN_CLASS(Scm_SysSigsetClass, sigset_print,
                          NULL, NULL, sigset_allocate, SCM_CLASS_DEFAULT_CPL);
 
-void sigset_print(ScmObj obj, ScmPort *out, 
+void sigset_print(ScmObj obj, ScmPort *out,
                   ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_Printf(out, "#<sys-sigset [");
@@ -844,26 +844,7 @@ void Scm_SetMasterSigmask(sigset_t *set)
 
 /*
  * Convenience routines hiding platform-dependent stuff
- *
- * TRANSIENT: These used to be used in vm.c, but no longer.  As of 0.9.3 these
- * aren't used anywhere.  Scm_SysSigmask covers those functionalities,
- * so we'll drop them by 1.0.
  */
-#if GAUCHE_API_VERSION < 1000
-void Scm_GetSigmask(sigset_t *mask)
-{
-    if (SIGPROCMASK(SIG_SETMASK, NULL, mask) != 0) {
-        Scm_SysError("sigprocmask failed");
-    }
-}
-
-void Scm_SetSigmask(sigset_t *mask)
-{
-    if (SIGPROCMASK(SIG_SETMASK, mask, NULL) != 0) {
-        Scm_SysError("sigprocmask failed");
-    }
-}
-#endif /*GAUCHE_API_VERSION < 1000*/
 
 /*
  * set signal mask
@@ -987,7 +968,7 @@ int Scm_SigWait(ScmSysSigset *mask)
     (void)SCM_INTERNAL_MUTEX_LOCK(sigHandlers.mutex);
     /* we can't wait for the signals Gauche doesn't handle. */
     to_wait = mask->set;
-    for (int i=0; i<SCM_NSIG; i++) {
+    for (int i=1; i<SCM_NSIG; i++) {
         if (!sigismember(&sigHandlers.masterSigset, i)) {
             sigdelset(&to_wait, i);
         }
