@@ -1,7 +1,7 @@
 /*
  * string.c - string implementation
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,10 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
 #include "gauche/priv/stringP.h"
+#include "gauche/priv/writerP.h"
+#include "gauche/char_attr.h"
 
 #include <string.h>
 #include <ctype.h>
@@ -574,12 +577,12 @@ static const char *index2ptr(const ScmStringBody *body,
             break;
         default:
             Scm_Panic("String index contains unrecognized signature (%02x). "
-                      "Possible memory corruption.  Aborting...", 
+                      "Possible memory corruption.  Aborting...",
                       index->signature);
         }
     }
-    return forward_pos(body, 
-                       SCM_STRING_BODY_START(body) + off, 
+    return forward_pos(body,
+                       SCM_STRING_BODY_START(body) + off,
                        nchars & (STRING_INDEX_INTERVAL(index)-1));
 }
 
@@ -616,7 +619,7 @@ ScmChar Scm_StringRef(ScmString *str, ScmSmallInt pos, int range_error)
     } else {
         p = index2ptr(b, pos);
     }
-    
+
     if (SCM_STRING_BODY_SINGLE_BYTE_P(b)) {
         return (ScmChar)(*(unsigned char *)p);
     } else {
@@ -885,7 +888,7 @@ static ScmObj substring(const ScmStringBody *xb,
     ScmSmallInt len = byterange? SCM_STRING_BODY_SIZE(xb) : SCM_STRING_BODY_LENGTH(xb);
     u_long flags = SCM_STRING_BODY_FLAGS(xb);
     if (!immutable) flags &= ~SCM_STRING_IMMUTABLE;
-    
+
     SCM_CHECK_START_END(start, end, len);
 
     if (byterange) {
@@ -1065,14 +1068,10 @@ static ScmSmallInt boyer_moore_reverse(const char *ss1, ScmSmallInt siz1,
    (*bi) and character index (*ci) is calculted, FOUND_BYTE_INDEX
    if only byte index is calculated.
 
-   When the encoding is utf-8 or none, we can scan a string as if it is just
-   a bytestring.   The only caveat is that, with utf-8, we need to calculate
-   character index after we find the match.  It is still a total win, for
-   finding out non-matches using Boyer-Moore is a lot faster than naive way.
-
-   If the encoding is EUC-JP or SJIS, we can only use Boyer-Moore when we
-   know the strings have single-byte only.  Multibyte strings of those
-   encodings can have spurious matches when compared bytewise.
+   With utf-8, we can scan a string as if it si just a bytestring.  However,
+   we need to calculate character index after we find the match.  It is still
+   a total win, for finding out non-matches using Boyer-Moore is a lot
+   faster than naive way.
  */
 
 /* return value of string_scan */
@@ -1080,24 +1079,6 @@ static ScmSmallInt boyer_moore_reverse(const char *ss1, ScmSmallInt siz1,
 #define FOUND_BOTH_INDEX 1  /* string found, and both indexes are calculated */
 #define FOUND_BYTE_INDEX 2  /* string found, and only byte index is calc'd */
 
-/* In utf-8 multibyte case, we only count byte index and let the caller
-   figure out the character index.  In other encodings we can always find
-   both index. */
-#if defined(GAUCHE_CHAR_ENCODING_UTF_8)
-#define FOUND_MAYBE_BOTH FOUND_BYTE_INDEX
-#else
-#define FOUND_MAYBE_BOTH FOUND_BOTH_INDEX
-#endif
-
-/* In euc-jp and sjis case, we use faster method only when (size == len). */
-#if defined(GAUCHE_CHAR_ENCODING_EUC_JP) || defined(GAUCHE_CHAR_ENCODING_SJIS)
-#define BYTEWISE_SEARCHABLE(siz, len)  ((siz) == (len))
-#define MULTIBYTE_NAIVE_SEARCH_NEEDED 1
-#else
-#define BYTEWISE_SEARCHABLE(siz, len)  TRUE
-#define MULTIBYTE_NAIVE_SEARCH_NEEDED 0
-#endif
-
 /* glibc has memrchr, but we need to provide fallback anyway and
    we don't need it to be highly tuned, so we just roll our own. */
 static const void *my_memrchr(const void *s, int c, size_t n)
@@ -1122,55 +1103,34 @@ static int string_search(const char *s1, ScmSmallInt siz1,
         return FOUND_BOTH_INDEX;
     }
 
-    /* Single-byte case. */
-    if (BYTEWISE_SEARCHABLE(siz1, len1)) {
-        if (siz2 == 1) {
-            /* Single ASCII character search case.  This is a huge win. */
-            const char *z = memchr(s1, s2[0], siz1);
-            if (z) { *bi = *ci = z - s1; return FOUND_MAYBE_BOTH; }
-            else return NOT_FOUND;
-        }
-        if (BYTEWISE_SEARCHABLE(siz2, len2)) {
-            ScmSmallInt i;
-            /* Shortcut for single-byte strings */
-            if (siz1 < siz2) return NOT_FOUND;
-            if (siz1 < 256 || siz2 >= 256) {
-                /* brute-force search */
-                for (i=0; i<=siz1-siz2; i++) {
-                    if (memcmp(s2, s1+i, siz2) == 0) break;
-                }
-                if (i == siz1-siz2+1) return NOT_FOUND;
-            } else {
-                i = boyer_moore(s1, siz1, s2, siz2);
-                if (i < 0) return NOT_FOUND;
+    if (siz2 == 1) {
+        /* Single ASCII character search case.  This is a huge win. */
+        const char *z = memchr(s1, s2[0], siz1);
+        if (z) { *bi = *ci = z - s1; return FOUND_BYTE_INDEX; }
+        else return NOT_FOUND;
+    } else {
+        ScmSmallInt i;
+        /* Shortcut for single-byte strings */
+        if (siz1 < siz2) return NOT_FOUND;
+        if (siz1 < 256 || siz2 >= 256) {
+            /* brute-force search */
+            for (i=0; i<=siz1-siz2; i++) {
+                if (memcmp(s2, s1+i, siz2) == 0) break;
             }
-            *bi = *ci = i;
-            return FOUND_MAYBE_BOTH;
+            if (i == siz1-siz2+1) return NOT_FOUND;
+        } else {
+            i = boyer_moore(s1, siz1, s2, siz2);
+            if (i < 0) return NOT_FOUND;
         }
-        /* FALLTHROUGH */
+        *bi = *ci = i;
+        return FOUND_BYTE_INDEX;
     }
-
-#if MULTIBYTE_NAIVE_SEARCH_NEEDED
-    /* Multibyte case. */
-    if (len1 >= len2) {
-        const char *sp = s1;
-        for (ScmSmallInt i=0; i<=len1-len2; i++) {
-            if (memcmp(sp, s2, siz2) == 0) {
-                *bi = (ScmSmallInt)(sp - s1);
-                *ci = i;
-                return FOUND_BOTH_INDEX;
-            }
-            sp += SCM_CHAR_NFOLLOWS(*sp) + 1;
-        }
-    }
-#endif /*MULTIBYTE_NAIVE_SEARCH_NEEDED*/
-    return NOT_FOUND;
 }
 
 /* NB: len2 is only used in some internal CES */
-static int string_search_reverse(const char *s1, ScmSmallInt siz1, 
+static int string_search_reverse(const char *s1, ScmSmallInt siz1,
                                  ScmSmallInt len1,
-                                 const char *s2, ScmSmallInt siz2, 
+                                 const char *s2, ScmSmallInt siz2,
                                  ScmSmallInt len2 SCM_UNUSED,
                                  ScmSmallInt *bi /* out */,
                                  ScmSmallInt *ci /* out */)
@@ -1181,56 +1141,28 @@ static int string_search_reverse(const char *s1, ScmSmallInt siz1,
         return FOUND_BOTH_INDEX;
     }
 
-    /* Single-byte case. */
-    if (BYTEWISE_SEARCHABLE(siz1, len1)) {
-        if (siz2 == 1) {
-            /* Single ASCII character search case.  This is a huge win. */
-            const char *z = my_memrchr(s1, s2[0], siz1);
-            if (z) { *bi = *ci = z - s1; return FOUND_MAYBE_BOTH; }
-            else return NOT_FOUND;
-        }
-        if (BYTEWISE_SEARCHABLE(siz2, len2)) {
-            ScmSmallInt i;
-            /* short cut for single-byte strings */
-            if (siz1 < siz2) return NOT_FOUND;
-            if (siz1 < 256 || siz2 >= 256) {
-                /* brute-force search */
-                for (i=siz1-siz2; i>=0; i--) {
-                    if (memcmp(s2, s1+i, siz2) == 0) break;
-                }
-                if (i < 0) return NOT_FOUND;
-            } else {
-                i = boyer_moore_reverse(s1, siz1, s2, siz2);
-                if (i < 0) return NOT_FOUND;
+    if (siz2 == 1) {
+        /* Single ASCII character search case.  This is a huge win. */
+        const char *z = my_memrchr(s1, s2[0], siz1);
+        if (z) { *bi = *ci = z - s1; return FOUND_BYTE_INDEX; }
+        else return NOT_FOUND;
+    } else {
+        ScmSmallInt i;
+        /* short cut for single-byte strings */
+        if (siz1 < siz2) return NOT_FOUND;
+        if (siz1 < 256 || siz2 >= 256) {
+            /* brute-force search */
+            for (i=siz1-siz2; i>=0; i--) {
+                if (memcmp(s2, s1+i, siz2) == 0) break;
             }
-            *bi = *ci = i;
-            return FOUND_MAYBE_BOTH;
+            if (i < 0) return NOT_FOUND;
         } else {
-            return NOT_FOUND;   /* sbstring can't contain mbstring. */
+            i = boyer_moore_reverse(s1, siz1, s2, siz2);
+            if (i < 0) return NOT_FOUND;
         }
+        *bi = *ci = i;
+        return FOUND_BYTE_INDEX;
     }
-
-#if MULTIBYTE_NAIVE_SEARCH_NEEDED
-    /* Multibyte case. */
-    if (len1 >= len2) {
-        const char *sp = s1 + siz1, *p;
-        for (ScmSmallInt i=0; i<len2; i++) {
-            SCM_CHAR_BACKWARD(sp, s1, p);
-            SCM_ASSERT(*p);
-            sp = p;
-        }
-        for (ScmSmallInt i=len1-len2; i>=0; i--) {
-            if (memcmp(sp, s2, siz2) == 0) {
-                *bi = (ScmSmallInt)(sp - s1);
-                *ci = i;
-                return FOUND_BOTH_INDEX;
-            }
-            SCM_CHAR_BACKWARD(sp, s1, p);
-            sp = p;
-        }
-    }
-#endif /*MULTIBYTE_NAIVE_SEARCH_NEEDED*/
-    return NOT_FOUND;
 }
 
 /* Scan s2 in s1, and calculates appropriate return value(s) according to
@@ -1387,9 +1319,6 @@ ScmObj Scm_StringScanCharRight(ScmString *s1, ScmChar ch, int retmode)
 #undef NOT_FOUND
 #undef FOUND_BOTH_INDEX
 #undef FOUND_BYTE_INDEX
-#undef FOUND_MAYBE_BOTH
-#undef BYTEWISE_SEARCHABLE
-#undef MULTIBYTE_NAIVE_SEARCH_NEEDED
 
 /* Split string by char.  Char itself is not included in the result.
    If LIMIT >= 0, up to that number of matches are considered (i.e.
@@ -1542,9 +1471,12 @@ char **Scm_ListToCStringArray(ScmObj lis, int errp, void *(*alloc)(size_t))
 /*----------------------------------------------------------------
  * printer
  */
+
+/* ch is single byte if bytemode is true */
 static inline void string_putc(ScmChar ch, ScmPort *port, int bytemode)
 {
-    char buf[6];
+    const int ESCAPE_BUF_MAX = 12;
+    char buf[ESCAPE_BUF_MAX];
     switch (ch) {
     case '\\': SCM_PUTZ("\\\\", -1, port); break;
     case '"':  SCM_PUTZ("\\\"", -1, port); break;
@@ -1554,13 +1486,32 @@ static inline void string_putc(ScmChar ch, ScmPort *port, int bytemode)
     case '\f': SCM_PUTZ("\\f", -1, port); break;
     case '\0': SCM_PUTZ("\\0", -1, port); break;
     default:
-        if (ch < ' ' || ch == 0x7f || (bytemode && ch >= 0x80)) {
-            /* TODO: Should we provide 'legacy-compatible writer mode,
-               which does not use ';' terminator? */
-            snprintf(buf, 6, "\\x%02x;", (unsigned char)ch);
-            SCM_PUTZ(buf, -1, port);
+        if (ch < 0x80 || bytemode) {
+            if (ch < ' ' || ch == 0x7f || bytemode) {
+                /* TODO: Should we provide 'legacy-compatible writer mode,
+                   which does not use ';' terminator? */
+                snprintf(buf, ESCAPE_BUF_MAX, "\\x%02x;", (unsigned char)ch);
+                SCM_PUTZ(buf, -1, port);
+            } else {
+                SCM_PUTC(ch, port);
+            }
         } else {
-            SCM_PUTC(ch, port);
+            switch (Scm_CharGeneralCategory(ch)) {
+            case SCM_CHAR_CATEGORY_Cc:
+            case SCM_CHAR_CATEGORY_Cf:
+            case SCM_CHAR_CATEGORY_Cs:
+            case SCM_CHAR_CATEGORY_Co:
+            case SCM_CHAR_CATEGORY_Cn:
+                if (ch < 0x10000) {
+                    snprintf(buf, ESCAPE_BUF_MAX, "\\x%04x;", (u_int)ch);
+                } else {
+                    snprintf(buf, ESCAPE_BUF_MAX, "\\x%x;", (u_int)ch);
+                }
+                SCM_PUTZ(buf, -1, port);
+                break;
+            default:
+                SCM_PUTC(ch, port);
+            }
         }
     }
 }
@@ -1568,13 +1519,21 @@ static inline void string_putc(ScmChar ch, ScmPort *port, int bytemode)
 static void string_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
     ScmString *str = SCM_STRING(obj);
+    int limit = (ctx->controls? ctx->controls->stringLength : -1);
+    int trimmed = FALSE;
+
     if (Scm_WriteContextMode(ctx) == SCM_WRITE_DISPLAY) {
+        /* Display mode isn't affected by string-length control */
         SCM_PUTS(str, port);
     } else {
         const ScmStringBody *b = SCM_STRING_BODY(str);
         if (SCM_STRING_BODY_SINGLE_BYTE_P(b)) {
             const char *cp = SCM_STRING_BODY_START(b);
             ScmSmallInt size = SCM_STRING_BODY_SIZE(b);
+            if (limit >= 0 && limit < size) {
+                trimmed = TRUE; size = limit;
+            }
+
             if (SCM_STRING_BODY_INCOMPLETE_P(b)) {
                 /* TODO: Should we provide legacy-compatible writer mode,
                    which puts #*"..." instead? */
@@ -1582,21 +1541,27 @@ static void string_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             } else {
                 SCM_PUTC('"', port);
             }
-            while (size--) {
+            while (--size >= 0) {
                 string_putc(*cp++, port, SCM_STRING_BODY_INCOMPLETE_P(b));
             }
         } else {
             const char *cp = SCM_STRING_BODY_START(b);
             ScmSmallInt len = SCM_STRING_BODY_LENGTH(b);
+            if (limit >= 0 && limit < len) {
+                trimmed = TRUE; len = limit;
+            }
 
             SCM_PUTC('"', port);
-            while (len--) {
+            while (--len >= 0) {
                 ScmChar ch;
                 SCM_CHAR_GET(cp, ch);
                 string_putc(ch, port, FALSE);
                 cp += SCM_CHAR_NBYTES(ch);
             }
         }
+        if (trimmed) {
+            SCM_PUTZ(SCM_WRITTEN_ELLIPSIS, -1, port);
+        }
         SCM_PUTC('"', port);
     }
 }
@@ -1702,23 +1667,23 @@ void Scm_StringBodyIndexDump(const ScmStringBody *sb, ScmPort *port)
     size_t index_size = 0;
 
     switch (STRING_INDEX_TYPE(index)) {
-    case STRING_INDEX8:  
-        Scm_Printf(port, "index8  "); 
+    case STRING_INDEX8:
+        Scm_Printf(port, "index8  ");
         index_size = (size_t)index->index8[1];
         break;
-    case STRING_INDEX16: 
-        Scm_Printf(port, "index16 ");   
+    case STRING_INDEX16:
+        Scm_Printf(port, "index16 ");
         index_size = (size_t)index->index16[1];
         break;
     case STRING_INDEX32:
-        Scm_Printf(port, "index32 "); 
+        Scm_Printf(port, "index32 ");
         index_size = (size_t)index->index32[1];
         break;
-    case STRING_INDEX64: 
+    case STRING_INDEX64:
         Scm_Printf(port, "index64 ");
         index_size = (size_t)index->index64[1];
         break;
-    default:    
+    default:
         Scm_Printf(port, "unknown(%02x) ", (uint8_t)STRING_INDEX_TYPE(index));
     }
     Scm_Printf(port, " interval %d  size %d\n", interval, index_size-1);
@@ -1738,164 +1703,6 @@ void Scm_StringBodyIndexDump(const ScmStringBody *sb, ScmPort *port)
 }
 
 
-/*==================================================================
- *
- * String pointer (OBSOLETED)
- *
- */
-
-#if GAUCHE_STRING_POINTER
-
-SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_StringPointerClass, NULL);
-
-ScmObj Scm_MakeStringPointer(ScmString *src, ScmSmallInt index,
-                             ScmSmallInt start, ScmSmallInt end)
-{
-    const ScmStringBody *srcb = SCM_STRING_BODY(src);
-    ScmSmallInt len = SCM_STRING_BODY_LENGTH(srcb);
-    ScmSmallInt effective_size;
-    const char *sptr, *ptr, *eptr;
-
-    SCM_CHECK_START_END(start, end, len);
-    while (index < 0) index += (end - start) + 1;
-    if (index > (end - start)) goto badindex;
-
-    sptr = forward_pos(srcb, SCM_STRING_BODY_START(srcb), start);
-    ptr = forward_pos(srcb, sptr, index);
-    if (end == len) {
-        eptr = SCM_STRING_BODY_END(srcb);
-    } else {
-        eptr = forward_pos(srcb, sptr, end - start);
-    }
-    effective_size = eptr - sptr;
-
-    ScmStringPointer *sp = SCM_NEW(ScmStringPointer);
-    SCM_SET_CLASS(sp, SCM_CLASS_STRING_POINTER);
-    sp->length = (SCM_STRING_BODY_INCOMPLETE_P(srcb)? -1 : (end-start));
-    sp->size = effective_size;
-    sp->start = sptr;
-    sp->index = index;
-    sp->current = ptr;
-    return SCM_OBJ(sp);
-  badindex:
-    Scm_Error("index out of range: %ld", index);
-    return SCM_UNDEFINED;
-}
-
-ScmObj Scm_StringPointerRef(ScmStringPointer *sp)
-{
-    ScmChar ch;
-    if (sp->length < 0 || sp->size == sp->length) {
-        if (sp->index >= sp->size) return SCM_EOF;
-        ch = *(const unsigned char*)sp->current;
-    } else {
-        if (sp->index >= sp->length) return SCM_EOF;
-        SCM_CHAR_GET(sp->current, ch);
-    }
-    return SCM_MAKE_CHAR(ch);
-}
-
-ScmObj Scm_StringPointerNext(ScmStringPointer *sp)
-{
-    ScmChar ch;
-    if (sp->length < 0 || sp->size == sp->length) {
-        if (sp->index >= sp->size) return SCM_EOF;
-        sp->index++;
-        ch = *(const unsigned char*)sp->current++;
-    } else {
-        if (sp->index >= sp->length) return SCM_EOF;
-        SCM_CHAR_GET(sp->current, ch);
-        sp->index++;
-        sp->current += SCM_CHAR_NFOLLOWS(*sp->current) + 1;
-    }
-    return SCM_MAKE_CHAR(ch);
-}
-
-ScmObj Scm_StringPointerPrev(ScmStringPointer *sp)
-{
-    ScmChar ch;
-    if (sp->index <= 0) return SCM_EOF;
-    if (sp->length < 0 || sp->size == sp->length) {
-        sp->index--;
-        ch = *(const unsigned char*)--sp->current;
-    } else {
-        const char *prev;
-        SCM_CHAR_BACKWARD(sp->current, sp->start, prev);
-        SCM_ASSERT(prev != NULL);
-        SCM_CHAR_GET(prev, ch);
-        sp->index--;
-        sp->current = prev;
-    }
-    return SCM_MAKE_CHAR(ch);
-}
-
-ScmObj Scm_StringPointerSet(ScmStringPointer *sp, ScmSmallInt index)
-{
-    if (index < 0) goto badindex;
-    /* NB: Safe to cast index, for too large index would be rejected
-       by the check. */
-    if (sp->length < 0 || sp->size == sp->length) {
-        if (index > sp->size) goto badindex;
-        sp->index = index;
-        sp->current = sp->start + index;
-    } else {
-        if (index > sp->length) goto badindex;
-        sp->index = index;
-        sp->current = forward_pos(NULL, sp->start, index);
-    }
-    return SCM_OBJ(sp);
-  badindex:
-    Scm_Error("index out of range: %ld", index);
-    return SCM_UNDEFINED;
-}
-
-ScmObj Scm_StringPointerSubstring(ScmStringPointer *sp, int afterp)
-{
-    /* TODO: set SCM_STRING_TERMINATED if applicable. */
-    if (sp->length < 0) {
-        if (afterp)
-            return SCM_OBJ(make_str(-1, sp->size - sp->index, sp->current,
-                                    0, NULL));
-        else
-            return SCM_OBJ(make_str(-1, sp->index, sp->start, 0, NULL));
-    } else {
-        if (afterp)
-            return SCM_OBJ(make_str(sp->length - sp->index,
-                                    sp->start + sp->size - sp->current,
-                                    sp->current, 0, NULL));
-        else
-            return SCM_OBJ(make_str(sp->index,
-                                    sp->current - sp->start,
-                                    sp->start, 0, NULL));
-    }
-}
-
-/* Copy string pointer.
-   Thanks to Alex Shinn (foof@synthcode.com) */
-ScmObj Scm_StringPointerCopy(ScmStringPointer *sp1)
-{
-    ScmStringPointer *sp2 = SCM_NEW(ScmStringPointer);
-    SCM_SET_CLASS(sp2, SCM_CLASS_STRING_POINTER);
-    sp2->length  = sp1->length;
-    sp2->size    = sp1->size;
-    sp2->start   = sp1->start;
-    sp2->index   = sp1->index;
-    sp2->current = sp1->current;
-    return SCM_OBJ(sp2);
-}
-
-/* Dump string pointer info for debugging.
-   Thanks to Alex Shinn (foof@synthcode.com) */
-void Scm_StringPointerDump(ScmStringPointer *sp1)
-{
-    Scm_Printf(SCM_CUROUT,
-               "<sp addr: %p len: %d size: %d start: %p index: %d cur: %p>\n",
-               sp1, sp1->length, sp1->size, sp1->start, sp1->index,
-               sp1->current);
-}
-
-#endif /*GAUCHE_STRING_POINTER*/
-
 /*==================================================================
  *
  * String cursor API
@@ -1999,7 +1806,7 @@ ScmObj Scm_StringCursorForward(ScmString* s, ScmObj sc, int nchars)
     if (SCM_INTEGERP(sc)) {
         return Scm_MakeStringCursorFromIndex(s, Scm_GetInteger(sc) + nchars);
     }
-    
+
     const ScmStringBody *srcb = SCM_STRING_BODY(s);
     const char *ptr = string_cursor_ptr(srcb, sc);
     if (ptr == NULL) {
@@ -2023,7 +1830,7 @@ ScmObj Scm_StringCursorBack(ScmString* s, ScmObj sc, int nchars)
     if (ptr == NULL) {
         Scm_Error("must be either an index or a cursor: %S", sc);
     }
-    
+
     if (SCM_STRING_BODY_SINGLE_BYTE_P(srcb) ||
         SCM_STRING_BODY_INCOMPLETE_P(srcb)) {
         return make_string_cursor(s, ptr - nchars);
@@ -2048,6 +1855,13 @@ ScmChar Scm_StringRefCursor(ScmString* s, ScmObj sc, int range_error)
     }
 
     const ScmStringBody *srcb = SCM_STRING_BODY(s);
+
+    /* we can't allow string-ref on incomplete strings, since it may yield
+       invalid character object. */
+    if (SCM_STRING_BODY_INCOMPLETE_P(srcb)) {
+        Scm_Error("incomplete string not allowed : %S", s);
+    }
+
     const char *ptr = string_cursor_ptr(srcb, sc);
     if (ptr == NULL) {
         Scm_Error("must be either an index or a cursor: %S", sc);
@@ -2055,9 +1869,8 @@ ScmChar Scm_StringRefCursor(ScmString* s, ScmObj sc, int range_error)
     if (ptr == SCM_STRING_BODY_END(srcb)) {
         if (range_error) {
             Scm_Error("cursor is at the end: %S", sc);
-        } else {
-            return SCM_CHAR_INVALID;
         }
+        return SCM_CHAR_INVALID;
     }
     ScmChar ch;
     SCM_CHAR_GET(ptr, ch);
@@ -2094,7 +1907,7 @@ int Scm_StringCursorCompare(ScmObj sc1, ScmObj sc2,
 
     ScmSmallInt i1 = string_cursor_offset(sc1);
     ScmSmallInt i2 = string_cursor_offset(sc2);
-    if (i1 < 0 || i2 < 0) { 
+    if (i1 < 0 || i2 < 0) {
         Scm_Error("arguments must be either both cursors or both indexes: %S vs %S", sc1, sc2);
     }
     return numcmp(SCM_MAKE_INT(i1), SCM_MAKE_INT(i2));
@@ -2195,9 +2008,9 @@ void Scm__DStringRealloc(ScmDString *dstr, ScmSmallInt minincr)
 }
 
 /* Retrieve accumulated string. */
-static const char *dstring_getz(ScmDString *dstr, 
-                                ScmSmallInt *psiz, 
-                                ScmSmallInt *plen, 
+static const char *dstring_getz(ScmDString *dstr,
+                                ScmSmallInt *psiz,
+                                ScmSmallInt *plen,
                                 int noalloc)
 {
     ScmSmallInt size, len;
