@@ -1,7 +1,7 @@
 /*
  * symbol.c - symbol implementation
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,6 +33,8 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
+#include "gauche/priv/configP.h"
+#include "gauche/priv/atomicP.h"
 #include "gauche/priv/builtin-syms.h"
 #include "gauche/priv/moduleP.h"
 
@@ -157,13 +159,16 @@ static SCM_DEFINE_STRING_CONST(default_prefix, "G", 1, 1);
 ScmObj Scm_Gensym(ScmString *prefix)
 {
     char numbuf[50];
-    /* We don't need mutex for this variable, since the race on it is
-       tolerated---multiple threads may get the same name symbols,
-       but they are uninterned and never be eq? to each other. */
-    static intptr_t gensym_count = 0;
+    static ScmAtomicVar gensym_count = 0;
+    ScmAtomicWord mycount = Scm_AtomicLoad(&gensym_count);
+
+    while (!Scm_AtomicCompareExchange(&gensym_count, &mycount, mycount+1)) {
+        mycount = Scm_AtomicLoad(&gensym_count);
+        Scm_YieldCPU();
+    }
 
     if (prefix == NULL) prefix = &default_prefix;
-    int nc = snprintf(numbuf, 49, "%"PRIdPTR, gensym_count++);
+    int nc = snprintf(numbuf, 49, "%"PRIdPTR, mycount);
     numbuf[49] = '\0';
     ScmObj name = Scm_StringAppendC(prefix, numbuf, nc, nc);
     ScmSymbol *sym = make_sym(SCM_CLASS_SYMBOL, SCM_STRING(name), FALSE);
@@ -194,7 +199,7 @@ static int escape_required_p(const ScmStringBody *b, u_int flags, int casefold)
 {
     int siz = SCM_STRING_BODY_SIZE(b);
     if (siz == 0) return !(flags & SCM_SYMBOL_WRITER_NOESCAPE_EMPTY);
-    
+
     const char *p = SCM_STRING_BODY_START(b);
     const char *e = p + siz;
     ScmChar ch;
