@@ -6,7 +6,8 @@
   down to main() to see how.)
 */
 
-#include <gauche.h>
+#include "gauche.h"
+#include "gauche/priv/configP.h"
 
 void Scm_Init_libextra(ScmModule *);
 
@@ -22,7 +23,3 @@ int main(int argc, const char **argv)
     Scm_SimpleMain(argc, argv, NULL, 0);
     return 0;
 }
-
-
-
-
