@@ -1,7 +1,7 @@
 /*
  * treemap.c - tree map implementation
  *
- *   Copyright (c) 2007-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2007-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 
 /*================================================================
  * ScmTreeCore
@@ -322,7 +322,7 @@ void Scm_TreeCoreCheckConsistency(ScmTreeCore *tc)
  * ScmTreeMap
  */
 
-static void treemap_print(ScmObj obj, ScmPort *port, 
+static void treemap_print(ScmObj obj, ScmPort *port,
                           ScmWriteContext *ctx SCM_UNUSED)
 {
     ScmTreeMap *tm = SCM_TREE_MAP(obj);
