@@ -188,8 +188,11 @@
         }
 #endif
 
-        SCM_PROF_COUNT_CALL(vm, VAL0);
-        VAL0 = SCM_SUBR(VAL0)->func(ARGP, argc, SCM_SUBR(VAL0)->data);
+        do {
+            SCM_PROF_COUNT_CALL(vm, VAL0);
+            vm->trampoline = -1;
+            VAL0 = SCM_SUBR(VAL0)->func(ARGP, argc, SCM_SUBR(VAL0)->data);
+        } while ((argc = vm->trampoline) >= 0);
         /* the subr may have substituted pc, so we need to check
            if we can pop the continuation immediately. */
         if (TAIL_POS()) RETURN_OP();
