@@ -1,7 +1,7 @@
 /*
  * write.c - writer
  *
- *   Copyright (c) 2000-2020  Shiro Kawai  <shiro@acm.org>
+ *   Copyright (c) 2000-2024  Shiro Kawai  <shiro@acm.org>
  *
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -33,7 +33,7 @@
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
-#include "gauche/class.h"
+#include "gauche/priv/configP.h"
 #include "gauche/writer.h"
 #include "gauche/priv/builtin-syms.h"
 #include "gauche/priv/macroP.h" /* PVREF stuff.  Will go in future */
@@ -49,8 +49,11 @@ static void write_ss(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static void write_object(ScmObj obj, ScmPort *out, ScmWriteContext *ctx);
 static ScmObj write_object_fallback(ScmObj *args, int nargs, ScmGeneric *gf);
+static int double_quote_closed_p(ScmString *s);
 SCM_DEFINE_GENERIC(Scm_GenericWriteObject, write_object_fallback, NULL);
 
+static void call_pprint(ScmObj obj, ScmPort *out);
+
 static const ScmWriteControls *defaultWriteControls;
 
 /*============================================================
@@ -129,6 +132,9 @@ ScmWriteControls *Scm_MakeWriteControls(const ScmWriteControls *proto)
         p->printBase = 10;
         p->printRadix = FALSE;
         p->printPretty = FALSE;
+        p->printIndent = 0;
+        p->stringLength = -1;
+        p->exactDecimal = FALSE;
     }
     return p;
 }
@@ -298,7 +304,7 @@ int Scm_WriteLimited(ScmObj obj, ScmObj p, int mode, int width)
     } else {
         write_rec(obj, SCM_PORT(out), &ctx);
     }
-    
+
     ScmString *str = SCM_STRING(Scm_GetOutputString(SCM_PORT(out), 0));
     int nc = SCM_STRING_BODY_LENGTH(SCM_STRING_BODY(str));
     if (nc > width) {
@@ -338,7 +344,7 @@ static void write_general(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
    write-object.   We can't use VMApply here since this function can be
    called deep in the recursive stack of Scm_Write, so the control
    may not return to VM immediately. */
-static void write_object(ScmObj obj, ScmPort *port, 
+static void write_object(ScmObj obj, ScmPort *port,
                          ScmWriteContext *ctx SCM_UNUSED)
 {
     Scm_ApplyRec(SCM_OBJ(&Scm_GenericWriteObject),
@@ -354,6 +360,20 @@ static ScmObj write_object_fallback(ScmObj *args, int nargs,
                   Scm_ArrayToList(args, nargs));
     }
     ScmClass *klass = Scm_ClassOf(args[0]);
+    /* If any base class has print() pointer, we use it. */
+    for (ScmClass **cpa = klass->cpa; *cpa; cpa++) {
+        if ((*cpa)->print) {
+            /* NB: For historical reasons, the context passed to
+               write_object isn't passed down to the generic
+               function.  We'll address it later, but for now
+               we use the default context. */
+            ScmWriteContext ctx;
+            write_context_init(&ctx, SCM_WRITE_WRITE, 0, 0);
+            (*cpa)->print(args[0], SCM_PORT(args[1]), &ctx);
+            return SCM_TRUE;
+        }
+    }
+    /* Fallback printer */
     Scm_Printf(SCM_PORT(args[1]), "#<%A%s%p>",
                klass->name,
                (SCM_FALSEP(klass->redefined)? " " : ":redefined "),
@@ -428,9 +448,9 @@ static size_t write_char(ScmChar ch, ScmPort *port, ScmWriteContext *ctx)
  */
 ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
-    const ScmWriteControls *wp = 
+    const ScmWriteControls *wp =
         Scm_GetWriteControls(ctx, Scm_PortWriteState(port));
-    
+
 #define CASE_ITAG_RET(obj, str)                 \
     case SCM_ITAG(obj):                         \
         Scm_PutzUnsafe(str, -1, port);          \
@@ -467,6 +487,7 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
         Scm_NumberFormatInit(&fmt);
         fmt.radix = wp->printBase;
         if (wp->printRadix) fmt.flags |= SCM_NUMBER_FORMAT_ALT_RADIX;
+        if (wp->exactDecimal) fmt.flags |= SCM_NUMBER_FORMAT_EXACT_DECIMAL_POINT;
         return SCM_MAKE_INT(Scm_PrintNumber(port, obj, &fmt));
     }
     /* PVREF only appears in pattern temlate in the current macro expander.
@@ -481,7 +502,7 @@ ScmObj Scm__WritePrimitive(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     else if (SCM_STRING_CURSOR_SMALL_P(obj)) {
         char buf[SPBUFSIZ];
         int k = snprintf(buf, SPBUFSIZ, "#<string-cursor %ld>",
-			 SCM_STRING_CURSOR_SMALL_OFFSET(obj));
+                         SCM_STRING_CURSOR_SMALL_OFFSET(obj));
         Scm_PutzUnsafe(buf, -1, port);
         return SCM_MAKE_INT(k);
     }
@@ -594,7 +615,7 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             }                                                           \
         }                                                               \
     } while (0)
-    
+
 
     for (;;) {
     write1:
@@ -666,7 +687,7 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             if (wp->printLength == 0) {
                 /* in this case we don't print the elements at all, so we need
                    to treat this specially. */
-                Scm_PutzUnsafe("(...)", -1, port);
+                Scm_PutzUnsafe("(" SCM_WRITTEN_ELLIPSIS ")", -1, port);
                 if (st) st->currentLevel--;
                 goto next;
             }
@@ -683,7 +704,7 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             if (wp->printLength == 0) {
                 /* in this case we don't print the elements at all, so we need
                    to treat this specially. */
-                Scm_PutzUnsafe("#(...)", -1, port);
+                Scm_PutzUnsafe("#(" SCM_WRITTEN_ELLIPSIS ")", -1, port);
                 if (st) st->currentLevel--;
                 goto next;
             }
@@ -715,7 +736,7 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                     Scm_PutcUnsafe(')', port);
                     POP();
                 } else if (wp->printLength >= 0 && wp->printLength <= i) {
-                    Scm_PutzUnsafe(" ...)", -1, port);
+                    Scm_PutzUnsafe(" " SCM_WRITTEN_ELLIPSIS ")", -1, port);
                     POP();
                 } else {
                     Scm_PutcUnsafe(' ', port);
@@ -746,14 +767,14 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                         POP();
                     } else {
                         obj = v;
-                        SCM_SET_CAR_UNCHECKED(SCM_CDR(top), 
+                        SCM_SET_CAR_UNCHECKED(SCM_CDR(top),
                                               SCM_MAKE_INT(count+1));
                         SCM_SET_CDR_UNCHECKED(SCM_CDR(top), SCM_NIL);
                         goto write1;
                     }
                 } else if (wp->printLength >= 0 && wp->printLength <= count) {
                     /* print-length limit reached */
-                    Scm_PutzUnsafe(" ...)", -1, port);
+                    Scm_PutzUnsafe(" " SCM_WRITTEN_ELLIPSIS ")", -1, port);
                     POP();
                 } else if (ht && !SCM_EQ(Scm_HashTableRef(ht, v, SCM_MAKE_INT(1)), SCM_MAKE_INT(1)))  {
                     /* cdr part is shared */
@@ -780,6 +801,18 @@ static void write_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 #undef CHECK_DEPTH
 }
 
+/* Returns TRUE if OBJ is 'pretty-printable', that is, pass2 need to be
+   handled with %pretty-print. */
+static int pprintable_p(ScmObj obj)
+{
+    if (!SCM_PTRP(obj)) return FALSE;
+    if (SCM_PAIRP(obj)) return TRUE;
+    if (SCM_VECTORP(obj) && SCM_VECTOR_SIZE(obj) > 0) return TRUE;
+    if (SCM_UVECTORP(obj) && SCM_UVECTOR_SIZE(obj) > 0) return TRUE;
+    if (SCM_ISA(obj, SCM_CLASS_DICTIONARY)) return TRUE;
+    return FALSE;
+}
+
 /* Write/ss main driver
    This should never be called recursively.
    We modify port->flags and port->writeState; they are cleaned up
@@ -800,7 +833,7 @@ static void write_ss(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     port->flags &= ~(SCM_PORT_WALKING|SCM_PORT_WRITESS);
 
     /* pass 2 */
-    if (ctx->controls && ctx->controls->printPretty) {
+    if (ctx->controls && ctx->controls->printPretty && pprintable_p(obj)) {
         static ScmObj proc = SCM_UNDEFINED;
         SCM_BIND_PROC(proc, "%pretty-print", Scm_GaucheInternalModule());
         Scm_ApplyRec4(proc, obj, SCM_OBJ(port),
@@ -841,9 +874,9 @@ void Scm_Format(ScmPort *out SCM_UNUSED,
  *
  *        %[width][.prec]A    - Same as %S, but use DISPLAY mode.
  *
- *        %C                  - Take ScmChar argument and outputs it.
+ *        %W                  - Call pprint.
  *
- *  Both functions return a number of characters written.
+ *        %C                  - Take ScmChar argument and outputs it.
  *
  *  A pound flag '#' for the S directive causes circular-safe output.
  *
@@ -859,6 +892,10 @@ void Scm_Format(ScmPort *out SCM_UNUSED,
  * called by Scm_WithPortLocking.  On some architecture, we can't pass
  * va_list type of argument in a closure packet easily.
  */
+/* NB: It is tempting to write the logic in Scheme, but we keep the main
+ * feature in C so that Scm_Printf is available to debug Scheme code.
+ * The exception is %W which need to call pprint.
+ */
 
 /* Pass 1.  Pop vararg and make a list of arguments.
  * NB: If we're "walking" pass, and the argument is a Lisp object,
@@ -918,7 +955,7 @@ static ScmObj SCM_NOINLINE vprintf_pass1(ScmPort *out,
                     SCM_APPEND1(h, t, Scm_MakeIntegerU((u_long)(intptr_t)val));
                     break;
                 }
-            case 'S': case 'A':
+            case 'S': case 'A': case 'W':
                 {
                     ScmObj o = va_arg(ap, ScmObj);
                     SCM_APPEND1(h, t, o);
@@ -1062,18 +1099,35 @@ static void vprintf_pass2(ScmPort *out, const char *fmt, ScmObj args)
                                 : ((c == 'A')
                                    ? SCM_WRITE_DISPLAY
                                    : SCM_WRITE_WRITE));
-                    int n = 0;
                     if (width <= 0) {
                         Scm_Write(val, SCM_OBJ(out), mode);
-                    } else {
+                    } if (prec <= 0) {
                         Scm_WriteLimited(val, SCM_OBJ(out), mode, width);
+                    } else {
+                        ScmObj buf = Scm_MakeOutputStringPort(TRUE);
+                        Scm_WriteLimited(val, buf, mode, width);
+                        ScmObj s = Scm_GetOutputString(SCM_PORT(buf), 0);
+                        Scm_PutsUnsafe(SCM_STRING(s), out);
+                        ScmSmallInt n = SCM_STRING_LENGTH(s);
+                        if (n > width) {
+                            if (double_quote_closed_p(SCM_STRING(s)))
+                                Scm_PutzUnsafe(" " SCM_WRITTEN_ELLIPSIS,
+                                               -1, out);
+                            else
+                                Scm_PutzUnsafe("\"" SCM_WRITTEN_ELLIPSIS,
+                                               -1, out);
+                        } else {
+                            for (; n < prec; n++) Scm_PutcUnsafe(' ', out);
+                        }
                     }
-                    if (n < 0 && prec > 0) {
-                        Scm_PutzUnsafe(" ...", -1, out);
-                    }
-                    if (n > 0) {
-                        for (; n < prec; n++) Scm_PutcUnsafe(' ', out);
-                    }
+                    break;
+                }
+            case 'W':
+                {
+                    SCM_ASSERT(SCM_PAIRP(args));
+                    ScmObj val = SCM_CAR(args);
+                    args = SCM_CDR(args);
+                    call_pprint(val, out);
                     break;
                 }
             case 'C':
@@ -1121,9 +1175,44 @@ static void vprintf_pass2(ScmPort *out, const char *fmt, ScmObj args)
     }
 }
 
+/* NB: libfmt has similar routine (quote-terminated?) in 'chop-and-out'.
+   Since Scm_Printf can be called during initialiation before Scheme runtime
+   is fully up, we need this implemented in C.  */
+static int double_quote_closed_p(ScmString *s)
+{
+    ScmSmallInt siz;
+    const char *p = Scm_GetStringContent(s, &siz, NULL, NULL);
+    int inside = FALSE;
+    int escaped = FALSE;
+    for (int i = 0; i < siz; i++, p++) {
+        if (!escaped) {
+            if (*p == '"') {
+                inside = !inside;
+            }
+            if (*p == '\\') {
+                escaped = TRUE;
+            } else {
+                escaped = FALSE;
+            }
+        } else {
+            escaped = FALSE;
+        }
+    }
+    return !inside;
+}
+
+static void call_pprint(ScmObj obj, ScmPort *out)
+{
+    ScmWriteControls *wc = Scm_MakeWriteControls(NULL);
+    wc->printPretty = TRUE;
+    wc->printWidth = 79;
+    wc->printIndent = Scm_PortColumn(out);
+    Scm_WriteWithControls(obj, SCM_OBJ(out), SCM_WRITE_WRITE, wc);
+}
+
 /* Public APIs */
 
-void Scm_Vprintf(ScmPort *out, const char *fmt, va_list ap, 
+void Scm_Vprintf(ScmPort *out, const char *fmt, va_list ap,
                  int sharedp SCM_UNUSED)
 {
     if (!SCM_OPORTP(out)) Scm_Error("output port required, but got %S", out);
