@@ -299,7 +299,7 @@
 (chibi-test
  (include "include/ideque-tests"))
 
-;; non-srfi-134 api
+;; non-SRFI-134 api
 (test* "make-ideque" '(a a a a a)
        (ideque->list (make-ideque 5 'a)))
 
@@ -383,187 +383,20 @@
   )
 
 ;;;========================================================================
-(test-section "data.random")
-(use data.random)
-(test-module 'data.random)
 
-(define (test-regular-strings regexp)
-  (let1 g (regular-strings$ regexp)
-    (test* #"regular-string$ ~regexp"
-           regexp
-           (generator->list g 10)
-           (^[rx ss] (every (^s (equal? (rxmatch->string rx s) s)) ss)))))
-
-(test-regular-strings #/a/)
-(test-regular-strings #/[a-z]/)
-(test-regular-strings #/[^a-z]/)
-(test-regular-strings #/abc|def/)
-(test-regular-strings #/ab*/)
-(test-regular-strings #/(ab|cd)+/)
-(test-regular-strings #/(abc){3}/)
-(test-regular-strings #/(abc){3,5}/)
-(test-regular-strings #/(abc){,3}/)
-(test-regular-strings #/(?i:abc)/)
-(test-regular-strings #/abc/i)
+;; data.random is tested in ext/data, for it depends on math.mt-random
 
 ;;;========================================================================
-;; depends on data.random
-(test-section "data.heap")
-(use data.heap)
-(test-module 'data.heap)
 
-(use srfi-1)
-(use srfi-27)
-(use gauche.sequence)
-(use util.match)
-
-(let ((rs (make-random-source)))
-  (define (do-test data comparator) ; data must be sorted
-    (define len (length data))
-    (define heap (make-binary-heap :storage (make-vector len)
-                                   :comparator comparator))
-    (let1 input (shuffle data rs)
-      (test* (format "heap(~s) insertion ~s" len input)
-             (map-with-index (^[i e] (list i e #t #t)) input)
-             (let ([zmin #f]
-                   [zmax #f])
-               (map-with-index
-                (^[i e]
-                  (begin
-                    (binary-heap-push! heap e)
-                    (binary-heap-check heap)
-                    (when (or (not zmin)
-                              (< (comparator-compare comparator e zmin) 0))
-                      (set! zmin e))
-                    (when (or (not zmax)
-                              (> (comparator-compare comparator e zmax) 0))
-                      (set! zmax e))
-                    (list i e
-                          (equal? (binary-heap-find-min heap) zmin)
-                          (equal? (binary-heap-find-max heap) zmax))))
-                input))))
-    (let1 hp1 (binary-heap-copy heap)
-      (test* (format "heap(~s) clear" len) #t
-             (begin
-               (binary-heap-clear! hp1)
-               (binary-heap-empty? hp1))))
-    (let1 hp1 (binary-heap-copy heap)
-      (test* (format "heap(~s) deletion from min" len)
-             data
-             (map-in-order (^_ (begin0 (binary-heap-pop-min! hp1)
-                                 (binary-heap-check hp1)))
-                           (iota len))))
-    (let1 hp1 (binary-heap-copy heap)
-      (test* (format "heap(~s) deletion from max" len)
-             (reverse data)
-             (map-in-order (^_ (begin0 (binary-heap-pop-max! hp1)
-                                 (binary-heap-check hp1)))
-                           (iota len))))
-    (let1 hp1 (binary-heap-copy heap)
-      (test* (format "heap(~s) deletion mixed" len)
-             (map list
-                  (take data (quotient len 2))
-                  (take (reverse data) (quotient len 2)))
-             (map-in-order (^_ (let* ([x (binary-heap-pop-min! hp1)]
-                                      [y (binary-heap-pop-max! hp1)])
-                                 (binary-heap-check hp1)
-                                 (list x y)))
-                           (iota (quotient len 2)))))
-    )
-  
-  (do-test (iota 15) default-comparator)
-  (do-test (iota 253) default-comparator)
-  (do-test (reverse (iota 33))
-           (make-comparator/compare number? #t (^[a b] (- (compare a b))) #f))
-  (do-test '(#\a #\b #\c #\d #\e) default-comparator)
-  )
-
-(let ((rs (make-random-source)))
-  (define (suck-all heap)
-    (do ([r '() (cons (binary-heap-pop-min! heap) r)])
-        [(binary-heap-empty? heap) (reverse r)]
-      ))
-  (define (do-heapify lis builder comparator)
-    (test* (format "heapify ~s" lis)
-           lis
-           (let* ([src  (builder (shuffle lis rs))]
-                  [heap (build-binary-heap src :comparator comparator)])
-             (binary-heap-check heap)
-             (suck-all heap))))
-  (define (do-scan lis pred item)
-    (test* (format "find, remove, delete ~s" lis)
-           (list (boolean (find pred lis))
-                 (remove pred lis)
-                 (delete item lis))
-           (let ([heap (build-binary-heap (list->vector (shuffle lis rs)))])
-             (list (if-let1 r (binary-heap-find heap pred)
-                     (pred r)
-                     #f)
-                   (let1 h (binary-heap-copy heap)
-                     (binary-heap-remove! h pred)
-                     (binary-heap-check h)
-                     (suck-all h))
-                   (let1 h (binary-heap-copy heap)
-                     (binary-heap-delete! h item)
-                     (binary-heap-check h)
-                     (suck-all h))))))
-
-  (do-heapify '() list->vector default-comparator)
-  (do-heapify '(1) list->vector default-comparator)
-  (do-heapify (iota 33) list->vector default-comparator)
-  (do-heapify '("a" "aa" "b" "bb" "c" "cc" "d" "dd") list->vector
-              string-comparator)
-
-  (do-scan '() odd? 1)
-  (do-scan (iota 23) odd? 5)
-  (do-scan (iota 42) (^n (< (modulo n 3) 2)) 91)
-  )
-
-(let ()
-  (define (test-swap source actions)
-    ;; actions : ((min|max item expected-result expected-min expected-max) ...)
-    (let1 hp (build-binary-heap source)
-      (dolist [action actions]
-        (match-let1 (minmax item xresult xmin xmax) action
-          (test* (format "swap ~s ~s ~s" source minmax item)
-                 (list xresult xmin xmax)
-                 (let1 r ((ecase minmax
-                            [(min) binary-heap-swap-min!]
-                            [(max) binary-heap-swap-max!])
-                          hp item)
-                   (list r
-                         (binary-heap-find-min hp)
-                         (binary-heap-find-max hp))))))))
-
-  (test-swap (vector 1 3 5 7 9 11 13 15 17)
-             '((min 4 1 3 17)
-               (min 2 3 2 17)
-               (max 16 17 2 16)
-               (max 1 16 1 15)
-               (min 20 1 2 20)
-               (min 10 2 4 20)))
-  (test-swap (vector 1)
-             '((min 2 1 2 2)
-               (max 1 2 1 1)
-               (min 0 1 0 0)))
-  (test-swap (vector 3 1)
-             '((min 4 1 3 4)
-               (max 5 4 3 5)
-               (min 6 3 5 6)
-               (max 0 6 0 5)))
-  (test-swap (vector 4 2 5)
-             '((max 3 5 2 4)
-               (max 1 4 1 3)
-               (max 1 3 1 2)
-               (max 1 2 1 1)
-               (max 1 1 1 1)))
-  )
+;; data.heap is tested in ext/data, for its test depends on math.mt-random.
 
 ;;;========================================================================
 (test-section "data.priority-map")
 (use data.priority-map)
 (test-module 'data.priority-map)
 
+(use scheme.list)
+
 (let* ([data '((a . 5) (b . 2) (c . 9) (d . -1) (e . -59) (f . 0) (g . 9))])
   (define (make-populated)
     (rlet1 m (make-priority-map)
@@ -582,11 +415,15 @@
     (test* "get" (map cdr data)
            (map (^p (dict-get m (car p))) data))
     (test* "min" '((e) . -59)
-           (priority-map-min m)
+           (priority-map-min-all m)
            compare-minmax-result)
+    (test* "min" '(e . -59)
+           (priority-map-min m))
     (test* "max" '((c g) . 9)
-           (priority-map-max m)
+           (priority-map-max-all m)
            compare-minmax-result)
+    (test* "max" (test-one-of '(g . 9) '(c . 9))
+           (priority-map-max m))
     (test* "delete" (remove (^p (eq? (car p) 'c)) data)
            (begin (dict-delete! m 'c)
                   (sort-by (dict->alist m) car))))
@@ -607,8 +444,89 @@
                            (let1 z (priority-map-pop-max! m)
                              (cons z (pop m)))))
                        m))))
+
+  (let ([tm (alist->tree-map data)])
+    (test* "alist->priority-map" data
+           (sort (dict->alist (alist->priority-map data))))
+    (test* "dict->priority-map" data
+           (sort (dict->alist (dict->priority-map tm)))))
   )
 
+;;;========================================================================
+;; range
+(test-section "data.range")
+(use data.range)
+(test-module 'data.range)
+
+;; SRFI-196 is a subset of data.range, and we test it here.
+
+(define-module srfi-196-tests
+  (use gauche.test)
+  (test-include-r7 "include/srfi-196-test"))
+
+;; gauche extensions
+
+(test* "range-ref out of range" (test-error <error> #/Index out of range/)
+       (range-ref (iota-range 10) 11))
+(test* "range-ref fallback" 'oops
+       (range-ref (iota-range 10) 11 'oops))
+(test* "range-ref fallback" 'oops
+       (range-ref (iota-range 10) -1 'oops))
+(test* "range-first fallback" 'wow
+       (range-first (iota-range 0) 'wow))
+(test* "range-last fallback" 'boo
+       (range-last (iota-range 0) 'boo))
+
+;; https://github.com/shirok/Gauche/pull/1004
+(test* "range-append" '(0 1 2 3 4 10 11 12 13 14)
+       (range->list (range-append (numeric-range 0 5) (numeric-range 10 15))))
+
+(test* "range-reverse start/end" '(5 4 3)
+       (range->list (range-reverse (iota-range 7) 3 6)))
+(test* "range-reverse start/end nested" '(5 4 3)
+       (range->list (range-reverse (subrange (iota-range 10) 1 9) 2 5)))
+(test* "range-reverse start/end nested" '(6 5 4)
+       (range->list (range-reverse
+                     (subrange (subrange (iota-range 10) 1 9) 1 6)
+                     2 5)))
+
+(test* "vector-range start/end" '(c d e f g)
+       (range->list (vector-range '#(a b c d e f g) 2)))
+(test* "vector-range start/end" '(c d e)
+       (range->list (vector-range '#(a b c d e f g) 2 5)))
+(test* "uvector-range" '(0 1 2 3 4)
+       (range->list (uvector-range '#u8(0 1 2 3 4))))
+(test* "uvector-range start/end" '(1 2 3 4)
+       (range->list (uvector-range '#u8(0 1 2 3 4 5) 1 5)))
+(test* "string-range start/end" '(#\c #\d #\e #\f #\g)
+       (range->list (string-range "abcdefg" 2)))
+(test* "string-range start/end" '(#\c #\d #\e)
+       (range->list (string-range "abcdefg" 2 5)))
+(test* "bitvector/bool-range" '(#t #f #t #t)
+       (range->list (bitvector/bool-range #*0101100 1 5)))
+(test* "bitvector/int-range" '(0 1 1 0 0)
+       (range->list (bitvector/int-range #*0101100 2)))
+
+;; sequence protocol
+(test* "range as a sequence (ref)" 3 (~ (numeric-range 1 6) 2))
+(test* "range as a sequence (iterate)" '(1 2 3 4 5)
+       (coerce-to <list> (numeric-range 1 6)))
+(test* "range as a sequence (build from list)" '(a b c d e)
+       (range->list (coerce-to <range> '(a b c d e))))
+(test* "range as a sequence (build from vector)" '(a b c d e)
+       (range->list (coerce-to <range> '#(a b c d e))))
+(test* "range as a sequence (subseq)" '(3 4 5 6 7)
+       (range->list (subseq (numeric-range 0 10) 3 8)))
+
+;; SRFI-42 extension
+(define-module range-srfi-42-test
+  (use gauche.test)
+  (use data.range)
+  (use srfi.42)
+  (test* "SRFI-42 :range" '((a 0) (b 1) (c 2) (d 3) (e 4))
+         (list-ec (: v (index k) (vector-range '#(a b c d e)))
+                  (list v k)))
+  )
 
 ;;;========================================================================
 ;; skew-list
@@ -733,9 +651,9 @@
   (t-take 6 1)    ; 6 = [3 3]
   (t-take 6 2)
   ;; general case
-  (t-take 6 3)  
-  (t-take 6 4)  
-  (t-take 6 5)  
+  (t-take 6 3)
+  (t-take 6 4)
+  (t-take 6 5)
   (t-take 26 2)
   (t-take 26 3)
   (t-take 26 5)
@@ -743,7 +661,7 @@
   (t-take 26 15)
   )
 
-;; append 
+;; append
 (let ()
   (define (t-append i j)
     (let ([x (iota i)]
@@ -774,7 +692,7 @@
   )
 
 ;; sequence protocol
-(test* "skew-list size-of" 5 
+(test* "skew-list size-of" 5
        (size-of (list->skew-list '(a b c d e))))
 (test* "skew-list iterator" '(a b c d e f g h i j k l m n o p q)
        (coerce-to <list> (list->skew-list '(a b c d e f g h i j k l m n o p q))))
@@ -783,6 +701,15 @@
 (test* "skew-list ref" '(a b c d e)
        (map (cute ref (list->skew-list '(a b c d e)) <>) (iota 5)))
 
+;;;========================================================================
+;; ulid
+(test-section "data.ulid")
+(use data.ulid)
+(test-module 'data.ulid)
 
+(define-module ulid-test
+  (use gauche.test)
+  (use data.ulid)
+  (test-include-r7 "include/ulid-test" (exclude (ulid))))
 
 (test-end)
