@@ -3,7 +3,7 @@
 ;;
 
 (use gauche.test)
-(use srfi-1)
+(use scheme.list)
 
 (test-start "dictionary framework")
 (use gauche.dictionary)
@@ -19,9 +19,9 @@
   (test* "fold" '((a . 1) (b . 2)) (dict-fold d acons '())
          (cut lset= equal? <> <>))
   (test* "seek" '(4 b 2)
-         (dict-seek d (^[k v] (and (eq? k 'b) (* v 2))) list (^[] 'huh?)))
+         (dict-seek d (^[k v] (and (eq? k 'b) (* v 2))) (^[] 'huh?) list))
   (test* "seek" 'huh?
-         (dict-seek d (^[k v] (and (eq? k 'c) (* v 2))) list (^[] 'huh?)))
+         (dict-seek d (^[k v] (and (eq? k 'c) (* v 2))) (^[] 'huh?) list))
   (test* "find" '(a 1) (values->list (dict-find d (^[k v] (eqv? v 1)))))
   (test* "find" '(#f #f) (values->list (dict-find d (^[k v] (eqv? v 3)))))
   (test* "find" 'boo (dict-find d (^[k v] (eqv? v 3)) (^[] 'boo)))
@@ -128,4 +128,3 @@
   )
 
 (test-end)
-
