@@ -1,4 +1,4 @@
-;; test exception handling system 
+;; test exception handling system
 ;; this must come after primsyn, error, macro and object tests.
 
 (use gauche.test)
@@ -25,7 +25,7 @@
          (map (cut ref e <>) '(message errno))))
 
 ;;--------------------------------------------------------------------
-(test-section "srfi-35 constructors, predicates and accessors")
+(test-section "SRFI-35 constructors, predicates and accessors")
 
 (test* "make-condition <error>" '(#t #t #f "moo")
        (let ((e (make-condition <error> 'message "moo")))
@@ -76,7 +76,7 @@
           (condition-ref e 'message)
           (condition-ref e 'errno)
           (condition-ref e 'port))))
-          
+
 (test* "make-compound-condition"
        `(#t #t #t "message" 12 ,(current-input-port))
        (let ((e (make-compound-condition
@@ -113,9 +113,9 @@
           (list (condition-ref i 'message) (condition-ref i 'port))
           (list (condition-ref s 'message) (condition-ref s 'errno)))
          ))
-       
+
 ;;--------------------------------------------------------------------
-(test-section "srfi-35 style condition definitions")
+(test-section "SRFI-35 style condition definitions")
 
 (define-condition-type &c &condition
   c?
@@ -166,6 +166,36 @@
            (lambda () (error &c1 :a "ok"))))
   )
 
+;;--------------------------------------------------------------------
+(test-section "with-exception-handler")
+
+(test* "with-exception-handler" '(got oops)
+       (call/cc
+        (lambda (ret)
+          (with-exception-handler
+           (lambda (e) (ret `(got ,e)))
+           (lambda () (raise 'oops) 'huh?)))))
+
+(test* "with-exception-handler (returning)" 'huh?
+       (with-exception-handler
+        (lambda (e) `(got ,e))
+        (lambda () (raise 'oops) 'huh?)))
+
+(test* "with-exception-handler (returning mv)" '(aha! oops)
+       (values->list
+        (with-exception-handler
+         (lambda (e) (values 'aha! e))
+         (lambda () (raise 'oops)))))
+
+;; Now built-in with-exception-hanlder has r7rs semantics
+(test* "with-exception-handler (raise in handler)"  '(outer (inner oops))
+       (with-exception-handler
+        (lambda (e) `(outer ,e))
+        (lambda ()
+          (with-exception-handler
+           (lambda (e) (raise `(inner ,e)))
+           (lambda () (raise 'oops))))))
+
 ;;--------------------------------------------------------------------
 (test-section "guard")
 
@@ -174,7 +204,7 @@
                ((symbol? x) (cons 'symbol x))
                ((is-a? x <error>) 'caught-error))
          (raise 'a)))
-       
+
 (test* "guard" 'caught-error
        (guard (x
                ((symbol? x) (cons 'symbol x))
