@@ -5,8 +5,8 @@
 (use gauche.test)
 (use gauche.generator)
 (test-start "file utilities")
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 
 ;; shorthand of normalizing pathname.  this doesn't do anything on
 ;; unix, but on Windows the separator in PATHNAME is replaced.
@@ -17,11 +17,10 @@
 
 (define (md p) (sys-mkdir p #o777))
 (define (mf p) (with-output-to-file p (cut display "z")))
-(define (rmrf p) (sys-system #"rm -rf ~p"))
 
-(and glob (test-module 'gauche.fileutil)) ;; trigger autoload
+(and sys-tm->alist (test-module 'gauche.fileutil)) ;; trigger autoload
 
-(rmrf "tmp1.o")
+(test-remove-files "tmp1.o")
 
 (define (file-pred-tests path expected)
   (test* #"file-exists? (~path)" (car expected) (file-exists? path))
@@ -44,152 +43,150 @@
 ;;
 ;; glob test.
 ;; Note: on Windows the file/directory name can't end with a period.
-;; 
-(let ()
-  (md "tmp1.o")
-  (md "tmp1.o/a")
-  (mf "tmp1.o/a/b")
-  (md "tmp1.o/a/cc")
-  (mf "tmp1.o/a/cc/a")
-  (mf "tmp1.o/a/.d")
-  (md "tmp1.o/.a")
-  (mf "tmp1.o/.a/b")
-  (md "tmp1.o/.a/.d")
-  (md "tmp1.o/aa")
-  (mf "tmp1.o/aa/b")
-  (mf "tmp1.o/aa/.d")
-  (mf "tmp1.o/a.a")
-  (mf "tmp1.o/a.b")
-  (mf "tmp1.o/a.a.a")
+;;
+($ test-with-temporary-directory "tmp1.o"
+   (^[]
+     (md "tmp1.o/a")
+     (mf "tmp1.o/a/b")
+     (md "tmp1.o/a/cc")
+     (mf "tmp1.o/a/cc/a")
+     (mf "tmp1.o/a/.d")
+     (md "tmp1.o/.a")
+     (mf "tmp1.o/.a/b")
+     (md "tmp1.o/.a/.d")
+     (md "tmp1.o/aa")
+     (mf "tmp1.o/aa/b")
+     (mf "tmp1.o/aa/.d")
+     (mf "tmp1.o/a.a")
+     (mf "tmp1.o/a.b")
+     (mf "tmp1.o/a.a.a")
 
-  ;; literal
-  (test* "glob a.a" (n "tmp1.o/a.a")
-         (glob "tmp1.o/a.a")
-         (pa$ lset= equal?))
+     ;; literal
+     (test* "glob a.a" (n "tmp1.o/a.a")
+            (glob "tmp1.o/a.a")
+            (pa$ lset= equal?))
 
-  ;; nomatch
-  (test* "glob z" '()
-         (glob "tmp1.o/z")
-         (pa$ lset= equal?))
+     ;; nomatch
+     (test* "glob z" '()
+            (glob "tmp1.o/z")
+            (pa$ lset= equal?))
 
-  ;; wildcard
-  (test* "glob *" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/a.a"
-                     "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/*")
-         (pa$ lset= equal?))
+     ;; wildcard
+     (test* "glob *" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/a.a"
+                        "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/*")
+            (pa$ lset= equal?))
 
-  (test* "glob a.*" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/a.*")
-         (pa$ lset= equal?))
+     (test* "glob a.*" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/a.*")
+            (pa$ lset= equal?))
 
-  (test* "glob .*" (n "tmp1.o/.a" "tmp1.o/." "tmp1.o/..")
-         (glob "tmp1.o/.*")
-         (pa$ lset= equal?))
+     (test* "glob .*" (n "tmp1.o/.a" "tmp1.o/." "tmp1.o/..")
+            (glob "tmp1.o/.*")
+            (pa$ lset= equal?))
 
-  (test* "glob ?" (n "tmp1.o/a")
-         (glob "tmp1.o/?")
-         (pa$ lset= equal?))
+     (test* "glob ?" (n "tmp1.o/a")
+            (glob "tmp1.o/?")
+            (pa$ lset= equal?))
 
-  (test* "glob *?" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/a.a"
-                      "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/*?")
-         (pa$ lset= equal?))
+     (test* "glob *?" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/a.a"
+                         "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/*?")
+            (pa$ lset= equal?))
 
-  (test* "glob ??" (n "tmp1.o/aa")
-         (glob "tmp1.o/??")
-         (pa$ lset= equal?))
+     (test* "glob ??" (n "tmp1.o/aa")
+            (glob "tmp1.o/??")
+            (pa$ lset= equal?))
 
-  (test* "glob *.*" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/*.*")
-         (pa$ lset= equal?))
+     (test* "glob *.*" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/*.*")
+            (pa$ lset= equal?))
 
-  (test* "glob */*" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
-         (glob "tmp1.o/*/*")
-         (pa$ lset= equal?))
+     (test* "glob */*" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
+            (glob "tmp1.o/*/*")
+            (pa$ lset= equal?))
 
-  (test* "glob */?" (n "tmp1.o/a/b" "tmp1.o/aa/b")
-         (glob "tmp1.o/*/?")
-         (pa$ lset= equal?))
+     (test* "glob */?" (n "tmp1.o/a/b" "tmp1.o/aa/b")
+            (glob "tmp1.o/*/?")
+            (pa$ lset= equal?))
 
-  (test* "glob *  (chdir)" (n "a" "aa" "a.a" "a.b" "a.a.a")
-         (begin (sys-chdir "tmp1.o") (begin0 (glob "*") (sys-chdir "..")))
-         (pa$ lset= equal?))
+     (test* "glob *  (chdir)" (n "a" "aa" "a.a" "a.b" "a.a.a")
+            (begin (sys-chdir "tmp1.o") (begin0 (glob "*") (sys-chdir "..")))
+            (pa$ lset= equal?))
 
-  (test* "glob */" (n "tmp1.o/a/" "tmp1.o/aa/")
-         (glob "tmp1.o/*/")
-         (pa$ lset= equal?))
+     (test* "glob */" (n "tmp1.o/a/" "tmp1.o/aa/")
+            (glob "tmp1.o/*/")
+            (pa$ lset= equal?))
 
-  ;; **
-  (test* "glob tmp1.o/**/?" (n "tmp1.o/a" "tmp1.o/a/b" "tmp1.o/a/cc/a"
-                               "tmp1.o/aa/b")
-         (glob "tmp1.o/**/?")
-         (pa$ lset= equal?))
+     ;; **
+     (test* "glob tmp1.o/**/?" (n "tmp1.o/a" "tmp1.o/a/b" "tmp1.o/a/cc/a"
+                                  "tmp1.o/aa/b")
+            (glob "tmp1.o/**/?")
+            (pa$ lset= equal?))
 
-  ;; multi
-  (test* "glob * .* (multi)" (n "tmp1.o/." "tmp1.o/.." "tmp1.o/.a" "tmp1.o/a"
-                                "tmp1.o/aa" "tmp1.o/a.a" "tmp1.o/a.b"
-                                "tmp1.o/a.a.a")
-         (glob '("tmp1.o/*" "tmp1.o/.*"))
-         (pa$ lset= equal?))
+     ;; multi
+     (test* "glob * .* (multi)" (n "tmp1.o/." "tmp1.o/.." "tmp1.o/.a" "tmp1.o/a"
+                                   "tmp1.o/aa" "tmp1.o/a.a" "tmp1.o/a.b"
+                                   "tmp1.o/a.a.a")
+            (glob '("tmp1.o/*" "tmp1.o/.*"))
+            (pa$ lset= equal?))
 
-  ;; braces
-  (test* "glob {a,aa}/{b,cc}" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
-         (glob '("tmp1.o/{a,aa}/{b,cc}"))
-         (pa$ lset= equal?))
-  (test* "glob {a{,a,.{a,b}}}" (n "tmp1.o/a" "tmp1.o/aa"
-                                 "tmp1.o/a.a" "tmp1.o/a.b")
-         (glob '("tmp1.o/{a{,a,.{a,b}}}"))
-         (pa$ lset= equal?))
-  (test* "glob {a/*,aa/*}" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
-         (glob '("tmp1.o/{a/*,aa/*}"))
-         (pa$ lset= equal?))
-  (test* "glob {,?/}*" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/a" "tmp1.o/aa"
-                          "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob '("tmp1.o/{,?/}*"))
-         (pa$ lset= equal?))
-  (test* "glob {,.}*" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/.a" "tmp1.o/."
-                         "tmp1.o/.." "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob '("tmp1.o/{,.}*"))
-         (pa$ lset= equal?))
+     ;; braces
+     (test* "glob {a,aa}/{b,cc}" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
+            (glob '("tmp1.o/{a,aa}/{b,cc}"))
+            (pa$ lset= equal?))
+     (test* "glob {a{,a,.{a,b}}}" (n "tmp1.o/a" "tmp1.o/aa"
+                                     "tmp1.o/a.a" "tmp1.o/a.b")
+            (glob '("tmp1.o/{a{,a,.{a,b}}}"))
+            (pa$ lset= equal?))
+     (test* "glob {a/*,aa/*}" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/aa/b")
+            (glob '("tmp1.o/{a/*,aa/*}"))
+            (pa$ lset= equal?))
+     (test* "glob {,?/}*" (n "tmp1.o/a/b" "tmp1.o/a/cc" "tmp1.o/a" "tmp1.o/aa"
+                             "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob '("tmp1.o/{,?/}*"))
+            (pa$ lset= equal?))
+     (test* "glob {,.}*" (n "tmp1.o/a" "tmp1.o/aa" "tmp1.o/.a" "tmp1.o/."
+                            "tmp1.o/.." "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob '("tmp1.o/{,.}*"))
+            (pa$ lset= equal?))
 
-  ;; charset
-  (test* "glob a.[ab]" (n "tmp1.o/a.a" "tmp1.o/a.b")
-         (glob "tmp1.o/a.[ab]")
-         (pa$ lset= equal?))
-  (test* "glob a.[[:alpha:]]" (n "tmp1.o/a.a" "tmp1.o/a.b")
-         (glob "tmp1.o/a.[[:alpha:]]")
-         (pa$ lset= equal?))
-  (test* "glob *.[[:alpha:]]" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/*.[[:alpha:]]")
-         (pa$ lset= equal?))
-  (test* "glob *.[![:alpha:]]" '()
-         (glob "tmp1.o/*.[![:alpha:]]")
-         (pa$ lset= equal?))
-  (test* "glob *.[^[:alpha:]]" '()
-         (glob "tmp1.o/*.[^[:alpha:]]")
-         (pa$ lset= equal?))
-  (test* "glob *.[^A-Z]" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-         (glob "tmp1.o/*.[^A-Z]")
-         (pa$ lset= equal?))
+     ;; charset
+     (test* "glob a.[ab]" (n "tmp1.o/a.a" "tmp1.o/a.b")
+            (glob "tmp1.o/a.[ab]")
+            (pa$ lset= equal?))
+     (test* "glob a.[[:alpha:]]" (n "tmp1.o/a.a" "tmp1.o/a.b")
+            (glob "tmp1.o/a.[[:alpha:]]")
+            (pa$ lset= equal?))
+     (test* "glob *.[[:alpha:]]" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/*.[[:alpha:]]")
+            (pa$ lset= equal?))
+     (test* "glob *.[![:alpha:]]" '()
+            (glob "tmp1.o/*.[![:alpha:]]")
+            (pa$ lset= equal?))
+     (test* "glob *.[^[:alpha:]]" '()
+            (glob "tmp1.o/*.[^[:alpha:]]")
+            (pa$ lset= equal?))
+     (test* "glob *.[^A-Z]" (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "tmp1.o/*.[^A-Z]")
+            (pa$ lset= equal?))
 
-  ;; specifying current/root dir
-  (test* "glob w/alt root dir"  (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-          (glob "/*.*" :folder (make-glob-fs-fold :root-path "tmp1.o"))
-          (pa$ lset= equal?))
-  (test* "glob w/alt current dir"  (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
-          (glob "*.*" :folder (make-glob-fs-fold :current-path "tmp1.o"))
-          (pa$ lset= equal?))
-
-  (rmrf "tmp1.o")
-  )
+     ;; specifying current/root dir
+     (test* "glob w/alt root dir"  (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "/*.*" :folder (make-glob-fs-fold :root-path "tmp1.o"))
+            (pa$ lset= equal?))
+     (test* "glob w/alt current dir"  (n "tmp1.o/a.a" "tmp1.o/a.b" "tmp1.o/a.a.a")
+            (glob "*.*" :folder (make-glob-fs-fold :current-path "tmp1.o"))
+            (pa$ lset= equal?))
+     ))
 
 ;;------------------------------------------------------------------
 (test-section "file.filter")
 (use file.filter)
 (test-module 'file.filter)
 
-(rmrf "tmp1.o")
-(rmrf "tmp2.o")
+(test-remove-files "tmp1.o" "tmp2.o")
+
 (with-output-to-file "tmp1.o"
   (cut display "aaa bbb ccc ddd\neee fff ggg hhh\n"))
 
@@ -213,7 +210,7 @@
                 :output "tmp2.o"))
          (call-with-input-file "tmp2.o" port->string)))
 
-(sys-unlink "tmp2.o")
+(test-remove-files "tmp2.o")
 
 (test* "file.filter cleanup" #f
        (with-error-handler (^e (file-exists? "tmp2.o"))
@@ -221,7 +218,7 @@
               (cut file-filter (^[in out] (error "yyy"))
                    :output "tmp2.o"))))
 
-(sys-unlink "tmp2.o")
+(test-remove-files "tmp2.o")
 
 (test* "file.filter cleanup" #t
        (with-error-handler (^e (file-exists? "tmp2.o"))
@@ -230,7 +227,7 @@
                    :output "tmp2.o"
                    :keep-output? #t))))
 
-(sys-unlink "tmp2.o")
+(test-remove-files "tmp2.o")
 
 (let ()
   (define (t tempfile)
@@ -251,8 +248,7 @@
   (t "foo")
   (t #t))
 
-(sys-unlink "tmp1.o")
-(sys-unlink "tmp2.o")
+(test-remove-files "tmp1.o" "tmp2.o")
 
 (test* "file.filter rename-hook" #t
        (begin
@@ -265,7 +261,7 @@
                         :leave-unchanged #t)
            (eqv? i1 (~ (sys-stat "tmp1.o")'ino)))))
 
-(sys-unlink "tmp1.o")
+(test-remove-files "tmp1.o")
 
 (let ()
   (define (f2s f) (call-with-input-file f
@@ -294,15 +290,14 @@
               :output "test2.o"
               :reader reader)
              (list r (f2s "test2.o")))))
-  
+
   (with-output-to-file "test1.o" (cut display "abc\n(def\nghi)"))
   (unwind-protect
       (begin
         (t read-line 3 "\"abc\"\"(def\"\"ghi)\"")
         (t read-char 13 "#\\a#\\b#\\c#\\newline#\\(#\\d#\\e#\\f#\\newline#\\g#\\h#\\i#\\)")
         (t read 2 "abc(def ghi)"))
-    (sys-unlink "test1.o")
-    (sys-unlink "test2.o")))
+    (test-remove-files "test1.o" "test2.o")))
 
 (test* "file-filter-fold example"
        "  1: abc\n  3: def\n"
@@ -318,9 +313,8 @@
                                    (format out "~3d: ~a\n" nc line))
                                  (+ nc 1))
                                1 :input "test1.o" :output "test2.o")
-             (call-with-input-file "test2.o" 
+             (call-with-input-file "test2.o"
                (^i (call-with-output-string (^o (copy-port i o))))))
-         (sys-unlink "test1.o")
-         (sys-unlink "test2.o")))
+         (test-remove-files "test1.o" "test2.o")))
 
 (test-end)
