@@ -114,36 +114,18 @@
   (test* "port-attribute-set! (autocreate)" 'ok
          (begin (port-attribute-set! p 'my-attribute 'ok)
                 (port-attribute-ref p 'my-attribute)))
-  (test* "port-attribute-create! (existent)" (test-error)
-         (port-attribute-create! p 'my-attribute))
-  (test* "port-attribute-create! (procedural)" 1
-         (let1 val 0
-           (define (get port :optional fallback) (inc! val) val)
-           (define (set port newval) (set! val newval))
-           (port-attribute-create! p 'my-attribute2 get set)
-           (port-attribute-ref p 'my-attribute2)))
-  (test* "port-attribute (procedural)" 101
-         (begin
-           (port-attribute-set! p 'my-attribute2 100)
-           (port-attribute-ref p 'my-attribute2)))
-  (test* "port-attribute again (procedural)" 102
-         (port-attribute-ref p 'my-attribute2))
-  (test* "port-attribute-create! (procedural, read-only)" 11
-         (let1 val 10
-           (define (get port :optional fallback) (inc! val) val)
-           (port-attribute-create! p 'my-attribute3 get)
-           (port-attribute-ref p 'my-attribute3)))
-  (test* "port-attribute-set! (procedural, read-only)" (test-error)
-         (port-attribute-set! p 'my-attribute3 20))
   (test* "port-attributes" '((my-attribute . ok)
                              (my-attribute2 . 103)
                              (my-attribute3 . 12)
                              (name . "tmp.o"))
-         (alist-delete 'reader-lexical-mode
-                       (sort (port-attributes p)
-                             string<?
-                             ($ symbol->string $ car $))))
-  (test* "port-attribute-delete!" '((my-attribute3 . 13) (name . "tmp.o"))
+         (begin
+           (port-attribute-set! p 'my-attribute2 103)
+           (port-attribute-set! p 'my-attribute3 12)
+           (alist-delete 'reader-lexical-mode
+                         (sort (port-attributes p)
+                               string<?
+                               ($ symbol->string $ car $)))))
+  (test* "port-attribute-delete!" '((my-attribute3 . 12) (name . "tmp.o"))
          (begin
            (port-attribute-delete! p 'my-attribute)
            (port-attribute-delete! p 'my-attribute2)
@@ -154,8 +136,6 @@
          (port-attribute-set! p 'name "foo"))
   (test* "name attribute is undeletable" (test-error <error> #/not deletable/)
          (port-attribute-delete! p 'name))
-  (test* "name attribute is overridable" (test-error <error> #/Couldn't create/)
-         (port-attribute-create! p 'name (^ _ #f)))
   )
 
 ;;-------------------------------------------------------------------
@@ -344,7 +324,7 @@
              [y (open-input-string "cd")]
              [r '()]
              [restart #f])
-         (if (call/cc 
+         (if (call/cc
               (^[escape]
                 (with-input-from-port x
                   (^()
@@ -365,7 +345,7 @@
           (let ([x (open-output-string)]
                 [y (open-output-string)]
                 [restart #f])
-            (if (call/cc 
+            (if (call/cc
                  (^[escape]
                    (with x
                      (^()
@@ -662,10 +642,14 @@
        (format "~10,,,'*,15s" '("abc" "def" "ghi" "jkl")))
 (test* "format ~s" "(\"abc\" \"def\" \"g"
        (format "~10,,,'*,15@s" '("abc" "def" "ghi" "jkl")))
-(test* "format ~s" "(\"abc\" \"def ..."
+(test* "format ~s" "(\"abc\" \"def\"..."
        (format "~10,,,'*,15:s" '("abc" "def" "ghi" "jkl")))
-(test* "format ~s" "(\"abc\" \"def ..."
+(test* "format ~s" "(\"abc\" \"def\"..."
        (format "~10,,,'*,15@:s" '("abc" "def" "ghi" "jkl")))
+(test* "format ~s" "(\"ab\\\"c\" #\\\" \"de\"..."
+       (format "~10,,,'*,20@:s" '("ab\"c" #\" "defghi")))
+(test* "format ~s" "(\"ab\\\"c\" #\\\" \"d\\\\\".."
+       (format "~10,,,'*,20@:s" '("ab\"c" #\" "d\nfghi")))
 
 (test* "format ~a" "abc"           (format "~a" "abc"))
 (test* "format ~a" "abc       "    (format "~10a" "abc"))
@@ -677,14 +661,14 @@
 (test* "format ~a" "abc*******"    (format "~10,,,'*a" "abc"))
 (test* "format ~a" "*******abc"    (format "~10,,,'*@a" "abc"))
 
-(test* "format ~a" "(abc def ghi jk"
-       (format "~10,,,'*,15a" '("abc" "def" "ghi" "jkl")))
-(test* "format ~a" "(abc def ghi jk"
-       (format "~10,,,'*,15@a" '("abc" "def" "ghi" "jkl")))
-(test* "format ~a" "(abc def gh ..."
-       (format "~10,,,'*,15:a" '("abc" "def" "ghi" "jkl")))
-(test* "format ~a" "(abc def gh ..."
-       (format "~10,,,'*,15@:a" '("abc" "def" "ghi" "jkl")))
+(test* "format ~a" "(abc def ghi j"
+       (format "~10,,,'*,14a" '("abc" "def" "ghi" "jkl")))
+(test* "format ~a" "(abc def ghi j"
+       (format "~10,,,'*,14@a" '("abc" "def" "ghi" "jkl")))
+(test* "format ~a" "(abc def g ..."
+       (format "~10,,,'*,14:a" '("abc" "def" "ghi" "jkl")))
+(test* "format ~a" "(abc def g ..."
+       (format "~10,,,'*,14@:a" '("abc" "def" "ghi" "jkl")))
 
 (test* "format ~w" "(#0=(1) #0#)"
        (format "~w" (let1 x '(1) (list x x))))
@@ -840,6 +824,14 @@
 (test* "format ~@*" "1 2 5"
        (format "~a ~a ~4@*~a" 1 2 3 4 5))
 
+(test* "format ~&" "abc" (format "~&abc"))
+(test* "format ~&" "abc\ndef" (format "abc~&def"))
+(test* "format ~&" "abc\ndef" (format "abc\n~&def"))
+(test* "format ~&" "abc\ndef" (format "abc~%~&def"))
+(test* "format ~&" "abc\ndef" (format "abc~%~&~&def"))
+(test* "format ~n&" "\n\nabc" (format "~3&abc"))
+(test* "format ~n&" "abc\n\n\ndef" (format "abc~3&def"))
+
 (test* "format single character"
        '(("~" "~~~" "~~~~~")
          ("\t" "\t\t\t" "\t\t\t\t\t")
@@ -850,6 +842,9 @@
                       (format (string #\~ #\v c) 5)))
             '(#\~ #\t #\% #\|)))
 
+(test* "format string node concatenation"
+       "~az" (format "~~a~a" 'z))
+
 (test* "format ~?" "ding[\"dang\" dong]ping"
        (format "~s~?~a" 'ding "[~s ~a]" '("dang" "dong") "ping"))
 (test* "format ~?" "ding[\"dang\" gong]ping" ; extra args for ~? are ignored
@@ -861,6 +856,26 @@
 (test* "format nested ~@?" "<[<<{\"O\" O}>>]>"
        (format "~s[~@?]~s" '< "~s{~@?}~s" '<< "~s ~a" "O" "O" '>> '>))
 
+
+(test* "format ~ + newline" "abcdef"
+       (format "abc~\ndef"))
+(test* "format ~ + newline" "abcdef"
+       (format "abc~\n     def"))
+(test* "format ~ + newline" "abcdef"
+       (format "abc~\n    \n    def"))
+(test* "format ~@ + newline" "abc\ndef"
+       (format "abc~@\ndef"))
+(test* "format ~@ + newline" "abc\ndef"
+       (format "abc~@\n    def"))
+(test* "format ~@ + newline" "abc\ndef"
+       (format "abc~@\n    \ndef"))
+(test* "format ~: + newline" "abcdef"
+       (format "abc~:\ndef"))
+(test* "format ~: + newline" "abc   def"
+       (format "abc~:\n   def"))
+(test* "format ~: + newline" "abc   \n   def"
+       (format "abc~:\n   \n   def"))
+
 (test* "format incomplete tilde sequence" (test-error)
        (format "~"))
 (test* "format incomplete tilde sequence" (test-error)
@@ -907,7 +922,7 @@
        '(foo bar this is outside of comment "hence this is in a string")
        (read-from-string
         "(foo bar #| grok
- \"the following bar-and-sharp terminates the comment |# 
+ \"the following bar-and-sharp terminates the comment |#
    this is outside of comment
  \"hence this is in a string\")
  "))
@@ -916,7 +931,7 @@
              "#| this is a string, not a comment"
              |# and this is not a comment terminator but an escaped symbol|)
        (read-from-string
-        "(foo bar 
+        "(foo bar
  \"#| this is a string, not a comment\"
  |# and this is not a comment terminator but an escaped symbol|)"))
 
@@ -1050,4 +1065,3 @@
        (map uvector-immutable? (load-read-test-uvectors)))
 
 (test-end)
-
