@@ -4,7 +4,8 @@
 ;; has to be done after their tests.
 
 (use gauche.test)
-(use srfi-1)
+(use scheme.list)
+(use file.util)
 (use util.isomorph)
 
 (test-start "advanced read/write features")
@@ -51,7 +52,7 @@
 ;; SRFI-10 Reader constructor (#,)
 ;;
 
-(test-section "srfi-10 reader constructor")
+(test-section "SRFI-10 reader constructor")
 
 (test "read ctor 1a" '(1 2 #f "4 5")
       (lambda ()
@@ -157,6 +158,20 @@
 (test* "circular list involving abbrev syntax" "#0=((quote . #0#))"
        (write-to-string (cdr #0='#0#) write/ss))
 
+
+(let* ([ht1 (make-hash-table 'eq?)]
+       [ht2 (make-hash-table 'eq?)]
+       [circ (list ht1 ht2)])
+  (hash-table-put! ht1 'circ circ)
+  (hash-table-put! ht1 'ht1 ht1)
+  (hash-table-put! ht2 'circ (cdr circ))
+  (hash-table-put! ht2 'ht1 ht1)
+  (test* "circular structure involving hashtable"
+         "#0=(#1=#<hash-table eq[2] @> . #2=(#<hash-table eq[2] @>))"
+         (regexp-replace-all #/@(0x)?[0-9a-f]+/
+                             (write-to-string circ write/ss)
+                             "@")))
+
 (define-class <foo> ()
   ((a :init-keyword :a)
    (b :init-keyword :b)))
@@ -247,6 +262,82 @@
          (set-cdr! a a)
          (format/ss "The answer is ~s ~s" a a)))
 
+;;---------------------------------------------------------------
+(test-section "obscure format features")
+
+(test* "format plural" "0 books, 1 book, 2 books, 1.0 books"
+       (format "~a book~p, ~a book~p, ~a book~p, ~a book~p"
+               0 0 1 1 2 2 1.0 1.0))
+(test* "format plural" "0 pens, 1 pen, 2 pens, 1.0 pens"
+       (format "~a pen~:p, ~a pen~:p, ~a pen~:p, ~a pen~:p"
+               0 1 2 1.0))
+(test* "format plural" "0 ponies, 1 pony, 2 ponies, 1.0 ponies"
+       (format "~a pon~:@p, ~a pon~:@p, ~a pon~:@p, ~a pon~:@p"
+               0 1 2 1.0))
+
+(test* "format roman numerals"
+       '("I" "II" "III" "IV" "V" "VI" "VII" "VIII" "IX" "X")
+       (map (^i (format "~@R" i)) (iota 10 1)))
+(test* "format roman numerals"
+       '("i" "ii" "iii" "iv" "v" "vi" "vii" "viii" "ix" "x")
+       (map (^i (format "~@r" i)) (iota 10 1)))
+(test* "format roman numerals"
+       '("I" "II" "III" "IIII" "V" "VI" "VII" "VIII" "VIIII" "X")
+       (map (^i (format "~:@R" i)) (iota 10 1)))
+(test* "format roman numerals"
+       '("X" "XX" "XXX" "XL" "L" "LX" "LXX" "LXXX" "XC" "C")
+       (map (^i (format "~@R" i)) (iota 10 10 10)))
+(test* "format roman numerals"
+       '("C" "CC" "CCC" "CD" "D" "DC" "DCC" "DCCC" "CM" "M")
+       (map (^i (format "~@R" i)) (iota 10 100 100)))
+
+(let1 data '((39 "XXXIX")
+             (246 "CCXLVI")
+             (789 "DCCLXXXIX")
+             (2421 "MMCDXXI")
+             (160 "CLX")
+             (207 "CCVII")
+             (1009 "MIX")
+             (1066 "MLXVI")
+             (1776 "MDCCLXXVI")
+             (1918 "MCMXVIII")
+             (2014 "MMXIV"))
+  (dolist [d data] (test* "format roman numeral" (cadr d)
+                          (format "~@R" (car d)))))
+
+(test* "format conditionals (case)"
+       '("It is a cat."
+         "It is a Siamese cat."
+         "It is a Manx cat."
+         "It is a Persian cat."
+         "It is a cat.")
+       (map (cut format "It is a ~[Siamese ~;Manx ~;Persian ~]cat." <>)
+            '(-1 0 1 2 3)))
+(test* "format conditionals (case, default)"
+       '("It is a Calico cat."
+         "It is a Siamese cat."
+         "It is a Manx cat."
+         "It is a Persian cat."
+         "It is a Calico cat.")
+       (map (cut format "It is a ~[Siamese ~;Manx ~;Persian ~:;Calico ~]cat." <>)
+            '(-1 0 1 2 3)))
+(test* "format conditionals (case, error)"
+       (test-error <error> #/Argument for ~\[ must/)
+       (format "It is a ~[Siamese ~;Manx ~;Persian ~]cat." 'a))
+
+(test* "format conditionals (alterative)"
+       '("Green Lamp" "Blue \"Lamp\"")
+       (map (cut format "~:[Green ~a~;Blue ~s~]" <> "Lamp") '(#f #t)))
+
+(test* "format conditionals (maybe)"
+       '("Don't worry." "It's OK. Don't worry.")
+       (map (cut format "~@[It's ~a. ~]~a." <> "Don't worry") '(#f "OK")))
+
+;; Ensure `format` does _not_ cache in this case.
+(test* "formatter cache"
+       '("a" "~s")
+       (map (cut format <> "~s" 'a) '(#f "~0@*~a")))
+
 ;;---------------------------------------------------------------
 (test-section "read/ss basic")
 
@@ -297,24 +388,17 @@
        (read-from-string "(#0=\"aa\" #1=\"aa\" #0# #1#)")
        isomorphic?)
 
-;;===============================================================
-;; These test is here since srfi-0 must have been tested before.
-;;
 (test-section "whitespaces")
 
 (test* "skipws" 'a
        (read-from-string
-        (cond-expand
-         [gauche.ces.utf8 "\u00a0\u1680\u2000\u200a\u2028\u2029\
-                           \u202f\u205f\u3000a"]
-         [(or gauche.ces.eucjp gauche.ces.sjis) "\u3000a"]
-         [else "a"])))
+        "\u00a0\u1680\u2000\u200a\u2028\u2029\u202f\u205f\u3000a"))
 
 ;;===============================================================
-;; Interference between srfi-10 and shared structure
+;; Interference between SRFI-10 and shared structure
 ;;
 
-(test-section "combine srfi-10 and srfi-38")
+(test-section "combine SRFI-10 and SRFI-38")
 
 ;; NB: this is an experimental feature.  Do not count on this API!
 (define-reader-ctor 'foo
@@ -417,15 +501,17 @@
   (define (write-to-string/ctx obj . args)
     ($ write-to-string obj
        (^x (write x (apply make-write-controls args)))))
-  
-  (test* "print-length" '("(0 1 2 3 4)"
-                          ("(0 1 2 3 4)"   "#(0 1 2 3 4)"   "#u8(0 1 2 3 4)")
-                          ("(0 1 2 3 ...)" "#(0 1 2 3 ...)" "#u8(0 1 2 3 ...)")
-                          ("(0 1 2 ...)"   "#(0 1 2 ...)"   "#u8(0 1 2 ...)")
-                          ("(0 1 2 3 4)"   "#(0 1 2 3 4)"   "#u8(0 1 2 3 4)")
-                          ("(0 1 ...)"     "#(0 1 ...)"     "#u8(0 1 ...)")
-                          ("(0 ...)"       "#(0 ...)"       "#u8(0 ...)")
-                          ("(...)"         "#(...)"         "#u8(...)"))
+  (define elli (with-module gauche.internal (string-ellipsis)))
+
+  (test* "print-length"
+         `("(0 1 2 3 4)"
+           ("(0 1 2 3 4)"       "#(0 1 2 3 4)"       "#u8(0 1 2 3 4)")
+           (,#"(0 1 2 3 ~elli)" ,#"#(0 1 2 3 ~elli)" ,#"#u8(0 1 2 3 ~elli)")
+           (,#"(0 1 2 ~elli)"   ,#"#(0 1 2 ~elli)"   ,#"#u8(0 1 2 ~elli)")
+           ("(0 1 2 3 4)"       "#(0 1 2 3 4)"       "#u8(0 1 2 3 4)")
+           (,#"(0 1 ~elli)"     ,#"#(0 1 ~elli)"     ,#"#u8(0 1 ~elli)")
+           (,#"(0 ~elli)"       ,#"#(0 ~elli)"       ,#"#u8(0 ~elli)")
+           (,#"(~elli)"         ,#"#(~elli)"         ,#"#u8(~elli)"))
          (let ([z (map (^n (list (write-to-string/ctx data :length n)
                                  (write-to-string/ctx (list->vector data)
                                                       :length n)
@@ -440,16 +526,16 @@
          '("()" "#()" "#u8()")
          (map (^x (write-to-string/ctx x :length 0))
               '(() #() #u8())))
-  
+
   (test* "print-length (nested)"
-         '(("(...)"
-            "#(...)")
-           ("((0 ...) ...)"
-            "#(#(0 ...) ...)")
-           ("((0 1 ...) (0 1 ...) ...)"
-            "#(#(0 1 ...) #(0 1 ...) ...)")
-           ("((0 1 2 ...) (0 1 2 ...) (0 1 2 ...) ...)"
-            "#(#(0 1 2 ...) #(0 1 2 ...) #(0 1 2 ...) ...)"))
+         `((,#"(~elli)"
+            ,#"#(~elli)")
+           (,#"((0 ~elli) ~elli)"
+            ,#"#(#(0 ~elli) ~elli)")
+           (,#"((0 1 ~elli) (0 1 ~elli) ~elli)"
+            ,#"#(#(0 1 ~elli) #(0 1 ~elli) ~elli)")
+           (,#"((0 1 2 ~elli) (0 1 2 ~elli) (0 1 2 ~elli) ~elli)"
+            ,#"#(#(0 1 2 ~elli) #(0 1 2 ~elli) #(0 1 2 ~elli) ~elli)"))
          (map (^n (list (write-to-string/ctx data2 :length n)
                         (write-to-string/ctx
                          (list->vector (map list->vector data2))
@@ -471,7 +557,7 @@
                                                    (make-write-controls
                                                     :level n)))))
               (iota 8))))
-         
+
 (let* ([data '(a (b (c (d (e) (f) g) h) i) #(j (k #(l #(m) (n) o) p) q) r)])
   (test* "print-level"
          '("(a (b (c (d (e) (f) g) h) i) #(j (k #(l #(m) (n) o) p) q) r)"
@@ -487,21 +573,22 @@
               '(6 5 4 3 2 1 0))))
 
 ;; another example from CLHS
-(let* ([level-length '((0 1) (1 1) (1 2) (1 3) (1 4) 
+(let* ([level-length '((0 1) (1 1) (1 2) (1 3) (1 4)
                        (2 1) (2 2) (2 3) (3 2) (3 3) (3 4))]
        [data '(if (member x y) (+ (car x) 3) '(foo . #(a b c d "Baz")))])
+  (define elli (with-module gauche.internal (string-ellipsis)))
   (test* "print-level & print-length"
-         '("0 1 -- #"
-           "1 1 -- (if ...)"
-           "1 2 -- (if # ...)"
-           "1 3 -- (if # # ...)"
-           "1 4 -- (if # # #)"
-           "2 1 -- (if ...)"
-           "2 2 -- (if (member x ...) ...)"
-           "2 3 -- (if (member x y) (+ # 3) ...)"
-           "3 2 -- (if (member x ...) ...)"
-           "3 3 -- (if (member x y) (+ (car x) 3) ...)"
-           "3 4 -- (if (member x y) (+ (car x) 3) '(foo . #(a b c d ...)))")
+         `(,#"0 1 -- #"
+           ,#"1 1 -- (if ~elli)"
+           ,#"1 2 -- (if # ~elli)"
+           ,#"1 3 -- (if # # ~elli)"
+           ,#"1 4 -- (if # # #)"
+           ,#"2 1 -- (if ~elli)"
+           ,#"2 2 -- (if (member x ~elli) ~elli)"
+           ,#"2 3 -- (if (member x y) (+ # 3) ~elli)"
+           ,#"3 2 -- (if (member x ~elli) ~elli)"
+           ,#"3 3 -- (if (member x y) (+ (car x) 3) ~elli)"
+           ,#"3 4 -- (if (member x y) (+ (car x) 3) '(foo . #(a b c d ~elli)))")
          (map (^z (let1 c (make-write-controls
                            :level (car z) :length (cadr z))
                     (format c "~d ~d -- ~s" (car z) (cadr z) data)))
@@ -584,36 +671,28 @@
 
 (use gauche.charconv)
 
-(cond-expand
- [gauche.ces.none]
- [else
-  (let ()
-    ;; This test checks peek-byte won't disturb subsequent read-char
-    ;; even the input contains illegal byte sequence.
-    ;; related: https://github.com/shirok/Gauche/pull/177
-    (define (read1 p) (read-char p))
-    (define (read2 p) (let1 c (read-char p) `(,c ,(read-char p))))
-    (define (peek1 p) (peek-char p) (read1 p))
-    (define (peek2 p) (peek-char p) (read2 p))
-    (define (fetch s reader ces)
-      (call-with-input-string (ces-convert s #f ces) reader))
-    
-    (define (t ces)
-      (test* #"peek-char - read-char (~ces)"
-             (fetch "\u3042" read1 ces) (fetch "\u3042" peek1 ces))
-      (test* #"peek-char - read-char^2 (~ces) case 1"
-             (fetch "\u3042" read2 ces) (fetch "\u3042" peek2 ces))
-      (test* #"peek-char - read-char^2 (~ces) case 2"
-             (fetch "\u3042\u3044" read2 ces)
-             (fetch "\u3042\u3044" peek2 ces)))
+(let ()
+  ;; This test checks peek-byte won't disturb subsequent read-char
+  ;; even the input contains illegal byte sequence.
+  ;; related: https://github.com/shirok/Gauche/pull/177
+  (define (read1 p) (read-char p))
+  (define (read2 p) (let1 c (read-char p) `(,c ,(read-char p))))
+  (define (peek1 p) (peek-char p) (read1 p))
+  (define (peek2 p) (peek-char p) (read2 p))
+  (define (fetch s reader ces)
+    (call-with-input-string (ces-convert s #f ces) reader))
 
-    ;; NB: If internal encoding is sjis or eucjp, we skip utf8 tests
-    ;; for they raise "encountered EOF in the middle of multibyte character"
-    ;; error.
-    (cond-expand
-     [gauche.ces.utf8 (for-each t '(utf8 sjis eucjp))]
-     [(or gauche.ces.sjis gauche.ces.eucjp) (for-each t '(sjis eucjp))]
-     [else]))])
+  (define (t ces)
+    (test* #"peek-char - read-char (~ces)"
+           (fetch "\u3042" read1 ces) (fetch "\u3042" peek1 ces))
+    (test* #"peek-char - read-char^2 (~ces) case 1"
+           (fetch "\u3042" read2 ces) (fetch "\u3042" peek2 ces))
+    (test* #"peek-char - read-char^2 (~ces) case 2"
+           (fetch "\u3042\u3044" read2 ces)
+           (fetch "\u3042\u3044" peek2 ces)))
+
+  (for-each t '(utf8 sjis eucjp))
+  )
 
 ;;===============================================================
 ;; Pretty printer
@@ -630,6 +709,7 @@
   (define (t name expect data . args)
     (test* #"~|name| ~|args|" expect
            (with-output-to-string (^[] (apply pprint data args)))))
+  (define elli (with-module gauche.internal (string-ellipsis)))
   (let-syntax
       ([t* (syntax-rules ()
              [(_ (data . args) expect)
@@ -654,13 +734,13 @@
        \n eiusmod tempor incididunt ut labore\
        \n et dolore)\n")
     (t* (data1 :length 5)
-        "(Lorem ipsum dolor sit amet ....)\n")
+        #"(Lorem ipsum dolor sit amet ~elli)\n")
     (t* (data1 :length 1)
-        "(Lorem ....)\n")
+        #"(Lorem ~elli)\n")
     (t* (data1 :length 0)
-        "(....)\n")
+        #"(~elli)\n")
     (t* (data1 :level 1 :length 5)
-        "(Lorem ipsum dolor sit amet ....)\n")
+        #"(Lorem ipsum dolor sit amet ~elli)\n")
     (t* (data1 :level 0 :length 5)
         "#\n")
     (t* ('a :level 0)
@@ -684,7 +764,7 @@
     ))
 
 (test* "no newline" "(a\n a)"
-       (call-with-output-string 
+       (call-with-output-string
          (cut pprint '(a a) :width 3 :newline #f :port <>)))
 
 (test* "label (issue #484)" "#0=(a #1=(b #0# #1#))\n"
@@ -715,6 +795,12 @@
        (call-with-output-string
          (cut pprint ''(,@(the quick brown fox jumps over the ,@lazy dog))
               :port <> :width 25)))
+(test* "shorthand notation" (test-none-of)
+       (call-with-output-string
+         (cut pprint `(,((with-module gauche.internal make-identifier)
+                         'quote (current-module) '())
+                       foo)
+              :port <>)))
 
 ;;===============================================================
 ;; utf-8 with BOM
@@ -722,7 +808,7 @@
 
 (test* "utf-8 with BOM" "foo!"
        (begin
-         (load "data/utf-8-bom.scm" :paths `(,(sys-dirname (current-load-path))))
+         (load "data/utf-8-bom.scm" :paths `(,(build-path 'cld)))
          (foo)))
 
 (test-end)
