@@ -71,6 +71,21 @@
   (test-length '(a . b) 1)
   (test-length '(a b c d e) 5))
 
+(test* "num-pairs (proper)" 5
+       (num-pairs '(a b c d e)))
+(test* "num-pairs (proper)" 0
+       (num-pairs '()))
+(test* "num-pairs (dotted)" 4
+       (num-pairs '(a b c d . e)))
+(test* "num-pairs (dotted)" 0
+       (num-pairs 'a))
+(test* "num-pairs (circular)" 1
+       (num-pairs '#0=(a . #0#)))
+(test* "num-pairs (circular)" 5
+       (num-pairs '#0=(a b c d e . #0#)))
+(test* "num-pairs (circular)" 5
+       (num-pairs '(a b . #0=(c d e . #0#))))
+
 ;;--------------------------------------------------------------------------
 (test-section "copy, append and reverse")
 
@@ -443,6 +458,17 @@
               [y (alist-copy x)])
          (set-cdr! (assq 'a y) (list 2))
          (list (assq 'a x) (assq 'a y))))
+(test* "alist-copy (circular)" (test-error)
+       (alist-copy '((a . 1) . #0=((b . 2) (c . 3) . #0#))))
+(test* "alist-copy (dotted)" (test-error)
+       (alist-copy '((a . 1) (b . 2) . c)))
+(test* "alist-copy (non-pair)" '(((a . 1) b (c . 2))
+                                 ((a . 1) b (c . 3)))
+       (let* ([orig '((a . 1) b (c . 2))]
+              [copy (alist-copy orig)])
+         (set-cdr! (assq 'c copy) 3)
+         (list orig copy)))
+
 (test* "alist-delete" '((a 1) (b 2) (c 3))
        (alist-delete 'x '((a 1) (b 2) (c 3))))
 (test* "alist-delete" '((b 2) (c 3))
@@ -522,6 +548,23 @@
 (test* "assoc-adjoin" '((a . 1))
        (assoc-adjoin '() 'a 1))
 
+(test* "alist-merge" '()
+       (alist-merge + '()))
+(test* "alist-merge" '((a . 1) (b . 2))
+       (alist-merge + '((a . 1) (b . 2))))
+(test* "alist-merge" '((a . 1) (b . 2) (c . 3) (d . 4))
+       (alist-merge + '((a . 1) (b . 2)) '((c . 3) (d . 4))))
+(test* "alist-merge" '((a . 5) (b . 2) (d . 3))
+       (alist-merge + '((a . 1) (b . 2)) '((d . 3) (a . 4))))
+(test* "alist-merge" '((a . 1) (b . 6) (c . 3))
+       (alist-merge + '((a . 1) (b . 2)) '((c . 3) (b . 4))))
+
+(test* "alist-merge" '((a 1 5) (b 2 4) (c 3))
+       (alist-merge append '((a 1) (b 2)) '((c 3) (b 4)) '() '((a 5))))
+(test* "alist-merge" '(("a" 1 4) ("b" 2 3))
+       (alist-merge string=? append '(("a" 1) ("b" 2)) '(("b" 3) ("a" 4))))
+
+
 (test* "assoc-update-in" '((a (aa . 1) (ab . 2))
                            (b (ba . 3) (bb (bba . 104) (bbb . 5))))
        (assoc-update-in '((a (aa . 1) (ab . 2))
@@ -680,21 +723,23 @@
 ;;--------------------------------------------------------------------------
 (test-section "extended pairs")
 
-(let ([extended-cons  (with-module gauche.internal extended-cons)]
-      [extended-pair? (with-module gauche.internal extended-pair?)]
-      [pair-attributes (with-module gauche.internal pair-attributes)]
-      [pair-attribute-get (with-module gauche.internal pair-attribute-get)]
-      [pair-attribute-set! (with-module gauche.internal pair-attribute-set!)])
+(let ()
   (test* "extended pair" #f (extended-pair? (cons 1 2)))
   ;(test* "extended pair" #t (extended-pair? '(1 2)))
   (test* "extended pair" #t (extended-pair? (extended-cons 1 2)))
   (test* "pair-attributes" '() (pair-attributes (cons 1 2)))
   (test* "pair-attributes" '() (pair-attributes (ipair 1 2)))
   (test* "pair-attributes" '() (pair-attributes (extended-cons 1 2)))
+  (test* "pair-attributes" '((a . b) (c d))
+         (pair-attributes (extended-cons 1 2 '((a . b) (c d)))))
   (test* "pair-attributes-get/set!" 'hoi
          (let1 z (extended-cons 1 2)
            (pair-attribute-set! z 'foo 'hoi)
            (pair-attribute-get z 'foo)))
+  (test* "pair-attributes-get/set!" 'poi
+         (let1 z (extended-cons 1 2 '((foo . hoi)))
+           (pair-attribute-set! z 'foo 'poi)
+           (pair-attribute-get z 'foo)))
   (test* "set-car! to extended pair" '(3 . 2)
          (rlet1 z (extended-cons 1 2)
            (set-car! z 3)))
@@ -703,4 +748,28 @@
            (set-cdr! z 4)))
   )
 
+(let* ((data1 '((a . b) (c . d)))
+       (data2 (list (cons 'A 'B) (cons 'C 'D)))
+       (p1 (extended-cons 1 2 data1))
+       (p2 (extended-cons 1 2 data2)))
+  ;; extended-pair-set! shouldn't clobber the original list
+  (test* "pair-attribute-set! noclobber 1" '((e . z) (a . b) (c . y))
+         (begin
+           (pair-attribute-set! p1 'e 'z)
+           (pair-attribute-set! p1 'c 'y)
+           (pair-attributes p1)))
+  (test* "pair-attribute-set! noclobber 2" '((E . Z) (A . Y) (C . D))
+         (begin
+           (pair-attribute-set! p2 'A 'Y)
+           (pair-attribute-set! p2 'E 'Z)
+           (pair-attributes p2)))
+  (test* "pair-attribute-set! noclobber" #t
+         (equal? data1 '((a . b) (c . d))))
+  (test* "pair-attribute-set! noclobber" #t
+         (equal? data2 '((A . B) (C . D))))
+  ;; prevent optimization
+  (set! data1 #f)
+  (set! data2 #f)
+  )
+
 (test-end)
