@@ -272,6 +272,7 @@
 (define-module Oc
   (export ichi (rename shi yon) go)
   (define ichi 1)
+  (define ni 2) ; not exported
   (define shi 4)
   (define go (+ ichi shi))
   (define kyu (+ shi go))
@@ -282,12 +283,31 @@
 (test "export-time renaming" '(1 4 5 9 #f #f)
       (lambda ()
         (let [(m (find-module 'Oc-1))]
-          (list (global-variable-ref m 'ichi #f)
-                (global-variable-ref m 'yon #f)
-                (global-variable-ref m 'go #f)
-                (global-variable-ref m 'ku #f)
-                (global-variable-ref m 'shi #f)
-                (global-variable-ref m 'kyu #f)))))
+          (list (module-binding-ref m 'ichi #f)
+                (module-binding-ref m 'yon #f)
+                (module-binding-ref m 'go #f)
+                (module-binding-ref m 'ku #f)
+                (module-binding-ref m 'shi #f)
+                (module-binding-ref m 'kyu #f)))))
+
+(test "export-time renaming and defines?/exports? (0c)"
+      '((ichi #t #t) (ni #t #f) (shi #t #f) (yon #f #t) (go #t #t) (kyu #t #f) (ku #f #t))
+      (lambda ()
+        (let [(m (find-module 'Oc))]
+          (map (lambda (n)
+                 (list n
+                       (module-binds? m n)
+                       (module-exports? m n)))
+               '(ichi ni shi yon go kyu ku)))))
+(test "export-time renaming and defines?/exports? (0c-1)"
+      '((ichi #t #t) (ni #f #f) (shi #f #f) (yon #t #t) (go #t #t) (kyu #f #f) (ku #t #t))
+      (lambda ()
+        (let [(m (find-module 'Oc-1))]
+          (map (lambda (n)
+                 (list n
+                       (module-binds? m n)
+                       (module-exports? m n)))
+               '(ichi ni shi yon go kyu ku)))))
 
 (define-module Od
   (import (Oc :rename ((ku kokono)) :prefix mm:)))
@@ -295,8 +315,8 @@
 (test "export-time renaming plus import renaming" '(#f 9)
       (lambda ()
         (let [(m (find-module 'Od))]
-          (list (global-variable-ref m 'mm:ku #f)
-                (global-variable-ref m 'mm:kokono #f)))))
+          (list (module-binding-ref m 'mm:ku #f)
+                (module-binding-ref m 'mm:kokono #f)))))
 
 (define-module Oe
   (export (rename a x) (rename x a))
@@ -308,8 +328,8 @@
 (test "export-time renaming with swapping" '(X A)
       (lambda ()
         (let [(m (find-module 'Oe-1))]
-          (list (global-variable-ref m 'a #f)
-                (global-variable-ref m 'x #f)))))
+          (list (module-binding-ref m 'a #f)
+                (module-binding-ref m 'x #f)))))
 
 (define-module Of
   (export foo)
@@ -320,7 +340,7 @@
   (export foo))
 
 (test "transitive export" 1
-      (lambda () (global-variable-ref (find-module 'Of-1) 'foo #f)))
+      (lambda () (module-binding-ref (find-module 'Of-1) 'foo #f)))
 
 (define-module Of-2
   (import Of)
@@ -331,8 +351,8 @@
 (test "export-time renaming and transitive export" '(1 #f)
       (lambda ()
         (list
-         (global-variable-ref (find-module 'Of-3) 'foo-alias #f)
-         (global-variable-ref (find-module 'Of-3) 'foo #f))))
+         (module-binding-ref 'Of-3 'foo-alias #f)
+         (module-binding-ref 'Of-3 'foo #f))))
 
 (define-module Of-4
   (import (Of-2 :rename ((foo-alias foo-newname)))))
@@ -341,9 +361,9 @@
       '(1 #f #f)
       (lambda ()
         (list
-         (global-variable-ref (find-module 'Of-4) 'foo-newname #f)
-         (global-variable-ref (find-module 'Of-4) 'foo-alias #f)
-         (global-variable-ref (find-module 'Of-4) 'foo #f))))
+         (module-binding-ref 'Of-4 'foo-newname #f)
+         (module-binding-ref 'Of-4 'foo-alias #f)
+         (module-binding-ref 'Of-4 'foo #f))))
 
 (define-module Of-5
   (extend Of)
@@ -358,8 +378,8 @@
       '((#f 1) (#f #f) (1 #f) (2 #f) (#f 2) (#f #f))
       (lambda ()
         (map (lambda (s)
-               (list (global-variable-ref (find-module 'Of-6) s #f)
-                     (global-variable-ref (find-module 'Of-7) s #f)))
+               (list (module-binding-ref 'Of-6 s #f)
+                     (module-binding-ref 'Of-7 s #f)))
              '(who foo-alias foo bah bar-alias bar))))
 
 ;;------------------------------------------------------------------
@@ -433,21 +453,21 @@
       (lambda ()
         (eval '(with-module V (extend Q MM)) (interaction-environment))))
 
-(test "global-variable-ref" 'gamma
+(test "module-binding-ref" 'gamma
       (lambda ()
-        (global-variable-ref 'U 'c)))
+        (module-binding-ref 'U 'c)))
 
-(test "global-variable-ref" (test-error)
+(test "module-binding-ref" (test-error)
       (lambda ()
-        (global-variable-ref 'U 'e)))
+        (module-binding-ref 'U 'e)))
 
-(test "global-variable-ref" 'huh?
+(test "module-binding-ref" 'huh?
       (lambda ()
-        (global-variable-ref 'U 'e 'huh?)))
+        (module-binding-ref 'U 'e 'huh?)))
 
-(test "global-variable-ref" 'huh?
+(test "module-binding-ref" 'huh?
       (lambda ()
-        (global-variable-ref 'U 'c 'huh? #t)))
+        (module-binding-ref 'U 'c 'huh? #t)))
 
 ;;------------------------------------------------------------------
 ;; mpl search bug in 0.9.1 reported by Ryo Akagi
@@ -464,7 +484,7 @@
 
 (define (mplbug-test mod var)
   (test* #"mpl search (~mod,~var)" #t
-         (global-variable-bound? (find-module mod) var)))
+         (module-binds? (find-module mod) var)))
 
 (mplbug-test 'mplbug-user1 'a)
 (mplbug-test 'mplbug-user1 'b)
@@ -503,12 +523,12 @@
   (import (mpl-prefix-b :prefix X:))
   (import (mpl-prefix-b :prefix Y:))
   (import mpl-prefix-b))
-  
+
 
 (let ()
   (define (check mod . syms)
     (map (lambda (sym)
-           (list sym (global-variable-ref (find-module mod) sym 'undef)))
+           (list sym (module-binding-ref mod sym 'undef)))
          syms))
   (test* "mpl prefix simple" '((a undef) (b undef) (X:a 0) (X:b 1))
          (check 'mpl-prefix-1 'a 'b 'X:a 'X:b))
@@ -528,7 +548,7 @@
   (test* "mpl prefix simple dupe (parent prefix)"
          '((a 0) (b 1) (Y:a 0) (Y:b undef))
          (check 'mpl-prefix-5 'a 'b 'Y:a 'Y:b))
-  
+
   (test* "mpl prefix simple merge (different prefix)"
          '((a undef) (b undef)   (c undef)
            (X:a 0)   (X:b 1)     (X:c undef)
@@ -540,7 +560,7 @@
            (X:a 0)   (X:b 1)     (X:c undef)
            (Y:a 0)   (Y:b undef) (Y:c 2))
          (check 'mpl-prefix-7 'a 'b 'c 'X:a 'X:b 'X:c 'Y:a 'Y:b 'Y:c))
-  
+
   (test* "mpl prefix simple dupe (different prefix)"
          '((a   0)   (b 1)
            (X:a 0)   (X:b 1)
@@ -625,7 +645,7 @@
         (let ((m0 (make-module #f)))
           (eval '(define x 13) m0)
           (eval 'x m0))))
-              
+
 (test "anonymous module" (test-error)
       (lambda ()
         (let ((m0 (make-module #f))
@@ -644,8 +664,7 @@
                             set! syntax-rules else =>))
                    (missing '()))
           (cond [(null? names) missing]
-                [(global-variable-bound? 'null (car names))
-                 (loop (cdr names) missing)]
+                [(module-binds? 'null (car names)) (loop (cdr names) missing)]
                 [else (loop (cdr names) (cons (car names) missing))]))))
 
 (test "scheme module bindings" '()
@@ -694,10 +713,55 @@
                             write write-char zero?))
                    (missing '()))
           (cond [(null? names) missing]
-                [(global-variable-bound? 'scheme (car names))
-                 (loop (cdr names) missing)]
+                [(module-binds? 'scheme (car names)) (loop (cdr names) missing)]
                 [else (loop (cdr names) (cons (car names) missing))]))))
 
+;;-------------------------------------------------------------------
+;; Module exports
+
+(define-module exports-test-A
+  (export a b c)
+  (define a 0)
+  (define b 1)
+  (define c 2))
+
+(define-module exports-test-B
+  (extend exports-test-A)
+  (define a 3)
+  (define d 4)
+  (export d))
+
+(define-module exports-test-C
+  (extend exports-test-A)
+  (export-all)
+  (define e 5)
+  (define f 6))
+
+(define-module exports-test-D
+  (extend exports-test-B exports-test-C)
+  (export g)
+  (define g 7))
+
+(test "module-exports"
+      '(a b c d e f g)
+      (lambda ()
+        ;; NB: We haven't tested 'sort', so we need to roll our own
+        (define (sym<? a b)
+          (string<? (symbol->string a) (symbol->string b)))
+        (define (sort-sym xs)
+          (if (or (null? xs) (null? (cdr xs)))
+            xs
+            (let loop ((pivot (car xs))
+                       (xs (cdr xs))
+                       (lt '())
+                       (ge '()))
+              (if (null? xs)
+                (append (sort-sym lt) (list pivot) (sort-sym ge))
+                (if (sym<? (car xs) pivot)
+                  (loop pivot (cdr xs) (cons (car xs) lt) ge)
+                  (loop pivot (cdr xs) lt (cons (car xs) ge)))))))
+        (sort-sym (module-exports (find-module 'exports-test-D)))))
+
 ;;-------------------------------------------------------------------
 ;; Macro and builtin inliner
 ;; https://twitter.com/tk_riple/status/647865265154326528
