@@ -304,6 +304,32 @@
        (list (method-leaf? (car (slot-ref nm-1 'methods)))
              (method-leaf? (car (slot-ref nm-2 'methods)))))
 
+;;----------------------------------------------------------------
+(test-section "immutable slots")
+
+(define-class <immutable-slot-class> ()
+  ((a :init-keyword :a :immutable #t)
+   (b :init-thunk (^[] *immutable-b*) :immutable #t)))
+(define *immutable-b* 5)
+
+(test* "immutable, initialization" 3
+       (let1 v (make <immutable-slot-class> :a 3)
+         (slot-ref v 'a)))
+
+(test* "immutable, reject" (test-error)
+       (let1 v (make <immutable-slot-class> :a 3)
+         (slot-set! v 'a 5)
+         (slot-ref v 'a)))
+
+(test* "immutable, delayed initialization" 3
+       (let1 v (make <immutable-slot-class>)
+         (slot-set! v 'a 3)
+         (slot-ref v 'a)))
+
+(test* "immutable, initialization by a thunk" 5
+       (let1 v (make <immutable-slot-class>)
+         (slot-ref v 'b)))
+
 ;;----------------------------------------------------------------
 (test-section "method sorting")
 
@@ -469,8 +495,8 @@
 ;; Another tricky interference between modules and implicit setter.
 ;; The :accessor slot option may create the setter generic function
 ;; with the weird name |setter of <getter-name>|.  The name isn't supposed
-;; to be visible from the user ever.  However, until 0.9.6, when the 
-;; MB.base below does not export the setter name, the derived class 
+;; to be visible from the user ever.  However, until 0.9.6, when the
+;; MB.base below does not export the setter name, the derived class
 ;; couldn't find the setter gf and created a separate setter gf, making it
 ;; incompatible with base class and derived class.
 (define-module MB.base
@@ -751,7 +777,7 @@
        (equal? (hash (make <cmp> :x (cons 1 2)))
                (hash (make <cmp> :x (cons 2 1)))))
 
-(use srfi-1)
+(use scheme.list)
 
 (define xht (make-hash-table 'equal?))
 
@@ -824,7 +850,7 @@
 ;; Allowing to use user-defined objects in a hashtable with
 ;; a comparator with default-hash.
 ;; We use eq-/eqv-hash regardless of the compoarator's hash setting
-;; If the equality predicate is eq?/eqv?.  See srfi-125.
+;; If the equality predicate is eq?/eqv?.  See SRFI-125.
 ;; Also https://github.com/shirok/Gauche/issues/708
 (define-class <user-defined-hashed> ()
   ((a :init-keyword :a)))
@@ -1040,6 +1066,16 @@
        '(A b C x Y z)
        (map (cut slot-ref *docu-sub* <>) '(a b c x y z)))
 
+
+;; https://github.com/shirok/Gauche/issues/845
+(define-class <issue845-meta> (<class>) ())
+(define-class <issue845> () () :metaclass <issue845-meta>)
+(define issue845 (make <issue845>))
+(define-class <issue845-meta> (<class>) ())
+(define-class <issue845> () () :metaclass <issue845-meta>)
+(test* "refefinition of metaclass" #t
+       (is-a? issue845 <issue845>))
+
 ;;----------------------------------------------------------------
 (test-section "metaclass/bound-slot")
 
@@ -1056,6 +1092,52 @@
 (test* "<bound-slot-A> check" #t
        (is-a? (make <bound-slot-A> :a 3) <bound-slot-A>))
 
+;;----------------------------------------------------------------
+(test-section "metaclass/typed-slot")
+
+(use gauche.mop.typed-slot)
+(test-module 'gauche.mop.typed-slot)
+
+(define-class <typed-slot-A> (<typed-slot-mixin>)
+  ((a :init-keyword :a
+      :type <integer>)
+   (b :init-keyword :b
+      :type (</> <string> <symbol>))
+   (c :init-keyword :c
+      :type (<?> <string>))
+   ))
+
+(test* "<typed-slot-A> ok" '(1 a #f)
+       (let1 z (make <typed-slot-A> :a 1 :b 'a :c #f)
+         (map (cut ~ z <>) '(a b c))))
+
+(test* "<typed-slot-A> set! check" '(2 "b" "c")
+       (let1 z (make <typed-slot-A>)
+         (set! (~ z 'a) 2)
+         (set! (~ z 'b) "b")
+         (set! (~ z 'c) "c")
+         (map (cut ~ z <>) '(a b c))))
+
+(test* "<typed-slot-A> init error a"
+       (test-error <error> #/Slot a of <typed-slot-A> must be of type/)
+       (make <typed-slot-A> :a 3.4))
+(test* "<typed-slot-A> init error b"
+       (test-error <error> #/Slot b of <typed-slot-A> must be of type/)
+       (make <typed-slot-A> :b 3))
+(test* "<typed-slot-A> init error c"
+       (test-error <error> #/Slot c of <typed-slot-A> must be of type/)
+       (make <typed-slot-A> :c 'c))
+
+(test* "<typed-slot-A> set! error a"
+       (test-error <error> #/Slot a of <typed-slot-A> must be of type/)
+       (set! (~ (make <typed-slot-A>) 'a) #f))
+(test* "<typed-slot-A> set! error b"
+       (test-error <error> #/Slot b of <typed-slot-A> must be of type/)
+       (set! (~ (make <typed-slot-A>) 'b) #f))
+(test* "<typed-slot-A> set! error c"
+       (test-error <error> #/Slot c of <typed-slot-A> must be of type/)
+       (set! (~ (make <typed-slot-A>) 'c) 5))
+
 ;;----------------------------------------------------------------
 (test-section "metaclass/singleton")
 
@@ -1178,7 +1260,7 @@
 ;;----------------------------------------------------------------
 (test-section "metaclass/instance-pool")
 
-(use srfi-1)
+(use scheme.list)
 (use gauche.mop.instance-pool)
 
 (define-class <pool-meta> (<instance-pool-meta>)
@@ -1456,7 +1538,7 @@
   ;; This triggers redefinition of <foo> AND <foobar>.
   (eval '(define-class <foo> () ()) (current-module))
 
-  ;; Make sure reference to the old <foobar> isn't 
+  ;; Make sure reference to the old <foobar> isn't
   (test* "multiple inheriance redefinition" '(#t #f)
          (list (boolean (memq <foobar> (~ <bar>'direct-subclasses)))
                (boolean (memq <foobar>-save (~ <bar>'direct-subclasses)))))
@@ -1653,5 +1735,5 @@
 (import (object-hygiene-2 :prefix oh2:))
 (test* "hygienic keyword match" oh2:<foo-meta>
        (class-of oh2:<foo>))
-  
+
 (test-end)
