@@ -15,6 +15,7 @@
    (("help" => help)
     ("a"        ()          (format #t "a,"))
     ("bb=s"     (arg)       (format #t "bb=~a," arg))
+    ("b=s"      (arg)       (format #t "b=~a," arg))
     ("cc=ss"    (arg1 arg2) (format #t "cc=(~a ~a)," arg1 arg2))
     ("ddd=sss"  args        (format #t "ddd=~a," args))
     ("eee=i"    (arg)       (format #t "eee=~s," arg))
@@ -85,6 +86,11 @@
 (test "-k"   '("k=(1 2 3),a,") (tester "-k" "1" "2" "3" "-a"))
 (test "-k"   '("k=(#f #f #f),a,") (tester "-k" "-a"))
 
+(test "-b 1"   '("b=abc,") (tester "-b" "abc"))
+(test "-b 2"   '("b=abc,") (tester "-b=abc"))
+(test "-b 3"   '("b=abc,") (tester "-babc"))
+(test "-b (long option match)"   '("bb=abc,") (tester "-bb" "abc"))
+
 (test "--" '("bb=x," "-a" "-cc") (tester "-bb" "x" "--" "-a" "-cc"))
 (test "--" '("" "-bb" "x" "--" "-a" "-cc") (tester "--" "-bb" "x" "--" "-a" "-cc"))
 
@@ -233,4 +239,90 @@
                  (else _ (ret x)))
               11)))))
 
+(let ()
+  (define (run args)
+    (let-args args
+        ((a "a*")
+         (b "b*=s"))
+      (list a b)))
+
+  (test* "plural args" '(() ())
+         (run '()))
+  (test* "plural args" '((#t) ("foo"))
+         (run '("-a" "-b" "foo")))
+  (test* "plural args" '((#t #t) ("foo" "bar"))
+         (run '("-a" "-b" "foo" "-a" "-b" "bar")))
+  )
+
+(let ()
+  (define (run args help-string-args)
+    (let/cc ret
+      (let-args args
+          ((a "a" ? "An option with no arg")
+           (b "b=s{PARAM_B}" ? "Taking string argument {PARAM_B}.")
+           (c "c=s{PARAM_C}" "wow" ? "Taking string argument and default. \
+                             And this is a long help string that will span \
+                             more than one line.")
+           (d "d" => (^[] 'yo) ? "Callback, no default")
+           (e "e" 'boo => (^[] 'yay) ? "Callback, default")
+           (f "f|floccinaucinihilipilification=s" ? "Long option name")
+           (g "g=i{PARAM_G1}i{PARAM_G2}" => (^[g1 g2] (list g1 g2))
+              ? "Taking two integer arguments, {PARAM_G1} and {PARAM_G2}.")
+           (h "h")                      ;no help string
+           (else => (^[badopt rest loop]
+                      (ret (string-append "Unknown option: " badopt "\n"
+                                          "Options:\n"
+                                          (apply option-parser-help-string
+                                                 help-string-args))))))
+        (list a b c d e f g))))
+
+  (define (t args expect)
+    (test* #"help string ~|args|" expect (run args '())))
+
+  ;; Just to make sure help string doesn't affect normal operations.
+  (t '()                     '(#f #f "wow" #f boo #f #f))
+  (t '("-a")                 '(#t #f "wow" #f boo #f #f))
+  (t '("-b" "bee")           '(#f "bee" "wow" #f boo #f #f))
+  (t '("-c" "sea")           '(#f #f "sea" #f boo #f #f))
+  (t '("-d")                 '(#f #f "wow" yo boo #f #f))
+  (t '("-e")                 '(#f #f "wow" #f yay #f #f))
+  (t '("-f" "foo")           '(#f #f "wow" #f boo "foo" #f))
+
+  ;; And now, help string generation
+  ($ test*/diff "help string generation 1"
+     '("Unknown option: zzz"
+       "Options:"
+       "  -a           An option with no arg"
+       "  -b PARAM_B   Taking string argument PARAM_B."
+       "  -c PARAM_C   Taking string argument and default. And this is a long help"
+       "               string that will span more than one line."
+       "  -d           Callback, no default"
+       "  -e           Callback, default"
+       "  -f, --floccinaucinihilipilification s"
+       "               Long option name"
+       "  -g PARAM_G1 PARAM_G2"
+       "               Taking two integer arguments, PARAM_G1 and PARAM_G2."
+       "  -h           (No help available)")
+     (run '("--zzz") '()))
+
+  ($ test*/diff "help string generation 2"
+     '("Unknown option: zzz"
+       "Options:"
+       "    -a                   An option with no arg"
+       "    -b PARAM_B           Taking string argument PARAM_B."
+       "    -c PARAM_C           Taking string argument and default."
+       "                         And this is a long help string that"
+       "                         will span more than one line."
+       "    -d                   Callback, no default"
+       "    -e                   Callback, default"
+       "    -f, --floccinaucinihilipilification s"
+       "                         Long option name"
+       "    -g PARAM_G1 PARAM_G2 Taking two integer arguments,"
+       "                         PARAM_G1 and PARAM_G2.")
+     (run '("--zzz") '(:omit-options-without-help #t
+                       :option-indent 4
+                       :description-indent 25
+                       :width 60)))
+  )
+
 (test-end)
