@@ -3,8 +3,8 @@
 ;;
 
 (use gauche.test)
-(use srfi-1)
-(use srfi-14)
+(use scheme.list)
+(use srfi.14)
 
 (test-start "regexp")
 
@@ -102,7 +102,7 @@
             "(?>abc)" "(?=a*b*c)" "(?!a*b*c)" "(?<=a[bc])" "(?<!a[bc])"
             "(a)(?(1)b)" "(a)(?(1)b|c)" "(a)(?(1)|c)"
             "(?(?=ab)cd|ef)" "(?(?!ab)cd|ef)"
-            "(?(?<=ab)cd|ef)" "(?(?<!ab)cd|ef)" 
+            "(?(?<=ab)cd|ef)" "(?(?<!ab)cd|ef)"
             ))
 
 ;;-------------------------------------------------------------------------
@@ -353,7 +353,7 @@
 
 (test-re #/a+@a+\.a+/ "aaaaaaaaaaa@aaaaaaaa@@a.aaaaa@aaaaa" '())
 (test-re #/a+@a+\.a+/ "aaaaaaaaaaa@aaaaaaaa@@a.aaaaa@aaaaa.a" '("aaaaa@aaaaa.a"))
- 
+
 ;; Repetition of empty string
 (test-re #/a(?:)*b/ "zabr" '("ab"))
 (test-re #/a(?:)?b/ "zabr" '("ab"))
@@ -434,13 +434,10 @@
 (test-re #/[\s\d]+/   "a 1 2 3 b " '(" 1 2 3 "))
 (test-re #/[\s\D]+/   "a 1 2 3 b " '("a "))
 
-(cond-expand
- [gauche.ces.none #f]
- [else
-  (test-re #/\u3000*x/ "\u3000\u3000\u3000x" '("\u3000\u3000\u3000x"))
-  (test-re #/\u3000+/  "\u3001\u3000\u3000" '("\u3000\u3000"))
-  (test-re #/\u3000+x/ "\u3000\u3000\u3001x" '())
-  (test-re #/\u3000+[xy]/ "\u3000\u3000\u3001x\u3001\u3000y" '("\u3000y"))])
+(test-re #/\u3000*x/ "\u3000\u3000\u3000x" '("\u3000\u3000\u3000x"))
+(test-re #/\u3000+/  "\u3001\u3000\u3000" '("\u3000\u3000"))
+(test-re #/\u3000+x/ "\u3000\u3000\u3001x" '())
+(test-re #/\u3000+[xy]/ "\u3000\u3000\u3001x\u3001\u3000y" '("\u3000y"))
 
 ;; this tests optimizer
 (test-re #/^\[?([^\]]*)\]?:(\d+)$/ "127.0.0.1:80"
@@ -809,7 +806,7 @@
                            "**\\k<match>**"))
 
 (test* "regexp-replace" "abfedhi"
-       (regexp-replace #/c(.*)g/ "abcdefghi" 
+       (regexp-replace #/c(.*)g/ "abcdefghi"
                        (lambda (m)
                          (list->string
                           (reverse
@@ -840,6 +837,16 @@
                             #/aba/ "abc"
                             #/bc/  "zz"))
 
+;; Substitution procedure's current-output-port.
+;; https://github.com/shirok/Gauche/issues/967
+(test* "regexp-replace and current-output-port"
+       '("ac" "yo")
+       (let* ((r #f)
+              (out (with-output-to-string
+                     (^[]
+                      (set! r (regexp-replace #/b/ "ab" (^_ (display "yo") "c")))))))
+         (list r out)))
+
 ;;-------------------------------------------------------------------------
 (test-section "regexp cimatch")
 
@@ -1102,22 +1109,22 @@
           (seq "ab" "cd"))
 
 (test-ast (1 #f
-	     (2 #f (seq #\a #\b #\c))
-	     (3 #f (seq #\d #\e #\f))
-	     (seq #\g #\h #\i))
+             (2 #f (seq #\a #\b #\c))
+             (3 #f (seq #\d #\e #\f))
+             (seq #\g #\h #\i))
           ($ ($ "abc") ($ "def") "ghi"))
 
 (test-ast (1 result (seq #\a #\b #\c)) (-> result "abc"))
 
 (test-ast (seq (seq (seq (seq #\a #\b #\c))
-		    (seq (seq #\d #\e #\f))
-		    (seq #\g #\h #\i)))
+                    (seq (seq #\d #\e #\f))
+                    (seq #\g #\h #\i)))
           (w/nocapture ($ ($ "abc") ($ "def") "ghi")))
 
 (test-ast (seq (seq #\a #\b)) (w/case "ab"))
 (test-cset `(seq (seq ,(string->char-set "Aa")
-		      ,(string->char-set "Bb")))
-	   (w/nocase "ab"))
+                      ,(string->char-set "Bb")))
+           (w/nocase "ab"))
 (test-ast (seq (seq (seq #\a #\b))) (w/nocase (w/case "ab")))
 (test-ast (seq (seq #[Aa] #[Bb]) (seq (seq #\c #\d)) (seq #[Ee] #[Ff]))
           (w/nocase "ab" (w/case "cd") "ef"))
@@ -1283,7 +1290,7 @@
             (neg-look-behind #\a ("bc"))))
 
 ;;-------------------------------------------------------------------------
-(test-section "SRE")		    ; based on chibi's regexp-test.sld
+(test-section "SRE")                ; based on chibi's regexp-test.sld
 
 (define (test-sre expected sre input)
   (test* sre
@@ -1434,8 +1441,6 @@
           '(: (* numeric) (w/nocase ($ (* (/ "af")))))
           "12345BeeF")
 
-(cond-expand
- [gauche.ces.utf8 (include "include/regexp-sre-utf8")]
- [else])
+(include "include/regexp-sre-utf8")
 
 (test-end)
