@@ -2,10 +2,11 @@
 
 (use gauche.test)
 (use gauche.process)
-(use srfi-13)
+(use srfi.13)
 (use file.util)
 (use file.filter)
 (use gauche.config)
+(use gauche.threads)
 
 (test-start "utility scripts")
 
@@ -82,7 +83,7 @@
              (process-output->string '("./gosh" "-ftest" "test.o")))
          (delete-files "test.o")))
 
-(test* "r7rs + srfi-22 by interpreter name" "baz bar"
+(test* "r7rs + SRFI-22 by interpreter name" "baz bar"
        (unwind-protect
            (begin
              (delete-files "test.o" "scheme-r7rs")
@@ -123,8 +124,9 @@
 (test-section "gauche-config")
 
 (define (run-gauche-config . opts)
-  (let1 p (run-process `("./gauche-config" ,@opts) :output :pipe)
-    (begin0 (read-line (process-output p))
+  (let1 p (run-process `("./gauche-config" ,@opts)
+                       :redirects '((>& 2 1) (> 1 out)))
+    (begin0 (port->string (process-output p 'out))
       (process-wait p))))
 
 (define *config-options*
@@ -137,13 +139,133 @@
     "--so-libs" "--dylib-suffix" "--dylib-ldflags" "--rpath-flag"))
 
 (dolist [opt *config-options*]
-  (test* #"gauche-config ~opt" (gauche-config opt) (run-gauche-config opt)))
+  (test* #"gauche-config ~opt" (gauche-config opt)
+         (car (string-split (run-gauche-config opt) #[\r\n]))))
+
+(test*/diff
+ "gauche-config --help"
+ '("Usage: gauche-config [option ...]"
+   "Multiple options can be given."
+   ""
+   "General parameter"
+   "  -V"
+   "        The current Gauche version."
+   ""
+   "Parameters to compile applications using Gauche"
+   "  -I"
+   "        Include path options required to compile programs using Gauche"
+   "        (Note: This doesn't work if Gauche installation directory"
+   "        path contains whitespaces.  See --incdirs below.)"
+   "  -L"
+   "        Library path options required to link programs using Gauche"
+   "        (Note: This doesn't work if Gauche installation directory"
+   "        path contains whitespaces.  See --archdirs below.)"
+   "  -l"
+   "        Link library options required to link programs using Gauche."
+   "  --cc"
+   "        The name of the C compiler used to compile this Gauche."
+   "  --cpp"
+   "        The command to run the C preprocessor."
+   "  --ac"
+   "        The directory that contains Gauche-specific autoconf macros."
+   "  --reconfigure"
+   "        The command line used to configure the current installation."
+   "  --arch"
+   "        The autoconf-style architecture signature (cpu-vendor-kernel-os)."
+   "  --incdirs"
+   "  --archdirs"
+   "        The list of directory names to be looked for include files and"
+   "        libraries, respectively.  Each directory name may be quoted if"
+   "        it contains whitespaces, and separated by ':' on Unix platforms, or"
+   "        by ';' on Windows platforms."
+   "  --local-incdir"
+   "  --local-libdir"
+   "        These are '-I' and '-L' flags for additional local headers/libraries"
+   "        to search, given by '--with-local' configure flags.  Note that those"
+   "        are also included in '-I', '-L', '--incdirs', and '--archdirs'."
+   ""
+   "Parameters to install files"
+   "  --prefix"
+   "        The directory prefix set by configure."
+   "  --sysincdir"
+   "  --siteincdir"
+   "  --pkgincdir"
+   "        Directories where system|site|package header files of extensions go."
+   "  --syslibdir"
+   "  --sitelibdir"
+   "  --pkglibdir"
+   "        Directories where system|site|package scheme files go."
+   "  --sysarchdir"
+   "  --sitearchdir"
+   "  --pkgarchdir"
+   "        Directories where system|site|package DSO files go."
+   "  --mandir"
+   "  --infodir"
+   "        Directories where gauche manpage and info docs are installed."
+   ""
+   "Parameters to help building extensions"
+   "  --object-suffix"
+   "        The extension of the compiled objects (e.g. 'o' or 'obj')."
+   "  --executable-suffix"
+   "        The extension of the executable including a period (empty on Unix systems,"
+   "        '.exe' on Windows."
+   "  --so-suffix"
+   "        The extension for dynamically loadable (dlopen-able) modules (e.g. 'so')."
+   "  --so-cflags"
+   "        Additional CFLAGS to create dynamically loadable modules."
+   "  --so-ldflags"
+   "        Additional LDFLAGS to create dynamically loadable modules."
+   "  --so-libs"
+   "        Additional libraries required to create dynamically loadable modules."
+   "  --dylib-suffix"
+   "        The extension for dynamically linked libraries (as opposed to dlopen()ed)"
+   "        Usually the same as --so-suffix, but OSX wants 'dylib'."
+   "  --dylib-ldflags"
+   "        LDFLAGS to create dynamically linked libraries."
+   "  --rpath-flag"
+   "        Compiler flag(s) to embed RPATH"
+   "  --default-cflags"
+   "        This is the CFLAGS used to compile Gauche, and to be used to compile"
+   "        extensions as well."
+   "  --cppflags"
+   "        Compiler flag(s) passed to C preprocessor"
+   "  --static-libs"
+   "        List of library link flags ('-llib') required to link Gauche statically."
+   "        Similar to '-l', but this includes the libraries that are used"
+   "        for extension modules, and also the static library itself,"
+   "        that is -lgauche-static-X.X."
+   "  --libgauche-so"
+   "        The base name of dynamically linked libgauche."
+   "")
+ (string-split (run-gauche-config "--help")
+               (cond-expand
+                [gauche.os.windows "\r\n"]
+                [else "\n"])))
 
 ;;=======================================================================
 (test-section "configure")
 
 (use gauche.configure)
 (test-module 'gauche.configure)
+(test-module 'gauche.configure.base)
+(test-module 'gauche.configure.lang)
+(test-module 'gauche.configure.prog)
+(test-module 'gauche.configure.output)
+(test-module 'gauche.configure.compile)
+(test-module 'gauche.configure.init)
+
+;; cf-help-string test
+(test* "cf-help-string (run-in)"
+       "  -C                      ignored for the compatibility to autoconf-generated\
+      \n                          configure\n"
+       ($ cf-help-string "-C"
+          "ignored for the compatibility to autoconf-generated configure"))
+(test* "cf-help-string (description next line)"
+       "  -C, --config-cache,--cache-file=FILE\
+      \n                          ignored for the compatibility to autoconf-generated\
+      \n                          configure\n"
+       ($ cf-help-string "-C, --config-cache,--cache-file=FILE"
+          "ignored for the compatibility to autoconf-generated configure"))
 
 ;; When we run configure, we need to include directories of gosh and
 ;; other scripts in PATH.  Before installing Gauche, where we find them
@@ -160,6 +282,44 @@
      [else
       (apply run-process `("env" ,#"PATH=~|paths-new|" ,@cmd) args)])))
 
+(define (with-cf-subst-test)
+  (make-directory* "test.o")
+  (with-output-to-file "test.o/configure"
+    (^[]
+      (write '(use gauche.configure))
+      (write '(cf-init "foo" "1.0"))
+      (write '(cf-subst 'A "1"))
+      (write '(cf-subst 'B "2"))
+      (write '(cf-subst 'C "3"))
+      (write '(define (dump)
+                (for-each (^v (cf-msg-notice "~a=~a" v (cf$ v)))
+                          '(A B C D E F))))
+      (write '(with-cf-subst
+               ((A "11") (B + "zz") (C "nn" +) (D "44") (E + "55") (F "66" +))
+               (dump)))
+      (write '(dump))))
+
+  (test* "running `configure' script for with-cf-subst" 0
+         (process-exit-status
+          (run-with-parent-directory-in-paths
+           `("../gosh" "-ftest" "./configure")
+           :output :null :wait #t :directory "test.o")))
+  (test*/diff "check result of with-cf-subst"
+              '("Configuring foo 1.0"
+                "A=11"
+                "B=2 zz"
+                "C=nn 3"
+                "D=44"
+                "E=55"
+                "F=66"
+                "A=1"
+                "B=2"
+                "C=3"
+                "D="
+                "E="
+                "F=")
+              (file->string "test.o/config.log")))
+
 (define (configure-test-1)
   (make-directory* "test.o/src")
   (make-directory* "test2.o")
@@ -169,6 +329,7 @@
                      (dolist [line (port->string-list in)]
                        (display ($ regexp-replace-all* line
                                    #/@@author@@/ ""
+                                   #/@@license@@/ ""
                                    #/@@/ ""
                                    #/\(cf-output-default\)/
                                    "(cf-define 'HAVE_STDIO_H \"1\")\n\
@@ -180,8 +341,6 @@
                    :output outfile))
     (filter-copy (build-path extdir "package-templates" "configure")
                  "test.o/configure")
-    (filter-copy (build-path extdir "package-templates" "configure-compat")
-                 "test.o/configure-compat")
     (filter-copy (build-path extdir "package-templates" "package.scm")
                  "test.o/package.scm")
     (filter-copy (build-path extdir "package-templates" "Makefile.in")
@@ -213,14 +372,14 @@
   (test* "Makefile substitution" '()
          (and (file-exists? "test.o/Makefile")
               (filter #/@\w+@/ (file->string-list "test.o/Makefile"))))
-  (test* "VERSION generation" "1.0\n"
+  (test* "VERSION generation" "0.1\n"
          (and (file-exists? "test.o/VERSION")
               (file->string "test.o/VERSION")))
   (test* "srcdir etc."
          '("srcdir = ." "top_srcdir = ../" "builddir = ." "top_builddir = ../")
          (file->string-list "test.o/src/Makefile"))
 
-  (test* "configure --version" "package configure 1.0"
+  (test* "configure --version" "package configure 0.1"
          (read-line
           (process-output
            (run-with-parent-directory-in-paths
@@ -238,7 +397,7 @@
                (file-exists? "test2.o/src/Makefile")))
 
   (test* "srcdir etc."
-         '("srcdir = ../test.o/src" "top_srcdir = ../../test.o" "builddir = ." "top_builddir = ../")
+         '("srcdir = ../../test.o/src" "top_srcdir = ../../test.o" "builddir = ." "top_builddir = ../")
          (file->string-list "test2.o/src/Makefile"))
 
   (dolist [d '("test.o" "test2.o")]
@@ -278,6 +437,7 @@
                      (dolist [line (port->string-list in)]
                        (display (regexp-replace-all* line
                                                      #/@@author@@/ ""
+                                                     #/@@license@@/ ""
                                                      #/@@/ "")
                                 out)
                        (newline out)))
@@ -293,20 +453,20 @@
         ($ for-each write
            '((use gauche.configure)
              (cf-init-gauche-extension)
-             (cf-check-headers '("stdio.h" "stdlib.h" 
+             (cf-check-headers '("stdio.h" "stdlib.h"
                                  "no-such-header-should-exist.h"))
              (cf-check-types '("a_t" "b_t" "struct c_t" "d_t")
-                             :includes '("typedef int a_t;\n"
-                                         "int b_t;\n"
-                                         "struct c_t { int i; };\n"))
+                             :includes '("typedef int a_t;"
+                                         "int b_t;"
+                                         "struct c_t { int i; };"))
              (cf-check-decls '("a" "b" "c" "d" "e" "f")
-                             :includes '("#define a 1\n"
-                                         "int b;\n"
-                                         "const int c = 1;\n"
-                                         "extern int d();\n"
-                                         "struct foo { int i; } e;\n"))
+                             :includes '("#define a 1"
+                                         "int b;"
+                                         "const int c = 1;"
+                                         "extern int d();"
+                                         "struct foo { int i; } e;"))
              (cf-check-members '("struct foo.a" "struct foo.b")
-                               :includes '("struct foo { int b; };\n"))
+                               :includes '("struct foo { int b; };"))
              (cf-check-funcs '("printf" "nonexistent_weird_function"))
              (unless (#/darwin/ (gauche-architecture))
                ;; this test fails on OSX since 'sin' is recognized as built-in
@@ -355,32 +515,33 @@
            '("LIBS = -lm"))
          (file->string-list "test.o/src/Makefile"))
 
-  (test* "cf-check-headers and cf-check-lib to set defines"
-         `("#define HAVE_STDIO_H 1"
-           "#define HAVE_STDLIB_H 1"
-           "/* #undef HAVE_NO_SUCH_HEADER_SHOILD_EXIST_H */"
-           "#define HAVE_A_T 1"
-           "/* #undef HAVE_B_T */"
-           "#define HAVE_STRUCT_C_T 1"
-           "/* #undef HAVE_D_T */"
-           "#define HAVE_DECL_A 1"
-           "#define HAVE_DECL_B 1"
-           "#define HAVE_DECL_C 1"
-           "#define HAVE_DECL_D 1"
-           "#define HAVE_DECL_E 1"
-           "#define HAVE_DECL_F 0"
-           "/* #undef HAVE_STRUCT_FOO_A */"
-           "#define HAVE_STRUCT_FOO_B 1"
-           "#define HAVE_PRINTF 1"
-           "/* #undef HAVE_NONEXISTENT_WEIRD_FUNCTION */"
-           ,(if (#/darwin/ (gauche-architecture))
-              "/* #undef HAVE_LIBM */"
-              "#define HAVE_LIBM 1")
-           "/* #undef HAVE_LIBNO_SUCH_LIBRARY_SHOULD_EXIST */")
-         ($ filter #/HAVE_/
-            $ file->string-list "test.o/config.h"))
+  (test*/diff "cf-check-headers and cf-check-lib to set defines"
+              `("#define HAVE_STDIO_H 1"
+                "#define HAVE_STDLIB_H 1"
+                "/* #undef HAVE_NO_SUCH_HEADER_SHOILD_EXIST_H */"
+                "#define HAVE_A_T 1"
+                "/* #undef HAVE_B_T */"
+                "#define HAVE_STRUCT_C_T 1"
+                "/* #undef HAVE_D_T */"
+                "#define HAVE_DECL_A 1"
+                "#define HAVE_DECL_B 1"
+                "#define HAVE_DECL_C 1"
+                "#define HAVE_DECL_D 1"
+                "#define HAVE_DECL_E 1"
+                "#define HAVE_DECL_F 0"
+                "/* #undef HAVE_STRUCT_FOO_A */"
+                "#define HAVE_STRUCT_FOO_B 1"
+                "#define HAVE_PRINTF 1"
+                "/* #undef HAVE_NONEXISTENT_WEIRD_FUNCTION */"
+                ,(if (#/darwin/ (gauche-architecture))
+                   "/* #undef HAVE_LIBM */"
+                   "#define HAVE_LIBM 1")
+                "/* #undef HAVE_LIBNO_SUCH_LIBRARY_SHOULD_EXIST */")
+              ($ filter #/HAVE_/
+                 $ file->string-list "test.o/config.h"))
   )
 
+(wrap-with-test-directory with-cf-subst-test '("test.o") #f)
 (wrap-with-test-directory configure-test-1 '("test.o" "test2.o") #f)
 (wrap-with-test-directory configure-test-2 '("test.o" "test2.o") #f)
 
@@ -493,8 +654,8 @@
   (define generate-command
     (if in-place?
       `(../gosh -ftest
-                ,(build-path *top-srcdir* "src" "gauche-package.in") 
-                generate 
+                ,(build-path *top-srcdir* "src" "gauche-package.in")
+                generate
                 ,@(cond-list [scheme-only? "--scheme-only"])
                 --template-dir ,(build-path *top-srcdir* "ext"
                                             "package-templates")
@@ -506,8 +667,8 @@
     (if in-place?
       `(../../gosh -q -I../../../src -I../../../lib
                    ../run compile
-                   --verbose test test.c testlib.stub)
-      `(,gauche-package compile --verbose test test.c testlib.stub)))    
+                   --verbose test test.c testlib.scm)
+      `(,gauche-package compile --verbose test test.c testlib.scm)))
   (define test-command
     (if in-place?
       `(../../gosh -ftest -I. ./test.scm)
@@ -535,7 +696,7 @@
              (load #"~|top-srcdir|/src/gauche-package.in")))))
 
   (test-log "Running ~a" generate-command)
-  
+
   ($ run-process generate-command :wait #t :directory "test.o/")
 
   (for-each file-check
@@ -543,7 +704,7 @@
               '("configure" "Makefile.in" "package.scm"
                 "test.scm" "test/tester.scm")
               '("configure" "Makefile.in" "package.scm"
-                "test.c" "test.h" "test.scm" "testlib.stub"
+                "test.c" "test.h" "test.scm" "testlib.scm"
                 "test/tester.scm")))
 
   (test-log "Running ~a" compile-command)
@@ -552,7 +713,7 @@
     (test* "gauche-package compile" #t
            (let* ([p ($ run-process
                         compile-command
-                        :redirects '((>& 2 1) (> 1 out)) 
+                        :redirects '((>& 2 1) (> 1 out))
                         :directory "test.o/Test")]
                   [o (port->string (process-output p 'out))])
              (process-wait p)
@@ -562,7 +723,7 @@
   (test* "running test" #t
          (let* ([p ($ run-process
                       test-command
-                      :redirects  '((>& 2 1) (> 1 out)) 
+                      :redirects  '((>& 2 1) (> 1 out))
                       :directory "test.o/Test")]
                 [o (port->string (process-output p 'out))])
            (process-wait p)
@@ -575,14 +736,60 @@
 ;;=======================================================================
 (test-section "precomp")
 
+(define (do-precomp! files extra-options)
+  (do-process! `("../../src/gosh" "-ftest"
+                 ,#"-I~|*top-srcdir*|/test/test-precomp"
+                 ,(build-path *top-srcdir* "lib/tools/precomp")
+                 "--strip-prefix"
+                 ,(fix-path (build-path *top-srcdir* "test/test-precomp"))
+                 ,@extra-options
+                 ,@(map (^[file] (fix-path (build-path *top-srcdir* "test/test-precomp" file)))
+                        files))
+               :directory "test.o"))
+
+(define (do-compile! output files)
+  (do-process!
+   `("../../src/gosh" "-ftest"
+     ,(build-path *top-srcdir* "src/gauche-package.in")
+     "compile"
+     ,#"--cppflags=-I~(fix-path (build-path *top-srcdir* \"src\")) \
+                   -I~(fix-path (build-path *top-srcdir* \"gc/include\")) \
+                   -I~(fix-path (build-path *top-builddir* \"src\")) \
+                   -I~(fix-path (build-path *top-builddir* \"gc/include\"))"
+     ,#"--ldflags=-L~(fix-path (build-path *top-srcdir* \"src\")) \
+                  -L~(fix-path (build-path *top-builddir* \"src\"))"
+     ,output
+     ,@files)
+   :directory "test.o"))
+
+;; Removing a DSO file while it is opened tend to cause issues, so we invoke
+;; a child gosh to load and test it.
+(define-syntax dynload-and-eval
+  (syntax-rules ()
+    [(_ libname expr)
+     (begin
+       (with-output-to-file "test.o/t.scm"
+         (^[]
+           (write '(add-load-path "."))
+           (write '(use gauche.threads)) ; some tests need this
+           (write '(use file.util))
+           (write `(define *top-srcdir*
+                     ,(if (absolute-path? *top-srcdir*)
+                        *top-srcdir*
+                        (build-path ".." *top-srcdir*))))
+           (write '(load libname))
+           (write '(write expr))
+           (write '(exit 0))))
+       (let* ([p (run-process '("../../src/gosh" "-ftest" "./t.scm")
+                              :output :pipe :directory "test.o")]
+              [result (read (process-output p))])
+         (process-wait p)
+         result))]))
+
 (define (precomp-test-1)
-  (do-process `("../../src/gosh" "-ftest"
-                ,#"-I~|*top-srcdir*|/test/test-precomp"
-                ,(build-path *top-srcdir* "src/precomp") "--strip-prefix"
-                ,(fix-path (build-path *top-srcdir* "test/test-precomp"))
-                ,@(map (^[file] (fix-path (build-path *top-srcdir* "test/test-precomp" file)))
-                       '("foo.scm" "foo/bar1.scm" "foo/bar2.scm" "foo/bar3.scm")))
-              :directory "test.o")
+  (test* "running precomp 1" #t
+         (do-precomp! '("foo.scm" "foo/bar1.scm" "foo/bar2.scm" "foo/bar3.scm")
+                      '()))
   (test* "precomp generated files"
          '("test.o/foo--bar1.c"
            "test.o/foo--bar2.c"
@@ -595,9 +802,8 @@
          (sort (map fix-path (directory-fold "test.o" cons '()))))
   )
 
-
 (add-load-path "test.o")
-(use srfi-42)
+(use srfi.42)
 (use scheme.vector :only (vector-every))
 
 (define (precomp-test-2)
@@ -615,18 +821,13 @@
           [(number? x)
            (and (number? y)
                 (cond [(nan? x) (nan? y)]
-                      [(-zero? x) (-zero? y)]
+                      [(negative-zero? x) (negative-zero? y)]
                       [else (= x y)]))]
           [else (equal? x y)]))
 
-  (do-process `("../../src/gosh" "-ftest"
-                ,#"-I~|*top-srcdir*|/test/test-precomp"
-                ,(build-path *top-srcdir* "src/precomp") "--strip-prefix"
-                ,(fix-path (build-path *top-srcdir* "test/test-precomp"))
-                "--single-interface"
-                ,@(map (^[file] (fix-path (build-path *top-srcdir* "test/test-precomp" file)))
-                       '("foo.scm" "foo/bar1.scm" "foo/bar2.scm" "foo/bar3.scm")))
-              :directory "test.o")
+  (test* "running precomp 2" #t
+         (do-precomp! '("foo.scm" "foo/bar1.scm" "foo/bar2.scm" "foo/bar3.scm")
+                      '("--single-interface")))
   (test* "precomp generated files with --single-interface"
          '("test.o/foo--bar1.c"
            "test.o/foo--bar2.c"
@@ -635,88 +836,157 @@
            "test.o/foo.sci")
          (sort (map fix-path (directory-fold "test.o" cons '()))))
 
-  (test* "consolidated sci file"
-         '(";; generated automatically.  DO NOT EDIT"
-           "#!no-fold-case"
-           "(define-module foo.bar2 (use util.match) (export bar2))"
-           "(select-module foo.bar2)"
-           "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar2\")"
-           "(provide \"foo/bar2\")"
-           "(define-module foo.bar3 (use foo.bar2) (export bar3))"
-           "(select-module foo.bar3)"
-           "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar3\")"
-           "(provide \"foo/bar3\")"
-           "(define-module foo.bar1 (use foo.bar2) (export bar1))"
-           "(select-module foo.bar1)"
-           "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar1\")"
-           "(provide \"foo/bar1\")"
-           "(define-module foo (use foo.bar1) (use foo.bar3) (export foo-master foo-literals foo-begin1 foo-begin2) (export foo-include1 foo-include2))"
-           "(select-module foo)"
-           "(dynamic-load \"foo\" :init-function \"Scm_Init_foo\")")
-         (file->string-list "test.o/foo.sci"))
+  (test*/diff "consolidated sci file"
+              '(";; generated automatically.  DO NOT EDIT"
+                "#!no-fold-case"
+                "(define-module foo.bar2 (use util.match) (export bar2))"
+                "(select-module foo.bar2)"
+                "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar2\")"
+                "(provide \"foo/bar2\")"
+                "(define-module foo.bar3 (use foo.bar2) (export bar3))"
+                "(select-module foo.bar3)"
+                "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar3\")"
+                "(provide \"foo/bar3\")"
+                "(define-module foo.bar1 (use foo.bar2) (export bar1))"
+                "(select-module foo.bar1)"
+                "(dynamic-load \"foo\" :init-function \"Scm_Init_foo__bar1\")"
+                "(provide \"foo/bar1\")"
+                "(define-module foo (use gauche.uvector) (use foo.bar1) (use foo.bar3) (export foo-master foo-literals foo-shared-literals foo-begin1 foo-begin2) (export foo-include1 foo-include2))"
+                "(select-module foo)"
+                "(dynamic-load \"foo\" :init-function \"Scm_Init_foo\")")
+              (file->string-list "test.o/foo.sci"))
 
-  (test* "compile and dynload" 
+  (test* "compile 2" #t
+         (do-compile! "foo"
+                      '("foo.c" "foo--bar1.c" "foo--bar2.c" "foo--bar3.c")))
+  (test* "dynload and literals 1"
          (list (include "test-precomp/literals.scm")
                'begin1
                'begin2
                'include1
                'include2)
-         (begin
-           (do-process! `("../../src/gosh" "-ftest"
-                          ,(build-path *top-srcdir* "src/gauche-package.in")
-                          "compile"
-                          ,#"--cppflags=-I~(fix-path (build-path *top-srcdir* \"src\")) \
-                                        -I~(fix-path (build-path *top-srcdir* \"gc/include\")) \
-                                        -I~(fix-path (build-path *top-builddir* \"src\")) \
-                                        -I~(fix-path (build-path *top-builddir* \"gc/include\"))"
-                          ,#"--ldflags=-L~(fix-path (build-path *top-srcdir* \"src\")) \
-                                       -L~(fix-path (build-path *top-builddir* \"src\"))"
-                          "foo"
-                          "foo.c" "foo--bar1.c" "foo--bar2.c" "foo--bar3.c")
-                        :directory "test.o")
-           (cond-expand
-            [gauche.os.windows
-             ;; On Windows, foo.dll can't be removed if process using it exists,
-             ;; so we run another gosh in subprocess.
-             (with-output-to-file "test.o/t.scm"
-               (^[]
-                 (write '(add-load-path "."))
-                 (write '(load "foo"))
-                 (write '(write
-                          `(,((global-variable-ref 'foo 'foo-literals))
-                            ,((global-variable-ref 'foo 'foo-begin1))
-                            ,((global-variable-ref 'foo 'foo-begin2))
-                            ,((global-variable-ref 'foo 'foo-include1))
-                            ,((global-variable-ref 'foo 'foo-include2)))))
-                 (write '(exit 0))))
-             (let* ([p (run-process '("../../src/gosh" "-ftest" "./t.scm")
-                                    :output :pipe :directory "test.o")]
-                    [result (read (process-output p))])
-               (process-wait p)
-               result)]
-            [else
-             (load "foo" :paths '("./test.o"))
-             (list ((global-variable-ref 'foo 'foo-literals))
-                   ((global-variable-ref 'foo 'foo-begin1))
-                   ((global-variable-ref 'foo 'foo-begin2))
-                   ((global-variable-ref 'foo 'foo-include1))
-                   ((global-variable-ref 'foo 'foo-include2)))]))
+         (dynload-and-eval
+          "foo"
+          (list ((module-binding-ref 'foo 'foo-literals))
+                ((module-binding-ref 'foo 'foo-begin1))
+                ((module-binding-ref 'foo 'foo-begin2))
+                ((module-binding-ref 'foo 'foo-include1))
+                ((module-binding-ref 'foo 'foo-include2))))
          literal=?)
+
+  (test* "literal sharing" '(#t #t #t #t #t #t)
+         (dynload-and-eval
+          "foo"
+          (let1 vs ((module-binding-ref 'foo 'foo-shared-literals))
+            (list
+             ;; See if partial lists are shared
+             (eq? (cdr (assq-ref vs 'list1))
+                  (assq-ref vs 'list2))
+             (eq? (cdr (assq-ref vs 'list2))
+                  (assq-ref vs 'list3))
+             ;; See if vectors are shared
+             (eq? (assq-ref vs 'vec1)
+                  (assq-ref vs 'vec2))
+             (eq? (assq-ref vs 'uvec1)
+                  (assq-ref vs 'uvec2))
+             ;; See if strings are shared
+             (eq? (assq-ref vs 'str1)
+                  (assq-ref vs 'str2))
+             ;; Descriptive types
+             (eq? (assq-ref vs 'dtype1)
+                  (assq-ref vs 'dtype2))
+             ))))
+  )
+
+(define (precomp-test-3)
+  (test* "running precomp 3" #t (do-precomp! '("types-test.scm") '("-e")))
+  (test* "compile 3" #t (do-compile! "types-test" '("types-test.c")))
+
+  (test* "type reconstruction" '(#t #t #f #t)
+         (dynload-and-eval
+          "types-test"
+          (list
+           (eq? (module-binding-ref 'types-test '<A>)
+                (</> (<Tuple> (<?> <int8>) <string> <integer>)
+                     (<List> <integer> 3 10)))
+           (eq? (module-binding-ref 'types-test '<A>)
+                (module-binding-ref 'types-test '<B>))
+           (eq? (module-binding-ref 'types-test '<A>)
+                (module-binding-ref 'types-test '<C>))
+           ;; NB: The following may not be eq?, since (<List> <integer>) is
+           ;; serialized as (<List> <integer> #f #f).  The memoization is
+           ;; based on the argument list of the type constructor, so it
+           ;; doesn't match with (<List> <integer>) at runtime.
+           (equal? (module-binding-ref 'types-test '<C>)
+                   (</> (<Tuple> (<?> <int8>) <string> <integer>)
+                        (<List> <integer>)))
+           )))
+
+  (test* "assertion with reconstructed type" '((3 "ok" 8483958394)
+                                               (#f "ok" -8483958394)
+                                               (1 2 3 4 5))
+         (dynload-and-eval
+          "types-test"
+          (list
+           ((module-binding-ref 'types-test 'foo) '(3 "ok" 8483958394))
+           ((module-binding-ref 'types-test 'foo) '(#f "ok" -8483958394))
+           ((module-binding-ref 'types-test 'foo) '(1 2 3 4 5)))))
+
+  (test* "assertion with reconstructed type"
+         #t
+         (dynload-and-eval
+          "types-test"
+          (guard (e ((<error> e)
+                     (boolean (#/supposed to be of type/ (~ e'message)))))
+            ((module-binding-ref 'types-test 'foo) "ng"))))
+  )
+
+(define (precomp-test-4)
+  (test* "running precomp 4" #t (do-precomp! '("literal-mutex.scm") '("-e")))
+  (test* "compile 4" #t (do-compile! "literal-mutex" '("literal-mutex.c")))
+
+  (test* "run-once" '(1 1 1 1 1)
+         (dynload-and-eval
+          "literal-mutex"
+          (let* ([foo (module-binding-ref 'literal-mutex 'foo)]
+                 [thrs (map (^_ (make-thread foo)) (iota 5))])
+            (for-each thread-start! thrs)
+            (map thread-join! thrs))))
+  )
+
+(define (precomp-test-5)
+  (test* "running precomp 5" #t (do-precomp! '("macros.scm") '("-e")))
+  (test* "compile 5" #t (do-compile! "macros" '("macros.c")))
+
+  (test* "compiled macro"
+         '(("compiled macro (er)" ((apple) (banana) bonk))
+           ("compiled macro (id-macro)" (#f 1 2)))
+         (dynload-and-eval
+          "macros"
+          (begin
+            (load (build-path *top-srcdir* "test" "test-precomp" "macros-user"))
+            `(("compiled macro (er)"
+               ,(map er-aif-test '(apple banana dragonfruit)))
+              ("compiled macro (id-macro)"
+               ,(im-state-test))))))
   )
 
 (wrap-with-test-directory precomp-test-1 '("test.o"))
 (wrap-with-test-directory precomp-test-2 '("test.o"))
+(wrap-with-test-directory precomp-test-3 '("test.o"))
+(wrap-with-test-directory precomp-test-4 '("test.o"))
+(wrap-with-test-directory precomp-test-5 '("test.o"))
 
 ;;=======================================================================
 (test-section "build-standalone")
 
 (test-script
- (build-path (or (sys-getenv "top_srcdir") "..") "src" "build-standalone"))
+ (build-path (or (sys-getenv "top_srcdir") "..") "lib" "tools" "build-standalone"))
 
 (define (run-build-static opts args)
   (do-process `("../../src/gosh" "-ftest"
+                ,(build-path *top-srcdir* "lib/tools/build-standalone")
                 ,#"-I=~|*top-srcdir*|/test/test-static"
-                ,(build-path *top-srcdir* "src/build-standalone")
                 ,#"--header-dir=~|*top-srcdir*|/src"
                 ,#"--header-dir=~|*top-builddir*|/src"
                 ,#"--header-dir=~|*top-srcdir*|/gc/include"
@@ -727,14 +997,14 @@
 
 (define (static-test-1)
   (test* "static link test" #t
-         (run-build-static `("-o" "staticmain" 
+         (run-build-static `("-o" "staticmain"
                              "-I" ,(build-path *top-srcdir* "test/test-static"))
-                           `(,(fix-path (build-path *top-srcdir* 
+                           `(,(fix-path (build-path *top-srcdir*
                                                     "test/test-static/main.scm"))
-                             "foo/bar.scm" 
+                             "foo/bar.scm"
                              "foo/bar-impl.scm"
                              "foo/baz.scm")))
-  (test* "static link executable" "ARGS: (A b CdE)"
+  (test* "static link executable" "ARGS: #(A b CdE)"
          (process-output->string '(test.o/staticmain A b CdE)))
   )
 
@@ -747,7 +1017,7 @@
 
 (define (test-script-test-1)
   (with-output-to-file "test.o/script"
-    (^[] 
+    (^[]
       (write `(use gauche.uvector))
       (write `(print "bar"))
       (write `(define (bar) (foo) (foo)))
