@@ -6,8 +6,8 @@
 
 (use gauche.test)
 (use gauche.config)
-(use srfi-1)
-(use srfi-13)
+(use scheme.list)
+(use srfi.13)
 
 (test-start "system")
 
@@ -23,6 +23,7 @@
 (define (n pathname) (sys-normalize-pathname pathname))
 
 ;; some common operations via command
+;; NB: We do not use sys-* interface yet, for we're going to test them!
 (define (cmd-rmrf dir)
   (cond-expand
    (gauche.os.windows
@@ -193,13 +194,16 @@
        (sys-normalize-pathname "." :absolute #t))
 (test* "normalize" (n (string-append (get-pwd-via-pwd) "/"))
        (sys-normalize-pathname "" :absolute #t))
-(cond-expand
- (gauche.os.windows #t)
- (else
-  (test* "normalize"
-         (n (string-append (slot-ref (sys-getpwuid (sys-getuid)) 'dir) "/abc"))
-         (sys-normalize-pathname "~/abc" :expand #t))))
-
+(test* "normalize"
+       (cond-expand
+        [gauche.os.windows
+         (n (string-append (or (sys-getenv "HOME") ; MSYS
+                               (sys-getenv "USERPROFILE") ; cmd.exe
+                               "")
+                           "/abc"))]
+        [else
+         (n (string-append (slot-ref (sys-getpwuid (sys-getuid)) 'dir) "/abc"))])
+       (sys-normalize-pathname "~/abc" :expand #t))
 (test* "normalize" (n "/a/b/c/d/e")
        (sys-normalize-pathname "/a/b//.///c//d/./e"
                                :canonicalize #t))
@@ -229,6 +233,37 @@
        (sys-normalize-pathname "a/b/c/../../../../../x/y"
                                :canonicalize #t))
 
+;;-------------------------------------------------------------------
+(test-section "glob")
+
+;; NB: Actual globbing on filesystems is tested in file.scm.
+;; Here we test the internal component, glob-component->regexp.
+
+(let ((data '("a.c" ".a.c" "...c" ".c" "a?.c" "a*.c")))
+  (define (t mode pattern expected)
+    (test* (string-append "glob pattern " (x->string mode) " " pattern)
+           expected
+           (filter (glob-component->regexp pattern :mode mode) data)))
+  (define (t-glob  pattern expected) (t :glob pattern expected))
+  (define (t-shell pattern expected) (t :shell pattern expected))
+
+  (t-glob "*.c" '("a.c" "a?.c" "a*.c"))
+  (t-glob "??.c" '("a?.c" "a*.c"))
+  (t-glob ".*.c" '(".a.c" "...c"))
+  (t-glob ".*c"  '(".a.c" "...c" ".c"))
+  (t-glob "*?.c" '("a.c" "a?.c" "a*.c"))
+  (t-glob "*\\?.c" '("a?.c"))
+  (t-glob "*\\*.c" '("a*.c"))
+
+  (t-shell "*.c" '("a.c" ".a.c" "...c" ".c" "a?.c" "a*.c"))
+  (t-shell "??.c" '(".a.c" "...c" "a?.c" "a*.c"))
+  (t-shell ".*.c" '(".a.c" "...c" ))
+  (t-shell ".*c"  '(".a.c" "...c" ".c"))
+  (t-shell "*?.c" '("a.c" ".a.c" "...c" "a?.c" "a*.c"))
+  (t-shell "*\\?.c" '("a?.c"))
+  (t-shell "*\\*.c" '("a*.c"))
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "filesystem")
 
@@ -406,12 +441,57 @@
 (test* "srfi time seconds->time" '(98765432109876 500000000)
        (let1 t (seconds->time 98765432109876.5)
          (list (ref t'second) (ref t'nanosecond))))
-(test* "srfi time setter" (make <time> :second -98765432109876 :nanosecond 4)
+
+;; <time> should be immutable.
+'(test* "srfi time setter" (make <time> :second -98765432109876 :nanosecond 4)
        (let1 t (make <time>)
          (set! (ref t'second) -98765432109876)
          (set! (ref t'nanosecond) 4)
          t))
 
+(let* ([t0 (current-time)]
+       [t1 (make <time> :second (~ t0'second) :nanosecond #e999_999_999)]
+       [t2 (make <time> :second (~ t0'second) :nanosecond 0)])
+  (define (t->list t :optional (delta-s 0) (delta-ns 0))
+    (list (+ (~ t'second) delta-s) (+ (~ t'nanosecond) delta-ns)))
+  (test* "absolute-time (integer second)"
+         (t->list t0 1)
+         (t->list (absolute-time 1 t0)))
+  (test* "absolute-time (negative integer second)"
+         (t->list t0 -1)
+         (t->list (absolute-time -1 t0)))
+  (test* "absolute-time (real second, t0)"
+         (test-one-of
+          (t->list t0 1 #e5e8)
+          (t->list t0 2 #e-5e8))
+         (t->list (absolute-time 1.5 t0)))
+  (test* "absolute-time (real second, t1)"
+         (t->list t1 2 #e-5e8)
+         (t->list (absolute-time 1.5 t1)))
+  (test* "absolute-time (negative real second, t0)"
+         (test-one-of
+          (t->list t0 0 #e-25e7)
+          (t->list t0 -1 #e75e7))
+         (t->list (absolute-time -0.25 t0)))
+  (test* "absolute-time (negative real second, t2)"
+         (t->list t2 -1 #e75e7)
+         (t->list (absolute-time -0.25 t2)))
+  (test* "absolute-time (duration)"
+         (t->list t1 11 -999_999_999)
+         (t->list (absolute-time (make <time>
+                                   :type 'time-duration
+                                   :second 10
+                                   :nanosecond 1)
+                                 t1)))
+  (test* "absolute-time (negative duration)"
+         (t->list t2 -10 1)
+         (t->list (absolute-time (make <time>
+                                   :type 'time-duration
+                                   :second -10
+                                   :nanosecond 1)
+                                 t2)))
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "stat")
 
@@ -555,7 +635,7 @@
          (let ((pid (sys-fork)))
            (if (= pid 0)
              (begin (sys-pause) (sys-exit 0))
-             (begin 
+             (begin
                (sys-kill pid SIGKILL)
                (receive (rpid code) (sys-wait)
                  (and (= rpid pid)
@@ -566,8 +646,8 @@
          (let1 pid (sys-fork)
            (if (= pid 0)
              (begin (nap) (sys-exit 0))
-             (begin 
-               (sys-kill pid SIGSTOP) 
+             (begin
+               (sys-kill pid SIGSTOP)
                (receive (rpid code) (sys-waitpid pid :untraced #t)
                  (and (= rpid pid)
                       (sys-wait-stopped? code)
@@ -639,22 +719,23 @@
                                  (>= c 50)) ; takes about 10s
                              (write (list (sys-getpid) (sys-getpgrp)))]
                             [else (loop (+ c 1))])))))))
-    (define (run-and-read-test msg ppid detached)
-      (test* msg detached
-             (receive (in out) (sys-pipe :buffering :none)
-               (let1 pid (sys-fork-and-exec "./gosh"
-                                            `("./gosh" "-ftest" "./test.out"
-                                              ,(x->string ppid))
-                                            :iomap `((1 . ,out))
-                                            :detached detached)
-                 (close-port out)
-                 (let1 result (read in)
-                   (sys-waitpid pid)
-                   (eqv? (car result) (cadr result)))))))
-    (run-and-read-test "fork, exec and detached process (not detached)" 
-                       (sys-getpid) #f)
-    (run-and-read-test "fork, exec and detached process (detached)"
-                       1 #t))
+    (let ()
+      (define (run-and-read-test msg ppid detached)
+        (test* msg detached
+               (receive (in out) (sys-pipe :buffering :none)
+                 (let1 pid (sys-fork-and-exec "./gosh"
+                                              `("./gosh" "-ftest" "./test.out"
+                                                ,(x->string ppid))
+                                              :iomap `((1 . ,out))
+                                              :detached detached)
+                   (close-port out)
+                   (let1 result (read in)
+                     (sys-waitpid pid)
+                     (eqv? (car result) (cadr result)))))))
+      (run-and-read-test "fork, exec and detached process (not detached)"
+                         (sys-getpid) #f)
+      (run-and-read-test "fork, exec and detached process (detached)"
+                         1 #t)))
   (cmd-rmrf "test.out")
 
   ;; Testing GC in forked process---we don't explicitly spawn a thread here,
@@ -676,7 +757,7 @@
   ;;              (let ((line (read-line in)))
   ;;                (sys-waitpid pid)
   ;;                line)))))
-  
+
   ] ; !gauche.os.windows
  [else])
 
@@ -814,7 +895,7 @@
               (sys-alarm 1)
               (read in)))))
 
-  (when (global-variable-bound? 'gauche 'sys-select)
+  (when (module-binds? 'gauche 'sys-select)
     (test* "sigalrm6 (interrupting syscall - restart)" '(#t 0)
            (let1 r #f
              (with-signal-handlers
@@ -935,10 +1016,10 @@
          (not gauche.os.cygwin))
     (let ()
       (define z (^n (raise 'foo)))
-      
+
       (set-signal-handler! SIGCHLD #t)
       (set-signal-handler! SIGINT z)
-      
+
       (test* "sys-sigwait" SIGHUP
              (receive (in out) (sys-pipe)
                (let1 pid (sys-fork)
@@ -965,8 +1046,7 @@
     ]
    [else]) ; (not (and gauche.sys.sigwait (not gauche.os.cygwin)))
 
-  ] 
+  ]
  [else]) ; gauche.os.windows
 
 (test-end)
-
