@@ -89,7 +89,7 @@
 
   (test* "make-csv-header-parser (regexps)" '#(3 2 6 5)
          (any (make-csv-header-parser header-slots2) data))
-  
+
   (test* "make-csv-record-parser (strings)"
          '(("Land of Lisp" "1958" "551,435,453" "39994")
            ("United States of Formula Translators" "1957" "4,343,225,434"
@@ -111,7 +111,7 @@
                                              '((#/year/i #/^\d+$/)
                                                #/country/i #/popu/i #/gdp/i))
                      data))
-  
+
   (test* "csv-rows->tuples (allow-gap? #f)"
          '(("Land of Lisp" "1958" "551,435,453" "39994")
            ("United States of Formula Translators" "1957" "4,343,225,434"
@@ -133,7 +133,7 @@
 ;;-------------------------------------------------------------------
 (test-section "diff")
 (use text.diff)
-(use srfi-13)
+(use srfi.13)
 (test-module 'text.diff)
 
 (define diff-a "foo
@@ -164,11 +164,70 @@ fuga
        (with-output-to-string
          (lambda () (diff-report diff-a diff-b))))
 
+(test* "diff-report/context"
+       (string-join '("***************"
+                      "*** 1,6 ****"
+                      "  foo"
+                      "  bar"
+                      "- bar"
+                      "  baz"
+                      "! baz"
+                      "  hoge"
+                      "--- 1,6 ----"
+                      "  foo"
+                      "  bar"
+                      "  baz"
+                      "! fuga"
+                      "  hoge"
+                      "+ fuga")
+                    "\n" 'suffix)
+       (with-output-to-string
+         (lambda () (diff-report/context diff-a diff-b))))
+
+(test* "diff-report/context (one-side empty)"
+       (string-join '("***************"
+                      "*** 0 ****"
+                      "--- 1,6 ----"
+                      "+ foo"
+                      "+ bar"
+                      "+ baz"
+                      "+ fuga"
+                      "+ hoge"
+                      "+ fuga")
+                    "\n" 'suffix)
+       (with-output-to-string
+         (lambda () (diff-report/context "" diff-b))))
+
+(test* "diff-report/unified"
+       (string-join '("@@ -1,6 +1,6 @@"
+                      " foo"
+                      " bar"
+                      "-bar"
+                      " baz"
+                      "-baz"
+                      "+fuga"
+                      " hoge"
+                      "+fuga")
+                    "\n" 'suffix)
+       (with-output-to-string
+         (lambda () (diff-report/unified diff-a diff-b))))
+
+(test* "diff-report/context (one-side empty)"
+       (string-join '("@@ -1,6 +0,0 @@"
+                      "-foo"
+                      "-bar"
+                      "-bar"
+                      "-baz"
+                      "-baz"
+                      "-hoge")
+                    "\n" 'suffix)
+       (with-output-to-string
+         (lambda () (diff-report/unified diff-a ""))))
 
 ;;-------------------------------------------------------------------
 (test-section "edn")
 (use text.edn)
-(use srfi-13)
+(use srfi.13)
 (use gauche.uvector)
 (test-module 'text.edn)
 
@@ -180,7 +239,7 @@ fuga
              (namespace/name . " namespace/name")
              (:keyword . ":keyword")
              ("string" . "\"string\"")
-             ("weird \" characters \t\n\r\\" 
+             ("weird \" characters \t\n\r\\"
               . "\"weird \\\" characters \\t\\n\\r\\\\\"")
              (#\a . "\\a")
              (#\newline . "\\newline")
@@ -233,7 +292,7 @@ fuga
            edn-equal?)))
 
 ;; custom object
-(register-edn-object-handler! 'u8vector 
+(register-edn-object-handler! 'u8vector
                               (^[tag vec] (vector->u8vector vec)))
 (define-method edn-write ((x <u8vector>))
   (display "#u8vector")
@@ -247,11 +306,11 @@ fuga
 ;; invalid
 (let1 invalids '("foo/bar/baz" "/foo" "foo/")
   (dolist [d invalids]
-    (test* (format "parse (invalid) ~s" d) 
+    (test* (format "parse (invalid) ~s" d)
            (test-error <edn-parse-error> #/invalid token/)
            (parse-edn-string d))))
 (test* (format "valid symbol name") '(#f #f #f #f #t)
-       (map edn-valid-symbol-name? 
+       (map edn-valid-symbol-name?
             '("nil" "bar/nil" "true" "false" "nile")))
 
 ;;-------------------------------------------------------------------
@@ -259,10 +318,13 @@ fuga
 (use text.external-editor)
 (test-module 'text.external-editor)
 
+;;-------------------------------------------------------------------
+;; text.fill depends on gauche.unicode, and tested in ext/gauche.
+
 ;;-------------------------------------------------------------------
 (test-section "html-lite")
 (use text.html-lite)
-(use srfi-13)
+(use srfi.13)
 (test-module 'text.html-lite)
 
 (test* "html-escape-string"
@@ -296,7 +358,7 @@ fuga
        (map string-trim-both
             (string-split (html-doctype :type :xhtml-1.0) #\newline)))
 
-(use srfi-13)
+(use srfi.13)
 
 (let ()
   ;; NB: avoid using tree->string, for we haven't tested it yet.
@@ -309,7 +371,7 @@ fuga
           ((null? (car x)) (flatten-rec (cdr x) r))
           (else (flatten-rec (cdr x) (cons (tos (car x)) r)))))
   (define (flatten x) (string-concatenate-reverse (flatten-rec x '())))
-  
+
   (test* "html, head, body"
          "<html><head><title>foo</title></head><body>foo</body></html>"
          (flatten (html:html (html:head (html:title "foo"))
@@ -323,6 +385,226 @@ fuga
          (flatten (html:img :src "foo" :alt "bar baz")))
   )
 
+;;-------------------------------------------------------------------
+(test-section "info")
+(use text.info)
+(test-module 'text.info)
+
+;;-------------------------------------------------------------------
+(test-section "multicolumn")
+(use text.multicolumn)
+(test-module 'text.multicolumn)
+
+(let ([data '("alike" "ancient" "attend" "blood" "both"
+              "break" "bury" "but" "children's" "civil" "continuance"
+              "could" "death" "death-mark'd" "dignity" "do" "ears"
+              "end" "fair" "fatal" "fearful" "foes" "forth" "from"
+              "grudge" "hands" "here" "hours" "households" "if"
+              "in" "is" "lay" "life" "loins" "love" "lovers" "makes"
+              "mend" "misadventured" "miss" "mutiny" "new" "nought"
+              "now" "of" "our" "overthrows" "pair" "parents"
+              "passage" "patient" "piteous" "rage" "remove" "scene"
+              "shall" "stage" "star-cross'd" "strife" "strive" "take"
+              "the" "their" "these" "to" "toil" "traffic" "two" "unclean"
+              "verona" "we" "what" "where" "which" "whose" "with" "you")])
+  (define (t expect . args)
+    (test*/diff (write-to-string `(display-multicolumn data ,@args))
+                expect
+                (with-output-to-string
+                  (cut apply display-multicolumn data args))))
+
+  (t '(
+       "alike          fearful        miss           strive"
+       "ancient        foes           mutiny         take"
+       "attend         forth          new            the"
+       "blood          from           nought         their"
+       "both           grudge         now            these"
+       "break          hands          of             to"
+       "bury           here           our            toil"
+       "but            hours          overthrows     traffic"
+       "children's     households     pair           two"
+       "civil          if             parents        unclean"
+       "continuance    in             passage        verona"
+       "could          is             patient        we"
+       "death          lay            piteous        what"
+       "death-mark'd   life           rage           where"
+       "dignity        loins          remove         which"
+       "do             love           scene          whose"
+       "ears           lovers         shall          with"
+       "end            makes          stage          you"
+       "fair           mend           star-cross'd"
+       "fatal          misadventured  strife"))
+
+  (t '(
+       "alike                    here                     piteous"
+       "ancient                  hours                    rage"
+       "attend                   households               remove"
+       "blood                    if                       scene"
+       "both                     in                       shall"
+       "break                    is                       stage"
+       "bury                     lay                      star-cross'd"
+       "but                      life                     strife"
+       "children's               loins                    strive"
+       "civil                    love                     take"
+       "continuance              lovers                   the"
+       "could                    makes                    their"
+       "death                    mend                     these"
+       "death-mark'd             misadventured            to"
+       "dignity                  miss                     toil"
+       "do                       mutiny                   traffic"
+       "ears                     new                      two"
+       "end                      nought                   unclean"
+       "fair                     now                      verona"
+       "fatal                    of                       we"
+       "fearful                  our                      what"
+       "foes                     overthrows               where"
+       "forth                    pair                     which"
+       "from                     parents                  whose"
+       "grudge                   passage                  with"
+       "hands                    patient                  you")
+      :minimum-width 25)
+
+  (t '("alike          ears           lay            pair           these"
+       "ancient        end            life           parents        to"
+       "attend         fair           loins          passage        toil"
+       "blood          fatal          love           patient        traffic"
+       "both           fearful        lovers         piteous        two"
+       "break          foes           makes          rage           unclean"
+       "bury           forth          mend           remove         verona"
+       "but            from           misadventured  scene          we"
+       "children's     grudge         miss           shall          what"
+       "civil          hands          mutiny         stage          where"
+       "continuance    here           new            star-cross'd   which"
+       "could          hours          nought         strife         whose"
+       "death          households     now            strive         with"
+       "death-mark'd   if             of             take           you"
+       "dignity        in             our            the"
+       "do             is             overthrows     their")
+     :max-columns 7)
+
+  (t '("alike          ancient        attend         blood"
+       "both           break          bury           but"
+       "children's     civil          continuance    could"
+       "death          death-mark'd   dignity        do"
+       "ears           end            fair           fatal"
+       "fearful        foes           forth          from"
+       "grudge         hands          here           hours"
+       "households     if             in             is"
+       "lay            life           loins          love"
+       "lovers         makes          mend           misadventured"
+       "miss           mutiny         new            nought"
+       "now            of             our            overthrows"
+       "pair           parents        passage        patient"
+       "piteous        rage           remove         scene"
+       "shall          stage          star-cross'd   strife"
+       "strive         take           the            their"
+       "these          to             toil           traffic"
+       "two            unclean        verona         we"
+       "what           where          which          whose"
+       "with           you")
+     :order 'row)
+
+  (t '(
+       "alike          here           piteous"
+       "ancient        hours          rage"
+       "attend         households     remove"
+       "blood          if             scene"
+       "both           in             shall"
+       "break          is             stage"
+       "bury           lay            star-cross'd"
+       "but            life           strife"
+       "children's     loins          strive"
+       "civil          love           take"
+       "continuance    lovers         the"
+       "could          makes          their"
+       "death          mend           these"
+       "death-mark'd   misadventured  to"
+       "dignity        miss           toil"
+       "do             mutiny         traffic"
+       "ears           new            two"
+       "end            nought         unclean"
+       "fair           now            verona"
+       "fatal          of             we"
+       "fearful        our            what"
+       "foes           overthrows     where"
+       "forth          pair           which"
+       "from           parents        whose"
+       "grudge         passage        with"
+       "hands          patient        you")
+     :width 50)
+
+  (t '(
+       "     alike          here           piteous"
+       "     ancient        hours          rage"
+       "     attend         households     remove"
+       "     blood          if             scene"
+       "     both           in             shall"
+       "     break          is             stage"
+       "     bury           lay            star-cross'd"
+       "     but            life           strife"
+       "     children's     loins          strive"
+       "     civil          love           take"
+       "     continuance    lovers         the"
+       "     could          makes          their"
+       "     death          mend           these"
+       "     death-mark'd   misadventured  to"
+       "     dignity        miss           toil"
+       "     do             mutiny         traffic"
+       "     ears           new            two"
+       "     end            nought         unclean"
+       "     fair           now            verona"
+       "     fatal          of             we"
+       "     fearful        our            what"
+       "     foes           overthrows     where"
+       "     forth          pair           which"
+       "     from           parents        whose"
+       "     grudge         passage        with"
+       "     hands          patient        you")
+     :width 50 :indent 5)
+
+  (t '(
+       "     alike          misadventured"
+       "     ancient        miss"
+       "     attend         mutiny"
+       "     blood          new"
+       "     both           nought"
+       "     break          now"
+       "     bury           of"
+       "     but            our"
+       "     children's     overthrows"
+       "     civil          pair"
+       "     continuance    parents"
+       "     could          passage"
+       "     death          patient"
+       "     death-mark'd   piteous"
+       "     dignity        rage"
+       "     do             remove"
+       "     ears           scene"
+       "     end            shall"
+       "     fair           stage"
+       "     fatal          star-cross'd"
+       "     fearful        strife"
+       "     foes           strive"
+       "     forth          take"
+       "     from           the"
+       "     grudge         their"
+       "     hands          these"
+       "     here           to"
+       "     hours          toil"
+       "     households     traffic"
+       "     if             two"
+       "     in             unclean"
+       "     is             verona"
+       "     lay            we"
+       "     life           what"
+       "     loins          where"
+       "     love           which"
+       "     lovers         whose"
+       "     makes          with"
+       "     mend           you")
+     :width 45 :indent 5)
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "parse")
 (use text.parse)
@@ -434,7 +716,7 @@ fuga
 (test* "skip-while" '(eof eof)
        (test-parseutil skip-while "yxxxd"
                        (^x (and (char? x)
-				(char-alphabetic? x)))))
+                                (char-alphabetic? x)))))
 
 (test* "next-token" '("" #\d)
        (test-parseutil next-token "xxxd" #[ax ] #[d] "next token"))
@@ -449,9 +731,9 @@ fuga
 (test* "next-token" '("aeio" #\tab)
        (test-parseutil next-token "   aeio\tnjj"
                        (^x (and (char? x)
-				(char-whitespace? x)))
+                                (char-whitespace? x)))
                        (^x (or (eof-object? x)
-			       (char-whitespace? x)))
+                               (char-whitespace? x)))
                        "next token"
                        ))
 
@@ -492,6 +774,105 @@ fuga
 
 ;; WRITEME
 
+;;-------------------------------------------------------------------
+(test-section "sh")
+(use text.sh)
+(test-module 'text.sh)
+
+(let ()
+  (define (t input expected flavor)
+    (test* #"shell-escape-string(~flavor) - ~input" expected
+           (shell-escape-string input flavor)))
+
+  (t ""          "\"\""              'windows)
+  (t "abc"       "abc"               'windows)
+  (t "a b c"     "\"a b c\""         'windows)
+  (t "a \"b\" c" "\"a \\\"b\\\" c\"" 'windows)
+  (t "a\\ \\\""  "\"a\\ \\\\\\\"\""  'windows)
+  (t ""          "''"                'posix)
+  (t "abc"       "abc"               'posix)
+  (t "a b c"     "'a b c'"           'posix)
+  (t "$abc"      "'$abc'"            'posix)
+  (t "[abc]"     "'[abc]'"           'posix)
+  (t "\\abc"     "'\\abc'"           'posix)
+  (t ">abc"      "'>abc'"            'posix)
+  (t "<abc"      "'<abc'"            'posix)
+  (t "a\"b\"c"   "'a\"b\"c'"         'posix)
+  (t "a(b)c"     "'a(b)c'"           'posix)
+  (t "a{b}c"     "'a{b}c'"           'posix)
+  (t "a'c"       "'a'\"'\"'c'"       'posix)
+  )
+
+(let ()
+  (define (t input expected flavor)
+    (test* #"shell-tokenize-string(~flavor) - ~input" expected
+           (shell-tokenize-string input flavor)))
+
+  (t ""                    '()                         'posix)
+  (t "   "                 '()                         'posix)
+  (t "abc"                 '("abc")                    'posix)
+  (t "  abc "              '("abc")                    'posix)
+  (t "ab   c d"            '("ab" "c" "d")             'posix)
+  (t "ab\\ c\\ d"          '("ab c d")                 'posix)
+  (t "ab\\'c\\\"d"         '("ab'c\"d")                'posix)
+  (t "'ab c d' ef \"g h\"" '("ab c d" "ef" "g h")      'posix)
+  (t "'ab\\\\cd' \"ab\\\\cd\"" '("ab\\\\cd" "ab\\cd")  'posix)
+  (t "'ab\\xcd' \"ab\\xcd\"" '("ab\\xcd" "ab\\xcd")    'posix)
+  (t "a\"\"b"              '("ab")                     'posix)
+  (t "a''b"                '("ab")                     'posix)
+  (t "a'\"'b"              '("a\"b")                   'posix)
+  (t "a\"'\"b"             '("a'b")                    'posix)
+
+  (t "$abc"                (test-error)                'posix)
+  (t "\"$abc\""            (test-error)                'posix)
+  (t "'$abc'"              '("$abc")                   'posix)
+  (t "abc >z"              (test-error)                'posix)
+  (t "abc \">z\""          '("abc" ">z")               'posix)
+  (t "abc \"`z`\""         (test-error)                'posix)
+  (t "abc '`z`'"           '("abc" "`z`")              'posix)
+
+  (t ""                    '()                         'windows)
+  (t "   "                 '()                         'windows)
+  (t "abc"                 '("abc")                    'windows)
+  (t "  abc "              '("abc")                    'windows)
+  (t "ab   c d"            '("ab" "c" "d")             'windows)
+  (t "\"a b c\" d e"       '("a b c" "d" "e")          'windows)
+  (t "\"ab\\\"c\" \"\\\\\" d" '("ab\"c" "\\" "d")      'windows)
+  (t "a\\\\\\b d\"e f\"g h" '("a\\\\\\b" "de fg" "h")  'windows)
+  (t "a\\\\\\\"b c d"      '("a\\\"b" "c" "d")         'windows)
+  (t "a\\\\\\\\\"b c\" d e" '("a\\\\b c" "d" "e")      'windows)
+  (t "\""                  '("")                       'windows)
+  (t "\"\""                '("")                       'windows)
+  (t "\"\"\""              '("\"")                     'windows)
+  (t "\"\"\"\""            '("\"")                     'windows)
+  (t "\"\"\"\"\""          '("\"")                     'windows)
+  (t "\"\"\"\"\"\""        '("\"\"")                   'windows)
+  (t "\"\"\"\"\"\"\""      '("\"\"")                   'windows)
+  (t "a\"b\"\" c"          '("ab\"" "c")               'windows)
+  (t "a\"b\"\"\" c"        '("ab\" c")                 'windows)
+  (t "a\"b\"\"\"\" c"      '("ab\"" "c")               'windows)
+  )
+
+(let ()
+  (define (t-case input expect)
+    (test* #"shell-case: ~input"
+           expect
+           ($ shell-case input
+              ["foo"  'foo]
+              ["foo?" 'foo?]
+              ["foo*" 'foo*]
+              ["*foo" '*foo]
+              [("bar*" "baz*") 'barz]
+              [else   'else])))
+  (t-case "foo" 'foo)
+  (t-case "foof" 'foo?)
+  (t-case "foogh" 'foo*)
+  (t-case "fufoo" '*foo)
+  (t-case "bazooka" 'barz)
+  (t-case "zbar" 'else)
+  (t-case "zfooz" 'else)
+  )
+
 ;;-------------------------------------------------------------------
 (test-section "sql")
 (use text.sql)
@@ -524,7 +905,7 @@ fuga
                                            - #\. "E" (number "-3")
                                            (number "1.2") (number ".3")
                                            )
-                                           
+
        (sql-tokenize "0 -12 +12 .123 123. 123.45 -.123 -123. -123.45
                       +.123 +123. +123.45 0E0 -1E3 -1.E3 -.1E3
                       -1.2E3 1E-3 1.E-3 .1E-3 -.E-3 1.2.3"))
@@ -563,6 +944,36 @@ fuga
 (use text.pager)
 (test-module 'text.pager)
 
+;;-------------------------------------------------------------------
+(test-section "segmented-match")
+(use text.segmented-match)
+(test-module 'text.segmented-match)
+
+(let ([data '(;; pattern (success-word ...) (fail-word ...)
+              ("a-b-c"
+               ("a-b-c" "a-b-c-d" "a-b-c-d-e" "a-b-c-"
+                "abc-bcd-cde" "abc-bcd-cde-def")
+               ("a-b" "ab-c" "abc-cde" "a-b-"))
+              ("abc"
+               ("abc" "abcd" "abc-abc" "abc-")
+               ("ab" "-abc" "a-bc"))
+              ("a--b"
+               ("a--b" "aaa--bbb")
+               ("a-b")))])
+  (dolist [pattest data]
+    (let ([pattern (car pattest)])
+      (dolist [succ (cadr pattest)]
+        (test* `(segmented-prefix ,pattern ,succ) #t
+               (segmented-prefix? pattern succ)))
+      (dolist [fail (caddr pattest)]
+        (test* `(segmented-prefix ,pattern ,fail) #f
+               (segmented-prefix? pattern fail)))
+      ))
+  )
+
+(test* "segmented match (alternative separator)" #t
+       (segmented-prefix? "/u/b/g" "/usr/bin/gosh" #\/))
+
 ;;-------------------------------------------------------------------
 (test-section "template")
 (use text.template)
@@ -575,15 +986,15 @@ fuga
         (make-template-environment
          :bindings `(abc 3 def 9 jkl ,format pqr 1000))))
 
-(test* "expand-template-file"
-       "<html><head><title>Test</title></head><body><p><a href=\"foo\">bar</a\n>baz</p\n></body></html>\n"
-       (expand-template-file
-        "template.txt"
-        (make-template-environment
-         :imports '(text.tree)
-         :bindings `(title "Test"
-                     body ,(html:p (html:a :href "foo" "bar") "baz")))
-        '("../test/data")))
+(test*/diff "expand-template-file"
+            "<html><head><title>Test</title></head><body><p><a href=\"foo\">bar</a\n>baz</p\n></body></html>\n"
+            (expand-template-file
+             "template.txt"
+             (make-template-environment
+              :imports '(text.tree)
+              :bindings `(title "Test"
+                                body ,(html:p (html:a :href "foo" "bar") "baz")))
+             '("../test/data")))
 
 ;;-------------------------------------------------------------------
 (test-section "tree")
