@@ -144,7 +144,7 @@
   (define (i . args) (dolist (k args) (tree-map-put! tree k k)))
   (define (d . args) (dolist (k args) (tree-map-delete! tree k)))
   (define (c) (%tree-map-check-consistency tree))
-  
+
   ;; Insertion
   ;;  case 0. adding to an empty tree
   ;;    B:0
@@ -419,7 +419,7 @@
                     (^[k v _] (unless (eq? k s) (tree-map-delete! tm k)))
                     #f)
                  (sort (tree-map-keys tm))))
-              '(a b c)))  
+              '(a b c)))
   (test* "deletion during traversal 3" '(() () ())
          (map (^s
                (let1 tm (make-tm)
@@ -431,7 +431,7 @@
                (let1 tm (make-tm)
                  (tree-map-fold-right tm (^[k v _] (tree-map-delete! tm k)) #f)
                  (sort (tree-map-keys tm))))
-              '(a b c)))  
+              '(a b c)))
   )
 
 ;;
@@ -488,7 +488,7 @@
   )
 
 (let1 tree (make-tree-map = <)
-  (test* "tree-map-pop-min! and num-entries" 0 
+  (test* "tree-map-pop-min! and num-entries" 0
          (begin
            (tree-map-put! tree 0 0)
            (tree-map-pop-min! tree)
@@ -567,5 +567,75 @@
          (tree-map-compare-as-sequences tm1 tm7 string-ci-comparator))
   )
 
-(test-end)
+;; key range generator
+(let ((tm (alist->tree-map '((0 . a) (1 . b) (2 . c) (3 . d) (4 . e)
+                             (5 . f))
+                           default-comparator)))
+  ;; avoid depending on gauche.generator
+  (define (->list gen)
+    (let loop ((r '()))
+      (let1 v (gen)
+        (if (eof-object? v)
+          (reverse r)
+          (loop (cons v r))))))
 
+  (test* "key-range generator (no limits)"
+         '((0 . a) (1 . b) (2 . c) (3 . d) (4 . e) (5 . f))
+         (->list (tree-map->generator/key-range tm)))
+  (test* "key-range generator (no limits, reverse)"
+         '((5 . f) (4 . e) (3 . d) (2 . c) (1 . b) (0 . a))
+         (->list (tree-map->generator/key-range tm :descending #t)))
+  (test* "key-range generator (> 1)"
+         '((2 . c) (3 . d) (4 . e) (5 . f))
+         (->list (tree-map->generator/key-range tm :> 1)))
+  (test* "key-range generator (> 1, reverse)"
+         '((5 . f) (4 . e) (3 . d) (2 . c))
+         (->list (tree-map->generator/key-range tm :> 1 :descending #t)))
+  (test* "key-range generator (>= 1)"
+         '((1 . b) (2 . c) (3 . d) (4 . e) (5 . f))
+         (->list (tree-map->generator/key-range tm :>= 1)))
+  (test* "key-range generator (>= 1, reverse)"
+         '((5 . f) (4 . e) (3 . d) (2 . c) (1 . b))
+         (->list (tree-map->generator/key-range tm :>= 1 :descending #t)))
+  (test* "key-range generator (< 4)"
+         '((0 . a) (1 . b) (2 . c) (3 . d))
+         (->list (tree-map->generator/key-range tm :< 4)))
+  (test* "key-range generator (< 4, reverse)"
+         '((3 . d) (2 . c) (1 . b) (0 . a))
+         (->list (tree-map->generator/key-range tm :< 4 :descending #t)))
+  (test* "key-range generator (<= 4)"
+         '((0 . a) (1 . b) (2 . c) (3 . d) (4 . e))
+         (->list (tree-map->generator/key-range tm :<= 4)))
+  (test* "key-range generator (<= 4, reverse)"
+         '((4 . e) (3 . d) (2 . c) (1 . b) (0 . a))
+         (->list (tree-map->generator/key-range tm :<= 4 :descending #t)))
+  (test* "key-range generator (> 1, <= 4)"
+         '((2 . c) (3 . d) (4 . e))
+         (->list (tree-map->generator/key-range tm :<= 4 :> 1)))
+  (test* "key-range generator (> 1, <= 4, reverse)"
+         '((4 . e) (3 . d) (2 . c))
+         (->list (tree-map->generator/key-range tm :<= 4 :> 1 :descending #t)))
+  (test* "key-range generator (>= 1, < 4)"
+         '((1 . b) (2 . c) (3 . d))
+         (->list (tree-map->generator/key-range tm :< 4 :>= 1)))
+  (test* "key-range generator (>= 1, < 4, reverse)"
+         '((3 . d) (2 . c) (1 . b))
+         (->list (tree-map->generator/key-range tm :< 4 :>= 1 :descending #t)))
+  (test* "key-range generator (> 5)"
+         '()
+         (->list (tree-map->generator/key-range tm :> 5)))
+  (test* "key-range generator (> 5, descending)"
+         '()
+         (->list (tree-map->generator/key-range tm :> 5 :descending #t)))
+  (test* "key-range generator (< 0)"
+         '()
+         (->list (tree-map->generator/key-range tm :< 0)))
+  (test* "key-range generator (< 0, descending)"
+         '()
+         (->list (tree-map->generator/key-range tm :< 0 :descending #t)))
+  (test* "key-range generator (> 4, < 4)"
+         '()
+         (->list (tree-map->generator/key-range tm :< 4 :> 4)))
+  )
+
+(test-end)
