@@ -3,7 +3,7 @@
 (use gauche.test)
 
 (test-start "multibyte (utf-8)")
-(use srfi-1)
+(use scheme.list)
 
 ;;-------------------------------------------------------------------
 (test-section "string builtins")
@@ -170,7 +170,7 @@
        [cycle2 (make-cycle #\α 25)]
        [cycle3 (make-cycle #\あ 85)])
   (define (make-str max-size)
-    (let loop ([size 0] [r '()] 
+    (let loop ([size 0] [r '()]
                [srcs (circular-list cycle1 cycle2 cycle3)]
                [counts (circular-list 1 2 3)])
       (let ([next-char (caar srcs)]
@@ -214,7 +214,7 @@
 (test-section "incomplete strings")
 
 (test "string-length" 9 (lambda () (string-length #*"あいう")))
-(test "string-complete->incomplete" #*"あいう" 
+(test "string-complete->incomplete" #*"あいう"
       (lambda () (string-complete->incomplete "あいう")))
 (test "string-complete->incomplete" #*"あいう"
       (lambda () (string-complete->incomplete #*"あいう")))
@@ -287,7 +287,7 @@
 
 ;;-------------------------------------------------------------------
 (test-section "string-library")
-(use srfi-13)
+(use srfi.13)
 
 (test "string-every" #t (lambda () (string-every #\あ "")))
 (test "string-every" #t (lambda () (string-every #\あ "ああああ")))
@@ -331,7 +331,7 @@
 ;;-------------------------------------------------------------------
 (test-section "char set")
 
-(use srfi-14)
+(use scheme.charset)
 
 (let ((data '(#[ぁ-ん他-唖ァ-ン]
               #[aあ唖ア])))
