@@ -5,7 +5,7 @@
 (use gauche.test)
 (test-start "util")
 
-(use srfi-1)
+(use scheme.list)
 
 ;;-----------------------------------------------
 (test-section "util.match")
@@ -349,6 +349,49 @@
             (^n (assoc-ref graph1 n '()))
             eq-comparator)))
 
+;;-----------------------------------------------
+(test-section "util.identifier-syntax")
+
+(define-module test-identifier-syntax
+  (use gauche.test)
+  (use util.identifier-syntax)
+  (test-module 'util.identifier-syntax)
+
+  (define p (cons 4 5))
+  (define-syntax p.car (identifier-syntax (car p)))
+  (test* "identifier-syntax form 1" 4 p.car)
+
+  (test* "identifier-syntax form 1 no set!"
+         (test-error <error> #/id macro can't be used with set!/i)
+         (eval '(set! p.car 99) (current-module)))
+
+  (let ((p (cons 6 7)))
+    (test* "identifier-syntax hygiene" 4 p.car))
+
+  (define-syntax p.cdr (identifier-syntax
+                        [_ (cdr p)]
+                        [(set! _ expr) (set-cdr! p expr)]))
+  (test* "identifier-sytnax form 2" 5 p.cdr)
+  (test* "identidier-syntax set!" '(4 . 99)
+         (begin
+           (set! p.cdr 99)
+           p))
+  (let ((p (cons 6 7)))
+    (set! p.cdr 101)
+    (test* "identifier-syntax hygiene" 101 p.cdr))
+
+  (let ((p (cons 8 9)))
+    (let-syntax ([p.cdr (identifier-syntax
+                         [_ (cdr p)]
+                         [(set! _ expr) (set-cdr! p (+ 1 expr))])])
+      (let ((p (cons 0 1)))
+        (set! p.cdr 9999)
+        (test* "local identifier syntax hygine" '(0 . 1) p)
+        (test* "local identifier syntax hygine" 10000 p.cdr))
+      (test* "local identifier syntax hygine" '(8 . 10000) p)))
+  (test* "local identifier syntax hygine" '(4 . 101) p)
+  )
+
 ;;-----------------------------------------------
 (test-section "util.isomorph")
 (use util.isomorph)
@@ -444,7 +487,7 @@
 (test* "lcs edit-list"
        '(((+ 0 a) (+ 1 b) (+ 2 c) (+ 3 d)))
        (lcs-edit-list '() '(a b c d)))
-      
+
 (test* "lcs edit-list"
        '(((+ 0 b)) ((- 1 b)))
        (lcs-edit-list '(a b) '(b a)))
@@ -457,11 +500,148 @@
        '(((+ 1 b)))
        (lcs-edit-list '(a) '(a b)))
 
+(test* "lcs edit-list/context (no diff)"
+       '()
+       (lcs-edit-list/context '(a b c) '(a b c)))
+
+(test* "lcs edit-list/context (a empty)"
+       '(#((0 0) (0 3 (+ a) (+ b) (+ c))))
+       (lcs-edit-list/context '() '(a b c)))
+
+(test* "lcs edit-list/context (b empty)"
+       '(#((0 3 (- a) (- b) (- c)) (0 0)))
+       (lcs-edit-list/context '(a b c) '()))
+
+(test* "lcs edit-list/context (deletion)"
+       '(#((0 4 (= a) (- b) (= c) (= d))
+           (0 3 (= a) (= c) (= d))))
+       (lcs-edit-list/context '(a b c d) '(a c d)))
+
+(test* "lcs edit-list/context (addition)"
+       '(#((0 4 (= a) (= b) (= c) (= d))
+           (0 5 (= a) (= b) (+ z) (= c) (= d))))
+       (lcs-edit-list/context '(a b c d) '(a b z c d)))
+
+(test* "lcs edit-list/context (change)"
+       '(#((0 3 (= a) (! b) (= c)) (0 3 (= a) (! d) (= c))))
+       (lcs-edit-list/context '(a b c) '(a d c)))
+
+(test* "lcs edit-list/context size"
+       '(#((3 12 (= d) (= e) (= f) (! g) (! h) (! i) (= j) (= k) (= l))
+           (3 12 (= d) (= e) (= f) (! x) (! y) (! z) (= j) (= k) (= l))))
+       (lcs-edit-list/context '(a b c d e f g h i j k l m n)
+                              '(a b c d e f x y z j k l m n)))
+
+(test* "lcs edit-list/context size"
+       '(#((5 10 (= f) (! g) (! h) (! i) (= j))
+           (5 10 (= f) (! x) (! y) (! z) (= j))))
+       (lcs-edit-list/context '(a b c d e f g h i j k l m n)
+                              '(a b c d e f x y z j k l m n)
+                              eq? :context-size 1))
+
+(test* "lcs edit-list/context multiple hunks"
+       '(#((1 4 (= b) (! c) (= d))
+           (1 5 (= b) (! C) (! C) (= d)))
+         #((9 13 (= j) (- k) (- l) (= m))
+           (10 12 (= j) (= m))))
+       (lcs-edit-list/context '(a b c   d e f g h i j k l m n)
+                              '(a b C C d e f g h i j     m n)
+                              eq? :context-size 1))
+
+(test* "lcs edit-list/context hunk merging"
+       '(#((0 11 (= a) (! b) (! c) (= d) (= e) (= f) (= g) (! h) (! i) (= j) (= k))
+           (0 11 (= a) (! B) (! C) (= d) (= e) (= f) (= g) (! H) (! I) (= j) (= k)))
+         #((12 17 (= m) (= n) (! o) (! p) (= q))
+           (12 17 (= m) (= n) (! O) (! P) (= q))))
+       (lcs-edit-list/context '(a b c d e f g h i j k l m n o p q)
+                              '(a B C d e f g H I j k l m n O P q)
+                              eq? :context-size 2))
+
+(test* "lcs edit-list/context boundary condition 1"
+       '(#((0 5 (- a) (- b) (= c) (= d) (= e))
+           (0 3 (= c) (= d) (= e)))
+         #((8 11 (= i) (= j) (= k))
+           (6 11 (= i) (= j) (= k) (+ l) (+ m))))
+       (lcs-edit-list/context '(a b c d e f g h i j k)
+                              '(c d e f g h i j k l m)))
+
+(test* "lcs edit-list/context boundary condition"
+       '(#((0 5 (! a) (! b) (= c) (= d) (= e))
+           (0 4 (! B) (= c) (= d) (= e)))
+         #((8 12 (= i) (= j) (= k) (! L))
+           (7 12 (= i) (= j) (= k) (! l) (! m))))
+       (lcs-edit-list/context '(a b c d e f g h i j k L)
+                              '(B c d e f g h i j k l m)))
+
+(test* "lcs edit-list/unified (no diff)"
+       '()
+       (lcs-edit-list/unified '(a b c) '(a b c)))
+
+(test* "lcs edit-list/unified (a empty)"
+       '(#(0 0 0 3 ((+ a) (+ b) (+ c))))
+       (lcs-edit-list/unified '() '(a b c)))
+
+(test* "lcs edit-list/unified (b empty)"
+       '(#(0 3 0 0 ((- a) (- b) (- c))))
+       (lcs-edit-list/unified '(a b c) '()))
+
+(test* "lcs edit-list/unified (deletion)"
+       '(#(0 4 0 3 ((= a) (- b) (= c) (= d))))
+       (lcs-edit-list/unified '(a b c d) '(a c d)))
+
+(test* "lcs edit-list/unified (addition)"
+       '(#(0 4 0 5 ((= a) (= b) (+ z) (= c) (= d))))
+       (lcs-edit-list/unified '(a b c d) '(a b z c d)))
+
+(test* "lcs edit-list/unified (change)"
+       '(#(0 3 0 3 ((= a) (- b) (+ d) (= c))))
+       (lcs-edit-list/unified '(a b c) '(a d c)))
+
+(test* "lcs edit-list/unified size"
+       '(#(3 9 3 9 ((= d) (= e) (= f) (- g) (- h) (- i)
+                    (+ x) (+ y) (+ z) (= j) (= k) (= l))))
+       (lcs-edit-list/unified '(a b c d e f g h i j k l m n)
+                              '(a b c d e f x y z j k l m n)))
+
+(test* "lcs edit-list/unified size"
+       '(#(5 5 5 5 ((= f) (- g) (- h) (- i)
+                    (+ x) (+ y) (+ z) (= j))))
+       (lcs-edit-list/unified '(a b c d e f g h i j k l m n)
+                              '(a b c d e f x y z j k l m n)
+                              eq? :context-size 1))
+
+(test* "lcs edit-list/unified multiple hunks"
+       '(#(1 3 1 4 ((= b) (- c) (+ C) (+ C) (= d)))
+         #(9 4 10 2 ((= j) (- k) (- l) (= m))))
+       (lcs-edit-list/unified '(a b c   d e f g h i j k l m n)
+                              '(a b C C d e f g h i j     m n)
+                              eq? :context-size 1))
+
+(test* "lcs edit-list/unified hunk merging"
+       '(#(0 11 0 11 ((= a) (- b) (- c) (+ B) (+ C) (= d) (= e) (= f) (= g)
+                      (- h) (- i) (+ H) (+ I) (= j) (= k)))
+         #(12 5 12 5 ((= m) (= n) (- o) (- p) (+ O) (+ P) (= q))))
+       (lcs-edit-list/unified '(a b c d e f g h i j k l m n o p q)
+                              '(a B C d e f g H I j k l m n O P q)
+                              eq? :context-size 2))
+
+(test* "lcs edit-list/unified boundary condition 1"
+       '(#(0 5 0 3 ((- a) (- b) (= c) (= d) (= e)))
+         #(8 3 6 5 ((= i) (= j) (= k) (+ l) (+ m))))
+       (lcs-edit-list/unified '(a b c d e f g h i j k)
+                              '(c d e f g h i j k l m)))
+
+(test* "lcs edit-list/unified boundary condition"
+       '(#(0 5 0 4 ((- a) (- b) (+ B) (= c) (= d) (= e)))
+         #(8 4 7 5 ((= i) (= j) (= k) (- L) (+ l) (+ m))))
+       (lcs-edit-list/unified '(a b c d e f g h i j k L)
+                              '(B c d e f g h i j k l m)))
+
 ;;-----------------------------------------------
 (test-section "util.rbtree")
 (use util.rbtree)
 (use gauche.sequence)
-(use srfi-27)
+(use srfi.27)
 (test-module 'util.rbtree)
 
 (define rbtree-check
@@ -727,7 +907,7 @@
 
   (test* "simple-relation (relation-fold)" 1264
          (relation-fold r + 0 'a 'b))
-           
+
   )
 
 (define-class <object-set-test> ()
@@ -779,7 +959,7 @@
                (setter (relation-column-setter r 'c)))
            (for-each (lambda (row) (setter row (+ (getter row) 300))) r)
            (map getter r)))
-           
+
   (test* "simple-relation (relation-fold)" 1644
          (relation-fold r + 0 'a 'c))
   )
@@ -837,7 +1017,7 @@
            (stream-unfolds (^s (if (stream-null? s)
                                  (values s '() '())
                                  (if (odd? (stream-car s))
-                                   (values (stream-cdr s) 
+                                   (values (stream-cdr s)
                                            (list (stream-car s))
                                            #f)
                                    (values (stream-cdr s)
@@ -1067,6 +1247,72 @@
                    (list->stream '("1" "2"))
                    (list->stream '())))
 
+(test* "stream-intersperse" '(1 0 2 0 3)
+       (stream->list (stream-intersperse (stream 1 2 3) 0)))
+(test* "stream-intersperse" '()
+       (stream->list (stream-intersperse (stream) 0)))
+
+(test* "stream-split" '((1 3) (5 7 9) () (1))
+       (stream->list
+        (stream-map stream->list
+                    (stream-split (stream 1 3 2 5 7 9 6 8 1) even?))))
+
+(test* "stream-last" 3
+       (stream-last (stream 1 2 3)))
+(test* "stream-last" (test-error)
+       (stream-last (stream)))
+
+(test* "stream-last-n" '(3 4 5)
+       (stream->list (stream-last-n (stream 1 2 3 4 5) 3)))
+(test* "stream-last-n" '(1 2)
+       (stream->list (stream-last-n (stream 1 2) 3)))
+(test* "stream-last" '()
+       (stream->list (stream-last-n (stream) 3)))
+
+;;-----------------------------------------------
+(test-section "util.temporal-relation")
+(use util.temporal-relation)
+(test-module 'util.temporal-relation)
+
+(define-module temporal-relation-test
+  (use gauche.test)
+  (use util.temporal-relation :prefix tr:)
+  (use util.match)
+
+  (define (interval a b) (cons a b))
+  (define p tr:pair-interval-protocol)
+  (define data
+    `([,(interval 0 5) ,(interval 6 9) before]
+      [,(interval 0 5) ,(interval 5 9) meets]
+      [,(interval 0 6) ,(interval 5 9) overlaps]
+      [,(interval 0 9) ,(interval 5 9) finished-by]
+      [,(interval 0 9) ,(interval 3 8) contains]
+      [,(interval 3 8) ,(interval 3 9) starts]
+      [,(interval 3 9) ,(interval 3 9) equal]
+      [,(interval 3 9) ,(interval 3 8) started-by]
+      [,(interval 4 8) ,(interval 0 9) during]
+      [,(interval 4 9) ,(interval 0 9) finishes]
+      [,(interval 4 9) ,(interval 0 8) overlapped-by]
+      [,(interval 5 9) ,(interval 0 5) met-by]
+      [,(interval 6 9) ,(interval 0 4) after]))
+  (define point-data
+    `([,(interval 0 5) 9 before]
+      [,(interval 0 8) 8 finished-by]
+      [,(interval 0 8) 7 contains]
+      [,(interval 5 9) 5 started-by]
+      [,(interval 5 9) 0 after]))
+
+  (dolist [d data]
+    (match-let1 [x y rel] d
+      (test* #"relation? ~rel" #t (tr:relation? rel))
+      (test* #"relate ~x ~y" rel (tr:relate p x y))
+      (test* #"relate ~y ~x" rel (tr:inverse (tr:relate p y x)))))
+
+  (dolist [d point-data]
+    (match-let1 [x y rel] d
+      (test* #"relate ~x ~y" rel (tr:relate-point p x y))))
+  )
+
 ;;-----------------------------------------------
 (test-section "util.toposort")
 (use util.toposort)
