@@ -7,7 +7,7 @@
 ;; objects collected, but may fail on certain occasions.
 
 (use gauche.test)
-(use srfi-1)
+(use scheme.list)
 
 (test-start "weak pointers")
 
@@ -83,7 +83,7 @@
 ;        (lambda (expected got)
 ;          (and (lset= equal? (car expected) (car got))
 ;               (lset= equal? (cadr expected) (cadr got)))))
-         
+
 ; (define x (make-weak-hash-table 'equal? 'key 'gone))
 
 ; (test* "make-weak-hash-table (key-weak)" <weak-hash-table>
@@ -122,5 +122,3 @@
 ;              (weak-hash-table-values x)))
 
 (test-end)
-
-
